/*									tab:2
 *
 * am.h - Generic Active Message header file
 *
 * "Copyright (c) 1994 by Lok Tin Liu and The Regents of the University 
 * of California.  All rights reserved."
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 			Lok Tin Liu
 * Version:			4
 * Creation Date:	Fri Jul  8 13:23:24 1994
 * Filename:		am.h
 * History:
 *	LTL	1	Mon Feb 20 09:37:46 1995
 *		Initial release
 */

/* This is the am.h used by the Split-C library. It includes all GAM v1.1
 * functions. Each platform can have a different but compatible version of
 * this file.
 */
#ifndef __AM_H_
#define __AM_H_

typedef unsigned int vnn_t;
typedef	void (*handler_t)();


#define barrier() HPAM_barrier()
#define am_barrier() HPAM_barrier()
#define am_barrier_init() HPAM_barrier_enable()
#define splitc_poll() am_poll()
#define am_wait(flag,value) HPAM_wait(flag,value) 

#define am_store_async(dest,lva,rva,len,lhandler,harg,endfun,earg) (am_store(dest,lva,rva, len, lhandler, harg),(*endfun)(dest,lva,len,earg))
	   
/* Must be called before any other am calls */
extern int am_enable();

/* Returns virtual node number */
extern int am_my_proc();

/* Returns the number of nodes */
extern int am_procs();

/* Must be called for a graceful shutdown */
extern void am_disable();

/* Not in spec. Must be called to clean up exit */
extern void am_exit(int code);

/* Receive messages by polling */
extern void am_poll();

/* Send request message with up to 4 integer arguments */
extern int am_request_4(vnn_t dest, void *fun, ...);

/* Send request message with 2 integer arguments and one double argument */
extern int am_request_fd(vnn_t dest, void *fun, double arg1, int arg2,
												 int arg3);

/* Send reply message with up to 4 integer arguments */
extern int am_reply_4(vnn_t dest, void *fun, ...);

/* Send reply message with 2 integer arguments and one double argument */
extern int am_reply_fd(vnn_t dest, void *fun, double arg1, int arg2, int arg3);

/* am_get in spec is renamed as am_store to be consistent with Split-C.
 * Copy nbytes bytes of data from address lva on calling node to address
 * rva on dest node. Invoke handler with handler_arg on dest node when
 * all data are received on dest node. 
 */
extern int am_store(vnn_t dest, void *lva, void *rva, int nbytes,
						 handler_t request_handler, void *handler_arg);

		
/* Retrive nbytes bytes of data from address rva on dest node to address
 * lva on calling node. Invoke handler with handler_arg on calling node when
 * all data are received on calling node.
 */
extern int am_get(vnn_t dest, void *sva, void *dva, int nbytes,
					 handler_t reply_handler, void *handler_arg);

/* Return maximum buffer size in bytes for put and get */
extern int am_max_size();

/* Macros for variable number of arguments */
#define am_reply_0(dest, fun) \
	am_reply_4(dest, fun) 
#define am_reply_1(dest, fun, one) \
	am_reply_4(dest, fun, one)
#define am_reply_2(dest, fun, one, two) \
	am_reply_4(dest, fun, one, two)
#define am_reply_3(dest, fun, one, two, three) \
	am_reply_4(dest, fun, one, two, three)

#define am_request_0(node, fun) \
	am_request_4(node, fun)
#define am_request_1(node, fun, one) \
	am_request_4(node, fun, one)
#define am_request_2(node, fun, one, two) \
	am_request_4(node, fun, one, two)
#define am_request_3(node, fun, one, two, three) \
	am_request_4(node, fun, one, two, three)

/* Macros for variable number of arguments -- for backward compatibility
	 with Split-C library
 */


#define am_request0(node, fun) \
	am_request_4(node, fun)
#define am_request1(node, fun, one) \
	am_request_4(node, fun, one)
#define am_request2(node, fun, one, two) \
	am_request_4(node, fun, one, two)
#define am_request3(node, fun, one, two, three) \
	am_request_4(node, fun, one, two, three)
#define am_request4(node, fun, one, two, three, four) \
	am_request_4(node, fun, one, two, three, four)

#define am_reply0(dest, fun) \
	am_reply_4(dest, fun) 
#define am_reply1(dest, fun, one) \
	am_reply_4(dest, fun, one)
#define am_reply2(dest, fun, one, two) \
	am_reply_4(dest, fun, one, two)
#define am_reply3(dest, fun, one, two, three) \
	am_reply_4(dest, fun, one, two, three)
#define am_reply4(dest, fun, one, two, three, four) \
	am_reply_4(dest, fun, one, two, three, four)


#endif





