#ifndef __GLIB_H__
#define __GLIB_H__
/******************************************************************************
 *
 * GLUnix group
 * NOW Project
 *
 * glib.h --
 *
 *     Description of the purpose and function of this file.
 *
 * $Id$
 *
 * $Log$
 *
 *
 * Copyright (c) 1994 by The Regents of the University of California.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * In no event shall the University of California be liable to any party for
 * direct, indirect, special, incidental, or consequential damages arising out
 * of the use of this software and its documentation, even if the University of
 * California has been advised of the possibility of such damage.
 * 
 * The University of California specifically disclaims any warranties,
 * including, but not limited to, the implied warranties of merchantability
 * and fitness for a particular purpose.  The software provided hereunder is
 * on an "as is" basis, and the University of California has no obligation to
 * provide maintenance, support, updates, enhancements, or modifications.
 *
 *****************************************************************************/

#include <stdio.h>
#include <netinet/in.h>

#include "types.h"

/* Let everybody know that GLUnix stuff has been included. */
#define __GLUNIX

typedef void *Am_FastAddr;

/*****************************************************************************/

#ifdef __cplusplus
extern "C" {
#endif

/*****************************************************************************/

extern char **environ;

extern int  Glib_Initialize(void);
extern int  Glib_AmIStartup(void);
extern int  Glib_Spawn(int num,   char *prog, char **argv);
extern int  Glib_Spawne(int num,  char *prog, char **argv, char **envp);
extern int  Glib_Spawnf(int num,  int flags, char *prog, char **argv);
extern int  Glib_Spawnef(int num, int flags, char *prog, char **argv,
			 char **envp);

/*
 * For the CoSpawn calls:
 *     If a shadowNpid is specified, we try to run
 *     on the same nodes as that Npid (as much as possible); if the number of
 *     nodes specified is fewer than that process is running on, then
 *     this program won't be running on some of the nodes; if it is
 *     greater, then this program will be started on some nodes
 *     which the shadowNpid is not on.  If, however, a shadowNpid
 *     is specified and the numNodes is 0, then this program will
 *     be run on exactly and only the nodes of the shadowNpid.
 *
 *     If the shadowNpid is 0 or NPID_NOBODY, then the CoSpawn calls
 *     act just like the Spawn calls.
 */
extern int  Glib_CoSpawn(Npid npid, int num, char *prog, char **argv);
extern int  Glib_CoSpawne(Npid npid, int num, char *prog, char **argv,
			  char **envp);
extern int  Glib_CoSpawnf(Npid npid, int num, int flags, char *prog,
			  char **argv);
extern int  Glib_CoSpawnef(Npid npid, int num, int flags, char *prog,
			   char **argv, char **envp);

/* The flags for Glib_Spawnf() :
   These can be logically ORed together */
#define    GLIB_SPAWN_NONE     0x00
#define    GLIB_SPAWN_ONE_EXIT 0x01  /* The whole program should
					exit when any one process exits */
#define    GLIB_SPAWN_NO_COSCH 0x02  /* The program need not be
					coscheduled */
#define    GLIB_SPAWN_TRACE    0x04  /* The child program will be debugged
				        and should call TRACME on startup */

/* Process status definitions */
#define    GLIB_PROC_STATUS_LOADING    0
#define    GLIB_PROC_STATUS_READY      1
#define    GLIB_PROC_STATUS_RUNNING    2
#define    GLIB_PROC_STATUS_WAITING    3
#define    GLIB_PROC_STATUS_STOPPED    4
#define    GLIB_PROC_STATUS_ZOMBIE     5
#define    GLIB_PROC_STATUS_MIGRATING  6

extern Npid            Glib_GetMyNpid(void);
extern VNN             Glib_GetMyVnn(void);
extern int             Glib_GetParallelDegree(void);
extern int             Glib_GetParallelDegreeByNpid(Npid npid);

extern struct in_addr *Glib_GetHostByVnn(VNN vnn);
extern int             Glib_GetTotalNodes(void);
extern int             Glib_GetNumAvailableNodes(void);

/* Given a local startup pid on the local machine, this routine returns
   the Npid. */
extern Npid  Glib_GetNpidByStartupPid(pid_t startupPid);

/* Gets a list of all the local pids for the specified Npid.  The localPids
   array must have room for at least Glib_GetParallelDegreeByNpid(npid)
   entries in it.  localPids[0] will be the local pid of VNN 0, etc.
   Returns 0 on success, -1 on failure. */
extern int Glib_GetLocalPidVector(Npid npid, pid_t *pidVector);

extern int Glib_Signal(Npid npid, VNN vnn, int sig);

/* These functions may want to go somewhere else */
extern Am_FastAddr    *Am_GetFastAddrByVnn(VNN vnn);
extern int             Io_WriteData(int fd, char *data, int len);
extern int             Io_ReadData(int fd, char *data, int len);

#ifdef __cplusplus
}
#endif

/*****************************************************************************/

#endif /* __GLIB_H__ */

