#ifndef __GLIB_STAT_H__
#define __GLIB_STAT_H__
/******************************************************************************
 *
 * GLUnix group
 * NOW Project
 *
 * glib_stat.h --
 *
 *     Description of the purpose and function of this file.
 *
 * $Id$
 *
 * $Log$
 *
 *
 * Copyright (c) 1994 by The Regents of the University of California.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * In no event shall the University of California be liable to any party for
 * direct, indirect, special, incidental, or consequential damages arising out
 * of the use of this software and its documentation, even if the University of
 * California has been advised of the possibility of such damage.
 * 
 * The University of California specifically disclaims any warranties,
 * including, but not limited to, the implied warranties of merchantability
 * and fitness for a particular purpose.  The software provided hereunder is
 * on an "as is" basis, and the University of California has no obligation to
 * provide maintenance, support, updates, enhancements, or modifications.
 *
 *****************************************************************************/

#include "glib.h"
#include "types.h"

#ifdef __cplusplus
extern "C" {
#endif

/*****************************************************************************/

typedef struct Glib_ProcessInfo {
    Npid         npid;
    uint         startupNode; /* The node that the startup is on. */
    pid_t        startupPid;  /* The local pid of the startup */
    int          numNodes;    /* The degree of parallelism of the program. */
    uid_t        uid;
    gid_t        gid;
    char        *cwd;
    char        *path;
    int          argc;
    char       **argv;
    int          envc;
    char       **envp;
    uint        *nodeVector; /* An array of ip addrs indexed by VNN */
    pid_t       *pidVector;  /* An array of local pids indexed by VNN */
    int          flags;
    int          status;
} Glib_ProcessInfo;

/* The flags values are defined in glib.h */

/* The status values are defined in glib.h */

/*****************************************************************************/
/* Given an Npid, this routine returns a structure containing beaucoup
   information about that GLUnix process.  The structure returned must
   be freed by the user explicitly, using the Glib_FreeProcStruct command.
   Do not free the structure manually since the library caches the
   structure for a little while. */
extern int Glib_GetNumNpids(void);
extern int Glib_GetAllNpids(Npid **npids);
extern struct Glib_ProcessInfo *Glib_GetProcessInfo(Npid npid);
extern void Glib_FreeProcessInfo(struct Glib_ProcessInfo *ps);

typedef struct Glib_NodeInfo {
    int  nodeNum;
    uint ipAddr;
    int  freeMem;
    int  cpu;
    int  xActivity;
} Glib_NodeInfo;

extern Glib_NodeInfo *Glib_GetNodeInfo(int nodeNum, Glib_NodeInfo *nodeInfo);

#ifdef __cplusplus
}
#endif

/*****************************************************************************/

#endif /* __GLIB_STAT_H__ */
