/*									Tab:8
 *
 * glunixstub.c - simulate glunix library
 *
 * 
 * "Copyright (c) 1994 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 		Richard P. Martin
 * Version:		1
 * Creation Date:	Sat Apr 30 20:31:16 1994
 * Filename:		glunixstub.c
 * History:
 */

#include <stdio.h>
#include <sys/shm.h>
#include "med.h"
#include "hpam_int.h"
#include "glib.h"

#define GLU_STUB_KEY 8563

#define PERMS 0666

static char *shmem_segment;
static int shmem_id;

extern int hpam_pid;
extern buf_desc *hpam_req_state; 
extern buf_desc *hpam_rep_state;

static char config_filename[256];
static FILE *config_file;
static int num_procs;
static char str_mac_addrs[MAXNODES][15];

int Glib_GetNpidByPid(int pid) {
  char *config_path;
  int i;

  /* if user does not specify full path of configuration file */
  if ( ((config_path = getenv("HPAM_CONFIG")) == (char *)NULL) ) {
    /* use HPAM_CONFIG by default */
    strcpy(config_filename, "./hpam.conf");
  }
  else
    strcpy(config_filename, config_path);

  config_file = fopen(config_filename, "r");
  if (config_file == (FILE *)0) {			
    printf("Cannot open configuration file %s\n", config_filename);
    exit(-1);
  }
	
  if (fscanf(config_file, "%d", &num_procs) == EOF) {
    printf("Cannot read number of nodes in configuration file\n");
    exit(-1);
  }

  for (i=0; i<num_procs; i++) {
     if (fscanf(config_file, "%s", str_mac_addrs[i]) == EOF) {
      fprintf(stderr,"Not enough nodes in configuration file\n");
      exit(-1);
    }
   }
  close(config_file);
  return getuid();
}

void *getshmbyNPID(int pid) {

  if ((shmem_id = shmget(GLU_STUB_KEY, 2*sizeof(struct t_NI_save_area),
				       (PERMS|IPC_CREAT)))<0){
    perror("getshmbyNPID: can't open shared segment");
    exit(1);
  }

  shmem_segment = (void *) shmat(shmem_id,(char *) 0 , 0);

  if (pid == 0)
    return (void *) shmem_segment;
  else 
    return (void *) (shmem_segment+sizeof(struct t_NI_save_area));
}

int getmypartition() {
  return 2;
}


/* get physical node number by virtual node number */
Am_FastAddr *Am_GetFastAddrByVnn(int node) {

  static  fddi_mac_addr_t addr;
  fddi_mac_addr_t *fddi_addr;

  fddi_addr= &addr;

  if ( (node<0) || (node>=num_procs) ) {
    fprintf(stderr,"Am_GetFastAddrByVnn: passed invalid node number\n",
	    node);
    return NULL;
  }
  if (sscanf (str_mac_addrs[node], "0x%04hx%04hx%04hx", 
	      &(fddi_addr->c), &(fddi_addr->b), &(fddi_addr->a))
      <= 0) {
    fprintf (stderr,"%s: () Am_GetFastAddrByVnn bad mac address node %d\n",
	     str_mac_addrs[node],node);
    exit(1);
  }

  return (Am_FastAddr *) fddi_addr;
}

int Glib_GetMyVnn() {
  int i;
  fddi_mac_addr_t my_addr;
  fddi_mac_addr_t  *remote_addr;

  Am_GetMyFastAddr( (void *) &my_addr);

  for(i=0;i<num_procs;i++) {
    remote_addr =  (fddi_mac_addr_t *) Am_GetFastAddrByVnn(i);
    if (my_addr.a == remote_addr->a)
      return i;
  }

  fprintf(stderr,"Glib_GetMyVnn(): unable to resolve Virtual Node Number, exiting\n");
  exit(1);
}

int Glib_GetPartitionSize() {
  return num_procs;
}

#ifndef HPAM_USE_GLUNIX
int GLUnix_init()
{
  return 1;
}

void pexecv(int nodes, char *prog_name, char *argv[])
{
  return;
}
#endif
