/*                                                                      
 * HPAM - HP precision architecture/Medusa Active Message layer, V3.0
 *        HPAM,HPAM_4,am_poll etc
 *
 * Version 3 close to GAM compliant, missing FP functions
 * "Copyright (c) 1994 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author:              Richard P. Martin
 * Version:		19
 * Creation Date:       Mon Oct 25 23:01:45 PDT 1993
 * Filename:            hpam.c
 * History: 
 *	RPM	19	Sat Oct 22 22:51:08 1994
 *		fixed bug in store/get, preamble computed before lock
 *	RPM	18	Fri Oct 21 17:51:48 1994
 *		fixed bug resulting from GAM spec when send to self
 *	RPM	17	Thu Jul 21 01:52:03 1994
 *		added src as first arg invoked in dispatch loop
 *	RPM	16	Mon May 30 13:46:15 1994
 *		re-implemented profiling code
 *	RPM	15	Sun May  8 00:32:54 1994
 *		changed where state is updated to lock buffers
 *	RPM	14	Sat Mar 19 21:46:26 1994
 *		version 2 passes tests. Implements an D*P model.
 *	RPM	13	Sat Mar 12 14:01:42 1994
 *		changed names to hpam
 *	RPM	12	Mon Jan 31 22:16:24 1994
 *		fixed bug in hpam_reply_4, didn't add base to rep_size
 *	RPM	11	Sun Dec  5 02:33:21 1993
 *		added wait_until_zero to support sync in splitc 
 *	RPM	10	Sun Dec  5 02:12:28 1993
 *		fixed bug in HPAM_4/xfer, state was wrong after a stall
 *	RPM	9	Mon Nov 29 01:19:37 1993
 * 		took out strings.h 
 *	RPM	8	Sat Nov 27 15:47:16 1993
 * 		fixed bug in poll, coredumped if target was unaligned 
 *	RPM	7	Fri Nov 26 17:46:19 1993
 * 		fixed bug in poll, target on extracting an xfer
 * 		was incorrect
 *      RPM  	6 	Sat Nov 20 22:28:22 PST 1993
 * 		fixed bug in poll, len on xfer was incorrect
 * 	RPM 	5 	Wed Nov 10 00:42:00 PST 1993
 * 		Fixed bug in reply_xfer
 * 		was not setting acked flag properly
 * 		Added the sweep to reXmit in HPAM_wait.
 * 	RPM 	4 	Mon Nov  1 17:52:29 PST 1993
 * 		reply_xfer working           
 * 	RPM 	3 	Fri Oct 29 14:43:00 PDT 1993
 * 		xfer working, lengthend timeout. 
 * 	RPM 	2 	Thu Oct 28 02:25:22 PDT 1993
 * 		Re-transmit for hpam_4 
 * 	RPM 	1 	Mon Oct 25 23:01:45 PDT 1993
 * 		Initial versions of HPAM_4,poll,and wait running.
 */

/* HPAM interface: 

   some of the GAM functions are built on top of HPAM primitives

 invoke fun(src,arg0,arg1,arg2,arg3) on node dest
 int am_request_4(int dest, void (*fun)(), unsigned long arg0,
                  unsigned long arg1, unsigned long arg2,unsigned long arg3)

 int am_request_fd(vnn_t dest, void *fun, double arg1, int arg2, int arg3);
 
 invoke fun(src,arg0,arg1,arg2,arg3) on sending node in a handler
 int am_reply_4(int dest, void (*fun)(), unsigned long arg0, 
                unsigned long arg1, unsigned long arg2, unsigned long arg3) 

  copy len bytes from sender starting at src_buff to node dest starting
  at dest_buff. Once copy is complete, invoke fun(dest_buff,len,arg).
  NOTE: len must be <= HPAM_MAX_PKT
 * HPAM_xfer(int dest,void (*fun)(),void *arg,void *src_buff,
             void *dest_buff,int len) 
 
  Same as above for replies.	     
 * HPAM_reply_xfer(void (*fun)(),void *arg,void *src_buff,
             void *dest_buff,int len)

  Save as bcast_4 for xfers. Not reliable.
 * HPAM_bcast_xfer(void (*fun)(),void *arg,void *src_buff,
     void *dest_buff,int len)

  Service the network. Must be called in code to drain the net.
 * am_poll() 

  Make sure all outstanding requests to dest have been acknowledged. 
 * HPAM_flush(int dest) 

  Poll until *flag==value
 * HPAM_wait(int *flag, int value)

  Poll until *flag==0
 * HPAM_wait_until_zero(int *flag)

*/

#include <stdio.h>
#include <stdlib.h>

#include "med.h"
#include "hpam_int.h"

extern unsigned long med_card_base;
extern int hpam_pid;               /* network pid                      */
extern int HPAM_self_address;      /* local part of network pid number */
extern int HPAM_partition_size;    /* number of nodes current executing */

extern int hpam_free[MAXNODES];    /* index to a possibly free buffer */
extern int hpam_reply_dest;        /* has the hander replied          */
extern int hpam_reply_buf;         /* buffer to use for reply         */

extern int hpam_check_proc;        /* next processor to check         */
extern int hpam_check_buf;         /* next buffer to check            */

/* start of the buffer descriptor headers */
extern buf_desc (*hpam_req_state)[MAXDEPTH];   /* request area */
extern buf_desc (*hpam_rep_state)[MAXDEPTH];   /* reply area   */

volatile unsigned long *hpam_busy;
volatile unsigned long *hpam_recv_nack;

extern struct t_NI_save_area *hpam_save_area;
extern unsigned long *hpam_in_q;

void am_poll();

#ifdef PROFILE
unsigned long  hpam_4_tick0,hpam_4_tick1, hpam_4_tick2,hpam_4_tick3,
               hpam_4_tick4,hpam_4_tick5,hpam_4_tick6,hpam_4_tick7;
unsigned long  hpam_4_check,hpam_4_lock, hpam_4_store, hpam_4_update, 
               hpam_4_service;

unsigned long hpam_4_searches,hpam_4s,hpam_4_stalls;

unsigned long  hpam_X_tick0,hpam_X_tick1, hpam_X_tick2,hpam_X_tick3,
               hpam_X_tick4,hpam_X_tick5,hpam_X_tick6,hpam_X_tick7;
unsigned long  hpam_X_check,hpam_X_lock, hpam_X_store, hpam_X_insert,
               hpam_X_update, hpam_X_service;
unsigned long hpam_X_searches,hpam_Xs,hpam_X_stalls;

unsigned long  hpam_P_calls, hpam_P_tc, hpam_P_critical, 
               hpam_P_status, hpam_P_crc, hpam_P_protection,
               hpam_P_reqrep, hpam_Ps,
               hpam_P_4xfer,hpam_P_dispatch, hpam_P_ack,hpam_P,
               hpam_P_release,hpam_P_rechecks,hpam_P_retrans;

unsigned long  hpam_P_tick0,hpam_P_tick1, hpam_P_tick2,hpam_P_tick3,
               hpam_P_tick4,hpam_P_tick5,hpam_P_tick6,hpam_P_tick7,
               hpam_P_tick8,hpam_P_tick9, hpam_P_tick10, hpam_P_tick11,
               hpam_P_tick12,hpam_P_tick13;
#endif

static int timecheck( buf_desc *req_state) {
  unsigned long tick,interval,retries;
  static int panic_messages=0;

  GETITMR(tick);

  if (req_state->timestamp >= tick)  /* has the interval timer wrapped */
    interval = tick - req_state->timestamp;
  else
    interval = (MAXITMR-req_state->timestamp) +tick;

  retries = (req_state->state&RETRY_MASK)>>RETRY_SHIFT;
      
  if (retries > MAX_RETRIES) {
    if (panic_messages < 16) {
      fprintf(stderr,"node %d not responding retries %d\n",hpam_check_proc,
	      retries);
      panic_messages++;
    }
    return -1;
  }
  if (interval > ((BASE_RETRIES<<(retries))+RETRIES_X)) {

#ifdef DEBUG
    printf("retransmit %x interval %x > %x node %0x buf %d state %0x\n",
	   retries+1,
	   interval,((BASE_RETRIES<<(retries))+RETRIES_X),
	   (((int)req_state-(int)hpam_req_state)/sizeof(buf_desc))/MAXDEPTH,
	   (((int)req_state-(int)hpam_req_state)/sizeof(buf_desc))%MAXDEPTH,
	   req_state->state);
#endif
    out_word(TX_READY_FIFO,req_state->size);
    req_state->state=((req_state->state&(~RETRY_MASK))|
		      ((retries+1)<<RETRY_SHIFT));
    GETITMR(req_state->timestamp);
  }

  return 0;
}

/* when all buffers to a given node are used, stall */
static int stall(int dest)
{

  register int buf;
  int fail;

  buf =0;
  while ((hpam_req_state[dest][hpam_free[dest]].state&TX_RD) ==0)  { 
    fail = timecheck(&(hpam_req_state[dest][buf])); /* try a bit harder */
    am_poll();
    buf = (buf+1)%MAXDEPTH;
    if (fail == -1)
      return fail;
  }
} /* stall */

/* remove all outstanding messages in the network to a given node*/
int  HPAM_flush(dest) {
  register int i,nr;
  int fail;

  if (dest == HPAM_self_address)  /* broadcast use the self window */
    return;  
  
  do {
    for (i=0,nr=0;i<MAXDEPTH;i++) {
      am_poll();
      if ((hpam_req_state[dest][i].state&TX_RD) ==0) {
	fail = timecheck(&(hpam_req_state[dest][i]));
	
	if (fail == -1) 
	  return fail;
	nr++;
      }
    }
  } while (nr>0);
}

int am_request_4(int dest, void (*fun)(), unsigned long arg0,
		 unsigned long arg1, unsigned long arg2,unsigned long arg3)
{
  register buf_desc *state_buf;        /* buffer descriptor */
  register unsigned long state_word;   /* tmp state word        */
  register unsigned long base;         /* points to head of packet */
  unsigned long id, free;              /* ID, buffer pair */
  unsigned long i;

  if (dest != HPAM_self_address)  {

    if ((hpam_req_state[dest][hpam_free[dest]].state)&TX_RD) {
      state_buf = &(hpam_req_state[dest][hpam_free[dest]]);
      free = hpam_free[dest];

    sendit:    
      base = (unsigned long) state_buf->buf;

    /* lock buffer */
      state_buf->size = base+PKT_4_DATA+16;
      state_buf->state = (state_buf->state&(~TX_RD)) | TX_NR; 
      state_word = state_buf->state;

    /* next ID number */
      id = (state_word+1)&ST_ID_MASK;

      out_word(base+PKT_CTL,ARGS4|REQ|(id<<ID_SHIFT)|free);
      out_word(base+PKT_FUN,fun);
      out_word(base+PKT_4_DATA,arg0);
      out_word(base+PKT_4_DATA+4,arg1);
      out_word(base+PKT_4_DATA+8,arg2);
      out_word(base+PKT_4_DATA+12,arg3);
      out_word(TX_READY_FIFO,base+PKT_4_DATA+16); /* send it */

      state_word = (state_word&(~ST_ID_MASK))|(id&ST_ID_MASK); /* id number */
      state_word = (state_word&(~RETRY_MASK)); /* retries = 0 */
      state_buf->state = state_word;

      hpam_free[dest]= (hpam_free[dest]+1)&(MAXDEPTH-1); /* inc & mod */
      GETITMR(state_buf->timestamp); /* set timestamp */

      am_poll();
      return 0;
    }
    else {			/* We have to find a free buffer */
      for (i=0;i<MAXDEPTH;i++) {
	if ( (hpam_req_state[dest][i].state) & TX_RD) {
	  state_buf = &(hpam_req_state[dest][i]);
	  free=i;     
	  goto sendit;
	}
      }
      if ( stall(dest) != -1) {
	state_buf = &(hpam_req_state[dest][hpam_free[dest]]);
	free = hpam_free[dest];
	goto sendit;
      }
      else 
	return -1;
    }
  }
  else{				/* message is to self */
    hpam_reply_dest=dest;
    (*fun)(HPAM_self_address,arg0, arg1, arg2, arg3);
    hpam_reply_dest=-1;
  }

  am_poll();
  return 0;
} /* am_request_4 */

int am_request_5(int dest, void (*fun)(), unsigned long arg0,
		 unsigned long arg1, unsigned long arg2,unsigned long arg3,
		 unsigned long arg4)
{
  register buf_desc *state_buf;        /* buffer descriptor */
  register unsigned long state_word;   /* tmp state word        */
  register unsigned long base;         /* points to head of packet */
  unsigned long id, free;              /* ID, buffer pair */
  unsigned long i;

  if (dest != HPAM_self_address)  {

    if ((hpam_req_state[dest][hpam_free[dest]].state)&TX_RD) {
      state_buf = &(hpam_req_state[dest][hpam_free[dest]]);
      free = hpam_free[dest];

    sendit:    
      base = (unsigned long) state_buf->buf;

    /* lock buffer */
      state_buf->size = base+PKT_4_DATA+16;
      state_buf->state = (state_buf->state&(~TX_RD)) | TX_NR; 
      state_word = state_buf->state;

    /* next ID number */
      id = (state_word+1)&ST_ID_MASK;

      out_word(base+PKT_CTL,GET|REQ|(id<<ID_SHIFT)|free);
      out_word(base+PKT_FUN,fun);
      out_word(base+PKT_4_DATA,arg0);
      out_word(base+PKT_4_DATA+4,arg1);
      out_word(base+PKT_4_DATA+8,arg2);
      out_word(base+PKT_4_DATA+12,arg3);
      out_word(base+PKT_4_DATA+16,arg4);

      out_word(TX_READY_FIFO,base+PKT_4_DATA+20); /* send it */

      state_word = (state_word&(~ST_ID_MASK))|(id&ST_ID_MASK); /* id number */
      state_word = (state_word&(~RETRY_MASK)); /* retries = 0 */
      state_buf->state = state_word;

      hpam_free[dest]= (hpam_free[dest]+1)&(MAXDEPTH-1); /* inc & mod */
      GETITMR(state_buf->timestamp); /* set timestamp */

      am_poll();
      return 0;
    }
    else {			/* We have to find a free buffer */
      for (i=0;i<MAXDEPTH;i++) {
	if ( (hpam_req_state[dest][i].state) & TX_RD) {
	  state_buf = &(hpam_req_state[dest][i]);
	  free=i;     
	  goto sendit;
	}
      }
      if ( stall(dest) != -1) {
	state_buf = &(hpam_req_state[dest][hpam_free[dest]]);
	free = hpam_free[dest];
	goto sendit;
      }
      else 
	return -1;
    }
  }
  else{				/* message is to self */
    hpam_reply_dest=dest;
    (*fun)(HPAM_self_address,arg0, arg1, arg2, arg3,arg4);
    hpam_reply_dest=-1;
  }

  am_poll();
  return 0;

} /* am_request_ 5*/

int am_reply_4(int dest, void (*fun)(), unsigned long arg0, 
		unsigned long arg1, unsigned long arg2, unsigned long arg3)
{
  register buf_desc *state_buf;
  register unsigned long base; /* points to head of packet */
  unsigned long id;

  if (hpam_reply_dest != -1) {
    if (hpam_reply_dest != HPAM_self_address)  {

      state_buf = &(hpam_rep_state[hpam_reply_dest][hpam_reply_buf]);
      base = (unsigned long) state_buf->buf;
      
      /* lock buffer */

      state_buf->size = base+PKT_4_DATA+16;
      /* sequence number */
      id = (state_buf->state &ST_ID_MASK);

      out_word(base+PKT_CTL,REP|ARGS4|(id<<ID_SHIFT)|hpam_reply_buf);
      out_word(base+PKT_FUN,fun);
      out_word(base+PKT_4_DATA,arg0);
      hpam_reply_dest = -1;
      out_word(base+PKT_4_DATA+4,arg1);
      out_word(base+PKT_4_DATA+8,arg2);
      out_word(base+PKT_4_DATA+12,arg3);
      out_word(TX_READY_FIFO,base+PKT_4_DATA+16); /* send it */
    
      state_buf->state = (id+1)&ST_ID_MASK;
    }
    else {
      (*fun)(HPAM_self_address,arg0, arg1, arg2, arg3);
      return;
    }
  }
  else{
    fprintf(stderr,"HPAM_reply_4() called outside handler\n");
  }
  return 0;
} /* HPAM_reply_4 */

void HPAM_bcast_4(void (*fun)(), unsigned long arg0, unsigned long arg1, 
	    unsigned long arg2, unsigned long arg3)
{

  register unsigned long base; /* points to head of packet */
  unsigned long control; 
  hpam_reply_dest = -1;

  base= (unsigned long) hpam_req_state[HPAM_self_address][0].buf;

  hpam_req_state[HPAM_self_address][0].size=base+PKT_4_DATA+12+sizeof(long);

  /* no sequence number for broadcasts,*/
  out_word(base+PKT_CTL,BCAST|ARGS4); 
  out_word(base+PKT_FUN,fun);
  out_word(base+PKT_4_DATA,arg0);
  out_word(base+PKT_4_DATA+4,arg1);
  out_word(base+PKT_4_DATA+8,arg2);
  out_word(base+PKT_4_DATA+12,arg3);
  out_word(TX_READY_FIFO,base+PKT_4_DATA+12+sizeof(long)); /* send it */

  hpam_req_state[HPAM_self_address][0].size =0;

  (*fun)(HPAM_self_address,arg0,arg1,arg2,arg3);
  am_poll();

} /* HPAM_bcast_4 */

/* basic routine to pull messages from the Medusa Card
 * 1: see if we got something
 * 2: if its not a status packet then....
 * 3: if the crc is good then ...
 * 4: If its a valid pid then....
 * 5: see if its a request, reply or broadcast?
 * 6: does the sequence number match?
 *    Y cool! 
 *    N this sucks... if a req, retransmit reply. If a rep, just drop it.
 * 7: Is it a 4 call? call handler
 * 8: else Xfer copy data, then call handler
 */

void am_poll()
{
  int i;
  unsigned long rx_buffer;           /* pointer from RX_READY     */
  unsigned long recv_start,recv_len; /* start address and length  */
  buf_desc *req_state;               /* pointer to request state buffer   */
  buf_desc *rep_state;               /* pointer to reply state buffer   */
  register unsigned long control;    /* control word in message   */
  unsigned long base;                /* frame to send ack         */
  unsigned long pid_s_d;             /* pid part of message       */
  unsigned long crc;                 /* crc bit                   */           
  int src;                           /* src node                  */
  int id,buf;                        /* id, buffer pair           */
  int len;                           /* data length of xfer       */ 
  unsigned long preamble;            /* preamble (in bytes) for xfer */
  unsigned long source,target;       /* pointers for xfer call    */

#ifdef PROFILE
 hpam_P_calls++;
#endif

 start:

  if (((rx_buffer = in_word(RX_READY_FIFO)) & 0x80000000) != 0) {
    recv_len  = rx_buffer & PKT_LEN_MASK;
    recv_start = rx_buffer & ~(PKT_LEN_MASK);
    if (recv_len > 0x12) {	/* is the message valid? */
      
      crc = in_word((recv_start+(recv_len-2))&~0x3);
      pid_s_d = in_word(recv_start+PKT_DST_SRC);
      control = in_word(recv_start+PKT_CTL);

      /*is the MSB of the second to last byte==0 then we have a good CRC. */
      if ((crc>>(((4-((recv_len-2)&0x3))<<3)-1) &0x1) ==0 ) {  /* good crc */
	
	if ( ((pid_s_d&PKT_PID_MASK)>>PID_SHIFT) == hpam_pid) { /*pid check*/
	  src = (pid_s_d&PKT_SRC_MASK);
	  id = (control&ID_MASK)>>ID_SHIFT;
	  buf = control&BUF_MASK;

	  if ((control&NET_MASK) == REQ)  { /* a request */
	    rep_state = &(hpam_rep_state[src][buf]);
	  
	    if (id ==((rep_state->state)&ST_ID_MASK)) {
	      hpam_reply_dest = src;
	      hpam_reply_buf = buf;

	      if ( (control&ARGS_MASK) == ARGS4) { /* an HPAM_4 call */

		( *( (void (*)()) in_word(recv_start+PKT_FUN)))
		  (src, 
		   in_word(recv_start+PKT_4_DATA),
		   in_word(recv_start+PKT_4_DATA+sizeof(unsigned long)),
		   in_word(recv_start+PKT_4_DATA+2*sizeof(unsigned long)),
		   in_word(recv_start+PKT_4_DATA+3*sizeof(unsigned long)));
	      }
	      else { /*  Xfer (store) or get call  */
		if ( control & GET_MASK ) { 
		  ( *( (void (*)()) in_word(recv_start+PKT_FUN)))
		    (src, 
		     in_word(recv_start+PKT_4_DATA),
		     in_word(recv_start+PKT_4_DATA+sizeof(unsigned long)),
		     in_word(recv_start+PKT_4_DATA+2*sizeof(unsigned long)),
		     in_word(recv_start+PKT_4_DATA+3*sizeof(unsigned long)),
		     in_word(recv_start+PKT_4_DATA+4*sizeof(unsigned long)));
		}
		else { /* an Xfer call */
		  target=in_word(recv_start+PKT_X_BASE);
		  source=recv_start+PKT_X_DATA;
		  preamble = (control&OFFSET_MASK) >> OFFSET_SHIFT;
		  recv_len -= (PKT_X_DATA+FDDI_TRAILER+preamble);
		  
		  len=recv_len;

		  /* med_vcopy crashes with unaligned data and len == 0 */
		  if (recv_len > 0) {
		    /* punt to med_vcopy if the data is unaliagned in the 
		     * pre or postamble, or if there is enough
		     * to justify the overhead */

		    if (preamble||(recv_len&0x3)||(target&0x3)||
			(recv_len>USE_VENOM_SIZE)){
		      med_vcopy((char*) (source+preamble),(char*) target,recv_len);
		    }
		    else {		/* aligned and under the size */
		      while (recv_len) {
			* ((unsigned long*)target)=in_word(source);
			source+=4; target+=4; recv_len-=4;
		      }
		      target-=len;
		    } /* aligned and under size*/
		  } /* end if len > 0 */

		(*( (void (*)()) in_word(recv_start+PKT_FUN)))
		  (src, 
		   (void *) target, 
		   len,
   		   (void *) in_word(recv_start+PKT_X_ARG));
		}/* else xfer */
	      } /* store or get call */
	      /*next id (sequence) number */
	      rep_state->state=(rep_state->state&(~ST_ID_MASK)) |
		((id+1)&ST_ID_MASK);
	  
	      /* if nobody sent and ack, send one */
	      if (hpam_reply_dest != -1) { /* send an empty ack */
		base = (unsigned long) rep_state->buf;
		out_word(base+PKT_CTL,ACK|REP|(id<<ID_SHIFT)|buf);
		out_word(TX_READY_FIFO,base+PKT_CTL+sizeof(long)); /*send it*/
		rep_state->size = base+PKT_CTL+sizeof(long);
		hpam_reply_dest = -1;
	      } /* need to ack */
	    } /* sequence numebrs match */
	    else {		/* wrong sequence number maybe a duplicate */
	      out_word(TX_READY_FIFO,rep_state->size);
#ifdef DEBUG
	      printf("Req wrong seq number, retransmit rep to %0x id %0x buf %0x exp %0x\n",
		     src,id,buf,(rep_state->state)&ST_ID_MASK);
#endif 
	    }
	  } /* end if a request */
	  else {			/* may be a reply, broadcast */ 
	    
	    if ( (control&NET_MASK)==REP )  { /* a reply */
	      req_state =  &(hpam_req_state[src][buf]);	  
	      
	      /* this conditional is tricky. When we get the reply,we may */
	      /* get two of the same id in a row. So have to */
	      /* check the second condition */
	      
	      if (id ==((req_state->state)&ST_ID_MASK) &&
		  (!(req_state->state&TX_RD)) ) {
		
		if(!(control&ACK_MASK)) { /* Not an empty ack */
		  
		  if (control&ARGS_MASK) { /* a HPAM_4 call */
		    ( *( (void (*)()) in_word(recv_start+PKT_FUN)))
		      (src,
		       in_word(recv_start+PKT_4_DATA),
		       in_word(recv_start+PKT_4_DATA+sizeof(unsigned long)),
		       in_word(recv_start+PKT_4_DATA+2*sizeof(unsigned long)),
		       in_word(recv_start+PKT_4_DATA+3*sizeof(unsigned long)));
		  }
		  else {	/* Xfer call  */
		    target=in_word(recv_start+PKT_X_BASE);
		    source=recv_start+PKT_X_DATA;
		    preamble = (control&OFFSET_MASK) >> OFFSET_SHIFT;
		    recv_len -= (PKT_X_DATA+FDDI_TRAILER+preamble);

		    /* punt to med_vcopy if the data is unaligned in the 
		     * head or tail portions, or if the target address is 
		     * unaligned, or if there is enough to justify the overhead 
		     */
		    len=recv_len;		

		    if (recv_len >0 ) {
		      if (preamble||(recv_len&0x3)||(target&0x3)||
			  (recv_len>USE_VENOM_SIZE)){
			med_vcopy((char*) (source+preamble),(char*) target,recv_len);
		      }
		      else {	/* aligned and under the size */

			while (recv_len) {
			  * ((unsigned long*)target)=in_word(source);
			  source+=4; target+=4; recv_len-=4;
			}
			target -=len;
		      }
		    } /* if len > 0 */
		    (*( (void (*)()) in_word(recv_start+PKT_FUN)))
		      (src,
		       (void *) target,
		       len,
		       (void *) in_word(recv_start+PKT_X_ARG));
		  } /* end if xfer */
		} /* not ack */

		/* free buffer */
		req_state->state = (req_state->state|TX_RD);
		hpam_free[src] = buf;

	      } /* sequence # ok */
	    else {		/* wrong seq number */
#ifdef DEBUG
	      printf("Rep from %d wrong seq number 0x%0x buf %0x, exp %0x drop\n",
		     src,id,buf,(req_state->state)&ST_ID_MASK);
#endif
	    }
	  } /* end reply sequence */
	  else {		/* a broadcast packet */
	    if (control&ARGS_MASK) { /* a HPAM_4 call */
	      ( *( (void (*)()) in_word(recv_start+PKT_FUN)))
		(src,
		 in_word(recv_start+PKT_4_DATA),
		 in_word(recv_start+PKT_4_DATA+sizeof(unsigned long)),
		 in_word(recv_start+PKT_4_DATA+2*sizeof(unsigned long)),
		 in_word(recv_start+PKT_4_DATA+3*sizeof(unsigned long)));
	    }
	    else {		/* bcast xfer call  */
	      target=in_word(recv_start+PKT_X_BASE);
	      source=recv_start+PKT_X_DATA;
	      preamble = (control&OFFSET_MASK) >> OFFSET_SHIFT;
	      recv_len -= (PKT_X_DATA+FDDI_TRAILER+preamble);

	      /* punt to med_vcopy if the data is unaliagned in the 
	       * head or tail portions, or if the there is enough
	       * to justify the overhead */
	      len=recv_len;
	      if (recv_len >0 ) {
		if (preamble||(recv_len&0x3)||(target&0x3)||
		    (recv_len>USE_VENOM_SIZE)){
		  med_vcopy((char*) (source+preamble),(char*) target,recv_len);
		}
		else {		/* aligned and under the size */

		  while (recv_len) {
		    *((unsigned long*)target)=in_word(source);
		    source+=4; target+=4; recv_len-=4;
		  }
		  target -=len;
		}
	      } /* if len > 0 */
	      if (in_word(recv_start+PKT_FUN) != NULL) {
		(*( (void (*)()) in_word(recv_start+PKT_FUN)))
		  (src,
		   (void *) target, len,
		   (void * ) in_word(recv_start+PKT_X_ARG));
	      }
	    } /* xfer */	     
	  } /* end bcast sequence */
	}
	} /* end if pid check */
	else {
	fprintf(stderr,"am_poll NPID %0x got packet for NPID %0x\n",hpam_pid,
		 ((pid_s_d&PKT_PID_MASK)>>PID_SHIFT) );
	}
    }  /* end if crc is good */
     else {
       fprintf(stderr,"am_poll(): received packet with a bad CRC\n");
     }
    } /* end if not a status packet */
    out_word (RX_FREE_FIFO, rx_buffer);
    goto start;
  } /* end if anything in the card */
  /* timeout and retransmission check */

  if (hpam_check_buf ==0) {
    hpam_check_proc= (hpam_check_proc+1) % HPAM_partition_size;

    if (hpam_check_proc == HPAM_self_address) 
      hpam_check_proc= (hpam_check_proc+1) % HPAM_partition_size;
  }

  hpam_check_buf = (hpam_check_buf == (MAXDEPTH-1)) ? 0 : hpam_check_buf+1 ;

  if ((hpam_req_state[hpam_check_proc][hpam_check_buf].state&TX_RD)==0) {
    timecheck(&(hpam_req_state[hpam_check_proc][hpam_check_buf]));
  }
}   /* am_poll */


int am_store(int dest,void *src_buff, void *dest_buff, int len, void (*fun)(),
	     void *arg)
{
  buf_desc *state_buf;                /* buffer descriptor */
  register unsigned long state_word;  /* word of buffer descriptor */
  register unsigned long base;        /* points to head of packet */
  unsigned long free,id;
  unsigned long i;                    /* classic loop counter */
  unsigned long wl_start,wl_end;      /* word- aligned pointers to data */
  int preamble;                       /* which byte does the data start */

  if (len > HPAM_MAX_PAYLOAD) {
    fprintf(stderr,"HPAM_xfer() got packet longer than %d\n",HPAM_MAX_PAYLOAD);
    len = HPAM_MAX_PAYLOAD;
  }
  
  if (dest != HPAM_self_address) {

    if ((hpam_req_state[dest][hpam_free[dest]].state)&TX_RD) {
      state_buf = &(hpam_req_state[dest][hpam_free[dest]]);
      free = hpam_free[dest];
    
  sendit:    
      base = (unsigned long) state_buf->buf;

    /* lock buffer */
      preamble = ((int) src_buff) & 0x3;
      state_buf->state = (state_buf->state&(~TX_RD)) | TX_NR; 
      state_buf->size =base+PKT_X_DATA+preamble+len;  

      state_word = state_buf->state;
    /* next ID number */
      id = (state_word+1)&ST_ID_MASK;
      wl_start = ((unsigned long)src_buff&(~0x3));

      out_word(base+PKT_FUN,fun);    
      out_word(base+PKT_CTL,(preamble<<OFFSET_SHIFT)|XFER|REQ|
	       (id<<ID_SHIFT)|free);
      out_word(base+PKT_X_ARG,arg);
      out_word(base+PKT_X_BASE,dest_buff);
  
      i=base+PKT_X_DATA;
      wl_end =wl_start+((len+0x3)&(~0x3));

      if (len >0) {
	while(wl_start<=wl_end) {   /* need <= here because might be a tail */
	  out_word(i,*((unsigned long *)wl_start)); /*if postamble unaligned */
	  wl_start+=4,i+=4;
	}
      }

      out_word(TX_READY_FIFO,base+PKT_X_DATA+preamble+len); /* send it */

      state_word = (state_word&(~ST_ID_MASK))|(id&ST_ID_MASK); /* id number */
      state_word = (state_word&(~RETRY_MASK)); /* retries = 0 */
      state_buf->state = state_word;
      hpam_free[dest]= (hpam_free[dest]+1)&(MAXDEPTH-1); /* inc & mod */
      GETITMR(state_buf->timestamp); /* set timestamp */

    } /* found a free buffer */
    else {			/* We have to find a free buffer */
      for (i=0;i<MAXDEPTH;i++) {
	if ( (hpam_req_state[dest][i].state) & TX_RD) {
	  state_buf = &(hpam_req_state[dest][i]);
	  free=i;     
	  goto sendit;
	}
      }
      if (stall(dest) != -1) {
	state_buf = &(hpam_req_state[dest][hpam_free[dest]]);
	free = hpam_free[dest];
	goto sendit;
      }
      else {
	return -1;
      }
    }
  } /* if dest != self */
  else {
    hpam_reply_dest=dest;
    memcpy(dest_buff,src_buff,len);
    (*fun)(HPAM_self_address,arg, dest_buff,len);    
    hpam_reply_dest=-1;
  }
  
  am_poll();
  return 0;
} /* HPAM_xfer */

static int HPAM_reply_xfer(void (*fun)(),void *arg,void *src_buff,
     void *dest_buff,int len)
{
  buf_desc *state_buf; 
  register unsigned long base;         /* points to head of packet */
  unsigned long control;               /* control word */
  unsigned long id,i;
  unsigned long wl_start,wl_end;       /* word- aligned pointers to data */
  int preamble;                        /* which byte does the data start */

  if (hpam_reply_dest != -1) {

    if (len > HPAM_MAX_PAYLOAD) {
      fprintf(stderr,"HPAM_reply_xfer() got packet longer then %d\n",
	      HPAM_MAX_PAYLOAD);
      len = HPAM_MAX_PAYLOAD;
    }

    if (hpam_reply_dest != HPAM_self_address) {

      state_buf = &(hpam_rep_state[hpam_reply_dest][hpam_reply_buf]);
      base = (unsigned long) state_buf->buf;

      /* lock buffer */
      preamble = ((int) src_buff) & 0x3;
      state_buf->size=base+PKT_X_DATA+preamble+len;  

      /* sequence number */
      id = (state_buf->state &ST_ID_MASK);
      out_word(base+PKT_FUN,fun);
      out_word(base+PKT_X_ARG,arg);
      out_word(base+PKT_CTL,(preamble<<OFFSET_SHIFT)|REP|XFER|
	       (id<<ID_SHIFT)|hpam_reply_buf);
      out_word(base+PKT_X_BASE,dest_buff);

      hpam_reply_dest = -1;
      /* copy the data */
      wl_start = (int)src_buff&(~0x3);
      wl_end =wl_start+((len+0x3)&(~0x3));
      i=base+PKT_X_DATA;

      if (len > 0) {
	while(wl_start<=wl_end) {
	  out_word(i,*((unsigned long *)wl_start));
	  wl_start+=4,i+=4;
	}
      }
      out_word(TX_READY_FIFO,base+PKT_X_DATA+preamble+len); /* send it */

      state_buf->state = (id+1)&ST_ID_MASK;
    }
    else {
      hpam_reply_dest=-1;
      memcpy(dest_buff,src_buff,len);
      (*fun)(HPAM_self_address,arg, dest_buff,len);    
    }
  }
  else{
    fprintf(stderr,"HPAM_reply_xfer() called outside handler\n");
  }
  return 0;
} /* HPAM_reply_xfer */

void HPAM_bcast_xfer(void (*fun)(),void *arg,void *src_buff,
     void *dest_buff,int len)
{
  register unsigned long base;         /* points to head of packet */
  unsigned long i;
  unsigned long wl_start,wl_end;      /* word- aligned pointers to data */
  int preamble;                        /* which byte does the data start */

  base = (unsigned long) hpam_req_state[HPAM_self_address][0].buf;

  /* lock buffer */
  hpam_req_state[HPAM_self_address][0].size=base+PKT_X_DATA+preamble+len;  

  if (len > HPAM_MAX_PAYLOAD) {
    fprintf(stderr,"HPAM_bcast_xfer() got packet longer then %d\n",
	    HPAM_MAX_PAYLOAD);
  }

  out_word(base+PKT_FUN,fun);

  preamble = ((int) src_buff) & 0x3;
  wl_start = ((unsigned long)src_buff&(~0x3));

  out_word(base+PKT_CTL,BCAST|XFER);
  out_word(base+PKT_X_ARG,arg);
  out_word(base+PKT_X_BASE,dest_buff);
  
  i=base+PKT_X_DATA;
  wl_end =wl_start+((len+0x3)&(~0x3));
  while(wl_start<=wl_end) {
    out_word(i,*((unsigned long *)wl_start));
    wl_start+=4,i+=4;
  }

  out_word(TX_READY_FIFO,base+PKT_X_DATA+preamble+len); /* send it */
  hpam_req_state[HPAM_self_address][0].size=0;

  /* now handle function call to self */
  
  memcpy(dest_buff,src_buff,len);
  (*fun)(HPAM_self_address,arg,dest_buff,len);
  am_poll();

}
 /* HPAM_bcast_xfer */

/* synthesize get out of am_request_4 and reply xfer */
static void HPAM_get_handler(int src, void (*fun)(), void *arg,void *src_buf, 
			     void *dest_buf, unsigned int len) {
  HPAM_reply_xfer(fun,arg,src_buf,dest_buf,len);
}

int am_get(int dest, void *rva, void *lva, unsigned int len, 
	   void (*handler)(), void *arg) {

  register buf_desc *state_buf;        /* buffer descriptor */
  register unsigned long state_word;   /* tmp state word        */
  register unsigned long base;         /* points to head of packet */
  unsigned long id, free;              /* ID, buffer pair */
  unsigned long i;

  if (dest != HPAM_self_address)  {

    if ((hpam_req_state[dest][hpam_free[dest]].state)&TX_RD) {
      state_buf = &(hpam_req_state[dest][hpam_free[dest]]);
      free = hpam_free[dest];

    sendit:    
      base = (unsigned long) state_buf->buf;

    /* lock buffer */
      state_buf->size = base+PKT_4_DATA+20;
      state_buf->state = (state_buf->state&(~TX_RD)) | TX_NR; 
      state_word = state_buf->state;

    /* next ID number */
      id = (state_word+1)&ST_ID_MASK;

      out_word(base+PKT_CTL,GET|REQ|(id<<ID_SHIFT)|free);
      out_word(base+PKT_FUN,HPAM_get_handler);
      out_word(base+PKT_4_DATA,handler);
      out_word(base+PKT_4_DATA+4,arg);
      out_word(base+PKT_4_DATA+8,rva);
      out_word(base+PKT_4_DATA+12,lva);
      out_word(base+PKT_4_DATA+16,len);
      out_word(TX_READY_FIFO,base+PKT_4_DATA+20); /* send it */

      state_word = (state_word&(~ST_ID_MASK))|(id&ST_ID_MASK); /* id number */
      state_word = (state_word&(~RETRY_MASK)); /* retries = 0 */
      state_buf->state = state_word;

      hpam_free[dest]= (hpam_free[dest]+1)&(MAXDEPTH-1); /* inc & mod */
      GETITMR(state_buf->timestamp); /* set timestamp */
      am_poll();
      return 0;
    }
    else {			/* We have to find a free buffer */
      for (i=0;i<MAXDEPTH;i++) {
	if ( (hpam_req_state[dest][i].state) & TX_RD) {
	  state_buf = &(hpam_req_state[dest][i]);
	  free=i;     
	  goto sendit;
	}
      }
      if ( stall(dest) != -1) {
	state_buf = &(hpam_req_state[dest][hpam_free[dest]]);
	free = hpam_free[dest];
	goto sendit;
      }
      else 
	return -1;
    }
  }

  else{				/* message is to self */
    memmove(lva,rva,len); /* must use mmove in case of overlap */
    (*handler)(arg,rva,len);
  }

  am_poll();
  return 0;



  
}

void HPAM_wait(volatile int *flag,int value)
{
  do {
    am_poll();
  } while (*flag < value);
 
  *flag -= value;
}

void HPAM_wait_until_zero(volatile int *flag)
{
  while (*flag)
    am_poll();
}

void am_wait(volatile int *flag,int value) {
  HPAM_wait(flag,value);
}

int am_max_size() {
  return HPAM_MAX_PAYLOAD;
}


#ifdef PROFILE
initstats()
{
  hpam_4_check=hpam_4_lock= hpam_4_store= hpam_4_update=hpam_4_service =0;
  
  hpam_X_check=hpam_X_lock= hpam_X_store= hpam_X_update=hpam_X_service =0;

  hpam_P_status= hpam_P_calls =hpam_P_protection = hpam_P_reqrep=
               hpam_P_4xfer  =  hpam_P_dispatch  = hpam_P_ack= hpam_Ps =
               hpam_P_rechecks = hpam_P_retrans = hpam_P_tc = hpam_P_critical =
	       hpam_P_release =0 ;

}

printstats()

{
  printf("total HPAM_4s %d stalls %d\n",hpam_4s,hpam_4_stalls);
  if (hpam_4s >0 ) {

    printf("HPAM_4 check \t%d \t%lf\n",hpam_4_check,
	   (double)hpam_4_check/(double)hpam_4s);
    printf("HPAM_4 lock \t%d \t%lf \n",hpam_4_lock,
	   (double)hpam_4_lock/(double)hpam_4s);
    printf("HPAM_4 store \t%d \t%lf \n",hpam_4_store,
	   (double)hpam_4_store/(double)hpam_4s);
    printf("HPAM_4 update \t%d \t%lf \n",hpam_4_update, 
	   (double)hpam_4_update/(double)hpam_4s);
    printf("HPAM_4 service \t%d \t%lf \n",hpam_4_service, 
	   (double)hpam_4_service/(double)hpam_4s);

  }

  printf("total HPAM_xferss %d stalls %d\n",hpam_Xs,hpam_X_stalls);
  if (hpam_Xs >0 ) {

    printf("HPAM_X check \t%d \t%lf\n",hpam_X_check,
	   (double)hpam_X_check/(double)hpam_Xs);
    printf("HPAM_X lock \t%d \t%lf \n",hpam_X_lock,
	   (double)hpam_X_lock/(double)hpam_Xs);
    printf("HPAM_X store \t%d \t%lf \n",hpam_X_store,
	   (double)hpam_X_store/(double)hpam_Xs);
    printf("HPAM_X insert \t%d \t%lf \n",hpam_X_insert,
       (double)hpam_X_insert/(double)hpam_Xs);
    
    printf("HPAM_X update \t%d \t%lf \n",hpam_X_update, 
	   (double)hpam_X_update/(double)hpam_Xs);
    printf("HPAM_X service \t%d \t%lf \n",hpam_X_service, 
	   (double)hpam_X_service/(double)hpam_Xs);

  }

  printf("poll calls %d polls-requests %d status %d rechecks %d \n",
	 hpam_P_calls,hpam_Ps,hpam_P_status,
	 hpam_P_rechecks);

  if (hpam_Ps >0 ) {
    printf("HPAM_P time lock  \t%d \t%lf\n",  hpam_P_tc ,
	   (double) hpam_P_tc /(double)hpam_Ps);

    printf("HPAM_P critical  \t%d \t%lf\n", hpam_P_critical,
	   (double)   hpam_P_critical /(double)hpam_Ps);

    printf("HPAM_P status \t%d \t%lf\n",  hpam_P_status ,
	   (double) hpam_P_status /(double)hpam_Ps);
    
    printf("HPAM_P crc \t%d \t%lf\n",  hpam_P_crc ,
	   (double) hpam_P_crc /(double)hpam_Ps);
    
    printf("HPAM_P protection \t%d \t%lf\n",  hpam_P_protection ,
	   (double) hpam_P_protection /(double)hpam_Ps);

    printf("HPAM_P req or rep \t%d \t%lf \n", hpam_P_reqrep ,
	   (double) hpam_P_reqrep  /(double)hpam_Ps);

    printf("HPAM_P 4 or xfer \t%d \t%lf \n", hpam_P_4xfer,
	   (double) hpam_P_4xfer /(double)hpam_Ps);

    printf("HPAM_P dispatch   \t%d \t%lf \n",hpam_P_dispatch,
       (double) hpam_P_dispatch/(double)hpam_Ps);

    printf("HPAM_P ack \t%d \t%lf \n", hpam_P_ack,
       (double) hpam_P_ack /(double)hpam_Ps);

    printf("HPAM_P check \t%d \t%lf \n", hpam_P_retrans,
       (double) hpam_P_retrans/(double)hpam_P_rechecks);    
  }

}

#endif
