/*                                                                      
 * HPAM - Precision architecture/Medusa Active Message layer, V2.0
 *        HPAM_barrier,etc
 *
 * "Copyright (c) 1993 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 		Richard P. Martin
 * Version:		2
 * Creation Date:	Fri Nov 26 18:16:02 EST 1993
 * Filename:		hpam_barrier.c
 * History:
 *	RPM	2	Fri Nov 26 15:19:43 1993
 *		new version on barrier based on ssl's version 
 */


#include <stdio.h>
#include <stdlib.h>
#include "hpam.h"

#define MYPROC HPAM_self_address
#define PROCS  HPAM_partition_size

static int up_buf[2];
static int down_buf;

void HPAM_barrier_enable()
{
	up_buf[0] = up_buf[1] = down_buf = 0;
}

void HPAM_incr_handler(int src,void *val)
{
	(*((int*)val))++;
}

void HPAM_barrier_up()
{
    int parent = (MYPROC - 1) / 2;
    int odd_child = 2 * MYPROC + 1;
    int even_child = 2 * MYPROC + 2;
    int parity = MYPROC & 1;

    if (MYPROC == 0) {
      if (PROCS != 1) 
	if (PROCS == 2) {
	  while (up_buf[1] == 0)
	    am_poll();
	} else {
	  while (up_buf[0] == 0 || up_buf[1] == 0)
	    am_poll();
	}
    } else {
      if (odd_child >= PROCS) {
	am_request_4(parent, HPAM_incr_handler, (void*)&up_buf[parity]);
      } else if (even_child >= PROCS) {
	while (up_buf[1] == 0)
	  am_poll();
	am_request_4(parent, HPAM_incr_handler, (void*)&up_buf[parity]);
      } else {
	while (up_buf[0] == 0 || up_buf[1] == 0)
	  am_poll();
	am_request_4(parent, HPAM_incr_handler, (void*)&up_buf[parity]);
      }
    } /* not zero */
    up_buf[0] = up_buf[1] = 0;
}

/* the nxam version uses a down tree. For HPAM use a broadcast 
 */

void HPAM_barrier_down()
{
    int left = 2 * MYPROC + 1;
    int right = 2 * MYPROC + 2;
    if (MYPROC != 0) {
      HPAM_wait(&down_buf,1);
    }

    if (left < PROCS)
      am_request_4(left, HPAM_incr_handler, (void*)&down_buf);
    if (right < PROCS)
      am_request_4(right, HPAM_incr_handler, (void*)&down_buf);
    
    down_buf = 0;
} 

void HPAM_barrier()
{
	HPAM_barrier_up();
	HPAM_barrier_down();
}

void 
am_barrier() {

  HPAM_barrier();

}

