/*                                                                      
 * HPAM- HP-PA/Medusa  Active Message layer, V2.1
 *        Initialization, Open card, etc.
 *
 * "Copyright (c) 1994 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author:              Richard P. Martin
 * Version: 		3
 * Creation Date:       Mon Oct 25 23:01:45 PDT 1993
 * Filename:            hpam_init.c
 * History:
 *	RPM	3	Wed May 11 16:35:19 1994
 *		moved glunix code to hpam_utils.c
 *	RPM	2	Wed May  4 00:28:33 1994
 *		added glunix code 
 */

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/un.h>
#include <sys/utsname.h>
#include <sys/ioctl.h>
#include <sys/framebuf.h>
#include <sys/netio.h>

#include "med.h"
#include "hpam_int.h"

extern void HPAM_barrier_enable();

/* globals */
static fddi_mac_addr_t node_macs[MAXNODES];/* Mac Addrs of the various nodes */
static unsigned short local_mac[3];       
extern unsigned long med_card_base;   /* base addr where card is mapped in */
extern int med_cardFD;                /* file ptr top open/close card*/

struct t_NI_save_area *hpam_save_area;/* glunix client shared memory segment */
int hpam_pid;                       /* now pid number            */
int HPAM_self_address;              /* local part of GPID number */
int HPAM_partition_size;            /* number of nodes currently executing */

int hpam_free[MAXNODES];            /* index to a possibly free buffer */
int hpam_reply_dest;                /* has the hander replied          */
int hpam_reply_buf;                 /* buffer to use for reply         */

int hpam_check_proc;                /* next processor to check         */
int hpam_check_buf;                 /* next buffer to check            */

/* start of the buffer descriptor headers */
buf_desc (*hpam_req_state)[MAXDEPTH];   /* request area */
buf_desc (*hpam_rep_state)[MAXDEPTH];   /* reply area   */

extern unsigned long *hpam_busy;    /* which buffer is currently in use*/
extern unsigned long *hpam_recv_nack; /* recv OK, 1= recv failed       */
unsigned long *hpam_in_q;           /* is the in_q non-zero?         */

/* enable the card */
/* set up the various buffers and such */
void am_enable()
{
  unsigned long req_buf, rep_buf;
  int i,j,k;
  fddi_mac_addr_t *addr;

  if (!initialise_medusa())
    {
        fprintf (stderr,"hpam_enable() couldn't map medusa --- exiting\n");
        exit(1);
    }

  /* try to attach to the scheduling daemon */

  hpam_pid = Glib_GetNpidByPid(getpid());

  
  local_mac[0] = in_word(MAC_MLA_C) & 0xffff;
  local_mac[1] = in_word(MAC_MLA_B) & 0xffff;
  local_mac[2] = in_word(MAC_MLA_A) & 0xffff;


  HPAM_partition_size=Glib_GetPartitionSize();
  HPAM_self_address=Glib_GetMyVnn();

  hpam_reply_dest = -1;        
  hpam_check_proc =0;
  hpam_check_buf =0;

  for (i=0;i<HPAM_partition_size;i++) {
    addr = (fddi_mac_addr_t *)Am_GetFastAddrByVnn(i);
    node_macs[i].a = addr->a;
    node_macs[i].b = addr->b;    
    node_macs[i].c = addr->c;    
  }

  if ( hpam_save_area == NULL) {
    hpam_save_area = (struct t_NI_save_area *) getshmbyNPID(hpam_pid);

    if (hpam_save_area == NULL) {
      fprintf(stderr,"hpam_enable(): unable to attach shared segment\n");
      exit(1);
    }
  }

  hpam_busy = &(hpam_save_area->lock);

  if (*hpam_busy) {
    fprintf(stderr,"am_enable: Medusa card appears to be locked by pid %d\n",
	    *hpam_busy);
    fprintf(stderr,"am_enable: Call hpam_clear_lock() to remove lock\n");
    fprintf(stderr,"am_enable: Make sure to call am_disable() before exiting\n");
    exit(1);
  }

  *hpam_busy= (int) getpid();

  hpam_save_area->in_q_size=0;
  hpam_save_area->req_state = (buf_desc *) hpam_save_area->req_buffer;
  hpam_save_area->rep_state = (buf_desc *) hpam_save_area->rep_buffer;
  hpam_save_area->in_q = hpam_save_area->in_q_buffer;
  hpam_save_area->in_q_head =0;
  hpam_save_area->in_q_tail =0; 
  hpam_save_area->out_q = hpam_save_area->out_q_buffer;
  hpam_save_area->out_q_head =0;
  hpam_save_area->out_q_tail =0;
  hpam_save_area->out_q_lock =0;
  hpam_recv_nack = &(hpam_save_area->rx_nack);
  hpam_in_q = &(hpam_save_area->in_q_size);
  
  /* reset the whole damn card */
  out_word (INTERRUPT_EVENT, 0xFFFFFFFF);   /* Reset this lot, too */
  out_word (INTERRUPT_MASK, 0x00000000);    /* Disallow all interrupts */
  hpam_medusa_init(2*(MAXNODES*MAXDEPTH),63,0x00000000);

  hpam_req_state = (buf_desc (*)[MAXDEPTH]) hpam_save_area->req_state;
  hpam_rep_state = (buf_desc (*)[MAXDEPTH]) hpam_save_area->rep_state;
  
  /* init the data seq bits and state  */
  for (i=0;i<HPAM_partition_size;i++) {

    hpam_free[i]=0;
    
    for (j=0;j<MAXDEPTH;j++) {
      hpam_req_state[i][j].timestamp=0;
      hpam_rep_state[i][j].timestamp=0;

      hpam_req_state[i][j].state = TX_RD;
      hpam_rep_state[i][j].state = 0x1;  /* init sequence number */

      hpam_req_state[i][j].buf = (unsigned long *)
	((unsigned long) CARD_MEMORY_BASE+ ((i*MAXDEPTH+j)*MED_BUFSIZE));

      
      hpam_rep_state[i][j].buf = (unsigned long *)
	( (unsigned long) CARD_MEMORY_BASE+
	 ((MAXDEPTH*MAXNODES+i*MAXDEPTH+j)*MED_BUFSIZE));	

      
      req_buf = (unsigned long) hpam_req_state[i][j].buf;
      rep_buf = (unsigned long) hpam_rep_state[i][j].buf;

      /* pre-load the request and reply buffers */
      if (i != HPAM_self_address) {

	out_word (req_buf, PKT_HDR);
	out_word (req_buf+4, ((node_macs[i].c & 0x0000ff00) << 16) |
		  ((node_macs[i].c & 0x000000ff) << 16) |
		  ((node_macs[i].b & 0x0000ff00)) |
		  ((node_macs[i].b & 0x000000ff)));
	out_word (req_buf+8,((node_macs[i].a & 0x0000ff00) << 16) |
		  ((node_macs[i].a & 0x000000ff) << 16) |
		  ((local_mac[0]  & 0x0000ff00)) |
		  ((local_mac[0]  & 0x000000ff)));
	out_word (req_buf+12, ((local_mac[1]  & 0x0000ff00) << 16) |
		  ((local_mac[1]  & 0x000000ff) << 16) |
		  ((local_mac[2]  & 0x0000ff00)) |
		  ((local_mac[2]  & 0x000000ff)));
	out_word(req_buf+16, 
		 (unsigned long )((hpam_pid<<PID_SHIFT) |
				  (i<<8) | HPAM_self_address));

	out_word (rep_buf, PKT_HDR);
	out_word (rep_buf+4,  ((node_macs[i].c & 0x0000ff00) << 16) |
		  ((node_macs[i].c & 0x000000ff) << 16) |
		  ((node_macs[i].b & 0x0000ff00)) |
		  ((node_macs[i].b & 0x000000ff)));
	out_word (rep_buf+8,  ((node_macs[i].a & 0x0000ff00) << 16) |
		  ((node_macs[i].a & 0x000000ff) << 16) |
		  ((local_mac[0]  & 0x0000ff00)) |
		  ((local_mac[0]  & 0x000000ff)));
	out_word (rep_buf+12, ((local_mac[1]  & 0x0000ff00) << 16) |
		  ((local_mac[1]  & 0x000000ff) << 16) |
		  ((local_mac[2]  & 0x0000ff00)) |
		  ((local_mac[2]  & 0x000000ff)));
	out_word(rep_buf+16, 
		 (unsigned long )((hpam_pid<<PID_SHIFT) |
				  (i<<8) | HPAM_self_address));    

      }
      /* for node n, use req buffer space n for broadcast */
      /* the rep buffer goes unused */
      else {

	out_word (req_buf, PKT_HDR);
	out_word (req_buf+4,  ((0xffff & 0x0000ff00) << 16) |
		  ((0xffff & 0x000000ff) << 16) |
		  ((0xffff & 0x0000ff00)) |
		  ((0xffff & 0x000000ff)));
	out_word (req_buf+8,  ((0xffff & 0x0000ff00) << 16) |
		  ((0xffff & 0x000000ff) << 16) |
		  ((local_mac[0]  & 0x0000ff00)) |
		  ((local_mac[0]  & 0x000000ff)));
	out_word (req_buf+12, ((local_mac[1]  & 0x0000ff00) << 16) |
		  ((local_mac[1]  & 0x000000ff) << 16) |
		  ((local_mac[2]  & 0x0000ff00)) |
		  ((local_mac[2]  & 0x000000ff)));
	out_word(req_buf+16, 
		 (unsigned long )((hpam_pid<<PID_SHIFT) |
				  (0xff<<8) | HPAM_self_address));
      }
    }
  }
  HPAM_barrier_enable();
  HPAM_barrier(); /* GAM spec says am_enable is implicit barrier */
} /* am_enable */


/*
 * Close the device file for the medusa card
 */
  
int shutdown_medusa()
{
  int result=0;
  if ((result = close (med_cardFD)) < 0)
    perror("shutdown_medusa(): close medusa card failed");
  return result;
}

/* store a buffer on the per-process output queue */
hpam_buffer(unsigned long recv_start, unsigned long recv_len)
{
  /* under construction */
}

/* called if the input queue is non-empty, returns a pointer to the next
 * packet
 */

unsigned long hpam_get_input()
{
  /* also under construction */
}

int am_my_proc() {
  return HPAM_self_address;
}

int am_procs() {
  return HPAM_partition_size;
}
int am_myproc() {
  return HPAM_self_address;
}

void am_disable() {
  int i;
  for (i=0; i<HPAM_partition_size;i++) {
    HPAM_flush(i);
  }
  *hpam_busy=0;
  shutdown_medusa();
}

void hpam_clear_lock() {

  if (hpam_save_area == NULL) {
    hpam_save_area = (struct t_NI_save_area *) getshmbyNPID(1);

    if (hpam_save_area == NULL) {
      fprintf(stderr,"hpam_clear_lock: unable to attach shared segment\n");
      exit(1);
    }
  }
  hpam_save_area->lock =0;
}
