/*                                                                      tab:8
 * HPAM - HP Medusa Active Message layer, V2.1
 *        Internal Private Header file
 *
 * "Copyright (c) 1993 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author:              Richard P. Martin
 * Version: 		1
 * Creation Date: 	Mon Oct 25 22:58:45 PDT 1993
 * Filename: 		hpam_int.h
 * History:
 * 	RPM	1	Mon Oct 25 22:58:45 PDT 1993
 */

#ifndef _HPAM_INT_H_
#define _HPAM_INT_H_
void HPAM_enable_barrier();
void HPAM_poll();
unsigned long med_copy(unsigned char *src, unsigned char *dest, unsigned len);
unsigned long med_vcopy(unsigned char *src, unsigned char *dest, unsigned len);

#define INT_TX_READY  0x00000080   /* card register values */
#define INT_TX_FREE   0x00000040
#define INT_RX_READY  0x00000020
#define INT_RX_FREE   0x00000010
#define INT_ELM       0x00000008
#define INT_MAC       0x00000004
#define INT_ALL       0x000000FC
#define INT_NONE      0x00000000

#define MAXNODES 4
#define MAXDEPTH 4
#define TRUE 1
#define FALSE 0

#define out_word(addr, datum) *((volatile unsigned long *)(addr)) = (unsigned long)(datum)
#define in_word(addr) *(volatile unsigned long *)(addr)

/* bit fields for the state word in the buffer descriptor */
#define TX_RD        0x80000000   /* Ready--- OK to X-mit       */
#define TX_NR        0x00000000   /* Not Ready to X-mit         */
#define ST_ID_MASK   0x000001ff   /* ID number                  */
#define RETRY_MASK   0x000f0000   /* number of re-transmissions */
#define RETRY_SHIFT  0x00000010

/* The following a packet/card buffer descriptor. It holds all the state
 * associated with a packet
 */
struct buffer_descriptor {
 unsigned long *buf;            /* base addr in the medusa      */
 unsigned long timestamp;       /* time the packet was sent     */
 unsigned long state;           /* state word                   */
 unsigned long size;            /* re-transmission start/length */
};

typedef struct buffer_descriptor buf_desc;

/* bit fields for the control word in the message */
#define OFFSET_MASK  0xf0000000   /* where does the data start? */
#define OFFSET_SHIFT 0x1c
#define GET_MASK     0x02000000   /* a get message tag type    */
#define GET          0x02000000   /* a get message tag type    */
#define ARGS_MASK    0x01000000   /* args or 4 word message?   */
#define XFER         0x00000000   /* an Xfer packet            */
#define ARGS4        0x01000000   /* a packet with 4 args      */
#define ACK_MASK     0x00400000   /* Is this an empty ack?     */
#define ACK          0x00400000   /* ack packet. Not ack = 0   */
#define NET_MASK     0x00300000   /* Is this a req/rep bit     */
#define REQ          0x00000000   /* a request packet          */
#define REP          0x00100000   /* a reply   packet          */
#define BCAST        0x00200000   /* a broadcast packet        */
#define ID_SHIFT     0x0000000a   /* shift to get the ID       */
#define ID_MASK      0x0007fc00   /* the sequence number       */
#define BUF_MASK     0x000003ff   /* the buffer number         */

#define FDDI_TRAILER 0x6          /* size of FDDI trailer       */
#define FDDI_CRC_OK  0x80         /* mask to each CRC on FDDI   */

/* all the following values are in itmr ticks */
#define MAXITMR      0xffffffff 
#define MAX_RETRIES  0x00000009  /* multiplier for timeout        */
#define BASE_RETRIES 0x00800000   /* min cycles between re-txmists */
#define RETRIES_X    0x003fffff   /* extra time to wait  */


/* get the value of the interval timer */ 
#ifdef __GNUC__
#define GETITMR(x) asm("mfctl %%itmr,%0" : "=r" (x));
#define INLINE inline
#else
#define GETITMR(x) (x=mfctl(16))
#define INLINE 
#endif

#ifdef PROFILE 
#define PROF_GETITMR(x) GETITMR(x) 
#else
#define PROF_GETITMR(x) 
#endif 

/* the fddi mac addresses is formatted as c.b.a */

struct fddi_mac_addr_s {
  unsigned short c;
  unsigned short b;
  unsigned short a;
};
typedef struct fddi_mac_addr_s fddi_mac_addr_t;


 /* offsets from start of packet  */
#define PKT_MAGIC 0      /* magic at start of FDDI */
#define PKT_MACA 4       /* 2 48 bit MAC addresses */
#define PKT_MACB 8    
#define PKT_MACC 12
#define PKT_DST_SRC 16   /* Net PID/destination and source numbers       */
#define PKT_CTL 20       /* req/rep network,  sequence numbers           */
#define PKT_FUN 24       /* function pointer                             */
#define PKT_4_DATA 28    /* user data for HPAM_4                         */
#define PKT_X_BASE 28    /* base addr pointer for X_fer                  */
#define PKT_X_ARG  32    /* arg to function for the xfer                 */
#define PKT_X_DATA 36    /* xfer usr data start                          */

#define PKT_PID_MASK 0xffff0000   /* network PID */
#define PID_SHIFT    0x00000010   /* shift to get network pid */
#define PKT_SRC_MASK 0x000000ff   /* source number (virtual node number) */
#define PKT_DST_MASK 0x0000ff00   /* dest number   */
#define MAGIC_PID 0xf0f0          /* used when not using glunix */

#define HEADER_SIZE PKT_X_DATA
#define HPAM_MAX_PAYLOAD (8176-HEADER_SIZE)
#define MAX_WORDS (MAX_PAYLOAD/sizeof(unsigned long))
#define USE_VENOM_SIZE 128  /* cut off to use venom */

/* glunix defines for describing the state of the card */
#define MED_RX_BUSY 0xffffffff   /* going to pop the fifo ... */
#define MED_RX_FREE 0x00000000   /* fifo not in use           */
#define MED_RX_FAILED 0xffffffff  /*got contexted switched at the wrong time*/

#define MAX_SAVE_MSGS 4
#define MAX_SLACK_PKT  MED_BUFSIZE  /* a slack buffer packet (bytes)*/
#define MAX_Q_ELEMENTS (MAX_SAVE_MSGS)
#define BUF_Q_SIZE (MAX_Q_ELEMENTS*MAX_SLACK_PKT)

#define BUF_DESC_SIZE (MAXNODES*MAXDEPTH*(sizeof(buf_desc)))

struct t_NI_save_area { /* temp ni save area */
  unsigned long lock;        /* flag for set when about to pop the rx_fifo */
  unsigned long rx_nack;     /* rx failed   */
  unsigned long in_q_size;   /* keep here to keep on same cache line */
  buf_desc *req_state;       /* request state pointer */
  buf_desc *rep_state;       /* reply state           */
  char *in_q;                /* input queue (slack buffer ) */
  unsigned long in_q_head;
  unsigned long in_q_tail;  

  char *out_q;                /* output queue   */
  unsigned long out_q_lock;   /* lock the out queue, in case we take a hit */
  unsigned long out_q_head; 
  unsigned long out_q_tail;
  
  /* these shadow the buffers in the card */
  char req_shadow[(MAXNODES*MAXDEPTH)*MAX_SLACK_PKT]; 
  char rep_shadow[(MAXNODES*MAXDEPTH)*MAX_SLACK_PKT];
  char current_pkt [MAX_SLACK_PKT]; /* if in the middle of a poll */

  char req_buffer[BUF_DESC_SIZE]; /*space for request descriptors */
  char rep_buffer[BUF_DESC_SIZE]; /* reply descriptors */

  char in_q_buffer[BUF_Q_SIZE];  /* input slack buffer */
  char out_q_buffer[BUF_Q_SIZE]; /* output slack buffer */
};

#endif



