/*									tab:8
 *
 * hpam_utils.c - utilites to deal with medusa card descriptors, etc
 *
 * 
 * "Copyright (c) 1994 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 		Richard P. Martin
 * Version:		1
 * Creation Date:	Wed May 11 16:32:15 1994
 * Filename:		hpam_utils.c
 * History:
 */

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/un.h>
#include <sys/utsname.h>
#include <sys/ioctl.h>
#include <sys/framebuf.h>
#include <sys/netio.h>

#include "med.h"
#include "hpam_int.h"
#include "glib.h"

unsigned long med_card_base;   /* base addr where card is mapped in */
int med_cardFD=0;              /* file ptr top open/close card*/

int hpam_medusa_init (unsigned int tx_buffers, unsigned int rx_buffers, 
		 unsigned int_mask)
{
  unsigned long result;
  int i;
  int sm;
  int errors = 0;

  out_word (RESET_FIFO, 0x00000000);
  out_word (INTERRUPT_MASK, INIT_INTERRUPT_MASK);
  out_word (RESET_FIFO, 0xFFFFFFFF);

  if ((result = (in_word(RESET_FIFO) & 0xFF)) != INIT_RESET_FIFO)
    {
      fprintf(stderr, "ERR-2: RESET FIFO incorrectly set to 0x%08lx instead of 0x%08lx\n",
	      result, INIT_RESET_FIFO);
    }

  if ((result = (in_word(INTERRUPT_MASK) & 0xFF)) != INIT_INTERRUPT_MASK)
    {
      fprintf(stderr, "ERR-7: INTERRUPT MASK incorrectly set to 0x%08lx instead of 0x%08lx\n",
	      result, INIT_INTERRUPT_MASK);
      errors++;
    }

  if ((result = (in_word(STATUS_BUFF_FIFO) & 0xFF)) != INIT_STATUS_BUFF_FIFO)
    {
      fprintf (stderr, "ERR-11: BUFFER FIFO STATUS incorrectly set to 0x%08lx instead of 0x%08lx\n",
              result, INIT_STATUS_BUFF_FIFO);
      errors++;
    }

  out_word(RESET_FIFO, 0x00000000);
  out_word(RESET_FIFO, 0xffffffff);

  for (i=0; i < tx_buffers; i++) {
    out_word (TX_READY_FIFO, CARD_MEMORY_BASE + (i * MED_BUFSIZE));
  }
#ifdef DEBUG
  printf( "created %d transmit buffers\n", i);
#endif


  /* rx_buffers, create at most 64 in the upper half of the card */
  for (i=0; i < rx_buffers; i++) {
    out_word (RX_FREE_FIFO, CARD_MEMORY_BASE +
	                    (64+i) * MED_BUFSIZE);
  }
#ifdef DEBUG	      
  printf("created %d receive buffers\n", i);
#endif

  out_word (INTERRUPT_MASK, int_mask);

#ifdef DEBUG
  printf( "interrupts enabled:  ");
  if (int_mask & INT_TX_READY)
	printf( "TX_READY ");
  if (int_mask & INT_TX_FREE)
	printf( "TX_FREE ");
  if (int_mask & INT_RX_READY)
	printf( "RX_READY ");
  if (int_mask & INT_RX_FREE)
	printf( "RX_FREE ");
  if (int_mask & INT_ELM)
	printf( "ELM ");
  if (int_mask & INT_MAC)
	printf( "MAC ");
  if (!int_mask)
	printf( "NONE ");
  printf( "\n");
#endif

  if (errors) {
      fprintf(stderr, "med_hw_init:  FAILED\n\n");
      out_word (INTERRUPT_MASK, INIT_INTERRUPT_MASK);
      return errors;
    }
}  /* hpam_medusa_init */



int initialise_medusa()
{
  int result;
  int gd;
  char *gd_path = "/dev/crt";
  struct med_ioctl_arg ioctl_args;

#ifdef __hp9000s700

  /*
   * Open the device file, and get the card base address.
   */

  if (med_cardFD != 0) {
    fprintf(stderr,"initialise_medusa(): Warning, called >1 times\n");
    return TRUE;
  }

  med_cardFD = open ("/dev/medusa", O_RDWR);
  if (med_cardFD < 0)
    {
      perror("Couldn't open /dev/medusa");
      return(FALSE);
    }

  ioctl_args.location = NULL;
  ioctl_args.reg_value = 0;


  if (ioctl(med_cardFD, MED_BASE_ADDR, &ioctl_args) < 0)
    {
      perror("ioctl(MED_BASE_ADDR)");
      return(FALSE);
    }
  
  med_card_base = ioctl_args.reg_value;

  if (med_card_base != 0xF4000000 && med_card_base != 0xF8000000)
    {
      fprintf (stderr,"initialise_medusa(): illegal card base address 0x%08lx\n", med_card_base);
      return (FALSE);
    }

/*
 * Call the device driver to map the card into user address space
 */
  
  if (ioctl(med_cardFD, MED_VENOM_ENABLE) < 0)
    {
      perror ("ioctl(MED_VENOM_ENABLE)");
      return (FALSE);
    }
  
  if (ioctl(med_cardFD, MED_MAP) < 0)
    {
      perror ("ioctl(MED_MAP)");
      return (FALSE);
    }

  /* enable venom med_vcopy */
  /* code to open it as /dev/crt as well */

  if ((gd = open(gd_path, O_RDWR)) == -1)
    {
      perror("Could not open /dev/crt" );
      return(FALSE);
    }

  if (ioctl(gd, GCMAP) == -1)
    {
      perror("Could not map /dev/crt");
      return(FALSE);
    }
#else  /* !__hp9000s700 */
  fprintf (stderr, "This program must be run on a HP s700 machine");
  exit(255);
#endif /* __hp9000s700 */

  return (TRUE);

} /* initialise_medusa */


/*
 * Get the local FDDI MAC address from the Medusa card
 */

void med_get_MAC_addr (unsigned short *dest_addr)
{
  int i;
  unsigned char dest[6];
  unsigned long *rom;

  rom = (unsigned long *) CARD_BASE;

  for (i=0; i < 6; i++) {
    dest[i] = in_word(rom + (MED_MAC_OFFSET >> 2) + i);
  }; /* for */

  dest_addr[0] = (dest[0] << 8) + dest[1];
  dest_addr[1] = (dest[2] << 8) + dest[3];
  dest_addr[2] = (dest[4] << 8) + dest[5];

} /* med_get_MAC_addr */

/* stuff the glunix layer needs */

int Am_GetFastAddrSize()
{
  return(sizeof(fddi_mac_addr_t));
}

int Am_GetMyFastAddr(void *addr)
{
  fddi_mac_addr_t *fddi_addr;

  if (med_card_base == NULL) {
    if (!initialise_medusa()) {
      fprintf (stderr,"Am_GetMyFastAddr()couldn't map medusa\n");
      return -1;
    }   
  }

  fddi_addr = (fddi_mac_addr_t *) addr;

  fddi_addr->c = (short) (in_word(MAC_MLA_C) & 0xffff);
  fddi_addr->b = (short) (in_word(MAC_MLA_B) & 0xffff);
  fddi_addr->a = (short) (in_word(MAC_MLA_A) & 0xffff);

  return 0;
}

char * Am_GetFastAddrStr(Am_FastAddr fastAddr)
{

  static char fddi_ascii[16];
  fddi_mac_addr_t *addr;

  if (fastAddr == NULL) {
    sprintf(fddi_ascii,"error\0");
  } 
  else {
    addr= (fddi_mac_addr_t *) fastAddr;
    sprintf(fddi_ascii,"%x0%0x%0x",addr->a,addr->b,addr->c);
  }

  return &fddi_ascii[0];
}


/* routine which gets called in the glunix daemon to context 
 * switch out a process. Walks the descriptors and saves the state */

int med_save_state(int npid)
{
  struct t_NI_save_area *local_proc;
  struct t_NI_save_area *remote_proc;
  buf_desc (*req_state)[MAXDEPTH];
  buf_desc (*rep_state)[MAXDEPTH];
  int i,j,rem_npid;
  unsigned long start,len;

  local_proc = ( struct t_NI_save_area *)  getshmbyNPID(npid);

  if (local_proc == NULL) {
    fprintf(stderr,"med_save_state() PANIC! can't save state of npid %d\n",
	    npid);
    return -1;
  }


/* save the state of the current request/ reply buffers */
  req_state = (buf_desc (*)[MAXDEPTH]) local_proc->req_state;
  rep_state = (buf_desc (*)[MAXDEPTH]) local_proc->rep_state;

  for (i=0;i<MAXNODES;i++) {
    for(j=0;j<MAXDEPTH;j++) {
      if ( (req_state[i][j].state&TX_RD) == 0) {
	start = req_state[i][j].size&PKT_LEN_MASK;
	len = req_state[i][j].size& ~(PKT_LEN_MASK);
      }
      else 
	len = PKT_FUN;

      med_vcopy((char*) start, (char *)
		&(local_proc->req_shadow[(i*MAXDEPTH+j)*MAX_SLACK_PKT]),
		len);

      start = rep_state[i][j].size&PKT_LEN_MASK;
      len = rep_state[i][j].size& ~(PKT_LEN_MASK);
      
      if (len ==0 ) 
	len = PKT_DST_SRC;

      med_vcopy((char *) start, (char *)
		&(local_proc->rep_shadow[(i*MAXDEPTH+j)*MAX_SLACK_PKT]),
		len);
    } /* end for j */
  } /* end for i */

  if (local_proc->lock == MED_RX_BUSY) { /* app just popped the fifo */
    local_proc->rx_nack =1;
  }
  else {
    if (local_proc->lock != MED_RX_FREE) {/* app was in the middle of a poll */
      start = local_proc->lock&PKT_LEN_MASK;
      len = rep_state[i][j].size& ~(PKT_LEN_MASK);

      rem_npid = ((in_word(start+PKT_DST_SRC))&PKT_PID_MASK)>>PID_SHIFT;

      if (rem_npid != npid) { /* remote pid was different send to out_queue*/
	remote_proc =   (struct t_NI_save_area *) getshmbyNPID(rem_npid);
      
	if (((remote_proc->in_q_head+1)%MAX_Q_ELEMENTS) !=
	    remote_proc->in_q_tail) {  /* not full */
	  
	  memcpy( (char* ) 
		 remote_proc->in_q_head+(remote_proc->in_q_head*MAX_SLACK_PKT),
		 (char *) start,
		 len);
	  *( (int*) remote_proc->in_q_head+(remote_proc->in_q_head*MAX_SLACK_PKT)) = len;
	  remote_proc->in_q_head=((remote_proc->in_q_head+1)%MAX_Q_ELEMENTS) ;
	  *( (int *) local_proc->current_pkt) =0;
	} 
      }
      else {
	med_vcopy((char*) start, (char *) local_proc->current_pkt,len);
	*( (int*)local_proc->current_pkt) = len;
      }
    }
    else {  /* app was not in the middle of a poll */
      /* set current packet len to zero */
      *( (int *) local_proc->current_pkt) =0;
    }
  } /* end else */

  /* distribute the out queues */
  while (local_proc->out_q_head != local_proc->out_q_tail) {
    start =(unsigned long)
      (local_proc->out_q+(local_proc->out_q_head)*MAX_SLACK_PKT);
    len = *((int *) 
	      (local_proc->out_q+(local_proc->out_q_head)*MAX_SLACK_PKT));

    rem_npid = ((in_word(start+PKT_DST_SRC))&PKT_PID_MASK)>>PID_SHIFT;
    remote_proc = (struct t_NI_save_area *) getshmbyNPID(rem_npid);
    if (((remote_proc->in_q_head+1)%MAX_Q_ELEMENTS) !=
	    remote_proc->in_q_tail) {  /* not full */
      
      memcpy((char *) remote_proc->in_q+(remote_proc->in_q_head*MAX_SLACK_PKT),
	     (char *)start,len);
    }

    local_proc->out_q_tail = (local_proc->out_q_tail+1)%MAX_Q_ELEMENTS;

  }
} /* end med_save_state */

/* routine which gets called in the glunix daemon to context 
 * switch in a process. walks the descriptors and restores the state */
int med_restore_state(int npid) 
{
  
  struct t_NI_save_area *local_proc;

  buf_desc (*req_state)[MAXDEPTH];
  buf_desc (*rep_state)[MAXDEPTH];
  int i,j;
  unsigned long start,dest,len;

  local_proc = ( struct t_NI_save_area *)  getshmbyNPID(npid);

  if (local_proc == NULL) {
    fprintf(stderr,"med_save_state() PANIC! can't save state of npid %d\n",
	    npid);
    return -1;
  }


/* save the state of the current request/ reply buffers */
  req_state =  (buf_desc (*)[MAXDEPTH]) local_proc->req_state;
  rep_state =  (buf_desc (*)[MAXDEPTH]) local_proc->rep_state;

  for (i=0;i<MAXNODES;i++) {
    for(j=0;j<MAXDEPTH;j++) {
      if ( (req_state[i][j].state&TX_RD) == 0) {
	start = req_state[i][j].size&PKT_LEN_MASK;
	len = req_state[i][j].size& ~(PKT_LEN_MASK);
      }
      else 
	len = PKT_FUN;
      }

      start = rep_state[i][j].size&PKT_LEN_MASK;
      len = rep_state[i][j].size& ~(PKT_LEN_MASK);
      
      if (len >0 ) {
	med_vcopy((char *)
		  &(local_proc->rep_shadow[(i*MAXDEPTH+j)*MAX_SLACK_PKT]),
		  (char*) rep_state[i][j].buf,
		  len);
      } /* end for j */
    }  /* end for i */

  /* if local proc was in the middle of pulling out a packet,  */
  if (*( (int *) local_proc->current_pkt) !=0 ) {
    start =(unsigned long)  local_proc->current_pkt;
    len   =  (unsigned long) *( (int *) local_proc->current_pkt);
    dest = (unsigned long) local_proc->lock&PKT_LEN_MASK;

#ifdef DEBUG
    printf("restoring lock buffer %d %d to %d \n",start,dest,len);
#endif 

    med_vcopy((char*) start,(char*) dest,len);
  }

}

void GLUnix_medusa_init()
{
  if ( initialise_medusa() ==FALSE ) {
    fprintf(stderr,"GLUnix_medusa_init(): open Medusa card failed\n");
    return;
  }
  hpam_medusa_init(0,63,0);
}

