/* -*-C-*-
*******************************************************************************
*
* File:         med.h
* RCS:          $Header: med.h,v 4.0 93/06/24 16:25:41 aje Exp $
* Description:  Header for Medusa Network Interface Driver
* Author:       Mike Prudence, Hyper Project
* Created:      Fri Oct 11 13:22:00 1991
* Modified:     Fri May  8 17:16:37 1992 (mjp) mjp@prudence
* Language:     C
*
*
* Copyright (c) 1992   Hewlett-Packard Company.  All rights reserved.
* 
* All copies of this program, whether in part or in whole, and whether
* modified or not, must display this and all other embedded copyright
* and ownership notices in full.
* 
* The contents of this software are proprietary and confidential to the
* Hewlett-Packard Company.  Transfer to any party is strictly forbidden
* other than as expressly permitted in writing by Hewlett-Packard Company.
* Unauthorized transfer to or possession by any unauthorized party may be
* a criminal offense.
* 
* The program materials are licensed 'as is' with no warrenty of any kind.
* HP will not be responsible for the correction of any bugs or deficiencies.
* In no event shall HP be liable for any damages of any kind, including
* special, indirect or consequential damages, arising out of or in
* connection with the use or performance of the program materials.
* 
*******************************************************************************
*/

#ifndef _MED_H_INCLUDED
#define _MED_H_INCLUDED $Header: med.h,v 4.0 93/06/24 16:25:41 aje Exp $

#if defined(MEDUSA_SUPPORT) && !defined(MEDUSA_NINJA_STACK)
# define MEDUSA_NINJA_STACK
#endif

#define CARD_BASE         med_card_base + 0x00000000

#define STI_ROM           med_card_base + 0x00000000

#define ELM               med_card_base + 0x00800000
#define MAC               med_card_base + 0x00c00000

#define RESET_MISC        med_card_base + 0x01000000
#define RESET_FIFO        med_card_base + 0x01000004
#define RESET_FDDI        med_card_base + 0x01000008

#define TEST_FCG          med_card_base + 0x0100000C
#define TEST_ELM          med_card_base + 0x01000010
#define TEST_MAC          med_card_base + 0x01000014

#define INTERRUPT_MASK    med_card_base + 0x01000018
#define INTERRUPT_EVENT   med_card_base + 0x0100001C

#define MISC_OUTPUT       med_card_base + 0x01000024
#define MISC_INPUT        med_card_base + 0x01000030

#define STATUS_SYNCH_FIFO med_card_base + 0x01000028
#define STATUS_BUFF_FIFO  med_card_base + 0x0100002C

#define CHECKSUM_UNIT     med_card_base + 0x01400000

#define TX_READY_FIFO     med_card_base + 0x02000000

#define TX_FREE_FIFO      med_card_base + 0x02400000

#define RX_READY_FIFO     med_card_base + 0x02800000
#define RX_FREE_FIFO      med_card_base + 0x02C00000

#define CARD_MEMORY_BASE  med_card_base + 0x03000000
#define CARD_CKSUM_BASE   med_card_base + 0x03800000

#define CARD_MEMORY_SIZE  0x00100000

/*
 * Bitmask definitions for the Interrupt Status Register
 */

#define ISR_TX_READY_FULL     0x80
#define ISR_TX_FREE_EMPTY     0x40
#define ISR_RX_READY_EMPTY    0x20
#define ISR_RX_FREE_FULL      0x10
#define ISR_ELM_INTERRUPT     0x08
#define ISR_MAC_INTERRUPT     0x04

#define PKT_LEN_MASK      0x00001FFF

/*
 * Bitmask definitions for the FIFO Reset Register
 */

#define RST_TX_READY     0x80
#define RST_TX_FREE      0x40
#define RST_RX_READY     0x20
#define RST_RX_FREE      0x10
#define RST_TX_SYN       0x08
#define RST_RX_SYN       0x04

/*
 * Bitmask definitions for the Misc RESET register
 */

#define RST_FDDI_CONTROLLER   0x80
#define RST_SERIAL_CONTROLLER 0x40

/*
 * Bitmask definitions for the Misc Output register
 */

#define SERIALRX_TEST       0x80

/*
 * Bitmask definitions for the buffer FIFO status register 
 */

#define BSR_TX_READY_FULL  0x80
#define BSR_TX_READY_EMPTY 0x40
#define BSR_TX_FREE_FULL   0x20
#define BSR_TX_FREE_EMPTY  0x10
#define BSR_RX_READY_FULL  0x08
#define BSR_RX_READY_EMPTY 0x04
#define BSR_RX_FREE_FULL   0x02
#define BSR_RX_FREE_EMPTY  0x01

/*
 * MAC Control and status registers
 */

#define MAC_CNTRL_A             med_card_base + 0x00C00000
#define MAC_CNTRL_B             med_card_base + 0x00C00004
#define MAC_INTR_MASK_A         med_card_base + 0x00C00008
#define MAC_INTR_MASK_B         med_card_base + 0x00C0000C

#define MAC_MSA                 med_card_base + 0x00C00040
#define MAC_MLA_A               med_card_base + 0x00C00044
#define MAC_MLA_B               med_card_base + 0x00C00048
#define MAC_MLA_C               med_card_base + 0x00C0004C
#define MAC_T_REQ               med_card_base + 0x00C00050
#define MAC_TVX_VALUE           med_card_base + 0x00C00054
#define MAC_T_MAX               med_card_base + 0x00C00054

#define MAC_FRAME_CT            med_card_base + 0x00C00080
#define MAC_LOST_CT             med_card_base + 0x00C00084
#define MAC_ERROR_CT            med_card_base + 0x00C00084
#define MAC_INTR_EVENT_A        med_card_base + 0x00C00088
#define MAC_INTR_EVENT_B        med_card_base + 0x00C0008C

#define MAC_RX_STATUS           med_card_base + 0x00C00090
#define MAC_TX_STATUS           med_card_base + 0x00C00094
#define MAC_T_NEG_A             med_card_base + 0x00C00098
#define MAC_T_NEG_B             med_card_base + 0x00C0009C
#define MAC_INFO_REG_A          med_card_base + 0x00C000A0
#define MAC_INFO_REG_B          med_card_base + 0x00C000A4

#define MAC_BIST_SIGNATURE      med_card_base + 0x00C000A8
#define MAC_TVX_TIMER           med_card_base + 0x00C000AC
#define MAC_TRT_TIMER_A         med_card_base + 0x00C000B0
#define MAC_TRT_TIMER_B         med_card_base + 0x00C000B4
#define MAC_THT_TIMER_A         med_card_base + 0x00C000B8

#define MAC_THT_TIMER_B         med_card_base + 0x00C000BC
#define MAC_SENTCOUNT           med_card_base + 0x00C000BC
#define MAC_PKT_REQUEST         med_card_base + 0x00C000C0
#define MAC_RC_CRC_A            med_card_base + 0x00C000C4
#define MAC_RC_CRC_B            med_card_base + 0x00C000C8
#define MAC_TX_CRC_A            med_card_base + 0x00C000CC
#define MAC_TX_CRC_B            med_card_base + 0x00C000D0


/*
 * TX Status masks
 */

#define MAC_RING_OPERATIONAL    0x0800

/*
 * MAC_CNTRL_A definitions
 */

#define MAC_ON                      0x8000
#define MAC_SETBIT5                 0x4000
#define MAC_SETBIT4                 0x2000
#define MAC_REVERSE_ADDR            0x1000
#define MAC_FLUSH_SA47              0x0800
#define MAC_COPYALL                 0x0600
#define MAC_COPYOWN                 0x0100
#define MAC_COPY_EXTRA_SMT          0x00C0
#define MAC_COPY_IND_LLC            0x0020
#define MAC_COPY_GROUP_LLC          0x0010
#define MAC_DISABLE_BRDCST          0x0008
#define MAC_RUN_BIST                0x0004
#define MAC_RX_PARITY               0x0002
#define MAC_NOTE_ALL_FRAMES         0x0001

/*
 * MAC_CNTRL_B defines
 */

#define MAC_RING_PURGE              0x8000
#define MAC_FDX_MODE                0x4000
#define MAC_BRIDGE_STRIP            0x2000
#define MAC_TX_PARITY_ON            0x1000
#define MAC_REPEAT_ONLY             0x0800
#define MAC_LOSE_CLAIM              0x0400
#define MAC_RESET_FIELD             0x0300
#define MAC_FSI_BEACON              0x0080
#define MAC_DELAY_TOKEN             0x0040
#define MAC_IGNORE_SACAM            0x0020

/*
 * MAC_RX_STATUS defines
 */

#define MAC_RC_FSM_STATE            0xE000
#define MAC_R_FLAG                  0x1000
#define MAC_E_FLAG                  0x0800
#define MAC_FSM_STATE               0x0700
#define MAC_N_FLAG                  0x0080
#define MAC_FR_PARS_STATE           0x0070
#define MAC_L_FLAG                  0x0008
#define MAC_H_FLAG                  0x0004
#define MAC_M_FLAG                  0x0002
#define MAC_A_FLAG                  0x0001

/*
 * MAC TX_STATUS defines
 */

#define MAC_TX_FSM_STATE            0xF000
#define MAC_RING_OP                 0x0800
#define MAC_PURGING                 0x0400
#define MAC_FLD_SEQ_STATE           0x03C0
#define MAC_FLD_CNT_STATE           0x0038
#define MAC_LATE_CT                 0x0007

/*
 * Configuration parameters
 */

#define MAC_MAS_CONFIG          0x0000
#define MAC_T_REQ_CONFIG        0xFF3C
#define MAC_T_MAX_CONFIG        0x80E0
#define MAC_CNTRL_A_CONFIG      (MAC_ON|MAC_REVERSE_ADDR)
#define MAC_CNTRL_B_CONFIG      0x0000


/*
 * ELM Control and status registers
 */

#define ELM_CNTRL_A             med_card_base + 0x00800000
#define ELM_CNTRL_B             med_card_base + 0x00800004
#define ELM_INTR_MASK           med_card_base + 0x00800008
#define ELM_XMIT_VECTOR         med_card_base + 0x0080000C
#define ELM_VECTOR_LENGTH       med_card_base + 0x00800010
#define ELM_LE_THRESHOLD        med_card_base + 0x00800014
#define ELM_A_MAX               med_card_base + 0x00800018
#define ELM_LS_MAX              med_card_base + 0x0080001C
#define ELM_TB_MIN              med_card_base + 0x00800020
#define ELM_T_OUT               med_card_base + 0x00800024
#define ELM_LC_SHORT            med_card_base + 0x0080002C
#define ELM_T_SCRUB             med_card_base + 0x00800030
#define ELM_NS_MAX              med_card_base + 0x00800034
#define ELM_TPC_LOAD_VALUE      med_card_base + 0x00800038
#define ELM_TNE_LOAD_VALUE      med_card_base + 0x0080003C
#define ELM_STATUS_A            med_card_base + 0x00800040
#define ELM_STATUS_B            med_card_base + 0x00800044
#define ELM_TPC                 med_card_base + 0x00800048
#define ELM_TNE                 med_card_base + 0x0080004C
#define ELM_CLK_DIV             med_card_base + 0x00800050
#define ELM_BIST_SIGNATURE      med_card_base + 0x00800054
#define ELM_RCV_VECTOR          med_card_base + 0x00800058
#define ELM_INTR_EVENT          med_card_base + 0x0080005C
#define ELM_VIOL_SYN_CTR        med_card_base + 0x00800060
#define ELM_MIN_IDLE_CTR        med_card_base + 0x00800064
#define ELM_LINK_ERR_CTR        med_card_base + 0x00800068

/*
 * Defines for ELM_CNTRL_A
 */

#define ELM_NOISE_TIMER         0x4000
#define ELM_TNE_16BIT           0x2000
#define ELM_TPC_16BIT           0x1000
#define ELM_REQ_SCRUB           0x0800
#define ELM_VSYM_CTR_INTRS      0x0200
#define ELM_MINI_CTR_INTRS      0x0100
#define ELM_FCG_LOOP_CNTRL      0x0080
#define ELM_FOT_OFF             0x0040
#define ELM_EB_LOC_LOOP         0x0020
#define ELM_LOC_LOOP            0x0010
#define ELM_SC_BYPASS           0x0008
#define ELM_SC_REM_LOOP         0x0004
#define ELM_RF_DISABLE          0x0002
#define ELM_RUN_BIST            0x0001

/*
 * Defines for ELM_CNTRL_B
 */

#define ELM_CONFG_CNTRL         0x8000
#define ELM_MATCH_LS            0x7800
#define ELM_MAINT_LS            0x0700
#define ELM_CLASS_S             0x0080
#define ELM_PC_LOOP             0x0060
#define ELM_PC_JOIN             0x0010
#define ELM_LONG                0x0008
#define ELM_PC_MAINT            0x0004
#define ELM_PCM_CNTRL           0x0003

/*
 * Defines for ELM_STATUS_A
 */

#define ELM_SIGNAL_DETECT       0x0400
#define ELM_PREV_LINE_ST        0x0300
#define ELM_LINE_ST             0x00E0
#define ELM_LSM_STATE           0x0010
#define ELM_UNKN_LINE_ST        0x0008
#define ELM_SYM_PR_CTR          0x0007

/*
 * Defines for ELM_STATUS_B
 */

#define ELM_RF_STATE            0xC000
#define ELM_PCI_STATE           0x3000
#define ELM_PCI_SCRUB           0x0800
#define ELM_PCM_STATE           0x0780
#define ELM_PCM_SIGNALLING      0x0040
#define ELM_LSF                 0x0020
#define ELM_RCF                 0x0010
#define ELM_TCF                 0x0008
#define ELM_BREAK_REASON        0x0007

/*
 * Defines for ELM interrupt EVENT
 */

#define ELM_NP_ERR              0x8000
#define ELM_LSD                 0x4000
#define ELM_LE_CTR              0x2000
#define ELM_MINI_CTR            0x1000
#define ELM_VSYM_CTR            0x0800
#define ELM_PHYINV              0x0400
#define ELM_EBUF_ERR            0x0200
#define ELM_TNE_EXPIRED         0x0100
#define ELM_TPC_EXPIRED         0x0080
#define ELM_PCM_ENABLED         0x0040
#define ELM_PCM_BROKE           0x0020
#define ELM_SELF_TEST           0x0010
#define ELM_TRACE_PROP          0x0008
#define ELM_PCM_CODE            0x0004
#define ELM_LS_MATCH            0x0002
#define ELM_PARITY_ERR          0x0001

/*
 * Defines for PCM Control
 */

#define ELM_PCM_CNTRL_NOP       0x0000
#define ELM_PCM_CNTRL_BREAK     0x0001
#define ELM_PCM_CNTRL_TRACE     0x0002
#define ELM_PCM_CNTRL_STOP      0x0003

/*
 * PCM States
 */

#define ELM_PCM_OFF             0x0000
#define ELM_PCM_BREAK           0x0080
#define ELM_PCM_TRACE           0x0100
#define ELM_PCM_CONNECT         0x0180
#define ELM_PCM_NEXT            0x0200
#define ELM_PCM_SIGNAL          0x0280
#define ELM_PCM_JOIN            0x0300
#define ELM_PCM_VERIFY          0x0380
#define ELM_PCM_ACTIVE          0x0400
#define ELM_PCM_MAINT           0x0480
 
/*
 * Configuration parameters for the ELM registers,
 * taken from the ELM data sheet, page 19.
 */

#define ELM_INTR_MASK_CONFIG      0x0000
#define ELM_LE_THRESHOLD_CONFIG   0x0001
#define ELM_A_MAX_CONFIG          0xFFF6
#define ELM_LS_MAX_CONFIG         0xFFFF
#define ELM_TB_MIN_CONFIG         0xFC2F
#define ELM_T_OUT_CONFIG          0xECED
#define ELM_LC_SHORT_CONFIG       0xF676
#define ELM_T_SCRUB_CONFIG        0xFF6D
#define ELM_NS_MAX_CONFIG         0xE796
#define ELM_CNTRL_A_CONFIG        0x0000

/*
 * PC_LOOP codes for ELM LCT
 */

#define ELM_PC_LOOP_NONE          0x0000
#define ELM_PC_LOOP_MAC           0x0020
#define ELM_PC_LOOP_IDLE          0x0040
#define ELM_PC_LOOP_REMOTE        0x0060

/*
 * Station type codes
 */

#define FDDI_STATION_A            0x00
#define FDDI_STATION_B            0x04
#define FDDI_STATION_S            0x02
#define FDDI_STATION_M            0x06

/*
 * Lengths of LCT tests
 */

#define FDDI_LCT_SHORT            0x00
#define FDDI_LCT_MEDIUM           0x02
#define FDDI_LCT_LONG             0x01
#define FDDI_LCT_EXTENDED         0x03

/*
 * MAC for LCT codes
 */

#define FDDI_MAC_LCT_NO           0x00
#define FDDI_MAC_LCT_YES          0x01


/*
 * Initial Board Register Values
 */

#define INIT_RESET_MISC        0xC0
#define INIT_RESET_FIFO        0xFC
#define INIT_RESET_FDDI        0xC0
#define INIT_TEST_FCG          0xC0
#define INIT_TEST_ELM          0x00
#define INIT_TEST_MAC          0x00
#define INIT_INTERRUPT_MASK    0x00
#define INIT_INTERRUPT_EVENT   0x00
#define INIT_MISC_OUTPUT       0xC0
#define INIT_MISC_INPUT        0xe0
#define INIT_STATUS_SYNCH_FIFO 0xA0
#define INIT_STATUS_BUFF_FIFO  0xAA
#define INIT_CHECKSUM_UNIT     0x00

#define MED_MAJOR_NUMBER    66
#define MED_MAJOR(a)	    (a >> 24)
#define MED_HW_PATH(a)	    ((a >> 12) & 0xfff)
#define MED_VSC_MOD_ID(a)   ((a & 0x00f00000) >> 20)
#define MED_SLOT(a)	    ((a & 0x000f0000) >> 16)
#define MED_PROTOCOL(a)     (a & 0xff)

#define MED_INDIVIDUAL_ADDRESS(address)		\
  (((u_char)(address)[0] & 0x01) == 0)

#define MED_BROADCAST_ADDRESS(address)			\
  ( ((u_char)(address)[0] == 0xFF) &&			\
    ((u_char)(address)[1] == 0xFF) &&			\
    ((u_char)(address)[2] == 0xFF) &&			\
    ((u_char)(address)[3] == 0xFF) &&			\
    ((u_char)(address)[4] == 0xFF) &&			\
    ((u_char)(address)[5] == 0xFF) )

#define MED_SAME_ADDRESS(addr1,addr2)			\
  ( ((u_char)(addr1)[5] == (u_char)(addr2)[5]) &&	\
    ((u_char)(addr1)[4] == (u_char)(addr2)[4]) &&	\
    ((u_char)(addr1)[3] == (u_char)(addr2)[3]) &&	\
    ((u_char)(addr1)[2] == (u_char)(addr2)[2]) &&	\
    ((u_char)(addr1)[1] == (u_char)(addr2)[1]) &&	\
    ((u_char)(addr1)[0] == (u_char)(addr2)[0]) )

#define MED_TIMEOUT 20  /* seconds */

#define MED_STI_ID 0x280B31AF	/* Medusa STI ROM graphics ID */
#define MED_MAC_OFFSET 0x3ffe3  /* offset to MAC address in STI ROM */

#define VENOM_SUP_CONTROL 0xfffbd080 /* Venom Supervisor Control address */

/* MEDUSA BUFFERS ------------------------------------------ */

#define NUM_BUFFERS    (128)
#define MED_BUFSIZE (0x2000)	/* Size in bytes of each Medusa buffer */
#define MEDUSA_MCLBYTES (4096)

#ifdef MEDUSA_NINJA_STACK
# define NUM_TX_CLUSTERS (56)   /* used as TX mbuf clusters */
# define NUM_TX_RESERVE  (16)   /* used to TX non-single-copy msgs */
# define NUM_TX_BUFFERS  (NUM_TX_CLUSTERS + NUM_TX_RESERVE)
#else
# define NUM_TX_CLUSTERS (0)
# define NUM_TX_RESERVE  (0)
# define NUM_TX_BUFFERS (56)
#endif
#define NUM_RX_BUFFERS  (NUM_BUFFERS - NUM_TX_BUFFERS)

#define TX_CLUSTER_0 (CARD_MEMORY_BASE)
#define TX_RESERVE_0 (TX_CLUSTER_0 + NUM_TX_CLUSTERS*MED_BUFSIZE)
#define RX_BUFFER_0  (TX_CLUSTER_0 + NUM_TX_TOTAL*MED_BUFSIZE)

#define MED_BUF_ADDR(b)	  ((b) & -(MED_BUFSIZE))
#define MED_BUF_OFFSET(b) ((b) % MED_BUFSIZE)
#define MED_BUF_NUM(b)    (((b)/MED_BUFSIZE) % NUM_BUFFERS)

/* The following macros assume the user has already established that `b' is
   a valid Medusa buffer address */

#define IS_TX_CLUSTER(b) (MED_BUF_NUM(b) < NUM_TX_CLUSTERS)
#define IS_TX_RESERVE(b) ((NUM_TX_CLUSTERS <= MED_BUF_NUM(b)) && (MED_BUF_NUM(b) < (NUM_TX_BUFFERS)))
#define IS_TX_BUFFER(b)  (MED_BUF_NUM(b) < (NUM_TX_BUFFERS))
#define IS_RX_BUFFER(b)  (NUM_TX_BUFFERS <= MED_BUF_NUM(b))

/* --------------------------------------------------------------- */

#define PKT_HDR (0x30780050)	/* Magic at start of FDDI packet */

#ifdef _KERNEL
#define out_word(addr, datum) *((unsigned long *)(addr)) = (unsigned long)(datum)
#define in_word(addr) *(unsigned long *)(addr)
#endif

struct med_ioctl_arg
{
  unsigned reg_value;
  unsigned *location;
};

#define MED_READ         _IOWR ('M', 0, struct med_ioctl_arg)
#define MED_WRITE        _IOW  ('M', 1, struct med_ioctl_arg)
#define MED_MAP          _IO   ('M', 2)
#define MED_VENOM_ENABLE _IO   ('M', 3)
#define MED_BASE_ADDR    _IOR  ('M', 4, struct med_ioctl_arg)

#define MED_MAC_HDR_SIZE (16)
#define MED_SNAP_HDR_SIZE (8)
#define MED_ETHER_HDR_SIZE (14)
#define MED_SPACE_FOR_HDRS (MED_MAC_HDR_SIZE + MED_SNAP_HDR_SIZE)
#define MAC_HEADER_SIZE (10)
  /* A bare-faced lie.  The present implementation uses an ersatz ether_hdr
     internally (in lanc and the upper-level protocols), and a proper IEEE
     802.2 LLC/SNAP header on the fibre.  This is the offset in a medusa
     buffer at which the ether header may be found.
  */

extern unsigned long med_card_base;     /* Base address of card in IO space */

#if defined(MEDUSA_SUPPORT) || defined (MEDUSA_NINJA_STACK)
extern struct ifnet *medusa_ifp;
#endif

#endif /* _MED_H_INCLUDED */
