
#include    <sys/types.h>
#include    <sys/errno.h>
#include    <sys/debug.h>
#include    <sys/stropts.h>
#include    <sys/stream.h>
#include    <sys/strlog.h>
#include    <sys/cmn_err.h>
#include    <sys/kmem.h>
#include    <sys/conf.h>
#include    <sys/ksynch.h>
#include    <sys/stat.h>
#include    <sys/dlpi.h>
#include    <sys/modctl.h>
#ifdef  KSTAT
#include        <sys/kstat.h>
#endif
#include    <sys/time.h>

#include    <sys/ddi.h>
#include    <sys/sunddi.h>


struct myri_packet {
	unsigned short pad;
	unsigned short edst[3];
	unsigned short esrc[3];
	unsigned short etype;

	/* ip header */
	unsigned short version;
	unsigned short totlen;
	unsigned short stuff[2];

	unsigned char ttl;
	unsigned char protocol;
	unsigned short hdrcksum;

	unsigned short addrs[4];


	/* tcp header */
	unsigned short ports[2];
	unsigned int seq;
	unsigned int ack;

	unsigned char hlen:4, other:4;
	unsigned char bits;

	unsigned short window;
	unsigned short tcp_cksum;
	unsigned short urgent;

};

int
verify_cksum(unsigned char *head, mblk_t *m)
{
#ifdef DO_CKSUM
	struct myri_packet *packet = (struct myri_packet *) head;
	unsigned int cksum = 0;
	int length, copy_length;
	int i;
	unsigned short *sptr;
	int first = 1;

	if (packet->protocol != 6) {
		return (1);
	}

	if (packet->version != 0x4500) {
		return (1);
	}

	/* length = total_len - IP_hdr */
	copy_length = length = (packet->totlen & 0xFFFF) - 20;

	if (length & 0x1) {
		copy_length = length + 1;
	}

	sptr = &packet->ports[0];
	for (i = 0; i < copy_length / 2; i++) {
		cksum += *sptr++;
	}

	/* do the psuedo header */
	cksum += packet->addrs[0];
	cksum += packet->addrs[1];
	cksum += packet->addrs[2];
	cksum += packet->addrs[3];
	cksum += packet->protocol;
	cksum += length;

	cksum = ((cksum >> 16) & 0xFFFF) + (cksum & 0xFFFF);
	if (cksum & 0xFFFF0000) {
		cksum++;
	}
	cksum = cksum & 0xFFFF;


	if ((cksum == 0) || (cksum == 0xFFFF)) {
		return (1);
	}
	else {
		cmn_err(CE_CONT, "\nchecksum not verified = 0x%x len = %d",
			cksum, length);
		cmn_err(CE_CONT, "sizeof header = %d\n",
			sizeof(struct myri_packet));
		cmn_err(CE_CONT, "tcp packet cksum field = 0x%x\n",
			packet->tcp_cksum);
		cmn_err(CE_CONT, "totlen = %d  length = %d\n",
			packet->totlen, length);

		sptr = (unsigned short *) head;
		for (i = 0; i < (packet->totlen + 16) / 2; i++) {
			if ((i < 128) || (i + 128) > ((packet->totlen + 16) / 2)) {
				if (first && (((packet->totlen + 16) / 2) >= 128) && (i >= 128)) {
					printf("%x %x %x\n", 0x2a2a, 0x2a2a, 0x2a2a);
					first = 0;
				}

				cmn_err(CE_CONT, "%x ", *sptr);
				if (i && !(i % 7)) {
					cmn_err(CE_CONT, "\n");
				}
			}
			sptr++;
		}
		return (0);
	}
#endif DO_CKSUM
}
