/* myri.c */
/*************************************************************************
 *                                                                       *
 * Myricom Myrinet Software                                              *
 *                                                                       *
 * Copyright (c) 1994 by Myricom, Inc.                                   *
 * All rights reserved.                                                  *
 *                                                                       *
 * This software is released under a non-disclosure agreement with       *
 * Myricom, Inc. Redistribution in any form is strictly prohibited.      *
 *                                                                       *
 * Myricom, Inc. makes no representations about the suitability of this  *
 * software for any purpose.                                             *
 *                                                                       *
 * THIS FILE IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING  *
 * THE WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A           *
 * PARTICULAR PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR     *
 * TRADE PRACTICE. MYRICOM, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO *
 * THE INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY THIS  *
 * FILE OR ANY PART THEREOF.                                             *
 *                                                                       *
 * In no event will Myricom, Inc. be liable for any lost revenue         *
 * or profits or other special, indirect and consequential damages, even *
 * if Myricom has been advised of the possibility of such damages.       *
 *                                                                       *
 *                                                                       *
 * Other copyrights might apply to parts of this software and are so     *
 * noted when applicable.                                                *
 *                                                                       *
 * Myricom, Inc.                                                         *
 * 325 N. Santa Anita Ave.                                               *
 * Arcadia, CA 91024                                                     *
 *                                                                       *
 *************************************************************************/

#include    <sys/types.h>
#include    <sys/errno.h>
#include    <sys/debug.h>
#include    <sys/stropts.h>
#include    <sys/stream.h>
#include    <sys/strlog.h>
#include    <sys/cmn_err.h>
#include    <sys/kmem.h>
#include    <sys/conf.h>
#include    <sys/ksynch.h>
#include    <sys/stat.h>
#include    <sys/dlpi.h>
#include    <sys/modctl.h>
#ifdef  KSTAT
#include        <sys/kstat.h>
#endif
#include    <sys/time.h>
/*
#include    <string.h>
*/
#include    <sys/ddi.h>
#include    <sys/sunddi.h>

#include "config.h"
#include "myri_api.h"
#include "myri.h"
#ifdef RELATIVE
#include "myri_code_rel.h"
#else
#include "myri_code.h"
#endif

ddi_dma_lim_t myri_dma_lim = {
	0x0,			/* lower */
	0xFFFFFFFF,		/* upper */
	0xFFFFFFFF,		/* counter */
	0x7F,			/* default burst size should be 0x7C? */
	4,			/* min transfer size in bytes */
	56320,			/* 55 MBytes/sec */
};

/* this needs to be defined for host_ifc.c */
u_short *LANAI[MYRI_MAX_BOARDS] = {0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0};
u_short *LANAI_WAKE[MYRI_MAX_BOARDS] = {0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0};
u_short *BOARD[MYRI_MAX_BOARDS] = {0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0};
extern struct myri *myridev;

extern int myridebug;

extern struct timeval time_start[32], time_stop[32];
extern int send_immediate, had_to_copy;

long long buffer[(10*1024)/sizeof(long long)];

/******************   myriLoadLCPkernel()  *************************/
/*
 * Load a LANai control program into the LANai chip. The program was
 * included in this file by "#include myri_code.h".
 *
 */

int
myriLoadLCPkernel(struct myri * myrip)
{
	int rv = 0;
	int unit = ddi_get_instance(myrip->myri_dip);


	CMN_ERR(0) (CE_CONT, "\nmyriLoadLCPkernel()\n");

	/* turn interrupts off and hold the lanai in reset */
#ifdef i86pc
	AMCC_INT_OFF(myrip->myri_regsp);
	AMCC_RESET_ON(myrip->myri_regsp);
#else
	myrip->myri_regsp->int_off = 1;
	myrip->myri_regsp->rst_on = 1;
#endif

	/* For LANai 3.0 */
	LANAI[unit][0x1007E] = 0xEFFF;

#ifdef i86pc
	CMN_ERR(0) (CE_CONT, "NOT loading LANai\n");
/*
	bzero((caddr_t) & myrip->lanai_memory[0], LANAI_SIZE);

	bcopy((caddr_t) & lanai_code[0],
	      (caddr_t) & myrip->lanai_memory[(int) lanai_code_off],
	      sizeof(lanai_code));

	bcopy((caddr_t) & lanai_data[0],
	      (caddr_t) & myrip->lanai_memory[lanai_data_off],
	      sizeof(lanai_data));
*/
#else
	bzero((caddr_t) & myrip->myri_regsp->lanai_memory[0], LANAI_SIZE);

	bcopy((caddr_t) & lanai_code[0],
	  (caddr_t) & myrip->myri_regsp->lanai_memory[(int) lanai_code_off],
	      sizeof(lanai_code));

	bcopy((caddr_t) & lanai_data[0],
	      (caddr_t) & myrip->myri_regsp->lanai_memory[lanai_data_off],
	      sizeof(lanai_data));
#endif


	CMN_ERR(1) (CE_CONT, "Done copying LANai code\n");

	/* Turn off reset */
	/* Turn off interrupts to be sure not to have a race condition */
#ifdef i86pc
	AMCC_INT_OFF(myrip->myri_regsp);
	AMCC_RESET_OFF(myrip->myri_regsp);
#else
	myrip->myri_regsp->int_off = 1;
	myrip->myri_regsp->rst_off = 1;
#endif

	{
		register int count = 5000;

		while (count && !myriWasReset(unit, KERNEL_INTERFACE)) {
			count--;
		}
	}

	rv = myri_lanai_handshake(myrip);

#ifdef i86pc
	AMCC_INT_ON(myrip->myri_regsp);
#else
	myrip->myri_regsp->int_on = 1;
#endif

	CMN_ERR(0) (CE_CONT, "Finished with myriLoadLCPkernel()  rv = %d\n\n", rv);

	return (rv);
}



/**********************   myri_lanai_handshake()  *************/
/*
 * Do the handshake with the LANai after a reset.
 */
int
myri_lanai_handshake(struct myri * myrip)
{
	int count = 0;
	int unit = ddi_get_instance(myrip->myri_dip);

	if (!myriWasReset(unit, KERNEL_INTERFACE)) {
		CMN_ERR(0) (CE_CONT, "myri_lanai_handshake: LANai trigger not set!\n");
		return (0);
	}

#ifdef i86pc
	AMCC_INT_OFF(myrip->myri_regsp);
#else
	myrip->myri_regsp->int_off = 1;
#endif

	while (++count <= 25) {
		if (myriHandshake(unit, KERNEL_INTERFACE)) {
			CMN_ERR(RESET_PRINT) (CE_CONT, "myri_lanai_handshake: LANai reset OK - %d\n",
					      count);
#ifdef i86pc
			AMCC_INT_ON(myrip->myri_regsp);
#else
			myrip->myri_regsp->int_on = 1;
#endif
			return (1);
		}
	}

	CMN_ERR(0) (CE_CONT, "myri_lanai_handshake: LANai[%d] NO handshake!\n",
		    unit);

#ifdef i86pc
	AMCC_INT_ON(myrip->myri_regsp);
#else
	myrip->myri_regsp->int_on = 1;
#endif
	return (0);
}


/****************************  hex_dump()  ********************/
void
hex_dump(u_char * cptr, int len)
{
	int i;

	if (len > 100)
		len = 100;

	for (i = 1; i <= len; i++) {
		CMN_ERR(0) (CE_CONT, "%x ", *cptr++);
		if (!(i % 16))
			CMN_ERR(0) (CE_CONT, "\n");
	}
	if ((i % 16))
		CMN_ERR(0) (CE_CONT, "\n");
}


/***********************  myri_tolanai()  *****************************/

int
myri_tolanai(queue_t * wq, mblk_t * mp, struct myri * myrip)
{
	int len;
	struct medium_addr myri_dest;
	int unit = ddi_get_instance(myrip->myri_dip);
	int rv = 1;
	int retval;
	int tries = MYRI_MAX_SEND_TRIES;


	CMN_ERR(3) (CE_CONT, "        myri_tolanai() called\n");

	if (!mp) {
		CMN_ERR(0) (CE_CONT, "  mp is NULL\n");
		return (0);
	}

	GET_TIME(time_stop[2]);
	MUTEX_ENTER(&myrip->myri_xmitlock);
	ASSERT(++myrip->xmit_count == 1);
	GET_TIME(time_stop[3]);

	/*
	 * pick off the myrinet destination address from the head of the
	 * message
	 */
	bcopy((caddr_t) mp->b_rptr, (caddr_t) & myri_dest,
	      sizeof(struct medium_addr));


	/* FIX hack for MULTICAST */
	/* Turn all Multicast packets into Myrinet broadcast packets */
	if (myri_dest.medaddr[0] & 0x1) {
		/* multicast or broadcast packet */
		myri_dest.medaddr[5] = 0xFF;
	}


	/* send the one message pointed to by mp */

	len = msgdsize(mp);

	/* save space for the alignment padding on the front */
	len += MYRI_PAD_LEN;

	if (len & 0x3) {
		/* make length a multiple of 4 */
		len = (len + 4) & 0xFFFFFFFC;
	}

	if (len <= MYRI_IMMED_LEN) {

		/* hand to LANai */

		send_immediate = 1;

		GET_TIME(time_stop[4]);
		if ((retval = myriAddSendImmediate(unit, KERNEL_INTERFACE,
						   KERNEL_INTERFACE,
						   myri_dest.medaddr[5],
						   len,
						   0,	/* ignore send_ack */
						   RECV0_Q,
						   mp,
						copy_mblk_to_buffer) > 0)) {

			myrip->myri_opackets++;
			rv = 1;
		}
		else if (retval == 0) {
			CMN_ERR(1) (CE_CONT, "myriAddSendImm failed(%d) len = %d\n",
				    retval, len);
			/*
			 * not enough send space for immediate try to send
			 * using DMA
			 */
			goto send_with_dma;
		}
		else if (retval < 0) {
			/*
			 * should do something different since this indicates
			 * reset
			 */
			CMN_ERR(0) (CE_CONT, "myriAddSendImm failed (%d) (reset?) len = %d\n",
				    retval, len);

			goto send_error;
		}

		GET_TIME(time_stop[5]);

		/* free chain */
		freemsg(mp);
		mp = NULL;

		/* check the send_ack queue */
		myri_check_send_ack(myrip);
		myrip->xmit_count--;
		MUTEX_EXIT(&myrip->myri_xmitlock);

		if (!rv) {
			/*
			 * FIX- ?(rv) or (!rv) check for stuff that got
			 * blocked
			 */
			if (myrip->myri_wantw) {
				myriwenable(myrip);
			}
		}
		return (rv);
	}			/* send immediate */



send_with_dma:
	/* else send using DMA */
	{
		register int index;
		int save_index;

		send_immediate = 0;

		save_index = myrip->last_send_buff;

		/* start with the next one */
		index = CYCLE(save_index, NUM_SEND_BUFF);

		/* find a send buffer */
		while (myrip->sendinfo[index].status != BUF_UNUSED) {
			index = CYCLE(index, NUM_SEND_BUFF);
			if (index == save_index) {
				break;
			}
		}

		if (myrip->sendinfo[index].status == BUF_UNUSED) {
			/* found a buffer */

			myrip->sendinfo[index].status = BUF_LANAI;
			myrip->last_send_buff = index;

			/* copy the chain to the buffer */
			copy_mblk_to_buffer(mp,
			     (unsigned short *) myrip->sendinfo[index].kptr,
					    len);

			/* hand it to the LANai */

			{
				Scatter scatter;
				Gather gather;

				myriFillScatter(scatter, KERNEL_INTERFACE, len,
				    (len <= RECV0_LEN) ? RECV0_Q : RECV1_Q);

				gather.pointer = (unsigned char *) myrip->sendinfo[index].dptr;
				gather.lengthInBytes = len;

				/* FIX - SYNC??? */
				GET_TIME(time_stop[14]);
				if (ddi_dma_sync(myrip->sendinfo[index].handle,
						 0, len, DDI_DMA_SYNC_FORDEV) != DDI_SUCCESS) {
					CMN_ERR(0) (CE_CONT, "DDI_DMA_SYNC_FORDEV index = %d len = %d\n",
						    index, len);
				}
				GET_TIME(time_stop[15]);
				GET_TIME(time_stop[4]);

				while (tries--) {
					if ((retval = myriAddSend(unit, KERNEL_INTERFACE,
						       myri_dest.medaddr[5],
								  len,
								  NO_CKSUM,
								  NO_CKSUM,
								  1,
								  &gather,
								  1,
								  &scatter,
								  (unsigned) &myrip->sendinfo[index])) > 0) {

						myrip->myri_opackets++;
						rv = 1;
						myrip->last_send_buff = index;
						break; /* break from while tries */
					}
					else if (retval == 0) {
						if (tries <= 0) {
							CMN_ERR(0) (CE_CONT, "myriAddSend failed(%d) len = %d  tries=%d\n",
								    retval, len, MYRI_MAX_SEND_TRIES);
							rv = 0;
							myrip->sendinfo[index].status = BUF_UNUSED;
							goto send_error;
						}
					}
					else if (retval < 0) {
						/*
						 * should do something
						 * different since this
						 * indicates reset
						 */
						CMN_ERR(0) (CE_CONT, "myriAddSend failed (%d) (reset?) len = %d\n",
							    retval, len);
						myrip->sendinfo[index].status = BUF_UNUSED;

						goto send_error;
					}
				}	/* while */

				GET_TIME(time_stop[5]);

				/* free chain */
				freemsg(mp);
				mp = NULL;

				/* check the send_ack queue */
				myri_check_send_ack(myrip);
				myrip->xmit_count--;
				MUTEX_EXIT(&myrip->myri_xmitlock);

				/* check for stuff that got blocked */
				if (!rv) {
					/*
					 * FIX - ?? (rv) or (!rv) in test
					 * here?
					 */
					if (myrip->myri_wantw)
						myriwenable(myrip);
				}

				return (rv);

			}	/* hand to lanai */
		}		/* if found a buffer */

		else {
			/* no send buffers free! */
			CMN_ERR(0) (CE_CONT, "myri_tolanai: no send buffers free!\n");
			myrip->myri_oerrors++;
			/* free chain */
			freemsg(mp);
			mp = NULL;

			/* check the send_ack queue */
			myri_check_send_ack(myrip);
			myrip->xmit_count--;
			MUTEX_EXIT(&myrip->myri_xmitlock);

			return (0);
		}
	}			/* send using DMA */


send_error:
	myrip->myri_oerrors++;
	if (mp) {
		freemsg(mp);
	}
	mp = NULL;

	myrip->xmit_count--;
	MUTEX_EXIT(&myrip->myri_xmitlock);
	CMN_ERR(RESET_PRINT) (CE_CONT, "send_error(): myriWASreset\n");
	myri_lanai_error(myrip);

	return (0);
}


/******************   myri_check_send_ack()  ********************/
/*
 * Check the send_ack queue. This is called while
 * myri_xmitlock is already held.
 *
 */

void
myri_check_send_ack(struct myri * myrip)
{
	unsigned id;
	int retval;
	register int index;
	register struct buff_info *infop;
	int unit = ddi_get_instance(myrip->myri_dip);


	while ((retval = myriGetSendAck(unit, KERNEL_INTERFACE, &id)) > 0) {

		CMN_ERR(2) (CE_CONT, "myriGetSendAck - 0x%x\n", id);

		if (!id) {
			/*
			 * "0" for id is from a send_immediate command
			 */
			continue;
		}

		infop = (struct buff_info *) id;
		index = (infop - &myrip->sendinfo[0]);
		if ((index >= 0) && (index < NUM_SEND_BUFF)) {
			/* pointer is in the proper range */
			infop->status = BUF_UNUSED;
			CMN_ERR(2) (CE_CONT, "myriGetSendAck - freeing send_buff[%d]\n", index);
		}
		else {
			CMN_ERR(0) (CE_CONT, "SEND_ACK free buff 0x%x  index = %d out of range\n",
				    id, index);

			myrip->xmit_count--;
			MUTEX_EXIT(&myrip->myri_xmitlock);
			myri_lanai_error(myrip);
			MUTEX_ENTER(&myrip->myri_xmitlock);
			ASSERT(++myrip->xmit_count == 1);
			return;
		}
	}			/* while retval>0 */

	if (retval < 0) {
		/* LANai was reset */
	}
}


/******************   myri_mutex_enter()   *********************/
/*
 * Grab all the semaphores.
 *
 */
void
myri_mutex_enter(struct myri * myrip)
{
	MUTEX_ENTER(0);
	MUTEX_ENTER(&myrip->myri_xmitlock);
	ASSERT(++myrip->xmit_count == 1);
	MUTEX_ENTER(&myrip->myri_recvlock);
	ASSERT(++myrip->recv_count == 1);
/*
	MUTEX_ENTER(&myrip->myri_bufflock);
*/
}


/******************   myri_mutex_exit()   *********************/
/*
 * Release all the semaphores.
 *
 */
void
myri_mutex_exit(struct myri * myrip)
{
/*
	MUTEX_EXIT(&myrip->myri_bufflock);
*/
	myrip->recv_count--;
	MUTEX_EXIT(&myrip->myri_recvlock);
	myrip->xmit_count--;
	MUTEX_EXIT(&myrip->myri_xmitlock);
	MUTEX_EXIT(0);
}

/******************   myri_lanai_error()  ********************/
/*
 * Called whenever we get some strange error condition in the LANai.
 * If we see too many errors, then we'll try to reload the LANai.
 * We'll only try to reload the LANai a small number of times
 * before we declare it dead.
 * MUST be called with NO mutex held.
 *
 */
void
myri_lanai_error(struct myri * myrip)
{
	int unit = ddi_get_instance(myrip->myri_dip);

	CMN_ERR(1) (CE_CONT, "         myri_lanai_error\n");

	myri_mutex_enter(myrip);

	if (myriWasReset(unit, KERNEL_INTERFACE)) {
		if (myri_lanai_handshake(myrip)) {
			myrip->lanai_errors = 0;
			myrip->lanai_reload = 0;
			myri_init_lanai(myrip);
			myri_mutex_exit(myrip);
			CMN_ERR(RESET_PRINT) (CE_CONT, "myri_lanai_error(): Completed handshake\n");
			return;
		}
		else {
			CMN_ERR(0) (CE_CONT, "myri_lanai_error(): LANai wouldn't reset\n");
			myrip->lanai_errors = MYRI_MAX_LANAI_ERRORS;
		}
	}
	/* LANai got error without a reset or didn't reset */
	if (++myrip->lanai_errors > MYRI_MAX_LANAI_ERRORS) {
		/* try to reload */
		while (myrip->lanai_reload++ <= MYRI_MAX_LANAI_RELOAD) {
#ifdef RELOAD_IF_ERROR
			if (myriLoadLCPkernel(myrip)) {
				/*
				 * reloaded and reset, need to initialize
				 */
				myrip->lanai_errors = 0;
				myrip->lanai_reload = 0;
				myri_init_lanai(myrip);
				myri_mutex_exit(myrip);
				return;
			}
#endif				/* RELOAD_IF_ERROR */
		}
		CMN_ERR(0) (CE_CONT, "lanai_error: couldn't reload LANai!!\n");
	}
	myri_mutex_exit(myrip);
}



/***********************  myri_dma_error()   *****************************/
void
myri_dma_error(int rv)
{
	switch (rv) {
			case DDI_DMA_MAPPED:
			cmn_err(CE_CONT, "dma_setup DDI_DMA_MAPPED - OK\n");
			break;

		case DDI_DMA_TOOBIG:
			cmn_err(CE_CONT, "dma_setup DDI_DMA_TOOBIG\n");
			break;

		case DDI_DMA_NOMAPPING:
			cmn_err(CE_CONT, "dma_setup DDI_DMA_NOMAPPING\n");
			break;

		case DDI_DMA_NORESOURCES:
			cmn_err(CE_CONT, "dma_setup DDI_DMA_NORESOURCES\n");
			break;

		case DDI_DMA_PARTIAL_MAP:
			cmn_err(CE_CONT, "dma_setup DDI_DMA_PARTIAL_MAP\n");
			break;

		default:
			cmn_err(CE_CONT, "dma_setup rv=%d not understood!\n", rv);
			break;
	}			/* switch */
}








/*********************  myri_reset_lanai()  ****************/
/*
 * Reset the LANai chip.
 */

int
myri_reset_lanai(struct myri * myrip)
{
	int rv;

	/* turn interrupts off and hold the lanai in reset */
#ifdef i86pc
	AMCC_INT_OFF(myrip->myri_regsp);
	AMCC_RESET_ON(myrip->myri_regsp);
	AMCC_RESET_OFF(myrip->myri_regsp);
#else
	myrip->myri_regsp->int_off = 1;
	myrip->myri_regsp->rst_on = 1;
	myrip->myri_regsp->rst_off = 1;
#endif

	rv = myri_lanai_handshake(myrip);

#ifdef i86pc
	AMCC_INT_ON(myrip->myri_regsp);
#else
	myrip->myri_regsp->int_on = 1;
#endif

	return (rv);
}

/*********************   myri_init_lanai() ********************/
/*
 * Initialize the LANai after a reset or reload.
 * Should be holding all the mutex vars here.
 *
 */
void
myri_init_lanai(struct myri * myrip)
{
	int unit = ddi_get_instance(myrip->myri_dip);
	struct buff_info binfo;


	CMN_ERR(3) (CE_CONT, "\nmyri_init_lanai()\n");


	/* free the buffers held by the LANai */
	myri_free_buff_all(myrip);

	if (!myrip->sts) {
		myrip->sts = myri_what_sts(myrip);
	}

	/* FIX should we do this if the interface is not up? */
	if (myrip->myri_ouraddr.medaddr[5]) {

		if (myriSetHostInfo(unit, KERNEL_INTERFACE,
				    myrip->myri_ouraddr.medaddr[5],
				    myrip->sts, EthernetPacketType,
				    MYRI_INT_DELAY, MYRI_INT_MASK)) {
			CMN_ERR(1) (CE_CONT, "myri_init_lanai: Setting Myrinet address ==> 0x%x\n",
				    myrip->myri_ouraddr.medaddr[5]);
		}
		else {
			CMN_ERR(0) (CE_CONT, "init_lanai: host_info NOT accepted\n");
			return;
		}

		binfo.myrip = myrip;
		binfo.index = -1;
		binfo.queue = RECV0_Q;
		myri_add_buff(&binfo);

		binfo.myrip = myrip;
		binfo.index = -1;
		binfo.queue = RECV1_Q;
		myri_add_buff(&binfo);

	}			/* interface has ether addr */
	else {
		CMN_ERR(3) (CE_CONT, "myri_init_lanai: No ether address yet\n");
	}

	CMN_ERR(3) (CE_CONT, "Exiting myri_init_lanai()\n\n");
}






/*********************   myri_what_sts()   ************************/
/*
 * Try to discover what STS we should use for the Sbus LANai board.
 *
 */

unsigned short
myri_what_sts(struct myri * myrip)
{
	int burst_sizes;
	unsigned short sts = 0;
	int rv;

	/* do a fake addr_setup to get a handle */

	rv = ddi_dma_addr_setup(myrip->myri_dip, NULL,
				(caddr_t) myrip, 16,
				(DDI_DMA_WRITE | DDI_DMA_READ),
				DDI_DMA_SLEEP, NULL,
				&myri_dma_lim, &myrip->send_handle);

	if ((rv == DDI_DMA_MAPPED) || (rv == DDI_DMA_PARTIAL_MAP)) {
		burst_sizes = ddi_dma_burstsizes(myrip->send_handle);
		cmn_err(CE_CONT, "sbus 'burst-sizes' (bytes)= \n");
		cmn_err(CE_CONT, "  %b\n",
			burst_sizes,
			"\20\7burst64\6burst32\5burst16\4burst8\3burst4\2burst2\1burst1");
		sts = (burst_sizes & 0xF8) >> 3;
		cmn_err(CE_CONT, "myri_sts = %b\n", sts,
		    "\20\4burst16word\3burst8word\2burst4word\1burst2word");

#ifdef NOT_NEEDED
		/*
		 * FIX hack to handle stupid controllers that don't support
		 * all bursts less than the maximum
		 */
		if (!((sts == 0xf) ||
		      (sts == 0x7) ||
		      (sts == 0x3) ||
		      (sts == 0x1) ||
		      (sts == 0x0))) {
			sts = 0;
			cmn_err(CE_CONT, "Brain-dead SBus controller:\n");
			cmn_err(CE_CONT, "       doesn't support all sizes less than the max.\n");
		}
#endif

		ddi_dma_free((ddi_dma_handle_t) myrip->send_handle);
	}
	else {
		cmn_err(CE_CONT, "what_sts couldn't do an addr setup\n");
		myri_dma_error(rv);
	}

	return (sts);
}


/*****************   myri_handle_lanai_intr()  ************/
/*
 * Called from myriintr.
 *
 */
int
myri_handle_lanai_intr(struct myri * myrip, int count)
{
	mblk_t *mp;
	int unit = ddi_get_instance(myrip->myri_dip);
	int retval = 0;

#ifdef i86pc
	AMCC_INT_OFF(myrip->myri_regsp);
#else
	myrip->myri_regsp->int_off = 1;
#endif

	if (myriWasReset(unit, KERNEL_INTERFACE)) {
		CMN_ERR(RESET_PRINT) (CE_CONT, "myri_handle_lanai_intr(): myriWASreset\n");
		myri_lanai_error(myrip);
	}

	/* look in the recv_ack queue and l2h queue */
	/* FIX - mutex once around the while loop, or each time? */

	MUTEX_ENTER(&myrip->myri_recvlock);
	ASSERT(++myrip->recv_count == 1);

	while (--count) {
		int save_len;

		GET_TIME(time_start[16]);

		mp = myri_do_recv(myrip);

		GET_TIME(time_stop[20]);

		if (mp) {
			/*
			 * function for handing the packet to the higher
			 * layers
			 */

			save_len = msgdsize(mp);

			myrisendup(myrip, mp, myriaccept);
			GET_TIME(time_stop[21]);
			retval = 1;
		}
		else {
			break;
		}

#ifdef DO_TIMING

		SUB_TIME(time_stop[17], time_start[16]);
		SUB_TIME(time_stop[18], time_start[16]);
		SUB_TIME(time_stop[19], time_start[16]);
		SUB_TIME(time_stop[20], time_start[16]);
		SUB_TIME(time_stop[21], time_start[16]);


		CMN_ERR(0) (CE_CONT, "Recv time: %d  %d  %d  %d  %d (len=%d) %s\n",
			    time_stop[17].tv_usec,
			    time_stop[18].tv_usec,
			    time_stop[19].tv_usec,
			    time_stop[20].tv_usec,
			    time_stop[21].tv_usec,
			    save_len,
			    (had_to_copy) ? "COPY" : "");

#endif				/* DO_TIMING */

	}			/* while */

	retval += myri_do_l2h(myrip);

	myrip->recv_count--;
	MUTEX_EXIT(&myrip->myri_recvlock);

#ifdef i86pc
	AMCC_INT_ON(myrip->myri_regsp);
#else
	myrip->myri_regsp->int_on = 1;
#endif

	return (retval);
}


/***************** myri_do_recv()  **********************/
/*
 * Check the recv_ack queue.
 * called while already holding the recv_lock mutex
 * ASSERT(mutex_owned(&myrip->recvlock));
 *
 */

mblk_t *
myri_do_recv(struct myri * myrip)
{
	mblk_t *m = NULL;
	int index;
	unsigned id;
	u_short length;
	u_short cksum = 0;
	int tail = -1;
	unsigned type;
	u_short queue;
	int unit = ddi_get_instance(myrip->myri_dip);
	int retval;
	struct buff_info *infop;
	int max;

	ASSERT(mutex_owned(&myrip->myri_recvlock));

	if ((retval = myriGetReceiveAck(unit, KERNEL_INTERFACE,
				&id, &length, &cksum, &type, &queue)) > 0) {
		CMN_ERR(2) (CE_CONT, "myri_do_recv()\n");
		CMN_ERR(2) (CE_CONT, "       rv=%d  id=0x%x len=%d cksum=0x%x type=%d queue=%d\n",
			    retval, id, length, cksum, type, queue);

		CMN_ERR(BUFF_PRINT) (CE_CONT, "recv_ack (queue%d,%x) len =%d \n",
				     queue, id, length);

		switch (type) {
			case ReceiveEndType:
				tail = 1;
				break;

			case ReceiveBodyFlushType:
			case ReceiveEndFlushType:
			case ReceiveBodyType:
			default:
				CMN_ERR(0) (CE_CONT, "recv_ack: bad type returned = %d\n", type);
				CMN_ERR(0) (CE_CONT, "          id = 0x%x len=%d cksum=0x%x type=%d queue=%d\n",
					    id, length, cksum, type, queue);
				goto recv_error;
		}


		CMN_ERR(3) (CE_CONT, "\tgot recv_ack\n");


		infop = (struct buff_info *) id;
		if (queue == RECV0_Q) {
			index = infop - &myrip->recv0info[0];
			max = NUM_RECV_BUFF0;
		}
		else if (queue == RECV1_Q) {
			index = infop - &myrip->recv1info[0];
			max = NUM_RECV_BUFF1;
		}
		else {
			CMN_ERR(0) (CE_CONT, "recv_ack: bad queue = %d\n", queue);
			goto recv_error;
		}


		if ((index >= 0) && (index < max) &&
				(infop->status == BUF_LANAI)) {

			/* FIX - SYNC?? */
			GET_TIME(time_stop[17]);
			if (ddi_dma_sync(infop->handle,
			0, length, DDI_DMA_SYNC_FORKERNEL) != DDI_SUCCESS) {
				CMN_ERR(0) (CE_CONT, "DDI_DMA_SYNC_FORKERNEL failed index = %d len = %d\n",
					    infop->index, length);
			}
			GET_TIME(time_stop[18]);

			/* loan it out if possible */

			myrip->buff_at_lanai[infop->queue]--;

#ifdef NO_LOAN
			if (TRUE) {
#else
			if (myrip->buff_at_lanai[infop->queue] <= MYRI_LBUF_MIN) {
#endif				/* NO_LOAN */

				had_to_copy = 1;

				/*
				 * not enough buffers at the LANai, will copy
				 * msg
				 */
				LANAI[unit][COPIED_BUFF_LOC]++;
				m = allocb(length, BPRI_HI);
				if (m) {
					bcopy((caddr_t) infop->kptr, (caddr_t) m->b_rptr, length);
					m->b_wptr = (u_char *) ((unsigned) m->b_rptr +
							 (unsigned) length);
					m->b_datap->db_type = M_DATA;

					infop->status = BUF_UNUSED;
					myri_add_buff(infop);
					{
						/* now try to find more buffs */
						struct buff_info binfo;
						bcopy((caddr_t) infop, (caddr_t) & binfo,
						  sizeof(struct buff_info));
						binfo.index = -1;
						myri_add_buff(&binfo);
					}
				}
				else {
					CMN_ERR(0) (CE_CONT, "do_recv: failed allocb\n");
					infop->status = BUF_UNUSED;
					myri_add_buff(infop);
					myrip->lanai_errors = 0;
					goto recv_error;
				}
			}
			else {
				had_to_copy = 0;
				/*
				 * lanai has enough buffers, will loan this
				 * one to higher layers
				 */
				/* this avoids a copy */
				if ((m = esballoc((unsigned char *) infop->kptr,
				 (queue == RECV0_Q) ? RECV0_LEN : RECV1_LEN,
						  BPRI_HI,
						  &infop->free_func))) {

					m->b_wptr = (u_char *) ((unsigned) m->b_rptr +
							 (unsigned) length);
					m->b_datap->db_type = M_DATA;
					infop->status = BUF_LOANED;
				}
				else {
					CMN_ERR(0) (CE_CONT, "do_recv: failed esballoc\n");
					infop->status = BUF_UNUSED;
					myri_add_buff(infop);
					myrip->lanai_errors = 0;
					goto recv_error;
				}
			}	/* else lanai count OK */

		}
		else {
			if ((index >= 0) && (index < max)) {
				CMN_ERR(0) (CE_CONT, "do_recv: bad STATUS id = 0x%x idx=%d %s\n",
				 id, index, BUFFINFO_STATUS(infop->status));
			}
			else {
				CMN_ERR(0) (CE_CONT, "do_recv: bad INDEX  id = 0x%x  index = %d \n",
					    id, index);
			}

#if (BUFF_PRINT<=PRINT_LEVEL)
			if (queue == 0) {
				myri_dump_buff_info(&myrip->recv0info[0]);
			}
			else if (queue == 1) {
				myri_dump_buff_info(&myrip->recv1info[0]);
			}
#endif				/* BUFF_PRINT */
#ifndef i86pc
			myrip->myri_regsp->rst_on = 1;
			myrip->myri_regsp->rst_off = 1;
#endif				/* i86pc */

			goto recv_error;
		}

		GET_TIME(time_stop[19]);

		m->b_cont = NULL;

	}			/* if retval > 0 */

	else if (retval < 0) {
		goto recv_error;
	}

	else if ((retval == 0) || (!m)) {
		return (NULL);
	}

	myrip->myri_ipackets++;

	/* remove the initial padding */
	/* 1st byte indicates the number of bytes to skip */
	{
		register u_char skip = *(m->b_rptr);

		if ((skip > 0) && (skip <= MYRI_MAX_PAD_LEN)) {
			m->b_rptr += skip;
		}
		else {
			CMN_ERR(0) (CE_CONT, "Bad myri_pad value = %d\n", skip);
			myri_dump_msg(m);
			freemsg(m);
			m = NULL;
			return (NULL);
		}
	}

	/* m->b_rptr should now be pointing at the ether header */

	return (m);


recv_error:
	myrip->myri_ierrors++;

	myrip->recv_count--;
	MUTEX_EXIT(&myrip->myri_recvlock);
	CMN_ERR(RESET_PRINT) (CE_CONT, "recv_error(): myriWASreset\n");
	myri_lanai_error(myrip);
	MUTEX_ENTER(&myrip->myri_recvlock);
	ASSERT(++myrip->recv_count == 1);

	if (m) {
		freemsg(m);
	}

	CMN_ERR(1) (CE_CONT, "Got to recv_error.\n");
	return (NULL);
}

/*****************   myri_do_l2h()  *********************/
/*
 * Check the l2h queue and respond.
 *
 */

int
myri_do_l2h(struct myri * myrip)
{
	int retval;
	unsigned cmd = 0, arg;
	int unit = ddi_get_instance(myrip->myri_dip);
	mblk_t *mp;
	int count;
	int dump[2] = {0, 0};


	if ((retval = myriCheckLanaiToHost(unit, KERNEL_INTERFACE, &cmd, &arg)) > 0) {
		CMN_ERR(1) (CE_CONT, "myriCheckLanaiToHost got cmd = %d  arg = %d\n", cmd, arg);

		switch (cmd) {

			case ReceiveQueueEmpty:
				{
					struct buff_info binfo;

					binfo.myrip = myrip;
					binfo.index = -1;

					CMN_ERR(1) (CE_CONT, "RECV_Q_EMPTY: No more host rcv buffs q=%d\n",
						    arg);

					switch (arg) {
						case RECV0_Q:
						case RECV1_Q:

							count = RECV_FEW_MSGS;
							while (--count) {
								mp = myri_do_recv(myrip);

								if (mp) {
									myrisendup(myrip, mp, myriaccept);
								}
								else {
									break;
								}
							}
							binfo.queue = arg;
							myri_add_buff(&binfo);
							dump[arg] = 1;
							break;

						default:
							CMN_ERR(0) (CE_CONT, "myriCheckLanaiToHost got cmd = %d  arg = %d\n",
								  cmd, arg);
							CMN_ERR(0) (CE_CONT, "RECVQ Arg not understood?? = %d\n", arg);
							myrip->recv_count--;
							MUTEX_EXIT(&myrip->myri_recvlock);
							myri_lanai_error(myrip);
							MUTEX_ENTER(&myrip->myri_recvlock);
							ASSERT(++myrip->recv_count == 1);
							break;
					}	/* switch arg */
				}	/* case q empty */

				break;

			case SendAckQueueFull:
				CMN_ERR(1) (CE_CONT, "SEND_ACK_Q_FULL: SEND_ACK_Q full\n");
				break;

			case ReceiveAckQueueFull:
				CMN_ERR(1) (CE_CONT, "RECV_ACK_Q_FULL: RECV_ACK_Q full\n");
				mp = myri_do_recv(myrip);

				if (mp) {
					/*
					 * function for handing the packet to
					 * the higher layers
					 */
					myrisendup(myrip, mp, myriaccept);
				}

				break;

			default:
				CMN_ERR(0) (CE_CONT, "cmd NOT UNDERSTOOD = %d\n",
					    cmd);
				break;

		}		/* switch */
	}			/* if l2h */

	if (retval < 0) {
		/* LANai was reset */
		return (0);
	}

	if (PRINT_LEVEL >= 1) {
		if (dump[RECV0_Q])
			myri_dump_buff_info(&myrip->recv0info[0]);
		if (dump[RECV1_Q])
			myri_dump_buff_info(&myrip->recv1info[0]);
	}


	return (retval);
}


/**********************   lanai_set_ex1()  *********************/
/*
 * Set the wake bit in a particular LANai.
 * This function replaces the one in lanai_dev.c. The API when
 * used from the kernel can't use lanai_dev.c.
 */

void
lanai_set_ex1_unit(int unit, int n)
{
	struct myri *myrip;

	myrip = myridev;

	while (myrip) {
		if (unit == ddi_get_instance(myrip->myri_dip))
			break;

		myrip = myrip->myri_nextp;
	}

#ifdef i86pc
	if (myrip) {
		if (n) {
			myrip->lanai_memory_wake[0] = myrip->lanai_memory[0];
			CMN_ERR(6) (CE_CONT, "lanai_set_ex1(%d)  wake_ON\n", n);
		}
	}
#else
	if (myrip) {
		if (n) {
			myrip->myri_regsp->wake_on = 0;
			CMN_ERR(6) (CE_CONT, "lanai_set_ex1(%d)  wake_ON\n", n);
		}
		else {
			myrip->myri_regsp->wake_off = 0;
			CMN_ERR(6) (CE_CONT, "lanai_set_ex1(%d)  wake_OFF\n", n);
		}
	}
#endif
	else {
		CMN_ERR(0) (CE_CONT, "set_ex1: unit = %d not found\n");
	}
}

void
myri_stop_lanai(struct myri * myrip)
{
	CMN_ERR(3) (CE_CONT, "          myri_STOP_lanai called\n");
	return;			/* FIX */

#ifdef DELETE_ME
	/* turn off interrupts and reset the LANai */
	myrip->myri_regsp->int_off = 1;
	myrip->myri_regsp->rst_on = 1;

	/* zero out the LANai memory just in case */
	bzero((caddr_t) & myrip->myri_regsp->lanai_memory[0], LANAI_SIZE);
#endif				/* DELETE_ME */
}


void
myri_dump_msg(mblk_t * mp)
{
	if (mp) {
		int len = 0, cnt = 0;
		mblk_t *mp_tmp;

		len = msgdsize(mp);

		CMN_ERR(0) (CE_CONT, "   MSG: len = %d bytes\n", len);

		mp_tmp = mp;
		while (mp_tmp) {
			CMN_ERR(0) (CE_CONT, "  %d: rptr = 0x%x  len = %d   type = %d\n",
				    cnt++, mp_tmp->b_rptr,
				    (mp_tmp->b_wptr - mp_tmp->b_rptr),
				    mp_tmp->b_datap->db_type);
			if (PRINT_LEVEL >= 0)
				hex_dump(mp_tmp->b_rptr, (mp_tmp->b_wptr - mp_tmp->b_rptr));
			mp_tmp = mp_tmp->b_cont;
		}		/* while */
	}
}



/******************  copy_mblk_to_buffer()  ***************************/
/*
 * Note: length is the exact length of the packet after padding with zeros
 *       The true length of the packet before padding comes from the mblk itself
 *       as we copy it.
 *
 */

void
copy_mblk_to_buffer(void *vptr, unsigned short *lptr, unsigned int length)
{
	register int len = MYRI_PAD_LEN;
	register mblk_t *m = (mblk_t *) vptr;
	register unsigned char *ptr;
	register int buf_len;

	GET_TIME(time_stop[7]);
	/* put padding on the front */
	lptr[0] = MYRI_TAG;
	ptr = (unsigned char *) &lptr[1];

	while (m) {
		buf_len = (m->b_wptr - m->b_rptr);
		if (buf_len>0) {
			bcopy((caddr_t) m->b_rptr, (caddr_t) ptr, buf_len);
			ptr += buf_len;
			len += buf_len;
		}
		else if (buf_len<0) {
			CMN_ERR(0)(CE_CONT,"Bad mblk? %08x %08x  len = %d\n",
						(unsigned)m->b_wptr,
						(unsigned)m->b_rptr,
						buf_len);
		}
		m = m->b_cont;
	}

	if (len & 0x3) {
		/* handle the padding of extra zeros */
		register int copy;
		copy = (length - len);

		PRINTF(2) ("adding %d '0' bytes (%d-%d) lptr = 0x%x  ptr = 0x%x\n",
			   copy, length, len,
			   (unsigned) lptr, (unsigned) ptr);

		while (copy--) {
			*ptr++ = 0;
		}
	}

#ifdef DO_CKSUM
	if (!verify_cksum((unsigned char *)&lptr[0]),m) {
	 	CMN_ERR(0)(CE_CONT, "TCP checksum not verified! len = %d\n",length);
	}
#endif DO_CKSUM


	GET_TIME(time_stop[8]);
}


/* This is called in the API library. For kernel operation,
 * we need to supply the function.
 */
void
usleep(unsigned usec)
{

	DELAY(usec);
}


/* end of myri.c */
