#include    <sys/types.h>
#include    <sys/errno.h>
#include    <sys/debug.h>
#include    <sys/stropts.h>
#include    <sys/stream.h>
#include    <sys/strlog.h>
#include    <sys/cmn_err.h>
#include    <sys/kmem.h>
#include    <sys/conf.h>
#include    <sys/ksynch.h>
#include    <sys/stat.h>
#include    <sys/dlpi.h>
#include    <sys/modctl.h>
#ifdef  KSTAT
#include        <sys/kstat.h>
#endif
/*
#include    <string.h>
*/
#include    <sys/ddi.h>
#include    <sys/sunddi.h>

#include "config.h"
#include "myri_api.h"
#include "myri.h"

static int
setup_buff_ptr(struct myri * myrip,
	       caddr_t rptr, struct buff_info * infop, int length, int dir);




/************************  myri_alloc_buffers ****************************/

int
myri_alloc_buffers(struct myri * myrip)
{
	unsigned char *tmp_ptr;
	unsigned char *rptr;
	int i, rv = 0;
	struct buff_info *infop;

	if (!myrip) {
		return (0);
	}

	if (myrip->buffer_ptr) {
		/* already allocated */
		return (1);
	}

	CMN_ERR(0) (CE_CONT, "myri_allocate: allocating buffer structures\n");
	CMN_ERR(5) (CE_CONT, "myri_allocate: sizeof buff_info = %d handle = %d\n",
		    sizeof(struct buff_info),
		    sizeof(ddi_dma_handle_t));

	/* allocate an array of recv/send buffers */
	myrip->buffer_ptr = (unsigned int *) kmem_zalloc(BUFFER_BYTES, KM_NOSLEEP);

	CMN_ERR(1) (CE_CONT, "myri_alloc_buffers: kernel addr for buffer_ptr = 0x%x\n",
		    (unsigned) myrip->buffer_ptr);

	if (!myrip->buffer_ptr) {
		/*
		 * Try again with SLEEP. Allocate an array of recv/send
		 * buffers
		 */
		myrip->buffer_ptr = (unsigned int *) kmem_zalloc(BUFFER_BYTES, KM_SLEEP);
		if (!myrip->buffer_ptr) {
			CMN_ERR(0) (CE_CONT, "myri_allocate: no space for bufs size = %dbytes\n",
				    BUFFER_BYTES);

			return (0);
		}
	}

	MUTEX_ENTER(&myrip->myri_bufflock);
	ASSERT(++myrip->buff_count==1);

	/* get buffers page aligned */
	rptr = (unsigned char *) ROUNDUP((unsigned) myrip->buffer_ptr, PAGESIZE);

	tmp_ptr = rptr;
	infop = &myrip->recv0info[0];
	for (i = 0; i < NUM_RECV_BUFF0; i++, infop++) {
		infop->myrip = myrip;
		infop->kptr = (unsigned int *) tmp_ptr;
		infop->index = i;
		infop->status = BUF_NOT_MAPPED;
		infop->free_func.free_func = myri_add_buff;
		infop->free_func.free_arg = (char *) infop;
		infop->queue = RECV0_Q;
		tmp_ptr += RECV0_LEN;
	}

	infop = &myrip->recv1info[0];
	for (i = 0; i < NUM_RECV_BUFF1; i++, infop++) {
		infop->myrip = myrip;
		infop->kptr = (unsigned int *) tmp_ptr;
		infop->index = i;
		infop->status = BUF_NOT_MAPPED;
		infop->free_func.free_func = myri_add_buff;
		infop->free_func.free_arg = (char *) infop;
		infop->queue = RECV1_Q;
		tmp_ptr += RECV1_LEN;
	}

	infop = &myrip->sendinfo[0];
	for (i = 0; i < NUM_SEND_BUFF; i++, infop++) {
		infop->myrip = myrip;
		infop->kptr = (unsigned int *) tmp_ptr;
		infop->index = i;
		infop->status = BUF_NOT_MAPPED;
		infop->free_func.free_func = myri_add_buff;
		infop->free_func.free_arg = (char *) infop;
		infop->queue = SENDQ;
		tmp_ptr += SEND_LEN;
	}

	infop = &myrip->recv0info[0];
	for (i = 0; i < NUM_RECV_BUFF0; i++, infop++) {
		if ((rv = setup_buff_ptr(myrip, (caddr_t) rptr,
					 infop, RECV0_LEN, DDI_DMA_READ))) {
			if (rv == -1) {
				goto setup_fail;
			}
			else if (rv == -2) {
				goto failed2;
			}
		}
		rptr += RECV0_LEN;
	}

	infop = &myrip->recv1info[0];
	for (i = 0; i < NUM_RECV_BUFF1; i++, infop++) {
		if ((rv = setup_buff_ptr(myrip, (caddr_t) rptr,
					 infop, RECV1_LEN, DDI_DMA_READ))) {
			if (rv == -1) {
				goto setup_fail;
			}
			else if (rv == -2) {
				goto failed2;
			}
		}
		rptr += RECV1_LEN;
	}

	infop = &myrip->sendinfo[0];
	for (i = 0; i < NUM_SEND_BUFF; i++, infop++) {
		if ((rv = setup_buff_ptr(myrip, (caddr_t) rptr,
					 infop, SEND_LEN, DDI_DMA_WRITE))) {
			if (rv == -1) {
				goto setup_fail;
			}
			else if (rv == -2) {
				goto failed2;
			}
		}
		rptr += SEND_LEN;
	}

	CMN_ERR(5) (CE_CONT, "Got addresses for the buffers\n");

#if (PRINT_LEVEL>=5)
	myri_dump_buff_info(&myrip->recv0info[0]);
	myri_dump_buff_info(&myrip->recv1info[0]);
	myri_dump_buff_info(&myrip->sendinfo[0]);
#endif				/* PRINT_LEVEL */

	myrip->buff_count--;
	MUTEX_EXIT(&myrip->myri_bufflock);
	return (1);

setup_fail:
failed2:
	CMN_ERR(0) (CE_CONT, "myri_alloc_buffers FAILED!\n");
	kmem_free(myrip->buffer_ptr, BUFFER_BYTES);
	myrip->buffer_ptr = NULL;
	{
		int j;
		struct buff_info *infop;

		infop = &myrip->recv0info[0];
		for (j = 0; j < NUM_RECV_BUFF0; j++, infop++) {
			if (infop->status == BUF_UNUSED) {
				ddi_dma_free(infop->handle);
				infop->status = BUF_NOT_MAPPED;
			}
		}
		infop = &myrip->recv1info[0];
		for (j = 0; j < NUM_RECV_BUFF1; j++, infop++) {
			if (infop->status == BUF_UNUSED) {
				ddi_dma_free(infop->handle);
				infop->status = BUF_NOT_MAPPED;
			}
		}
		infop = &myrip->sendinfo[0];
		for (j = 0; j < NUM_SEND_BUFF; j++, infop++) {
			if (infop->status == BUF_UNUSED) {
				ddi_dma_free(infop->handle);
				infop->status = BUF_NOT_MAPPED;
			}
		}

	}

	myrip->buff_count--;
	MUTEX_EXIT(&myrip->myri_bufflock);
	return (0);
}



/*******************   setup_buff_ptr()  *****************************/
static int
setup_buff_ptr(struct myri * myrip,
	       caddr_t rptr, struct buff_info * infop, int length, int dir)
{
	int rv;
	ddi_dma_cookie_t cookie;

	rv = ddi_dma_addr_setup(myrip->myri_dip, NULL,
				(caddr_t) rptr, length, dir,
				DDI_DMA_DONTWAIT, NULL,
				&myri_dma_lim, &infop->handle);

	if (rv != DDI_DMA_MAPPED) {
		CMN_ERR(0) (CE_CONT, "alloc_buffers: dma_addr_setup failed infop = 0x%x\n",
			    (unsigned) infop);
		myri_dma_error(rv);
		return (-1);
	}

	infop->status = BUF_UNUSED;

	rv = ddi_dma_htoc(infop->handle, 0, &cookie);

	if (rv != DDI_SUCCESS) {
		CMN_ERR(0) (CE_CONT, "failed to get cookie\n");
		myri_dma_error(rv);
		return (-2);
	}

	infop->dptr = (unsigned int *) cookie.dmac_address;
	CMN_ERR(5) (CE_CONT, "infop->dptr = 0x%x\n",
		    (unsigned) infop->dptr);

	return (0);
}



/********************* myri_free_buff_all() ******************/

int
myri_free_buff_all(struct myri * myrip)
{
	int i;
	register struct buff_info *infop;

	CMN_ERR(BUFF_PRINT) (CE_CONT, "myri_free_buff_all() called\n");

	MUTEX_ENTER(&myrip->myri_bufflock);
    ASSERT(++myrip->buff_count==1);

	infop = &myrip->recv0info[0];
	for (i = 0; i < NUM_RECV_BUFF0; i++, infop++) {
		if (infop->status != BUF_LOANED) {
			infop->status = BUF_UNUSED;
		}
	}

	infop = &myrip->recv1info[0];
	for (i = 0; i < NUM_RECV_BUFF1; i++, infop++) {
		if (infop->status != BUF_LOANED) {
			infop->status = BUF_UNUSED;
		}
	}

	infop = &myrip->sendinfo[0];
	for (i = 0; i < NUM_SEND_BUFF; i++, infop++) {
		if (infop->status != BUF_LOANED) {
			infop->status = BUF_UNUSED;
		}
	}
	myrip->last_send_buff = 0;
	myrip->buff_at_lanai[RECV0_Q] = 0;
	myrip->buff_at_lanai[RECV1_Q] = 0;

	myrip->buff_count--;
	MUTEX_EXIT(&myrip->myri_bufflock);

	return (1);
}




/************************   myri_add_buff()  ************************/
/*
 * Called to free a recv_buff and hand it to the LANai
 * If index == -1 then try to add all bufs for a particular queue
 *
 */
void
myri_add_buff(struct buff_info * infop)
{
	struct myri *myrip = infop->myrip;
	int unit = ddi_get_instance(myrip->myri_dip);
	int retval;
	int buf_len = 0;
	register int i;
	int max;
	int queue;

	CMN_ERR(BUFF_PRINT) (CE_CONT, "myri_add_buff(%s,%d,%s)\n",
			     (infop->queue == RECV0_Q) ? "queue0" : "queue1",
			     infop->index, 
				(infop->index == -1) ? "-" : BUFFINFO_STATUS(infop->status));

	switch (infop->queue) {
		case RECV0_Q:
			buf_len = RECV0_LEN;
			break;

		case RECV1_Q:
			buf_len = RECV1_LEN;
			break;

		default:
			CMN_ERR(0) (CE_CONT, "myri_add_buff: bad queue = %d\n",
				    infop->queue);
			return;
			break;
	}

	MUTEX_ENTER(&myrip->myri_bufflock);
    ASSERT(++myrip->buff_count==1);

	if (infop->index != -1) {
		switch (infop->status) {
			case BUF_LOANED:
			case BUF_UNUSED:
			case BUF_NULL:
				infop->status = BUF_LANAI;
				if ((retval = myriAddReceiveBuffer(unit, KERNEL_INTERFACE,
							   (unsigned) infop,
					      (unsigned char *) infop->dptr,
					      buf_len, infop->queue)) > 0) {
					myrip->buff_at_lanai[infop->queue]++;
				}
				else {
					/* add failed */
					if (retval <= 0) {
						CMN_ERR(BUFF_PRINT) (CE_CONT, "myri_add_buff: indiv. add FAILED queue%d  idx=%d rv=%d\n",
								     infop->queue, infop->index, retval);
					}
					if (retval < 0) {
						/* need to handshake */
					}
					infop->status = BUF_UNUSED;
				}
				break;

			case BUF_LANAI:
			default:
				/* invalid status for adding to LANai */
				CMN_ERR(0) (CE_CONT, "myri_add_buff: bad status? = %s queue%d idx=%d\n",
					    BUFFINFO_STATUS(infop->status), infop->queue, infop->index);
				break;
		}		/* switch */
	}
	else {
		/* try all the bufs */
		/*
		 * reset infop to start at the appropriate array of
		 * buff_infos
		 */

		queue = infop->queue;
		if (queue == RECV0_Q) {
			infop = &myrip->recv0info[0];
			max = NUM_RECV_BUFF0;
		}
		else if (queue == RECV1_Q) {
			infop = &myrip->recv1info[0];
			max = NUM_RECV_BUFF1;
		}
		else {
			CMN_ERR(0) (CE_CONT, "myri_add_buff: bad queue = %d for index = -1\n",
				    queue);
			myrip->buff_count--;
			MUTEX_EXIT(&myrip->myri_bufflock);
			return;
		}


		for (i = 0; i < max; i++, infop++) {
			if (!((infop->status == BUF_LOANED) ||
			      (infop->status == BUF_LANAI))) {

				infop->status = BUF_LANAI;
				if ((retval = myriAddReceiveBuffer(unit, KERNEL_INTERFACE,
							   (unsigned) infop,
					      (unsigned char *) infop->dptr,
						     buf_len, queue)) > 0) {

					CMN_ERR(BUFF_PRINT) (CE_CONT, "    ---myri_add_buff(%s,%d,%d)\n",
							     (queue == RECV0_Q) ? "queue0" : "queue1",
					       infop->index, infop->status);

					myrip->buff_at_lanai[queue]++;
				}
				else {
					/* add failed */
					infop->status = BUF_UNUSED;

					CMN_ERR(1) (CE_CONT, "    ---myri_add_buff(%s,%d,%d) *** FAILED = %d\n",
						    (queue == RECV0_Q) ? "queue0" : "queue1",
					infop->index, infop->status, retval);

					break;
				}
			}
			else {
				/*
				 * don't touch the buffer, it is not
				 * available
				 */
			}
		}		/* for i */
	}			/* all the bufs */


	CMN_ERR(5) (CE_CONT, "----->myri_add_buff(EXIT)\n");

	myrip->buff_count--;
	MUTEX_EXIT(&myrip->myri_bufflock);
}


/************************   myri_dump_buff_info()  ************************/
/*
 * Called to print the information in a buff_info array
 *
 */
void
myri_dump_buff_info(struct buff_info * infop)
{
	int i;
	int max;
	int unit = ddi_get_instance(infop->myrip->myri_dip);

	if ((unsigned) infop == (unsigned) &infop->myrip->recv0info[0]) {
		max = NUM_RECV_BUFF0;
		CMN_ERR(0) (CE_CONT, "Recv 0 buffers\n");
	}
	else if ((unsigned) infop == (unsigned) &infop->myrip->recv1info[0]) {
		max = NUM_RECV_BUFF1;
		CMN_ERR(0) (CE_CONT, "Recv 1 buffers\n");
	}
	else if ((unsigned) infop == (unsigned) &infop->myrip->sendinfo[0]) {
		max = NUM_SEND_BUFF;
		CMN_ERR(0) (CE_CONT, "SEND buffers\n");
	}
	else {
		CMN_ERR(0) (CE_CONT, "infop - no match\n");
		return;
	}

	CMN_ERR(0) (CE_CONT, "index\tkptr\tdptr\tunit\tqueue\tstatus\n");
	for (i = 0; i < max; i++, infop++) {
		CMN_ERR(0) (CE_CONT, "%d\t%x\t%x\t%d\t%d\t%s\n",
			    infop->index,
			    infop->kptr,
			    infop->dptr,
			    unit,
			    infop->queue,
			    BUFFINFO_STATUS(infop->status));
	}
}


/* end of myri_buff.c */
