/* mlanai.c (mmap)*/
/*************************************************************************
 *                                                                       *
 * Myricom Myrinet Software                                              *
 *                                                                       *
 * Copyright (c) 1994 by Myricom, Inc.                                   *
 * All rights reserved.                                                  *
 *                                                                       *
 * This software is released under a non-disclosure agreement with       *
 * Myricom, Inc. Redistribution in any form is strictly prohibited.      *
 *                                                                       *
 * Myricom, Inc. makes no representations about the suitability of this  *
 * software for any purpose.                                             *
 *                                                                       *
 * THIS FILE IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING  *
 * THE WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A           *
 * PARTICULAR PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR     *
 * TRADE PRACTICE. MYRICOM, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO *
 * THE INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY THIS  *
 * FILE OR ANY PART THEREOF.                                             *
 *                                                                       *
 * In no event will Myricom, Inc. be liable for any lost revenue         *
 * or profits or other special, indirect and consequential damages, even *
 * if Myricom has been advised of the possibility of such damages.       *
 *                                                                       *
 *                                                                       *
 * Other copyrights might apply to parts of this software and are so     *
 * noted when applicable.                                                *
 *                                                                       *
 * Myricom, Inc.                                                         *
 * 325 N. Santa Anita Ave.                                               *
 * Arcadia, CA 91024                                                     *
 *                                                                       *
 *************************************************************************/
static char *myricopyright = "\t\t\tCopyright (c) 1994 by Myricom, Inc.";


#ifndef _KERNEL
#define _KERNEL
#endif
#ifndef SUNDDI
#define SUNDDI
#endif

#include <sys/types.h>
#include <sys/param.h>
#include <sys/errno.h>
#include <sys/file.h>
#include <sys/conf.h>
#include <sys/uio.h>
#include <sys/map.h>
#include <sys/debug.h>
#include <sys/modctl.h>
#include <sys/kmem.h>
#include <sys/cmn_err.h>
#include <sys/open.h>
#include <sys/stat.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>

#include "config.h"
#include "myri_api.h"
#include "myri.h"

#define AM_INTR  0xf211 /* flag used for interrupts */

/* needed for linking mmap infor for dlpi info */
char _depends_on[] = "drv/myri";
extern void *myri_state;

/* ------------------------------------------------------------------- */


#define MLANAI_NAME1 "mlanai"
#define MLANAI_NAME2 "MYRICOM,mlanai"
 /* FIX - are these correct?? */
#define MLANAI_MINOR(instance) (instance)
#define MLANAI_INSTANCE(minor) (minor)

#define roundup(x, y)   ((((x)+((y)-1))/(y))*(y))

static ddi_dma_lim_t mlanai_dma_lim = {
	0x0,			/* lower */
	0xFFFFFFFF,		/* upper */
	0xFFFFFFFF,		/* counter */
	0x7F,			/* default burst size */
	4,			/* min transfer in bytes */
	56320,			/* 55 MBytes/sec */
};


#define POLL_RV_ALL (POLLIN | POLLPRI | POLLOUT | POLLRDNORM | POLLRDBAND | POLLWRBAND)

static void *state_head;	/* opaque handle top of state structs */

uint mlanai_copy_block_size;

/* defined in dlpi.c */
extern int myri_use_am_intr ;  /* to myriintr, check am_flags? */
extern int am_intr_count;      /* how many interrupts to use?  */
extern kmutex_t   myri_am_mu;
extern kcondvar_t myri_am_cv;

/*
 * These are the entry points into our driver that are called when the
 * driver is loaded, during a system call, or in response to an interrupt.
 */
static int
mlanai_getinfo(dev_info_t * dip, ddi_info_cmd_t infocmd, void *arg,
	       void **result);
static int mlanai_identify(dev_info_t * dip);
static int mlanai_attach(dev_info_t * dip, ddi_attach_cmd_t cmd);
static int mlanai_open(dev_t * dev, int openflags, int otyp, cred_t * credp);
static int mlanai_close(dev_t dev, int openflags, int otyp, cred_t * credp);
static int mlanai_ioctl(dev_t dev, int cmd, int arg,int mode, cred_t *credp,
	                   int *rvalp);
static int mlanai_read(dev_t dev, struct uio *uiop, cred_t *credp);
static int mlanai_detach(dev_info_t * dip, ddi_detach_cmd_t cmd);
static int mlanai_probe(dev_info_t * dip);
static int mlanai_mmap(dev_t dev, off_t off, int prot);
static u_int mlanai_intr();
static int
mlanai_chpoll(dev_t dev, short events, int anyyet, short *reventsp,
	      struct pollhead ** phpp);
static int mlanai_do_dma_setup(caddr_t arg);


/*
 * When our driver is loaded or unloaded, the system calls our _init or
 * _fini routine with a modlinkage structure.  The modlinkage structure
 * contains:
 *
 *	modlinkage->
 *		modldrv->
 *			dev_ops->
 * 				cb_ops
 *
 * cb_ops contains the normal driver entry points and is roughly equivalent
 * to the cdevsw & bdevsw structures in previous releases.
 *
 * dev_ops contains, in addition to the pointer to cb_ops, the routines
 * that support loading and unloading our driver.
 */
static struct cb_ops mlanai_cb_ops = {
	mlanai_open,
	mlanai_close,
	nodev,			/* not a block driver	 */
	nodev,			/* no print routine	 */
	nodev,			/* no dump routine	 */
	mlanai_read,		/* dummy read routine 	 */
	nodev,			/* no write routine 	 */
	mlanai_ioctl,		/* ioctl routine 	 */
	nodev,			/* no devmap routine	 */
	mlanai_mmap,
	ddi_segmap,
	mlanai_chpoll,		/* chpoll routine	 */
	ddi_prop_op,
	0,			/* not a STREAMS driver	 */
	D_NEW | D_MP,		/* safe for multi-thread/multi-processor */
};

static struct dev_ops mlanai_ops = {
	DEVO_REV,		/* DEVO_REV indicated by manual	 */
	0,			/* device reference count	 */
	mlanai_getinfo,
	mlanai_identify,
	mlanai_probe,		/* device probe for non-self-id */
	mlanai_attach,
	mlanai_detach,
	nodev,			/* device reset routine		 */
	&mlanai_cb_ops,
	(struct bus_ops *) 0,	/* bus operations		 */
};

extern struct mod_ops mod_driverops;
static struct modldrv modldrv = {
	&mod_driverops,
	"Myrinet mmap Driver(2.2-4M)",	/* Module Name	 */
	&mlanai_ops,
};

static struct modlinkage modlinkage = {
	MODREV_1,		/* MODREV_1 indicated by manual */
	{(void *) &modldrv,
	NULL}			/* termination of list of linkage structures */
};

/*
 * _init, _info, and _fini support loading and unloading the driver.
 */
int
_init(void)
{
	register int error;

	CMN_ERR(3) (CE_CONT, "mlanai:  _init  called\n");

	if ((error =
	     ddi_soft_state_init(&state_head, sizeof(LANAI_DATA), 1)) != 0) {
		CMN_ERR(0) (CE_CONT, "soft state init failed\n");
		return (error);
	}

	if ((error = mod_install(&modlinkage)) != 0) {
		CMN_ERR(3) (CE_CONT, "modinstall failed\n");
		ddi_soft_state_fini(&state_head);
	}

	return (error);
}

int
_info(struct modinfo * modinfop)
{
	CMN_ERR(3) (CE_CONT, "mlanai   _info  called\n");
	return (mod_info(&modlinkage, modinfop));
}

int
_fini(void)
{
	int status;

	CMN_ERR(3) (CE_CONT, "mlanai   _fini  called\n");
	if ((status = mod_remove(&modlinkage)) != 0)
		return (status);

	ddi_soft_state_fini(&state_head);

	return (status);
}

/*
 * This is a pretty generic getinfo routine as described in the manual.
 */
/*ARGSUSED*/
static int
mlanai_getinfo(dev_info_t * dip, ddi_info_cmd_t infocmd, void *arg, void **result)
{
	register int error;
	register LANAI_DATA *mlanai_p;

	CMN_ERR(3) (CE_CONT, "mlanai_getinfo  called\n");

	switch (infocmd) {
		case DDI_INFO_DEVT2DEVINFO:
			CMN_ERR(3) (CE_CONT, "        DDI_INFO_DEVT2DEVINFO\n");
			mlanai_p = (LANAI_DATA *) ddi_get_soft_state(state_head,
				    MLANAI_INSTANCE(getminor((dev_t) arg)));
			if (mlanai_p == NULL) {
				*result = NULL;
				error = DDI_FAILURE;
			}
			else {
				/*
				 * don't need to use a MUTEX even though we
				 * are accessing our instance structure;
				 * mlanai_p->devinfo_p never changes.
				 */
				*result = mlanai_p->devinfo_p;
				error = DDI_SUCCESS;
			}
			break;
		case DDI_INFO_DEVT2INSTANCE:
			CMN_ERR(3) (CE_CONT, "        DDI_INFO_DEVT2INSTANCE\n");
			*result = (void *) MLANAI_INSTANCE(getminor((dev_t) arg));
			error = DDI_SUCCESS;
			break;
		default:
			CMN_ERR(0) (CE_CONT, "        default!!\n");
			*result = NULL;
			error = DDI_FAILURE;
	}

	return (error);
}

static int
mlanai_identify(dev_info_t * dip)
{
	char *dev_name;
	dev_name = ddi_get_name(dip);

	CMN_ERR(0) (CE_CONT, "mlanai_identify called: string = '%s'\n",
		    dev_name);

	if (strcmp(dev_name, MLANAI_NAME1) == 0) {
		CMN_ERR(0) (CE_CONT, "        IDENTIFIED\n");
		return (DDI_IDENTIFIED);
	}
/*
	else if (strcmp(dev_name, MLANAI_NAME2) == 0) {
		CMN_ERR(0) (CE_CONT, "IDENTIFIED\n");
		return (DDI_IDENTIFIED);
	}
*/
	else {
		CMN_ERR(3) (CE_CONT, "NOT identified\n");
		return (DDI_NOT_IDENTIFIED);
	}
}


/*
 * mlanai_attach gets called if mlanai_identify returns DDI_IDENTIFIED
 */
static int
mlanai_attach(dev_info_t * dip, ddi_attach_cmd_t cmd)
{
	int instance, minor;
	register LANAI_DATA *mlanai_p;
	char name[16];
	struct myri *myrip = NULL;

	CMN_ERR(0) (CE_CONT, "mlanai_attach(): called\n");
	mlanai_copy_block_size = COPY_BLOCK_SIZE;

	if (cmd != DDI_ATTACH)
		return (DDI_FAILURE);

	instance = ddi_get_instance(dip);

	CMN_ERR(0) (CE_CONT, "    instance = %d\n", instance);

	/****************************************************************/
	/*
	 * Here's the strange stuff, we're trying to locate the "myri" driver
	 * to use its data structures to get pointers to the LANAI.
	 */

	myrip = ddi_get_soft_state(myri_state, instance);
	if (!myrip) {
		CMN_ERR(0) (CE_CONT, "mlanai_attach():  get_soft_state (myri): failed");
		CMN_ERR(0) (CE_CONT, "mlanai_attach():  myri DLPI device myri%d not yet 'plumbed'\n ",
			    instance);
		return (DDI_FAILURE);
	}

	/****************************************************************/


	if (ddi_soft_state_zalloc(state_head, instance) != 0)
		return (DDI_FAILURE);

	if (!(mlanai_p = (LANAI_DATA *) ddi_get_soft_state(state_head, instance)))
		return (DDI_FAILURE);
	mlanai_p->devinfo_p = dip;
	mlanai_p->state = 1;
	ddi_set_driver_private(dip, (caddr_t) mlanai_p);
	mlanai_p->myrip = myrip;/* save the myrip pointer for DMA use */

	/* grab the board address from the other driver (myri) */
#ifdef i86pc
	mlanai_p->cntrl = (volatile LANAI_REG *) myrip->myri_regsp;
	mlanai_p->memory = (unsigned int *) (&myrip->lanai_memory[0]);
	mlanai_p->sts = mlanai_p->myrip->sts;
#else
	mlanai_p->cntrl = (volatile LANAI_REG *) myrip->myri_regsp;
	mlanai_p->memory = (unsigned int *) ((unsigned) mlanai_p->cntrl + CNTRL_SIZE);
	mlanai_p->sts = mlanai_p->myrip->sts;
#endif

	/* "register" my interrupt routine with the DLPI driver */
	/* this replaces the add_intr() function call */
	myrip->mlanai_intr = mlanai_intr;
	myrip->mlanai_p = mlanai_p;


	/*
	 * initialize MUTEX; we will use this MUTEX later on to lock our
	 * instance structure
	 */
	mutex_init(&mlanai_p->mutex, "mlanai mutex",
		   MUTEX_DRIVER, (void *) &mlanai_p->iblock_cookie);
	mlanai_p->ok = 1;

	/*
	 * ddi_create_minor_node creates an entry in an internal kernel
	 * table; the actual entry in the file system is created by
	 * drvconfig(1) in /devices and devlinks(1) in /dev when you run
	 * add_drv(1). SBus:1, VME:1 entry; note
	 * MLANAI_MINOR,MLANAI_INSTANCE. INSTANCE per board; minor per device
	 * on board. sid == self-identifying device; no probe needed NAME
	 * gets added to /devices/.../...:NAME.
	 */

	minor = MLANAI_MINOR(instance);

	(void) cmn_err(CE_CONT, "Myricom mmap() driver mlanai%d\n",
		       instance);
	(void) cmn_err(CE_CONT, "  %s\n", myricopyright);

	sprintf(name, "mlanai%d", minor);

	if (ddi_create_minor_node(dip, name, S_IFCHR, minor,
				  DDI_PSEUDO, 0) == DDI_FAILURE) {
		goto bailout;
	}
	mlanai_p->minor = 1;
	ddi_report_dev(dip);
#ifdef DELETE
	mlanai_p->cntrl->int_off = 0;	/* Clear Interrupt */
#endif				/* DELETE */
	return (DDI_SUCCESS);
bailout:
	cmn_err(CE_CONT, "Failed in attach\n");
	(void) mlanai_detach(dip, DDI_DETACH);
	return (DDI_FAILURE);
}

/*
 * When our driver is unloaded, mlanai_detach cleans up and frees the resources
 * we allocated in mlanai_attach.
 */
static int
mlanai_detach(dev_info_t * dip, ddi_detach_cmd_t cmd)
{
	register LANAI_DATA *mlanai_p;	/* will point to this */
	int instance;

	if (cmd != DDI_DETACH)
		return (DDI_FAILURE);

	instance = ddi_get_instance(dip);
	if (!(mlanai_p = (LANAI_DATA *) ddi_get_soft_state(state_head, instance)))
		return (DDI_FAILURE);

	if (mlanai_p->open)
		return (DDI_FAILURE);	/* Is this needed? */

	if (mlanai_p->copy_block_save) {
		/* turn off interrupts and write zeros to the lanai */
#ifdef i86pc
		AMCC_INT_OFF(mlanai_p->cntrl);
#else
		mlanai_p->cntrl->int_off = 1;	/* Turn off interrupts */
#endif
		bzero((caddr_t) mlanai_p->memory, LANAI_SIZE - LANAI_RESERVED);
		/*
		 * delay to let any pending DMA to complete, then reset the
		 * LANai
		 */
		drv_usecwait(250000);
#ifdef i86pc
		AMCC_RESET_ON(mlanai_p->cntrl);
#else
		mlanai_p->cntrl->rst_on = 1;	/* Hold Lanai in reset */
#endif

		if (mlanai_p->handle) {
			ddi_dma_free(mlanai_p->handle);
			mlanai_p->handle = NULL;
		}


#ifndef MYRI_COHERENT
		kmem_free(mlanai_p->copy_block_save,
			  (COPY_BLOCK_SIZE + 2 * PAGESIZE));
#else				/* MYRI_COHERENT */
		ddi_iopb_free((caddr_t) mlanai_p->copy_block_save);
#endif				/* MYRI_COHERENT */
		mlanai_p->copy_block_save = NULL;
	}

	if (mlanai_p->state)
		ddi_soft_state_free(state_head, instance);
	if (mlanai_p->addint)
		ddi_remove_intr(mlanai_p->devinfo_p, 0, mlanai_p->iblock_cookie);
	if (mlanai_p->ok)
		mutex_destroy(&mlanai_p->mutex);
	if (mlanai_p->minor)
		ddi_remove_minor_node(dip, NULL);

	return (DDI_SUCCESS);
}

/*
 * mlanai_probe is called to test the existence of the expected device.
 */

static int
mlanai_probe(dev_info_t * dip)
{
	struct myri *myrip = NULL;
	int instance;

	CMN_ERR(3) (CE_CONT, "mlanai_probe called\n");

	/* need to see if the myri0, myri1 etc is there */

	instance = ddi_get_instance(dip);

	CMN_ERR(3) (CE_CONT, "mlanai_probe: instance = %d\n",
					instance);

	/****************************************************************/
	/*
	 * Here's the strange stuff, we're trying to locate the "myri" driver
	 * to use its data structures to get pointers to the LANAI.
	 */

	myrip = ddi_get_soft_state(myri_state, instance);
	if (!myrip) {
		CMN_ERR(1) (CE_CONT,
		    "mlanai_probe(): ddi_get_soft_state (myri%d): failed\n",
			    instance);
		CMN_ERR(1) (CE_CONT,
			 "mlanai_probe(): 'myri%d' device not yet loaded\n",
			    instance);
		CMN_ERR(0) (CE_CONT,
			    "mlanai_probe(): Need to plumb IP Myrinet device unit = %d\n",
			    instance);
		CMN_ERR(0) (CE_CONT,
			    "       or remove this instance=%d from /kernel/drv/mlanai.conf\n",
			    instance);
		return (DDI_PROBE_FAILURE);
	}

	/****************************************************************/

	CMN_ERR(3) (CE_CONT, "mlanai_probe returning DDI_PROBE_SUCCESS\n");

	return (DDI_PROBE_SUCCESS);
}

/*
 * mlanai_open is called in response to the open(2) system call
 */
/*ARGSUSED*/
static int
mlanai_open(dev_t * dev, int openflags, int otyp, cred_t * credp)
{
	register LANAI_DATA *mlanai_p;

	if (!(mlanai_p = (LANAI_DATA *) ddi_get_soft_state(state_head,
					  MLANAI_INSTANCE(getminor(*dev)))))
		return (ENXIO);

	if (otyp != OTYP_CHR)
		return (EINVAL);
	mutex_enter(&mlanai_p->mutex);	/* Start critical section */
	if (!mlanai_p->open) {
		mlanai_p->int_pend = 0;
		mlanai_p->open = 1;
	}
	mutex_exit(&mlanai_p->mutex);	/* End critical section */

	return (0);
}

/*
 * mlanai_close is called after the last process that has the device open
 * calls close(2)
 *
 * Note that the file may be closed by the process while the mapping is
 * still in the effect, so don't do too much cleanup!
 *
 * The kernel does not actually need the driver to maintain the mapping
 * provided by mmap(2), only to establish the mapping.  So the driver
 * does not get notified when the memory is unmapped, either by munmap(2)
 * or by the death of the process.
 */
/*ARGSUSED*/
static int
mlanai_close(dev_t dev, int openflags, int otyp, cred_t * credp)
{
	register LANAI_DATA *mlanai_p;

	if (!(mlanai_p = (LANAI_DATA *) ddi_get_soft_state(state_head,
					   MLANAI_INSTANCE(getminor(dev)))))
		return (ENXIO);
	mutex_enter(&mlanai_p->mutex);	/* Start critical section */
	if (mlanai_p->open)
		mlanai_p->open = 0;
	mutex_exit(&mlanai_p->mutex);	/* End critical section */
	return (0);
}

static int 
mlanai_ioctl(dev_t dev, int cmd, int arg,int mode, cred_t *credp,
	     int *rvalp) {

  register LANAI_DATA *mlanai_p;
  int minor_dev;

  if (!(mlanai_p = (LANAI_DATA *) 
	ddi_get_soft_state(state_head,
			   MLANAI_INSTANCE(getminor(dev)))))
    return (ENXIO);

  switch(cmd) {
  default:
    return(ENOTTY);
  }
}

static int 
mlanai_read(dev_t dev, struct uio *uiop, cred_t *credp) {

  register LANAI_DATA *mlanai_p;
  short *lanai_memory;

  if (!(mlanai_p = (LANAI_DATA *) ddi_get_soft_state(state_head,
						     MLANAI_INSTANCE(getminor(dev)))))
    return (ENXIO);
  
  CMN_ERR(3) (CE_CONT,"inside mlanai_read\n");

  lanai_memory = (short *) mlanai_p->memory;


  mutex_enter(&myri_am_mu);

  CMN_ERR(3) (CE_CONT,"mlanai: inside mutex\n");
  myri_use_am_intr = 1;
  mlanai_p->myrip->myri_regsp->int_on = 1;  
  CMN_ERR(3) (CE_CONT,"lanai memory kaddr is 0x%0x \n",lanai_memory);

  lanai_memory[AM_INTR] = 1;

  /* goto sleep */
  cv_wait_sig(&myri_am_cv, &myri_am_mu);
  lanai_memory[AM_INTR] = 0;

  myri_use_am_intr = 0;

  mlanai_p->myrip->myri_regsp->int_on = 1;  
  mutex_exit(&myri_am_mu);

  CMN_ERR(3)  (CE_CONT,"mlanai: outside mutex\n");

  uiop->uio_resid -=4;
  return(4);

}


/*
 * Our mmap routine gets called for each page of memory the user process tries
 * to map using mmap(2).  Our routine actually gets called twice for each page;
 * the kernel checks to see that the mapping will succeed before doing the
 * mapping.
 *
 * To establish a user mapping, the kernel needs the PFN (page frame number)
 * of the physical memory, which we provide by calling hat_getkpfnum() on an
 * address mapped into the kernel.  Once we have the PFN, we no longer need
 * to have that kernel mapping.
 *
 * The system does not need the device driver in order to maintain a user
 * mapping; nor does it need for the device to remain open by the user process.
 * So while our mmap routine gets called to perform a mapping, we are not
 * called at all as a result of munmap(2); and a process may close our device
 * while retaining - and using - the user mapping.
 */
/*ARGSUSED*/
static int
mlanai_mmap(dev_t dev, off_t off, int prot) {
  register LANAI_DATA *mlanai_p;
  int burst_sizes = 0;

  if (!(mlanai_p = (LANAI_DATA *) ddi_get_soft_state(state_head,
						     MLANAI_INSTANCE(getminor(dev)))))
    return (-1);

  if ((unsigned int) off >= TOTSIZE_BYTES) {
    /*
     * We use a hack here to decide when to allocate a block for
     * copying. If mmap() is called with a length bigger than the
     * size of the board, then we allocate a copy_block and send
     * the DMA pointer back in the beginning of the block.
     */

    if (!mlanai_p->copy_block_save) {
      /* allocate the block */
      mlanai_p->copy_block_save = NULL;


#ifndef MYRI_COHERENT
      mlanai_p->copy_block_save = (unsigned int *) kmem_zalloc(
							       (COPY_BLOCK_SIZE + 2 * PAGESIZE),
							       KM_NOSLEEP);

      if (!mlanai_p->copy_block_save) {
	cmn_err(CE_CONT, "myri_mmap: Couldn't kmem_alloc() the copy_block!\n");
	return (-1);
      }
#else  /* MYRI_COHERENT */
      if (ddi_iopb_alloc(mlanai_p->devinfo_p,
			 &mlanai_dma_lim,
			 (COPY_BLOCK_SIZE + 2 * PAGESIZE),
			 (caddr_t) & mlanai_p->copy_block_save) != DDI_SUCCESS) {
	cmn_err(CE_CONT, "myri_mmap: Couldn't ddi_iopb_alloc() the copy_block!\n");
	goto dma_error;
      }
#endif /* MYRI_COHERENT */


      mlanai_p->copy_block = (unsigned int *)
	roundup((unsigned) mlanai_p->copy_block_save, PAGE_SIZE);

      CMN_ERR(0) (CE_CONT, "mlanai_mmap():call dma_addr_setup - possible SLEEP\n");
      if (mlanai_do_dma_setup((caddr_t) mlanai_p)) {
	CMN_ERR(0) (CE_CONT, "mlanai_do_dma_setup returns true.\n");
      }
      else {
	CMN_ERR(0) (CE_CONT, "mlanai_do_dma_setup returns FALSE!\n");
	goto dma_error;
      }


      if (ddi_dma_htoc(mlanai_p->handle, 0, &mlanai_p->cookie) == DDI_SUCCESS) {

	mlanai_p->io_copy_block = (unsigned int *)
	  mlanai_p->cookie.dmac_address;
      }
      else {
	CMN_ERR(0) (CE_CONT, "mmap(): ddi_dma_htoc() didn't succeed!\n");
	goto dma_error;
      }

      burst_sizes = ddi_dma_burstsizes(mlanai_p->handle);
      cmn_err(CE_CONT, "sbus 'burst-sizes' (bytes)= %b\n",
	      burst_sizes,
	      "\20\7burst64\6burst32\5burst16\4burst8\3burst4\2burst2\1burst1");
      mlanai_p->sts = (burst_sizes & 0xF8) >> 3;
      cmn_err(CE_CONT, "myri_sts = %b\n",
	      mlanai_p->sts,
	      "\20\4burst16word\3burst8word\2burst4word\1burst2word");

      /*
       * FIX hack to handle stupid controllers that don't
       * support all bursts less than the maximum
       */
#ifdef NOT_NEEDED
      if (!((mlanai_p->sts == 0xf) ||
	    (mlanai_p->sts == 0x7) ||
	    (mlanai_p->sts == 0x3) ||
	    (mlanai_p->sts == 0x1))) {
	mlanai_p->sts = 0;
	cmn_err(CE_CONT, "Brain-dead SBus controller:\n");
	cmn_err(CE_CONT, "       doesn't support all sizes less than the max.\n");
      }
#endif

      mlanai_p->copy_block[0] = (unsigned) mlanai_p->io_copy_block;
      mlanai_p->copy_block[1] = (unsigned) mlanai_p->cookie.dmac_size;
      mlanai_p->copy_block[2] = (unsigned) mlanai_p->sts;
      mlanai_p->copy_block[3] = (unsigned) mlanai_p->myrip->lanai_memory;

      cmn_err(CE_CONT, "myri_mmap: copy_blk allocated kaddr = %x  io = %x\n",
	      (unsigned) mlanai_p->copy_block,
	      (unsigned) mlanai_p->io_copy_block);
      cmn_err(CE_CONT, "myri_mmap: data[0] = 0x%x data[1] = 0x%x data[2] = 0x%x data[3] = 0x%x\n",
	      mlanai_p->copy_block[0],
	      mlanai_p->copy_block[1],
	      mlanai_p->copy_block[2],
	      mlanai_p->copy_block[3]);
    }


    mlanai_p->copy_block[0] = (unsigned) mlanai_p->io_copy_block;
    mlanai_p->copy_block[1] = (unsigned) mlanai_p->cookie.dmac_size;
    mlanai_p->copy_block[2] = (unsigned) mlanai_p->sts;
    mlanai_p->copy_block[3] = (unsigned) mlanai_p->myrip->lanai_memory;

    return (hat_getkpfnum((caddr_t) mlanai_p->copy_block +
			  (off - TOTSIZE_BYTES)));
  }

  /*
   * the first address must be in the first page of the LANai
   */
#ifdef i86pc
  if (off < (4 * 1024)) {
    return (hat_getkpfnum((caddr_t) (mlanai_p->cntrl)));
  }
  off -= (4 * 1024);
  if (off < (132 * 1024)) {
    if (mlanai_p->myrip->lanai_memory) {
      return (hat_getkpfnum((caddr_t) (mlanai_p->myrip->lanai_memory))
	      + (off >> MMU_PAGESHIFT));
    }
    else {
      return (hat_getkpfnum((caddr_t) (mlanai_p->cntrl)));
    }
  }
  off -= (132 * 1024);
  if (mlanai_p->myrip->lanai_memory_wake) {
    return (hat_getkpfnum((caddr_t) (mlanai_p->myrip->lanai_memory_wake))
	    + (off >> MMU_PAGESHIFT));
  }
  else {
    return (hat_getkpfnum((caddr_t) (mlanai_p->cntrl)));
  }
#else
  return (hat_getkpfnum((caddr_t) & (mlanai_p->cntrl->int_off))
	  + (off >> MMU_PAGESHIFT));
#endif

 dma_error:
  if (mlanai_p->handle) {
    ddi_dma_free(mlanai_p->handle);
    mlanai_p->handle = NULL;
  }


  if (mlanai_p->copy_block_save) {
#ifndef MYRI_COHERENT
    kmem_free(mlanai_p->copy_block_save,
	      (COPY_BLOCK_SIZE + 2 * PAGESIZE));
#else  /* MYRI_COHERENT */
    ddi_iopb_free((caddr_t) mlanai_p->copy_block_save);
#endif /* MYRI_COHERENT */
    mlanai_p->copy_block_save = NULL;
  }
  return (-1);
}

/*****************   mlanai_do_dma_setup()  ****************/

int
mlanai_do_dma_setup(caddr_t argument)
{
  LANAI_DATA *mlanai_p = (LANAI_DATA *) argument;

  CMN_ERR(0) (CE_CONT, "mlanai_do_dma_setup(0x%x) called\n",
	      (unsigned) mlanai_p);

  /* watch out, using the device_ptr from myri device here!! */

  switch (ddi_dma_addr_setup(mlanai_p->myrip->myri_dip, NULL,
			     (caddr_t) mlanai_p->copy_block, COPY_BLOCK_SIZE,
			     (DDI_DMA_RDWR | DDI_DMA_CONSISTENT),
			     mlanai_do_dma_setup, argument,
			     &mlanai_dma_lim, &mlanai_p->handle)) {
  case DDI_DMA_MAPPED:
    CMN_ERR(3) (CE_CONT, "dma_setup DDI_DMA_MAPPED - OK\n");
    return (1);
    break;

  case DDI_DMA_TOOBIG:
    CMN_ERR(0) (CE_CONT, "mlanai_mmap(): dma_setup DDI_DMA_TOOBIG\n");
    return (0);
    break;

  case DDI_DMA_NOMAPPING:
    CMN_ERR(0) (CE_CONT, "mlanai_mmap() dma_setup DDI_DMA_NOMAPPING\n");
    return (0);
    break;

  case DDI_DMA_NORESOURCES:
    CMN_ERR(0) (CE_CONT, "mlanai_mmap(): dma_setup DDI_DMA_NORESOURCES\n");
    return (0);
    break;

  case DDI_DMA_PARTIAL_MAP:
    CMN_ERR(0) (CE_CONT, "mlanai_mmap(): dma_setup DDI_DMA_PARTIAL_MAP\n");
    return (0);
    break;

  default:
    CMN_ERR(0) (CE_CONT, "mlanai_mmap(): dma_setup rv not understood!\n");
    return (0);
    break;
  } /* switch */
}




/*
 * Whenever the SBus sees an interrupt at the level we specified with
 * ddi_add_intr in mlanai_attach, we get a chance to determine if it is our
 * interrupt or somebody else's.
 *
 * This routine will get called for each instance of each driver that uses
 * that level interrupt until somebody returns DDI_INTR_CLAIMED.
 *
 * Note to 4.x developers: in 4.x your xxpoll() routine had to check each
 * of your instances (units) to determine if any were interrupting.  This
 * is no longer necessary as the system now calls your xxintr() routine
 * for each attached instance of your device.
 */
static u_int
mlanai_intr(register LANAI_DATA * mlanai_p)
{
  if (!mlanai_p->ok)
    return (DDI_INTR_UNCLAIMED); /* intr before mutex_init */

  CMN_ERR(5) (CE_CONT, "mlanai_intr (interrupt routine) called\n");

  mutex_enter(&mlanai_p->mutex); /* Start critical section */
#ifdef i86pc
  AMCC_INT_OFF(mlanai_p->cntrl);
#else
  mlanai_p->cntrl->int_off = 1;	/* Clear Interrupt */
#endif
  mlanai_p->int_pend = 1;
  if (mlanai_p->polling) {
    pollwakeup(&mlanai_p->pollhead, POLL_RV_ALL);
    mlanai_p->polling = 0;
    CMN_ERR(5) (CE_CONT, "mlanai_intr: waking up a process\n");
  }
#ifdef i86pc
  AMCC_INT_CLEAR(mlanai_p->cntrl);
  AMCC_INT_ON(mlanai_p->cntrl);
#else
  mlanai_p->cntrl->int_on = 1;	/* Clear Interrupt */
#endif

  mutex_exit(&mlanai_p->mutex);	/* End critical section */

  return (DDI_INTR_CLAIMED);
}

static int
mlanai_chpoll(dev_t dev, short events, int anyyet, short *reventsp,
	      struct pollhead ** phpp)
{
  register LANAI_DATA *mlanai_p;

  CMN_ERR(5) (CE_CONT, "mlanai_chpoll called\n");

  if (!(mlanai_p = (LANAI_DATA *) ddi_get_soft_state(state_head,
						     MLANAI_INSTANCE(getminor(dev))))) {
    CMN_ERR(5) (CE_CONT, "mlanai_chpoll exiting - no soft state\n");
    return (ENXIO);
  }

  CMN_ERR(5) (CE_CONT, "mlanai_chpoll called   events = %d  anyyet = %d\n",
	      events, anyyet);

  mutex_enter(&mlanai_p->mutex); /* Start critical section */

  if (mlanai_p->int_pend) {
    *reventsp = POLL_RV_ALL;
    mlanai_p->int_pend = 0;
  }
  else {
    *reventsp = 0;
    if (!anyyet) {
      CMN_ERR(5) (CE_CONT, "mlanai_chpoll: Suspending the process\n");
      *phpp = &mlanai_p->pollhead;
      mlanai_p->polling = 1;
    }
    else {
      CMN_ERR(5) (CE_CONT, "mlanai_chpoll:  anyyet = %d??\n", anyyet);
    }
  }
  mutex_exit(&mlanai_p->mutex);	/* End critical section */
  return (0);
}

/* end of mlanai.c */
