
/*************************************************************************
 *                                                                       *
 * Myricom Myrinet Software                                              *
 *                                                                       *
 * Copyright (c) 1994, 1995 by Myricom, Inc.                             *
 * All rights reserved.                                                  *
 *                                                                       *
 * This software is released under a non-disclosure agreement with       *
 * Myricom, Inc. Redistribution in any form is strictly prohibited.      *
 *                                                                       *
 * Myricom, Inc. makes no representations about the suitability of this  *
 * software for any purpose.                                             *
 *                                                                       *
 * THIS FILE IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING  *
 * THE WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A           *
 * PARTICULAR PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR     *
 * TRADE PRACTICE. MYRICOM, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO *
 * THE INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY THIS  *
 * FILE OR ANY PART THEREOF.                                             *
 *                                                                       *
 * In no event will Myricom, Inc. be liable for any lost revenue         *
 * or profits or other special, indirect and consequential damages, even *
 * if Myricom has been advised of the possibility of such damages.       *
 *                                                                       *
 *                                                                       *
 * Other copyrights might apply to parts of this software and are so     *
 * noted when applicable.                                                *
 *                                                                       *
 * Myricom, Inc.                                                         *
 * 325 N. Santa Anita Ave.                                               *
 * Arcadia, CA 91024                                                     *
 *                                                                       *
 *************************************************************************/

#include "myri_sizes.h"

#define BUFF_PRINT	5	/* debug buff stuff  if <= PRINT_LEVEL */
#define RESET_PRINT	5	/* debug reset stuff if <= PRINT_LEVEL */

#define MUTEX_ENTER(p)  { \
            if (p) mutex_enter(p); \
            CMN_ERR(5)(CE_CONT, "mutex_ENTER(0x%x)\n", (unsigned)(p)); \
            }

#define MUTEX_EXIT(p)   { \
            CMN_ERR(5)(CE_CONT, " mutex_EXIT(0x%x)\n", (unsigned)(p)); \
            if (p) mutex_exit(p); \
            }

#ifndef i86pc

/*************************************************************************
* The SBus board occupies 3 consecutive regions starting at the base     *
* address of the SBus slot. The addresses below are BYTE addresses.      *
*                                                                        *
*   0000 0000   ROM          [512KBytes]                                 *
*   0008 0000   CNTRL        [256KBytes]                                 *
*   000C 0000   LANAI memory [128KBytes]                                 *
*                                                                        *
*   The control region is further subdivided into these 8 sectors.       *
*                                                                        *
*  CNTRL + 0000 turn off board reset.                                    *
*  CNTRL + 0004 turn on  board reset.                                    *
*  CNTRL + 0008 turn off interrupt enable.                               *
*  CNTRL + 000C turn on  interrupt enable.                               *
*  CNTRL + 0010 turn off lanai wake signal.                              *
*  CNTRL + 0014 turn on  lanai wake signal.                              *
*  CNTRL + 0018 read interrupt bit.                                      *
*                                                                        *
*************************************************************************/

#define ROM_SIZE        (512*1024)
#define CNTRL_SIZE      (256*1024)
#define LANAI_SIZE      ((128*1024)+4096)	/* + 4k for L3.0 */
#define LANAI_RESERVED  (64)	/* reserved locations at end of LANai */


#define TOTSIZE_BYTES  (CNTRL_SIZE + LANAI_SIZE)

#define IBIT 0x01000000

/* CTRL, section in shorts */
#define BLOCKSIZE_SHORT	((CNTRL_SIZE)/sizeof(u_short))

struct myriregs {
	volatile unsigned long rst_off;
	volatile unsigned long rst_on;
	volatile unsigned long int_off;
	volatile unsigned long int_on;
	volatile unsigned long wake_off;
	volatile unsigned long wake_on;
	volatile unsigned long int_bit;
	unsigned long filler[(CNTRL_SIZE / sizeof(long)) - 7];
	volatile unsigned short lanai_memory[LANAI_SIZE / sizeof(short)];
};


/* typedefs for the mmap() driver */

typedef struct {
	volatile unsigned long rst_off;
	volatile unsigned long rst_on;
	volatile unsigned long int_off;
	volatile unsigned long int_on;
	volatile unsigned long wake_off;
	volatile unsigned long wake_on;
	volatile unsigned long int_bit;
}   LANAI_REG;
#endif !i86pc

#ifdef i86pc
#include "pci1.h"
#endif i86pc

typedef struct {
	int open:1;
	int ok:1;
	int state:1;
	int mapped:1;
	int addint:1;
	int minor:1;
	int polling:1;
	int int_pend:1;
	int size;
	volatile LANAI_REG *cntrl;
	unsigned int *memory;
	kmutex_t mutex;
	ddi_iblock_cookie_t iblock_cookie;	/* for mutexes */
	struct pollhead pollhead;
	dev_info_t *devinfo_p;
	unsigned *copy_block;
	unsigned *io_copy_block;
	unsigned *copy_block_save;
	ddi_dma_handle_t handle;
	ddi_dma_cookie_t cookie;
	unsigned sts;
	struct myri *myrip;
}   LANAI_DATA;


#define MYRI_MAX_BOARDS	(16)	/* maximum number of SBus boards */

#define MYRI_INT_DELAY	(250)	/* units are usec */
#define MYRI_INT_MASK	(RECV_ACK_MASK | L2H_MASK)
#define MAX_GATHER  ((int)512)
#define MAX_GATHER_ALLOWED  ((int)32)

#define MYRI_MAX_LANAI_ERRORS	(500)	/* num errors before reload */
#define MYRI_MAX_LANAI_RELOAD	(5)	/* num times to try reload */

#define RECV_FEW_MSGS		(10)	/* only rcv a few msgs per interrupt */
#define RECV_ALL_MSGS		(0xFFFF)	/* recv all msgs after send */
#define MYRI_MAX_SEND_TRIES	5		/* if AddSend fails, then retry */

/*for readability*/
#define RECV0_Q 0
#define RECV1_Q 1
#define SENDQ 4
#define KERNEL_INTERFACE 0	/* cast in stone */


/* padding 1/2 word (MYRI_PAD_LEN<<8 | 0xAB) */
/* this is to make sure that the IP packet is appropriately aligned */
/* for the higher layers */

#define MYRI_TAG    ((unsigned short)(0x02AB))
#define MYRI_BYTE_PAD	((unsigned char)0xAB)


#define HOWMANY(x, y)   (((x)+((y)-1))/(y))
#define ROUNDUP(x, y)   ((((x)+((y)-1))/(y))*(y))
#define CYCLE(x,max)	(((x+1)>=max)?0:(x+1))

enum OPS {
	NO_OP, INSERT, DELETE, FREE_BUF, FIND_DELETE
};


#define PRINTF(level)	if (level<=PRINT_LEVEL) printf
#define CMN_ERR(level)	if (level<=PRINT_LEVEL) cmn_err


/*
 * XXX Note that values used for #defines are bogus. You need to use
 * values that are appropriate for your driver/device. Also feel free
 * to add fields in the structs defined in this file.
 */

/*
 * XXX Declarations specific to the medium
 */

#define MEDIUM_MTU	(MYRINET_MTU)	/* max frame w/o header */
#define MEDIUM_MIN	(0)
#define MEDIUM_MAX	(HOST_MTU)	/* max frame w/header w/o fcs */


#define MEDIUMTYPE_MAX	(0xffff)/* max valid medium type */

/*
 * XXX Medium address length
 */

#define MEDIUM_ADDRL		(6)

/*
 * XXX structure for the medium specific address
 */

struct medium_addr {
	u_char medaddr[MEDIUM_ADDRL];
};

/*
 * XXX structure for the medium specific header
 */

struct medium_header {
	struct medium_addr med_dhost;
	struct medium_addr med_shost;
	u_short med_type;
};


/*
static struct medium_addr myri_zero_addr = {{0, 0, 0, 0, 0, 0}};
*/



/*
 * XXX Maximum # of multicast addresses per Stream
 */

#define DEVMAXMC	64
#define DEVMCALLOC	(DEVMAXMC * sizeof (struct medium_addr))

/*
 * XXX Declarations and definitions specific to the device. Add fields
 * as necessary.
 */


/*
 * XXX Definitions for modumyri_info.
 * Fill in values appropriate for your driver.
 * Contact USL to obtain a unique "registered" module ID number
 */

#define         MYRIIDNUM         (999)	/* module ID number */
#define         MYRINAME          "myri"	/* module name */
#define         MYRIMINPSZ        (0)	/* min packet size */
#define         MYRIMAXPSZ         (MYRINET_MTU)	/* max packet size */
#define         MYRIHIWAT         (0x7fffffff)	/* hi-water mark */
#define         MYRILOWAT         (1)	/* lo-water mark */


/*
 * Per-Stream instance state information.
 *
 * Each instance is dynamically allocated at open() and free'd at
 * close(). Each per-stream instance points to at most one per-device
 * structure using the sl_myrip field. All instances are threaded
 * together into one list of active instances ordered on minor device
 * number.
 */

struct myristr {
	struct myristr *sl_nextp;	/* next in list */
	queue_t *sl_rq;		/* ptr to our rq */
	struct myri *sl_myrip;	/* attached device */
	u_long sl_state;	/* current state */
	u_long sl_sap;		/* bound sap */
	u_long sl_flags;	/* misc flags */
	u_int sl_mccount;	/* # enabled multicast addrs */
	struct medium_addr *sl_mctab;	/* table of multicast addrs */
	u_long sl_minor;	/* minor device number */
	kmutex_t sl_lock;	/* protect this struct */
};


/* per-stream flags */
#define SLRAW           0x02	/* M_DATA plain raw mode */
#define SLALLPHYS       0x04	/* "promiscuous mode" */
#define SLALLMULTI      0x08	/* enable all multicast addresses */
#define SLALLSAP        0x10	/* enable all ether type values */



#define MYRI_LBUF_MIN   (8)	/* minimum num bufs for LANai */

enum BUFF_STATUS {
	BUF_NULL = 0, BUF_LANAI, BUF_LOANED, BUF_UNUSED, BUF_NOT_MAPPED
};

#define BUFFINFO_STATUS(x) \
		(x==BUF_NULL)?"NULL": \
		(x==BUF_LANAI)?"LANAI": \
		(x==BUF_LOANED)?"LOANED": \
		(x==BUF_UNUSED)?"UNUSED": \
		(x==BUF_NOT_MAPPED)?"NOT_MAPPED":"UNKNOWN"

struct buff_info {
	unsigned int *kptr;	/* kernel virtual pointer */
	unsigned int *dptr;	/* dma pointer for LANai */
	struct myri *myrip;	/* pointer to the owner of the buff */
	int index;		/* buffer index */
	int queue;		/* buffer used for which queue */
	int status;		/* free, in use etc */
	int next;		/* used for chaining possibly */
	ddi_dma_handle_t handle;/* from ddi_dma_alloc */
	frtn_t free_func;	/* free_function and parameters */
};

/*
	buffer layout

	recv0_buff[NUM_RECV_BUFF0]
	recv1_buff[NUM_RECV_BUFF1]
	snd_buff[NUM_SEND_BUFF]
*/

/*
#define MYRI_IMMED_LEN  (256+TCP_IP_HDR_SIZE+ETHER_HEADER+ MYRI_MAX_PAD_LEN)
*/
#define MYRI_IMMED_LEN  (0)

#define RECV0_LEN       (1024)
#define RECV1_LEN       (ALIGNED_SIZE)
#define SEND_LEN        (ALIGNED_SIZE)

#define NUM_RECV_BUFF0	((3*NUM_RECEIVES))	/* was 3x */
#define NUM_RECV_BUFF1	((3*NUM_RECEIVES)/2)	/* was 2x */

#define NUM_SEND_BUFF	(16)

#define BUFFER_BYTES    ( (RECV0_LEN * NUM_RECV_BUFF0) + \
			  (RECV1_LEN * NUM_RECV_BUFF1) + \
			  (SEND_LEN * NUM_SEND_BUFF) + \
			PAGESIZE + 16)



/*
 * Per-Device instance state information.
 *
 * Each instance is dynamically allocated on first attach.
 */
struct myri {

	struct myri *myri_nextp;/* next in linked list */
	volatile struct myriregs *myri_regsp;	/* addr of register set */
	dev_info_t *myri_dip;	/* dev info */
	struct medium_addr myri_ouraddr;	/* local address */
	u_int myri_flags;	/* misc. flags */
	u_int myri_wantw;	/* out of resources */
	kmutex_t myri_xmitlock;	/* protect send and send_ack Q */
	kmutex_t myri_recvlock;	/* protect recv_ack Q */
	kmutex_t myri_bufflock;	/* protect recv_buf Q */
	int xmit_count;
	int recv_count;
	int buff_count;
	caddr_t myri_arg;	/* device specific arg */
	ddi_iblock_cookie_t iblock_cookie;	/* cookie for intr */
	ddi_dma_handle_t send_handle;
	ddi_dma_handle_t send_ack_handle;
	ddi_dma_handle_t recv_handle;
	unsigned short sts;	/* save the LANai sts */

	/* counters to keep stats for netstat support */
	ulong_t myri_ipackets;	/* # packets received */
	ulong_t myri_ierrors;	/* # total input errors */
	ulong_t myri_opackets;	/* # packets sent */
	ulong_t myri_oerrors;	/* # total output errors */
	ulong_t myri_collisions;/* # collisions */

	u_short *lanai_memory;
	u_short *lanai_registers;
	u_short *lanai_memory_wake;
	u_short *lanai_registers_wake;

	int alive;

	int lanai_errors;	/* tracks problems with LANai */
	int lanai_reload;	/* number of times we tried to reload */

	unsigned int *buffer_ptr;
	struct buff_info recv0info[NUM_RECV_BUFF0];
	struct buff_info recv1info[NUM_RECV_BUFF1];
	struct buff_info sendinfo[NUM_SEND_BUFF];
	int last_send_buff;
	int buff_at_lanai[2];

	    u_int(*mlanai_intr) (LANAI_DATA *);	/* interrupt function for
						 * mmap() device */
	LANAI_DATA *mlanai_p;	/* pointer to device info for mmap() device */

};

/* flags */
#define MYRIRUNNING       0x01	/* chip is initialized */
#define MYRISLAVE         0x02	/* slave device (no DMA) */
#define MYRIPROMISC       0x04	/* promiscuous mode enabled */


/*
 * XXX Define offset to receive descriptor.
 */

#define MYRIHEADROOM	34

/*
 * Private DLPI full dlsap address format.
 * You need to specify a physical address (dl_phys) appropriate for
 * your medium
 */


struct myridladdr {
	struct medium_addr dl_phys;
	u_short dl_sap;
};

/*
 * Full DLSAP address length
 * You need to customize this based on your dladdr struct.
 */

#define DEVICE_ADDRL (sizeof (u_short) + MEDIUM_ADDRL)

#define DLADDRL	(80)		/* large upper bound on DL ADDR LEN */

#ifdef KSTAT

/*
 * "Export" a few of the error counters via the kstats mechanism.
 */
struct myristat {
	struct kstat_named les_ipackets;
	struct kstat_named les_ierrors;
	struct kstat_named les_opackets;
	struct kstat_named les_oerrors;
	struct kstat_named les_collisions;
};

#endif				/* KSTAT */

/* Some streams defines */

#define	DB_BASE(mp)		((mp)->b_datap->db_base)
#define	DB_LIM(mp)		((mp)->b_datap->db_lim)
#define	DB_REF(mp)		((mp)->b_datap->db_ref)
#define	DB_TYPE(mp)		((mp)->b_datap->db_type)

#define	MBLKL(mp)		((mp)->b_wptr - (mp)->b_rptr)
#define	MBLKSIZE(mp)	((mp)->b_datap->db_lim - (mp)->b_datap->db_base)
#define	MBLKHEAD(mp)	((mp)->b_rptr - (mp)->b_datap->db_base)
#define	MBLKTAIL(mp)	((mp)->b_datap->db_lim - (mp)->b_wptr)
#define	MBLKIN(mp, off, len)	(((mp)->b_rptr + off + len) <= (mp)->b_wptr)

#define	OFFSET(base, p)	((caddr_t)(p) - (caddr_t)(base))



/* defined in myri_code.h */
/* generated from lcp.dat */
extern unsigned short lanai_code_off;
extern unsigned short lanai_code[];
extern unsigned short lanai_data_off;
extern unsigned short lanai_data[];


/*
 * Function prototypes.
 */

extern void myriwenable(struct myri * myrip);
extern struct myristr *
myriaccept(struct myristr * slp, struct myri * myrip,
	   int type, struct medium_addr * addrp);

void myrisendup(struct myri *, mblk_t *, struct myristr * (*) ());
int do_putnext(queue_t * rq, mblk_t * mp, int first_time);

#ifdef  KSTAT
static void myristatinit(struct myri *);
#endif

void hex_dump(u_char * ptr, int len);
int
myri_update_list(unsigned id, ddi_dma_handle_t * handle,
		 mblk_t ** head, mblk_t ** tail, int op);
int myri_bad_alignment(mblk_t ** mptr_ptr);
int myri_tolanai(queue_t * wq, mblk_t * mp, struct myri * myrip);
void calc_scatter_vector(int len, int hdr_len, u_short * s_cnt, Scatter * scatter);
void myri_dma_error(int rv);
void myri_check_send_ack(struct myri * myrip);
void myri_lanai_error(struct myri * myrip);
void myri_init_lanai(struct myri * myrip);
unsigned short myri_what_sts(struct myri * myrip);
void myri_mutex_enter(struct myri * myrip);
void myri_mutex_exit(struct myri * myrip);
int myri_lanai_handshake(struct myri * myrip);
int myri_reset_lanai(struct myri * myrip);
int myri_handle_lanai_intr(struct myri * myrip, int count);
int myri_do_l2h(struct myri * myrip);
mblk_t *myri_do_recv(struct myri * myrip);
void lanai_set_ex1_unit(int unit, int n);
void myri_stop_lanai(struct myri * myrip);
void myri_dump_msg(mblk_t * mp);
int myriLoadLCPkernel(struct myri * myrip);

int myri_alloc_buffers(struct myri * myrip);
int myri_free_buff_all(struct myri * myrip);
void myri_add_buff(struct buff_info * infop);
void myri_dump_buff_info(struct buff_info *);

void copy_mblk_to_buffer(void *vptr, unsigned short *lptr, unsigned int length);

#ifdef DO_CKSUM
int verify_cksum(unsigned char *head, mblk_t * m);
#endif DO_CKSUM


int putq(queue_t * q, mblk_t * mp);
int putnext(queue_t * q, mblk_t * mp);
int canputnext(queue_t * q);


extern ddi_dma_lim_t myri_dma_lim;
extern u_short *LANAI[MYRI_MAX_BOARDS];
extern struct myri *myridev;



#ifdef DO_TIMING
#define GET_TIME(t_val) uniqtime(&(t_val))
#define SUB_TIME(t,sub) {               \
                t.tv_sec-=sub.tv_sec;   \
                t.tv_usec-=sub.tv_usec; \
        if (t.tv_usec < 0) {    \
                    t.tv_sec--;         \
                    t.tv_usec+=1000000; \
                }                       \
                if (t.tv_sec < 0) {     \
                    t.tv_sec = 0;       \
                    t.tv_usec = 0;      \
                    } \
        }
#define TIMING_PRINT	0
#else
#define GET_TIME(t_val) {}
#define SUB_TIME(t,sub) {}
#define TIMING_PRINT	100
#endif 


/*
 * This file is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify this file without charge, but are not authorized to
 * license or distribute it to anyone else except as part of a product
 * or program developed by the user.
 *
 * THIS FILE IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 *
 * This file is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 *
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY THIS FILE
 * OR ANY PART THEREOF.
 *
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even
 * if Sun has been advised of the possibility of such damages.
 *
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */

/*
 * psli.h header for the Sample Streams DLPI Driver
 */



/* end of myri.h */
