/*									tab: 2
 *
 * lam.c - Lanai Active Messages 
 *
 * 
 * "Copyright (c) 1994 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 		Richard P. Martin and Lok T. Liu
 * Version:		144
 * Creation Date:	Tue Feb 28 18:15:42 1995
 * Filename:		lam.c
 * History:
 *	RPM	144	Sun Aug  6 20:41:28 1995
 *		added support for 2 endpoints
 *	LTL	19	Thu Jun 29 14:41:37 1995
 *		Added non-SPMD support
 *	RPM	2	Sat Jun  3 19:56:10 1995
 *		added relative addressing code. Fixed Q size problem.
 */

#include  <stdio.h>
#include  <stdlib.h>
#include  <unistd.h>
#include  "lanai_dev.h"
#include <signal.h>
#include "config.h"
#include <sys/types.h>

#include <sys/ipc.h>
#include <sys/shm.h>

#include <sys/time.h>
#include "ep.h"
#include "ifc.h"
#include "am1_5.h"

#ifdef USE_GLUNIX
/* GLUnix interface */
#include <cam.h>
#include <glib.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#endif 

#ifdef DEBUG
#include <assert.h>
#define __eprintf printf
#endif 

#define LAM_CONFIG_DEFAULT "/usr/now/etc/myrinet_routes/lam_conf_" 
#define LAM_LOCK_DEFAULT  "/usr/now/etc/myrinet_lock/" 

#define LAM_SHMEM_KEY     0x700328
#define LAM_SHMEM_SIZE    0x1000
#define PERMS             0666

#ifdef SOLARIS
#include <sys/systeminfo.h> 
#define gethostname(n,l)  sysinfo(SI_HOSTNAME,n,l)
#else
#define memmove(d,s,l) bcopy(s,d,l)
#endif

#ifdef __GNUC__
#define INLINE inline
#else 
#define INLINE 
#endif


unsigned short sts;
int copy_block_len,units;
int unit;

volatile ep_t *endpt;
volatile route_t *routes;
volatile desc_t *in_d,*out_d;
int in_num,out_num;

volatile int *dma_base;
volatile int *user_base;

int *lam_dma_base;

int lam_myproc;
int lam_procs =2;
int lam_reply =0;  /* which node to reply to */

int flow_cntl[MAX_REMOTE];
int req_flow_in[MAX_REMOTE];
int req_flow_out [MAX_REMOTE];

int am_poll_count =0;

int lam_my_ep;
int lam_other_ep;
int lam_ep_0,lam_ep_1;

unsigned int *lam_shmem_addr; 

int rep_flow_in[MAX_REMOTE];
int rep_flow_out[MAX_REMOTE];
int self_buf[LAM_PKT_WORDS];  /* buffer medium-sized message to self */

uint seq_glob;

typedef void (*short_handler_t)(int src,int arg0, int arg1,int arg2, int arg3,
													int arg4);
typedef void (*xfer_handler_t)(int src,int arg0, int arg1,int arg2);
typedef void (*medium_handler_t)(int src,void *buf,int len,int arg0,int arg1);
extern unsigned int LANAI_fd[64];


int  
am_request_4(int dest, int handler_index, int arg0, int arg1, 
		  int arg2, int arg3);

int  
am_reply_4(int dest, int handler_index, int arg0, int arg1, 
		  int arg2, int arg3);

int 
am_poll();

short_handler_t 
lam_null_handler(int src,int arg0, int arg1,int arg2, int arg3) {

}

extern int am_barrier_init();

/*
 * Support for indirect active messages
 *
 */

static short_handler_t handler_table[MAX_HANDLERS];
static seg_t seg_table[MAX_SEGMENTS];

/* return maximum number of entries in the handler table */
int 
am_max_handlers() {
  return MAX_HANDLERS;
}

/* This function sets handler table entry indexed by handler_index to func.
 * Valid handler_indexes are [0, am_max_handlers()] and -1.
 * handler_index of -1 will ask the system to pick an unused index.
 * Setting a used entry without removing it first is an error.
 * This function returns the  table index on success or -1 otherwise.
 */
int 
am_add_handler(int handler_index, short_handler_t func) {
  int rcode, i;
	
  /* if user specifies index */
  if ((handler_index >= 0) && (handler_index < MAX_HANDLERS)) {
    /* if entry not used */
    if (handler_table[handler_index] == (short_handler_t)abort) {
      handler_table[handler_index] = func;
      rcode = handler_index;
    }
    else {
      rcode = -1;
    }
  }
  else if (handler_index == -1) {
    /* look for a free index using linear search */
    i = 0;
    while ((handler_table[i] != (short_handler_t)abort) && (i < MAX_HANDLERS)) {
      ++i;
    }
    if (i < MAX_HANDLERS) {
      handler_table[i] = func;
      rcode = i;
    }
    else {
      rcode = -1;
    }
  }
  else
    rcode = -1;
  return rcode;
}

/* This function removes the handler table entry indexed by handler_index
 * and replaces it with abort(). Removing an unused entry is an error.
 * Return 0 on sucess or -1 otherwise.
 */
int 
am_remove_handler(int handler_index) {
  /* if entry used */
  if ((short_handler_t) handler_table[handler_index] != 
      (short_handler_t)abort) {
    handler_table[handler_index] = (short_handler_t)abort;
    return 0;
  }
  else {
    return -1;
  }
}

/* return maximum number of entries in the segment table */
int 
am_max_segments() {
  return MAX_SEGMENTS;
}

/* This function sets the base and length of a virtual memory segment
 * indexed by seg_index. Valid segment_indexes are [0, am_max_segments()]
 * and -1. segment_index of -1 will ask the system to pick an unused index.
 * Setting a used entry without removing it first is an error.
 * This function returns the  table index on success or -1 otherwise.
 */
int 
am_set_segment(int seg_index, void *base, int len) {
  int rcode, i;
	
  /* if user specifies index */
  if ((seg_index >= 0) && (seg_index < MAX_SEGMENTS)) {
    /* if entry not used */
    if (seg_table[seg_index].base == 0) {
      seg_table[seg_index].base = base;
      seg_table[seg_index].len = len;
      rcode = seg_index;
    }
    else {
      rcode = -1;
    }
  }
  else if (seg_index == -1) {
    /* look for a free index using linear search */
    i = 0;
    while ((seg_table[i].base != 0) && (i < MAX_SEGMENTS)) {
      ++i;
    }
    if (i < MAX_SEGMENTS) {
      seg_table[i].base = base;
      seg_table[i].len = len;
      rcode = i;
    }
    else {
      rcode = -1;
    }
  }
  else
    rcode = -1;
  return rcode;
}

/* This function frees the segment table entry indexed by seg_index
 * and replaces the base and length with 0. Freeing an unused entry is an
 * error. Return 0 on sucess or -1 otherwise. Sending data to an unused segment
 * will cause the program to abort.
 */
int 
am_free_segment(int seg_index) {
  /* if entry used */
  if (seg_table[seg_index].base != 0) {
    seg_table[seg_index].base = 0;
    seg_table[seg_index].len = 0;
    return 0;
  }
  else {
    return -1;
  }  
}



void 
am_wait(volatile int *flag,int value) {
  do {
    am_poll();
  } while (*flag < value);
 
  *flag -= value;
}

void 
am_poll_wait(volatile int *flag,int value) {

  while (*flag < value) {
    am_poll();
  }
 
  *flag -= value;
}

print_out_log () {
  int i;

}

print_in_log () {
  int i;
}

static int  
am_reply_0(int dest,short_handler_t fun) {

#ifdef DEBUG
  rep_flow_out[dest]++;
  assert( (dest >=0) && (dest < lam_procs));
#endif 

  lam_reply = -1;

  out_d->hdr.s_data[0] = (int) fun;
  out_d->op_dest = (int)  (AM_TYPE)|(dest<<DEST_SHIFT)|(REPLY)|(SHORT);

  out_num = (out_num +1) & OUT_SIZE_MASK;
  out_d = &(endpt->out[out_num]);

  return 0;
}


int 
am_poll(){
  register int op_dest,src,h_index;

#ifdef DEBUG
  am_poll_count++;
#endif 

  op_dest = in_d->op_dest;

  while ( (op_dest&OP_MASK)  != 0 ) { /* if buffer has anything */

    src = (unsigned) (op_dest & DEST_MASK) >> DEST_SHIFT;
    h_index = in_d->hdr.s_data[0]; 

    if ( (h_index <= 0) || ( h_index > MAX_HANDLERS)) {
      fprintf(stderr,"node %d: remote node %d sent invalid handler 0x%0x\n",
	      lam_myproc,src,h_index);
      goto error;
    }

    if ( (op_dest&REQ_MASK) == REQUEST) {
      lam_reply =src;
#ifdef DEBUG
      req_flow_in[src]++;
#endif 
    } else {
      flow_cntl[src]++;
#ifdef DEBUG
      rep_flow_in[src]++;
#endif 

    }

    if ( (op_dest &OP_MASK) == SHORT) {

      ( *( (void (*)()) handler_table[h_index]))
	(src,
	 in_d->hdr.s_data[1],
	 in_d->hdr.s_data[2],
	 in_d->hdr.s_data[3],
	 in_d->hdr.s_data[4],
	 in_d->hdr.s_data[5]);
      
      in_d->op_dest = in_d->op_dest & (~OP_MASK) ;
      in_num = (in_num+1)&IN_SIZE_MASK;
      in_d = &(endpt->in[in_num]);

    }
    else {
				
      if ((op_dest &OP_MASK) == MEDIUM ) {

	{
	  register uint word1,word2,word3,word4,word5;

	  word5 = in_d->hdr.b_hdr.data[3]; 

	  word1 = in_d->hdr.b_hdr.len; /* length the Lanai gives us */
	  /*word2 = in_d->hdr.b_hdr.data[0];  handler index */
	  word3 = in_d->hdr.b_hdr.data[1]; /* arg0 */
	  word4 = in_d->hdr.b_hdr.data[2]; /* arg1 */

	  word5 = word5 & 0x3;	/* number of unaligned bytes */


	  if (word5 != 0) {

	    if (word1 > 4) {
	      word1 = ( (word1-4)&(~0x3))+word5;
	    }
	    else {
	      word1 = word5;	    
	    }
	  }
	  ( (medium_handler_t) handler_table[h_index])
	    (src,
	     (char *) ((uint *) user_base + (LAM_PKT_WORDS*in_num)),
	     word1, word3, word4);		
	}
	in_d->op_dest = in_d->op_dest & (~OP_MASK) ;
	in_num = (in_num+1)&IN_SIZE_MASK;
	in_d = &(endpt->in[in_num]);
			
      }
				
      else if ((op_dest &OP_MASK) == BULK ) {

	{
	  register uint word1,word2,word3,word4,word5,offset;


	  word5 = in_d->hdr.b_hdr.data[3]; 
		offset =(word5 & OFFSET_MASK) >> OFFSET_SHIFT;
	  word5 = word5 & 0x3; /* number of unaligned bytes */
		
	  word1 = in_d->hdr.b_hdr.len;      /* length the Lanai gives us */
	  /* word2 = in_d->hdr.b_hdr.data[0];  handler index */
    /* dest addr */		
	  word3 = (unsigned)seg_table[in_d->hdr.b_hdr.data[1]].base + offset;
	  word4 = in_d->hdr.b_hdr.data[2]; /* arg */


#ifdef DEBUG
	  assert( ((word1 + offset)-word5)
		 <= seg_table[in_d->hdr.b_hdr.data[1]].len);
#endif

	  if (word5 == 0) {

	    memcpy((char *) word3,
		   (char*) ((uint *) user_base + (LAM_PKT_WORDS*in_num)),
		   word1) ;
	  }
	  else {
	    if (word1 > 4) {
	      word1 = ( (word1-4)&(~0x3))+word5;
	    }
	    else {
	      word1 = word5;
	    }
	    memcpy((char *) word3, 
		   (char *) ((uint *) user_base + (LAM_PKT_WORDS*in_num)),
		   word1);				
	  }
	  ( (xfer_handler_t) handler_table[h_index])(src,word3,word1,word4);
	}
	in_d->op_dest = in_d->op_dest & (~OP_MASK) ;
	in_num = (in_num+1)&IN_SIZE_MASK;
	in_d = &(endpt->in[in_num]);
			
      }			
      else {

	/* CRC ERROR */
	printf("Wank! got crc error.\n");
	printf("in_num = 0x%0x in_d = 0x%0x\n",in_num,in_d);
	printf("op_dest= 0x%0x s_data[0]= 0x%0x s_data[1]= 0x%0x\n",
	       in_d->op_dest, in_d->hdr.s_data[0], in_d->hdr.s_data[1]);
	printf("s_data[2]= 0x%0x s_data[3]= 0x%0x s_data[4]= 0x%0x\n",
	       in_d->hdr.s_data[2], in_d->hdr.s_data[3], in_d->hdr.s_data[4]);
	printf("s_data[5]= 0x%0x crc= 0x%0x\n",
	       in_d->hdr.s_data[5], in_d->crc);

      error:
	in_d->op_dest = in_d->op_dest & (~OP_MASK) ;
	in_num = (in_num+1)&IN_SIZE_MASK;
	in_d = &(endpt->in[in_num]);
	
      }
    }

    if ( (op_dest & REQ_MASK) == REQUEST ) { /* was a request or reply? */
      if (lam_reply == src) 
    	am_reply_0(src, NULL_HANDLER);
    }
    
    op_dest = in_d->op_dest;
  } /* end while op_dest */

  return 0;
}


/* The following are the indirect version of the AM functions. All handlers
 * are named by a handler index. Remote virtual memory regions are named
 * by a segment index.
 */

int 
am_request_4(int dest, int handler_index, int arg0, int arg1, 
	     int arg2, int arg3) {

  int poll_count =0;
  int i;
#ifdef DEBUG
  int print_times =0;
#endif 
  
  
  if (dest != lam_myproc) {

    while (flow_cntl[dest] ==0) {
      poll_count++;
      am_poll();

      if (poll_count > MAX_POLL_COUNT) {
	printf("am_request_4: failed to inject message to %d \n", dest);
	return -1;
      }
#ifdef DEBUG      
      print_times++;
      /*      printf("req_4 node %d out of credits to %d times %d count %d \r",
	      lam_myproc,dest,print_times,am_poll_count); */
#endif 
    }

#ifdef DEBUG
    assert( (dest >=0) && (dest < lam_procs));

    while ( ((out_d->op_dest)&OP_MASK) != 0) {
      am_poll();
      printf("\nWarning, am_request_4 found output queue full! \n"); 
    }
#endif 
    flow_cntl[dest]--;
#ifdef DEBUG
    req_flow_out[dest]++;
#endif 
    
    out_d->hdr.s_data[0] = (int) handler_index;
    out_d->hdr.s_data[1] = (int) arg0;
    out_d->hdr.s_data[2] = (int) arg1;  
    out_d->hdr.s_data[3] = (int) arg2;
    out_d->hdr.s_data[4] = (int) arg3;

    (volatile) out_d->op_dest = i =
      (int) (AM_TYPE)|(dest<<DEST_SHIFT)|(REQUEST)|(SHORT);
    
    out_num = (out_num +1) & OUT_SIZE_MASK;
    out_d = &(endpt->out[out_num]);

    am_poll();
  }
  else {
    (handler_table[handler_index])(lam_myproc,arg0, arg1, arg2, arg3, 0);
    am_poll();
  }
  return 0;
}


int 
am_request_5(int dest, int handler_index, int arg0, int arg1, 
	     int arg2, int arg3, int arg4) {
  int i;
  int poll_count = 0;

#ifdef DEBUG
  int print_times =0;
#endif 
  
  if (dest != lam_myproc ) {
    while (flow_cntl[dest] ==0) {
      poll_count++;
      am_poll();

      if (poll_count > MAX_POLL_COUNT) {
	printf("am_request_5: failed to inject message to %d \n", dest);
	return -1;
      }
#ifdef DEBUG
      print_times++;
      /*      printf("req_5 node %d out of credits to %d times %d count %d \r",
	      lam_myproc,dest,print_times,am_poll_count); */
#endif 

    }

    flow_cntl[dest]--;
#ifdef DEBUG
    assert( (dest >=0) && (dest < lam_procs));
    req_flow_out[dest]++;
#endif 
    
#ifdef DEBUG
    while ( ((out_d->op_dest)&OP_MASK) != 0) {
      am_poll();
      printf("\nWarning, am_request_5 found output queue full! \n");
    }
#endif 

    out_d->hdr.s_data[0] = (int) handler_index;
    out_d->hdr.s_data[1] = (int) arg0;
    out_d->hdr.s_data[2] = (int) arg1;  
    out_d->hdr.s_data[3] = (int) arg2;
    out_d->hdr.s_data[4] = (int) arg3;
    out_d->hdr.s_data[5] = (int) arg4;

    out_d->op_dest = i = 
      (int) (AM_TYPE)|(dest<<DEST_SHIFT)|(REQUEST)|(SHORT);

    out_num = (out_num +1) & OUT_SIZE_MASK;
    out_d = &(endpt->out[out_num]);

    am_poll();
  }
  else {
    (handler_table[handler_index])(lam_myproc,arg0, arg1, arg2, arg3,arg4);
    am_poll();
  }

  return 0;
}


/* almost the same implementation as am_request_xfer() */
int 
am_request(int dest, int handler_index, void *lva, int nbytes,
	   int arg0, int arg1) {

  int i;
  int poll_count =0;

  if (nbytes <= (LAM_PKT_WORDS*sizeof(int))) {

    if ( dest != lam_myproc ) {

      while (flow_cntl[dest] == 0) {
	poll_count++;
      	am_poll();
	if (poll_count > MAX_POLL_COUNT) {
	  printf("am_request: failed to inject message to %d \n", dest); 
	  return -1;
	}
      }

#ifdef DEBUG      
      assert(nbytes > 0 );	/* transfer should be >0 < max size */
      assert(nbytes <= LAM_PKT_WORDS*4 );

      assert( (dest >=0) && (dest < lam_procs)); /* dest oK? */
      while ( ((out_d->op_dest)&OP_MASK) != 0) {
        am_poll();
       	printf("\nWarning, am_request found output queue full! \n");
      }
#endif 
      
      flow_cntl[dest]--;  

#ifdef DEBUG
      req_flow_out[dest]++;
#endif 
      /* copy into io space buffer */
      /* need to check max len, assume that all buffers are word aligned */
      /* and in multiple of words */

      memcpy((int *) (user_base+((LAM_PKT_WORDS*IN_Q_SIZE)+
				 (out_num*LAM_PKT_WORDS))),
	     (char *) lva, nbytes);
      
      out_d->hdr.b_hdr.la = (int *) lam_dma_base +
       	(LAM_PKT_WORDS*IN_Q_SIZE)+(out_num*LAM_PKT_WORDS);
      
      out_d->hdr.b_hdr.len = nbytes;
      out_d->hdr.b_hdr.data[0] = (int) handler_index;
      out_d->hdr.b_hdr.data[1] = (int) arg0;
      out_d->hdr.b_hdr.data[2] = (int) arg1;
      /* store 'extra' bytes */			
      out_d->hdr.b_hdr.data[3] = (nbytes & 0x3);
      
      if ( (nbytes & 0x3) != 0) { /* round up to upper byte */
	out_d->hdr.b_hdr.len = (nbytes+4) & (~0x3);
      }
      
      out_d->op_dest = i = 
	(int) (AM_TYPE)|(dest<<DEST_SHIFT)|(REQUEST)|(MEDIUM); /*go*/

      out_num = (out_num +1) & OUT_SIZE_MASK;
      out_d = &(endpt->out[out_num]);
      am_poll();
      return 0;
    }
    else {
      memmove(self_buf, lva, nbytes);
      (handler_table[handler_index])(dest, (int)self_buf, nbytes,
				     arg0, arg1, 0);   
      am_poll();
      return 0;
    }
  }
  else {
    printf("am_request called with size > %d\n",LAM_PKT_WORDS*sizeof(int));
  }	

  return -1;

}


/* Copy nbytes bytes of data from address lva on the calling node to
 * address (base+offset) of the segment indexed by seg_index on dest node.
 * Invoke handler indexed by request_handler_index with handler_arg
 * on dest node when all data are received on dest node.
 * Offset is between 0 and 2^29 - 1 inclusive. 
 * Return to caller when the data is sent.
 */
int  
am_request_xfer(int dest, void *lva, int seg_index, int offset, int nbytes,
		int handler_index, void *arg0) {
  
  int i;
  int poll_count; 

  if (nbytes <= (LAM_PKT_WORDS*sizeof(int))) {

    if ( dest != lam_myproc ) {

      while (flow_cntl[dest] == 0) {
	poll_count++;
      	am_poll();
	if (poll_count > MAX_POLL_COUNT)  {
	  printf("am_request_xfer: failed to inject message to %d \n", dest);
	  return -1;
	}
      }

#ifdef DEBUG      
      assert(nbytes > 0 );	/* transfer should be >0 < max size */
      assert(nbytes <= LAM_PKT_WORDS*4 );

      assert( (dest >=0) && (dest < lam_procs)); /* dest oK? */
      while ( ((out_d->op_dest)&OP_MASK) != 0) {
	am_poll();
	printf("\nWarning, am_request_xfer found output queue full! \n");
      }
#endif 
      
      flow_cntl[dest]--;  

#ifdef DEBUG
      req_flow_out[dest]++;
#endif 
      /* copy into io space buffer */
      /* need to check max len, assume that all buffers are word aligned */
      /* and in multiple of words */

      memcpy((int *) (user_base+((LAM_PKT_WORDS*IN_Q_SIZE)+
				 (out_num*LAM_PKT_WORDS))),
	     (char *) lva, nbytes);
      
      out_d->hdr.b_hdr.la = (int *) lam_dma_base +
	(LAM_PKT_WORDS*IN_Q_SIZE)+(out_num*LAM_PKT_WORDS);
      
      out_d->hdr.b_hdr.len = nbytes;
      out_d->hdr.b_hdr.data[0] = (int) handler_index;
      out_d->hdr.b_hdr.data[1] = (int) seg_index;
      out_d->hdr.b_hdr.data[2] = (int) arg0;
      /* store 'extra' bytes */			
      out_d->hdr.b_hdr.data[3] = (nbytes & 0x3) | (offset << OFFSET_SHIFT); 
      
      if ( (nbytes & 0x3) != 0) { /* round up to upper byte */
       	out_d->hdr.b_hdr.len = (nbytes+4) & (~0x3);
      }
      
      out_d->op_dest = i =
	(int) (AM_TYPE)|(dest<<DEST_SHIFT)|(REQUEST)|(BULK); /*go*/

      out_num = (out_num +1) & OUT_SIZE_MASK;
      out_d = &(endpt->out[out_num]);
      am_poll();
      return 0;
    }
    else {
      memmove(seg_table[seg_index].base + offset ,lva, nbytes);
      (handler_table[handler_index])(dest,
				     (unsigned)seg_table[seg_index].base + offset,
				     nbytes, (int)arg0, 0, 0);    
      am_poll();
      return 0;
    }
  }
  else {
    printf("am_request_xfer called with size > %d\n",LAM_PKT_WORDS*sizeof(int));
  }
  return -1;

}

/* Copy nbytes bytes of data from address lva on the calling node to
 * address (base+offset) of the segment indexed by seg_index on dest node.
 * Invoke handler indexed by request_handler_index with handler_arg
 * on dest node when all data are received on dest node.
 * Offset is between 0 and 2^29 - 1 inclusive.
 * Return to caller when the request to send is posted. The source buffer
 * should not be modified until the matching reply handler is invoked.
 */
int  
am_request_xfer_async(int dest_node, void *src_buf, int seg_index, int offset,
		      int len, int handler_index, void *request_handler_arg1) {
  int i; 

  i =am_request_xfer(dest_node, src_buf, seg_index, offset, len,
		     handler_index, request_handler_arg1);

  return i;

}


int 
am_reply_4(int dest, int handler_index, int arg0, int arg1, 
	   int arg2, int arg3) {
  
  if ( dest != lam_myproc ) {

#ifdef DEBUG      
    assert( ((out_d->op_dest)&OP_MASK) == 0 );
    assert( (dest >=0) && (dest < lam_procs)); /* dest oK? */
    rep_flow_out[dest]++;
#endif 

    lam_reply = -1;

    out_d->hdr.s_data[0] = (int) handler_index;
    out_d->hdr.s_data[1] = (int) arg0;
    out_d->hdr.s_data[2] = (int) arg1;  
    out_d->hdr.s_data[3] = (int) arg2;
    out_d->hdr.s_data[4] = (int) arg3;

    out_d->op_dest = (int)  (AM_TYPE)|(dest<<DEST_SHIFT)|(REPLY)|(SHORT);

    out_num = (out_num +1) & OUT_SIZE_MASK;
    out_d = &(endpt->out[out_num]);
  }
  else {
    (handler_table[handler_index])(lam_myproc,arg0, arg1, arg2, arg3, 0);
  }

  return 0;
}


int 
am_reply(int dest, int handler_index, void *lva, int nbytes,
	     int arg0, int arg1) {
  if ( dest != lam_myproc ) {

#ifdef DEBUG
    assert( ((out_d->op_dest)&OP_MASK) == 0 );
    assert( (dest >=0) && (dest < lam_procs)); /* dest oK? */
    rep_flow_out[dest]++;
#endif 

    lam_reply = -1;

    /* copy into io space buffer */
    /* need to check max len, assume that all buffers are word aligned */
    /* and in multiple of words */
    memcpy( (char *)  (user_base+ ((LAM_PKT_WORDS*IN_Q_SIZE)+
				   (out_num*LAM_PKT_WORDS))),
	   (char *) lva, nbytes);
    out_d->hdr.b_hdr.la = (int *) lam_dma_base +
      (LAM_PKT_WORDS*IN_Q_SIZE)+(out_num*LAM_PKT_WORDS);

    out_d->hdr.b_hdr.len = nbytes;
    out_d->hdr.b_hdr.data[0] = handler_index;
    out_d->hdr.b_hdr.data[1] = (int) arg0;
    out_d->hdr.b_hdr.data[2] = (int) arg1;
    /* store 'extra' bytes */		
    out_d->hdr.b_hdr.data[3] = (nbytes & 0x3);

    if ( (nbytes & 0x3) != 0) {
      out_d->hdr.b_hdr.len = (nbytes+4) & (~0x3);
    }

    out_d->op_dest = (int) (AM_TYPE)|(dest<<DEST_SHIFT)|(REPLY)|(MEDIUM);
    out_num = (out_num +1) & OUT_SIZE_MASK;
    out_d = &(endpt->out[out_num]);
  }  
  else {
    memmove(self_buf, lva, nbytes);
    (handler_table[handler_index])(dest, (int)self_buf,
				   nbytes, arg0, arg1, 0);    
  }

  return 0; 

}


/* Copy nbytes bytes of data from address lva on the calling node to
 * base address of the segment indexed by seg_index on dest node.
 * Invoke handler indexed by reply_handler_index with handler_arg
 * on dest node when all data are received on dest node.
 * Offset is between 0 and 2^29 - 1 inclusive. 
 * Return to caller when the data is sent. Should only be called inside
 * a request handler.
 */
int  
am_reply_xfer(int dest, void *lva, int seg_index, int offset, int nbytes,
	      int handler_index, void *arg0) {

  if ( dest != lam_myproc ) {

#ifdef DEBUG
    assert( ((out_d->op_dest)&OP_MASK) == 0 );
    assert( (dest >=0) && (dest < lam_procs)); /* dest oK? */
    rep_flow_out[dest]++;
#endif 

    lam_reply = -1;

    /* copy into io space buffer */
    /* need to check max len, assume that all buffers are word aligned */
    /* and in multiple of words */
    memcpy( (char *)  (user_base+ ((LAM_PKT_WORDS*IN_Q_SIZE)+
				   (out_num*LAM_PKT_WORDS))),
	   (char *) lva, nbytes);
    out_d->hdr.b_hdr.la = (int *) lam_dma_base +
      (LAM_PKT_WORDS*IN_Q_SIZE)+(out_num*LAM_PKT_WORDS);

    out_d->hdr.b_hdr.len = nbytes;
    out_d->hdr.b_hdr.data[0] = handler_index;
    out_d->hdr.b_hdr.data[1] = (int) seg_index;
    out_d->hdr.b_hdr.data[2] = (int) arg0;
    /* store 'extra' bytes */		
    out_d->hdr.b_hdr.data[3] = (nbytes & 0x3) | (offset << OFFSET_SHIFT); 

    if ( (nbytes & 0x3) != 0) {
      out_d->hdr.b_hdr.len = (nbytes+4) & (~0x3);

    }

    out_d->op_dest = (int) (AM_TYPE)|(dest<<DEST_SHIFT)|(REPLY)|(BULK);
    out_num = (out_num +1) & OUT_SIZE_MASK;
    out_d = &(endpt->out[out_num]);
  }  
  else {
    memmove(seg_table[seg_index].base + offset, lva, nbytes);
    (handler_table[handler_index])
      (dest,(unsigned)seg_table[seg_index].base + offset,
       nbytes, (int)arg0, 0, 0);    
  }
  return 0; 

}


int 
am_procs() {
  return lam_procs;
}

int 
am_my_proc() {
  return lam_myproc;
}

int 
am_max_size() {
  return LAM_PKT_WORDS*4;
}

int 
am_init_tables() {
  int i;

  for (i = 0; i < MAX_HANDLERS; ++i)
    handler_table[i] = (short_handler_t)abort;

  for (i = 0; i < MAX_SEGMENTS; ++i) {
    seg_table[i].base = 0;
    seg_table[i].len = 0;
  }

  if (am_add_handler(NULL_HANDLER, (short_handler_t)lam_null_handler)
      != NULL_HANDLER) {
    fprintf(stderr, "Failed to add lam_null_handler()\n");
    exit(-1);
  }
}



#ifdef USE_GLUNIX
/* get configuration from GLUNIX */
void 
LAM_get_glu_config() {
  pid_t pid;
  Npid glunix_pid;
		
  /* get total number of nodes */
  lam_procs = Glib_GetParallelDegree();

    if (lam_procs <= 0) {
      printf("GLUNIX reports that this proces is not part of a parallel program: AM_procs = %d\n", lam_procs);
      exit(-1);
    }
    /* get my node number */
    lam_myproc = Glib_GetMyVnn();

  pid = getpid();
  if (pid < 0) {
    perror("Cannot get local PID");
    exit(-1);
  }
  /* get the GLUNIX pid to agree on a server port number */
  glunix_pid = Glib_GetMyNpid();

  if (glunix_pid == NPID_NOBODY) {
    printf("GLUNIX reports that this proces is not part of a parallel program: glunix_pid = %d\n", glunix_pid);
    exit(-1);
  }

}
#endif 


int 
am_grab_lock_or_die(char *my_name) {
  register int tmp_reg; 
  volatile int *mem_addr; 
  pid_t my_pid,other_pid; 
  int shmem_id;
  int tries,i; 
  
  if (( shmem_id = shmget(LAM_SHMEM_KEY, LAM_SHMEM_SIZE,
			       (PERMS|IPC_CREAT)))<0) {
    perror("can't open shared segment to get lock");
    exit(1);
  }

  my_pid = getpid();
  tries =0;
  /* check if someone is using the card */

  lam_shmem_addr = (void *) shmat(shmem_id,(char *) 0 , 0);

 try_lock:

  tmp_reg = 0; 
  AM_SWAP( (*lam_shmem_addr) ,tmp_reg);
  tries++;
  if (tmp_reg == 0) {
    /* see if this person is around */
    other_pid = (pid_t) lam_shmem_addr[1];

    if ( (other_pid == 0) || (other_pid == my_pid))  {
       lam_shmem_addr[1]= my_pid;
       AM_INIT_LOCK( lam_shmem_addr[0]);
       goto try_lock;
    }

    i = kill(other_pid,0); 
    if (i == 0) {  /* eek, process exists */
      fprintf(stderr,"Process %d is holding the lock for the LANai card on %s\n",other_pid,my_name);
      exit(1);
    }
    else { /* doesn't exist? reset and try again */
      if (tries > 100) {
	fprintf(stderr,"Process %d failed to aquire lock after %d tries on %s\n",my_pid,my_name,tries);
	exit(1);
      }
      AM_INIT_LOCK( lam_shmem_addr[0]);
      goto try_lock;
    }
  }
  /*ok, got the lock  */
  lam_shmem_addr[1] = my_pid; 
}
  

int
am_enable(int use_glunix) {	

  int i,j,port,hop;
  int port_num[MAX_REMOTE];
  char *p, *s,*first_dot;
  char *home;
  char *config_path;
  static FILE *config_file;
  static char config_filename[256];
  char lcp_file[256];
  char node_name[256];
  char my_hostname[256];
  char remote_host[256];
  int *remote_ip_num;
  char temp_name[256];
  char paths[MAX_REMOTE][MAX_HOPS]; /* eight hops for now */
  char *dest_route,temp_path[8];
  
  gethostname(my_hostname,80);
  
  /* truncate hostname to first dot */
  if ( (first_dot = (char *) strchr(my_hostname,'.')) != (char *) NULL) {
    first_dot[0] = '\0';
  }
  
  /* set up default filename path */
  strcpy(config_filename, LAM_CONFIG_DEFAULT);
  strcat(config_filename,my_hostname);
  
  /* if the environment var is set up, use that */
  if ( ((config_path = getenv("LAM_CONFIG")) != (char *)NULL) ) {
    strcpy(config_filename, config_path);

  }
  
  config_file = fopen(config_filename, "r");
  if (config_file == (FILE *)0) {			
    printf("Cannot open configuration file %s\n", config_filename);
    exit(-1);
  }
  
  (void) fscanf(config_file,"%s ",lcp_file);
  
  unit = 0;
  units = open_lanai_copy_block(&copy_block_len,&sts);	
  if (copy_block_len<=0) {
    perror("ops: open_copy_block()");
    printf("error! Length = %d  <= 0\n",copy_block_len);
  }
  else {
#ifdef DEBUG
    printf("ops: User opened copy_block \n"); 
    printf("ops: user_ptr = %x    dma_ptr = %x    copy_block_len = 0x%x\n",
	   (unsigned)UBLOCK[unit],(unsigned)DBLOCK[unit],copy_block_len);
#endif 
  }
  
  
  if ( copy_block_len < 
      ((IN_Q_SIZE+OUT_Q_SIZE)*LAM_PKT_WORDS*sizeof(int))) {
    
    fprintf(stderr, "warning! copy_block_len to short, len %d bytes wanted %d\n",
	    copy_block_len,
	    (IN_Q_SIZE+OUT_Q_SIZE)*LAM_PKT_WORDS*sizeof(int));
  }

  am_grab_lock_or_die(my_hostname);

  lanai_interrupt(0);
  lanai_reset(1);
  clear_load(0,lcp_file);
  memset( (char*) LANAI[0] ,0,128*1024) ;
  clear_load(0,lcp_file);
  
  lanai_reset(0);
  while (! L_INIT_DONE);     /* wait for lanai init done */
  
  /* set up routes */
  endpt = (ep_t *) ( (char *) LANAI[0] +  ((uint)BULK_PTR*2) );
  routes = (route_t *) ( (char *) LANAI[0] + ((uint) ROUTE_TAB*2) );

#ifdef DEBUG  
  printf("LANAI BASE %0x, ROUTE_TAB %0x, IN_BASE %0x, OUT_BASE %0x, BULK_PTR %0x\n", LANAI[0], ROUTE_TAB,IN_BASE, 
	 OUT_BASE , BULK_PTR);	   
  
  fflush(stdout);
#endif 
  
  if (fscanf(config_file, "%d ", &lam_procs) == EOF) {
    fprintf(stderr,"Cannot read number of nodes in configuration file\n");
    fflush(stderr);
    exit(-1);
  }
  
  lam_myproc = -1;
#ifdef USE_GLUNIX
  /* has side effect of resetting lam_procs and lam_myproc */
 if (use_glunix) {
   LAM_get_glu_config();
 }
#endif 
  
  if (lam_procs > MAX_REMOTE) {
    printf("too many nodes in config file, max is %d\n",MAX_REMOTE);
    exit(-1);
  }
  
  
  for (i=0; i< lam_procs; i++) {

#ifdef USE_GLUNIX
 if (use_glunix) {
   remote_ip_num = (int *) Am_GetFastAddrByVnn( (VNN) i);
   strcpy(remote_host,(char *)Am_GetFastAddrStr(remote_ip_num));

    if ( (first_dot = (char *) strchr(remote_host,'.')) != (char *) NULL) {
      first_dot[0] = '\0';
    }

    if ( fseek(config_file, 0, SEEK_SET) == -1) {
      fprintf(stderr,"Can't reset file pointer\n");
      fflush(stderr); fflush(stdout);
      exit(-1); 
    }

    /* get past top two line next line in the file */
    fscanf(config_file,"%s ",lcp_file);
    fscanf(config_file, "%d ", &j) ;

    j=0; 
    while ( (fscanf(config_file, "%s %s",temp_path,node_name) != EOF) && 
	   (j ==0) ){
      if (strcmp(remote_host,node_name) == 0) {
	j = 1;
	break;
      }
    }

    if (j ==0 ) {
      fprintf(stderr,"Node %s missing from in configuration file\n",
	      remote_host);
    }

    /* set the path */
    strcpy(paths[i],temp_path);
 }
#endif 

 if (!use_glunix) {

    if (fscanf(config_file, "%s %s",&paths[i],node_name) == EOF) {
      fprintf(stderr,"Not enough nodes in configuration file\n");
      fflush(stderr);
      exit(-1);
    }

    strcpy(temp_name,my_hostname);
    if (strcmp(temp_name,node_name) == 0) {
      lam_myproc = i; 
    }

  }

 } /* end for i = 1 to procs */


  fclose(config_file);
  
  if (lam_myproc == -1) {
    fprintf(stderr,"could not find my node \n"); fflush(stderr);
    exit(1);
  }      
  

  for (i = 0; i< lam_procs; i++) {
    routes[i].len = strlen(paths[i]);      
    for (j =0; j<  strlen(paths[i]); j++) {

      /* really dumb encoding */
      /* 01234567 is +0 ..  +7 */
      /* ABCDEFGH is -O ..  -7 */

      hop = (int) (paths[i][j] - '0') ;

      /* try a negative route */
      if ( (hop <0) || (hop > 7) ) {
	hop = paths[i][j] - 'A' ;
	hop = -hop;

	if ((hop > 0) || (hop < -7)) {
	  printf("invalid hop entry %c to node %d hop %d\n",(char) hop,i,j);
	  exit(1);
	}
      }
      temp_path[j] = ((hop&0x3f) |0x80);
    }
      
    dest_route = (char *) routes[i].path;

    switch (routes[i].len) {
    case 0:
      break;
    case 1:
      memcpy(&dest_route[1],temp_path,1);
      break;
    case 2:
      memcpy(dest_route+0,temp_path,2);
      break; 
    case 3:
      memcpy(dest_route+1,temp_path,3);	
      break;
    case 4:
      memcpy(dest_route+0,temp_path,4);	
      break;
    case 5:
      memcpy(dest_route+3,temp_path,5);	
      break;
    case 6:
      memcpy(dest_route+2,temp_path,6);	
      break;
    case 7:
      memcpy(dest_route+1,temp_path,7);	
      break;
    case 8:
      memcpy(dest_route+0,temp_path,8);	
      break;
    default:
      fprintf(stderr,"node %d: route to node %d too long \n",lam_myproc,i);
      exit(1);
      break; 
    } /* end switch */

  } /* end for lam_procs */
  
  /* set return path */
  LANAI[0][P_END_PT] = lam_myproc;
  
  in_num =out_num=0 ;
  in_d = &(endpt->in[in_num]);
  out_d = &(endpt->out[out_num]);

  /* check for double word alignment */
/*  if ( (((int) (in_d)&0x7) !=0) ) {
 *    fprintf(stderr,"wonk! input queue not double word aligned. Fix lcp\n");
 *    exit(1);
 *  }  
 */

  /* set up flow control info */
  for (i=0; i<lam_procs; i++) {
    flow_cntl[i] = (DEPTH); 
    req_flow_out[i] = rep_flow_in[i] = 0;
    req_flow_in[i] = rep_flow_out[i] = 0;
  }
  
  dma_base = lam_dma_base = (unsigned *)DBLOCK[0];
  user_base = (volatile int *) UBLOCK[0];
  
  /* set up receive bulk packets  */
  for (i = 0; i < IN_Q_SIZE ; i++) {
    endpt->bulk_d_p[i].hi = (uint *) dma_base + (LAM_PKT_WORDS*i); 
  }
  
  /* set up send */
  for (i=0; i< OUT_Q_SIZE ; i ++) {
    endpt->out[i].hdr.b_hdr.la = (int *) dma_base +
      (LAM_PKT_WORDS*2*DEPTH*MAX_REMOTE)+(i*LAM_PKT_WORDS); 

  }; 

  am_init_tables();  
  am2_barrier_init(0);

  H_DEBUG_GO = 0;
  
  H_INIT_DONE = 1;  

#ifdef USE_GLUNIX /*<coy>*/

 if (use_glunix) {
  Glib_Barrier();
 }
 
  return AM_OK;
#endif
}

int 
am_request_fd(int dest, void *fun, int arg2, int arg3, double d1) {

  int *i;

  /* code assume a double is 2x int. should work on Sparc v8 */
  i = (int *) &d1;
  am_request_4(dest,fun,arg2,arg3,(int) *i,(int) *(i+1)); 

}

int 
am_reply_fd(int dest, void *fun, int arg2, int arg3, double d1) {

  int *i;

  i = (int *) &d1;

  am_reply_4(dest,fun,arg2,arg3,*i,*(i+1)); 

}


#ifdef USE_GLUNIX

int 
lam_glunix_spawn(int *argc_p, char **argv,char *envp[] ) {

  int i;
  int num_nodes;
  char *node_names,*next_node;
  struct hostent *hp;
  int nodes_to_spawn[MAX_REMOTE];  /* IP addresses */


  if (!Glib_Initialize()) {
    fprintf(stderr,"Error - GLUnix couldn't initialize!\n");
    exit(-1);
  }


  if (Glib_AmIStartup()) {

    if ( *(argc_p) > 1) {
      /* argv[1] is number of nodes; argv[0] is program name; argv is
       * list of arguments including program name and number of nodes.
       * We need to remove argv[1] when the argument array is passed
       * to the splitc_main().
       */
      num_nodes = atoi(argv[1]);
      if (num_nodes < 0) {
	fprintf (stderr,"lam: Error - must have positive nodes.\n");
	exit(1);
      }

      i = Glib_GetNumAvailableNodes ();

      if (i < num_nodes) {
        fprintf (stderr,
		 "lam: Error - Only %d nodes available (requested %d).\n",
                 i, num_nodes);
        exit (1);
      }

      for (i = 2; i <= *argc_p; ++i) {
        argv[i - 1] = argv[i];
      }
      *argc_p --;

      node_names = getenv("LAM_GLUNIX_NODES") ;

      if (node_names != (char *) NULL) {

	next_node = (char *) strtok(node_names," ");
	
	for (i=0; i< num_nodes; i++) {

	  if (next_node == NULL) {
	    fprintf(stderr,"not enough nodes in LAM_GLUNIX_NODES\n");
	    exit(1);
	  }

	 /* get the IP address */
	  hp = gethostbyname(next_node);
	  if ((hp == 0) || (hp->h_addr == 0)) {			
	    fprintf(stderr,"lam_glunix_spawn cant get IP address of host %s\n",
		    next_node);
	    return -1;
	  }		
  
	  memcpy(&nodes_to_spawn[i], ((unsigned int *)hp->h_addr),
		 sizeof(int));
	  
	  next_node = (char *) strtok( NULL," ");
	}

	/* now we have all the IP addresses. Time to spawn off jobs */
	if (Glib_Spawnefp(num_nodes, nodes_to_spawn, GLIB_SPAWN_ONE_EXIT,
			  argv[0], argv,envp) == -1 ) {
	  fprintf(stderr," Glib_Spawnefp failed \n");
	}
	exit(0);
      } /* end if spawn on specific nodes */
      else {
	if ( Glib_Spawn(num_nodes, argv[0], &argv[0]) == -1 ) {
	  fprintf(stderr," Glib_Spawn failed \n");
	}
      } /* else spawn on any old nodes */
      exit(0);
    } /* if more than one arg */
    else {
      if (   Glib_Spawn(1, argv[0], &argv[0]) == -1 ) {
	fprintf(stderr," Glib_Spawn failed \n");
      }
      exit(0);
    }
  }
  else { /* startup returns failure */

  }
}
#endif 


int 
am_disable()
{
  AM_INIT_LOCK(lam_shmem_addr[0]);
  shmdt(lam_shmem_addr);
}


int 
am_sleep() {
  int buf[4];
  read(LANAI_fd[0],buf,4);
}


int 
lam_change_route(int v, int len, int *p) {
  char temp_path[8];
  char *dest_route;
  int j, hop;

#ifdef DEBUG
  assert((v < MAX_REMOTE) && (len < 9));
#endif 
  
  routes[v].len = len;
  for (j=0; j<len; j++) {
    hop = p[j];
    if ((hop > 7) || (hop < -7)) {
      printf("invalid hop entry %d to node %d hop %d\n",hop,v,j);
      return 0;
    }
    temp_path[j] = ((hop&0x3f) |0x80);
  }
  
  dest_route = (char *) routes[v].path;
  
  switch (routes[v].len) {
  case 0:
    break;
  case 1:
    memcpy(dest_route+1,temp_path,1);
    break;
  case 2:
    memcpy(dest_route+0,temp_path,2);
    break; 
  case 3:
    memcpy(dest_route+1,temp_path,3);	
    break;
  case 4:
    memcpy(dest_route+0,temp_path,4);	
    break;
  case 5:
    memcpy(dest_route+3,temp_path,5);	
    break;
  case 6:
    memcpy(dest_route+2,temp_path,6);	
    break;
  case 7:
    memcpy(dest_route+1,temp_path,7);	
    break;
  case 8:
    memcpy(dest_route+0,temp_path,8);	
    break;
  default:
    printf("route too long \n");
    exit(1);
    break; 
  } /* end switch */
  return 1;
}



void 
lam_restore_flow(int des) {
  flow_cntl[des]++;
}

void 
lam_route_display() {
  int i,u;

  for(i=0;i<lam_procs;i++) {
    printf("Route %d ",i);
    for(u=0;u<4;u++) {
      printf(" %x", routes[i].path[u]);
    }
    printf("\n");
  }
}

/* these are as close to the AM-II spec as possible */
int 
AM_Init(int use_glunix) {
  int i; 
  i = am_enable(use_glunix);

  if (i != -1) 
    return AM_OK ;

  return 
    AM_ERR_NOT_INIT ;
} 

int 
AM_Terminate() {
  am_disable();
  return AM_OK;
}

/* request functions */
int 
AM_Request4(ea_t request_endpoint, int reply_endpoint, handler_t handler, 
		       int arg0, int arg1, int arg2, int arg3) {
  int i; 
  
  i = am_request_4(reply_endpoint, handler, arg0, arg1,  arg2, arg3);
  
  if (i != -1) 
    return AM_OK; 

  return AM_ERR_NOT_INIT;

}

int 
AM_RequestI2(ea_t request_endpoint, int reply_endpoint,handler_t handler,
		  void *source_addr, int nbytes, int arg0, int arg1) {

  int i; 
  i = am_request(reply_endpoint, handler, source_addr, nbytes,
		 arg0, arg1); 
  
  if (i != -1) 
    return AM_OK; 

  return AM_ERR_NOT_INIT;
}

int 
AM_RequestXfer1(ea_t request_endpoint, int reply_endpoint, int dest_offset, 
		handler_t handler,void *source_addr, int nbytes, int arg0) {

  int i; 
  
  i = am_request_xfer(reply_endpoint, source_addr,0, dest_offset, nbytes,
		      handler, arg0);

  if (i != -1) 
    return AM_OK; 

  return AM_ERR_NOT_INIT;

}


/* 
 * Reply functions 
 */
int 
AM_Reply4(void *token, handler_t handler,
	     int arg0, int arg1, int arg2, int arg3) {
  
  int i,dest; 

  dest = (int) token; 
  
  i = am_reply_4(dest,handler,arg0, arg1, arg2, arg3);

  if (i != -1) 
    return AM_OK; 

  return AM_ERR_NOT_INIT;

}

int 
AM_ReplyI2(void *token, handler_t handler, int nbytes, void *source_addr,
	   int arg0, int arg1) {

  int i,dest;
  dest = (int ) token; 

  i = am_reply(dest, handler, source_addr, nbytes, arg0, arg1);

  if (i != -1) 
    return AM_OK; 

  return AM_ERR_NOT_INIT;

}

int 
AM_ReplyXfer1(void *token, int dest_offset,
	      handler_t handler, void *source_addr, int nbytes, 
	      int arg0) {

  int i,dest; 
  dest = (int ) token; 

  i = am_reply_xfer(dest,source_addr,0,dest_offset,nbytes,handler,arg0);

  if (i != -1) 
    return AM_OK; 

  return AM_ERR_NOT_INIT;

}


int 
AM_Poll(eb_t bundle) {
  int i;

  i = am_poll();
  return i;
}

int 
AM_MaxShort() {
  return 4; 
}


int 
AM_MaxMedium() {
  return LAM_PKT_BYTES;
}

int 
AM_MaxLong() {
  return LAM_PKT_BYTES;
}

int 
AM_MaxNumHandlers() {

  int i; 
  i = am_max_handlers();
  return i;
}

int 
AM_SetHandler(ea_t ea, handler_t *handler, void (*function)()) {
  int i; 
  volatile handler_t h; 

  h = *handler; 

  i = am_add_handler( h, function);

  if (i != -1) {
    *handler =i; 
    return AM_OK;
  }

  return AM_ERR_NOT_INIT;
}

int 
AM_SetSeg(ea_t ea, void *addr, int nbytes) {
  int i; 

  i = am_set_segment(0, addr, nbytes);
  
  if (i != -1) 
    return AM_OK;

  return AM_ERR_NOT_INIT;
}


int 
AM_MaxSegLength(int *nbytes) {
  return (1<<28);
}

int AM_Wait(eb_t eb) {
  am_sleep();
}
