#ifndef _AM_H
#define _AM_H   

#define AM_ALL  0xffffffff
#define AM_NONE 0x00000000

#define AM_PAR   1
#define AM_SEQ   0

#define AM_OK           0      /* Function completed successfully */
#define AM_ERR_NOT_INIT 1      /* Active message layer not initialized */
#define AM_ERR_NOT_TERM 2      /* Active message layer not initialized */
#define AM_ERR_BAD_ARG  3      /* Invalid function parameter passed */
#define AM_ERR_RESOURCE 4      /* Problem with requested resource */
#define AM_ERR_NOT_SENT 5      /* Asynchronous message not sent */
#define AM_ERR_IN_USE   6      /* Resource currently in use */

/* Error codes for error handler */
typedef int op_t;
typedef int ea_t;
typedef int handler_t;
typedef int eb_t; 

#define EBADARGS     0x1   
#define EBADENTRY    0x2   
#define EBADTAG      0x4
#define EBADHANDLER  0x8
#define EBADSEGOFF   0x10
#define EBADLENGTH   0x20
#define EBADENDPOINT 0x40   /* Not Applicable to UDPAM */
#define ECONGESTION  0x80
#define EUNREACHABLE 0x100  

#define AM_REQUEST        1
#define AM_REQUEST_MEDIUM 2 
#define AM_REQUEST_LONG   3
#define AM_REPLY          4
#define AM_REPLY_MEDIUM   5
#define AM_REPLY_LONG     6

/* Events */ 
#define AM_NOEVENTS     0x0 
#define AM_EMPTYTONOT   0x1
#define AM_MSGRECEIVED  0x2 
#define AM_NOTEMPTY     0x1

/* Types of Handlers */
typedef void (*ErrorHandler)(int status, op_t opcode, void *argblock); 

typedef void (*Handler4)(void *token, int arg0, int arg1, int arg2,int arg3);  

typedef void (*HandlerMedium2)(void *token, void *buf, int nbytes, 
			       int arg0, int arg1); 
typedef void (*HandlerXfer1)(void *token, int dest_offset, int nbytes, 
			     int arg0, int arg1, int arg2, int arg3);  

/*
 * Active Messages API 
 */
#define AM_USE_GLUNIX 1 
#define AM_NO_GLUNIX  0 
extern int AM_Init( int flags);

extern int AM_Terminate();
 
/* Request Functions */ 
extern int AM_Request4(ea_t request_endpoint, int reply_endpoint, 
		       handler_t handler, 
		       int arg0, int arg1, int arg2, int arg3); 

extern int AM_RequestI2(ea_t request_endpoint, int reply_endpoint,
			     handler_t handler, void *source_addr, int nbytes,
			     int arg0, int arg1);

extern int AM_RequestXfer1(ea_t request_endpoint, int reply_endpoint,
			   int dest_offset, handler_t handler, 
			   void *source_addr, int nbytes,
			   int arg0);

/*
 * Reply functions 
 */
extern int AM_Reply4(void *token, handler_t handler,
		     int arg0, int arg1, int arg2, int arg3);

extern int AM_ReplyI2(void *token, handler_t handler,
			   int nbytes, void *source_addr,
			   int arg0, int arg1);

extern int AM_ReplyXfer1(void *token, int dest_offset,
			 handler_t handler, void *source_addr, int nbytes, 
			 int arg0);

extern int AM_Poll(eb_t bundle);
extern int AM_MaxShort();
extern int AM_MaxMedium();
extern int AM_MaxLong();

/*
 * Endpoint and Bundle API  
 */

extern int AM_SetHandler(ea_t ea, handler_t *handler, void (*function)());
extern int AM_MaxNumHandlers();

extern int AM_SetSeg(ea_t ea, void *addr, int nbytes);
extern int AM_MaxSegLength(int *nbytes);

extern int AM_Wait(eb_t eb);

#endif /* _AM_H */
