/*									8
 *
 * lam.c - Lanai Active Messages 
 *
 * 
 * "Copyright (c) 1995 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 		Richard P. Martin and Lok T. Liu
 * Version:		10
 * Creation Date:	Tue Feb 28 18:15:42 1995
 * Filename:		lam.c
 * History:
 *	RPM	8	Sun Sep 24 19:54:02 1995
 *		Added glunix defines 
 *	RPM	7	Tue Aug 22 10:03:49 1995
 *		added long routes, sleep.
 *	RPM	2	Sat Jun  3 19:56:10 1995
 *		added relative addressing code. Fixed Q size problem.
 */

#include  <stdio.h>
#include  <stdlib.h>
#include  <unistd.h>
#include <string.h>
#include  "lanai_dev.h"
#include <signal.h>
#include "config.h"
#include <sys/types.h>

#include <sys/ipc.h>
#include <sys/shm.h>

#include <sys/time.h>
#include <sys/stat.h>

#include "ep.h"
#include "ifc.h"

#ifdef USE_GLUNIX
/* GLUnix interface */
#include <cam.h>
#include <glib.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#endif 


#ifdef DEBUG
#include <assert.h>
#endif

#define LAM_CONFIG_DEFAULT "/usr/now/etc/myrinet_routes/lam_conf" 
#define LAM_MAP_DEFAULT    "/usr/now/etc/myrinet_routes/master_routes" 

#define LAM_SHMEM_KEY     0x700328
#define LAM_SHMEM_SIZE    0x1000
#define PERMS             0666

#define LEN_SHIFT 8 
 
#ifdef SOLARIS

#include <sys/systeminfo.h> 
#define gethostname(n,l)  sysinfo(SI_HOSTNAME,n,l)
#else
#define memmove(d,s,l) bcopy(s,d,l)
#endif

#ifdef __GNUC__
#define INLINE inline
#else 
#define INLINE 
#endif

unsigned short sts;
int copy_block_len,units;
int unit;

volatile ep_t *endpt;
volatile route_t *routes;
volatile desc_t *in_d,*out_d;
int in_num,out_num;

volatile int *dma_base;
volatile int *user_base;

int *lam_dma_base;

int lam_myproc;
int lam_procs =2;
int lam_reply =0;  /* which node to reply to */

int flow_cntl[MAX_REMOTE];
int req_flow_in[MAX_REMOTE];
int req_flow_out [MAX_REMOTE];

int am_poll_count =0;
int good_polls =0;
int null_polls =0;
int no_credits =0; 

struct timespec poll_begin,poll_end,poll_lapsed;

int rep_flow_in[MAX_REMOTE];
int rep_flow_out[MAX_REMOTE];

uint seq_glob;

int lam_sync_flag[MAX_REMOTE];
int lam_sync_count, lam_old_count;

unsigned int *lam_shmem_addr; 

typedef void (*handler_t)(int src,int arg0, int arg1,int arg2, int arg3);
typedef void (*xfer_handler_t)(int src,int arg0, int arg1,int arg2);
extern unsigned int LANAI_fd[64];

#ifdef USE_GLUNIX
extern char *lam_my_hostname;
#endif 

int lam_frag_size; 

void 
am_request_4(int dest,handler_t fun, int arg0, int arg1, 
		  int arg2, int arg3);

void 
am_reply_4(int dest,handler_t fun, int arg0, int arg1, 
		  int arg2, int arg3);

void 
am_poll();

handler_t 
lam_null_handler(int src,int arg0, int arg1,int arg2, int arg3) {

}
extern int am_barrier_init();


handler_t 
lam_sync_rep_h(int src,int arg0, int arg1,int arg2, int arg3) {
  lam_sync_count = arg0;
}

handler_t 
lam_sync_req_h(int src,int arg0, int arg1,int arg2, int arg3) {
  if (lam_sync_flag[src] == 0) {
    lam_sync_flag[src] =1;
    lam_sync_count++;
  }
#if 0
  printf("node 0 got sync request from node %d\n",src);
  fflush(stdout);
#endif
  am_reply_4(src,(handler_t) lam_sync_rep_h,lam_sync_count,0,0,0);
}


void 
am_wait(volatile int *flag,int value)
{
  do {
    am_poll();
  } while (*flag < value);
 
  *flag -= value;
}

void 
am_poll_wait(volatile int *flag,int value)
{

 while (*flag < value) {
   am_poll();
 }
 
  *flag -= value;
}


/* nice little routing provided for by Yannis Schoinas <schoinas@cs.wisc.edu>
 */

void 
fmemcpy(void *to, void *from, int size) {

    if (((((int) (to)) & 0x7) == 0) && ((((int) (from)) & 0x7) == 0) &&
       (size % sizeof(double) == 0))
    {
      register int i;

      register double *tmp1 = (double *)from, *tmp2 = (double *)to;
      for (i = 0; i < size/sizeof(double); ++i)
      {
        tmp2[i] = tmp1[i];
      }
    }
    else
      memcpy(to, from, size);
}


/* take a string in the form AAA123 and return 123 */
static int 
name_to_number(char *name) {
  int len;
  int i;
  int value;
  char *first_char; 
  char *last_char;   

  value = -1;
  len= strlen(name);

  /* find first char */
  for (i=0, first_char  =name; (i< len) ; i++, first_char++) {
    if (isdigit(*first_char));
      break;
  }

  /* no digit found */
  if (i == len) 
    return -1; 

  /* find last digit  */
  for (last_char  = first_char; (i<len) ; i++, last_char++){
    if (!isdigit(*first_char)) {
      break;
    }
  }

  value = atoi(first_char);
  return value;
}



/* for the route from i to j, fill in the route as a string */
static int 
number_to_route(FILE *map,int source,int dest, char *path) {
  int checked_src; 
  int line_len; 
  char line[512];

  if ( fseek(map, 0, SEEK_SET) == -1) {
    fprintf(stderr,"number_to_route can't reset file pointer\n");
    fflush(stderr); fflush(stdout);
    exit(-1); 
  }

  checked_src = -1;

  /* find the line with the source node */
  while (checked_src != source) {
    if (fscanf(map,"%s",line) == 0 ) {
      fprintf(stderr,"could not find source %d in map file\n",source);
      exit(-1);
    }

    /* find line matching the 'source' number */

    if (sscanf(line,"Source %d",&checked_src) == 0) {
      fprintf(stderr,"invalid map file format\n",source);
      exit(-1);      
    }
  }

  /* find field matching the 'dest number */
/*   strtok() */

  

}



static INLINE void 
am_reply_0(int dest,handler_t fun) {

#ifdef DEBUG
  rep_flow_out[dest]++;
  assert( (dest >=0) && (dest < lam_procs));
#endif 

  lam_reply = -1;

  out_d->hdr.s_data[0] = (int) fun;
  out_d->op_dest = (int) (AM_TYPE)|(dest<<DEST_SHIFT)|(REPLY)|(SHORT);

  out_num = (out_num +1) & OUT_SIZE_MASK;
  out_d = &(endpt->out[out_num]);

}


void 
am_poll(){
  register int op_dest,src;
  register unsigned fun;  
  register long long temp0, temp1; 

#ifdef DEBUG
  am_poll_count++;
#endif 

  while (((op_dest = in_d->op_dest)&OP_MASK) != 0 ){ /*if buffer has anything*/

    src = (unsigned) (op_dest & DEST_MASK) >> DEST_SHIFT;

    if ( (op_dest&REQ_MASK) == REQUEST) {
      lam_reply =src;
#ifdef DEBUG
      req_flow_in[src]++;
#endif 
    } else {
      flow_cntl[src]++;
#ifdef DEBUG
      rep_flow_in[src]++;
#endif 

    }

    if ( (op_dest &OP_MASK) == SHORT) {

      temp0 = *( ((long long *) in_d->hdr.s_data) +1 ); 
      temp1 = *( ((long long *) in_d->hdr.s_data) +2 ); 


      ( *( (void (*)()) in_d->hdr.s_data[0]))(src,in_d->hdr.s_data[1],
					      temp0,
					      temp1);

      in_d->op_dest = in_d->op_dest & (~OP_MASK) ;
      in_num = (in_num+1)&IN_SIZE_MASK;
      in_d = &(endpt->in[in_num]);

    }
    else {
      if ((op_dest &OP_MASK) == BULK ) {

	{
	  register uint word1,word2,word3,word4,word5,word6;

	  word5 = in_d->hdr.b_hdr.data[3]; /* number of unaligned bytes */
	  word1 = in_d->hdr.b_hdr.len;      /* length the Lanai gives us */
	  word2 = in_d->hdr.b_hdr.data[0]; /* fun pointer */
	  word3 = in_d->hdr.b_hdr.data[1]; /* dest address */
	  word4 = in_d->hdr.b_hdr.data[2]; /* arg 0*/

	  word6 = word5 >> LEN_SHIFT; 
	  word5 = word5 & 0xf;

	  if (word5 == 0) {

	    memcpy((char *) word3,
		   (char*) ((uint *) user_base + (LAM_PKT_WORDS*in_num)),
		   word1) ;
	  }
	  else {
	    if (word1 > 4) {
	      word1 = ( (word1-4)&(~0x3))+word5;
	    }
	    else {
	      word1 = word5;	    
	    }
	    memcpy((char *) word3, 
		   (char *) ((uint *) user_base + (LAM_PKT_WORDS*in_num)),
		   word1); 
	  }

	  if ( (xfer_handler_t) word2 != (xfer_handler_t) lam_null_handler ) {
	    ( (xfer_handler_t) word2)(src,word3,word6,word4);
	  }

	}
	in_d->op_dest = in_d->op_dest & (~OP_MASK) ;
	in_num = (in_num+1)&IN_SIZE_MASK;
	in_d = &(endpt->in[in_num]);
      }
      else {
	/* CRC ERROR */
	printf("Wank! got crc error.\n");
	printf("in_num = 0x%0x in_d = 0x%0x\n",in_num,in_d);
	printf("op_dest= 0x%0x s_data[0]= 0x%0x s_data[1]= 0x%0x\n",
	       in_d->op_dest, in_d->hdr.s_data[0], in_d->hdr.s_data[1]);
	printf("s_data[2]= 0x%0x s_data[3]= 0x%0x s_data[4]= 0x%0x\n",
	       in_d->hdr.s_data[2], in_d->hdr.s_data[3], in_d->hdr.s_data[4]);
	printf("s_data[5]= 0x%0x crc= 0x%0x\n",
	       in_d->hdr.s_data[5], in_d->crc);

	in_d->op_dest = in_d->op_dest & (~OP_MASK) ;
	in_num = (in_num+1)&IN_SIZE_MASK;
	in_d = &(endpt->in[in_num]);
	
      }
    }

   if ( (op_dest & REQ_MASK) == REQUEST ) { /* was a request or reply? */
      if (lam_reply == src) 
	am_reply_0(src, (handler_t) lam_null_handler);
    }

  }

}



INLINE void 
am_request_4(int dest,handler_t fun, int arg0, int arg1, 
		  int arg2, int arg3) {

  int i;

#ifdef DEBUG
  int print_times =0;
#endif 

  if (dest != lam_myproc) {

    if (flow_cntl[dest] != 0) {

    send_top:
#ifdef DEBUG
      assert( (dest >=0) && (dest < lam_procs));

      while ( ((out_d->op_dest)&OP_MASK) != 0) {
	am_poll();
	printf("\nWarning, am_request_4 found output queue full! \n"); 
      }
#endif 

      flow_cntl[dest]--; 
#ifdef DEBUG
      req_flow_out[dest]++;
#endif 

    
      out_d->hdr.s_data[0] = (int) fun;
      out_d->hdr.s_data[1] = (int) arg0;
      out_d->hdr.s_data[2] = (int) arg1;  
      out_d->hdr.s_data[3] = (int) arg2;
      out_d->hdr.s_data[4] = (int) arg3;

      (volatile) out_d->op_dest =
	(int) (AM_TYPE)|(dest<<DEST_SHIFT)|(REQUEST)|(SHORT);
      
      out_num = (out_num +1) & OUT_SIZE_MASK;
      out_d = &(endpt->out[out_num]);
      
      am_poll();
    } /* end if have credits */
    else {
      while (flow_cntl[dest] == 0) 
	am_poll();
      goto send_top;
    }
  } /* dest != me */
  else {
    (*fun)(lam_myproc,arg0, arg1, arg2, arg3); 
    am_poll();
  }
}


INLINE void 
am_request_5(int dest, void (*fun)(), int arg0, int arg1, 
		  int arg2, int arg3, int arg4) {
  int i;

#ifdef DEBUG
  int print_times =0;
#endif 

  if (dest != lam_myproc ) {
    while (flow_cntl[dest] == 0) {
      am_poll();

#ifdef DEBUG
      print_times++;
/*      printf("req_5 node %d out of credits to %d times %d count %d \r",
	     lam_myproc,dest,print_times,am_poll_count); */
#endif 

    }

    flow_cntl[dest]--;
#ifdef DEBUG
    assert( (dest >=0) && (dest < lam_procs));
    req_flow_out[dest]++;
#endif 
    
#ifdef DEBUG
    while ( ((out_d->op_dest)&OP_MASK) != 0) {
      am_poll();
      printf("\nWarning, am_request_5 found output queue full! \n");
    }
#endif 

    out_d->hdr.s_data[0] = (int) fun;
    out_d->hdr.s_data[1] = (int) arg0;
    out_d->hdr.s_data[2] = (int) arg1;  
    out_d->hdr.s_data[3] = (int) arg2;
    out_d->hdr.s_data[4] = (int) arg3;
    out_d->hdr.s_data[5] = (int) arg4;

    out_d->op_dest = i = (int) (AM_TYPE)|(dest<<DEST_SHIFT)|(REQUEST)|(SHORT);

    out_num = (out_num +1) & OUT_SIZE_MASK;
    out_d = &(endpt->out[out_num]);

    am_poll();
  }
  else {
     (*fun)(lam_myproc,arg0, arg1, arg2, arg3,arg4);
     am_poll();
  }

}


void 
lam_store(int dest, void *lva, void *rva, int nbytes,
		  xfer_handler_t fun, void *arg0,int real_len) {

  int i;

  if (nbytes <= (LAM_PKT_WORDS*sizeof(int))) {

    if ( dest != lam_myproc ) {

      while (flow_cntl[dest] == 0) {
	am_poll();
      }
#ifdef DEBUG      
      assert(nbytes > 0 );   /* transfer should be >0 < max size */
      assert(nbytes <= LAM_PKT_WORDS*4 );

      assert( (dest >=0) && (dest < lam_procs));  /* dest oK? */
      while ( ((out_d->op_dest)&OP_MASK) != 0) {
	am_poll();
	printf("\nWarning, am_store found output queue full! \n");
      }
#endif 
      
      flow_cntl[dest]--;  

#ifdef DEBUG
      req_flow_out[dest]++;
#endif 
   /* copy into io space buffer */
   /* need to check max len, assume that all buffers are word aligned */
   /* and in multiple of words */

      memcpy((int *) (user_base+((LAM_PKT_WORDS*IN_Q_SIZE)+
				 (out_num*LAM_PKT_WORDS))),
	     (char *) lva, (nbytes));

      out_d->hdr.b_hdr.la = (int *) lam_dma_base +
	(LAM_PKT_WORDS*IN_Q_SIZE)+(out_num*LAM_PKT_WORDS);
      
      out_d->hdr.b_hdr.len = nbytes;
      out_d->hdr.b_hdr.data[0] = (int) fun;
      out_d->hdr.b_hdr.data[1] = (int) rva;
      out_d->hdr.b_hdr.data[2] = (int) arg0;
      /* store 'extra' bytes */
      out_d->hdr.b_hdr.data[3] = (real_len<<LEN_SHIFT) | (nbytes & 0x3); 

      if ( (nbytes & 0x3) != 0) {  /* round up to upper byte */
	out_d->hdr.b_hdr.len = (nbytes+4) & (~0x3);
      }
      
      out_d->op_dest = i = 
	(int) (AM_TYPE)|(dest<<DEST_SHIFT)|(REQUEST)|(BULK); /*go*/

      out_num = (out_num +1) & OUT_SIZE_MASK;
      out_d = &(endpt->out[out_num]);
      am_poll();
      return;
    }
    else {
      memmove(rva,lva,nbytes);
      (*fun)(dest,(void *) rva,nbytes,(void *)arg0);    
      am_poll();
      return;
    }
  }
  else {
    printf("lam_store called with size > %d\n",LAM_PKT_WORDS*sizeof(int));
  }
}

/* this function does fragmentation of outbound data */
/* assume the network is as in-order delivery semantics */

INLINE void
am_store(int dest,void *l_buf,void *r_buf,int nbytes,
	 void (*fun), void *arg0) {
  int i;

  if( nbytes <= lam_frag_size) {
   lam_store(dest,l_buf,r_buf,nbytes,fun,arg0,nbytes);  
  }
  else {
    /* chunk the data into maximum packet sizes */
    for (i=lam_frag_size; i<nbytes; i+=lam_frag_size) {
     lam_store(dest,l_buf,r_buf,lam_frag_size, (xfer_handler_t)
	       lam_null_handler, (void*)i, nbytes);
      (char *) l_buf+= lam_frag_size; (char *)r_buf+= lam_frag_size;  
    }
    i-=lam_frag_size;
    /*put the last packet */
    lam_store(dest,l_buf,r_buf,nbytes-i,fun,arg0,nbytes); 
  }

}



void 
am_reply_4(int dest,handler_t fun, int arg0, int arg1, 
		  int arg2, int arg3) {


  if ( dest != lam_myproc ) {


#ifdef DEBUG      
      assert( ((out_d->op_dest)&OP_MASK) == 0 );
      assert( (dest >=0) && (dest < lam_procs));  /* dest oK? */
      rep_flow_out[dest]++;
#endif 


    lam_reply = -1;

    out_d->hdr.s_data[0] = (int) fun;
    out_d->hdr.s_data[1] = (int) arg0;
    out_d->hdr.s_data[2] = (int) arg1;  
    out_d->hdr.s_data[3] = (int) arg2;
    out_d->hdr.s_data[4] = (int) arg3;

    out_d->op_dest = (int)  (AM_TYPE)|(dest<<DEST_SHIFT)|(REPLY)|(SHORT);

    out_num = (out_num +1) & OUT_SIZE_MASK;
    out_d = &(endpt->out[out_num]);
  }
  else {
    (*fun)(lam_myproc,arg0, arg1, arg2, arg3);
  }

}



void 
am_reply_xfer(int dest, void *lva, void *rva, int nbytes,
		 void (*fun)(), void *arg0) {

  int real_len;

  real_len = nbytes; 
  if ( dest != lam_myproc ) {

#ifdef DEBUG
    assert( ((out_d->op_dest)&OP_MASK) == 0 );
    assert( (dest >=0) && (dest < lam_procs));  /* dest oK? */
    rep_flow_out[dest]++;
#endif 


    lam_reply = -1;

  /* copy into io space buffer */
  /* need to check max len, assume that all buffers are word aligned */
  /* and in multiple of words */
    memcpy( (char *)  (user_base+ ((LAM_PKT_WORDS*IN_Q_SIZE)+
				   (out_num*LAM_PKT_WORDS))),
	   (char *) lva, nbytes);
    out_d->hdr.b_hdr.la = (int *) lam_dma_base +
      (LAM_PKT_WORDS*IN_Q_SIZE)+(out_num*LAM_PKT_WORDS);

    out_d->hdr.b_hdr.len = nbytes;
    out_d->hdr.b_hdr.data[0] = (int) fun;
    out_d->hdr.b_hdr.data[1] = (int) rva;
    out_d->hdr.b_hdr.data[2] = (int) arg0;
    out_d->hdr.b_hdr.data[3] = (real_len<<LEN_SHIFT) | (nbytes & 0x3); 

    if ( (nbytes & 0x3) != 0) {
      out_d->hdr.b_hdr.len = (nbytes+4) & (~0x3);
    }

    out_d->op_dest = (int) (AM_TYPE)|(dest<<DEST_SHIFT)|(REPLY)|(BULK);
    out_num = (out_num +1) & OUT_SIZE_MASK;
    out_d = &(endpt->out[out_num]);
  }  
  else {
    memmove(rva,lva,nbytes);
    (*fun)(dest,rva,nbytes,arg0);    
  }
}


void 
am_get(int dest, void *rva, void *lva, unsigned int len, 
	   void (*handler)(), void *arg) {

#ifdef DEBUG
  assert(len >0 ) ;
  assert(len <= LAM_PKT_WORDS*4 );
  assert( ((out_d->op_dest)&OP_MASK) == 0 );
#endif 

  am_request_5(dest, am_reply_xfer, (int) rva, (int) lva, len,
	       (int) handler, (int) arg);

}

int 
am_procs() {
  return lam_procs;
}

int 
am_my_proc() {
  return lam_myproc;
}

int 
am_max_size() {
  return LAM_PKT_WORDS*4; 
}

void 
am_store_async(int dest_node, void *src_buf, void *dest_buf,
	       int len, handler_t request_handler, void *request_handler_arg1,
	       void (*endfunc)(), void *endfunc_arg1) {

  am_store(dest_node, src_buf, dest_buf, len, request_handler, 
	   request_handler_arg1);

  (*endfunc)(lam_myproc,src_buf,len,endfunc_arg1);    

}



#ifdef USE_GLUNIX
/* get configuration from GLUNIX */
void LAM_get_glu_config()
{
  pid_t pid;
  Npid glunix_pid;
		
  /* get total number of nodes */
  lam_procs = Glib_GetParallelDegree();

    if (lam_procs <= 0) {
      printf("GLUNIX reports that this proces is not part of a parallel program: AM_procs = %d\n", lam_procs);
      exit(-1);
    }
    /* get my node number */
    lam_myproc = Glib_GetMyVnn();

  pid = getpid();
  if (pid < 0) {
    perror("Cannot get local PID");
    exit(-1);
  }
  /* get the GLUNIX pid to agree on a server port number */
  glunix_pid = Glib_GetMyNpid();

  if (glunix_pid == NPID_NOBODY) {
    printf("GLUNIX reports that this proces is not part of a parallel program: glunix_pid = %d\n", glunix_pid);
    exit(-1);
  }

}
#endif 




int 
am_grab_lock_or_die(char *my_name) {
  register int tmp_reg; 
  volatile int *mem_addr; 
  pid_t my_pid,other_pid; 
  int shmem_id;
  int tries,i; 
  struct stat stat_buf;
  char pid_str[64];
  char pid_str_padded[64];
  char stat_path[64];

  if (( shmem_id = shmget(LAM_SHMEM_KEY, LAM_SHMEM_SIZE,
			       (PERMS|IPC_CREAT)))<0) {
    perror("can't open shared segment to get lock");
    exit(1);
  }

  my_pid = getpid();
  tries =0;
  /* check if someone is using the card */

  lam_shmem_addr = (void *) shmat(shmem_id,(char *) 0 , 0);

 try_lock:

  tmp_reg = 0; 
  AM_SWAP( (*lam_shmem_addr) ,tmp_reg);
  tries++;
  if (tmp_reg == 0) {
    /* see if this person is around */
    other_pid = (pid_t) lam_shmem_addr[1];

    if ( (other_pid == 0) || (other_pid == my_pid))  {
       lam_shmem_addr[1]= my_pid;
       AM_INIT_LOCK( lam_shmem_addr[0]);
       goto try_lock;
    }

    /* i = kill(other_pid,0); */
    strcpy(pid_str_padded,"00000");
    sprintf(pid_str,"%d",other_pid);
    
    strcpy(&(pid_str_padded[5-strlen(pid_str)]),pid_str);
    strcpy(stat_path,"/proc/");
    strcat(stat_path,pid_str_padded);

    i = stat(stat_path, &stat_buf);
    if (i == 0) {  /* eek, process exists */
      fprintf(stderr,"Process %d is holding the lock for the LANai card on %s\n",other_pid,my_name);
      exit(1);
    }
    else { /* doesn't exist? reset and try again */
      if (tries > 100) {
	fprintf(stderr,"Process %d failed to aquire lock after %d tries on %s\n",my_pid,my_name,tries);
	exit(1);
      }
      AM_INIT_LOCK( lam_shmem_addr[0]);
      goto try_lock;
    }
  }
  /*ok, got the lock  */
  lam_shmem_addr[1] = my_pid; 
}


int
am_enable(int use_glunix) {	/* arg == list of up to 4 files, */

  int i,j,port,hop;
  int port_num[MAX_REMOTE];
  int node_num;
  char *p, *s,*first_dot;
  char *home;
  char *config_path;
  char *map_path;
  static FILE *config_file;
  static FILE *map_file;
  static char config_filename[256];
  static char map_filename[256];
  char lcp_file[256];
  char node_name[256];
  char my_hostname[256];
  char remote_host[256];
  pid_t my_pid,other_pid; 
  int shmem_id;
  int *remote_ip_num;
  char temp_name[256];
  char paths[MAX_REMOTE][MAX_HOPS]; /* eight hops for now */
  char *dest_route,temp_path[8];
  
  
  gethostname(my_hostname,80);
  
  /* truncate hostname to first dot */
  if ( (first_dot = (char *) strchr(my_hostname,'.')) != (char *) NULL) {
    first_dot[0] = '\0';
  }
  
  /* set up default filename path */
  strcpy(config_filename, LAM_CONFIG_DEFAULT);
  strcat(config_filename,my_hostname);
  
  /* if the environment var is set up, use that */
  if ( ((config_path = getenv("LAM_CONFIG")) != (char *)NULL) ) {
    strcpy(config_filename, config_path);
  }

  /* if the environment var is set up, use that for the map file */
  if ( ((map_path = getenv("LAM_MAP")) != (char *)NULL) ) {
    strcpy(map_filename, map_path);
  }
  
  config_file = fopen(config_filename, "r");
  if (config_file == (FILE *)0) {			
    printf("Cannot open configuration file %s\n", config_filename);
    exit(-1);
  }

  map_file = fopen(map_filename, "r");
  if (map_file == (FILE *)0) {			
    printf("Cannot open network map file %s\n", config_filename);
    exit(-1);
  }
  
  
  (void) fscanf(config_file,"%s ",lcp_file);
  
  unit = 0;
  units = open_lanai_copy_block(&copy_block_len,&sts);	
  if (copy_block_len<=0) {
    perror("ops: open_copy_block()");
    printf("error! Length = %d  <= 0\n",copy_block_len);
  }
  else {

#ifdef DEBUG
  printf("ops: User opened copy_block \n"); 
  printf("ops: user_ptr = %x    dma_ptr = %x    copy_block_len = 0x%x\n",
	 (unsigned)UBLOCK[unit],(unsigned)DBLOCK[unit],copy_block_len);

#endif 

  }
  
  
  if ( copy_block_len < 
      ((IN_Q_SIZE+OUT_Q_SIZE)*LAM_PKT_WORDS*sizeof(int))) {
    
    fprintf(stderr, "warning! copy_block_len to short, len %d bytes wanted %d\n",
	    copy_block_len,
	    (IN_Q_SIZE+OUT_Q_SIZE)*LAM_PKT_WORDS*sizeof(int));
  }


  am_grab_lock_or_die(my_hostname);

  lanai_interrupt(0);
  lanai_reset(1);
  clear_load(0,lcp_file);
  memset( (char*) LANAI[0] ,0,128*1024) ;
  clear_load(0,lcp_file);
  
  lanai_reset(0);
  while (! L_INIT_DONE);     /* wait for lanai init done */
  
  /* set up routes */
  endpt = (ep_t *) ( (char *) LANAI[0] +  ((uint)BULK_PTR*2) );
  routes = (route_t *) ( (char *) LANAI[0] + ((uint) ROUTE_TAB*2) );

#ifdef DEBUG  
  printf("LANAI BASE %0x, ROUTE_TAB %0x, IN_BASE %0x, OUT_BASE %0x, BULK_PTR %0x\n", LANAI[0], ROUTE_TAB,IN_BASE, 
	 OUT_BASE , BULK_PTR);	   
  
  fflush(stdout);
#endif 
  
  if (fscanf(config_file, "%d ", &lam_procs) == EOF) {
    fprintf(stderr,"Cannot read number of nodes in configuration file\n");
    fflush(stderr);
    exit(-1);
  }
  
  lam_myproc = -1;
#ifdef USE_GLUNIX
  /* has side effect of resetting lam_procs and lam_myproc */
  if (use_glunix) {
    LAM_get_glu_config();
  }
#endif 
  
  if (lam_procs > MAX_REMOTE) {
    printf("too many nodes in config file, max is %d\n",MAX_REMOTE);
    exit(-1);
  }
  
  
  for (i=0; i< lam_procs; i++) {

#ifdef USE_GLUNIX
    if (use_glunix) {
      remote_ip_num = (int *) Am_GetFastAddrByVnn( (VNN) i);
      strcpy(remote_host,(char *)Am_GetFastAddrStr(remote_ip_num));
    }
#else
    if (!use_glunix) {
      if (fscanf(config_file, "%s %s",temp_path,remote_host) == EOF) {
	fprintf(stderr,"not enough nodes in configuration file\n");
	fflush(stderr);
	exit(-1);
      }
    }
#endif 

    if ( (first_dot = (char *) strchr(remote_host,'.')) != (char *) NULL) {
      first_dot[0] = '\0';
    }      
    node_num = name_to_number(remote_host); 
    if (node_num == -1) {
      fprintf(stderr,"node names must have a number, sorry \n");
      exit(-1);
    }

    j = number_to_route(map_file,lam_myproc,node_num,&paths[i]);
      
    if (j ==0 ) {
      fprintf(stderr,"Node %s missing from in configuration file\n",
	      remote_host);
    }

    
  } /* end for i = 1 to procs */


  fclose(config_file); 
  fclose(map_file);

 
  if (lam_myproc == -1) {
    fprintf(stderr,"could not find my node \n"); fflush(stderr);
    exit(1);
  }      
  

  for (i = 0; i< lam_procs; i++) {
    routes[i].len = strlen(paths[i]);      
    for (j =0; j<  strlen(paths[i]); j++) {

      /* really dumb encoding */
      /* 01234567 is +0 ..  +7 */
      /* ABCDEFGH is -O ..  -7 */

      hop = (int) (paths[i][j] - '0') ;

      /* try a negative route */
      if ( (hop <0) || (hop > 7) ) {
	hop = paths[i][j] - 'A' ;
	hop = -hop;

	if ((hop > 0) || (hop < -7)) {
	  printf("invalid hop entry %c to node %d hop %d\n",(char) hop,i,j);
	  exit(1);
	}
      }
      temp_path[j] = ((hop&0x3f) |0x80);
    }
      
    dest_route = (char *) routes[i].path;

    switch (routes[i].len) {
    case 0:
      break;
    case 1:
      memcpy(&dest_route[1],temp_path,1);
      break;
    case 2:
      memcpy(dest_route+0,temp_path,2);
      break; 
    case 3:
      memcpy(dest_route+1,temp_path,3);	
      break;
    case 4:
      memcpy(dest_route+0,temp_path,4);	
      break;
    case 5:
      memcpy(dest_route+3,temp_path,5);	
      break;
    case 6:
      memcpy(dest_route+2,temp_path,6);	
      break;
    case 7:
      memcpy(dest_route+1,temp_path,7);	
      break;
    case 8:
      memcpy(dest_route+0,temp_path,8);	
      break;
    default:
      fprintf(stderr,"node %d: route to node %d too long \n",lam_myproc,i);
      exit(1);
      break; 
    } /* end switch */

  } /* end for lam_procs */
  
  /* set return path */
  LANAI[0][P_END_PT] = lam_myproc;
  
  in_num =out_num=0 ;
  in_d = &(endpt->in[in_num]);
  out_d = &(endpt->out[out_num]);


  /* check for double word alignment */
/* if ( (((int) (in_d)&0x7) !=0) ) {
 *    fprintf(stderr,"wonk! input queue not double word aligned. Fix lcp\n");
 *    exit(1); 
 *  }  
 */

  /* set up flow control info */
  for (i=0; i<lam_procs; i++) {
    flow_cntl[i] = (DEPTH); 
    req_flow_out[i] = rep_flow_in[i] = 0;
    req_flow_in[i] = rep_flow_out[i] = 0;
    lam_sync_flag[i] =0;
  }
  
  dma_base = lam_dma_base = (unsigned *)DBLOCK[0];
  user_base = (volatile int *) UBLOCK[0];
  
  /* set up receive bulk packets  */
  for (i = 0; i < IN_Q_SIZE ; i++) {
    endpt->bulk_d_p[i].hi = (uint *) dma_base + (LAM_PKT_WORDS*i); 
  }
  
  /* set up send */
  for (i=0; i< OUT_Q_SIZE ; i ++) {
    endpt->out[i].hdr.b_hdr.la = (int *) dma_base +
      (LAM_PKT_WORDS*2*DEPTH*MAX_REMOTE)+(i*LAM_PKT_WORDS); 

  }; 

  
  lam_frag_size = (LAM_PKT_WORDS*sizeof(int));
  if (lam_frag_size > (LAM_PKT_WORDS*sizeof(int))) {
    printf("frag size > packet size! \n");
  }

  am_barrier_init();
  H_DEBUG_GO = 0;
  
  H_INIT_DONE = 1;  


#ifdef USE_GLUNIX /*<coy>*/
  if (use_glunix) {
    Glib_Barrier();
  }
#endif
}

int 
am_request_fd(int dest, void *fun, int arg2, int arg3, double d1) {

  int *i;

  /* code assume a double is 2x int. should work on Sparc v8 */
  i = (int *) &d1;
  am_request_4(dest,fun,arg2,arg3,*i,*(i+1)); 

}

int 
am_reply_fd(int dest, void *fun, int arg2, int arg3, double d1) {

  int *i;

  i = (int *) &d1;

  am_reply_4(dest,fun,arg2,arg3,*i,*(i+1)); 

}


#ifdef USE_GLUNIX

int 
lam_glunix_spawn(int *argc_p, char **argv,char *envp[] ) {

  int i;
  int num_nodes;
  char *node_names,*next_node;
  struct hostent *hp;
  int nodes_to_spawn[MAX_REMOTE];  /* IP addresses */


  if (!Glib_Initialize()) {
    fprintf(stderr,"Error - GLUnix couldn't initialize!\n");
    exit(-1);
  }


  if (Glib_AmIStartup()) {

    if ( *(argc_p) > 1) {
      /* argv[1] is number of nodes; argv[0] is program name; argv is
       * list of arguments including program name and number of nodes.
       * We need to remove argv[1] when the argument array is passed
       * to the splitc_main().
       */
      num_nodes = atoi(argv[1]);
      if (num_nodes < 0) {
	fprintf (stderr,"lam: Error - must have positive nodes.\n");
	exit(1);
      }

      i = Glib_GetNumAvailableNodes ();

      if (i < num_nodes) {
        fprintf (stderr,
		 "lam: Error - Only %d nodes available (requested %d).\n",
                 i, num_nodes);
        exit (1);
      }

      for (i = 2; i <= *argc_p; ++i) {
        argv[i - 1] = argv[i];
      }
      *argc_p --;

      /* NB GLUNIX_NODES works now, this is redundant */
      node_names = getenv("LAM_GLUNIX_NODES") ;

      if (node_names != (char *) NULL) {

	next_node = (char *) strtok(node_names," ");
	
	for (i=0; i< num_nodes; i++) {

	  if (next_node == NULL) {
	    fprintf(stderr,"not enough nodes in LAM_GLUNIX_NODES\n");
	    exit(1);
	  }

	 /* get the IP address */
	  hp = gethostbyname(next_node);
	  if ((hp == 0) || (hp->h_addr == 0)) {			
	    fprintf(stderr,"lam_glunix_spawn cant get IP address of host %s\n",
		    next_node);
	    return -1;
	  }		
  
	  memcpy(&nodes_to_spawn[i], ((unsigned int *)hp->h_addr),
		 sizeof(int));
	  
	  next_node = (char *) strtok( NULL," ");
	}

	/* now we have all the IP addresses. Time to spawn off jobs */
	if (Glib_Spawnefp(num_nodes, nodes_to_spawn, GLIB_SPAWN_ONE_EXIT,
			  argv[0], argv,envp) == -1 ) {
	  fprintf(stderr," Glib_Spawnefp failed \n");
	}
	exit(0);
      } /* end if spawn on specific nodes */
      else {
	if ( Glib_Spawn(num_nodes, argv[0], &argv[0]) == -1 ) {
	  fprintf(stderr," Glib_Spawn failed \n");
	}
      } /* else spawn on any old nodes */
      exit(0);
    } /* if more than one arg */
    else {
      if (   Glib_Spawn(1, argv[0], &argv[0]) == -1 ) {
	fprintf(stderr," Glib_Spawn failed \n");
      }
      exit(0);
    }
  }
  else { /* startup returns failure */

  }
}
#endif 

int 
am_disable() {

  lam_shmem_addr =0;
  shmdt((void *) lam_shmem_addr);
}


int 
am_sleep() {
  int buf[4];
  read(LANAI_fd[0],buf,4);
}

int
lam_restore_flow(int des) {
  if (flow_cntl[des] < MAX_DEPTH) 
    flow_cntl[des]++;
}

int 
lam_change_route(int v, int len, int *p) {
  char temp_path[8];
  char *dest_route;
  int j, hop;

#ifdef DEBUG
  assert((v < MAX_REMOTE) && (len < 9));
#endif
  
  routes[v].len = len;
  for (j=0; j<len; j++) {
    hop = p[j];
    if ((hop > 7) || (hop < -7)) {
      printf("invalid hop entry %d to node %d hop %d\n",hop,v,j);
      return 0;
    }
    temp_path[j] = ((hop&0x3f) |0x80);
  }
  
  dest_route = (char *) routes[v].path;
  
  switch (routes[v].len) {
  case 0:
    break;
  case 1:
    memcpy(dest_route+1,temp_path,1);
    break;
  case 2:
    memcpy(dest_route+0,temp_path,2);
    break; 
  case 3:
    memcpy(dest_route+1,temp_path,3);	
    break;
  case 4:
    memcpy(dest_route+0,temp_path,4);	
    break;
  case 5:
    memcpy(dest_route+3,temp_path,5);	
    break;
  case 6:
    memcpy(dest_route+2,temp_path,6);	
    break;
  case 7:
    memcpy(dest_route+1,temp_path,7);	
    break;
  case 8:
    memcpy(dest_route+0,temp_path,8);	
    break;
  default:
    printf("route too long \n");
    exit(1);
    break; 
  } /* end switch */
  return 1;
}


void 
lam_route_display() {
  int i,u;

  for(i=0;i<lam_procs;i++) {
    printf("Route %d ",i);
    for(u=0;u<4;u++) {
      printf(" %x", routes[i].path[u]);
    }
    printf("\n");
  }
}


/* really dumb way to sync all nodes */
/* all send to zero, when they get back that all have sent, they */
/* continue */

/* try Glib_Signal(npid, vnn, sig)? instead */
int 
lam_sync_all() {
  int i;
  int done; 
  int tries; 
  int total_time;
  int num_nodes; 

  struct timeval tv_begin, tv_end,lapsed; 
  
  done = 0;
  tries = 0;
  lam_sync_count= lam_old_count = 0;
  num_nodes = am_procs();

  if (lam_myproc != 0) {
    while (!done) {
      
      /* issue request */
      tries++;
      am_request_4(0,(handler_t) lam_sync_req_h,lam_myproc,0,0,0);
      /* wait for reply or time out */
      gettimeofday(&tv_begin,NULL); 
      total_time =0;

      /* wait 1000 msec for zero to respond thread to change tick value */
      while ( (total_time < 20000) && (lam_sync_count == lam_old_count)) {
	gettimeofday(&tv_end,NULL); 
	if (tv_begin.tv_usec > tv_end.tv_usec) {
	  tv_end.tv_usec += 1000000;
	  tv_end.tv_sec--;
	}
	lapsed.tv_usec = tv_end.tv_usec - tv_begin.tv_usec;
	lapsed.tv_sec =  tv_end.tv_sec - tv_begin.tv_sec;  
	total_time = lapsed.tv_sec*1000000+ lapsed.tv_usec ;

	am_poll();
      }
      if (lam_sync_count == lam_procs) 
	done =1;

      if (tries > num_nodes*1000) 
	done =1;

      
      if (lam_old_count == lam_sync_count)
	lam_restore_flow(0);
      else 
	lam_old_count = lam_sync_count;

    }

    if (lam_old_count == 0) {
      printf(" node zero not responding \n");
    } 
    else { 
      if (lam_sync_count != lam_procs) {
	printf(" some nodes not responding \n");
      }
    }
  } /* if I am not zero */
  else { /* am zero, wait for up to 15 seconds before giving up */
    lam_sync_count++;

    gettimeofday(&tv_begin,NULL); 
    total_time =0;
    
    while ( !done ) {
      am_poll();
      
      gettimeofday(&tv_end,NULL); 
      if (tv_begin.tv_usec > tv_end.tv_usec) {
	tv_end.tv_usec += 1000000;
	tv_end.tv_sec--;
      }
      lapsed.tv_usec = tv_end.tv_usec - tv_begin.tv_usec;
      lapsed.tv_sec =  tv_end.tv_sec - tv_begin.tv_sec;  
      total_time = lapsed.tv_sec*1000000+ lapsed.tv_usec ;

      if (total_time > 15000000*num_nodes )
	done =1;

      if ( lam_sync_count == lam_procs) 
	done =1;
      
    }

    if (lam_sync_count != lam_procs) {
      printf("proc %d: nodes which did not respond: ",lam_myproc);
      for (i=1; i< lam_procs; i++ ) {
	if (lam_sync_flag[i] == 0) 
	  printf("%d ",i);
      }
      printf("%\n");
    }
  }

  if (lam_sync_count != lam_procs)
    return -1;

  return 1;
}
