/****************************************************************************/
/*                                                                          */
/* Copyright    **    **  ********  ********  **    **   ******     Limited */
/* Copyright    ***  ***  ********  ********  **   **   ********    Limited */
/* Copyright    ********  **           **     **  **    **    **    Limited */
/* Copyright    ** ** **  ******       **     ****      **    **    Limited */
/* Copyright    **    **  **           **     ** **     **    **    Limited */
/* Copyright    **    **  **           **     **  **    **    **    Limited */
/* Copyright    **    **  ********  ********  **   **   ********    Limited */
/* Copyright    **    **  ********  ********  **    **   ******     Limited */
/*                                                                          */
/*                      **       ****     ******    ******                  */
/*                     ***      ******   ********  ********                 */
/*                    ****     **    **  **    **  **   ***                 */
/*                      **      *******   *******      ***                  */
/*                      **           **        **    ***                    */
/*                      **           **        **   ***                     */
/*                      **     ********  ********  ********                 */
/*                      **      ******    ******   ********                 */
/*                                                                          */
/* Author:  Eric Barton - September 1992                                    */
/****************************************************************************/

/*
 * Copyright (c) 1992 by Meiko Ltd.
 */

#pragma ident "$Id: check.c,v 1.5 1994/08/30 14:55:21 jim Exp $"
/*             $Source: /home/cvs/master/cs2parts/ew/check.c,v $ */

#include <stdio.h>
#include <string.h>
#include <errno.h>

extern int      optind;
extern char    *optarg;

#define BLOCKSIZE	32
#define LABELBASE	"ew_cb_"

typedef struct
{
   int		offset;
   char		name[64];
}	LABEL;

main (int argc, char **argv)
{
   char		line[256];
   LABEL        labels[64];
   LABEL	label;
   int		n = 0;
   char	       *baseName = LABELBASE;
   int		block = BLOCKSIZE;
   int		verbose = 0;
   int		removeOnError = 0;
   char	       *filename = NULL;
   char	       *tmpFile;
   int		i;
   int		j;
   int		c;

   while ((c = getopt (argc, argv, "n:b:vr")) != -1)
      switch (c)
      {
      case 'b':
	 if ((block = atoi (optarg)) <= 0)
	 {
	    fprintf (stderr, "ERROR: bad block size %s\n", optarg);
	    exit (1);
	 }
	 break;
	 
      case 'n':
	 baseName = optarg;
	 break;

      case 'v':
	 verbose++;
	 break;

      case 'r':
	 removeOnError++;
	 break;
	 
      default:
	 fprintf (stderr, "ERROR: bad option\n");
	 exit (1);
      }

   if (optind < argc)
      filename = argv[optind++];
   
   for (;;)					/* parse */
   {
      if (!fgets (line, sizeof (line), stdin))
      {
	 if (!errno)
	    break;
	 
	 perror ("ERROR on stdin: ");

	 if (removeOnError && filename)
	    unlink (filename);
	 
	 exit (1);
      }
      
      if (sscanf (line, "%*s %i %*i %*i %*i %*i %*i %64s",
		  &labels[n].offset, labels[n].name) == 2 &&
	  !strncmp (baseName, labels[n].name, strlen (baseName)))
      {
	 if (++n == sizeof (labels)/sizeof (label))
	 {
	    fprintf (stderr, "ERROR: too many labels\n");

	    if (removeOnError && filename)
	       unlink (filename);
	    
	    exit (1);
	 }
      }
   }

   for (i = 0; i < n - 1; i++)			/* sort */
      for (j = i + 1; j < n; j++)
	 if (labels[i].offset > labels[j].offset)
	 {
	    label = labels[i];
	    labels[i] = labels[j];
	    labels[j] = label;
	 }
   
   for (i = c = 0; i < n - 1; i += 2)
   {
      for (j = 0; labels[i].name[j]; j++)
	 if (labels[i].name[j] != labels[i+1].name[j])
	    break;
      
      if (strcmp (labels[i].name + j, "begin") ||
	  strcmp (labels[i + 1].name + j, "end") ||
	  labels[i].offset/BLOCKSIZE != (labels[i+1].offset-1)/BLOCKSIZE)
      {
	 fprintf (stderr, "ERROR: %24s: %#8x, %24s: %#8x\n",
		  labels[i].name, labels[i].offset,
		  labels[i+1].name, labels[i+1].offset);
	 c++;
	 continue;
      }

      if (filename)
      {
	 tmpFile = tmpnam (NULL);

	 sprintf (line, "echo '%x,%x?ai' | adb %s > %s", 
		  labels[i].offset, (labels[i+1].offset - labels[i].offset)/4,
		  filename, tmpFile);
	 
	 if (system (line))
	 {
	    fprintf (stderr, "Can't do <%s>\n", line);
	    c++;
	    break;
	 }
	 
	 sprintf (line, "grep '\\[' %s", tmpFile);
	 
	 if (!system (line))
	 {
	    fprintf (stderr, "ERROR: memory references\n");
	    c++;
	    continue;
	 }

	 unlink (tmpFile);
      }
      
      if (verbose)
	 fprintf (stderr, "       %24s: %#8x, %24s: %#8x\n",
		  labels[i].name, labels[i].offset,
		  labels[i+1].name, labels[i+1].offset);
   }
   
   if (n & 1)
   {
      fprintf (stderr, "ERROR: odd label %s\n", labels[n-1].name);
      c++;
   }

   if (c && removeOnError && filename)
      unlink (filename);
   
   exit (c ? 1 : 0);
}
