#include <assert.h>
#include "meiam.h"
#include "queue.h"

/*                                                                      tab:4
 *
 * meiam.c - Active Message calls for the CS-2
 *
 * "Copyright (c) 1993 by Chad Yoshikawa and The Regents of the University
 * of California.  All rights reserved."
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * Author:                      Chad Yoshikawa
 * Version:                     1
 * Filename:                    meiam.c
 * History:
 */

 

/*Active Message format:
  packet type
  function pointer
  arg0
  arg1
  arg2
  arg3
  */

/*
 * Globals for the Split-C library
 */

int ___PROCS;
#define PROCS ___PROCS
int ___MYPROC;
#define MYPROC ___MYPROC

/*
 * Update the number of flow-control credits
 */

#define UPDATE_K_OUT() {unsigned int K_tmp;\
			    K_tmp = K_S;\
			    K_Out+=(K_tmp - K_S_Old);\
			    K_S_Old = K_tmp;}


unsigned volatile int K_S = 0; 
unsigned volatile int K_S_Old= 0; 

#define K_STORES_Out  (K_S - K_S_Old)

volatile int did_reply=0;        /*Request handler did reply?*/
static int K_Out = MAX_NUM_OUT;  /*max oustanding requests per processor*/



typedef void (*MEIAM_handler_4i)(int AM_myproc,int arg1, int arg2, int arg3, int arg4);

typedef void (*MEIAM_handler_2i1d)(int AM_myproc,int arg1, int arg2, double arg4);

queue_t *am_queue; /*Queue for all incoming/outgonig active messages*/



/* Structures needed for stores*/
ELAN_DMA *am_store_dmaDesc;
ELAN_EVENT *am_store_dmaEvent;


/*Just for flow-control reasons...*/
static void bogus_handler()
{
}


void timer_init()
{
}

/* 'Hardware barrier' -- no polling */
void am_hw_barrier()
{
	ew_fgsync(ew_base.segGroup);
}

/* Disable active messages - clean up*/
void am_disable()
{

/*Wait for requests to come back and elan to finish w/ 'stores'*/
while ((K_Out+K_STORES_Out)<MAX_NUM_OUT)
	am_poll();

old_am_barrier();

am_hw_barrier();

}

int am_max_size()
{
    return 2147483647;/*MAXINT*/
}


void am_store_init()
{
    /*One dma descriptor for store DMAs (store is syncrhonous)*/
    am_store_dmaDesc = (ELAN_DMA *)memalign(8,sizeof(ELAN_DMA));
    am_store_dmaDesc->dma_type = DMA_TYPE(TR_TYPE_BYTE,DMA_NORMAL,8);
    am_store_dmaDesc->dma_destEvent = NULL;
    
    /*One event for store DMAs (store is syncrhonous)*/
    am_store_dmaEvent = (ELAN_EVENT *)memalign(8,sizeof(ELAN_EVENT));
    am_store_dmaDesc->dma_sourceEvent = am_store_dmaEvent;
    ELAN_CLEAREVENT(am_store_dmaEvent);
}


/*
 * Enable active message functions
 */
void am_enable()
{
    int size; /*queue_size*/
    
    ew_baseInit();
    
    am_store_init();
    
    ___MYPROC = am_my_proc();
    ___PROCS =am_procs();
    
    
    size = ew_rup2(K_Out*___PROCS);  /*KP storage per node*/

    /* Create a queue of size KP for outgoing requests and DMAs
     * and outgoing replies
     */
    am_queue = CreateQueue(size*2,___PROCS); 
    am_barrier_init();
    am_hw_barrier();
}

int am_procs(void)
{
    return (ew_base.segGroup->g_size);
}

int am_my_proc(void)
{
    return (ew_base.segGroup->g_self);
}


/*
 * Send a one-way request, i.e. a message to a handler that doesn't
 * respond
 */
int am_request_4_one_way(vnn_t dest, handler_t handler, int arg0,
			 int arg1, int arg2, int arg3)
{
    int head;
    volatile int *message_ptr;
    
    /*loopback*/
    if (dest==___MYPROC){
	((MEIAM_handler_4i)handler)(dest,arg0,arg1,arg2,arg3);
    } else {
	/*Wait for responses if necessary*/
	while (!K_Out && (K_S==K_S_Old))
	    am_poll();
	
	head = am_queue->list_head;
	message_ptr = am_queue->message_ptr+(head*QUEUE_ENTRY_WORDS);
	
	assert(*(message_ptr+QREMOTE_NODE)==-1);
	message_ptr[QTYPE]=INT_PKT | STORE_PKT; /*Mark as one-way*/
	message_ptr[QARG1]=handler;
	message_ptr[QARG2]=___MYPROC;
	message_ptr[QARG3]=arg0;
	message_ptr[QARG4]=arg1;
	message_ptr[QARG5]=arg2;
	message_ptr[QARG6]=arg3;
	message_ptr[QREMOTE_NODE]=(dest|STORE_PKT);
	am_queue->list_head=(head+1)&(am_queue->queue_size-1);
	
	UPDATE_K_OUT();
	K_Out--; /* One less outstanding message */
	
    }
    am_poll(); /* Drain the network */
}


/*
 * Send a request message to 'dest' processor
 */
int am_request_4(vnn_t dest, handler_t handler, int arg0,
		 int arg1, int arg2, int arg3)
{
    int head;
    volatile int *message_ptr;
    
    /*loopback*/
    if (dest==___MYPROC){
	((MEIAM_handler_4i)handler)(dest,arg0,arg1,arg2,arg3);
    } else {
	/*Wait for responses if necessary*/
	while (!K_Out && (K_S==K_S_Old))
	    am_poll();

	head = am_queue->list_head;
	message_ptr = am_queue->message_ptr+(head*QUEUE_ENTRY_WORDS);
	
	assert(*(message_ptr+QREMOTE_NODE)==-1);
	
	message_ptr[QTYPE]=INT_PKT | REQUEST_PKT;
	message_ptr[QARG1]=handler;
	message_ptr[QARG2]=___MYPROC;
	message_ptr[QARG3]=arg0;
	message_ptr[QARG4]=arg1;
	message_ptr[QARG5]=arg2;
	message_ptr[QARG6]=arg3;
	message_ptr[QREMOTE_NODE]=(dest);
	am_queue->list_head=(head+1)&(am_queue->queue_size-1);
	
	UPDATE_K_OUT();
	K_Out--; /*One less outstanding message*/
	
    }
    am_poll();/*Drain the network*/
    
}

int am_request_df(vnn_t dest, handler_t handler, int arg0, int arg1, 
		  double arg2)
{
    int iarg2_0, iarg2_1;
    int head;
    volatile int *message_ptr;
    
    /*loopback*/
    if (dest==___MYPROC)
	((MEIAM_handler_2i1d)handler)(dest,arg0,arg1,arg2);
    else{
	
	/*Wait for responses if necessary*/
	while (!K_Out && K_S==K_S_Old)
	    am_poll();
	
	

	head = am_queue->list_head;
	message_ptr = am_queue->message_ptr+(head*QUEUE_ENTRY_WORDS);

	iarg2_0 = *((int*)&arg2);
	iarg2_1 = *(((int*)&arg2)+1);
	
	assert(*(message_ptr+QREMOTE_NODE)==-1);
	message_ptr[QTYPE]=DOUBLE_PKT | REQUEST_PKT;
	message_ptr[QARG1]=handler;
	message_ptr[QARG2]=___MYPROC;
	message_ptr[QARG3]=arg0;
	message_ptr[QARG4]=arg1;
	message_ptr[QARG5]=iarg2_0;
	message_ptr[QARG6]=iarg2_1;
	message_ptr[QREMOTE_NODE]=(dest);
	am_queue->list_head=(head+1)&(am_queue->queue_size-1);

	UPDATE_K_OUT();
	K_Out--;
	
    }
    am_poll(); /*Drain the network*/
}

int am_reply_4(vnn_t requesting_vnn, handler_t handler,
	       int arg0, int arg1, int arg2, int arg3)
{
    int head;
    volatile int *message_ptr ;
    
    if (requesting_vnn==___MYPROC)
	((MEIAM_handler_4i)handler)(requesting_vnn,arg0,arg1,arg2,arg3);
    else{
	head = am_queue->list_head;
	message_ptr = am_queue->message_ptr+(head*QUEUE_ENTRY_WORDS);
	
	assert(*(message_ptr+QREMOTE_NODE)==-1);
	
	message_ptr[QTYPE]=INT_PKT | REPLY_PKT;
	message_ptr[QARG1]=handler;
	message_ptr[QARG2]=___MYPROC;
	message_ptr[QARG3]=arg0;
	message_ptr[QARG4]=arg1;
	message_ptr[QARG5]=arg2;
	message_ptr[QARG6]=arg3;
	message_ptr[QREMOTE_NODE]=(requesting_vnn);
	am_queue->list_head=(head+1)&(am_queue->queue_size-1);
	
	/*Don't poll*/
	
	assert(!did_reply);
	did_reply = 1; /*The request handler did the reply*/
    }
}

int am_reply_df(vnn_t requesting_vnn, handler_t handler, int arg0, int arg1, 
		double arg2)
{
    int iarg2_0, iarg2_1;
    int head;
    volatile int *message_ptr;
    
    if (requesting_vnn==___MYPROC)
	((MEIAM_handler_2i1d)handler)(requesting_vnn,arg0,arg1,arg2);
    else{
	iarg2_0 = *((int*)&arg2);
	iarg2_1 = *(((int*)&arg2)+1);
	
	head = am_queue->list_head;
	message_ptr = am_queue->message_ptr+(head*QUEUE_ENTRY_WORDS);
	
	assert(*(message_ptr+QREMOTE_NODE)==-1);
	
	message_ptr[QTYPE]=DOUBLE_PKT | REPLY_PKT;
	message_ptr[QARG1]=handler;
	message_ptr[QARG2]=___MYPROC;
	message_ptr[QARG3]=arg0;
	message_ptr[QARG4]=arg1;
	message_ptr[QARG5]=iarg2_0;
	message_ptr[QARG6]=iarg2_1;
	message_ptr[QREMOTE_NODE]=(requesting_vnn);
	am_queue->list_head=(head+1)&(am_queue->queue_size-1);
	
	/*Poll? - No.  Presumably, we polled to get the request message
	  that called this function.  Thus, let am_poll poll and we 
	  will simply return...*/

	/*Don't set the reply in the loop back case*/		
	assert(!did_reply);
	did_reply = 1; /*The request handler did the reply*/
    } 
}

extern ELAN_EVENT done; /*Event declared in queueThread.s*/
/*Asynchronous reply_xfer*/
void am_reply_xfer(vnn_t dest_vnn,
		   void *rva,
		   void *lva,
		   int nbytes,
		   handler_t request_handler,
		   void *handler_arg)
{
    int head;
    volatile int *message_ptr;
    
    /*Store back the data, but don't do it synchonously*/
    /*Guaranteed storage--don't need to wait for anything*/    
    head = am_queue->list_head;
    message_ptr = am_queue->message_ptr+(head*QUEUE_ENTRY_WORDS);
    
    /*Also know that this function will never loopback*/
    /*Should never happen - should be guaranteed space for
      requests AND replies*/
    
    assert(*(message_ptr+QREMOTE_NODE)==-1);
    
    ((ELAN_DMA*)message_ptr)->dma_type=DMA_TYPE(TR_TYPE_BYTE,DMA_NORMAL,8);
    ((ELAN_DMA*)message_ptr)->dma_destEvent = NULL;
    ((ELAN_DMA*)message_ptr)->dma_sourceEvent = &done;
    
    ((ELAN_DMA*)message_ptr)->dma_size = nbytes;
    ((ELAN_DMA*)message_ptr)->dma_source = lva;
    ((ELAN_DMA*)message_ptr)->dma_dest = rva;
    ((ELAN_DMA*)message_ptr)->dma_destProc = dest_vnn;
    ((ELAN_DMA*)message_ptr)->dma_pad = DMA_PKT; /*signal ELAN*/
    
    am_queue->list_head=(head+1)&(am_queue->queue_size-1);
    
    /*Send completion event*/
    am_reply_4(dest_vnn,request_handler,rva,nbytes,handler_arg,0);
    
}

/* Drain the network.  If the incoming message is a response,
 * then increase K_Out.  If the incoming message is a request, then
 * make sure it sends a response.  Important:  Drain the network
 * by removing the args into registers or something, then 
 * clear the incoming message bit
 * HANDLERS CANNOT POLL!  did_reply should be 0
 */
void am_poll()
{
  double darg;
  int iarg[2];
  register int arg0,arg1,arg2,arg3,arg4,arg5,arg6;
  volatile int *message_address;

  assert(!did_reply);

  message_address = am_queue->queue_base_ptr;
  
 Start_Poll:
  if ((message_address[QFLAG]==-1))
      return;
  
  /*MESSAGE there*/
  /*Quickly pull it out....*/
  arg0=message_address[QTYPE];
  arg1=message_address[QARG1];
  arg2=message_address[QARG2];
  arg3=message_address[QARG3];
  arg4=message_address[QARG4];
  arg5=message_address[QARG5];
  arg6=message_address[QARG6];
  /*Clear incoming buffer*/
  message_address[QFLAG]=-1;
  message_address[QFLAG2]=-1;  /*Order here is very important*/
  

  if (arg0&INT_PKT){
      ((MEIAM_handler_4i)arg1)(arg2,arg3,arg4,arg5,arg6);
  } else if (arg0&DOUBLE_PKT){
      iarg[0] = arg5;
      iarg[1] = arg6;
      darg = *((double*)iarg);
      ((MEIAM_handler_2i1d)arg1)(arg2,arg3,arg4,darg);
      
  } else if (arg0&GET_PKT){
      am_reply_xfer(arg1,arg2,arg3,arg4,arg5,arg6);
  
  } else {
      ew_exception(EW_DEBUG,"Invalid packet type in poll()");
  }
  
  /*Have to be careful about the global did_reply value*/
  if (arg0&REQUEST_PKT)	{
      if (!did_reply){
	  am_reply_4(arg2,bogus_handler,0,0,0,0);
      }
      did_reply=0;/*reset*/
  }
  /*Doesn't matter if we move this up, since
    nothing in the context can am_request anyway...*/
  else if (arg0&REPLY_PKT)	
      K_Out++;
  
  
  goto Start_Poll;  /* Drain network completely*/
}


void am_exit(int code)
{
}

/* while *flag != value, poll for messages */
void am_wait(volatile int *flag, int value)
{
  while (*flag < value)
    {
      am_poll();
    }
  (*flag)-=value;
}


static int up_buf[2];
static int down_buf;

int am_barrier_init()
{
  up_buf[0] = up_buf[1] = down_buf = 0;
}

void am_incr_handler(vnn_t node, void *val, int foo, int bar)
{
  (*((int*)val))++;
}

void am_barrier_up()
{
    int parent = (MYPROC - 1) / 2;
    int odd_child = 2 * MYPROC + 1;
    int even_child = 2 * MYPROC + 2;
    int parity = MYPROC & 1;
    /*int *buf = reduce_buf;*/
    /*reset_store_sync();*/
    /*barrier();*/
    if (MYPROC == 0) {
		if (PROCS != 1) 
			if (PROCS == 2) {
				while (up_buf[1] == 0)
					am_poll();
				/*store_sync(sizeof(int));*/
				/*return (ADD(val, *((int *) &buf[1])));*/
			} else {
				while (up_buf[0] == 0 || up_buf[1] == 0)
					am_poll();
				/*store_sync(2 * sizeof(int));*/
				/*return (ADD(ADD(val, *((int *) &buf[1])), *((int *) buf)));*/
			}
    } else {
		/*___globPtrType dest = toglobal(parent, (void *) &reduce_buf[parity]);*/
		if (odd_child >= PROCS) {
			/*__i_store(dest, val);*/
			am_request_4_one_way(parent, am_incr_handler, (int)&up_buf[parity], 0, 0, 0);
		} else if (even_child >= PROCS) {
			while (up_buf[1] == 0)
				am_poll();
			/*store_sync(sizeof(int));*/
			/*__i_store(dest, ADD(val, *((int *) &buf[1])));*/
			am_request_4_one_way(parent, am_incr_handler, (int)&up_buf[parity], 0, 0, 0);
		} else {
			while (up_buf[0] == 0 || up_buf[1] == 0)
				am_poll();
			/*store_sync(2 * sizeof(int));*/
			/*__i_store(dest, ADD(ADD(val, *((int *) &buf[1])), *((int *) buf)));*/
			am_request_4_one_way(parent, am_incr_handler, (int)&up_buf[parity], 0, 0, 0);
		}
    }
	up_buf[0] = up_buf[1] = 0;
}

void am_barrier_down()
{
    int left = 2 * MYPROC + 1;
    int right = 2 * MYPROC + 2;
    /*reset_store_sync();*/
    /*barrier();*/
    if (MYPROC != 0) {
		while (down_buf == 0)
			am_poll();
		/*store_sync(sizeof(int));*/
		/*val = *((int *) &down_buf);*/
    }
    if (left < PROCS)
		am_request_4_one_way(left, am_incr_handler, (int)&down_buf, 0, 0, 0);
		/*__i_store(toglobal(left, (int *) &down_buf), val);*/
    if (right < PROCS)
		am_request_4_one_way(right, am_incr_handler, (int)&down_buf, 0, 0, 0);
	    /*__i_store(toglobal(right, (int *) &down_buf), val);*/
	down_buf = 0;
	/*barrier();*/
    /*return (val);*/
} 

/*
 * Barrier using active messages, time scales w/ PROCS
 */
void old_am_barrier()
{
	am_barrier_up();
	am_barrier_down();
}

/*
 * Barrier using modified elan widget functions
 */
void am_barrier()
{
	gam_fgsync(ew_base.segGroup);
}

/* After completion of the function, the buffer pointed
 * to by lva *can* be modified.  Thus, this is a syncrhonous store
 * After completion, invoke the handler on the remote processor
 */
int am_store(vnn_t dest_vnn,
	     void *lva,
	     void *rva,
	     int nbytes,
	     handler_t request_handler,
	     void *handler_arg)
{
    
    if (dest_vnn == MYPROC){
	bcopy(lva, rva, nbytes);
	((MEIAM_handler_4i)request_handler)(MYPROC,rva,nbytes,handler_arg,0);
    } else {
	am_store_dmaDesc->dma_size = nbytes;
	am_store_dmaDesc->dma_source = lva;
	am_store_dmaDesc->dma_dest = rva;
	am_store_dmaDesc->dma_destProc = dest_vnn;

	/*Initiate DMA, and wait*/
	elan_dma(ew_ctx,am_store_dmaDesc);
	elan_waitevent(ew_ctx,am_store_dmaEvent,ELAN_POLL_EVENT);

	/*Send completion event*/
	am_request_4(dest_vnn,request_handler,rva,nbytes,handler_arg,0);
    }	
}

/* After completion of the function, the buffer pointed
 * to by lva *cannot* be modified.  Thus, this is an asyncrhonous store
 * After completion, invoke the handler on the remote processor
 */
int am_store_async(vnn_t dest_vnn, 
		   void *lva, 
		   void *rva, 
		   int nbytes, 
		   handler_t request_handler, 
		   void *handler_arg)
{
    int head;
    volatile int *message_ptr;
    
    if (dest_vnn == MYPROC){
	bcopy(lva, rva, nbytes);
	((MEIAM_handler_4i)request_handler)(MYPROC,rva,nbytes,handler_arg,0);
    } else {
	/*This is a request so we must make sure there is a credit before
	  we fill in the DMA*/
	/*Wait for responses if necessary*/
	while (!K_Out && (K_S==K_S_Old))
	    am_poll();
	
	head = am_queue->list_head;
	message_ptr = am_queue->message_ptr+(head*QUEUE_ENTRY_WORDS);
	
	((ELAN_DMA*)message_ptr)->dma_type=DMA_TYPE(TR_TYPE_BYTE,DMA_NORMAL,8);
	((ELAN_DMA*)message_ptr)->dma_destEvent = NULL;
	((ELAN_DMA*)message_ptr)->dma_sourceEvent = &done;
	
	((ELAN_DMA*)message_ptr)->dma_size = nbytes;
	((ELAN_DMA*)message_ptr)->dma_source = lva;
	((ELAN_DMA*)message_ptr)->dma_dest = rva;
	((ELAN_DMA*)message_ptr)->dma_destProc = dest_vnn;
	((ELAN_DMA*)message_ptr)->dma_pad =  DMA_PKT; /*signal ELAN*/
	
	am_queue->list_head=(head+1)&(am_queue->queue_size-1);
	
	/*Send completion event*/
	am_request_4(dest_vnn,request_handler,rva,nbytes,handler_arg,0);
    }
}

/*After completion of the function, the buffer pointed
  to by lva *cannot* be modified.  Thus, this is an asyncrhonous store
  After completion, invoke the handler on the remote processor
  */
int am_store_one_way_async(vnn_t dest_vnn, 
			   void *lva, 
			   void *rva, 
			   int nbytes, 
			   handler_t request_handler, 
			   void *handler_arg)
{
    int head;
    volatile int *message_ptr;
  
    if (dest_vnn == MYPROC){
	bcopy(lva, rva, nbytes);
	((MEIAM_handler_4i)request_handler)(MYPROC,rva,nbytes,handler_arg,0);
    } else {
	/*This is a request so we must make sure there is a credit before
	  we fill in the DMA*/
	/*Wait for responses if necessary*/
	while (!K_Out && (K_S==K_S_Old))
	    am_poll();
	
	head = am_queue->list_head;
	message_ptr = am_queue->message_ptr+(head*QUEUE_ENTRY_WORDS);
	
	((ELAN_DMA*)message_ptr)->dma_type=DMA_TYPE(TR_TYPE_BYTE,DMA_NORMAL,8);
	((ELAN_DMA*)message_ptr)->dma_destEvent = NULL;
	((ELAN_DMA*)message_ptr)->dma_sourceEvent = &done;
      
	((ELAN_DMA*)message_ptr)->dma_size = nbytes;
	((ELAN_DMA*)message_ptr)->dma_source = lva;
	((ELAN_DMA*)message_ptr)->dma_dest = rva;
	((ELAN_DMA*)message_ptr)->dma_destProc = dest_vnn;
	((ELAN_DMA*)message_ptr)->dma_pad =  DMA_PKT; /*signal ELAN*/
	
	am_queue->list_head=(head+1)&(am_queue->queue_size-1);
	
	/*Set completion event*/
	am_request_4_one_way(dest_vnn,request_handler,rva,nbytes,handler_arg,0);
  }
}

/*Semantics: After completion of the function, the buffer pointed
  to by lva *can* be modified.  Thus, this is a syncrhonous store
  After completion, invoke the handler on the remote processor
  ******AND THE HANDLER WILL NOT SEND A REPLY MESSAGE*********
  */
int am_store_one_way(vnn_t dest_vnn, 
		     void *lva, 
		     void *rva, 
		     int nbytes, 
		     handler_t request_handler, 
		     void *handler_arg)
{
    
    if (dest_vnn == MYPROC){
	bcopy(lva, rva, nbytes);
	((MEIAM_handler_4i)request_handler)(MYPROC,rva,nbytes,handler_arg,0);
    } else {
	am_store_dmaDesc->dma_size = nbytes;
	am_store_dmaDesc->dma_source = lva;
	am_store_dmaDesc->dma_dest = rva;
	am_store_dmaDesc->dma_destProc = dest_vnn;
	/*Initiate DMA, and wait*/
	elan_dma(ew_ctx,am_store_dmaDesc);
	elan_waitevent(ew_ctx,am_store_dmaEvent,ELAN_POLL_EVENT);
	/*Send completion event*/
	am_request_4_one_way(dest_vnn,request_handler,rva,nbytes,handler_arg,0);
    }
}

int am_get(vnn_t source_vnn, 
	   void *rva, 
	   void *lva,
	   int nbytes,
	   handler_t reply_handler,
	   void *handler_arg)
{
    int head;
    volatile int *message_ptr;
    
    /*loopback*/
    if (source_vnn == MYPROC){
	bcopy(rva, lva, nbytes);
	((MEIAM_handler_4i)reply_handler)(MYPROC,lva,nbytes,handler_arg,0);
    } else {
	
	/*Wait for responses if necessary*/
	while (!K_Out && K_S == K_S_Old)
	    am_poll();
	
	head = am_queue->list_head;
	message_ptr = am_queue->message_ptr+(head*QUEUE_ENTRY_WORDS);
	
	assert(*(message_ptr+QREMOTE_NODE)==-1);

	/*This means that the handler is am_reply_xfer*/
	message_ptr[QTYPE]=GET_PKT|REQUEST_PKT;
	message_ptr[QARG1]=___MYPROC;
	message_ptr[QARG2]=lva;
	message_ptr[QARG3]=rva;
	message_ptr[QARG4]=nbytes;
	message_ptr[QARG5]=reply_handler;
	message_ptr[QARG6]=handler_arg;
	message_ptr[QREMOTE_NODE]=(source_vnn);
	am_queue->list_head=(head+1)&(am_queue->queue_size-1);

        UPDATE_K_OUT();
	K_Out--;
	
    }
    
}


