/*									tab:2
 *
 * meiam.h - Internal header file for Meiko Active Messages
 *
 * "Copyright (c) 1995 by Chad Yoshikawa and The Regents of the University 
 * of California.  All rights reserved."
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 			Chad Yoshikawa
 * Version:			5
 * Filename:		meiam.h
 * History:
 */



#ifndef __MEIAM_INT_H_
#define __MEIAM_INT_H_

typedef unsigned int vnn_t;
typedef void (*handler_t)();

extern int AM_myproc, AM_procs;

/* Must be called before any other am calls */
void am_enable();

/* Returns virtual node number */
int am_my_proc();

/* Returns the number of nodes */
int am_procs();

/* Must be called for a graceful shutdown */
void am_disable();

/* Not in spec. Must be called to close all sockets nefore exit */
void am_exit(int code);

/* Receive messages by polling */
void am_poll();

/* Send request message with up to 4 integer arguments */
int am_request_4(vnn_t dest, handler_t fun, int arg1, int arg2, int arg3,
								 int arg4);

/* Send request message with 3 integer arguments and one double argument */
int am_request_fd(vnn_t dest, handler_t fun, int arg1, int arg2, int arg3,
								 double arg4);

/* Send reply message with up to 4 integer arguments */
int am_reply_4(vnn_t dest, handler_t fun, int arg1, int arg2, int arg3,
							 int arg4);

/* Send reply message with 3 integer arguments and one double argument */
int am_reply_fd(vnn_t dest, handler_t fun, int arg1, int arg2, int arg3,
								 double arg4);

/* Copy nbytes bytes of data from address lva on calling node to address
 * rva on dest node. Invoke handler with handler_arg on dest node when
 * all data are received on dest node.
 */
int am_put(vnn_t dest, handler_t request_handler, void *handler_arg1,
					 void *handler_arg2, void *lva, void *rva, int nbytes);

/* Retrive nbytes bytes of data from address rva on dest node to address
 * lva on calling node. Invoke handler with handler_arg on calling node when
 * all data are received on calling node.
 */
#if 0
int am_get(vnn_t dest, handler_t reply_handler, void *handler_arg1,
				     void *handler_arg2, void *sva, void *dva, int nbytes);
#endif

/* Return maximum buffer size in bytes for put and get */
int am_max_size();
#define INT_PKT 1
#define DOUBLE_PKT 2
#define GET_PKT 4
#define REQUEST_PKT 8 /*Packet is a request*/
#define REPLY_PKT 16  /*Packet is a reply*/
#define STORE_PKT 512 /*Packet is a one-way request*/
#define DMA_PKT	1024 /*Elan sends a DMA*/

#define MAX_NUM_OUT 256 /*Maximum number of requests outstanding*/

#endif

