#include <stdio.h>
#include <sys/time.h>
#include "queue.h"

#define xprintf if (DEBUG) printf

/* Stack for the elan thread*/
static char stack[STACK_SIZE];

/*
 * Create a queue object and return a pointer to it
 * size *must* be a power of 2
 */
queue_t *CreateQueue(int size,int procs)
{
    queue_t *queue_ptr; 
    char *stack_ptr;
    int i;
    
    /*Check for power of 2 size...*/
    if ((size & (-size))!=size)
	ew_exception(EW_DEBUG, "Queue size not a power of 2!");
    
    /*Create queue space...*/
    queue_ptr = (queue_t *)ew_allocate(ew_base.alloc,EW_ALIGN,sizeof(queue_t)); 
    queue_ptr->message_ptr=(int*)EW_ALIGNUP(calloc((size)+1,QUEUE_ENTRY_SIZE),EW_ALIGN);
    /*Size *2 since each request/reply can include DMA descriptor*/
    
    /*clear elems - the remote node doubles as full/empty outgoing element*/
    for (i=0;i<size;i++)
	*(queue_ptr->message_ptr+QREMOTE_NODE+QUEUE_ENTRY_WORDS*i) = -1;
    
    queue_ptr->queue_size = size;
    queue_ptr->queue_base_ptr = ew_allocate(ew_base.alloc,EW_ALIGN,QUEUE_ENTRY_SIZE);
    /*First flag for incoming conditional checks, second flag for the 
      sparc to read in*/
    queue_ptr->queue_base_ptr[QFLAG]=-1; /*Clear the incoming queue element*/
    queue_ptr->queue_base_ptr[QFLAG2]=-1; /*Clear the incoming queue element*/
    queue_ptr->list_head = 0;
    
    if (procs>1) /*No communication if only 1 processor...*/
	elan_runthread (ew_ctx, Queue_Thread, stack, sizeof(stack),1,(int)queue_ptr);
    
    return (queue_ptr);
}

