/*									tab:8
 *
 * sp1am.c - Functions that make MPLp look like generic active messages.
 *
 * 
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 		Steve Luna
 * Version:		159
 * Creation Date:	Wed Oct 13 12:46:17 1993
 * Filename:		sp1am.c
 * History:
 *	LTL	13	Thu Feb 16 22:34:12 1995
 *		Made it Generic Active Message v1.1 compliant
 *	ssl	11	Mon Apr 18 20:11:19 1994
 *		modified for SP1
 */

#include <stdio.h>
#include <limits.h>

/* This implemenetation uses IBM's MPLp library, which is a prototype of the
 * current MPL in the PSSP system software. Not many people other than
 * Berkeley should have this.  This file is the same as the MPL implementation
 * except for the include file here.
 */
#include "/usr/lpp/euih/eui/mpceui.h"
 
#define NUM_seg 256
#define NUM_irecv 1
#define AM_INT 0
#define AM_DOUBLE 1
/* #define GAM_DEBUG */

/* use these to get unique message types. since messages to a given processor
   are ordered, only need separate ones for sending (put, store) and recving
   (get) - ssl */

#define SENDTYPE(proc) (proc)
#define RECVTYPE(proc) (am_numproc+(proc))

typedef unsigned int vnn_t;
typedef	void (*i_handler_t)(vnn_t src, int arg1, int arg2, int arg3, int arg4);
typedef	void (*d_handler_t)(vnn_t src, double arg1, int arg2, int arg3);
typedef void (*xfer_handler_t)(vnn_t src_node, void *buf, int len,
															 void *info);
		
typedef struct {
  i_handler_t handler;
	int type;
	vnn_t src_node; 
	int args[4];
} sp1am_pkt;


typedef struct {
	xfer_handler_t end_fun;
	void *buf;
	int len;
	void *info;
	int mid;
	int src_node;
} xfer_segment;

static sp1am_pkt sp1am_buf[NUM_irecv];
static int mid[NUM_irecv];
static int am_myproc;
static int am_numproc;

static xfer_segment SP1AM_seg[NUM_seg];
static int SP1AM_free_seg[NUM_seg/32];
static char fho[256];		/* table to find #shifts to highest order bit of byte */
static int dontcare, SP1AM_msg_type, source[NUM_irecv];
static int xfer_source[NUM_seg], xfer_type[NUM_seg];

int am_enable()
{
	int i, b, nbuf[4], rc;

	rc = mpc_task_query(nbuf, 4, 3);
	if (rc)
	{
		fprintf(stderr, "mpc_task_query failed in am_enable()\n");
		exit(-1);
	}
	dontcare = nbuf[0];
	SP1AM_msg_type = NUM_seg;
	mpc_environ(&am_numproc, &am_myproc);	
	
	for (i = 0; i < NUM_irecv; i++)
	{
		source[i] = dontcare;
		mpc_recv((char*)&sp1am_buf[i], sizeof(sp1am_pkt), &source[i],
				 &SP1AM_msg_type, &mid[i]);
	}

	for (i = 0; i < NUM_seg/32; i++)
		SP1AM_free_seg[i] = -1;
	fho[0] = -1;
	for (i = 1; i<256; i++)
	{
		b = 7;
		while(! (i&(1<<b))) b--;
		fho[i] = b;
	}
	am_barrier_init();
	return 0;
}

void am_disable()
{
}

int am_my_proc()
{
  return am_myproc;
}

int am_procs()
{
  return am_numproc;
}

int am_max_size()
{
  return INT_MAX;
}

void am_poll()
{
	int i, j, id;
	xfer_segment seg;
	static int in_handler = 0;
	int s, count, adj;
	union {
	    double d;
	    int i[2]; 
	} u;
	int received;
	
	if (in_handler)
	{
		fprintf(stderr, "FATAL error: "
				"am_poll called from within a handler on proc %d\n",
				am_myproc);
		exit(-1);
	} 

	received = 1;
  /* keep draining until nothing left; otherwise we may deadlock
	 * this increases the overhead a bit becuase we will poll at least once
	 * more if we get something -- ltl
	 */
	while (received) {
		received = 0;	
		/* poll for small messages */
		for (i = 0; i < NUM_irecv; i++)
		{
			s=mpc_status(mid[i]);
			if (s == sizeof(sp1am_pkt))			    
			{
				in_handler = 1;
				if (sp1am_buf[i].type == AM_INT) {
					((i_handler_t)sp1am_buf[i].handler)(sp1am_buf[i].src_node,
														 sp1am_buf[i].args[0], sp1am_buf[i].args[1],
														 sp1am_buf[i].args[2], sp1am_buf[i].args[3]);
				}
				else {
					u.i[0] = sp1am_buf[i].args[0];
					u.i[1] = sp1am_buf[i].args[1];
					((d_handler_t)sp1am_buf[i].handler)(sp1am_buf[i].src_node, u.d,
																 sp1am_buf[i].args[2], sp1am_buf[i].args[3]);
				}
				in_handler = 0;
				received = 1;
				source[i] = dontcare;
				mpc_recv((char*)&sp1am_buf[i], sizeof(sp1am_pkt), &source[i],
						 &SP1AM_msg_type, &mid[i]);
			}
		}

		/* poll for bulk messages */
		for (i = (NUM_seg/32)-1; i>=0; i--)
			if (SP1AM_free_seg[i] != -1)
				for (j = 31; j>=0; j--)
					if ((SP1AM_free_seg[i]&(1<<j)) == 0) {
					    id = (seg=SP1AM_seg[(i<<5)+j]).mid;
					    s = mpc_status(id);
						if (s >= 0)
						{
							SP1AM_free_seg[i] ^= (1<<j);
							seg.end_fun(seg.src_node, seg.buf, seg.len, seg.info);
							received = 1;
						}
}
  }
}

void am_wait(volatile int *flag, int value)
{
 	while (*flag < value)
	{
		am_poll();
	}
	(*flag)-=value;
}

void am_poll_wait(volatile int *flag, int value)
{
	do
		am_poll();
	while (*flag < value);
	(*flag)--;
}

void am_request_4(vnn_t dest, void *fun, int arg1, int arg2, int arg3,
									int arg4)
{
	sp1am_pkt sp1am_sbuf;

	sp1am_sbuf.handler = fun;
	sp1am_sbuf.src_node = am_myproc;
	sp1am_sbuf.type = AM_INT;
	sp1am_sbuf.args[0] = arg1;
	sp1am_sbuf.args[1] = arg2;
	sp1am_sbuf.args[2] = arg3;
	sp1am_sbuf.args[3] = arg4;
	mpc_bsend((char*)&sp1am_sbuf, sizeof(sp1am_pkt), dest, SP1AM_msg_type);
	am_poll();
}

void am_request_fd(vnn_t dest, void *fun, double arg1, int arg2, int arg3)
{
	union {
   double d;
	 int i[2]; 
	} u;
	sp1am_pkt sp1am_sbuf;

	u.d = arg1;
	sp1am_sbuf.handler = fun;
	sp1am_sbuf.type = AM_DOUBLE;	
	sp1am_sbuf.src_node = am_myproc;
	sp1am_sbuf.args[0] = u.i[0];
	sp1am_sbuf.args[1] = u.i[1];	
	sp1am_sbuf.args[2] = arg2;
	sp1am_sbuf.args[3] = arg3;

	mpc_bsend((char*)&sp1am_sbuf, sizeof(sp1am_pkt), dest, SP1AM_msg_type);
	am_poll();
}

void am_reply_4(int dest, void *fun, int arg1, int arg2, int arg3, int arg4)
{
	sp1am_pkt sp1am_sbuf;

	sp1am_sbuf.handler = fun;
	sp1am_sbuf.type = AM_INT;	
	sp1am_sbuf.src_node = am_myproc;	
	sp1am_sbuf.args[0] = arg1;
	sp1am_sbuf.args[1] = arg2;
	sp1am_sbuf.args[2] = arg3;
	sp1am_sbuf.args[3] = arg4;
	mpc_bsend((char*)&sp1am_sbuf, sizeof(sp1am_pkt), dest, SP1AM_msg_type);
}

void am_reply_fd(int dest, void *fun, double arg1, int arg2, int arg3)
{
	union {
   double d;
	 int i[2]; 
	} u;
	sp1am_pkt sp1am_sbuf;
	
	u.d = arg1;		
	sp1am_sbuf.handler = fun;
	sp1am_sbuf.type = AM_DOUBLE;	
	sp1am_sbuf.src_node = am_myproc;
	sp1am_sbuf.args[0] = u.i[0];
	sp1am_sbuf.args[1] = u.i[1];
	sp1am_sbuf.args[2] = arg2;
	sp1am_sbuf.args[3] = arg3;
	mpc_bsend((char*)&sp1am_sbuf, sizeof(sp1am_pkt), dest, SP1AM_msg_type);
}

int allocate_segment()
{
	int i, b, f, free;
	for (i = (NUM_seg/32)-1; i>=0; i--)
	{
		if (SP1AM_free_seg[i])
		{
			free = SP1AM_free_seg[i];
			if (f=(free>>24)&0xff) {
				b = 24+fho[f];
			} else if (f=(free>>16)&0xff) {
				b = 16+fho[f];
			} else if (f=(free>>8)&0xff) {
				b = 8+fho[f];
			} else {
				b = fho[free&0xff];
			}
		  SP1AM_free_seg[i] = free ^ (1<<b);
			return (i<<5)+b;
		}
	}
	fprintf(stderr, "error in sp1am.c, couldn't allocate segment\n");
}

void am_open_send_segment(int src_node, void *base, int count, void *fun,
										 void *info)
{
	int seg;
	
	seg = allocate_segment();
	xfer_source[seg] = dontcare;
	xfer_type[seg] = SENDTYPE(src_node);
	mpc_recv(base, count, &xfer_source[seg], &xfer_type[seg],
			 &SP1AM_seg[seg].mid);
	SP1AM_seg[seg].end_fun = fun;
	SP1AM_seg[seg].info = info;
	SP1AM_seg[seg].len = count;	
	SP1AM_seg[seg].buf = base;
	SP1AM_seg[seg].src_node = src_node;
}

void am_open_receive_segment(int src_node, void *base, int count, void *fun,
										 void *info)
{
	int seg;
	
	seg = allocate_segment();
	xfer_source[seg] = dontcare;
	xfer_type[seg] = RECVTYPE(src_node);
	mpc_recv(base, count, &xfer_source[seg], &xfer_type[seg],
			 &SP1AM_seg[seg].mid);
	SP1AM_seg[seg].end_fun = fun;
	SP1AM_seg[seg].info = info;
	SP1AM_seg[seg].len = count;	
	SP1AM_seg[seg].buf = base;
	SP1AM_seg[seg].src_node = src_node;
}

void am_xfer(int dest, int seg, void *buf, int size)
{
	mpc_bsend(buf, size, dest, seg);
	am_poll();
}

void am_reply_xfer(int dest, int seg, void *buf, int size)
{
	mpc_bsend(buf, size, dest, seg);
}


/* This used to be  __store() in Split-C -- ltl */
int am_store(vnn_t dest, void *src_buf, void *dest_buf, int len,
							xfer_handler_t request_handler, void *handler_arg)
{
	if (dest == am_myproc) {
		memcpy(dest_buf, src_buf, len);
		/* invoke the handler */
		(request_handler)(dest, dest_buf, len, handler_arg);
	}
	else {
#ifdef GAM_DEBUG
	    fprintf(stderr,"%d store msgid %d\n",am_myproc,
		    *((int*)handler_arg));
#endif		    
/*		am_request_4(dest, am_open_send_segment, (int)dest_buf, len,
								 (int)request_handler, (int)handler_arg); */
		am_request_4(dest, am_open_send_segment, (int)dest_buf, len,
								 (int)request_handler, *((int*)handler_arg));		
		am_xfer(dest, SENDTYPE(am_myproc), src_buf, len);
	}
}

void am_xfer_async(int dest, int seg, void *buf, int size, int *msgid)
{
	mpc_send(buf, size, dest, seg, msgid);
	am_poll(); 
}

int am_store_async(vnn_t dest, void *src_buf, void *dest_buf,
									 int len, xfer_handler_t request_handler,
									 void *request_handler_arg,
									 xfer_handler_t endfunc, void *endfunc_arg)
{
        int mid;
    
	if (dest == am_myproc) {
		memcpy(dest_buf, src_buf, len);
		/* invoke the handler */
		(request_handler)(dest, dest_buf, len, request_handler_arg);
	}
	else {
#ifdef GAM_DEBUG
	    fprintf(stderr,"%d store msgid %d\n",am_myproc,
		    *((int*)request_handler_arg));
#endif		    
/*		am_request_4(dest, am_open_send_segment, (int)dest_buf, len,
								 (int)request_handler, (int)request_handler_arg); */
		am_request_4(dest, am_open_send_segment, (int)dest_buf, len,
								 (int)request_handler, *((int*)request_handler_arg));
		am_xfer_async(dest, SENDTYPE(am_myproc), src_buf, len,
			      &mid);
	    }
}


/* This used to be __get_ctr() in Split-C -- ltl */
int am_get(vnn_t dest, void *src_buf, void *dest_buf, int len,
					 xfer_handler_t reply_handler, void *handler_arg)
{
	int segId;

	if (dest == am_myproc) {
    memcpy(dest_buf, src_buf, len);
		/* invoke the handler */
		(reply_handler)(dest, dest_buf, len, handler_arg);		
	} 
	else {
		/* Set up a segment */
		am_open_receive_segment(dest, dest_buf, len,
								 reply_handler, handler_arg);			
		/* Call the function in the remote node which does the xfer */
 		am_request_4(dest, (void *)am_reply_xfer, RECVTYPE(dest), src_buf,
					len, 0);
	}
}




