/*									tab:2
 *
 * am_int.h - Paragon Active Messages include file (for internal use only)
 *
 * "Copyright (c) 1994 by Lok Tin Liu and The Regents of the University 
 * of California.  All rights reserved."
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 			Lok Tin Liu
 * Version:			145
 * Creation Date:	Wed Aug 17 13:58:55 1994
 * Filename:		am_int.h
 * Description:
 *    This file contains the PAM header file developed by LTL at UCB.
 *    It contains no Intel proprietary information.
 * History:
 *	LTL	142	Sun Jul 23 13:42:14 1995
 *		Made am_request_df() and am_reply_df() consistent with GAM spec 1.1.
 *    Renamed PAM_REQUEST_3I1D to PAM_REQUEST_2I1D and PAM_REPLY_3I1D to
 *    PAM_REPLY_2I1D. Removed PAM_REQUEST_1I2D and PAM_REPLY_1I2D.
 *	LTL	140	Thu Jul  6 21:06:28 1995
 *		PAM Release 1.6
 */



#ifndef PAM_INT_H
#define PAM_INT_H
#define PAM_INTERNAL

#define PAM_PLACEHOLDER -1

/* message tags */
#define PAM_REQUEST_5I 1
#define PAM_REQUEST_2I1D 2
#define PAM_REPLY_5I 4
#define PAM_REPLY_2I1D 5
#define PAM_REQUEST_STORE 7
#define PAM_STORE_ENDFUNC 8
#define PAM_REPLY_GET 9
#define PAM_GET 10
#define PAM_SEND_5I 11

#define PAM_MAXPKT 8192 /* maximum packet size */

#include "am.h"
		
extern pam_recbuf_t *pam_recbuf_base; /* base of receive buffer */
extern pam_recbuf_t *pam_recbuf_bound; /* bound of receive buffer */
extern pam_recbuf_t *pam_rptr; /* read pointer */   
extern int *pam_outstanding;
extern int pam_max_outstanding;
extern vnn_t pam_mynode;
extern int pam_replied;
extern volatile int pam_store_counter;


extern int am_init_recbuf(pam_recbuf_t *recbuf_base,
													 pam_recbuf_t *recbuf_bound);  
extern int am_setup(pam_recbuf_t *base, pam_recbuf_t *bound);
extern int am_reset(pam_recbuf_t *base, pam_recbuf_t *bound);
extern int am_teardown();


typedef void (pam_handler_5i)(vnn_t src_node, int arg0, int arg1, int arg2,
															int arg3, int arg4);
typedef void (pam_handler_2i1d)(vnn_t src_node, int arg0, int arg1,
																double arg2);
typedef void (pam_endfunc)(vnn_t dest_node, void *src_buf, int len,
													 int arg0);
typedef void (pam_xfer_handler)(vnn_t src_node, void *dest_buf, int len,
																int arg0);



extern void pam_store_endfunc(vnn_t dest_node, void *src_buf, int len,
															void *arg0);

extern int am_store_async_internal(vnn_t dest_node,
													void *src_buf, void *dest_buf, int len,
													handler_t request_handler,
													void *request_handler_arg1,
													handler_t endfunc, void *endfunc_arg1);

/* the side that initiates the get is the destination */
extern int am_get_internal(vnn_t dest_node,
									void *src_buf, void *dest_buf, int len,
									handler_t reply_handler, void *handler_arg1);

extern int am_reply_get(unsigned long return_route,
									void *src_buf, void *dest_buf, int len,
									handler_t reply_handler, void *handler_arg1);

#undef PAM_INTERNAL
#endif
