/*									tab:2
 *
 * pamlib.c - PAM user library
 *
 * "Copyright (c) 1994 by Lok Tin Liu and The Regents of the University 
 * of California.  All rights reserved."
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 			Lok Tin Liu
 * Version:			432
 * Creation Date:	Tue Jun  7 11:18:04 1994
 * Filename:		pamlib.c
 * Description:
 *    This file contains the PAM user library developed by LTL at UCB.
 *    It contains no Intel proprietary information except for the
 *    way the post page is accessed.
 * History:
 *	LTL	407	Sun Jul 23 13:06:21 1995
 *		Made am_request_df() and am_reply_df() consistent with GAM spec 1.1.
 *    Removed case for PAM_REQUEST_1I2D and PAM_REPLY_1I2D in am_poll().
 *    Added calls to gsync() in am_enable() and am_disable().
 *		Removed inline declaration for am_enq_7i(), am_poll(), and am_wait()
 *    so that we can use cc to compile instead of gcc.
 *	LTL	403	Thu Jul  6 21:00:57 1995
 *		PAM Release 1.6
 */



/*
 * pamlib.c
 *
 * PAM interface library
 */

/*
 * Message processor interface initialization
 */


#include "am_int.h"
#include <nx.h>

#define PAM_ALIGN 32
#define MAXCOUNT 128

extern void	pam_post1();
extern void	pam_post2();
extern void	pam_post3();
extern void	pam_post4();
extern void	pam_post5();
extern void	pam_post6();
extern void	pam_post7();
extern void	pam_post8();
extern void	pam_post9();
extern void	pam_post10();


void	(*pam_intf_vec1)() = pam_post1;
void	(*pam_intf_vec2)() = pam_post2;
void	(*pam_intf_vec3)() = pam_post3;
void	(*pam_intf_vec4)() = pam_post4;
void	(*pam_intf_vec5)() = pam_post5;
void	(*pam_intf_vec6)() = pam_post6;
void	(*pam_intf_vec7)() = pam_post7;
void	(*pam_intf_vec8)() = pam_post8;
void	(*pam_intf_vec9)() = pam_post9;
void	(*pam_intf_vec10)() = pam_post10;

/* base of receive buffer (may not be aligned at 32-byte boundary) */
pam_recbuf_t *pam_rbase;
/* base of receive buffer (aligned at 32-byte boundary) */
pam_recbuf_t *pam_recbuf_base;
/* bound of receive buffer */
pam_recbuf_t *pam_recbuf_bound;
/* read pointer */
pam_recbuf_t *pam_rptr; 

int *pam_outstanding;
int pam_max_outstanding;
vnn_t pam_mynode;
volatile int pam_store_counter;
volatile int pam_nop_counter;
volatile int pam_replied;


volatile int pam_taskid;

void pam_nop_handler()
{
		++pam_nop_counter;
}


void pam_store_endfunc(vnn_t dest_node, void *src_buf, int len,
											 void *arg0)
{
  pam_store_counter += len;
}

vnn_t am_my_proc()
{
  return mynode();
}

int am_procs()
{
  return numnodes();
}

/* initialize PAM receive buffer */
int pam_init_recbuf(pam_recbuf_t *recbuf_base, pam_recbuf_t *recbuf_bound)
{
  pam_recbuf_t *p;

	if ((recbuf_base != 0) && (recbuf_bound != 0)
			&& (recbuf_base <= recbuf_bound)) {
 	  for (p = recbuf_base; p <= recbuf_bound; ++p) {
			p->flag = 0;
			p->arg0 = -1;
			p->arg1 = -1;
			p->arg2 = -1;
			p->arg3 = -1;
			p->arg4 = -1;
			p->arg5 = -1;
			p->arg6 = -1;
    }		
  }
	else {
    printf("Error: pam_init_recbuf() invalid buffer address on node %d\n",
					 mynode());
    return -1;
	}
}

/* enable PAM layer
 *
 * input: maximum number of outstanding requsts per destination
 * output: base and bound of receive buffer
 *         return -1 if error
 */

int am_enable(int max_outstanding, volatile pam_recbuf_t **recbuf_base,
							volatile pam_recbuf_t **recbuf_bound)
{
  int buf_size, i, count;
  unsigned ptr;
	volatile global_pam_taskid, tmp, done;
	
	/* allocate buffers for request, reply, and store endfunc */
	buf_size = numnodes()*max_outstanding*3;
  pam_rbase =	(pam_recbuf_t *)malloc(buf_size*sizeof(pam_recbuf_t)+PAM_ALIGN);
	if (pam_rbase) {
		/* wire the receive buffer so that it is not pageable; better hope that
		 * it is contagious in physical memory !!!
		 */
			
		if (mcmsg_wire(pam_rbase, buf_size*sizeof(pam_recbuf_t)+PAM_ALIGN) != 0) {
			printf("Error: can't wire PAM receive buffer on node %d\n", mynode());
			return -1;				
    }
		/* align at 32-byte boundary */
		pam_recbuf_base = (pam_recbuf_t *)
				(((unsigned)pam_rbase + (PAM_ALIGN - 1)) & ~(PAM_ALIGN - 1));
		*recbuf_base = pam_recbuf_base;
		pam_recbuf_bound = pam_recbuf_base + (buf_size - 1);
		*recbuf_bound = pam_recbuf_base + (buf_size - 1);
		pam_max_outstanding = max_outstanding;
		pam_mynode = mynode();
		pam_outstanding = (int *)malloc(numnodes()*sizeof(int));
		pam_store_counter = 0;
		if (pam_outstanding == 0) {
			printf("Error: can't allocate pam_outstanding[] on node %d\n", mynode());
			return -1;				
		}
		for (i = 0; i < numnodes(); ++i)
			pam_outstanding[i] = 0;
    done = count = 0;
    while ((count < MAXCOUNT) && !(done)) {
			pam_taskid = -1;
			pam_get_taskid(&pam_taskid);
			while (pam_taskid == -1); /* wait for pam_taskid to be set */
			global_pam_taskid = pam_taskid;
			/* find the lowest taskid among all nodes */
			gilow((long *)&global_pam_taskid, 1, (long *)&tmp);
			/* if any one of the nodes fails to allocate a taskid */
			if (global_pam_taskid <= -2) {
				printf("Fail to allocate PAM taskid\n");
				exit(-1);
			}
			else if (global_pam_taskid == pam_taskid) {
         done = 1;
			}
			/* check if consensus is reached */
			gland((long *)&done, 1, (long *)&tmp);			
  		++count;
    }
		if (!done) {
			printf("Fail to allocate PAM taskid\n");
  		exit(-1);
		}
		pam_setup(pam_recbuf_base, pam_recbuf_bound, PAM_MAXPKT);

		pam_rptr = pam_recbuf_base;
		pam_init_recbuf(pam_recbuf_base, pam_recbuf_bound);
	  am_barrier_init();
    pam_store_counter = 0;
		gsync();
		return 0;
  }
	else {
  	printf("Error: can't allocate PAM receive buffer on node %d\n", mynode());
    return -1;
	}
			
}


int am_disable()
{
	gsync();
  /* free receive buffers */
  if (pam_rbase)
		free(pam_rbase);
	if (pam_outstanding)
		free(pam_outstanding);	
	/* teardown PAM */
	pam_teardown();
	
}

int am_max_size()
{
		return AM_MAX_PKT;
}


void am_poll()
{
	register double *dp;
	register unsigned long arg0, tag, src_node;
	register handler_t handler;
	register unsigned long flag, arg4;


	flag = pam_rptr->flag;
	
	while (flag != 0) {
		handler = (handler_t)pam_rptr->arg1;
  	arg0 = pam_rptr->arg0;
		tag = arg0 & 0xffff;
		src_node = arg0 >> 16;

		switch (tag) {
			case PAM_REQUEST_5I:
				pam_replied = 0;
				((pam_handler_5i *)(pam_rptr->arg1))(src_node, pam_rptr->arg2,
									 pam_rptr->arg3, pam_rptr->arg4,
									 pam_rptr->arg5, pam_rptr->arg6);
				if (!pam_replied) {
					am_reply_5(src_node, pam_nop_handler, 0, 0, 0, 0, 0);
				}
				break;
			case PAM_SEND_5I:
				((pam_handler_5i *)(pam_rptr->arg1))(src_node, pam_rptr->arg2,
									 pam_rptr->arg3, pam_rptr->arg4,
									 pam_rptr->arg5, pam_rptr->arg6);
				break;				
			case PAM_REQUEST_2I1D:
				pam_replied = 0;				
				dp = (double *)&(pam_rptr->arg5);			
				((pam_handler_2i1d *)(pam_rptr->arg1))(src_node, pam_rptr->arg2,
									 pam_rptr->arg3, *dp);
				if (!pam_replied) {
					am_reply_5(src_node, pam_nop_handler, 0, 0, 0, 0, 0);
				}
				break;
			case PAM_REPLY_5I:
				((pam_handler_5i *)(pam_rptr->arg1))(src_node, pam_rptr->arg2,
									 pam_rptr->arg3, pam_rptr->arg4,
									 pam_rptr->arg5, pam_rptr->arg6);
				--pam_outstanding[src_node];
				break;
			case PAM_REPLY_2I1D:
				dp = (double *)&(pam_rptr->arg5);			
				((pam_handler_2i1d *)(pam_rptr->arg1))(src_node, pam_rptr->arg2, 
									 pam_rptr->arg3, *dp);
				--pam_outstanding[src_node];				
				break;
			case PAM_STORE_ENDFUNC:
				/* src_node is really dest_node here */
				((pam_endfunc *)(pam_rptr->arg1))(src_node,
																					(void *)pam_rptr->arg2,
																					pam_rptr->arg3,
																					pam_rptr->arg4);
				break;
			case PAM_REQUEST_STORE:
				pam_replied = 0;				
				((pam_xfer_handler *)(pam_rptr->arg1))(src_node,
																							 (void *)pam_rptr->arg2,
																							 pam_rptr->arg3,
																							 pam_rptr->arg4);
				if (!pam_replied) {
					am_reply_5(src_node, pam_nop_handler, 0, 0, 0, 0, 0);
				}				
				break;
			case PAM_REPLY_GET:
				((pam_xfer_handler *)(pam_rptr->arg1))(src_node,
																							 (void *)pam_rptr->arg2,
																							 pam_rptr->arg3,
																							 pam_rptr->arg4);
				--pam_outstanding[src_node];								
				break;
			case PAM_GET:
				/* reply the get -- arg6 is actually the buffer len; handler is
				 * the reply handler
				 */
				am_reply_get(pam_rptr->arg4, (void *)pam_rptr->arg2,
										 (void *)pam_rptr->arg3,
										pam_rptr->arg6, handler, (void *)pam_rptr->arg5);
				break;
			default:
				printf("Error: unrecognized message type %d on node %d\n",
							 tag, pam_mynode);
				printf("%x %x %x %x %x %x %x %x\n", pam_rptr->flag, pam_rptr->arg0,
							 pam_rptr->arg1, pam_rptr->arg2, pam_rptr->arg3,
 							 pam_rptr->arg4, pam_rptr->arg5, pam_rptr->arg6);
				while (1);
				break;
		}

		pam_rptr->flag = 0;			

		/* advance to next entry */
		++pam_rptr;

		if (pam_rptr > pam_recbuf_bound)
			pam_rptr = pam_recbuf_base;

		flag = pam_rptr->flag;
  }
}

void am_wait(volatile int *flag, int value)
{
  while (*flag < value)
			am_poll(); 
  *flag -= value; 
}

#if 0
int am_request_5(vnn_t dest, handler_t handler, int arg0, int arg1,
												int arg2, int arg3, int arg4)
{
	register int *ip;
	
  ip = pam_outstanding + dest; 
	while (*ip >= pam_max_outstanding) 
	  am_poll(); 
	am_enq_7i(dest, PAM_PLACEHOLDER, PAM_REQUEST_5I | (pam_mynode << 16), 
							(int)handler, (int)arg0, (int)arg1, (int)arg2, (int)arg3, 
							(int)arg4); 
  ++(*ip);
}

int am_reply_5(vnn_t dest, handler_t handler, int arg0, int arg1,
											int arg2, int arg3, int arg4)
{
	am_enq_7i(dest, PAM_PLACEHOLDER, PAM_REPLY_5I | (pam_mynode << 16), 
						(int)handler, arg0, arg1, arg2, arg3, arg4); 
	pam_replied = 1; 
}

int am_send_5(vnn_t dest, handler_t handler, int arg0, int arg1,
												int arg2, int arg3, int arg4)
{
	am_enq_7i(dest, PAM_PLACEHOLDER, PAM_SEND_5I | (pam_mynode << 16), 
							(int)handler, (int)arg0, (int)arg1, (int)arg2, (int)arg3, 
							(int)arg4); 
}



int am_store_async(vnn_t dest_node, void *src_buf, void *dest_buf, int len, 
									 handler_t request_handler, void *request_handler_arg1,
									 handler_t endfunc, void *endfunc_arg1)
{
	register int *ip;
	
	if (dest_node != pam_mynode) {
		ip = pam_outstanding + dest_node; 
		while (*ip >= pam_max_outstanding) 
			am_poll(); 
		am_store_async_internal(dest_node, src_buf, dest_buf, len, 
									request_handler, request_handler_arg1, 
									endfunc, endfunc_arg1);
		++(*ip);
  }
	else {
    memmove(dest_buf, src_buf, len);
		((pam_endfunc *)(endfunc))(dest_node, src_buf, len, endfunc_arg1);
		((pam_xfer_handler *)(request_handler))(dest_node, dest_buf, len,
																						request_handler_arg1);
	}
}


int am_get(vnn_t dest_node, void *src_buf, void *dest_buf, int len,
					 handler_t reply_handler, void *handler_arg1)
{
	register int *ip;

	if (dest_node != pam_mynode) {
		ip = pam_outstanding + dest_node; 
		while (*ip >= pam_max_outstanding) 
			am_poll(); 
		am_get_internal(dest_node, src_buf, dest_buf, len, 
									reply_handler, handler_arg1);
		++(*ip);
  }	
	else {
    memmove(dest_buf, src_buf, len);
		((pam_xfer_handler *)(reply_handler))(dest_node, dest_buf, len,
																					handler_arg1);
	}	
}

#endif

int am_request_df(vnn_t dest, handler_t handler, int arg1,
												 int arg2, double arg3)
{
	register int *ip;
  union {
		double full;
		struct {
			int lo;
			int hi;
  	} half;
  } u;
	
  ip = pam_outstanding + dest;
	u.full = arg3;
  while (*ip >= pam_max_outstanding) 
		am_poll(); 
	am_enq_7i(dest, PAM_PLACEHOLDER, PAM_REQUEST_2I1D | (pam_mynode << 16),
						(int)handler, arg1, arg2, 0, u.half.lo, u.half.hi);
  ++(*ip);
}



int am_reply_df(vnn_t dest, handler_t handler, int arg1,
											 int arg2, double arg3)
{
  union {
		double full;
		struct {
			int lo;
			int hi;
  	} half;
  } u;

	u.full = arg3;
	
	am_enq_7i(dest, PAM_PLACEHOLDER, PAM_REPLY_2I1D | (pam_mynode << 16),
						(int)handler, arg1, arg2, 0, u.half.lo, u.half.hi);  
	pam_replied = 1; 
}



int am_store(vnn_t dest_node, void *src_buf, void *dest_buf, int len, 
						 handler_t request_handler, void *request_handler_arg1)
{
	register int *ip;

  if (dest_node != pam_mynode) {
		ip = pam_outstanding + dest_node;

		while (*ip >= pam_max_outstanding) 
			am_poll(); 
		am_store_async_internal(dest_node, src_buf, dest_buf, len, 
									request_handler, request_handler_arg1,
									pam_store_endfunc, 0); 
		++(*ip);   
		while (pam_store_counter < len) 
			am_poll(); 
		pam_store_counter = 0;
  }
	else {
    memmove(dest_buf, src_buf, len);
		((pam_xfer_handler *)(request_handler))(dest_node, dest_buf, len,
																						(int)request_handler_arg1);
	}
}



/* PAM post routines
 *
 * -- make sure the POST interface routine numbers match those in
 *    mcmsg_config_pam.h
 */


mcmsg_pam_setup_holder()
{
		asm("_pam_setup::");
		asm(" POST_PAM_SETUP = 40");
		asm("   orh     ha%_pam_intf_vec2, r0, r30");
		asm("   ld.l    l%_pam_intf_vec2(r30), r30");
		asm("   bri     r30");
		asm("    or     POST_PAM_SETUP, r0, r31");
}



mcmsg_pam_teardown_holder()
{
		asm("_pam_teardown::");
		asm(" POST_PAM_TEARDOWN = 41");		
		asm("   orh     ha%_pam_intf_vec1, r0, r30");
		asm("   ld.l    l%_pam_intf_vec1(r30), r30");
		asm("   bri     r30");
		asm("    or     POST_PAM_TEARDOWN, r0, r31");
}


mcmsg_pam_get_taskid_holder()
{
		asm("_pam_get_taskid::");
		asm(" POST_PAM_GET_TASKID = 42");		
		asm("   orh     ha%_pam_intf_vec2, r0, r30");
		asm("   ld.l    l%_pam_intf_vec2(r30), r30");
		asm("   bri     r30");
		asm("    or     POST_PAM_GET_TASKID, r0, r31");
}

 
mcmsg_am_enq_7i_holder()
{
		asm("_am_enq_7i::");
		asm(" POST_PAM_ENQ = 43");		
		asm("   orh     ha%_pam_intf_vec9, r0, r30");
		asm("   ld.l    l%_pam_intf_vec9(r30), r30");
		asm("   bri     r30");
		asm("    or     POST_PAM_ENQ, r0, r31");
}



mcmsg_am_store_async_internal_holder()
{
		asm("_am_store_async_internal::");
		asm(" POST_PAM_STORE_ASYNC = 45");		
		asm("   orh     ha%_pam_intf_vec9, r0, r30");
		asm("   ld.l    l%_pam_intf_vec9(r30), r30");
		asm("   bri     r30");
		asm("    or     POST_PAM_STORE_ASYNC, r0, r31");
}


 
mcmsg_am_get_internal_holder()
{
		asm("_am_get_internal::");
		asm(" POST_PAM_GET = 46");		
		asm("   orh     ha%_pam_intf_vec7, r0, r30");
		asm("   ld.l    l%_pam_intf_vec7(r30), r30");
		asm("   bri     r30");
		asm("    or     POST_PAM_GET, r0, r31");
}

 
mcmsg_am_reply_get_holder()
{
		asm("_am_reply_get::");
		asm(" POST_PAM_REPLY_GET = 47");		
		asm("   orh     ha%_pam_intf_vec7, r0, r30");
		asm("   ld.l    l%_pam_intf_vec7(r30), r30");
		asm("   bri     r30");
		asm("    or     POST_PAM_REPLY_GET, r0, r31");
}


/* for testing or benchmarking */
mcmsg_pam_test_holder()
{
		asm("_pam_test::");
		asm(" POST_PAM_TEST = 48");		
		asm("   orh     ha%_pam_intf_vec9, r0, r30");
		asm("   ld.l    l%_pam_intf_vec9(r30), r30");
		asm("   bri     r30");
		asm("    or     POST_PAM_TEST, r0, r31");
}

/* for testing or benchmarking */
mcmsg_pam_test2_holder()
{
		asm("_pam_test2::");
		asm(" POST_PAM_TEST2 = 49");		
		asm("   orh     ha%_pam_intf_vec9, r0, r30");
		asm("   ld.l    l%_pam_intf_vec9(r30), r30");
		asm("   bri     r30");
		asm("    or     POST_PAM_TEST2, r0, r31");
}

