/*									tab:2
 *
 * am_glib.c - Active Message/GLUnix interface routines
 *
 * "Copyright (c) 1994 by Lok Tin Liu and The Regents of the University 
 * of California.  All rights reserved."
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 			Lok Tin Liu
 * Version:			52
 * Creation Date:	Fri Sep 16 10:59:39 1994
 * Filename:		am_glib.c
 * History:
 */

#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <stdio.h>
#include <string.h>

#include "am_glib.h"

#define MAX_HOSTNAME_LEN 256
#define USE_ATM 0

volatile static unsigned long fast_inaddr;

#if USE_ATM
/* Get the fast address of the caller. Return 0 if successful; -1 if error.
 * This is a hack that works only for the dawn machines. I assume the
 * ATM connection is named as dawnXX-sa.
 */

int Am_GetMyFastAddr(Am_Fast_Addr fast_addr)
{
	char my_hostname[MAX_HOSTNAME_LEN];
	char fast_hostname[MAX_HOSTNAME_LEN];
	char *cp;
	struct hostent *hp;


	if (gethostname(my_hostname, MAX_HOSTNAME_LEN) != 0) {
		perror("Am_GetMyFastAddr: cannot get hostname");
    return -1;
	}

/*	printf("my hostname is %s\n", my_hostname); */

	/* Find the first occurence of '.' to get the machine name without the
	 * domain name
	 */
	if (cp = strchr(my_hostname, '.')) {
    /* The hostname contains the domain name. So, truncate it. */
		*cp = '\0';
	}
/*	printf("my hostname is %s\n", my_hostname);	 */

  /* Assume that the ATM connection is dawnXX-sa */
	strcpy(fast_hostname, my_hostname);
	strcat(fast_hostname, "-sa");
/*	printf("my fast hostname is %s\n", fast_hostname); */

	hp = gethostbyname(fast_hostname);
	if ((hp == 0) || (hp->h_addr == 0)) {			
/*		printf("Cannot get IP address of local host on ATM %s\n", fast_hostname);*/
		/* try the TCP/IP address */
		hp = gethostbyname(my_hostname);
		if ((hp == 0) || (hp->h_addr == 0)) {			
			printf("Am_GetMyFastAddr: cannot get IP address of local host on Ethernet %s\n",
						 my_hostname);
			return -1;
	  }		
	}
	fast_inaddr = *((unsigned int *)hp->h_addr);
/*  printf("fast addr is %x\n", fast_inaddr); */
	memcpy(fast_addr, &fast_inaddr, sizeof(Am_Fast_Addr));
  return 0;
}
#else

int Am_GetMyFastAddr(Am_Fast_Addr fast_addr)
{
	char my_hostname[MAX_HOSTNAME_LEN];
	char fast_hostname[MAX_HOSTNAME_LEN];
	char *cp;
	struct hostent *hp;


	if (gethostname(my_hostname, MAX_HOSTNAME_LEN) != 0) {
		perror("Am_GetMyFastAddr: cannot get hostname");
    return -1;
	}
/*	printf("my hostname is %s\n", my_hostname); */

	/* Find the first occurence of '.' to get the machine name without the
	 * domain name
	 */
	if (cp = strchr(my_hostname, '.')) {
    /* The hostname contains the domain name. So, truncate it. */
		*cp = '\0';
	}
/*	printf("my hostname is %s\n", my_hostname);	 */



	/* try the TCP/IP address */
	hp = gethostbyname(my_hostname);
	if ((hp == 0) || (hp->h_addr == 0)) {			
		printf("Am_GetMyFastAddr: cannot get IP address of local host on Ethernet %s\n",
					 my_hostname);
		return -1;
  }		

	fast_inaddr = *((unsigned int *)hp->h_addr);
	memcpy(fast_addr, (void *)&fast_inaddr, sizeof(Am_Fast_Addr));
/*  printf("Am_GetMyFastAddr: fast addr is %x\n", *(unsigned *)fast_addr);	 */
	fflush(stdout);
  return 0;
}
#endif

/* Return the size of a fast address */
int Am_GetFastAddrSize()
{
  return sizeof(Am_Fast_Addr);
}



char tcpam_fast_hostname[MAX_HOSTNAME_LEN];

/* Return a pointer to a string containing a meaning name of the fast address.
 * Note that the string is stored in static storage.
 */
char *Am_GetFastAddrStr(Am_Fast_Addr fast_addr)
{

	struct hostent *hp;
	
	hp = gethostbyaddr((char *)fast_addr, sizeof(struct in_addr), AF_INET);
	if ((hp == 0) || (hp->h_name == 0)) {
		printf("Am_getFastAddrStr: cannot get hostname for IP address %s\n",
					 inet_ntoa(*(struct in_addr *)fast_addr));
		tcpam_fast_hostname[0] = '\0';
  }
	else
		strcpy(tcpam_fast_hostname, hp->h_name);
	return tcpam_fast_hostname;
}





