/*									tab:2
 *
 * tcpam_barrier.c - TCP Active Messages barrier
 *
 * "Copyright (c) 1994 by Lok Tin Liu and The Regents of the University 
 * of California.  All rights reserved."
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 			Lok Tin Liu
 * Version:			28
 * Creation Date:	Fri Jul  8 15:22:47 1994
 * Filename:		tcpam_barrier.c
 * History:
 *	LTL	5	Sun Jul 10 15:07:17 1994
 *		Modified from paragon/nxam_barrier.c for TCPAM
 */


#include "tcpam_int.h"

#define MYPROC AM_myproc
#define PROCS  AM_procs

static int up_buf[2];
static int down_buf;

int am_barrier_init()
{
	up_buf[0] = up_buf[1] = down_buf = 0;
}

void am_incr_handler(vnn_t node, void *val)
{
	(*((int*)val))++;
}

void am_barrier_up()
{
    int parent = (MYPROC - 1) / 2;
    int odd_child = 2 * MYPROC + 1;
    int even_child = 2 * MYPROC + 2;
    int parity = MYPROC & 1;
    /*int *buf = reduce_buf;*/
    /*reset_store_sync();*/
    /*barrier();*/
    if (MYPROC == 0) {
		if (PROCS != 1) 
			if (PROCS == 2) {
				while (up_buf[1] == 0)
					am_poll();
				/*store_sync(sizeof(int));*/
				/*return (ADD(val, *((int *) &buf[1])));*/
			} else {
				while (up_buf[0] == 0 || up_buf[1] == 0)
					am_poll();
				/*store_sync(2 * sizeof(int));*/
				/*return (ADD(ADD(val, *((int *) &buf[1])), *((int *) buf)));*/
			}
    } else {
		/*___globPtrType dest = toglobal(parent, (void *) &reduce_buf[parity]);*/
		if (odd_child >= PROCS) {
			/*__i_store(dest, val);*/
			am_request_4(parent, am_incr_handler, (int)&up_buf[parity], 0, 0, 0);
		} else if (even_child >= PROCS) {
			while (up_buf[1] == 0)
				am_poll();
			/*store_sync(sizeof(int));*/
			/*__i_store(dest, ADD(val, *((int *) &buf[1])));*/
			am_request_4(parent, am_incr_handler, (int)&up_buf[parity], 0, 0, 0);
		} else {
			while (up_buf[0] == 0 || up_buf[1] == 0)
				am_poll();
			/*store_sync(2 * sizeof(int));*/
			/*__i_store(dest, ADD(ADD(val, *((int *) &buf[1])), *((int *) buf)));*/
			am_request_4(parent, am_incr_handler, (int)&up_buf[parity], 0, 0, 0);
		}
    }
	up_buf[0] = up_buf[1] = 0;
}

void am_barrier_down()
{
    int left = 2 * MYPROC + 1;
    int right = 2 * MYPROC + 2;
    /*reset_store_sync();*/
    /*barrier();*/
    if (MYPROC != 0) {
		while (down_buf == 0)
			am_poll();
		/*store_sync(sizeof(int));*/
		/*val = *((int *) &down_buf);*/
    }
    if (left < PROCS)
		am_request_4(left, am_incr_handler, (int)&down_buf, 0, 0, 0);
		/*__i_store(toglobal(left, (int *) &down_buf), val);*/
    if (right < PROCS)
		am_request_4(right, am_incr_handler, (int)&down_buf, 0, 0, 0);
	    /*__i_store(toglobal(right, (int *) &down_buf), val);*/
	down_buf = 0;
	/*barrier();*/
    /*return (val);*/
} 

void am_barrier()
{
	am_barrier_up();
	am_barrier_down();
}

