/*									tab:2
 *
 * tcpam_int.h - Internal header file for TCP/IP Active Messages
 *
 * "Copyright (c) 1994 by Lok Tin Liu and The Regents of the University 
 * of California.  All rights reserved."
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 			Lok Tin Liu
 * Version:			33
 * Creation Date:	Fri Jul  8 13:33:23 1994
 * Filename:		tcpam_int.h
 * History:
 */



#ifndef __TCPAM_INT_H_
#define __TCPAM_INT_H_

typedef unsigned int vnn_t;
typedef void (*handler_t)();

extern int AM_myproc, AM_procs;

/* Must be called before any other am calls */
void am_enable();

/* Returns virtual node number */
int am_my_proc();

/* Returns the number of nodes */
int am_procs();

/* Must be called for a graceful shutdown */
void am_disable();

/* Not in spec. Must be called to close all sockets nefore exit */
void am_exit(int code);

/* Receive messages by polling */
int am_poll();

/* Returns the AM layer's errno */
int am_errno();

/* Send request message with up to 4 integer arguments */
int am_request_4(vnn_t dest, handler_t fun, int arg1, int arg2, int arg3,
								 int arg4);

/* Send request message with 2 integer arguments and one double argument */
int am_request_fd(vnn_t dest, handler_t fun, double arg1, int arg2, int arg3);

/* Send reply message with up to 4 integer arguments */
int am_reply_4(vnn_t dest, handler_t fun, int arg1, int arg2, int arg3,
							 int arg4);

/* Send reply message with 2 integer arguments and one double argument */
int am_reply_fd(vnn_t dest, handler_t fun, double arg1, int arg2, int arg3);

/* Copy nbytes bytes of data from address lva on calling node to address
 * rva on dest node. Invoke handler with handler_arg on dest node when
 * all data are received on dest node.
 */
int am_store(vnn_t dest, void *lva, void *rva, int nbytes,
						 handler_t request_handler, void *handler_arg);

/* Retrive nbytes bytes of data from address rva on dest node to address
 * lva on calling node. Invoke handler with handler_arg on calling node when
 * all data are received on calling node.
 */
int am_get(vnn_t dest, void *sva, void *dva, int nbytes,
					 handler_t reply_handler, void *handler_arg);

/* Return maximum buffer size in bytes for put and get */
int am_max_size();

#endif

