/*									tab:4
 *
 * X_handlers_cp.c - Split-C host functions for X
 *
 * "Copyright (c) 1993 by Steve Lumetta and The Regents of the University 
 * of California.  All rights reserved."
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 			Lok Tin Liu
 * Version:			3
 * Creation Date:	Sun Nov  7 16:59:46 1993
 * Filename:		X_handlers_cp.c
 * History:
 *	SL	3	Mon Jan 17 16:30:07 1994
 *		Fixed bug in declaration of CMPE_printGetChannel
 *	LTL	2	Thu Nov 25 15:41:14 1993
 *		Fixed bug in setting host handlers
 *	LTL	1	Sun Nov  7 17:00:02 1993
 *		Initial release. X stuff is not in libsplit-c anymore.
 */



#include <stdio.h>
#include "host.h"
#include "print.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>


static Display *X_Display = NULL;
static Window X_Window;
static GC X_GC;
static int X_Size;
static XImage *X_Image;
static char *X_Data;
static unsigned long X_Pixels[256];

extern Handler CMPE_printGetChannel;

void imageXDraw() 
{
	if (X_Image != NULL) {
		XPutImage(X_Display, X_Window, X_GC, X_Image, 0, 0, 0, 0,
				  X_Size, X_Size);
		listenX();
	}
}

listenX()
{
	XEvent event;

    /* XPending() does flush */
    while (XPending(X_Display) > 0) {
		XNextEvent(X_Display, &event);
		if (event.type == Expose) {
			/* purge other exposes */
			while (XCheckTypedEvent(X_Display, Expose, &event));
			imageXDraw();
        }
    }
}

#ifndef CMMD

void X_INIT_handler(Handler *f, int proc, int two, int three, int four)
{
	/* two - size of display window */
	
    int rgbOfs, rgbRange, i;
    Screen scr;
    int scrNum;
    XSizeHints size_hints;
	XColor SetColors[256];
	Colormap colorMap;
	Visual *vis;

	unsigned long masks[256]; /* These are ignored */

    X_Display = XOpenDisplay(NULL);
    if (X_Display == NULL) {
		fprintf(stderr, "Could not open display.\n");
		exit(1);
    }
    scrNum = DefaultScreen(X_Display);
    colorMap = DefaultColormap(X_Display, scrNum);
    X_GC = DefaultGC(X_Display, scrNum);
	
	X_Data = (char *)malloc(two*two);
    if (XAllocColorCells(X_Display, colorMap, True, masks, 0, X_Pixels, 4)) {
		for (i = 0; i < 4; i++) {
			SetColors[i].pixel = X_Pixels[i];
			SetColors[i].flags = DoRed|DoGreen|DoBlue;
		}
		SetColors[0].red = 0;         /* Blue, green, red, and white */
		SetColors[0].green = 0;
		SetColors[0].blue = 60000;
		SetColors[1].red = 0;
		SetColors[1].green = 60000;
		SetColors[1].blue = 0;
		SetColors[2].red = 60000;
		SetColors[2].green = 0;
		SetColors[2].blue = 0;
		SetColors[3].red = 60000;
		SetColors[3].green = 60000;
		SetColors[3].blue = 60000;
		XStoreColors(X_Display, colorMap, SetColors, 4);
		X_Image = XCreateImage(X_Display, DefaultVisual(X_Display, scrNum),
							   8, ZPixmap, 0, X_Data, two, two, 8, 0);
	}
	else {
		printf("You must use a color or grey-scale monitor.\n");
		exit(1);
/*		X_Pixels[0] = BlackPixel(X_Display, scrNum);
		X_Pixels[1] = X_Pixels[2] = X_Pixels[3] =
			WhitePixel(X_Display, scrNum);
		X_Image = XCreateImage(X_Display, DefaultVisual(X_Display, scrNum),
							   1, ZPixmap, 0, X_Data, two, two, 8, 0);*/

	}
	for (i = 0; i < two*two; i++)
		X_Data[i] = X_Pixels[0];
	X_Window = XCreateSimpleWindow(X_Display,
								   RootWindow(X_Display, scrNum),
								   100, 100, two, two, 5, 1, 0);
	if (X_Window == (Window)NULL) {
		fprintf(stderr, "Could not open window.\n");
		exit(1);
	}
	size_hints.flags = PPosition|PSize|PMinSize;
	size_hints.x = 100;
	size_hints.y = 100;
	size_hints.width = two;
	size_hints.height = two;
	size_hints.min_width = two;
    size_hints.min_height = two;
    XSetStandardProperties(X_Display, X_Window, "Split-C Display",
						   "Split-C Display", None, NULL, 0, &size_hints);
    XSelectInput(X_Display, X_Window, ExposureMask);
    XMapWindow(X_Display, X_Window);
    X_Size = two;
    listenX();
}

int doXBlock(info, base)
remoteFunction *info;
char *base;
{
	int x1, y1, x2, y2, len, i, j;
	char *p = base, *fill;
	int cnt, val, bit[4];
	
    i = (int)info->stream;
	x1 = (i>>16);
	y1 = (i&0xFFFF);
    i = (int)info->adr2;
	x2 = (i>>16);
	y2 = (i&0xFFFF);
	len = ((x2 - x1)*(y2 - y1) + 3)/4;
	
	for (i = y1, cnt = 4; i < y2; i++) {
		fill = X_Data + i*X_Size + x1;
		for (j = x1; j < x2; j++) {
			if (cnt == 4) {
				val = *p++;
				while (cnt-- > 0) {
					bit[cnt] = (val&3);
				    val >>= 2;
				}
				cnt = 0;
			}
			*fill++ = X_Pixels[bit[cnt++]];
		}
	}	

	am_reply2(info->pid, info->function, -3, info->adr);
	free(base);
	free(info);
	return 0;					/* deallocate this channel */
}


VOID X_POINT_handler(Handler *f, int proc, int two, int three, int four)
{
	/* two - X position of point */
	/* three - Y position of point */
	/* four - color of point */

	X_Data[three*X_Size + two] = X_Pixels[four];
}

VOID X_BLOCK_handler(Handler *f, int proc, int two, int three, int four)
{
	/* two - place to put channel on node */
	/* three - upper left corner of block */
	/* four - just outside lower right corner of block */

	Channel c;
	remoteFunction *rf;
	int x1, y1, x2, y2, len;
	char *p;
	
	x1 = (three>>16);
	y1 = (three&0xFFFF);
	x2 = (four>>16);
	y2 = (four&0xFFFF);
	len = ((x2 - x1)*(y2 - y1) + 3)/4;
	p = (char *)malloc(len);

	rf = (remoteFunction *)malloc(sizeof(remoteFunction));
	rf->pid = proc;
	rf->function = (Handler)CMPE_printGetChannel;
	rf->stream = (FILE *)three;
	rf->adr2 = (void *)four;
	rf->adr = (void *)two;
		
	c = am_open_segment(p, len, doXBlock, (void *)rf);

	am_reply2(proc, CMPE_printGetChannel, (int )c, two);
}
		
VOID X_SHOW_handler(Handler *f, int proc, int two, int three, int four)
{
	imageXDraw();
}

#else /* CMMD */

/* These functions are essential the same as the CMAM version, except
 * the handler pointer argument is not passed because indirect active
 * message is not used
 */

void doXBlock(int rport_num, remoteFunction *info)

{
	int x1, y1, x2, y2, len, i, j;
	char *base, *p, *fill;
	int cnt, val, bit[4];

	p = base = am_get_rport_addr(rport_num);
	
    i = (int)info->stream;
	x1 = (i>>16);
	y1 = (i&0xFFFF);
    i = (int)info->adr2;
	x2 = (i>>16);
	y2 = (i&0xFFFF);
	len = ((x2 - x1)*(y2 - y1) + 3)/4;
	
	for (i = y1, cnt = 4; i < y2; i++) {
		fill = X_Data + i*X_Size + x1;
		for (j = x1; j < x2; j++) {
			if (cnt == 4) {
				val = *p++;
				while (cnt-- > 0) {
					bit[cnt] = (val&3);
				    val >>= 2;
				}
				cnt = 0;
			}
			*fill++ = X_Pixels[bit[cnt++]];
		}
	}	

	am_reply2(info->pid, info->function, -3, info->adr);
	free(base);
	free(info);
        am_free_rport(rport_num);
}



VOID X_INIT_handler(int proc, int two, int three, int four)
{
	/* two - size of display window */
	
    int rgbOfs, rgbRange, i;
    Screen scr;
    int scrNum;
    XSizeHints size_hints;
	XColor SetColors[256];
	Colormap colorMap;
	Visual *vis;

	unsigned long masks[256]; /* These are ignored */

    X_Display = XOpenDisplay(NULL);
    if (X_Display == NULL) {
		fprintf(stderr, "Could not open display.\n");
		exit(1);
    }
    scrNum = DefaultScreen(X_Display);
    colorMap = DefaultColormap(X_Display, scrNum);
    X_GC = DefaultGC(X_Display, scrNum);
	
	X_Data = (char *)malloc(two*two);
    if (XAllocColorCells(X_Display, colorMap, True, masks, 0, X_Pixels, 4)) {
		for (i = 0; i < 4; i++) {
			SetColors[i].pixel = X_Pixels[i];
			SetColors[i].flags = DoRed|DoGreen|DoBlue;
		}
		SetColors[0].red = 0;         /* Blue, green, red, and white */
		SetColors[0].green = 0;
		SetColors[0].blue = 60000;
		SetColors[1].red = 0;
		SetColors[1].green = 60000;
		SetColors[1].blue = 0;
		SetColors[2].red = 60000;
		SetColors[2].green = 0;
		SetColors[2].blue = 0;
		SetColors[3].red = 60000;
		SetColors[3].green = 60000;
		SetColors[3].blue = 60000;
		XStoreColors(X_Display, colorMap, SetColors, 4);
		X_Image = XCreateImage(X_Display, DefaultVisual(X_Display, scrNum),
							   8, ZPixmap, 0, X_Data, two, two, 8, 0);
	}
	else {
		printf("You must use a color or grey-scale monitor.\n");
		exit(1);
/*		X_Pixels[0] = BlackPixel(X_Display, scrNum);
		X_Pixels[1] = X_Pixels[2] = X_Pixels[3] =
			WhitePixel(X_Display, scrNum);
		X_Image = XCreateImage(X_Display, DefaultVisual(X_Display, scrNum),
							   1, ZPixmap, 0, X_Data, two, two, 8, 0);*/

	}
	for (i = 0; i < two*two; i++)
		X_Data[i] = X_Pixels[0];
	X_Window = XCreateSimpleWindow(X_Display,
								   RootWindow(X_Display, scrNum),
								   100, 100, two, two, 5, 1, 0);
	if (X_Window == (Window)NULL) {
		fprintf(stderr, "Could not open window.\n");
		exit(1);
	}
	size_hints.flags = PPosition|PSize|PMinSize;
	size_hints.x = 100;
	size_hints.y = 100;
	size_hints.width = two;
	size_hints.height = two;
	size_hints.min_width = two;
    size_hints.min_height = two;
    XSetStandardProperties(X_Display, X_Window, "Split-C Display",
						   "Split-C Display", None, NULL, 0, &size_hints);
    XSelectInput(X_Display, X_Window, ExposureMask);
    XMapWindow(X_Display, X_Window);
    X_Size = two;
    listenX();
}

VOID X_POINT_handler(int proc, int two, int three, int four)
{
	/* two - X position of point */
	/* three - Y position of point */
	/* four - color of point */

	X_Data[three*X_Size + two] = X_Pixels[four];
}

VOID X_BLOCK_handler(int proc, int two, int three, int four)
{
	/* two - place to put channel on node */
	/* three - upper left corner of block */
	/* four - just outside lower right corner of block */

	Channel c;
	remoteFunction *rf;
	int x1, y1, x2, y2, len;
	char *p;
	
	x1 = (three>>16);
	y1 = (three&0xFFFF);
	x2 = (four>>16);
	y2 = (four&0xFFFF);
	len = ((x2 - x1)*(y2 - y1) + 3)/4;
	p = (char *)malloc(len);

	rf = (remoteFunction *)malloc(sizeof(remoteFunction));
	rf->pid = proc;
	rf->function = (Handler)CMPE_printGetChannel;
	rf->stream = (FILE *)three;
	rf->adr2 = (void *)four;
	rf->adr = (void *)two;
		
	c = am_open_segment(p, len, doXBlock, (void *)rf);

	am_reply2(proc, CMPE_printGetChannel, (int )c, two);
}
		
VOID X_SHOW_handler(int proc, int two, int three, int four)
{
	imageXDraw();
}

#endif /* CMMD */

static Handler XHostFunctions[] =
{	
	X_INIT_handler,
	X_POINT_handler,
	X_BLOCK_handler,
	X_SHOW_handler
};

extern int hostFunctionCount;
extern Handler *hostFunctions;
extern Handler defaultHostFunctions[];

/* Add the 4 X functions to the hostFunctions table.
 * Note that 4 slots are reserved for
 * the X handlers in defaultHostFunctions already
 */

void splitc_addXHandlers(Handler *newXHandlers)
{
	int base = hostFunctionCount;
	
	if (!hostFunctions)
	{
		hostFunctions = malloc(sizeof(Handler)*(hostFunctionCount));
		memcpy(hostFunctions,
			   defaultHostFunctions,
			   hostFunctionCount*sizeof(Handler));
	}

    memcpy(hostFunctions+DEFAULT_HOST_HANDLERS_NUM,
		   newXHandlers,
		   4 * sizeof(Handler));
 
}


void splitc_host_X_init()
{
  splitc_addXHandlers(XHostFunctions);   
}
