/*                              					tab:4
 *
 * am_cmmd.c - Miscellaneous functions to fix up active messages in CMMD
 *
 * "Copyright (c) 1993 by Lok T. Liu and The Regents of the University 
 * of California.  All rights reserved."
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 			Lok Tin Liu
 * Version:			10
 * Creation Date:	Mon Aug 16 14:50:49 1993
 * Filename:		am_cmmd.c
 * History:
 *	LTL	9	Wed Mar 16 11:19:30 1994
 *		Fixed bug in cmaml_open_this_segment 
 *	LTL	4	Mon Nov 29 19:54:16 1993
 *		Fixed alignment bug for cmaml_xfer
 *	LTL	3	Sun Aug 29 10:35:21 1993
 *		Added cmaml_open_this_segment
 *	LTL	1	Mon Aug 16 14:52:17 1993
 *		New code for version 112 of CMMD libsplitc. Fixed up deficiencies
 *		of CMAML.
 */



#ifdef CMMD

#include <cm/cmmd.h>
#include <cm/cmna.h>
#include <cmsys/ni_reg_offset.h>
#include <cmsys/ni_constants.h>


typedef void (*Handler)();

#ifdef pe_obj



/* Our own version of barrier
 * Cannot use CMMD_sync_with_nodes because it doesn't poll for 
 *   active messges
 */
int cmaml_barrier(int info_bit)
{  
    int gtemp; 

    CMNA_com_send_first(OR_SCAN, SCAN_REDUCE, 1, info_bit);
    while(!RECEIVE_OK(CMNA_com_status())) CMAML_poll(); 
    gtemp = CMNA_com_receive_word(); 
    return(gtemp);
}



/* Our own version of wait
 * Cannot use CMAML_wait because it doesn't subtract value from *flag
 *   when *flag >= value before polling the network
 */
void cmaml_wait(volatile int *flag, int value)
{
  while (*flag < value)
    CMAML_poll();
  *flag -= value;
}

#endif /* pe_obj */


/* use CMMD receive ports as if they were CMAM segments */
/* you pass in a rport number, we return a segment id */
int cmaml_open_this_segment(int rport_num, void *base_addr, int len, 
                              Handler func, void *data) 
{
  int sid = -1;		
  if (CMAML_allocate_this_rport(rport_num) != -1) {
    CMAML_set_rport_addr(rport_num, base_addr);
    CMAML_set_rport_byte_counter(rport_num, len);
    CMAML_set_rport_total_nbytes(rport_num, len);
    CMAML_set_rport_handler(rport_num, func);
    CMAML_set_rport_handler_arg2(rport_num, data);
    sid = (rport_num << CMAML_SERIAL_OFFSET_BITS) + ((int)base_addr & 7);
  }
  return (sid);
}


/* use CMMD receive ports as if they were CMAM segments */
/* return a segment id */
int cmaml_open_segment(void *base_addr, int len, 
                              Handler func, void *data) 
{
  int sid = -1;
  int rport_num;

  rport_num = CMAML_allocate_rport();
  if (rport_num != -1) {
    CMAML_set_rport_addr(rport_num, base_addr);
    CMAML_set_rport_byte_counter(rport_num, len);
    CMAML_set_rport_total_nbytes(rport_num, len);
    CMAML_set_rport_handler(rport_num, func);
    CMAML_set_rport_handler_arg2(rport_num, data);
    /* encode the base address alignment in sid */ 
    sid = (rport_num << CMAML_SERIAL_OFFSET_BITS) + ((int)base_addr & 7);
  }

  return (sid);
}

int cmaml_xfer(int dest, int sid, void *buf, int len)
{
  int rport_num = sid >> CMAML_SERIAL_OFFSET_BITS;  
  int align = sid & 7;
  int offset = (sid & 0x01ffffff) - align;  /* get the 25 offset bits */
  /* scopy handler is not used */
  if (len <= 0)
    return (-1);
  else
    return (CMAML_scopy(dest, rport_num, align, offset, buf, len, 0, 0));
}

#endif /* CMMD */





