/*									tab:8
 *
 * bulk.c - Handlers for bulk functions.
 *
 * 
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 		Steve Luna
 * Version:		1
 * Creation Date:	Sat Nov 27 23:07:14 1993
 * Filename:		bulk.c
 * History:
 */

#include <split-c/split-c.h>
#include <assert.h>

volatile int __seg_ctr = 0;
volatile int __seg_num;

static void setSeg(int segId)
{
	__seg_num = segId;
	__seg_ctr++;
}

typedef struct {
	Processor	pid;
	Counter		*ctr;
} __PutCtrInfo;

#ifndef CMMD

int __GetComplete(void *info, void *base)
{
	__decr_ctr(info);
	return 0;
}

#else

void __GetComplete(int rport_num, void *info)
{
	__decr_ctr(info);
	am_free_rport(rport_num);
}

#endif

void __PutCompleteAtSource(void)
{
	numPutOuts--;
}

#ifndef CMMD

int __PutCompleteAtDest(void *proc, void *base)
{
	am_reply0((int)proc, __PutCompleteAtSource);
	return 0;
}

#else

void __PutCompleteAtDest(int rport_num, void *proc)
{
	am_reply0((int)proc, __PutCompleteAtSource);
	am_free_rport(rport_num);
}

#endif

void __PutAllocateSegment(int proc, void *dest, int len)
{
	int segId = am_open_segment(dest, len, __PutCompleteAtDest,(void*)proc);
	am_reply1(proc, setSeg, segId);

}

void __PutCtrCompleteAtSource(Counter *ctr)
{
	(*ctr)--;
}

#ifndef CMMD

int __PutCtrCompleteAtDest(void *info, void *base)
{
	__PutCtrInfo *i = info;
	am_reply1(i->pid, __PutCtrCompleteAtSource, i->ctr);
	am_free(i);
	return 0;
}

#else

void __PutCtrCompleteAtDest(int rport_num, void *info)
{
	__PutCtrInfo *i = info;
	am_reply1(i->pid, __PutCtrCompleteAtSource, i->ctr);
	am_free(i);
	am_free_rport(rport_num);
}

#endif /*CMMD */

void __PutCtrAllocateSegment(int proc, void *dest, int len,
									Counter *ctr)
{
	__PutCtrInfo *info;
	int segId;
	
	info = (__PutCtrInfo *)am_malloc(sizeof(__PutCtrInfo));
	info->pid = proc;
	info->ctr = ctr;
	segId = am_open_segment(dest, len, __PutCtrCompleteAtDest, info);
	am_reply1(proc, setSeg, segId);
}

/* ............................
 *
 * STOREs
 *
 * ............................
 */

#ifndef CMMD

static int StoreCompleteAtDest(void *len, void *base)
{
	recvBytes += (int)len;
	return 0;
}

#else

static void StoreCompleteAtDest(int rport_num, void *len)
{
	recvBytes += (int)len;
        am_free_rport(rport_num);
}
#endif

static void StoreAllocateSegment(int proc, void *dest, int len)
{
	int segId = am_open_segment(dest, len,
		StoreCompleteAtDest, (void *)len);
	am_reply1(proc, setSeg, segId);
}

/* Implicit ctr is incremented */
static void __3i_store_handler(int *addr, int data1, int data2, int data3) {
	addr[0] = data1;
	addr[1] = data2;
	addr[2] = data3;
	recvBytes += 12;
}

/* STORE bulk */
void __store(___V_GPTR gPtr, void *lPtr, int len)
{
	Processor pid = toproc(gPtr);
	void *addr    = tolocal(gPtr);

	if(pid == MYPROC) {
		sentBytes += len;
		bcopy(lPtr, addr, len);
		recvBytes += len;
	} else if(len <= STORE_STRUCT_LEN && (len&0x3) == 0) {
		int numWords = len>>2;
		int *iPtr = lPtr;

		sentBytes += len;
		/* send 3-word packets */
		while(numWords >= 3) {
			am_request4(pid, __3i_store_handler, addr, iPtr[0],
			       iPtr[1], iPtr[2]);
			iPtr += 3;
			addr = (void *)((int)addr+12);
			numWords -= 3;
		}
		/* send odd stuff at end */
		if(numWords) {
			if(numWords == 1)
				am_request3(pid, __i_store_handler, addr,
				       iPtr[0], &recvBytes);
			else /* numWords == 2 */
				am_request4(pid, __ll_store_handler, addr,
				       iPtr[0], iPtr[1], &recvBytes);
		}
	} else {
		__seg_ctr = 0;
		am_request3(pid, StoreAllocateSegment, MYPROC, addr, len);
		am_wait(&__seg_ctr, 1);
		sentBytes += len;
		am_xferBuffer(pid, __seg_num, lPtr, len);
	}
}

/* ........................................
 *
 * BULK STORE without alignment restriction
 *
 * ........................................
 */

/* Clean this - too many paranthesis */
#define GET_BYTES(s,d,a) { \
			   switch (a) {	\
			     case 0: \
			       *(int *)(d) = *(int *)(s); break; \
			     default: \
			       *(int *)(d) = \
				   ((*(int *)((char *)(s) - (a))) << (8*(a))) + \
				   ((*(int *)((char *)(s)+4-(a))) >> (8*(4-(a)))); \
			       break; \
			   } \
		         }
			       
				      

void ___begin_chars_store_handler (int *addr, int data, int numChars) {
    int existing = *addr;
    switch (numChars) {
      case 1:	existing = (existing & 0xFFFFFF00) | (data & 0xFF); break;
      case 2:	existing = (existing & 0xFFFF0000) | (data & 0xFFFF); break;
	
      case 3: 	existing = (existing & 0xFF000000) | (data & 0xFFFFFF); break;
      case 4:	existing = data; break;
    }
    *addr = existing;
    recvBytes += numChars;    
}
void ___end_chars_store_handler (int *addr, int data, int numChars) {
    int existing = *addr;
    switch (numChars) {
      case 1:	existing = (existing & 0xFFFFFF) | (data & 0xFF000000); break;
      case 2:	existing = (existing & 0xFFFF) | (data & 0xFFFF0000); break;
      case 3:	existing = (existing & 0xFF) | (data & 0xFFFFFF00); break;
      case 4:	existing = data; break;
    }
    *addr = existing;
    recvBytes += numChars;
}

void
bulk_store(___V_GPTR gPtr, void *lPtr, int len)
{
	Processor pid = toproc(gPtr);
	void *addr    = tolocal(gPtr);
	static void __bulk_store_short(int node,void *dest,void *src,int len);

	if(len == 0) return;

	/* Local copy */
	sentBytes += len;
	if(pid == MYPROC) {
		bcopy(lPtr, addr, len);
		recvBytes += len;
	} else if(len <= STORE_STRUCT_LEN) {
		__bulk_store_short(pid, addr, lPtr, len);
	} else {
		/* open a segment at destination */
		__seg_ctr = 0;
		am_request3(pid, StoreAllocateSegment, MYPROC, addr, len);
		am_wait(&__seg_ctr, 1);
		/* transfer the data */
		am_xferBuffer(pid, __seg_num, lPtr, len);
	}
}

/*
 * Store up to eight bytes of any alignment
 */

static void
__bulk_store_Nb_handler(char *dest, int len, int w1, int w2)
{
	int l = len>8 ? 8 : len;
	switch(l) {
	case 8:	dest[7] = w2;
	case 7:	dest[6] = w2>>8;
	case 6:	dest[5] = w2>>16;
	case 5:	dest[4] = w2>>24;
	case 4:	dest[3] = w1;
	case 3: dest[2] = w1>>8;
	case 2:	dest[1] = w1>>16;
	case 1: dest[0] = w1>>24;
	case 0: ;
	}
	recvBytes += l;
}

static void
__bulk_store_Nb(int node, char *dest, char *src, int len)
{
	int w1=0, w2=0;
	switch(len) {
	default:
	case 8:	w2 |= (0xff&src[7]);
	case 7:	w2 |= (0xff&src[6])<<8;
	case 6:	w2 |= (0xff&src[5])<<16;
	case 5:	w2 |= (0xff&src[4])<<24;
	case 4:	w1 |= (0xff&src[3]);
	case 3: w1 |= (0xff&src[2])<<8;
	case 2:	w1 |= (0xff&src[1])<<16;
	case 1: w1 |= (0xff&src[0])<<24;
	case 0: ;
	}
	am_request4(node, __bulk_store_Nb_handler, dest, len, w1, w2);
}

/*
 * Store aligned words
 */

static void
__bulk_store_3i_handler(int *dest, int w1, int w2, int w3)
{
	dest[0] = w1; dest[1] = w2; dest[2] = w3;
	recvBytes += 12;
}

static void
__bulk_store_i(int node, int *dest, int *src, int len)
{
	/* full packets */
	while(len >= 12) {
		am_request4(node, __bulk_store_3i_handler, dest,
			src[0], src[1], src[2]);
		src += 3; len -= 12; dest += 3;
	}
	/* remainder */
	while(len > 0) {
		am_request4(node, __bulk_store_Nb_handler, dest, len,
			src[0], src[1]);
		src += 2; len -= 8; dest += 2;
	}
}

/*
 * Bulk store of only a "few" bytes. Use am_request4 to transfer the data.
 */

#define csrc(o) (*(unsigned char *)((int)src+o))
#define ssrc(o) (*(unsigned short *)((int)src+o))
#define isrc(o) (*(unsigned int *)((int)src+o))
#define grab_12 { \
	switch((long)src & 3) { \
	case 0:	w0 = isrc(0); w1 = isrc(4); \
		w2 = isrc(8); \
		break; \
	case 1: w0 = (csrc(0)<<24)+(ssrc(1)<<8)+(isrc(3)>>24); \
		w1 = (isrc(3)<<8)+(isrc(7)>>24); \
		w2 = (isrc(7)<<8)+(isrc(11)>>24); \
		break; \
	case 2: w0 = (ssrc(0)<<16)+(isrc(2)>>16); \
		w1 = (isrc(2)<<16)+(isrc(6)>>16); \
		w2 = (isrc(6)<<16)+(isrc(10)>>16); \
		break; \
	case 3: w0 = (csrc(0)<<24)+(isrc(1)>>8); \
		w1 = (isrc(1)<<24)+(isrc(5)>>8); \
		w2 = (isrc(5)<<24)+(isrc(9)>>8); \
		break; \
	default:w0=0; w1=0; w2=0; /* make gcc -W happy */ \
	} \
}

static void
__bulk_store_short(int node, void *dest, void *src, int len)
{
	int salign = (int)src & 3;
	int dalign = (int)dest & 3;
	int l;

	if((salign|dalign) == 0) {
		/* word-aligned source and word-aligned destination */
		__bulk_store_i(node, dest, src, len);
	} else if(len <= 8) {
		__bulk_store_Nb(node, dest, src, len);
	} else if(salign == dalign) {
		/* same source and destination alignment */
		/* send prefix to reach word-alignment */
		l = 4 - salign;
		__bulk_store_Nb(node, dest, src, l);
		dest = (void *)((int)dest+l);
		src = (void *)((int)src+l);
		len -= l;
		/* send full packets */
		__bulk_store_i(node, dest, src, len);
	} else {
		/* alignments screwed-up */
		/* send prefix to reach word-alignment of destination */
		l = 4 - dalign;
		__bulk_store_Nb(node, dest, src, l);
		dest = (void *)((int)dest + l);
		src = (void *)((int)src + l);
		len -= l;
		/* send full packets, destination being word-aligned */
		while(len >= 12) {
			unsigned int w0, w1, w2;
			grab_12;
			am_request4(node, __bulk_store_3i_handler, dest, w0, w1, w2);
			dest = (void *)((int)dest+12);
			src = (void *)((int)src+12);
			len -= 12;
		}
		/* send packet fraction remaining */
		while(len > 0) {
			__bulk_store_Nb(node, dest, src, len);
			src = (void *)((int)src+8);
			len -= 8;
			dest = (void *)((int)dest+8);
		}
	}
}
