/*									tab:4
 *
 * cm5.h - cm5 header stuff for tam
 *
 * Copyright (c) 1992 by Seth Copen Goldstein and 
 * Regents of the University of California
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  Copyright holder(s) make no
 * representation about the suitability of this software for
 * any purpose. It is provided "as is" without express or
 * implied warranty.
 *
 * Author: 			Seth Copen Goldstein
 * Version:			73
 * Creation Date:	Sun Jan 26 22:28:33 1992
 * Filename:		cm5.h
 * History:
 *	ssl	73	Tue Feb  8 17:51:50 1994
 *		barrier() should be defined even if SPLIT_C_DEBUG is on
 *	ssl	72	Thu Nov 18 18:38:42 1993
 *		Moved include of print.h from split-c.h
 *	ssl	69	Thu Oct 28 20:30:50 1993
 *		Added cm5 specifics (ni stuff) from split-c.h
 *	LTL	68	Wed Aug 25 22:24:34 1993
 *		Added more macros for polling
 *	LTL	66	Mon Aug 16 15:27:01 1993
 *		Renamed macros and added supports for CMMD
 */

#ifndef _CM5_H_
#define _CM5_H_

/* ........
 *
 * I/O hack
 *
 * ........
 */

/* HACK for printing etc. from the nodes
 * this requires the user to include <stdio.h> before split-c.h
 */
#include "print.h"

#ifndef CMMD

/* ...............................
 *
 * Access to the network interface
 *
 * ...............................
 */

/* Include CMAM - CM-5 Active Messages.
 * If we don't want Split-C inlining, we also don't want CMAM inlining
 */

#ifdef	SPLIT_C_DEBUG
#	 define CMAM_DEBUG
#endif

/* barrier needs to be defined even if SPLIT_C_DEBUG is on - ssl */
#define barrier() CMAM_barrier(1) /* Uhhhh, ohh, that simple? -TvE */


#include "cmam.h"
#define splitc_poll() CMAM_poll()
typedef void (*Handler)();

/*********/
/* types */
/*********/

/********************/
/* message routines */
/********************/

#define am_reply0(dest, fun) \
	CMAM_reply_4(dest, fun)
#define am_reply1(dest, fun, one) \
	CMAM_reply_4(dest, fun, one)
#define am_reply2(dest, fun, one, two) \
	CMAM_reply_4(dest, fun, one, two)
#define am_reply3(dest, fun, one, two, three) \
	CMAM_reply_4(dest, fun, one, two, three)
#define am_reply4(node, fun, one, two, three, four) \
	CMAM_reply_4(node, fun, one, two, three, four)
#define am_rpc4(node, fun, one, two, three, four) \
	CMAM_4(node, fun, one, two, three, four)
#define am_rpc3(node, fun, one, two, three) \
	CMAM_4(node, fun, one, two, three)
#define am_rpc2(node, fun, one, two) \
	CMAM_4(node, fun, one, two)
#define am_rpc1(node, fun, one) \
	CMAM_4(node, fun, one)
#define am_rpc0(node, fun) \
	CMAM_4(node, fun)
#define am_request4(node, fun, one, two, three, four) \
	CMAM_4(node, fun, one, two, three, four)
#define am_request3(node, fun, one, two, three) \
	CMAM_4(node, fun, one, two, three)
#define am_request2(node, fun, one, two) \
	CMAM_4(node, fun, one, two)
#define am_request1(node, fun, one) \
	CMAM_4(node, fun, one)
#define am_request0(node, fun) \
	CMAM_4(node, fun)
#define am_poll()	CMAM_poll()
#define am_request_poll()	CMAM_request_poll()
#define am_reply_poll()	CMAM_reply_poll()
#define am_wait(flag, value) CMAM_wait(flag, value)
#define am_poll_wait(flag, value) CMAM_poll_wait(flag, value)
#define am_barrier(info_bit) CMAM_barrier(info_bit)

#define am_open_segment(buffer, len, handler,  data) \
	CMAM_open_segment((void *)buffer, len, handler, data)
#define am_open_this_segment(seg, buffer, len, handler,  data) \
	CMAM_open_this_segment(seg, (void *)buffer, len, handler, data)
#define amQuerySegment		CMAM_query_segment
#define amShortenSegment	CMAM_shorten_segment

#define am_rxfer CMAM_reply_xfer
#define am_xferBuffer(pr, seg, buff, len)	CMAM_xfer(pr, seg, buff, len)
#define am_rxferBuffer(pr, seg, buff, len)	CMAM_reply_xfer(pr, seg, buff, len)
#define am_rxferBuffer2Node(pr, seg, buff, len)	CMAM_reply_xfer(pr, seg, buff, len)
 
#define am_request_tohost0(fun) CMAM_host_4(fun)
#define am_request_tohost1(fun, one) CMAM_host_4(fun, one)
#define am_request_tohost2(fun, one, two) CMAM_host_4(fun, one, two)
#define am_request_tohost3(fun, one, two, three) \
	CMAM_host_4(fun, one, two, three)
#define am_request_tohost4(fun, one, two, three, four) \
	CMAM_host_4(fun, one, two, three, four)

#define am_rpc_tohost0(fun) CMAM_host_4(fun)
#define am_rpc_tohost1(fun, one) CMAM_host_4(fun, one)
#define am_rpc_tohost2(fun, one, two) CMAM_host_4(fun, one, two)
#define am_rpc_tohost3(fun, one, two, three) \
	CMAM_host_4(fun, one, two, three)
#define am_rpc_tohost4(fun, one, two, three, four) \
	CMAM_host_4(fun, one, two, three, four)

 
#define am_reply_tohost0(fun) CMAM_host_reply_4(fun)
#define am_reply_tohost1(fun, one) CMAM_host_reply_4(fun, one)
#define am_reply_tohost2(fun, one, two) CMAM_host_reply_4(fun, one, two)
#define am_reply_tohost3(fun, one, two, three) \
	CMAM_host_reply_4(fun, one, two, three)
#define am_reply_tohost4(fun, one, two, three, four) \
	CMAM_host_reply_4(fun, one, two, three, four)

#define am_sendToHost(code, one, two, three)	\
	CMAM_host_indirect_4(code, myProc, one, two ,three)
#define am_sendExtraToHost(code, one, two, three, four)	\
	CMAM_host_indirect_4(code, myProc|(one<<16), two ,three, four)

#define am_free(ptr) CMAM_free(ptr) 
#define am_malloc(size) CMAM_malloc(size)

#ifdef pe_obj
#	define am_xferBuffer2Host(seg, buffer, len) \
		CMAM_host_xfer(seg, buffer, len)
#	define am_rxferBuffer2Host(seg, buffer, len) \
		CMAM_host_reply_xfer(seg, buffer, len)
#else
#	define am_xferBuffer2Node(pr, seg, buffer, len)	\
		CMAM_xfer(pr, seg, buffer, len)
#endif

#else /* CMMD */

#ifdef	SPLIT_C_DEBUG
#	 define CMAML_DEBUG
#else
#    define barrier() cmaml_barrier(1)
#endif

#include <cm/cmna.h>
#include <cmsys/ni_reg_offset.h>
#include <cm/cmmd.h>
#include "am_cmmd.h"
#define splitc_poll() CMAML_poll()

/*********/
/* types */
/*********/

/********************/
/* message routines */
/********************/


#define am_reply0(dest, fun) \
	CMAML_reply(dest, fun)
#define am_reply1(dest, fun, one) \
	CMAML_reply(dest, fun, one)
#define am_reply2(dest, fun, one, two) \
	CMAML_reply(dest, fun, one, two)
#define am_reply3(dest, fun, one, two, three) \
	CMAML_reply(dest, fun, one, two, three)
#define am_reply4(node, fun, one, two, three, four) \
	CMAML_reply(node, fun, one, two, three, four)
#define am_rpc4(node, fun, one, two, three, four) \
	CMAML_rpc(node, fun, one, two, three, four)
#define am_rpc3(node, fun, one, two, three) \
	CMAML_rpc(node, fun, one, two, three)
#define am_rpc2(node, fun, one, two) \
	CMAML_rpc(node, fun, one, two)
#define am_rpc1(node, fun, one) \
	CMAML_rpc(node, fun, one)
#define am_rpc0(node, fun) \
	CMAML_rpc(node, fun)
#define am_request4(node, fun, one, two, three, four) \
	CMAML_request(node, fun, one, two, three, four)
#define am_request3(node, fun, one, two, three) \
	CMAML_request(node, fun, one, two, three)
#define am_request2(node, fun, one, two) \
	CMAML_request(node, fun, one, two)
#define am_request1(node, fun, one) \
	CMAML_request(node, fun, one)
#define am_request0(node, fun) \
	CMAML_request(node, fun)
#define am_poll()	CMAML_poll()
#define am_request_poll()	CMAML_request_poll()
#define am_reply_poll()	CMAML_reply_poll()
#define am_wait(flag, value) cmaml_wait(flag, value)
#define am_poll_wait(flag, value) CMAML_poll_wait(flag, value)
#define am_barrier(info_bit) cmaml_barrier(info_bit)

#define am_open_segment(buffer, len, handler,  data) \
	cmaml_open_segment((void *)buffer, len, handler, data)
#define am_open_this_segment(seg, buffer, len, handler,  data) \
	cmaml_open_this_segment(seg, (void *)buffer, len, handler, data)
#define am_free_rport(rport_num) CMAML_free_rport(rport_num)
#define am_get_rport_addr(rport_num) CMAML_get_rport_addr(rport_num)

#define am_rxfer cmaml_xfer
#define am_xferBuffer(pr, seg, buff, len)	cmaml_xfer(pr, seg, buff, len)
#define am_rxferBuffer(pr, seg, buff, len)	cmaml_xfer(pr, seg, buff, len)
#define am_rxferBuffer2Node(pr, seg, buff, len)	cmaml_xfer(pr, seg, buff, len)
 
#define am_request_tohost0(fun) CMAML_request_tohost(fun)
#define am_request_tohost1(fun, one) CMAML_request_tohost(fun, one)
#define am_request_tohost2(fun, one, two) CMAML_request_tohost(fun, one, two)
#define am_request_tohost3(fun, one, two, three) \
	CMAML_request_tohost(fun, one, two, three)
#define am_request_tohost4(fun, one, two, three, four) \
	CMAML_request_tohost(fun, one, two, three, four)

#define am_rpc_tohost0(fun) CMAML_rpc_tohost(fun)
#define am_rpc_tohost1(fun, one) CMAML_rpc_tohost(fun, one)
#define am_rpc_tohost2(fun, one, two) CMAML_rpc_tohost(fun, one, two)
#define am_rpc_tohost3(fun, one, two, three) \
	CMAML_rpc_tohost(fun, one, two, three)
#define am_rpc_tohost4(fun, one, two, three, four) \
	CMAML_rpc_tohost(fun, one, two, three, four)

 
#define am_reply_tohost0(fun) CMAML_reply_tohost(fun)
#define am_reply_tohost1(fun, one) CMAML_reply_tohost(fun, one)
#define am_reply_tohost2(fun, one, two) CMAML_reply_tohost(fun, one, two)
#define am_reply_tohost3(fun, one, two, three) \
	CMAML_reply_tohost(fun, one, two, three)
#define am_reply_tohost4(fun, one, two, three, four) \
	CMAML_reply_tohost(fun, one, two, three, four)

#define am_sendToHost(code, one, two, three)	\
	CMAML_request_tohost((*(code)), myProc, one, two ,three)
#define am_sendExtraToHost(code, one, two, three, four)	\
	CMAML_request_tohost((*(code)), myProc|(one<<16), two ,three, four)

#define am_free(ptr) free(ptr) 
#define am_malloc(size) malloc(size)

#ifdef pe_obj
#	define am_xferBuffer2Host(seg, buffer, len) \
		cmaml_xfer(CMMD_host_node(), seg, buffer, len)
#	define am_rxferBuffer2Host(seg, buffer, len) \
		cmaml_xfer(CMMD_host_node(), seg, buffer, len)
#else
#	define am_xferBuffer2Node(pr, seg, buffer, len)	\
		cmaml_xfer(pr, seg, buffer, len)
#endif
#endif /* CMMD */

void am_terminate();

void am_request5();
void am_request8();
void am_request11();
#define am_request6 am_request8
#define am_request7	am_request8
#define am_request9 am_request11
#define am_request10 am_request11

#endif



