/*                                                                      tab:4

 * com.header - Header file to generate com.sc
 *
 * 
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author:                      Arvind Krishnamurthy
 * Version:                     6
 * Creation Date:       Apr  2 13:42:47 1993
 * Filename:            com.sc
 * History:              
 *      ssl     6       Thu Oct 28 23:41:04 1993
 *              com.h and cm/cmmd.h are now included with split-c.h
 *      LTL     4       Mon Aug 16 15:33:59 1993
 *              Fixed it up for version 112 to support CMMD
 *      AK      2       Fri Apr  2 13:42:47 1993
 *              Obtained from David's initial version of com library.
 */
static char _version_string_[] = "\nVersion:6:com.sc\0\n";

#include <split-c/split-c.h>

/*
   This library provides global communication support for Split-C, including
   broadcasts, reductions, and scans.  

   It provides operations in which all processors participate.
   There are two basic styles: a processor-oriented functional interface
   and a spread array interface.

   Currently only the processor interface is supported.

   Eventually it will also provide autonomous operations.

   An all_ routine end with a barrier, so that there is no conflict on
   use of internal state.  Routines that consume a spread array also
   begin with a barrier, to reduce the chance that it is still being
   produced when the routine is called.

 */

#define ADD(A, B)	(A + B)
#define MAX(A, B)	(A < B ? B : A)
#define MIN(A, B)	(A < B ? A : B)
#define MULT(A, B)	(A * B)
#define OR(A, B)	(A | B)
#define XOR(A, B)	(A ^ B)
#define AND(A, B)	(A & B)

int 
all_reduce_to_one_add (int val)
{
  barrier ();
  return (CMMD_reduce_int (val, CMMD_combiner_add));
}

unsigned int 
all_reduce_to_one_uadd (unsigned int val)
{
  barrier ();
  return (CMMD_reduce_uint (val, CMMD_combiner_uadd));
}

float 
all_reduce_to_one_fadd (float val)
{
  barrier ();
  return (CMMD_reduce_float (val, CMMD_combiner_fadd));
}

double 
all_reduce_to_one_dadd (double val)
{
  barrier ();
  return (CMMD_reduce_double (val, CMMD_combiner_dadd));
}

int 
all_reduce_to_one_max (int val)
{
  barrier ();
  return (CMMD_reduce_int (val, CMMD_combiner_max));
}

unsigned int 
all_reduce_to_one_umax (unsigned int val)
{
  barrier ();
  return (CMMD_reduce_uint (val, CMMD_combiner_umax));
}

float 
all_reduce_to_one_fmax (float val)
{
  barrier ();
  return (CMMD_reduce_float (val, CMMD_combiner_fmax));
}

double 
all_reduce_to_one_dmax (double val)
{
  barrier ();
  return (CMMD_reduce_double (val, CMMD_combiner_dmax));
}

int 
all_reduce_to_one_min (int val)
{
  barrier ();
  return (CMMD_reduce_int (val, CMMD_combiner_min));
}

unsigned int 
all_reduce_to_one_umin (unsigned int val)
{
  barrier ();
  return (CMMD_reduce_uint (val, CMMD_combiner_umin));
}

float 
all_reduce_to_one_fmin (float val)
{
  barrier ();
  return (CMMD_reduce_float (val, CMMD_combiner_fmin));
}

double 
all_reduce_to_one_dmin (double val)
{
  barrier ();
  return (CMMD_reduce_double (val, CMMD_combiner_dmin));
}

int 
all_reduce_to_one_or (int val)
{
  barrier ();
  return (CMMD_reduce_int (val, CMMD_combiner_ior));
}

unsigned int 
all_reduce_to_one_uor (unsigned int val)
{
  barrier ();
  return (CMMD_reduce_uint (val, CMMD_combiner_ior));
}

int 
all_reduce_to_one_xor (int val)
{
  barrier ();
  return (CMMD_reduce_int (val, CMMD_combiner_xor));
}

unsigned int 
all_reduce_to_one_uxor (unsigned int val)
{
  barrier ();
  return (CMMD_reduce_uint (val, CMMD_combiner_xor));
}

int 
all_reduce_to_one_and (int val)
{
  barrier ();
  return (CMMD_reduce_int (val, CMMD_combiner_and));
}

unsigned int 
all_reduce_to_one_uand (unsigned int val)
{
  barrier ();
  return (CMMD_reduce_uint (val, CMMD_combiner_and));
}

int 
all_bcast (int val)
{
  int temp = val;

  barrier ();
  if (MYPROC == 0)
    CMMD_bc_to_nodes (&temp, sizeof (int));
  else
    CMMD_receive_bc_from_node (&temp, sizeof (int));
  return (temp);
}

int 
all_bcast_i (int val)
{
  int temp = val;

  barrier ();
  if (MYPROC == 0)
    CMMD_bc_to_nodes (&temp, sizeof (int));
  else
    CMMD_receive_bc_from_node (&temp, sizeof (int));
  return (temp);
}

unsigned int 
all_bcast_u (unsigned int val)
{
  unsigned int temp = val;

  barrier ();
  if (MYPROC == 0)
    CMMD_bc_to_nodes (&temp, sizeof (unsigned int));
  else
    CMMD_receive_bc_from_node (&temp, sizeof (unsigned int));
  return (temp);
}

float 
all_bcast_f (float val)
{
  float temp = val;

  barrier ();
  if (MYPROC == 0)
    CMMD_bc_to_nodes (&temp, sizeof (float));
  else
    CMMD_receive_bc_from_node (&temp, sizeof (float));
  return (temp);
}

double 
all_bcast_d (double val)
{
  double temp = val;

  barrier ();
  if (MYPROC == 0)
    CMMD_bc_to_nodes (&temp, sizeof (double));
  else
    CMMD_receive_bc_from_node (&temp, sizeof (double));
  return (temp);
}

void *
all_bcast_p (void *val)
{
  void *temp = val;

  barrier ();
  if (MYPROC == 0)
    CMMD_bc_to_nodes (&temp, sizeof (void *));
  else
    CMMD_receive_bc_from_node (&temp, sizeof (void *));
  return (temp);
}

void 
all_bcast_buffer (___V_SPTR buf, int size)
{

  barrier ();
  if (MYPROC == 0)
    CMMD_bc_to_nodes (tolocal (buf), size);
  else
    CMMD_receive_bc_from_node (tolocal (buf), size);
}

int 
all_reduce_to_all_add (int val)
{
  barrier ();
  return (CMMD_reduce_int (val, CMMD_combiner_add));
}

unsigned int 
all_reduce_to_all_uadd (unsigned int val)
{
  barrier ();
  return (CMMD_reduce_uint (val, CMMD_combiner_uadd));
}

float 
all_reduce_to_all_fadd (float val)
{
  barrier ();
  return (CMMD_reduce_float (val, CMMD_combiner_fadd));
}

double 
all_reduce_to_all_dadd (double val)
{
  barrier ();
  return (CMMD_reduce_double (val, CMMD_combiner_dadd));
}

int 
all_reduce_to_all_max (int val)
{
  barrier ();
  return (CMMD_reduce_int (val, CMMD_combiner_max));
}

unsigned int 
all_reduce_to_all_umax (unsigned int val)
{
  barrier ();
  return (CMMD_reduce_uint (val, CMMD_combiner_umax));
}

float 
all_reduce_to_all_fmax (float val)
{
  barrier ();
  return (CMMD_reduce_float (val, CMMD_combiner_fmax));
}

double 
all_reduce_to_all_dmax (double val)
{
  barrier ();
  return (CMMD_reduce_double (val, CMMD_combiner_dmax));
}

int 
all_reduce_to_all_min (int val)
{
  barrier ();
  return (CMMD_reduce_int (val, CMMD_combiner_min));
}

unsigned int 
all_reduce_to_all_umin (unsigned int val)
{
  barrier ();
  return (CMMD_reduce_uint (val, CMMD_combiner_umin));
}

float 
all_reduce_to_all_fmin (float val)
{
  barrier ();
  return (CMMD_reduce_float (val, CMMD_combiner_fmin));
}

double 
all_reduce_to_all_dmin (double val)
{
  barrier ();
  return (CMMD_reduce_double (val, CMMD_combiner_dmin));
}

int 
all_reduce_to_all_or (int val)
{
  barrier ();
  return (CMMD_reduce_int (val, CMMD_combiner_ior));
}

unsigned int 
all_reduce_to_all_uor (unsigned int val)
{
  barrier ();
  return (CMMD_reduce_uint (val, CMMD_combiner_ior));
}

int 
all_reduce_to_all_xor (int val)
{
  barrier ();
  return (CMMD_reduce_int (val, CMMD_combiner_xor));
}

unsigned int 
all_reduce_to_all_uxor (unsigned int val)
{
  barrier ();
  return (CMMD_reduce_uint (val, CMMD_combiner_xor));
}

int 
all_reduce_to_all_and (int val)
{
  barrier ();
  return (CMMD_reduce_int (val, CMMD_combiner_and));
}

unsigned int 
all_reduce_to_all_uand (unsigned int val)
{
  barrier ();
  return (CMMD_reduce_uint (val, CMMD_combiner_and));
}

int 
all_scan_add (int val)
{
  barrier ();
  return (CMMD_scan_int (val, CMMD_combiner_add, CMMD_upward, CMMD_none,
			 0, CMMD_inclusive));
}

unsigned int 
all_scan_uadd (unsigned int val)
{
  barrier ();
  return (CMMD_scan_uint (val, CMMD_combiner_uadd, CMMD_upward, CMMD_none,
			  0, CMMD_inclusive));
}

float 
all_scan_fadd (float val)
{
  barrier ();
  return (CMMD_scan_float (val, CMMD_combiner_fadd, CMMD_upward, CMMD_none,
			   0, CMMD_inclusive));
}

double 
all_scan_dadd (double val)
{
  barrier ();
  return (CMMD_scan_double (val, CMMD_combiner_dadd, CMMD_upward, CMMD_none,
			    0, CMMD_inclusive));
}

int 
all_scan_max (int val)
{
  barrier ();
  return (CMMD_scan_int (val, CMMD_combiner_max, CMMD_upward, CMMD_none,
			 0, CMMD_inclusive));
}

unsigned int 
all_scan_umax (unsigned int val)
{
  barrier ();
  return (CMMD_scan_uint (val, CMMD_combiner_umax, CMMD_upward, CMMD_none,
			  0, CMMD_inclusive));
}

float 
all_scan_fmax (float val)
{
  barrier ();
  return (CMMD_scan_float (val, CMMD_combiner_fmax, CMMD_upward, CMMD_none,
			   0, CMMD_inclusive));
}

double 
all_scan_dmax (double val)
{
  barrier ();
  return (CMMD_scan_double (val, CMMD_combiner_dmax, CMMD_upward, CMMD_none,
			    0, CMMD_inclusive));
}

int 
all_scan_min (int val)
{
  barrier ();
  return (CMMD_scan_int (val, CMMD_combiner_min, CMMD_upward, CMMD_none,
			 0, CMMD_inclusive));
}

unsigned int 
all_scan_umin (unsigned int val)
{
  barrier ();
  return (CMMD_scan_uint (val, CMMD_combiner_umin, CMMD_upward, CMMD_none,
			  0, CMMD_inclusive));
}

float 
all_scan_fmin (float val)
{
  barrier ();
  return (CMMD_scan_float (val, CMMD_combiner_fmin, CMMD_upward, CMMD_none,
			   0, CMMD_inclusive));
}

double 
all_scan_dmin (double val)
{
  barrier ();
  return (CMMD_scan_double (val, CMMD_combiner_dmin, CMMD_upward, CMMD_none,
			    0, CMMD_inclusive));
}

int 
all_scan_or (int val)
{
  barrier ();
  return (CMMD_scan_int (val, CMMD_combiner_ior, CMMD_upward, CMMD_none,
			 0, CMMD_inclusive));
}

unsigned int 
all_scan_uor (unsigned int val)
{
  barrier ();
  return (CMMD_scan_uint (val, CMMD_combiner_ior, CMMD_upward, CMMD_none,
			  0, CMMD_inclusive));
}

int 
all_scan_xor (int val)
{
  barrier ();
  return (CMMD_scan_int (val, CMMD_combiner_xor, CMMD_upward, CMMD_none,
			 0, CMMD_inclusive));
}

unsigned int 
all_scan_uxor (unsigned int val)
{
  barrier ();
  return (CMMD_scan_uint (val, CMMD_combiner_xor, CMMD_upward, CMMD_none,
			  0, CMMD_inclusive));
}

int 
all_scan_and (int val)
{
  barrier ();
  return (CMMD_scan_int (val, CMMD_combiner_and, CMMD_upward, CMMD_none,
			 0, CMMD_inclusive));
}

unsigned int 
all_scan_uand (unsigned int val)
{
  barrier ();
  return (CMMD_scan_uint (val, CMMD_combiner_and, CMMD_upward, CMMD_none,
			  0, CMMD_inclusive));
}

static long long reduce_buf[2];
static long long bcast_buf;

static long long scan_buf[20];
static long long temp_scan_val[20];
static long long scan_result_buf;

int 
all_reduce_to_one_mult (int val)
{
  int parent = (MYPROC - 1) / 2;
  int odd_child = 2 * MYPROC + 1;
  int even_child = 2 * MYPROC + 2;
  int parity = MYPROC & 1;
  long long *buf = reduce_buf;
  reset_store_sync ();
  barrier ();
  if (MYPROC == 0)
    {
      if (PROCS == 1)
	{
	  return (val);
	}
      else if (PROCS == 2)
	{
	  store_sync (sizeof (int));
	  return (MULT (val, *((int *) &buf[1])));
	}
      else
	{
	  store_sync (2 * sizeof (int));
	  return (MULT (MULT (val, *((int *) &buf[1])), *((int *) buf)));
	}
    }
  else
    {
      ___globPtrType dest = toglobal (parent, (void *) &reduce_buf[parity]);
      if (odd_child >= PROCS)
	{
	  __i_store (dest, val);
	}
      else if (even_child >= PROCS)
	{
	  store_sync (sizeof (int));
	  __i_store (dest, MULT (val, *((int *) &buf[1])));
	}
      else
	{
	  store_sync (2 * sizeof (int));
	  __i_store (dest, MULT (MULT (val, *((int *) &buf[1])), *((int *) buf)));
	}
    }
  return (val);
}
unsigned int 
all_reduce_to_one_umult (unsigned int val)
{
  int parent = (MYPROC - 1) / 2;
  int odd_child = 2 * MYPROC + 1;
  int even_child = 2 * MYPROC + 2;
  int parity = MYPROC & 1;
  long long *buf = reduce_buf;
  reset_store_sync ();
  barrier ();
  if (MYPROC == 0)
    {
      if (PROCS == 1)
	{
	  return (val);
	}
      else if (PROCS == 2)
	{
	  store_sync (sizeof (unsigned int));
	  return (MULT (val, *((unsigned int *) &buf[1])));
	}
      else
	{
	  store_sync (2 * sizeof (unsigned int));
	  return (MULT (MULT (val, *((unsigned int *) &buf[1])), *((unsigned int *) buf)));
	}
    }
  else
    {
      ___globPtrType dest = toglobal (parent, (void *) &reduce_buf[parity]);
      if (odd_child >= PROCS)
	{
	  __i_store (dest, val);
	}
      else if (even_child >= PROCS)
	{
	  store_sync (sizeof (unsigned int));
	  __i_store (dest, MULT (val, *((unsigned int *) &buf[1])));
	}
      else
	{
	  store_sync (2 * sizeof (unsigned int));
	  __i_store (dest, MULT (MULT (val, *((unsigned int *) &buf[1])), *((unsigned int *) buf)));
	}
    }
  return (val);
}
float 
all_reduce_to_one_fmult (float val)
{
  int parent = (MYPROC - 1) / 2;
  int odd_child = 2 * MYPROC + 1;
  int even_child = 2 * MYPROC + 2;
  int parity = MYPROC & 1;
  long long *buf = reduce_buf;
  reset_store_sync ();
  barrier ();
  if (MYPROC == 0)
    {
      if (PROCS == 1)
	{
	  return (val);
	}
      else if (PROCS == 2)
	{
	  store_sync (sizeof (float));
	  return (MULT (val, *((float *) &buf[1])));
	}
      else
	{
	  store_sync (2 * sizeof (float));
	  return (MULT (MULT (val, *((float *) &buf[1])), *((float *) buf)));
	}
    }
  else
    {
      ___globPtrType dest = toglobal (parent, (void *) &reduce_buf[parity]);
      if (odd_child >= PROCS)
	{
	  __f_store (dest, val);
	}
      else if (even_child >= PROCS)
	{
	  store_sync (sizeof (float));
	  __f_store (dest, MULT (val, *((float *) &buf[1])));
	}
      else
	{
	  store_sync (2 * sizeof (float));
	  __f_store (dest, MULT (MULT (val, *((float *) &buf[1])), *((float *) buf)));
	}
    }
  return (val);
}
double 
all_reduce_to_one_dmult (double val)
{
  int parent = (MYPROC - 1) / 2;
  int odd_child = 2 * MYPROC + 1;
  int even_child = 2 * MYPROC + 2;
  int parity = MYPROC & 1;
  long long *buf = reduce_buf;
  reset_store_sync ();
  barrier ();
  if (MYPROC == 0)
    {
      if (PROCS == 1)
	{
	  return (val);
	}
      else if (PROCS == 2)
	{
	  store_sync (sizeof (double));
	  return (MULT (val, *((double *) &buf[1])));
	}
      else
	{
	  store_sync (2 * sizeof (double));
	  return (MULT (MULT (val, *((double *) &buf[1])), *((double *) buf)));
	}
    }
  else
    {
      ___globPtrType dest = toglobal (parent, (void *) &reduce_buf[parity]);
      if (odd_child >= PROCS)
	{
	  __d_store (dest, val);
	}
      else if (even_child >= PROCS)
	{
	  store_sync (sizeof (double));
	  __d_store (dest, MULT (val, *((double *) &buf[1])));
	}
      else
	{
	  store_sync (2 * sizeof (double));
	  __d_store (dest, MULT (MULT (val, *((double *) &buf[1])), *((double *) buf)));
	}
    }
  return (val);
}

int 
all_reduce_to_all_mult (int val)
{
  int result = all_reduce_to_one_mult (val);
  return (all_bcast_i (result));
}
unsigned int 
all_reduce_to_all_umult (unsigned int val)
{
  unsigned int result = all_reduce_to_one_umult (val);
  return (all_bcast_u (result));
}
float 
all_reduce_to_all_fmult (float val)
{
  float result = all_reduce_to_one_fmult (val);
  return (all_bcast_f (result));
}
double 
all_reduce_to_all_dmult (double val)
{
  double result = all_reduce_to_one_dmult (val);
  return (all_bcast_d (result));
}

int 
all_scan_mult (int val)
{
  int dest;
  int i, count;
  int bit_vec = 1;
  int scan_val;
  long long *local_buf = (long long *) scan_buf;
  reset_store_sync ();
  barrier ();
  *((int *) &temp_scan_val[0]) = scan_val = val;
  for (count = 0; bit_vec != PROCS; count++, bit_vec = bit_vec << 1)
    {
      dest = MYPROC & (~bit_vec);
      if (dest != MYPROC)
	break;
    }
  store_sync (count * sizeof (int));
  for (i = 1; i <= count; i++)
    {
      scan_val = MULT (scan_val, *((int *) &local_buf[i]));
      *((int *) &temp_scan_val[i]) = scan_val;
    }
  __i_store (toglobal (dest, &scan_buf[count + 1]), scan_val);
  count--;
  bit_vec = bit_vec >> 1;
  if (MYPROC != 0)
    {
      int result_val;
      store_sync (sizeof (int));
      result_val = *((int *) &scan_result_buf);
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store (toglobal (dest, &scan_result_buf), MULT (result_val, *((int *) &temp_scan_val[count])));
	}
      barrier ();
      return MULT (result_val, val);
    }
  else
    {
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store (toglobal (dest, &scan_result_buf), *((int *) &temp_scan_val[count]));
	}
      barrier ();
      return val;
    }
}
unsigned int 
all_scan_umult (unsigned int val)
{
  int dest;
  int i, count;
  int bit_vec = 1;
  unsigned int scan_val;
  long long *local_buf = (long long *) scan_buf;
  reset_store_sync ();
  barrier ();
  *((unsigned int *) &temp_scan_val[0]) = scan_val = val;
  for (count = 0; bit_vec != PROCS; count++, bit_vec = bit_vec << 1)
    {
      dest = MYPROC & (~bit_vec);
      if (dest != MYPROC)
	break;
    }
  store_sync (count * sizeof (unsigned int));
  for (i = 1; i <= count; i++)
    {
      scan_val = MULT (scan_val, *((unsigned int *) &local_buf[i]));
      *((unsigned int *) &temp_scan_val[i]) = scan_val;
    }
  __i_store (toglobal (dest, &scan_buf[count + 1]), scan_val);
  count--;
  bit_vec = bit_vec >> 1;
  if (MYPROC != 0)
    {
      unsigned int result_val;
      store_sync (sizeof (unsigned int));
      result_val = *((unsigned int *) &scan_result_buf);
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store (toglobal (dest, &scan_result_buf), MULT (result_val, *((unsigned int *) &temp_scan_val[count])));
	}
      barrier ();
      return MULT (result_val, val);
    }
  else
    {
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store (toglobal (dest, &scan_result_buf), *((unsigned int *) &temp_scan_val[count]));
	}
      barrier ();
      return val;
    }
}
float 
all_scan_fmult (float val)
{
  int dest;
  int i, count;
  int bit_vec = 1;
  float scan_val;
  long long *local_buf = (long long *) scan_buf;
  reset_store_sync ();
  barrier ();
  *((float *) &temp_scan_val[0]) = scan_val = val;
  for (count = 0; bit_vec != PROCS; count++, bit_vec = bit_vec << 1)
    {
      dest = MYPROC & (~bit_vec);
      if (dest != MYPROC)
	break;
    }
  store_sync (count * sizeof (float));
  for (i = 1; i <= count; i++)
    {
      scan_val = MULT (scan_val, *((float *) &local_buf[i]));
      *((float *) &temp_scan_val[i]) = scan_val;
    }
  __f_store (toglobal (dest, &scan_buf[count + 1]), scan_val);
  count--;
  bit_vec = bit_vec >> 1;
  if (MYPROC != 0)
    {
      float result_val;
      store_sync (sizeof (float));
      result_val = *((float *) &scan_result_buf);
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __f_store (toglobal (dest, &scan_result_buf), MULT (result_val, *((float *) &temp_scan_val[count])));
	}
      barrier ();
      return MULT (result_val, val);
    }
  else
    {
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __f_store (toglobal (dest, &scan_result_buf), *((float *) &temp_scan_val[count]));
	}
      barrier ();
      return val;
    }
}
double 
all_scan_dmult (double val)
{
  int dest;
  int i, count;
  int bit_vec = 1;
  double scan_val;
  long long *local_buf = (long long *) scan_buf;
  reset_store_sync ();
  barrier ();
  *((double *) &temp_scan_val[0]) = scan_val = val;
  for (count = 0; bit_vec != PROCS; count++, bit_vec = bit_vec << 1)
    {
      dest = MYPROC & (~bit_vec);
      if (dest != MYPROC)
	break;
    }
  store_sync (count * sizeof (double));
  for (i = 1; i <= count; i++)
    {
      scan_val = MULT (scan_val, *((double *) &local_buf[i]));
      *((double *) &temp_scan_val[i]) = scan_val;
    }
  __d_store (toglobal (dest, &scan_buf[count + 1]), scan_val);
  count--;
  bit_vec = bit_vec >> 1;
  if (MYPROC != 0)
    {
      double result_val;
      store_sync (sizeof (double));
      result_val = *((double *) &scan_result_buf);
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __d_store (toglobal (dest, &scan_result_buf), MULT (result_val, *((double *) &temp_scan_val[count])));
	}
      barrier ();
      return MULT (result_val, val);
    }
  else
    {
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __d_store (toglobal (dest, &scan_result_buf), *((double *) &temp_scan_val[count]));
	}
      barrier ();
      return val;
    }
}
