/*									tab:4
 *
 * handlers_cp.c - default handlers that run on host
 *
 * Copyright (c) 1992 by Seth Copen Goldstein and 
 * Regents of the University of California
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  Copyright holder(s) make no
 * representation about the suitability of this software for
 * any purpose. It is provided "as is" without express or
 * implied warranty.
 *
 * Author: 			Seth Copen Goldstein
 * Version:			54
 * Creation Date:	Wed Jan 20 15:15:52 1993
 * Filename:		handlers_cp.c
 * History:
 *	LTL	52	Thu Nov 25 15:41:50 1993
 *		Fixed bug in setting host handlers
 *	ACD	45	Wed Nov 24 17:03:44 1993
 *		Steve needed a label for debugger break point.
 *	LTL	42	Tue Nov 23 21:31:03 1993
 *		Set splitc_partition_size in splitc_start_nodes
 *	LTL	38	Sat Oct 30 16:46:32 1993
 *		Change USER_HANDLER_BASE to DEFAULT_HOST_HANDLERS_NUM
 *	LTL	30	Wed Oct 27 15:31:56 1993
 *		Pulled out the X functions and put them in a separate
 *      library
 *	LTL	26	Mon Aug 16 15:38:12 1993
 *		Fixed it up for version 112 to support CMMD
 *	AK	23	Tue Jun 22 17:54:11 1993
 *		Modified fgets handler to return the result.
 *	ACD	21	Mon Jun 14 10:18:38 1993
 *		Removed sbrk_handler interface
 *	SL	19	Tue May 25 12:54:57 1993
 *		Added basic X interface
 *  SL  18  Wed May 12 23:08:20 PDT 1993
 *      Added SC_PROFILE option
 * 	TvE	17	Mon Apr 12 09:00:31 PDT 1993
 *		Bug fix from Madison about FLOAT
 *	SCG	14	Fri Jan 22 13:50:29 1993
 *		added tve's set_procs_handler
 */
static char _version_string_[] = "\nVersion:52:handlers_cp.c\0\n";


#include "host.h"
#include "print.h"
#include <errno.h>

int splitc_notDone = 1;
volatile int splitc_exitCode = 0;

#ifdef SC_PROFILE  /* Count of processors ready to profile */
static int profReadyCount = 0;
#endif
static int finishedCount = 0;
int splitc_partition_size;

#ifndef CMMD
extern VOID CMPE_floatResponse();
extern VOID CMPE_getintResponse();
extern VOID CMPE_channelResponse();
extern VOID CMPE_fopenResponse();
extern VOID CMPE_fcloseResponse();
extern VOID CMPE_ungetcResponse();
extern VOID CMPE_incrFlag();
extern VOID CMPE_getcResponse();
extern VOID CMPE_printGetChannel();
#else
Handler CMPE_printGetChannel;   /* need this for X_BLOCK */
#endif /* CMMD */


/************************************************************/
/* some debugging routines.  TAKE THIS OUT BEFORE SHIPPING? */
/************************************************************/

#ifdef LOGGING_HOST
FILE *loger=NULL;
#define LOG_HOST_CALL(name, proc, two, three, four)	\
	fprintf(loger, "%3d:%s two:%d(%x) three:%d(%x) four:%d(%x)\n",	\
		   proc, name, two, two, three, three, four, four);		\
	fflush(loger);

#	define closeLogger()	fclose(loger);

#   define LogPrint(x)		DoLogPrint x

VOID DoLogPrint(format, VarArgBase)
char *format;
VarArgBaseDecl;
{
	VarArgDecl;
	VarArgInit(format);
	vfprintf(loger, format, VarArg);
	VarArgEnd();
	fflush(loger);
}

VOID openLogger(name)
char *name;
{
	char buffer[256];

	sprintf(buffer, "%s.split-c.log", name);
	loger = fopen(buffer, "w+");
	if (loger == NULL)
	{
		printf("unable to open %s as log file\n", buffer);
	}
}
#else
#	define LOG_HOST_CALL(name, proc, two, three, four)
#   define LogPrint(args)
#	define openLogger(name)
#	define closeLogger()
#endif

#ifndef CMMD
 
/*******************************************/
/* routines called by end of xfer function */
/*******************************************/
typedef struct {
        FILE *fp;
        long *offset;
} filePtrInfoType;

#define MAX_FILES_OPEN  100

filePtrInfoType fileInfo[MAX_FILES_OPEN];
int initialized = 0;

/* this is used for indiv and synchronized file accesses */
int doFopen(rf, base)
remoteFunction *rf;
char *base;
{
	FILE *f;

	f = fopen(base, strlen(base)+1+base);
	am_reply4(rf->pid, rf->function, f, rf->adr, rf->adr2, errno);
	return 0;
}


/* this is used for independent access to the same file */
int doGlobFopen(rf, base)
remoteFunction *rf;
char *base;
{
        FILE *f;
        int i, j;

        f = fopen(base, strlen(base)+1+base);
        for (i=0; i<MAX_FILES_OPEN; i++)
                if (fileInfo[i].fp == NULL) {
                        fileInfo[i].fp = f;
                        for (j=0; j<CMNA_partition_size; j++)
                                fileInfo[i].offset[j] = 0;
                        break;
                }
        am_reply4(rf->pid, rf->function, f, rf->adr, rf->adr2, errno);
        return 0;
}


int doPrint(info, base)
remoteFunction *info;
char *base;
{
	LogPrint(("    DoPrint: %x proc:%3d stream:%x base:%x\n",
			  info, info->pid, info->stream, base));
	
	fprintf(info->stream, "%s", base);
	am_reply2(info->pid, info->function, -3, info->adr);
	free(base);
	free(info);
	return 0;					/* deallocate this channel */
}


/************************************************/
/* functions called from PEs to get things done */
/************************************************/
VOID  GLOB_FOPEN_handler(Handler *f, int proc, int two, int three, int four)
{
        /* two - length in bytes of name + type */
        /* three - channel adress */
        /* four - file handle address */

        char *p;
        int i;
        Channel c;
        remoteFunction *rf;

        LOG_HOST_CALL("fopen", proc, two , three, four);

        if (!initialized) {
                for (i=0; i<MAX_FILES_OPEN; i++) {
                        fileInfo[i].fp = NULL;
                        fileInfo[i].offset = (int *)malloc(CMNA_partition_size*sizeof(long));
                }
                initialized = 1;
        }

        rf = (remoteFunction *)malloc(sizeof(remoteFunction));
        rf->pid = proc;
        rf->function = (Handler )CMPE_fopenResponse;
        rf->adr = (VOID_STAR )four;
        rf->adr2 = (VOID_STAR )three;

        p = malloc(two);
        c = am_open_segment(p, two, doGlobFopen, (void *)rf);

        am_reply2(proc, (Handler )CMPE_channelResponse, (int )c, three);
}


VOID  FOPEN_handler(Handler *f, int proc, int two, int three, int four)
{
	/* two - length in bytes of name + type */
	/* three - channel adress */
	/* four - file handle address */
	
	char *p;
	int i;
	Channel c;
	remoteFunction *rf;

	LOG_HOST_CALL("fopen", proc, two , three, four);

	rf = (remoteFunction *)malloc(sizeof(remoteFunction));
	rf->pid = proc;
	rf->function = (Handler )CMPE_fopenResponse;
	rf->adr = (VOID_STAR )four;
	rf->adr2 = (VOID_STAR )three;
	
	p = malloc(two);
	c = am_open_segment(p, two, doFopen, (void *)rf);
	
	am_reply2(proc, (Handler )CMPE_channelResponse, (int )c, three);
}

VOID  FCLOSE_handler(Handler *f, int proc, int two, int three, int four)
{
	/* two - stream to close */
	/* three - place to put errno */
	/* four - place to put completion flag */
	
	LOG_HOST_CALL("FCLOSE", proc, two, three, four);

	fclose((FILE *)two);
	am_reply3(proc, (Handler )CMPE_fcloseResponse, errno, three, four);
}

VOID  UNGETC_handler(Handler *f, int proc, int two, int three, int four)
{
		/* two - stream */
		/* three - character to unget */
		/* four - address of completeion flag */

	LOG_HOST_CALL("UNGETC", proc, two, three, four);

	ungetc(three, (FILE *)two);
	am_reply2(proc, CMPE_ungetcResponse, errno, four);
}


VOID  GETINT_handler(Handler *f, int proc, int two, int three, int four)
{
	/* two - stream */
	/* three - adr to put value */
	/* four - adr to put flag */

	int i;

	LOG_HOST_CALL("GETINT", proc, two, three, four);

	fscanf((FILE *)two, "%d", &i);
	am_reply4(proc, CMPE_getintResponse, i, three, four, errno);
}


VOID  GETFLOAT_handler(Handler *f, int proc, int two, int three, int four)
{
	/* two - stream */
	/* three - adr to put value */
	/* four - adr to put flag */
	
	double fl;

	LOG_HOST_CALL("GETFLOAT", proc, two, three, four);
		
	fscanf((FILE *)two, "%lf", &fl);
	am_reply3(proc, CMPE_floatResponse, fl, three, four);
}
		
VOID  FERROR_handler(Handler *f, int proc, int two, int three, int four)
{
	/* two - stream to get error on */
	/* three - place to put errno */
	/* four - place to put completion flag */
	
	int i;
	LOG_HOST_CALL("FERROR", proc, two, three, four);

	i = ferror((FILE *)two);
	am_reply3(proc, (Handler )CMPE_fcloseResponse, i, three, four);
}

		
VOID  TERMINATE_handler(Handler *f, int proc, int two, int three, int four)
{
	LOG_HOST_CALL("TERMINATE", proc, two, three, four);
}
		

VOID  FFLUSH_handler(Handler *f, int proc, int two, int three, int four)
{
        int retVal;

        LOG_HOST_CALL("FFLUSH", proc, two, three, four);

        retVal = fflush((FILE *)two);
        if (three)
                am_reply2(proc, CMPE_incrFlag, three, retVal);
}

VOID  PUTC_handler(Handler *f, int proc, int two, int three, int four)
{
	/* 2 - stream to put character to */
	/* 3 - character to put onto stream */
	/* 4 - if !0 -> adr of flag that will be incremeneted */
		
	LOG_HOST_CALL("PUTC", proc, two, three, four);

	putc(three, (FILE *)two);
	if (four)
		am_reply1(proc, CMPE_incrFlag, four);
}
		
VOID  GETC_handler(Handler *f, int proc, int two, int three, int four)
{
	/* 2 - stream */
	/* 3 - dest adr of char */
	/* 4 - dest adr of flag */

	int l;
	
	LOG_HOST_CALL("GETC", proc, two, three, four);
	
	l = getc((FILE *)two);
	am_reply3(proc, CMPE_getcResponse, l, three, four);
}

VOID  FGETS_handler(Handler *f, int one, int two, int three, int four)
{
	/* two is the segment # */
	/* three - filehandle */
	/* four - rpc if we need to prematurly close */

	int proc = one & 0xffff;	/* processor doing request */
	int l = one >> 16;			/* max length expecting */
	char *p;
	int actL;					/* length actually read */
	int retValue;

	LOG_HOST_CALL("FGETS", proc, two, three, four);

	p = malloc(l);
	retValue = (int)fgets(p, l, (FILE *)three);

	am_rxferBuffer2Node(proc, two, (int *)p, actL = (strlen(p)+1));
	if (retValue == (int)NULL)
		am_reply2(proc, (Handler )four, two, -1);
	else
		am_reply2(proc, (Handler )four, two, l-actL);
}
		
VOID  INIT_IO_handler(Handler *f, int proc, int two, int three, int four)
{
	LOG_HOST_CALL("INIT_IO", proc, two, three, four);

	am_reply3(proc, (Handler )two, (int )stdin, (int )stdout, (int )stderr);
}
		
VOID  FPRINT_handler(Handler *f, int proc, int two, int three, int four)
{
	/* two -   length in bytes */
	/* three - file stream to print on */
	/* four -  place to put channel on node */

	Channel c;
	remoteFunction *rf;
	char *p;
	
	rf = (remoteFunction *)malloc(sizeof(remoteFunction));
	rf->pid = proc;
	rf->function = (Handler )CMPE_printGetChannel;
	rf->stream = (FILE *)three;
	rf->adr = (VOID_STAR)four;
		

	p = (char *)malloc(two);
	c = am_open_segment(p, two, doPrint, (void * )rf);

	LogPrint(("%3d:FPRINT, stream:%x len:%3d buffer: %x | %x<-%3d calls %x\n",
			  proc, three, two, p, four, c, rf));

	am_reply2(proc, CMPE_printGetChannel, (int )c, four);
}
		
VOID  PRINT_handler(Handler *f, int proc, int two, int three, int four)
{
	/* 2 - size */
	/* three - place to put channel on node */
	/* four - handler to call on node */

	char *p;
	int i;
	Channel c;
	remoteFunction *rf;

	LOG_HOST_CALL("PRINT", proc, two, three, four);
	
	rf = (remoteFunction *)malloc(sizeof(remoteFunction));
	rf->pid = proc;
	rf->function = (Handler )four;
	rf->stream = stdout;
	rf->adr = (VOID_STAR )three;
		
	p = (char *)malloc(two);
	c = am_open_segment(p, two, doPrint, (void * )rf);
	am_reply2(proc, (Handler )four, (int )c, three);
}
			
static int argCount;
static char **argVector;
static int argSize=0;
static int *offsets;

int splitc_packageArgs(int argc, char **argv)
{
	int i;

	argCount = argc;
	argVector = argv;
	offsets = (int *)malloc(sizeof(int) * argc);
	for (i=0; i<argc; i++)
	{
		offsets[i] = argSize;
		argSize += (strlen(argv[i])+1);
	}
	return argSize;
}

VOID  GETARGS_handler(Handler *f, int proc, int sid)
{
	int offset=sizeof(int)*argCount;
	int i;
	
	LOG_HOST_CALL("GETARGS", proc, sid, 0, 0);

	/* send the integer offsets, then a string at a time */
	am_rxferBuffer(proc, sid, offsets, argCount*sizeof(int));
	for (i=0; i<argCount; i++)
		am_rxferBuffer(proc,
				   sid + offset + offsets[i],
				   argVector[i],
				   strlen(argVector[i])+1);
}




#ifdef SC_PROFILE
VOID PROF_READY_handler(Handler *f, int proc, int two, int three, int four)
{
	LOG_HOST_CALL("PROF_READY", proc, two, three, four);

	if (++profReadyCount == splitc_partition_size) {
		/* Could do better with a tree notification */
		while (profReadyCount-- > 0)
			am_reply1(profReadyCount, CMPE_incrFlag, two);
	}
}
#endif

VOID FSEEK_handler(Handler *f, int proc, int two, int three, int four)
{
        /* 2 - stream */
        /* 3 - offset<<2|ptrname */
        /* 4 - adr for indicating done */

        long offset = three >> 2;
        int ptrname = three & 0x3;
        int result;
        int i;

        for (i=0; i<MAX_FILES_OPEN; i++)
                if (fileInfo[i].fp == (FILE *)two) {
                        fseek((FILE *)two, fileInfo[i].offset[proc], SEEK_SET);
                        result = fseek((FILE *)two, offset, ptrname);
                        fileInfo[i].offset[proc] = ftell((FILE *)two);
                }

        if (i == MAX_FILES_OPEN)
                result = fseek((FILE *)two, offset, ptrname);
        am_reply2(proc, CMPE_printGetChannel, result, four);
}


typedef int Segment;

typedef struct
{
	FILE *stream;
	Segment segment;
	void *adr;
} freadPacket;

VOID FREAD_handler(Handler *f, int proc, int two, int three, int four)
{
	/* 2 - stream */
	/* 3 - segment */
	/* 4 - return address */
	freadPacket *r;
	
	r = (freadPacket *)malloc(sizeof(freadPacket));
	r->stream = (FILE *)two;
	r->segment = three;
	r->adr = (void *)four;
	am_reply2(proc, CMPE_printGetChannel, r, four);
}

VOID FREAD2_handler(Handler *f, int proc, freadPacket *r, int three, int four)
{
        /* 2 - adr of freadPacket info */
        /* 3 - size */
        /* 4 - nitems */

        int result;
        char *buffer;
        int i;

        buffer = malloc(three * four);
        for (i=0; i<MAX_FILES_OPEN; i++)
                if (fileInfo[i].fp == r->stream) {
                        fseek(r->stream, fileInfo[i].offset[proc], 0);
                        result = fread(buffer, three, four, r->stream);
                        fileInfo[i].offset[proc] += result*three;
                        break;
                }
        if (i == MAX_FILES_OPEN)
                result = fread(buffer, three, four, r->stream);
        am_rxferBuffer2Node(proc, r->segment, (int *)buffer, three * result);
        am_reply2(proc, CMPE_printGetChannel, result, r->adr);
        free(buffer);
        free(r);
}


VOID  EXIT_handler(Handler *f, int proc, int two, int three, int four)
{
	LOG_HOST_CALL("EXIT", proc, two, three, four);
	if (three == 0)
	{
		if (++finishedCount == splitc_partition_size)
			splitc_exitCode = splitc_notDone = 0;
	}
	else
	{
		splitc_notDone = 0;			/* some node decided to do an exit call */
		splitc_exitCode = two;
	}
}


VOID  SET_PROCS_handler(Handler *f, int proc, int two, int three, int four)
{
	LOG_HOST_CALL("SET_PROCS", proc, two, three, four);

	splitc_partition_size = two;
	if (finishedCount >= splitc_partition_size)
		splitc_exitCode = splitc_notDone = 0;
}

#else

VOID  EXIT_handler(int proc, int two, int three, int four)
{
	LOG_HOST_CALL("EXIT", proc, two, three, four);

	if (three == 0)
	{
		if (++finishedCount == splitc_partition_size)
			splitc_exitCode = splitc_notDone = 0;
	}
	else
	{
		splitc_notDone = 0;			/* some node decided to do an exit call */
		splitc_exitCode = two;
	}
}


VOID  SET_PROCS_handler(int proc, int two, int three, int four)
{
	LOG_HOST_CALL("SET_PROCS", proc, two, three, four);

	splitc_partition_size = two;
	if (finishedCount >= splitc_partition_size)
		splitc_exitCode = splitc_notDone = 0;
}

#endif /* ifndef CMMD */


/*						!!!!!!!	READ THIS !!!!!!!
 ***************************************************************************
 * DEFAULT_HOST_HANDLERS_NUM is defined in split-c/print.h and needs to be
 * changed to reflect the length of defaultHostFunctions.
 ***************************************************************************
 */

int hostFunctionCount=USER_HANDLER_BASE;

#ifndef CMMD
	
Handler defaultHostFunctions[] =
{
	FOPEN_handler,
	FCLOSE_handler,
	UNGETC_handler,
	GETINT_handler,
	GETFLOAT_handler,
	FERROR_handler,
	TERMINATE_handler,
	FFLUSH_handler,
	PUTC_handler,
	GETC_handler,
	FGETS_handler,
	INIT_IO_handler,
	FPRINT_handler,
	PRINT_handler,
	GETARGS_handler,
	EXIT_handler,
	FSEEK_handler,
	FREAD_handler,
	FREAD2_handler,
	SET_PROCS_handler,
    GLOB_FOPEN_handler,

#ifdef SC_PROFILE		/* Add a handler for profiling */
	,PROF_READY_handler
#endif
	0, 0, 0, 0               /* reserve 4 slots for X handlers */
};

#else /* CMMD */
	
Handler defaultHostFunctions[] =
{
	EXIT_handler,
	SET_PROCS_handler,
	0, 0, 0, 0               /* reserve 4 slots for X handlers */
};

#endif /* CMMD */

/*						!!!!!!!	READ THIS !!!!!!!
 ***************************************************************************
 * DEFAULT_HOST_HANDLERS_NUM is defined in split-c/print.h and needs to be
 * changed to reflect the length of defaultHostFunctions.
 ***************************************************************************
 */

/* hostFunctions must contain the address of the start of a handler table.  The
 * table must include the contents of defaultHostFunctions.
 */

Handler *hostFunctions = NULL;

/* add a set of new functions to the hostFunctions table returning the base
 * offset that will invoke the handler.
 */

int splitc_addHandlers(int num, Handler *newHandlers)
{
	int base = hostFunctionCount;
	
	if (hostFunctions)
		hostFunctions = realloc(hostFunctions,
								sizeof(Handler) * (num+hostFunctionCount));
	else
	{
		hostFunctions = malloc(sizeof(Handler)*(hostFunctionCount+num));
		memcpy(hostFunctions,
			   defaultHostFunctions,
			   hostFunctionCount*sizeof(Handler));
	}

	memcpy(hostFunctions+hostFunctionCount,
		   newHandlers,
		   num * sizeof(Handler));
	

	hostFunctionCount += num;

	return base;
}



#ifndef CMMD

/*************************************/
/* Host abstain from combine network */
/*************************************/

void splitc_abstain_com(void)
{
	*(volatile int *)(NI_BASE+NI_COM_CONTROL_O) = 0x3;
}

/*********************************/
/* default way to start up nodes */
/*********************************/

/* The debugger requires the ability to break the node programs after
   they have been created but before they have entered splitc_main.
   The only apparent way to do this is to stop them after they have
   entered the main loop (hidden in old versions of CMMD, in the
   node 'main' code in newer versions).
*/

static void _scd_node_prestart(int argc, char**argv)
{
	if (hostFunctions == NULL) hostFunctions = defaultHostFunctions;
}

void _scd_node_start(int argc, char**argv)
{
	SPLITC_MAIN((int)hostFunctions, argc, splitc_packageArgs(argc, argv));
}

#else 


/*********************************/
/* default way to start up nodes */
/*********************************/

/* The debugger requires the ability to break the node programs after
   they have been created but before they have entered splitc_main.
   The only apparent way to do this is to stop them after they have
   entered the main loop (hidden in old versions of CMMD, in the
   node 'main' code in newer versions).
*/

static void _scd_node_prestart(int argc, char**argv)
{

  if (hostFunctions == NULL) hostFunctions = defaultHostFunctions;

  CMMD_enable();
}

void _scd_node_start(int argc, char**argv)
{
  splitc_partition_size = CMMD_partition_size();
  /* get the PE handler address from Proc 0 */
  CMMD_receive_block(0, CMMD_ANY_TAG, &CMPE_printGetChannel,
                     sizeof(Handler));
  
  /* broadcast the number of host handlers to all nodes */ 
  CMMD_bc_from_host(&hostFunctionCount, sizeof(int));

  /* broadcast the host handler addresses to all nodes */ 
  CMMD_bc_from_host(hostFunctions, sizeof(Handler)*hostFunctionCount); 
}

#endif /* CMMD */

void splitc_startnodes(int argc, char**argv)
{
	_scd_node_prestart(argc, argv);
	_scd_node_start(argc, argv);
}
