/*									tab:4
 *
 * main.sc - PE side of general library support for split-c programs
 *
 * Copyright (c) 1992 by Seth Copen Goldstein and 
 * Regents of the University of California
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  Copyright holder(s) make no
 * representation about the suitability of this software for
 * any purpose. It is provided "as is" without express or
 * implied warranty.
 *
 * Author: 			Seth C Goldstein
 * Version:			131
 * Creation Date:	Wed Sep 23 22:08:17 1992
 * Filename:		main.sc
 * History:
 *	LTL	124	Wed Nov 10 21:16:20 1993
 *		Fixed bug in sending the address of CMPE_printGetChannel
 *		to host
 *	LTL	107	Mon Oct 25 17:02:35 1993
 *		Added barrier in fset_io_mode and io_mode
 *	LTL	105	Mon Oct 25 14:00:14 1993
 *		Changed stdin from CMMD_sync_bc mode to
 *		CMMD_indepedent mode in CMMD version
 *	LTL	102	Mon Aug 30 14:20:35 1993
 *		Added send_extra_to_host()
 *	LTL	100	Mon Aug 16 15:43:55 1993
 *		Fixed it up for version 112 to support CMMD.
 *		Most of the changes are in host handlers, I/O and
 *		main().
 *	AK	92	Tue Jun 22 17:53:51 1993
 *		Modified fgets to return NULL.
 *	ACD	87	Mon Jun 14 10:11:21 1993
 *		Removed sbrk_handler interface
 *	SL	86	Tue May 25 12:55:50 1993
 *		Added basic X interface
 *  SL  85  Wed May 12 23:06:32 PDT 1993
 *      Added SC_PROFILE option
 * 	TvE	84	Mon Apr 12 09:06:26 PDT 1993
 *		Fixed Madison bug in fgets: didn't reliably close the segment
 * 	TvE	83	Mon Apr 12 08:57:13 PDT 1993
 *		Cleaned-up FLOAT bug fix from Madison.
 *	AK	81	Tue Mar 16 10:48:51 1993
 *		Fixed bug report from Madison. Have to check with Seth to
 *		make sure.
 *	SCG	78	Fri Mar  5 17:35:39 1993
 *		made printf buffer bigger
 *	SCG	75	Fri Jan 22 13:10:21 1993
 *		added call for users to send to the host, send_to_host()
 *	SL	73	Wed Jan 20 12:48:40 1993
 *		Added call to start CMNA timer #0 for timer library
 *	SL	72	Wed Jan 20 10:31:44 1993
 *		Added a call to initialize atomic and rpc operations
 * 	TvE	71	Tue Jan 19 00:08:56 PST 1993
 *		Updated initializing of ___PROCS and CO. into split-c.sc
 *		Added a call to all_set_PROCS to get partition size checked
 *		against compiled-in parameters.
 *	SCG	60	Mon Dec  7 16:02:36 1992
 *		added argc, argv support for splitc_main
 *	SCG	48	Wed Dec  2 15:41:57 1992
 *		fixed some bugs in fprintf.  Added fread and fseek
 *	SCG	44	Thu Oct  1 10:49:25 1992
 *		added GMALLOC_SBRK handler
 */
static char _version_string_[] = "\nVersion:124:pe.c\0\n";

typedef int Segment;
typedef int Channel;
#define myProc MYPROC

#include <split-c/ansi-c.h>
#include <errno.h>

#include <split-c/split-c.h>
#include "cm5.h"

#ifdef SC_PROFILE /* All processors ready to profile flag */
static int all_proc_profile = 0;
#endif


/* define the host handlers */

static Handler *host_handler;


#ifndef CMMD

#define	FOPEN		(host_handler+0)
#define	FCLOSE		(host_handler+1)
#define	UNGETC		(host_handler+2)
#define	GETINT		(host_handler+3)
#define	GETFLOAT	(host_handler+4)
#define	FERROR		(host_handler+5)
#define	TERMINATE	(host_handler+6)
#define	FFLUSH		(host_handler+7)
#define	PUTC		(host_handler+8)
#define	GETC		(host_handler+9)
#define	FGETS		(host_handler+10)
#define	INIT_IO		(host_handler+11)
#define	FPRINT		(host_handler+12)
#define	PRINT		(host_handler+13)
#define	GETARGS		(host_handler+14)
#define	EXIT		(host_handler+15)
#define FSEEK			(host_handler+16)
#define FREAD			(host_handler+17)
#define FREAD2			(host_handler+18)
#define SET_PROCS		(host_handler+19)
#define GLOB_FOPEN            (host_handler+20)

#ifdef SC_PROFILE /* Extra handler for profiling */
#define PROF_READY		(host_handler+25)
#endif

#else /* CMMD */

#define	EXIT		(host_handler+0)
#define SET_PROCS	(host_handler+1)

#endif /* CMMD */

#define LARGEST_PRINT_STRING	4096+8 /* the +8 is so I can guarantee that it
					* is on a doubleword boundary
					*/

extern void spread_malloc_init();
extern void atomic_init();

#ifdef SC_PROFILE  /* Profile init and printout */
extern void profile_init();
extern char *prof_filename;
#endif



VOID am_broadcast();

VOID CMPE_printGetChannel(chan, dest)
Channel chan;
Channel *dest;
{
	*dest = chan;
}

/*******************************************************/
/* I/O stuff                                           */
/*******************************************************/

#ifndef CMMD

/*******************************************************/
/* printing on nodes requieres communicating with host */
/*******************************************************/

#undef printf


VOID localPrintf(s, VarArgBase)
char *s;
VarArgBaseDecl;
{
	VarArgDecl;
	char buffer[LARGEST_PRINT_STRING];
	char *bp;
	int l;
	volatile Channel chan;
	
	bp = (char *)((((unsigned )(buffer)) + 15) & 0xfffffff8);
	
	VarArgInit(s);
	vsprintf(bp, s, VarArg);
	if ((l = strlen(bp)) > 0)
	{
		l = l + 1;
		chan = -2;
		am_sendToHost(PRINT, l, &chan, CMPE_printGetChannel);
		while (chan == -2) am_poll();
		am_xferBuffer2Host(chan, bp, l);
		while (chan != -3)  {am_poll();}
	}
	VarArgEnd();
}

VOID DPRINTF(s, VarArgBase)
char *s;
VarArgBaseDecl;
{
	VarArgDecl;
	char buffer[256];
	char *bp;
	
	bp = (char *)((((unsigned )(buffer)) + 15) & 0xfffffff8);
	
	VarArgInit(s);
	vsprintf(bp, s, VarArg);
#undef printf
	printf("%s", bp);
	VarArgEnd();
}

/*************************************************************/
/* generic wait for host to increment the flag - waitForHost */
/*************************************************************/

static int waitForHost = 0;
static waitForInitCompleted = 0;
static volatile int hostRetValue;

VOID CMPE_incrFlag(adr)
int *adr;
{
	(*adr)++;
}

VOID CMPE_retValue(adr, val)
int *adr;
int val;
{
        hostRetValue = val;
        (*adr)++;
}


/******************************/
/* fflush a file opened on PM */
/******************************/

int localFflush(handle)
FILE *handle;
{
	int waitHere = 0;
	
	am_sendToHost(FFLUSH, (int )handle, &waitHere, 0);
	am_wait(&waitHere, 1); 
	return hostRetValue;
}

/****************************************************************/
/* fgets into a buffer from the PM, block user until it returns */
/****************************************************************/

static int fgetsDone(info, base)
int info;
int *base;
{
	waitForHost++;
	return 0;
}


static VOID fgetsTerminateChannel(chan, len)
Channel chan;
int len; /* extra length to be shortened */
{
	hostRetValue = chan;
	if (len == -1) {
		amShortenSegment(chan, amQuerySegment(chan));
		hostRetValue = (int) NULL;
	}
	else
		amShortenSegment(chan, len);
}

char *fgets(buffer, len, handle)
char *buffer;
int len;
FILE *handle;
{
	Channel chan;
	char *newBuff;
	
	newBuff = malloc(len);
	chan = am_open_segment(newBuff, len, fgetsDone, (char * )buffer);
	waitForHost = 0;
	am_sendExtraToHost(FGETS, len, chan, handle, fgetsTerminateChannel);
	am_wait(&waitForHost, 1);
	strcpy(buffer, newBuff);
	free(newBuff);
	if (hostRetValue == (int) NULL)
		return (char *)hostRetValue;
	else
		return buffer;
}

char *all_fgets(buffer, len, handle)
___V_SPTR buffer;
int len;
FILE *handle;
{
	char *retVal;

	on_one {
		retVal = fgets(tolocal(buffer), len, handle);
	}
	all_bcast_buffer(buffer, len);
	retVal = (char *)all_bcast_i((int)retVal);
	return retVal;
}

/***************************************************/
/* initIoSystem - gets stdin, stdout, stderr, etc. */
/***************************************************/

FILE *stdin;
FILE *stdout;
FILE *stderr;

static VOID getStdIo(one, two ,three)
FILE *one;
FILE *two;
FILE *three;
{
	stdin = one;
	stdout = two;
	stderr = three;
	waitForHost = 1;
	waitForInitCompleted++;
}

static VOID initIoSystem()
{
	am_sendToHost(INIT_IO, getStdIo, 0, 0);
	am_wait(&waitForHost, 1);
	am_broadcast(getStdIo, stdin, stdout, stderr);
}

/***************************************************/
/* terminate this node.  It has nothing else to do */
/***************************************************/

VOID am_terminate()
{
	am_sendToHost(TERMINATE, 0, 0, 0);
	while(1) am_poll();
}

/***************/
/* putc & getc */
/***************/

int localFputc(c, stream)
char c;
FILE *stream;
{
	int waitHere = 0;
	
	am_sendToHost(PUTC, stream, c, &waitHere);
	am_wait(&waitHere, 1);
	return 0;
}

VOID CMPE_getcResponse(c, dest, flag)
int c;
int *dest;
int *flag;
{
	*dest = c;
	*flag = 1;
}

int localFgetc(stream)
FILE *stream;
{
	int fgetcChar;
	int done = 0;
	
	am_sendToHost(GETC, stream, &fgetcChar, &done);
	am_wait(&done, 1);
	return fgetcChar;
}


static void fputsHost(bp, stream)
char *bp;
FILE *stream;
{
	int l;
	volatile Channel chan;
	
	if ((l = strlen(bp)) > 0)
	{
		l = l + 1;
		chan = -2;
		am_sendToHost(FPRINT, l, stream, &chan);
		while (chan == -2) am_poll();
		am_xferBuffer2Host(chan, (int *)bp, l);
		while (chan != -3) {am_poll();}
	}
}

int fprintf(stream, s, VarArgBase)
FILE *stream;
const char *s;
VarArgBaseDecl;
{
	VarArgDecl;
	char buffer[LARGEST_PRINT_STRING];
	char *bp;
	int l;
	
	bp = (char *)((((unsigned )(buffer)) + 15) & 0xfffffff8);
	
	VarArgInit(s);
	vsprintf(bp, s, VarArg);
	fputsHost(bp, stream);
	VarArgEnd();
}

VOID vfprintf(stream, s, VarArgBase)
FILE *stream;
char *s;
VarArgBaseDecl;
{
	VarArgDecl;
	char buffer[LARGEST_PRINT_STRING];
	char *bp;
	
	VarArgInit(s);
	bp = (char *)((((unsigned )(buffer)) + 15) & 0xfffffff8);
	vsprintf(bp, s, VarArg);
	fputsHost(bp, stream);
	VarArgEnd();
}

/* Support for scanf-related functions begins here. */

#include <values.h>

#define NCHARS	(1 << BITSPERBYTE)

extern double atof();
extern char *memset();
extern int ungetc();

static unsigned char *
    setup(fmt, tab)
register unsigned char *fmt;
register char *tab;
{
    register int b, c, d, t = 0;
    
    if(*fmt == '^') {
	t++;
	fmt++;
    }
    (void)memset(tab, !t, NCHARS);
    if((c = *fmt) == ']' || c == '-') { /* first char is special */
	tab[c] = t;
	fmt++;
    }
    while((c = *fmt++) != ']') {
	if(c == '\0')
	    return(NULL); /* unexpected end of format */
	if(c == '-' && (d = *fmt) != ']' && (b = fmt[-2]) < d) {
	    (void)memset(&tab[b], t, d - b + 1);
	    fmt++;
	} else
	    tab[c] = t;
    }
    return(fmt);
}

static void
    store_int(value,length,arg_list)
long 	value;
int	length;
va_list	*arg_list ;
{
    switch(length) {
      case 'l': *va_arg(*arg_list, long *)  = value ;
	break ;
      case 'h': *va_arg(*arg_list, short *) = (short)value ;
	break ;
      default:  *va_arg(*arg_list, int *)   = (int)value ;
	break ;
    }
}

static int
    number(stow, type, len, size, iop, chars_scanned, listp)
int stow, type, len, size;
register FILE *iop;
register int *chars_scanned ;
va_list *listp;
{
    char numbuf[64];
    register char *np = numbuf;
    register int c, base;
    int	digitseen = 0,
    dotseen = 0,
    expseen = 0,
    floater = 0,
    negflg = 0;
    long lcval = 0;
    extern char _lc_radix;	/* INTL. the international radix character */
    
    switch(type) {
      case 'e':
      case 'f':
      case 'g':
	floater++;
	/* FALLTHROUGH */
      case 'd':
      case 'u':
	base = 10;
	break;
      case 'o':
	base = 8;
	break;
      case 'x':
	base = 16;
	break;
      case 'i':
	base = 0;	/* General integer */
	break ;
      default:
	return(0); /* unrecognized conversion character */
    }
    
    switch(c = getc(iop)) {
      case '-':
	negflg++;
      case '+': /* fall-through */
	len--;
	c = getc(iop);
	(*chars_scanned)++ ;
    }
    if (base == 0) {	/* general integer, test format */
	if (c == '0') {
	    c = getc(iop) ;
	    if (c == 'x' || c == 'X') {
		/* 0x or 0X mean hex, get next digit and
		 * mark the 0[Xx] as scanned.
		 */
		base = 16 ;
		len -= 2;
		c = getc(iop);
		*chars_scanned += 2 ;
	    } else if (isdigit(c)) {
		/* 0 followed by a digit is octal,
		 * mark 0 as scanned.
		 */
		base = 8 ;
		len--;
		(*chars_scanned)++ ;
	    } else {
		/* Decimal 0, restart scan at 0.
		 */
		ungetc(c,iop); 
		c = '0';
		base = 10 ;
	    }
	} else
	    base = 10 ;
    } 
    for( ; --len >= 0 && np < &numbuf[63]; *np++ = c, c = getc(iop)) {
	if(isdigit(c) || base == 16 && isxdigit(c)) {
	    int digit = c - (isdigit(c) ? '0' :
			     isupper(c) ? 'A' - 10 : 'a' - 10);
	    if(digit >= base)
		break;
	    if(stow && !floater)
		lcval = base * lcval + digit;
	    digitseen++;
	    (*chars_scanned)++ ;
	    continue;
	}
	if(!floater)
	    break;
	if(c == '.' && !dotseen++) {	/* INTL */
	    (*chars_scanned)++ ;
	    continue;
	}
	if((c == 'e' || c == 'E') && digitseen && !expseen++) {
	    (*chars_scanned)++ ;
	    if (--len < 0 || np >= &numbuf[63])
		break;
	    *np++ = c;
	    c = getc(iop);
	    if(isdigit(c) || c == '+' || c == '-' || isspace(c)) {
		(*chars_scanned)++ ;
		continue;
	    }
	}
	break;
    }
    if(stow && digitseen)
	if(floater) {
	    register double dval;
	    
	    *np = '\0';
	    dval = atof(numbuf);
	    if(negflg)
		dval = -dval;
	    if(size == 'l' || size == 'L')
		*va_arg(*listp, double *) = dval;
	    else
		*va_arg(*listp, float *) = (float)dval;
	} else {
	    /* suppress possible overflow on 2's-comp negation */
	    if(negflg && lcval != HIBITL)
		lcval = -lcval;
	    store_int(lcval,size,listp);
	}
    ungetc(c,iop);
    return(digitseen); /* successful match if non-zero */
}

static int
    string(stow, type, len, tab, iop, chars_scanned, listp)
register int stow, type, len;
register char *tab;
register FILE *iop;
int *chars_scanned;
va_list *listp;
{
    register int ch;
    register char *ptr;
    char *start;
    
    start = ptr = stow ? va_arg(*listp, char *) : NULL;
    if(type == 'c' && len == MAXINT)
	len = 1;
    while((ch = getc(iop)) != EOF &&
	  !(type == 's' && isspace(ch) || type == '[' && tab[ch])) {
	if(stow)
	    *ptr = ch;
	ptr++;
	(*chars_scanned)++;
	if(--len <= 0)
	    break;
    }
    if(ch == EOF || len > 0 && ungetc(ch, iop) == EOF)
	*listp = NULL; /* end of input */
    if(ptr == start)
	return(0); /* no match */
    if(stow && type != 'c')
	*ptr = '\0';
    return(1); /* successful match */
}

static int
    DoScan(iop, fmt, va_alist)
register FILE *iop;
register unsigned char *fmt;
va_list va_alist;
{
    char tab[NCHARS];
    register int ch;
    int 	nmatch = 0,
    len,
    inchar,
    stow,
    size,
    chars_scanned = 0 ;
    
    /*******************************************************
     * Main loop: reads format to determine a pattern,
     *		and then goes to read input stream
     *		in attempt to match the pattern.
     *******************************************************/
    for( ; ; ) {
	if((ch = *fmt++) == '\0')
	    return(nmatch); /* end of format */
	if(isspace(ch)) {
	    while(isspace(inchar = getc(iop)))
		chars_scanned++;
	    if(ungetc(inchar, iop) != EOF)
		continue;
	    break;
	}
	if(ch != '%' || (ch = *fmt++) == '%') {
	    if((inchar = getc(iop)) == ch) {
		chars_scanned++;
		continue;
	    }
	    if(ungetc(inchar, iop) != EOF)
		return(nmatch); /* failed to match input */
	    break;
	}
	if(ch == '*') {
	    stow = 0;
	    ch = *fmt++;
	} else
	    stow = 1;
	
	for(len = 0; isdigit(ch); ch = *fmt++)
	    len = len * 10 + ch - '0';
	if(len == 0)
	    len = MAXINT;
	
	if((size = ch) == 'l' || size == 'h' || size == 'L')
	    ch = *fmt++;
	if(ch == '\0' ||
	   ch == '[' && (fmt = setup(fmt, tab)) == NULL)
	    return(EOF); /* unexpected end of format */
	if(isupper(ch)) { /* no longer documented */
#ifdef _POSIX_SOURCE		  /* 005 */
	    if(ch != 'E' && ch != 'F' && ch != 'G' && ch != 'X')
#endif
		size = 'l';
	    ch = ch - 'A' + 'a';
	}
	if (ch == 'n') {
	    if (stow) {
		store_int((long)chars_scanned,size,&va_alist);
	    }
	    continue;
	}
	if (ch == 'p') { /* pointer - convert to %8x */
	    ch = 'x';
	    len = 8;
	}
	if(ch != 'c' && ch != '[') {
	    while(isspace(inchar = getc(iop)))
		chars_scanned++;
	    if(ungetc(inchar, iop) == EOF)
		break;
	}
	if((size = (ch == 'c' || ch == 's' || ch == '[') ?
	    string(stow,ch,len,tab,iop,&chars_scanned,&va_alist) :
	    number(stow,ch,len,size,iop,&chars_scanned,&va_alist))
	   != 0)
	    nmatch += stow;
	if(va_alist == NULL) /* end of input */
	    break;
	if(size == 0)
	    return(nmatch); /* failed to match input */
    }
    return(nmatch != 0 ? nmatch : EOF); /* end of input */
}

localScanf(fmt, va_alist)
char *fmt;
va_dcl
{
    va_list ap;
    
    va_start(ap);
    return(DoScan(stdin, fmt, ap));
}

localFscanf(fp, fmt, va_alist)
FILE *fp;
char *fmt;
va_dcl
{
    va_list ap;
    
    va_start(ap);
    return(DoScan(fp, fmt, ap));
}

/* end of support routines for scanf-related functions. */

VOID CMPE_channelResponse(chan, dest)
Channel chan;
Channel *dest;
{
	*dest = chan;
}

VOID CMPE_fopenResponse(f, dest, flag, err)
FILE *f;
FILE **dest;
int *flag;
int err;
{
	*dest = f;
	*flag = -2;
	errno = err;
}

FILE *fopen(name, type)
const char *name;
const char *type;
{
	char buffer[512+8];
	char *bp;
	volatile Channel chan;
	int len;
	FILE *result = NULL;
	
	bp = (char *)((((unsigned )(buffer)) + 15) & 0xfffffff8);
	strcpy(bp, name);
	len = strlen(bp)+1;
	strcpy(bp+len, type);
	len += strlen(type);
	
	chan = -2;
	am_sendToHost(FOPEN, len+1, &chan, &result);
	while (chan == -2) am_poll();
	am_xferBuffer2Host(chan, (int *)bp, len+1);
	while (chan != -2) am_poll();
	return result;
}


/* This is used when everyone accesses the same file
   but want the accesses to take place independently. */
static FILE *global_fopen(name, type)
const char *name;
const char *type;
{
        char buffer[512+8];
        char *bp;
        volatile Channel chan;
        int len;
        FILE *result = NULL;

        bp = (char *)((((unsigned )(buffer)) + 15) & 0xfffffff8);
        strcpy(bp, name);
        len = strlen(bp)+1;
        strcpy(bp+len, type);
        len += strlen(type);

        chan = -2;
        am_sendToHost(GLOB_FOPEN, len+1, &chan, &result);
        while (chan == -2) am_poll();
        am_xferBuffer2Host(chan, (int *)bp, len+1);
        while (chan != -2) am_poll();
        return result;
}

FILE *all_fopen(name, type)
const char *name;
const char *type;
{
	FILE *stream;

	on_one {
		stream = fopen(name, type);
	}
	stream = (FILE *)all_bcast_i((int)stream);
	return stream;
}


FILE *all_independent_fopen(name, type)
const char *name;
const char *type;
{
        FILE *stream;

        on_one {
                stream = global_fopen(name, type);
        }
        stream = (FILE *)all_bcast_i((int)stream);
        return stream;

}

CMPE_fcloseResponse(value, dest, flag)
int value;
int *dest;
int *flag;
{
	*dest = value;
	*flag = 1;
}

VOID localFclose(stream)
FILE *stream;
{
	int done = 0;
	int err;
	
	am_sendToHost(FCLOSE, stream, &err, &done);
	am_wait(&done, 1);
	errno = err;
}

void all_fclose(stream)
FILE *stream;
{

	on_one {
		localFclose(stream);
	}
	barrier();
	return;
}

int localFerror(stream)
FILE *stream;
{
	int done = 0;
	int err;
	
	am_sendToHost(FERROR, stream, &err, &done);
	am_wait(&done, 1);
	errno = err;
}

CMPE_ungetcResponse(value, flag)
int value;
int *flag;
{
	errno = value;
	*flag = 1;
}

int localUngetc(d, stream)
int d;
FILE *stream;
{
	int done = 0;
	int err;
	
	am_sendToHost(UNGETC, stream, d, &done);
	am_wait(&done, 1);
	return 1;
}

CMPE_floatResponse(fl, three, four)
double fl;
double *three;
int *four;
{
	*three = fl;
	*four = 1;
}

double getFloat(stream)
FILE *stream;
{
	int done = 0;
	double fl;
	
	am_sendToHost(GETFLOAT, stream, &fl, &done);
	am_wait(&done, 1);
	return fl;
}

CMPE_getintResponse(value, dest, flag, err)
int value;
int *dest;
int *flag;
int err;
{
	*dest = value;
	*flag = 1;
	errno = err;
}


int getInt(stream)
FILE *stream;
{
	int done = 0;
	int i;
	
	am_sendToHost(GETINT, stream, &i, &done);
	am_wait(&done, 1);
	return i;
}

/*********/
/* fseek */
/*********/

int localFseek(stream, offset, ptrname)
FILE *stream;
long offset;
int ptrname;
{
	volatile int result;
	
	result = -2;
	am_sendToHost(FSEEK, stream, (offset<<2)|ptrname, &result);
	while (result == -2) am_poll();
	return result;
}

/*********/
/* fread */
/*********/

static int freadDone(flag, buffer)
int flag;
int *buffer;
{
	*(int *)flag = 1;
	return 0;
}

int localFread(ptr, size, nitems, stream)
char *ptr;
int size;
int nitems;
FILE *stream;
{
	volatile int result;
	Segment sid;
	int len = size * nitems;
	volatile int flag;
	int temp;
	
	sid = am_open_segment(ptr, len, freadDone, (void *)&flag);
	result = 0;
	am_sendToHost(FREAD, stream, sid, &result);
	while (result == 0) am_poll();

	/* send rest of info */

	temp = result;
	result = nitems+1;			/* an impossible return value */
	flag = 0;					/* set to 1 when transfer complete */
	am_sendToHost(FREAD2, temp, size, nitems);
	while (result == (nitems+1)) am_poll();

	/* result now has actual number of items that were read */
	/* this might require a shortening of the segment */

	if (nitems != result)
		amShortenSegment(sid, (nitems - result)*size);

	while (flag == 0) am_poll();
	return result;
}

int all_fread(ptr, size, nitems, stream)
___V_SPTR ptr;
int size;
int nitems;
FILE *stream;
{
	int retVal;

	on_one {
		retVal = localFread(tolocal(ptr), size, nitems, stream);
	}
	all_bcast_buffer(ptr, size*nitems);
	retVal = all_bcast_i(retVal);
	return retVal;
}

#else /* CMMD */

/* keep these functions for backward compatibility 
 * user probably needs to change code because stdin is in sync_bc mode
 * by default now
 */
int getInt(FILE *f)
{
  int result;
 
  fscanf(f, "%d", &result);
  return(result);
}

double getFloat(FILE *f)
{
  double result;
 
  fscanf(f, "%f", &result);
  return(result);
}

/* I/O functions on top of CMMD */

/* all nodes open file in global independent mode
 * return file descriptor
 */
int all_open_independent(char *path, int flags, int mode)
{
  barrier();
  return (CMMD_global_open(path, flags, mode));
}

/* all nodes open file in global synchronous broadcast mode
 * return file descriptor
 */
int all_open_sync_bc(char *path, int flags, int mode)
{
  int fd;

  barrier();
  fd = CMMD_global_open(path, flags, mode);
  CMMD_set_io_mode(fd, CMMD_sync_bc);
  return (fd);
}



/* all nodes open file in global synchronous sequential mode
 * return file descriptor
 */
int all_open_sync_seq(char *path, int flags, int mode)
{
  int fd;

  barrier();
  fd = CMMD_global_open(path, flags, mode);
  CMMD_set_io_mode(fd, CMMD_sync_seq);
  return (fd);
}


/****************/
/* buffered I/O */
/****************/

/* all nodes open file in global independent mode
 * return stream pointer
 */
FILE *all_fopen_independent(char *path, char *mode)
{
  barrier();
  return (CMMD_global_fopen(path, mode));
}


/* all nodes open file in global synchronous broadcast mode 
 * return stream pointer
 */
FILE *all_fopen_sync_bc(char *path, char *mode)
{
  FILE *stream;

  barrier();
  stream = CMMD_global_fopen(path, mode);
  CMMD_fset_io_mode(stream, CMMD_sync_bc);
  return (stream);
}


/* all nodes open file in global synchronous sequential mode 
 * return stream pointer
 */
FILE *all_fopen_sync_seq(char *path, char *mode)
{
  FILE *stream;

  barrier();
  stream = CMMD_global_fopen(path, mode);
  CMMD_fset_io_mode(stream, CMMD_sync_seq);
  return (stream);
}

/* set I/O mode */
int set_io_mode(int fd, int io_mode)
{
  barrier();
  return (CMMD_set_io_mode(fd, io_mode));
}

/* set mode for buffered I/O */
int fset_io_mode(FILE *stream, int io_mode)
{
  barrier();
  return (CMMD_fset_io_mode(stream, io_mode));
}

/* return I/O mode */
int get_io_mode(int fd)
{
  return (CMMD_get_io_mode(fd));
}

/* return mode for buffered I/O */
int fget_io_mode(FILE *stream)
{
  return (CMMD_fget_io_mode(stream));
}

#endif /* CMMD */

VOID localExit(flag)
int flag;
{
	am_sendToHost(EXIT, flag, 1, 0);
	while (1) am_poll();
}

/* set_procs on host */

_set_procs_on_host(int newval)
{
	am_sendToHost(SET_PROCS, newval, 0, 0);
}




/*******************************/
/* broadcast an active message */
/*******************************/

VOID am_upRightBroadcast(Handler func, int d1, int d2, int d3);
VOID am_upLeftBroadcast(Handler func, int d1, int d2, int d3);

VOID am_downBroadcast(func, d1, d2, d3)
Handler func;
int d1, d2, d3;
{
	int i = (2 * myProc) + 1;
	
	if (PROCS > i)
		am_request4(i++, am_downBroadcast, func, d1, d2, d3);
	if (PROCS > i)
		am_request4(i, am_downBroadcast, func, d1, d2, d3);
	(*func)(d1, d2, d3);
}

VOID am_upBroadcast(func, d1, d2, d3)
Handler func;
int d1, d2, d3;
{
	int parent = ((myProc-1) >> 1);

	if (myProc & 0x1)
		am_request4(parent, am_upRightBroadcast, func, d1, d2, d3);
	else
		am_request4(parent, am_upLeftBroadcast, func, d1, d2, d3);
}

VOID am_upRightBroadcast(func, d1, d2, d3)
Handler func;
int d1, d2, d3;
{
	if (myProc)
		am_upBroadcast(func, d1, d2, d3);
	am_request4(2 * myProc + 2, am_downBroadcast, func, d1, d2, d3);
	(*func)(d1, d2, d3);
}

VOID am_upLeftBroadcast(func, d1, d2, d3)
Handler func;
int d1, d2, d3;
{
	/* got this from my right child.  SO send to parent and left child */
	if (myProc)
		am_upBroadcast(func, d1, d2, d3);
	am_request4(2 * myProc + 1, am_downBroadcast, func, d1, d2, d3);
	(*func)(d1, d2, d3);
}

VOID am_broadcast(func, d1, d2, d3)
Handler func;
int d1, d2, d3;
{
	if (myProc != 0)
	{
		am_upBroadcast(func, d1, d2, d3);
		am_downBroadcast(func, d1, d2, d3);
		/*am_request4(0, am_downBroadcast, func, d1, d2, d3);*/
	}
	else
		am_downBroadcast(func, d1, d2, d3);
}

#ifndef CMMD
/****************************************/
/* lets get the arguments from the host */
/****************************************/

static int argCount;
static char **argVector;
static int argVectorLen;
static char *argVectorData;
static int argsRcvd=0;

int gotArgs(int *buffer, int flag)
{
	argsRcvd=1;
	return 0;
}

void setupArgs(int argc, int size)
{
	argVectorLen = size + (sizeof(int) * argc);
	argVectorData = (char *)malloc(argVectorLen);
	argVector = (char **)malloc(argc*sizeof(char **));
	argCount = argc;
	if (am_open_this_segment(0,
							   argVectorData,
							   argVectorLen,
							   gotArgs,
							   argVectorLen) == -1)
	{
		printf("failure to open segment 0\n");
		localExit(-1);
	}
}

void getArguments()
{
	am_sendToHost(GETARGS, 0, 0, 0);
}	

void all_bcastArgs()
{
	int dest = (MYPROC*2) + 1;
	int i;
	int offset = sizeof(int) * argCount;
	
	/* get and send the data */
		
	am_wait(&argsRcvd, 1);

	if (dest < PROCS)
		am_xferBuffer(dest, 0, argVectorData, argVectorLen);
	if (++dest < PROCS)
		am_xferBuffer(dest, 0, argVectorData, argVectorLen);
		
	/* process the data */

	for (i=0; i<argCount; i++)
		argVector[i] = argVectorData + offset + ((int *)argVectorData)[i];
}		

#endif /* CMMD */

/**********************************************/
/* external interface for sending to the host */
/**********************************************/

void send_to_host(int offset, int arg1, int arg2, int arg3)
{
	am_sendToHost(host_handler+offset, arg1, arg2, arg3);
}

/* arg1 is a 16-bit value */
void send_extra_to_host(int offset, int arg1, int arg2, int arg3, int arg4)
{
	am_sendExtraToHost(host_handler+offset, arg1, arg2, arg3, arg4);
}

/*****************************************************************
 * here we have the main used on each PE.  It initializes i/o, malloc, etc. and
 * then it calls splitc_main()
 *****************************************************************/

#ifndef CMMD

CMPE_SPLITC_MAIN(handler, argc, argSize)
CMAM_handler **handler;
int argc;
int argSize;
{
	extern int getLog(int);
	(int)___MYPROC = CMNA_self_address; /* get this done now */
	(int)___PROCS = CMNA_partition_size; /* until we know better */
	(int)___PROCSM1 = CMNA_partition_size - 1;
	(int)___PROCSLOG = getLog(CMNA_partition_size);

	host_handler = handler;
	setupArgs(argc, argSize);
	if (myProc == 0)
	{
		getArguments();
		initIoSystem();
	}
	else
		am_wait(&waitForInitCompleted, 1);
	all_bcastArgs();

	all_set_PROCS(CMNA_partition_size);

	spread_malloc_init();

	atomic_init();

	CMNA_timer_clear(0);
	CMNA_timer_start(0);

#ifdef SC_PROFILE		/* Profiling version of library */
	{					/* Process profile filename option */
		int i, j;

		for (i = 1; i < argc - 1; i++)
			if (!strcmp(argVector[i], "-P"))
				break;
		if (i < argc - 1) {
			prof_filename = argVector[i + 1];
			for (j = i; j < argc - 2; j++)
				argVector[j] = argVector[j + 2];
			argc -= 2;
		}
		profile_init(); /* Uses get_seconds() */
	}
#endif

	/* all initialization has been done, so lets go! */

	splitc_main(argc, argVector);

#ifdef SC_PROFILE		/* Profiling version of library */
	am_sendToHost(PROF_READY, (int)&all_proc_profile, 0, 0);
	am_wait(&all_proc_profile, 1);
	all_do_profile("program complete");
#endif
	barrier();
}

#else

static Handler *hostHandlerTable = NULL;
static int hostHandlerCount = 0;

main(int argc, char **argv)
{

	/* !!!!!! CAUTION !!!!!!
         * Not too safe to do any I/O from here 
         */
	Handler node_handler;
	extern int getLog(int);
	(int)___MYPROC = CMMD_self_address(); /* get this done now */
	(int)___PROCS = CMMD_partition_size(); /* until we know better */
	(int)___PROCSM1 = CMMD_partition_size() - 1;
	(int)___PROCSLOG = getLog(CMMD_partition_size());

	CMMD_enable_host();

	CMAML_disable_interrupts();
	CMMD_fset_io_mode(stdin, CMMD_independent);
	CMMD_fset_io_mode(stdout, CMMD_independent);
	CMMD_fset_io_mode(stderr, CMMD_independent);


	/* send the PE handler address to host; this address should be
		* the same on all nodes
			*/
	if (CMMD_self_address() == 0) {
		node_handler = CMPE_printGetChannel;	 	  
		CMMD_send_block(CMMD_host_node(), CMMD_DEFAULT_TAG, 
						&node_handler, sizeof(Handler));
	}

	barrier();

	/* get the host handler addresses from the host 
		* we need to do this because there is no indirect active 
			* message in CMMD 3.0
				*/
	CMMD_receive_bc_from_host(&hostHandlerCount, sizeof(int));
	hostHandlerTable = malloc(sizeof(Handler)*hostHandlerCount);
	if (hostHandlerTable == 0) {
		fprintf(stderr, "Fail to allocate memory for hostHandlerTable\n");
		localExit(-1);
	}  

	CMMD_receive_bc_from_host(hostHandlerTable, 
							  sizeof(Handler)*hostHandlerCount);
	host_handler = hostHandlerTable;


	/* !!!!!! CAUTION !!!!!!
         * Not too safe to do any I/O before here 
         */


	all_set_PROCS(CMMD_partition_size());

	spread_malloc_init();

	atomic_init();

	CMNA_timer_clear(0);
	CMNA_timer_start(0);
 
	/* all initialization has been done, so lets go! */

	splitc_main(argc, argv);

	fflush(stdout);
	fflush(stderr);

	am_sendToHost(EXIT, 0, 0, 0);
	while (1) am_poll();

}

#endif /* CMMD */
