/*									tab:8
 *
 * malloc_cp.c - Host code for gmalloc
 *
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 		Andrea Dusseau
 * Version:		5
 * Creation Date:	??
 * Filename:		malloc_cp.c
 * History:
 *	ssl	5	Thu Oct 28 22:57:43 1993
 *		Include am.h instead of cmam.h.
 *      Call am_reply4 instead of CMAM_reply_4.
 * 	TvE	4	Sat Apr  3 22:52:10 PST 1993
 *		Changed include <cmam/cmam.h> to "cmam.h"
 * 	TvE	3	Mon Jan  4 21:52:55 PST 1993
 *		Adapted to changes in #@$%^& CMOST: CM_sbrk now returns the
 *		new value of the break, i.e. the *top* of the allocated
 *		memory! The documentation says we should now use
 *		CM_set_iu_heap_ptr, but that's ill documented...
 * 	TvE	2	Thu Dec 24 03:26:57 PST 1992
 *		moved to malloc_cp.c
 *	AK	1	Sat Nov 28 21:14:49 1992
 *		changed include file name from cmam-2-4
 *	SCG	1	Thu Oct  1 10:55:54 1992
 *		made sbrk_handler handle indirect calls
 * 	TvE	1	Tue Sep 29 12:47:46 PDT 1992
 *		changed include file names
 */

#include <stdio.h>
#include "am.h"
#include <cmsys/cm_types.h>
#include "malloc_int.h"

void sbrk_handler(void *hinfo, int proc, int size,
		  gmalloc_t *info, void (*gmalloc_reply_handler)())
{
	CM_memaddr_t heapptr;

#if 0
	fprintf(stderr, "Allocating %d bytes on parallel heap. From 0x%x\n",
		size, CM_sbrk(0));
#endif
	heapptr = CM_sbrk(0);
	if(CM_sbrk(size) == -1) {
		fprintf(stderr,
		    "ERROR: Could not allocate %u bytes of parallel memory.\n",
		    size);
		am_reply4(proc, gmalloc_reply_handler, info, ERROR, NULL, 0);
		return;
	}
#if 0
	fprintf(stderr, "Allocated %d bytes on parallel heap. Got 0x%x\n",
		size, heapptr);
#endif

	am_reply4(proc, gmalloc_reply_handler,
		     info, ALLOCATED, (void *)heapptr, 0);
}
