/*									tab:4
 *
 * malloc.sc - Source code for the "{all_}spread_malloc" library procedures.
 *
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 			Andrea Dusseau
 * Version:			11
 * Creation Date:	??
 * Filename:		malloc.sc
 * History:
 *	ssl	11	Mon Dec 13 22:15:02 1993
 *		Changed include directory for lock.h
 *	ssl	10	Wed Nov 17 23:26:39 1993
 *		Removed global and spread tags. Added explicit writes to allow
 *      compilation with native compiler.
 *	LTL	8	Mon Aug 16 15:48:01 1993
 *		Renamed am_* macros for version 112 to support CMMD
 *	ACD	6	Mon Jun 21 13:11:35 1993
 *		Removed calls to host for CM_sbrk; Instead call CMPE_set_iu_heap_ptr
 *	AK	5	Sat Apr  3 15:14:09 1993
 *		Changed all_bcast call to all_bcast_i
 * 	TvE	4	Thu Dec 24 03:24:58 PST 1992
 *		moved to malloc.sc
 *	SCG	3	Wed Dec 16 21:16:13 1992
 *		changed PROCS to MAX_PROCS.  Needs to be done in init call
 *	SCG	2	Thu Oct  1 10:54:18 1992
 *		changed CMAM_4 into an indirect for host sbrk call
 * 	TvE	1	Tue Sep 29 12:47:46 PDT 1992
 *		changed include file names
 */

#include <stdio.h>
#include <split-c/split-c.h>
#include "malloc_int.h"
#include "lock.h"
#include <cmsys/cm_memory.h>

/*
 * Info for binned allocation.  See malloc_int.h for details.
 * These are only valid on SPREAD_MALLOC_PROC.
 */
Address		memFreeLists[BIN_BUCKETS];

/*
 * Info for large-block allocator.  See malloc_int.h for details.
 * These are only valid on SPREAD_MALLOC_PROC.
 */
Address         memFirst;
Address         memLast;
Address		memCurrentPtr;
int		memInitBegin = 0;
int		memInitialized = 0;
int             memLargestFree = 0;
int             memBytesFreed = 0;

/*
 * Info specific to global memory allocation
 */
int proc_save;
unsigned bytesNeeded_save;
int index_save;
int regionSize_save;

lock_struct(int, int_locked);
int_locked MemExclFlag;
#ifdef NO_SPLITCC
#define ___ILOCK_GPTR ___globPtrType
#else
#define ___ILOCK_GPTR int_locked *global
#endif
___ILOCK_GPTR MemExclPtr;

/* #define DEBUG */



/*
 * ----------------------------------------------------------------------------
 *
 * all_spread_malloc --
 *	This procedure is called by all processors to allocate a spread array
 *      dynamically.
 *
 * Inputs:
 *      num_elems -- the total number of elements to be allocated cyclicly 
 *                   across the processors
 *      bytes_per_elem -- the size (in bytes) of each element to be allocated
 *
 * Results:
 *	The return value is a spread pointer to an area of at least 
 *      "bytes_per_elem * num_elems" bytes of free storage.  
 *
 * Side effects:
 *	The returned block is marked as allocated within SPREAD_MALLOC_PROC's memory
 *      space and will not be allocated to anyone else until ANY processor 
 *      deallocates the space with a call to {all_}spread_free.
 *
 * ----------------------------------------------------------------------------
 */
___V_SPTR all_spread_malloc(unsigned num_elems, unsigned bytes_per_elem)
{
  int g_mallocPtr = NULL;

  am_barrier(1);
  if (!bytes_per_elem || !num_elems)
	  return tospread(0, 0);
  
  /* 
   * NOTE: should do check that all processors have requested the same
   * number of bytes. . . but I currently don't
   */

  if (MYPROC == SPREAD_MALLOC_PROC) {
    g_mallocPtr = (int)sc_malloc(num_elems, bytes_per_elem);
  }

  g_mallocPtr = all_bcast_i(g_mallocPtr); 
  return tospread(0, (void *)g_mallocPtr);
}



/*
 * ----------------------------------------------------------------------------
 *
 * spread_malloc --
 *	This procedure is called by one processor to allocate spread arrays
 *      dynamically.  
 *
 * Inputs:
 *      num_elems -- the total number of elements to be allocated cyclicly 
 *                   across the processors
 *      bytes_per_elem -- the size (in bytes) of each element to be allocated
 *
 * Results:
 *	The return value is a spread pointer to an area of at least 
 *      "bytes_per_elem * num_elems" bytes of free storage.  
 *
 * Side effects:
 *	The returned block is marked as allocated within SPREAD_MALLOC_PROC's memory
 *      space and will not be allocated to anyone else until ANY processor 
 *      deallocates the space with a call to {all_}spread_free.
 *
 * ----------------------------------------------------------------------------
 */
___V_SPTR spread_malloc(unsigned num_elems, unsigned bytes_per_elem)
{
  void *g_mallocPtr = NULL;

  if (!bytes_per_elem || !num_elems)
	  return tospread(0, 0);

  LockGet_Size(MemExclPtr, WRITER_LOCK, sizeof(int_locked));
  g_mallocPtr = sc_malloc(num_elems, bytes_per_elem);
  LockRelease_Size(MemExclPtr, WRITER_LOCK, sizeof(int_locked));
  return tospread(0, g_mallocPtr);
}

void *sc_malloc(unsigned num_elems, unsigned bytes_per_elem)
{
  gmalloc_t info;
  unsigned bytes_per_proc;
  ___I_GPTR memInitBeginPtr;

  info.flag = 0;
  bytes_per_proc = MAX_BLOCKS(num_elems) * bytes_per_elem;
  bytes_per_proc = BYTES_TO_BLOCKSIZE(bytes_per_proc);
  if (bytes_per_proc <= BIN_SIZE)  {
     am_request4(SPREAD_MALLOC_PROC, gmalloc_bin_handler, 
           MYPROC, bytes_per_proc, 
	   BLOCKSIZE_TO_INDEX(bytes_per_proc), &info);
    am_wait(&info.flag, 1);
    switch(info.result) {
    case ALLOCATED: 
      return info.ptr;
    case NEEDSBRK:
      info.flag = 0;
      am_request3(SPREAD_MALLOC_PROC, morebinmem_handler, MYPROC, 
	     MemChunkAlloc(info.sizetobrk), &info);
      am_wait(&info.flag, 1);
      return info.ptr;
    }
  }

  else {
    memInitBeginPtr = toglobal(0, &memInitBegin);
    if (!__i_read(memInitBeginPtr)) {
      __i_write(memInitBeginPtr, 1);
	  __i_write(toglobal(0, &memFirst), (int)MemChunkAlloc(MIN_REGION_SIZE));
#ifdef DEBUG
      printf("%d: Initialized large-allocator memory reserves.\n", MYPROC);
#endif
    }

    am_request3(SPREAD_MALLOC_PROC, gmalloc_big_handler, 
                MYPROC, bytes_per_proc, &info);
    am_wait(&info.flag, 1);
    switch(info.result) {
      case ALLOCATED: 
	return info.ptr;
      case NEEDSBRK:
	info.flag = 0;
	am_request3(SPREAD_MALLOC_PROC, morebigmem_handler, info.sizetobrk,
	       MemChunkAlloc(info.sizetobrk), &info);
	am_wait(&info.flag, 1);
	if (info.result != ALLOCATED) {
	  fprintf(stderr, "%d INTERNAL ERROR: Not enough memory to satisfy request after sbrk.\n", MYPROC);
	  exit(-1);
	}
	return info.ptr;
      }
  }
}



/*
 * ----------------------------------------------------------------------------
 *
 * spread_free --
 *      Return a previously-allocated block of parallel storage to the 
 *      free pool.
 *
 *      Any processor may free the memory, since all the state is held on
 *      the processor in charge of allocating spread memory. 
 *
 * Inputs: 
 *      blockPtr -- spread pointer to beginning of block allocated with
 *                  {all_}spread_malloc.
 *
 * Side effects:
 *      The storage pointed to by blockPtr is marked as free and returned
 *	to the free pool.  Nothing in the bytes pointed to by blockPtr is
 *	modified at this time:  no change will occur until at least the
 *	next call to {all_}spread_malloc.
 *
 * ----------------------------------------------------------------------------
 */
void spread_free(___V_SPTR blockPtr) 
{
  void *localptr;
  localptr = tolocal(blockPtr);

  if (localptr == NULL) return;

  localptr -= sizeof(AdminInfo);
  LockGet_Size(MemExclPtr, WRITER_LOCK, sizeof(int_locked));
  am_request1(SPREAD_MALLOC_PROC, spread_free_handler, localptr);
  LockRelease_Size(MemExclPtr, WRITER_LOCK, sizeof(int_locked));
}


/*
 * ----------------------------------------------------------------------------
 *
 * all_spread_free --
 *      Return a previously-allocated block of parallel storage to the 
 *      free pool.
 *
 *      All processors participate in this system call.
 *
 * Inputs: 
 *      blockPtr -- spread pointer to beginning of block allocated with
 *                  {all_}spread_malloc.
 *
 * Side effects:
 *      The storage pointed to by blockPtr is marked as free and returned
 *	to the free pool.  Nothing in the bytes pointed to by blockPtr is
 *	modified at this time:  no change will occur until at least the
 *	next call to {all_}spread_malloc.
 *
 * ----------------------------------------------------------------------------
 */
void all_spread_free(___V_SPTR blockPtr)
{

  am_barrier(1);

  if (MYPROC == SPREAD_MALLOC_PROC) {
    void *localptr;
    localptr = tolocal(blockPtr);

    if (localptr == NULL) return;
    localptr -= sizeof(AdminInfo);
    LockGet_Size(MemExclPtr, WRITER_LOCK, sizeof(int_locked));
    spread_free_handler(localptr);
    LockRelease_Size(MemExclPtr, WRITER_LOCK, sizeof(int_locked));
  }
}

/*
 * --------------------------------------------------------------------------------
 *         CMAM Handlers
 * --------------------------------------------------------------------------------
 */
 
void gmalloc_reply_handler(gmalloc_t *info, int result, void *ptr, int bytesNeeded)
{
  info->result = result;
  info->ptr = ptr;
  info->sizetobrk = bytesNeeded;
  info->flag++;
}

void gmalloc_bin_handler(int proc, unsigned size, int index, gmalloc_t *info)
{
  void *ptr;

  /* 
   * Check to see if we can satisfy the request from the memory 
   * already allocated.
   */
  if (ptr = trytobin(size, index)) {
    am_reply4(proc, gmalloc_reply_handler, info, ALLOCATED, ptr, 0);
    return;
  }
  
  /* 
   * Currently there is not any memory allocated of this size.  
   * Save the information and tell the processor it must allocate the memory.
   */
  bytesNeeded_save = size;
  index_save = index;
  regionSize_save =  size*BLOCKS_AT_ONCE;
  am_reply4(proc, gmalloc_reply_handler, info, NEEDSBRK, NULL, regionSize_save);
}


void morebinmem_handler(int proc, void *newRegion, gmalloc_t *info)
{
  Address result;
  int regionSize = regionSize_save;

  /*
   * Put the newly allocated memory into the list
   */
  while (regionSize >= bytesNeeded_save) {
    SET_ADMIN(newRegion, memFreeLists[index_save]);
    memFreeLists[index_save] = newRegion;
    newRegion += bytesNeeded_save;
    regionSize -= bytesNeeded_save;
#ifdef DEBUG    
    printf("b");
#endif
  }
#ifdef DEBUG
  printf("\n");
#endif
      
  /* 
   * Give the ptr to the processor which requested this memory
   */
  result = memFreeLists[index_save];
  memFreeLists[index_save] = (Address) GET_ADMIN(result);
  SET_ADMIN(result, MARK_DUMMY(bytesNeeded_save));
  am_reply4(proc, gmalloc_reply_handler, info, ALLOCATED, 
	       result + sizeof(AdminInfo), 0);
}


void gmalloc_big_handler(int proc, unsigned bytesNeeded, gmalloc_t *info)
{
  if (!memInitialized) {
    /* 
     * Initialize the large-object free list with two blocks that
     * mark its beginning and end.
     */
    if (!memInitBegin) {
      fprintf(stderr, "INTERNAL ERROR: Variables not initialized correctly.\n");
      exit(-1);
    }
    memLast = memFirst + MIN_REGION_SIZE - sizeof(AdminInfo);
    SET_ADMIN(memFirst, MARK_FREE(memLast - memFirst));
    SET_ADMIN(memLast, MARK_DUMMY(0));
    memCurrentPtr	= memFirst;
    memInitialized = 1;
  }
    
  trytoallocbig(proc, bytesNeeded, info);
}

void morebigmem_handler(int regionSize, void *newRegion, gmalloc_t *info)
{
  Address result;
  int admin, thisBlockSize;
  Address nextPtr;

  /*
   * If the new region immediately follows the end of the previous
   * region, merge them together.  At this point memCurrentPtr always
   * points to a block of memory adjacent to and preceding the block
   * of memory pointed to by memLast.   Thus it
   * may be possible to merge the new region with both the region
   * at result and the one at memLast.
   */
  admin = GET_ADMIN(memCurrentPtr);
  if (newRegion == (memLast+sizeof(AdminInfo))) {
    newRegion = memLast;
    regionSize += sizeof(AdminInfo);
    if (!IS_IN_USE(admin)) {
      newRegion = memCurrentPtr;
      regionSize += SIZE(admin);
    }
  }
  else {
    SET_ADMIN(memLast, MARK_DUMMY(newRegion - memLast));
  }

  /*
   * Create a dummy block at the end of the new region, which links
   * to "last".
   */
	
  SET_ADMIN(newRegion, MARK_FREE(regionSize - sizeof(AdminInfo)));
  memLast = newRegion + regionSize - sizeof(AdminInfo);
  SET_ADMIN(memLast, MARK_DUMMY(0));

  trytoallocbig(proc_save, bytesNeeded_save, info);
}


void trytoallocbig(int proc, unsigned bytesNeeded, gmalloc_t *info)
{
  int thisBlockSize, admin;
  Address nextPtr;
  Address result;

  result = memCurrentPtr;

  while (1) {
    admin = GET_ADMIN(result);
    thisBlockSize = SIZE(admin);
    nextPtr = result+thisBlockSize;
    if (!IS_IN_USE(admin)) {
	
      /*
       * Several blocks in a row could have been freed since the last
       * time we were here.  If so, merge them together.
       */
      while (!IS_IN_USE(GET_ADMIN(nextPtr))) {
	thisBlockSize += SIZE(GET_ADMIN(nextPtr));
	admin = MARK_FREE(thisBlockSize);
	SET_ADMIN(result, admin);
	nextPtr = result+thisBlockSize;
#ifdef DEBUG
	printf("l");
#endif
      }
#ifdef DEBUG
      printf(".\n");
#endif
      if (thisBlockSize >= bytesNeeded)  goto bigmalloc_satisfied;

      /*
       * We can't satisfy the request, but we may to need to update memLargestFree 
       */
      if (thisBlockSize > memLargestFree) memLargestFree = thisBlockSize;
    }
    
    /*
     * This block won't do the job.  Go on to the next block.
     */
    if (nextPtr != memLast) {
      result = nextPtr;
      continue;
    }

    /* 
     * Have reached the last element of the list of allocated chunks of memory.
     * If there's any chance that there's enough free storage to satisfy the 
     * smallest request, then go back to the beginning of the list of chunks and 
     * start again.
     */
    if ((memLargestFree + memBytesFreed) > bytesNeeded) {
      memLargestFree = 0;
      memBytesFreed = 0;
      result = memFirst;
      continue;
    }
    else break;
  }
 
  /* 
   * Not enough free space in the list.  Have the processor do an sbrk 
   */
  memCurrentPtr = result;
  proc_save = proc;
  bytesNeeded_save = bytesNeeded;
  am_reply4(proc, gmalloc_reply_handler, info, NEEDSBRK, NULL,
	       bytesNeeded + MIN_REGION_SIZE);
  return;

bigmalloc_satisfied:

  /*
   * At this point we've got a block that's large enough.  If it's
   * larger than needed for the object, put the rest back on the
   * free list (note: even if the remnant is smaller than the smallest
   * allowable large object, (which means it'll be used by itself,) we put it back
   * on the list so it can merge with either this block or the next,
   * whichever gets freed first).
   */

  if (thisBlockSize <= bytesNeeded + sizeof(AdminInfo)) {
    /* We don't have free space for more than the AdminInfo */
    SET_ADMIN(result, MARK_IN_USE(admin));
  } 
  else {
    SET_ADMIN(result+bytesNeeded, MARK_FREE(thisBlockSize - bytesNeeded));
    SET_ADMIN(result, MARK_IN_USE(bytesNeeded));
  }
  

  memCurrentPtr = result;
  am_reply4(proc, gmalloc_reply_handler, info, ALLOCATED, 
	       result+sizeof(AdminInfo), 0);
}



void spread_free_handler(blockPtr)
register Address blockPtr;	
{
    register int  admin;
    register int  index;
    register int  size;


    admin = GET_ADMIN(blockPtr);

  /* 
   *  Make sure that this block bears some resemblance to a
   *  well-formed storage block.
   */

    if (!IS_IN_USE(admin)) {
      printf("ERROR: gfree received corrupt storage block.\n");
      exit(-1);
    }

    /* This procedure is easier for un-binned blocks (those without the
     * DUMMY bit set) than for the binned ones.  If un-binned, just clear
     * the use bit and record how many bytes were freed, for use later
     * when deciding whether or not to allocate more storage.
     */

    if (!IS_DUMMY(admin)) {
      SET_ADMIN(blockPtr, MARK_FREE(admin));
      memBytesFreed += SIZE(admin);
    } else {
      /* 
       * For small blocks, add the block back onto its free list.
       */
      index = BLOCKSIZE_TO_INDEX(SIZE(admin));
      SET_ADMIN(blockPtr, MARK_FREE((int) memFreeLists[index]));
      memFreeLists[index] = blockPtr;
    }

}

void *trytobin(unsigned bytesNeeded, int index)
{

  int thisBlockSize, admin;
  Address result;
  Address nextPtr;

  result = memFreeLists[index];
  if (result == (Address) NULL) {
    /* There aren't currently any free objects of this size. */
      return NULL;
    }

  memFreeLists[index] = (Address) GET_ADMIN(result);
  SET_ADMIN(result, MARK_DUMMY(bytesNeeded));

  return (result + sizeof(AdminInfo));
}



void spread_malloc_init()
{
  
  LockInit();
  LockInitObj(&MemExclFlag);
  MemExclPtr = (___ILOCK_GPTR)toglobal(0, &MemExclFlag);

  if (MYPROC == SPREAD_MALLOC_PROC) {
    int i;
    /*
     * Clear out all of the bins.
     */
    for (i = 0; i < BIN_BUCKETS; i++) {
      memFreeLists[i] = NULL;
    }
  }
}


void *MemChunkAlloc(unsigned bytesNeeded)
{
  CM_memaddr_t old;

  if ((old = CMPE_get_iu_heap_ptr()) == NULL) {
    fprintf(stderr, "ERROR: CMPE_get_iu_heap_ptr failed.\n");
    exit(-1);
  }

  if (!CMPE_set_iu_heap_ptr(old + bytesNeeded)) {
    fprintf(stderr, "ERROR: spread_malloc failed to allocate memory.\n");
    exit(-1);
  }
  return (void *)old;
}






    
    






