/* 									tab:8
 *
 * split-c_print.h - Hack for printing from Split-C
 * 
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author:	Seth Copen Goldstein
 * Version:	17
 * Filename:	split-c_print.h
 * History:
 *	LTL	14	Wed Oct 27 15:27:43 1993
 *		Changed USER_HANDLER_BASE because X handlers are pulled
 *      out of libsplit-c and exist in a separate libray now
 *	AK	9	Tue Oct 26 10:43:10 1993
 *		Introduced support for scanf
 *	LTL	7	Mon Aug 16 15:51:28 1993
 *		Changed I/O stuff for version 112 to support CMMD
 *	SL	2	Tue May 25 11:33:34 1993
 *		Added handlers for basic X functions
 *	SCG	1	Mon May  3 10:31:08 1993
 *		incremented USER_HANDER_BASE to account for ALL functions
 *		in defaultHostFunctions
 *	SCG	1	Wed Jan 20 16:38:49 1993
 *		added USER_HANDLER_BASE so that users can install their own
 *		host handlers.
 */

#ifndef SPLIT_C_PRINT_H
#define SPLIT_C_PRINT_H

/* USER_HANDLER_BASE is used to define user host handlers.
 *   
 * DEFAULT_HOST_HANDLERS_NUM is the default number of host
 * handlers used by libsplit-c.
 * This must be updated when any new default functions are added in 
 * handlers_cp.c
 *
 * Four host handler slots are reserved for the X handlers.
 */

#ifndef CMMD

#ifdef SC_PROFILE /* Extra handler used in profiling version */
#define DEFAULT_HOST_HANDLERS_NUM 22
#else
#define DEFAULT_HOST_HANDLERS_NUM 21
#endif

#else /* CMMD */
#define DEFAULT_HOST_HANDLERS_NUM 2
#endif

#define USER_HANDLER_BASE (DEFAULT_HOST_HANDLERS_NUM + 4)

#ifndef CMMD
/*
 * HACK for printing etc. from the nodes
 */

/* this requires the user to include "stdio.h" before split-c.h */

#ifdef pe_obj
#	undef putc
#	undef getc
#	undef ferror
#	define printf 	localPrintf
#	define fscanf 	localFscanf
#	define scanf	localScanf
#	define fflush 	localFflush
#	define fputc	localFputc
#	define fgetc	localFgetc
#	define putc	fputc
#	define getc	fgetc
#	define ungetc	localUngetc
#	define exit	localExit
#	define ferror 	localFerror
#	define fclose 	localFclose
#   define fread  localFread
#   define fseek  localFseek
	extern int localFseek();
	extern int localFread();
        extern int all_fread();
        extern char *all_fgets();
        extern void all_fclose();
	extern void localPrintf();
	extern int localFflush();
	extern void PRINTF();
	extern void localExit();
#	undef stdin
#	undef stdout
#	undef stderr

#	ifdef FILE
		extern FILE *stdin;
		extern FILE *stdout;
		extern FILE *stderr;
	extern int getInt(FILE *stream);
	extern double getFloat(FILE *stream);
        extern FILE *all_fopen();
        extern FILE *all_independent_fopen();
#	endif


#endif /* pe_obj */

#else /* CMMD */


#ifdef pe_obj

#define exit	 localExit
    extern void localExit();


#ifdef FILE
/* keep these functions for backward compatibility */

    extern int getInt(FILE *stream);
    extern double getFloat(FILE *stream);


/* I/O functions build on top of CMMD */

/* I/O error codes */
#define BAD_IO_MODE ECMMDMADMODE
#define INCONSISTENT_IO ECMMDINCONSISTENT
#define IO_MISMATCH ECMMDNOMATCH

/* I/O modes */
#define LOCAL CMMD_local
#define INDEPENDENT CMMD_independent
#define SYNC_BC CMMD_sync_bc
#define SYNC_SEQ CMMD_sync_seq

    extern int all_open_independent(char *path, int flags, int mode);
    extern int all_open_sync_bc(char *path, int flags, int mode);
    extern int all_open_sync_seq(char *path, int flags, int mode);
    extern FILE *all_fopen_independent(char *path, char *mode);
    extern FILE *all_fopen_sync_bc(char *path, char *mode);
    extern FILE *all_fopen_sync_seq(char *path, char *mode);
    extern int set_io_mode(int fd, int io_mode);
    extern int fset_io_mode(FILE *stream, int io_mode);
    extern int get_io_mode(int fd);
    extern int fget_io_mode(FILE *stream);

#endif 

#endif /* pe_obj */

#endif /* CMMD */

#endif /* SPLIT_C_PRINT_H */
