/* 									tab:8
 *
 * split-c.sc - Support for Split-C extensions
 * 
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author:	Arvind Krishnamurthy
 * Version:	8
 * Filename:	split-c.sc
 * History:
 *	ssl	8	Thu Oct 28 22:04:40 1993
 *		Include am.h instead of cm5.h
 *	LTL	6	Mon Aug 16 15:54:02 1993
 *		Changed all_set_PROCS and __i_reduce for version 112
 *		to support CMMD
 *	SCG	5	Fri Jan 22 13:56:39 1993
 *		added host part of all_set_procs
 * 	TvE	4	Tue Jan 19 01:52:21 PST 1993
 *		Added all_set_PROCS. Still needs check against what has been
 *		compiled into the program, but that requires cc support.
 * 	TvE	3	Wed Dec 23 14:41:44 PST 1992
 *		Major overhaul.
 * 	TvE	2	Tue Sep 29 12:44:08 PDT 1992
 *		include file name changes
 *		removed include of gmalloc.pe.c (gag!)
 *	SCG	1	Wed Sep 23 00:11:02 1992
 *		made it compatible with split-c_int.h
 */
static char _version_string_[] = "\nVersion:8:split-c.c\0\n";

#include <stdio.h>
#define SPLIT_C_INTERNAL_DEFINE	/* turns extern's to not externs */
#include <split-c/split-c.h>

/* .....................................
 *
 * Check/change the number of processors
 *
 * .....................................
 */

const int ___PROCS;
const int ___PROCSM1;
const int ___PROCSLOG;
const int ___MYPROC;
static num_set_PROCS = 0; /* 0=never, 1=at init time, 2=once by user */

/* return log2 if exact, else return -1 */
int
getLog(int i)
{
	int exp = 0;

	if(i <= 0) return -1;
	for(exp=0; (i&1)==0; i>>=1) exp++;
	return i==1 ? exp : -1;
}

void
all_set_PROCS(int new_PROCS)
{
#if 0 /* dunno how to do that yet */
	extern int ___PROCS_linked;	/* symbol defined at link time */
	int compile_size = (int)&___PROCS_linked;
#endif
	int new_PROCSLOG = getLog(new_PROCS);

	barrier();

	/*
	 * Check the argument against partition size
	 */
	if(new_PROCS > CMNA_partition_size) {
		on_one {
			fprintf(stderr, "FATAL error: "
			    "attempt to set PROCS to %d\n"
			    "which is greater than partition_size (%d)\n",
			    new_PROCS, CMNA_partition_size);
		}
		barrier();
		exit(1);
	} else if(new_PROCS <= 0) {
		on_one {
			fprintf(stderr, "FATAL error: "
			    "attempt to set PROCS to %d\n", new_PROCS);
		}
		barrier();
		exit(1);
	}

#if 0 /* dunno how to do that yet */
	/*
	 * Check the argument against compilation flags of program
	 */
	if(compile_size > 0 && new_PROCS != compile_size) {
		on_one {
			fprintf(stderr, "FATAL error: "
			    "attempt to set PROCS (%d) to other than\n"
			    "value compiled into program (%d)\n", new_PROCS,
			    compile_size);
		}
		barrier();
		exit(1);
	} else if(compile_size == -1 && new_PROCSLOG < 0) {
		on_one {
			fprintf(stderr, "FATAL error: "
			    "attempt to set PROCS (%d) to non power-of-two\n"
			    "value with program compiled for power of two\n",
			    new_PROCS);
		}
		barrier();
		exit(1);
	}
#endif

	/*
	 * Check the argument against compilation flags of library
	 */
#ifdef	___PROCS_POWER_OF_TWO
	if(new_PROCSLOG == -1) {
		on_one {
			fprintf(stderr, "FATAL error: "
			    "attempt to set PROCS (%d) to non power-of-two\n"
			    "value with library compiled for power of two\n",
			    new_PROCS);
		}
		barrier();
		exit(1);
	}
#endif
#ifndef	___PROCS_VARIABLE
	if(new_PROCS != PROCS) {
		on_one {
			fprintf(stderr, "FATAL error: "
			    "attempt to set PROCS (%d) to other than\n"
			    "value compiled into library (%d)\n", new_PROCS,
			    PROCS);
		}
		barrier();
		exit(1);
	}
#endif

	/*
	 * We get called once at init time and possibly once by the user
	 */
	if(num_set_PROCS > 1) {
		on_one {
			fprintf(stderr, "FATAL error: "
			    "attempt to `all_set_PROCS' a 2nd time\n");
		}
		barrier();
		exit(1);
	}
	num_set_PROCS++;

	/*
	 * Now let's DOIT
	 */
#ifndef CMMD

	if(MYPROC >= new_PROCS) {
		/* sure looks like *I*'m out... */
		/* first abstain from barriers */
		*(volatile int *)(NI_BASE+NI_SYNC_GLOBAL_ABSTAIN_O) = 0x1;
		(void)___i_reduce(0); /* barrier without sync network */
		/* then abstain from reduces (i.e. all_store_syncs) */
		*(volatile int *)(NI_BASE+NI_COM_CONTROL_O) = 0x3;
		/* become a memory controller ... */
		while(1) am_poll();
	} else {
		/* Ha! I'm still in the game... */
		/* wait for extraneous procs to abstain from sync network */
		(void)___i_reduce(0);
		/* set ___PROCS and CO. */
		(int)___PROCS = new_PROCS;
		(int)___PROCSM1 = new_PROCS - 1;
		(int)___PROCSLOG = new_PROCSLOG;
		_set_procs_on_host(new_PROCS);
		/* now sync all remaining processors */
		barrier();
	}

#else
	CMMD_reset_partition_size(new_PROCS);

	if(MYPROC < new_PROCS) {
		/* Ha! I'm still in the game... */
		/* wait for extraneous procs to abstain from sync network */
		(void)___i_reduce(0);
		/* set ___PROCS and CO. */
		(int)___PROCS = new_PROCS;
		(int)___PROCSM1 = new_PROCS - 1;
		(int)___PROCSLOG = new_PROCSLOG;
		_set_procs_on_host(new_PROCS);
		/* now sync all remaining processors */
		barrier();
	}

#endif

}
	
/* ................................................
 *
 * Perform integer-add reduce using control network
 *
 * ................................................
 */

#ifndef CMMD

int
___i_reduce(int diff)
{
	int status;

	CMNA_com_send_first(ADD_SCAN, SCAN_REDUCE, 1, diff);
	status = *(volatile int *)(NI_BASE+NI_COM_STATUS_O);
	while(!(status & (1<<NI_SEND_OK_P))) {
		am_poll();
		CMNA_com_send_first(ADD_SCAN, SCAN_REDUCE, 1, diff);
		status = *(volatile int *)(NI_BASE+NI_COM_STATUS_O);
	}

	status = *(volatile int *)(NI_BASE+NI_COM_STATUS_O);
	while(!(status & (1<<NI_REC_OK_P))) {
		am_poll();
		status = *(volatile int *)(NI_BASE+NI_COM_STATUS_O);
	}
	return *(volatile int *)(NI_BASE+NI_COM_RECV_O);
}

#else

int ___i_reduce(int diff)
{
  barrier();
  return (CMMD_reduce_int(diff, CMMD_combiner_add));
}

#endif /* CMMD */









