/*                                                                      tab:8
 *
 * atomic_int.h - Internal include file for PE side of the atomic library procedures.
 *
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author:		Andrea Dusseau
 * Version:		11
 * Creation Date:	Sun Oct 25 15:50:36 PST 1992
 * Filename:		atomic_int.h
 * History:
 *	ssl	9	Wed Jan 19 11:46:06 1994
 *		removed rpc
 *	ssl	8	Tue Nov 16 11:53:40 1993
 *		Replaced global tags with macros.
 *	ssl	7	Thu Oct 28 22:52:39 1993
 *		Include am.h instead of cmam.h or am_cmmd.h
 *	LTL	6	Mon Aug 16 15:25:07 1993
 *		Fixed it up for version 112 to support CMMD
 * 	TvE	5	Sun Apr  4 12:35:09 PDT 1993
 *		Changed include <cmam/cmam.h> to "cmam.h"
 *	SL	3	Wed Jan 20 10:18:38 1993
 *		Added definitions for rpc return functions
 *	ACD    2      Tue Jan  5 14:50:58 PST 1993
 *		Moved definitions for atomic, atomic_x to atomic.sc
*/

#ifndef __ATOMIC_INT_H__
#define __ATOMIC_INT_H__

#include "am.h"

/* .........................................................
 *
 * Structure to receive reply from atomic and rpc operations
 *
 * .........................................................
 */

typedef struct {
	volatile int	flag;
	union {
		int	v_int;
		double	v_double;
	} value;
} atomic_t;

/* ..................................
 *
 * Send atomic or rpc operation reply
 *
 * ..................................
 */
EXTERN void atomic_return_i_rh(int src_node, int val);
EXTERN void atomic_return_f_rh(int src_node, float val);
EXTERN void atomic_return_d_rh(int src_node, double val);

#define atomic_return_i(val) { \
  am_reply_1(proc, atomic_return_i_rh, val); return; }

#ifdef PAM
#define atomic_return_d(val) { \
  am_reply__fd(proc, atomic_return_d_rh, val); return; }
#else
#define atomic_return_d(val) { \
  am_reply_1(proc, atomic_return_d_rh, val); return; }
#endif

/* ....................................
 *
 * Canned atomic operations on integers
 *
 * ....................................
 */

EXTERN void cmp_and_swap_atomic(int proc, short *loc, int test, int new);
EXTERN void test_and_set_atomic(int proc, int *loc);
EXTERN void exchange_atomic(int proc, int *loc, int new_val);
EXTERN void fetch_and_add_atomic(int proc, int *loc, int new_val);

EXTERN inline int cmp_and_swap(___I_GPTR loc, int test, int new)
INLINE({
	return atomic_i(cmp_and_swap_atomic, loc, test, new);
})

EXTERN inline int test_and_set(___I_GPTR loc)
INLINE({ 
	return atomic_i(test_and_set_atomic, loc);
})

EXTERN inline int exchange(___I_GPTR loc, int new_val)
INLINE({ 
	return atomic_i(exchange_atomic, loc, new_val);
})

EXTERN inline int fetch_and_add(___I_GPTR loc, int val)
INLINE({ 
	return atomic_i(fetch_and_add_atomic, loc, val);
})

#endif /* __ATOMIC_INT_H__ */



