/*									tab:8
 *
 * main.c - Main program for Generic Active Message layers
 *
 * 
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 		Steve Luna
 * Version:		49
 * Creation Date:	Sat Nov 20 16:47:42 1993
 * Filename:		main.c
 * History:
 *	LTL	41	Mon Jul 24 18:21:57 1995
 *		Added all_set_PROCS, which prints out an error when new_PROCS != PROCS
 *	LTL	34	Mon Jul 24 16:28:20 1995
 *		Removed #define's for the default version
 *	LTL	32	Mon Feb 20 09:36:06 1995
 *		Modified to use GAM. Used #ifdef instead of having separate platform
 *    specific versions.
 *	ssl	2	Mon Apr 18 20:54:41 1994
 *		modified for SP1
 */

#include <stdio.h>
#define SPLIT_C_INTERNAL_DEFINE	/* turns extern's to not externs */
#include <split-c/split-c.h>

/* No longer const to avoid cc compiler error. Must fix later. - ssl */
int ___PROCS;
int ___PROCSM1;
int ___PROCSLOG;
int ___MYPROC;
int ___AM_MAX_PKT;

/* return log2 if exact, else return -1 */
getLog(int i)
{
	int exp = 0;

	if(i <= 0) return -1;
	for(exp=0; (i&1)==0; i>>=1) exp++;
	return i==1 ? exp : -1;
}

init_consts()
{
	___PROCS = am_procs();
	___MYPROC = am_my_proc();
	___AM_MAX_PKT = am_max_size();		
	___PROCSM1 = (int) ___PROCS - 1;
	___PROCSLOG = getLog((int) ___PROCS);

#if __PROCSLOG__ > 0
	/* number of processors is fixed power-of-two */
	if (___PROCSLOG != __PROCSLOG__)
	{
		on_one
			fprintf(stderr, "FATAL error: "
			    "attempt to set PROCS (%d) to other than\n"
			    "value compiled into program (%d)\n", ___PROCS,
			    __PROCS__);
		exit(1);
	}
#elif defined(__PROCSLOG__)
	/* number of processors is unknown power-of-two */
	if (___PROCSLOG == -1)
	{
		on_one
			fprintf(stderr, "FATAL error: "
			    "attempt to set PROCS (%d) to non power-of-two\n"
			    "value with program compiled for power of two\n",
			    ___PROCS);
		exit(1);
	}
#elif __PROCS__ > 0
	/* number of processors is fixed non-power-of-two */
	if (___PROCS != __PROCS__)
	{
		on_one
			fprintf(stderr, "FATAL error: "
			    "attempt to set PROCS (%d) to other than\n"
			    "value compiled into program (%d)\n", ___PROCS,
			    __PROCS__);
		exit(1);
	}
#endif

	/*spread_malloc_init();*/
	atomic_init();
	com_init();
#ifdef SC_PROFILE	
	profile_init();
#endif	
}


void all_set_PROCS(int new_PROCS)
{
  if (new_PROCS != PROCS) {
    on_one {
      fprintf(stderr, "FATAL error: attempt to reset PROCS\n");
    }
    barrier();
    exit(1);
  }
}

int __gc_program_ended;   /* set this flag for malloc() fixes on SP2 */

main(int argc, char **argv)
{
  __gc_program_ended = 0;
  am_enable();
	init_consts();
	timer_init();
	barrier();
  __gc_program_ended = 1;	
	splitc_main(argc, argv);
  __gc_program_ended = 2;	
	barrier();
	am_disable();
	exit(0);
}

