/*									tab:8
 *
 * store.sc - signaling global memory accesses (store)
 *
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 		Thorsten von Eicken
 * Version:		71
 * Creation Date:	Thu Dec 24 02:39:52 PST 1992
 * Filename:		store.sc
 * History:
 *	ssl	61	Sun Feb 27 16:07:30 1994
 *		added all_store_sync_ctr
 *	ssl	60	Tue Feb  8 16:35:47 1994
 *		Added barrier to end of all_store_sync. Race condition if ___i_reduce
 *      is synthesized or if interrupts are turned on.
 *	ssl	59	Tue Dec 14 01:19:45 1993
 *		Moved store code to bulk.c
 *	ssl	58	Sun Dec 12 23:17:20 1993
 *		All bulk store stuff, except for simple case, removed for paragon.
 *	ssl	57	Sun Dec  5 23:07:07 1993
 *		am and handlers now use unsigned instead of long long as arguments.
 *      doubles will be passed in fp registers and long longs will be dword
 *      aligned.
 *	ssl	55	Tue Nov 16 12:22:50 1993
 *		Replaced global tags with macros.
 *	LTL	53	Mon Aug 16 15:56:37 1993
 *		Fixed it up for version 112 to support CMMD
 *	AK	49	Tue Mar 16 14:31:28 1993
 *		Made use of both networks for store operations.
 *	AK	37	Fri Jan 22 21:23:08 1993
 *		Merged bulk.sc and removed it.
 *	AK	35	Fri Jan 22 21:20:17 1993
 *		Made __3i_store_handler public. Hope to merge
 *		bulk.sc in this file and make it static again.
 * 	TvE	33	Sun Jan 17 02:50:37 PST 1993
 *		___store with a short length had bugs. Didn't increment
 *		all pointers and didn't take odd lengths into account. Still
 *		needs a fix if the start address isn't on a word boundary.
 * 	TvE	1	Thu Dec 24 02:39:52 PST 1992
 *		Extracted from split-c_int.h
 */
static char _version_string_[] = "\nVersion:61:store.c\0\n";

#include <split-c/split-c.h>

/* .....................
 *
 * Keep track of STORESs
 *
 * .....................
 */

Counter sentBytes = 0;
Counter recvBytes = 0;

/* Wait until all stores from all processors have completed */
void all_store_sync(void)
{
	int res;
	extern int ___i_reduce(int diff);

	am_barrier(0);
	do {
		res = ___i_reduce(recvBytes-sentBytes);
	} while(res != 0);

	recvBytes = sentBytes = 0;
	/* need this if ___i_reduce is synthesized or if interrupts are turned
	   on - ssl */
	am_barrier(0);
}


void all_store_sync_ctr(Counter *rbytes, Counter *sbytes)
{
	int res;
	extern int ___i_reduce(int diff);

	am_barrier(0);
	do {
		res = ___i_reduce(*rbytes - *sbytes);
	} while(res != 0);

	*rbytes = *sbytes = 0;
	/* need this if ___i_reduce is synthesized or if interrupts are turned
	   on - ssl */
	am_barrier(0);
}

/* ............................
 *
 * STOREs with explicit counter
 *
 * ............................
 */

void __c_store_handler (int src_node, char *addr, int data, Counter *ctr)
{
	addr[0] = data;
	*ctr += sizeof(char);
}

void __sh_store_handler(int src_node, short *addr, int data, Counter *ctr)
{
	addr[0] = data;
	*ctr += sizeof(short);
}

void __i_store_handler(int src_node, int *addr, int data, Counter *ctr,
											 int *addr2)
{
	addr[0] = data;
	*ctr += sizeof(int);
}

void __f_store_handler(int src_node, float *addr, unsigned word0, Counter *ctr)
{
	*((unsigned*)addr) = word0;
	*ctr += sizeof(float);
}

/* pass unsigned so it will not pass in fp regs */
void __d_store_handler(int src_node, double *addr, unsigned word0,
											 unsigned word1, Counter *ctr)
{
	*((unsigned*)addr) = word0;
	*((unsigned*)addr+1) = word1;
	*ctr += sizeof(double);
}

/* pass as two unsigned so long long won't be dword aligned. addr is unsigned*
   so that cc can understand it. - ssl */
void __ll_store_handler(int src_node, unsigned *addr, unsigned word0,
												unsigned word1,	Counter *ctr)
{
	*addr = word0;
	*(addr+1) = word1;
	*ctr += 2*sizeof(unsigned);
}

