/*									tab:8
 *
 * hppa.h - Macros for pmam
 *
 * 
 * "Copyright (c) 1993 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 		Richard P. Martin
 * Version:		2
 * Creation Date:	Sun Nov 21 17:30:24 PST 1993
 * Filename:		timer.c
 * History:
 *	ssl	2	Mon Apr 18 20:44:35 1994
 *		added HP_CYCLE_TIMER
 */

#include <split-c/split-c.h>
#include <time.h>

#ifndef HP_CYCLE_TIMER
extern struct timeval splitc_base_time;

double get_seconds()
{
  struct timeval t,lapsed;
  struct timezone z;
  
  gettimeofday(&t,&z);
  if (splitc_base_time.tv_usec > t.tv_usec) {
    t.tv_usec += 1000000;
    t.tv_sec--;
  }
    
  lapsed.tv_usec = t.tv_usec - splitc_base_time.tv_usec;
  lapsed.tv_sec = t.tv_sec - splitc_base_time.tv_sec;

  return (double)lapsed.tv_sec+((double)lapsed.tv_usec/(double)1000000.0) ;
}
#else
/* this version for accurate timing using cycle counter. must be careful
   because it is wall clock time and it overflows every 30 seconds - ssl */
#define CLK 99
#define MHZ 1e6
#define MAXITMR 0xffffffff
#define GETITMR(x) asm("mfctl %%itmr,%0" : "=r" (x));

double get_seconds()
{
	static unsigned x, init = 0;
	static double sec = 0;
	unsigned newx;

	if (!init)
	{		
		init++;
		GETITMR(x);
		return sec;
	}
	else
	{
		GETITMR(newx);
		/* if it wrapped */
		if (x >= newx)
			sec += (MAXITMR - (x - newx))/(CLK*MHZ);
		else
			sec += (newx - x)/(CLK*MHZ);
		x = newx;
		return sec;
	}
}
#endif

