/*									tab:4
 *
 * barrier.c - Code to implement Split-C barrier
 *
 * "Copyright (c) 1996 and The Regents of the University 
 * of California.  All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 			Andrea C. Dusseau
 * Version:			2
 * Creation Date:	Mon Feb 12 23:26:01 1996
 * Filename:		barrier.c
 * History:
 */

#include <stdio.h>
#include <split-c/split-c.h>

static Counter up_buf[2];
static Counter down_buf;

mutex_t	Barrier_mutex;

void barrier_init()
{
#ifndef SC_SEP_REQ_RPL
#ifndef SC_AM_ATOMIC
#ifdef  SC_MSG_THR
	if (mutex_init(&Barrier_mutex, USYNC_THREAD, NULL)) {
		perror("mutex_init of Store_mutex");
		exit(-1);
	}
#endif
#endif
#endif
	up_buf[0] = up_buf[1] = down_buf = 0;
}

void __barrier_incr_handler(void *token, int *val)
{
	__splitc_lock(&Barrier_mutex);
	(*val)++;
	__splitc_unlock(&Barrier_mutex);
	AM_Reply0(token, __NULL_RH);
}

void barrier_up()
{
    int parent = (MYPROC - 1) / 2;
    int odd_child = 2 * MYPROC + 1;
    int even_child = 2 * MYPROC + 2;
    int parity = MYPROC & 1;

    if (MYPROC == 0) {
		if (PROCS != 1) {
			if (PROCS == 2) {
				while (up_buf[1] == 0) {
					splitc_poll();
				}
			} else {
				while (up_buf[0] == 0 || up_buf[1] == 0) {
					splitc_poll();
				}
			}
		} else {
			splitc_poll();
		}
	} else {
		if (odd_child >= PROCS) {
			AM_Request1(__sc_endpoint_src, parent, __BARRIER_INCR_H, (int)&up_buf[parity]);
		} else if (even_child >= PROCS) {
			while (up_buf[1] == 0) {
				splitc_poll();
			}
			AM_Request1(__sc_endpoint_src, parent, __BARRIER_INCR_H, (int)&up_buf[parity]);
		} else {
			while (up_buf[0] == 0 || up_buf[1] == 0) {
				splitc_poll();
			}
			AM_Request1(__sc_endpoint_src, parent, __BARRIER_INCR_H, (int)&up_buf[parity]);
		}
	}
	/* not zero */
	__splitc_lock(&Barrier_mutex);
	up_buf[0] = up_buf[1] = 0;
	__splitc_unlock(&Barrier_mutex);
}

void barrier_down()
{
    int left = 2 * MYPROC + 1;
    int right = 2 * MYPROC + 2;
	
    if (MYPROC != 0) {
		while (down_buf == 0) {
			splitc_poll();
		}
    }

	if (left < PROCS) {
		AM_Request1(__sc_endpoint_src, left, __BARRIER_INCR_H, (int)&down_buf);
	}
	if (right < PROCS) {
		AM_Request1(__sc_endpoint_src, right, __BARRIER_INCR_H, (int)&down_buf);
	}
	__splitc_lock(&Barrier_mutex);
	down_buf = 0;
	__splitc_unlock(&Barrier_mutex);
} 



