/*									tab:8
 *
 * bulk.c - Handlers for bulk functions.
 *
 * 
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 		Steve Luna
 * Version:		27
 * Creation Date:	Sat Nov 27 23:07:14 1993
 * Filename:		bulk.c
 * History:
 *	LTL	8	Sun Jul 10 15:06:21 1994
 *		Modified for TCPAM
 *	RPM	2	Sun Nov 28 22:33:34 1993
 *		new version for PMAM interface, bonged segments
 */

#include <split-c/split-c.h>


void __get_handler(void *token, int local, int dest, int len, int ctr)
{
	AM_ReplyXfer4(token, dest, __GETCOMPLETE_RH, local, len, ctr, 0, 0, 0);
}
				  
void __get_complete_reply_handler(void *token, void *buf, int len, Counter *ctr)
{
	if (ctr != NULL) {
		__decr_ctr(ctr);
	}
}


void __bulk_put_handler(void *token, void *buf, int len, Counter *ctr)
{
	AM_Reply1(token, __PUTCTRCOMPSRC_RH, ctr);
}

void __put_ctr_comp_src_reply_handler(void *token, Counter *ctr)
{
	if (ctr != NULL) {
		__decr_ctr(ctr);
	}
}

void __null_reply_handler(void *token)
{
}

/* ............................
 *
 * STOREs with explicit counter
 *
 * ............................
 */


void __store_handler(void *token, void *buf, int len)
{
	__incr_ctr_size(&recvBytes, len);
	AM_Reply0(token, __NULL_RH);
}


/* STORE bulk */
void __store(___V_GPTR gPtr, void *lPtr, int len)
{
	Processor pid = toproc(gPtr);
	void *addr    = tolocal(gPtr);
	int i;

	if(pid == MYPROC) {
		__incr_ctr_size(&sentBytes, len);
		memcpy(addr,lPtr, len);
		__incr_ctr_size(&recvBytes, len);
	} else {
		if (len <= ___AM_MAX_LONG) {
			AM_RequestXfer4(__sc_endpoint_src, pid, (int)addr, __STORE_H, lPtr,
							len, 0, 0, 0, 0);
			/*am_store(pid, lPtr, addr, len, store_handler, 0);*/
		}
		else {
			for (i = ___AM_MAX_LONG; i < len; i += ___AM_MAX_LONG) {
				AM_RequestXfer4(__sc_endpoint_src, pid, (int)addr, __STORE_H, lPtr,
								___AM_MAX_LONG, 0, 0, 0, 0);
				/*am_store(pid, lPtr, addr, ___AM_MAX_PKT, store_handler, 0);*/
				(char *)addr += ___AM_MAX_LONG;
				(char *)lPtr += ___AM_MAX_LONG;
			}
			i -= ___AM_MAX_LONG;
			AM_RequestXfer4(__sc_endpoint_src, pid, (int)addr, __STORE_H, lPtr,
							len - i, 0, 0, 0, 0);
			/*am_store(pid, lPtr, addr, len - i, store_handler, 0);*/
		}
		__incr_ctr_size(&sentBytes, len);
	}
}

void
bulk_store(___V_GPTR gPtr, void *lPtr, int len)
{
  __store(gPtr, lPtr, len);
}







