/*									tab:8
 *
 * main.c - Main program for Generic Active Message layers
 *
 * 
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 		Steve Luna
 * Version:		43
 * Creation Date:	Sat Nov 20 16:47:42 1993
 * Filename:		main.c
 * History:
 *	LTL	41	Mon Jul 24 18:21:57 1995
 *		Added all_set_PROCS, which prints out an error when new_PROCS != PROCS
 *	LTL	34	Mon Jul 24 16:28:20 1995
 *		Removed #define's for the default version
 *	LTL	32	Mon Feb 20 09:36:06 1995
 *		Modified to use GAM. Used #ifdef instead of having separate platform
 *    specific versions.
 *	ssl	2	Mon Apr 18 20:54:41 1994
 *		modified for SP1
 */

#include <stdio.h>
#define SPLIT_C_INTERNAL_DEFINE	/* turns extern's to not externs */
#include <split-c/split-c.h>

#include <glib.h>
#include <enameserver.h>

/* No longer const to avoid cc compiler error. Must fix later. - ssl */
int ___PROCS;
int ___PROCSM1;
int ___PROCSLOG;
int ___MYPROC;

int	___AM_MAX_SHORT;
int	___AM_MAX_MEDIUM;
int	___AM_MAX_LONG;

#ifdef 	SC_SEP_REQ_RPL
eb_t	__sc_bundle_dst;
eb_t	__sc_bundle_src;
ea_t	__sc_endpoint_dst;
ea_t	__sc_endpoint_src;
#else
/* If request and reply bundles are the same, then share variable */
eb_t	__sc_bundle;
ea_t	__sc_endpoint;
#endif

/* return log2 if exact, else return -1 */
getLog(int i)
{
	int exp = 0;

	if(i <= 0) return -1;
	for(exp=0; (i&1)==0; i>>=1) exp++;
	return i==1 ? exp : -1;
}

init_consts()
{
	___PROCSM1 = (int) ___PROCS - 1;
	___PROCSLOG = getLog((int) ___PROCS);

#if __PROCSLOG__ > 0
	/* number of processors is fixed power-of-two */
	if (___PROCSLOG != __PROCSLOG__)
	{
		on_one
			fprintf(stderr, "FATAL error: "
			    "attempt to set PROCS (%d) to other than\n"
			    "value compiled into program (%d)\n", ___PROCS,
			    __PROCS__);
		exit(1);
	}
#elif defined(__PROCSLOG__)
	/* number of processors is unknown power-of-two */
	if (___PROCSLOG == -1)
	{
		on_one
			fprintf(stderr, "FATAL error: "
			    "attempt to set PROCS (%d) to non power-of-two\n"
			    "value with program compiled for power of two\n",
			    ___PROCS);
		exit(1);
	}
#elif __PROCS__ > 0
	/* number of processors is fixed non-power-of-two */
	if (___PROCS != __PROCS__)
	{
		on_one
			fprintf(stderr, "FATAL error: "
			    "attempt to set PROCS (%d) to other than\n"
			    "value compiled into program (%d)\n", ___PROCS,
			    __PROCS__);
		exit(1);
	}
#endif

	/*spread_malloc_init();*/
	/*atomic_init();*/

	barrier_init();
	com_init();
	store_init();
	memory_init();
	atomic_init();
	timer_init();

#ifdef SC_PROFILE
	profile_init();
#endif	
}


void all_set_PROCS(int new_PROCS)
{
  if (new_PROCS != PROCS) {
    on_one {
      fprintf(stderr, "FATAL error: attempt to reset PROCS\n");
    }
    barrier();
    exit(1);
  }
}


void __error_handler(int status, op_t opcode, void *argblock)
{
	printf("AM layer failed on node %d.\n", MYPROC);

	switch (status) {
	  case EBADARGS:
		printf("Invalid arguements to request or reply\n");
		break;
	  case EBADENTRY:
		printf("Unbound table entry\n");
		break;
	  case EBADTAG:
		printf("Sender Tag did not match Receiver Tag\n");
		break;
	  case EBADHANDLER:
		printf("Invalid index into receiver's handler table\n");
		break;
	  case EBADSEGOFF:
		printf("Invalid offset into destination memory segment\n");
		break;
	  case EBADLENGTH:
		printf("Bulk transfer length exceeds sengment length\n");
		break;
	  case EBADENDPOINT:
		printf("Destination endpoint does not exist\n");
		break;
	  case ECONGESTION:
		printf("Persistent congestion at destination\n");
		break;
	  case EUNREACHABLE:
		printf("Unreachable destination\n");
		break;
#if 0
	  case EREPLYREJECTED:
		printf("Destination refused reply message\n");
		break;
#endif
	  default:
		printf("Unknown status value: %d\n", status);
	}

	fflush(stdout);
	exit(-1);
}


#define NS_Safe(func) \
    { if (func != 0) {fprintf(stderr, "ERROR: %s failed\n", #func); exit(1);}}

int setupNames()			   
{
	char 	inName[32];
	byte	*result;
	int		result_len;
	int 	i;
	int		translations;
	en_t	global_endpoint;
	en_t	__sc_global[PROCS];


	AM_Safe(AM_MaxNumTranslations(&translations));
	if (translations < PROCS) {
		fprintf(stderr, "ERROR: AM_MaxNumTranslations (%d) less than PROCS (%d).\n",
				translations, PROCS);
		exit(-1);
	}

	/* I don't know the right way to do this! */
	global_endpoint.port = __sc_endpoint_dst->sockaddr.sin_port;
    global_endpoint.ip_addr = __sc_endpoint_dst->sockaddr.sin_addr.s_addr;

	/* Insert our global endpoint name into the name server */
	if (PROCS == 1) {
		AM_Safe(AM_Map(__sc_endpoint_src, 0, global_endpoint, SPLITC_TAG));
	}
	else {
		NS_Safe(NameServer_Init());
		sprintf(inName, "Pid%dVnn%d", Glib_GetMyNpid(), MYPROC);
		NS_Safe(NameServer_Bind(inName, (byte *)&global_endpoint, sizeof(en_t)));
		if (!Glib_Barrier()) {
			perror("Glib_Barrier Failed");
			exit(-1);
		}

		for (i = 0; i < PROCS; i++) {
			sprintf(inName, "Pid%dVnn%d", Glib_GetMyNpid(), i);
			NS_Safe(NameServer_Lookup(inName, &result, &result_len));
			if (result == NULL) {
				fprintf(stderr, "ERROR: result Null for lookup %s\n", inName);
				exit(-1);
			}			

			if (result_len != sizeof(en_t)) {
				fprintf(stderr, "ERROR: result_len(%d)!=sizeof(en_t)(%d) for lookup %s\n",
						result_len, sizeof(en_t), inName);
				exit(-1);
			}
			memcpy(&__sc_global[i], result, sizeof(en_t));
			
			AM_Safe(AM_Map(__sc_endpoint_src, i, __sc_global[i], SPLITC_TAG));
		}
	}
}


void setupAM()
{
	int 	maxSegLength;
	int 	maxHandlers;
	int		i;
	
	AM_Safe(AM_Init());
 	AM_Safe(AM_MaxSegLength(&maxSegLength));

	/* Allocate Signalling bundle and endpoint */
	AM_Safe(AM_AllocateBundle(AM_PAR, &__sc_bundle_signal));
	AM_Safe(AM_AllocateEndpoint(__sc_bundle_src, &__sc_endpoint_signal));
	AM_Safe(AM_SetSeg(__sc_endpoint_signal, (void *)0x00, maxSegLength));
	AM_Safe(AM_SetEventMask(__sc_bundle_signal, AM_NOEVENTS));

	/* Allocate Request bundle and endpoint */
	AM_Safe(AM_AllocateBundle(AM_PAR, &__sc_bundle_request));
	AM_Safe(AM_AllocateEndpoint(__sc_bundle_request, &__sc_endpoint_request));
	AM_Safe(AM_SetSeg(__sc_endpoint_request, (void *)0x00, maxSegLength));
	AM_Safe(AM_SetEventMask(__sc_bundle_request, AM_NOTEMPTY));


	/* Setup Signalling Handlers */
	if ((maxHandlers = AM_MaxNumHandlers()) < __LAST_SIGNAL_HANDLER) {
		fprintf(stderr, "ERROR: Only %d handler table entries available; need %d\n",
				maxHandlers, __LAST_SIGNAL_HANDLER);
		exit(1);
	}
	
	for (i = 0; 1; i++) {
		if (signal_handler_list[i].handler_index == __LAST_SIGNAL_HANDLER) {
			break;
		}
		AM_Safe(AM_SetHandler(__sc_endpoint_signal,
							  signal_handler_list[i].handler_index,
							  signal_handler_list[i].func));
	}

	/* Setup Request Handlers */
	if ((maxHandlers = AM_MaxNumHandlers()) < __LAST_REQUEST_HANDLER) {
		fprintf(stderr, "ERROR: Only %d handler table entries available; need %d\n",
				maxHandlers, __LAST_REQUEST_HANDLER);
		exit(1);
	}
	
	for (i = 0; 1; i++) {
		if (request_handler_list[i].handler_index == __LAST_REQEUST_HANDLER) {
			break;
		}
		AM_Safe(AM_SetHandler(__sc_endpoint_request,
							  request_handler_list[i].handler_index,
							  request_handler_list[i].func));
	}
		
	AM_Safe(AM_SetTag(__sc_endpoint_signal, SPLITC_TAG));
	AM_Safe(AM_SetTag(__sc_endpoint_request, SPLITC_TAG));

	___AM_MAX_SHORT = AM_MaxShort();
	___AM_MAX_MEDIUM = AM_MaxMedium();
	___AM_MAX_LONG = AM_MaxLong();
}

void setupGlunix(int argc, char *argv[])
{
	int 	num_nodes;
	int		i;
	char	hostname[256];
	
	if (Glib_Initialize() == 0) {
		fprintf(stderr, "Glib_Initialize failed.\n");
		fflush(stderr);
		exit(-1);
	}

	if (gethostname(hostname, 256)) {
		perror("gethostname");
		exit(-1);
	}
			
	if (Glib_AmIStartup()) {

		if (argc > 1) {
			/*
			 * argv[1] is number of nodes; argv[0] is program name; argv is
			 * list of arguments including program name and number of nodes.
			 * We need to remove argv[1] when the argument array is passed
			 * to the splitc_main().
			 */
			num_nodes = atoi(argv[1]);
			if (num_nodes < 1) {
				fprintf (stderr, "Split-C: Invalid number of nodes: %s\n", argv[1]);
				fprintf (stderr,
						 "Split-C: Usage '%s <num_nodes> {program arguments}'\n",
						 argv[0]);
						 
				exit (1);
			}
				
			i = Glib_GetNumAvailableNodes();
			if (i < num_nodes) {
				fprintf (stderr,
						 "Split-C: Error - Only %d nodes available (requested %d).\n",
						 i, num_nodes);
				exit (1);
			}
			
			for (i = 2; i <= argc; i++) {
				argv[i - 1] = argv[i];
			}
			argc--;
			/* Spawn only returns if it failed */

			printf("Split-C: Spawning %s on %d nodes\n", argv[0], num_nodes);
			fflush(stdout);
			if (Glib_Spawn(num_nodes, argv[0], argv)) {
				perror("Glib_Spawn Failed");
				exit(-1);
			}
		}
		else {
			___MYPROC = 0;				
			___PROCS = 1;
			/* Special case -- don't spawn; easier for debugging */
#if 0
			if (Glib_Spawn(1, argv[0], argv)) {
				perror("Glib_Spawn Failed");
				exit(-1);
			}
#endif

		}
	}
	/* This is a node in the parallel job */
	else {
		___MYPROC = Glib_GetMyVnn();
		___PROCS = Glib_GetParallelDegree();
		printf("Split-C: %s is node number %d of %d\n", hostname, MYPROC, PROCS);
		fflush(stdout);
	}
}


void message_thread()
{
	while (1) {
		
		AM_Safe(AM_WaitSema(__sc_bundle_request));
		AM_Poll(__sc_bundle_request);

#if 0
		AM_Safe(AM_SetEventMask(__sc_bundle_request, AM_NOTEMPTY));
#endif
	}
}



main(int argc, char **argv)
{
	thread_t 	thr_id;
	
	setupGlunix(argc, argv);
	setupAM();
	setupNames();

	/* Start up thread to handle request messages */
	if (thr_create(NULL, (size_t)0, message_thread, NULL, THR_NEW_LWP, &thr_id)) {
		perror("Creation of message_thread");
	}

	/*
	 * The message thread must run at a higher priority than the compute
	 * thread so that it will be given a chance to run.  The message thread
	 * yeilds the processor when it is waiting on a message.
	 */
	if (thr_setprio(thr_id, 1)) {
		perror("thr_setprio");
	}

	init_consts();

	on_one {
		printf("Split-C: Using dedicated message thread.\n");
		fflush(stdout);
		printf("Split-C: Using separate bundles for requests and replies.\n");
		fflush(stdout);
	}

	if (!Glib_Barrier()) {
		perror("Glib_Barrier Failed");
		exit(-1);
	}
	barrier();
	splitc_main(argc, argv);
	barrier();

	AM_Safe(AM_Terminate());
	exit(0);
}


