/* 									tab:8
 *
 * split-c.h - Support for Split-C extensions
 * 
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author:	Arvind Krishnamurthy
 * Version:	29
 * Filename:	split-c.h
 * History:
 *	ssl	20	Mon Mar 21 09:46:28 1994
 *		sync_ctr for HPAM can use poll but PMAM cannot
 *	ssl	19	Sun Mar 20 14:25:01 1994
 *		sync_ctr on hps can now use poll
 *	ssl	18	Fri Mar 11 16:10:16 1994
 *		sync_ctr must use am_wait_until_zero for hpam. am_poll does not
 *      retransmit
 *	ssl	15	Thu Nov 18 18:39:21 1993
 *		Moved include of print.h to cm5.h
 *	ssl	14	Tue Nov 16 11:48:02 1993
 *		Added macros for spread and global tags to allow compiling with native
 *      compiler.
 *	ssl	13	Thu Oct 28 20:30:03 1993
 *		Removed extra prototypes. Put cm5 specific stuff in cm5.h.
 *	LTL	12	Mon Aug 16 15:52:35 1993
 *		Fixed it up for version 112 to support CMMD
 *	ACD	10	Mon Jun 14 10:56:03 1993
 *		Changed spread_free interface
 *  SL  9   Wed May 12 23:23:40 PST 1993
 *      Added SC_PROFILE option
 * 	TvE	8	Sat Apr  3 22:50:44 PST 1993
 *		Changed include <cmam/cmam.h> to "cmam.h"
 *	SL	7	Wed Feb 10 16:25:30 1993
 *		Added debugging commands (SC_DEBUG).
 *	SL	6	Wed Jan 20 12:58:42 1993
 *		Added timer library prototypes.
 * 	TvE	5	Mon Jan 18 22:55:24 PST 1993
 *		Nuked barrier.h and added #define barrier just in front of
 *		#include of cmam.h.
 *		Added declaration of all_set_PROCS. Definition is in split-c.sc
 *		Changed definitions of PROCS, PROCSLOG, etc... slightly
 * 	TvE	4	Mon Jan  4 17:51:14 PST 1993
 *		Moved definitions of EXTERN and INLINE around because they
 *		conflicted with cmam.h.
 *		Added more public definitions and declarations.
 * 	TvE	3	Sat Dec 19 00:16:26 PST 1992
 *		Started major clean-up and overhaul
 *	SCG	1	Thu Dec  3 11:25:41 1992
 *		added fread and fseek
 * 	TvE	2	Tue Sep 29 12:43:37 PDT 1992
 *		include file name changes
 *	SCG	1	Wed Sep 23 00:11:02 1992
 *		made it compatible with split-c.internal.h
 */

#ifndef SPLIT_C_H
#define SPLIT_C_H

#include <stdio.h>

#ifndef __GNUC__
#include <split-c/ansi-c.h>
#endif

/* ............................
 *
 * Control debugging statements
 *
 * ............................
 */

#ifdef SPLIT_C_DEBUG
#define SC_DEBUG(a) a
#else
#define SC_DEBUG(a)
#endif

/* ........................................................
 * Macros for removing global and spread tags. For use when
 * compiling with native compiler. - ssl
 * ........................................................
 */

#ifdef NO_SPLITCC
#define ___C_GPTR ___globPtrType
#define ___I_GPTR ___globPtrType
#define ___V_GPTR ___globPtrType
#define ___V_SPTR ___globPtrType
#else
#define ___C_GPTR char *global
#define ___I_GPTR int *global
#define ___V_GPTR void *global
#define ___V_SPTR void *spread
#endif

/* ...........................
 *
 * Profiling routines and data
 *
 * ...........................
 */

#ifdef SC_PROFILE
#include <split-c/profile.h>
#else
#define prof_compute()	
#define prof_idle()	
#define all_do_profile(a)
#endif

/* ..........
 *
 * Processors
 *
 * ..........
 */

typedef int Processor;

/* Variables to hold convenient values relate to the number of processors */
extern int ___PROCS;	/* number of procs (<= partition size) */
extern int ___PROCSM1;	/* ___PROCS-1 */
extern int ___PROCSLOG;	/* log(___PROCS) or -1 if not power of two */
extern int ___MYPROC;	/* self address */

extern int ___AM_MAX_SHORT; /* size of short packet supported by AM */
extern int ___AM_MAX_MEDIUM; /* size of short packet supported by AM */
extern int ___AM_MAX_LONG; /* size of short packet supported by AM */


/* Sort out whether PROCS and CO. are constant or variable.
 * Note that __PROCS__ and __PROCSLOG__ are set by the compiler driver */
#if __PROCSLOG__ > 0
	/* number of processors is fixed power-of-two */
#	define MYPROC	(___MYPROC)
#	define PROCS	(1<<(__PROCSLOG__))
#	define PROCSM1	(PROCS-1)
#	define PROCSLOG	(__PROCSLOG__)
#	define PROCSMAX	PROCS
#	define ___PROCS_POWER_OF_TWO	/* compiled for power of two */
#	undef  ___PROCS_VARIABLE	/* not compiled for variable PROCS */
#elif defined(__PROCSLOG__)
	/* number of processors is unknown power-of-two */
#	define MYPROC	(___MYPROC)
#	define PROCS	(___PROCS)
#	define PROCSM1	(___PROCSM1)
#	define PROCSLOG	(___PROCSLOG)
#	define PROCSMAX	1024		/* Ha! Built a bigger one yet? */
#	define ___PROCS_POWER_OF_TWO	/* compiled for power of two */
#	define ___PROCS_VARIABLE	/* compiled for variable PROCS */
#elif __PROCS__ > 0
	/* number of processors is fixed non-power-of-two */
#	define MYPROC	(___MYPROC)
#	define PROCS	(__PROCS__)
#	define PROCSM1	(PROCS-1)
#	define PROCSLOG	(-1)
#	define PROCSMAX	PROCS
#	undef  ___PROCS_POWER_OF_TWO	/* not compiled for power of two */
#	undef  ___PROCS_VARIABLE	/* not compiled for variable PROCS */
#elif defined(__PROCS__)
	/* number of processors is unknown non-power-of-two */
#	define MYPROC	(___MYPROC)
#	define PROCS	(___PROCS)
#	define PROCSM1	(___PROCSM1)
#	define PROCSLOG	(-1)
#	define PROCSMAX	1024
#	undef  ___PROCS_POWER_OF_TWO	/* not compiled for power of two */
#	define ___PROCS_VARIABLE	/* compiled for variable PROCS */
#endif


/* .............................
 *
 * Control `extern' and `inline'
 *
 * .............................
 */

/* Defining SPLIT_C_DEBUG has the effect of disabling all inlining of functions
 * which helps when debugging. */
#if !defined(SPLIT_C_DEBUG)
#	define	SPLIT_C_INLINE
#	define	INLINE(code) {code}
#else
#	undef	SPLIT_C_INLINE
#	define	INLINE(code) ;
#endif

/* Defining SPLIT_C_INTERNAL_DEFINE has the effect of producing a
 * real (non-extern) version of all functions and variables. */
#undef EXTERN
#ifdef SPLIT_C_INTERNAL_DEFINE
#      define EXTERN
#      define SPLIT_C_INLINE
#      define INLINE(code) {code}
#else
#      define EXTERN extern
#endif

/* A hack for dealing with a native compiler that does not */
/* inline. - ssl */
#if !defined(__GNUC__) && !defined(SPLIT_C_INTERNAL_DEFINE)
#	undef	SPLIT_C_INLINE
#	define	INLINE(code) ;
#endif

/* ........................
 *
 * Message Layer
 *
 * ........................
 */
#include <thread.h>
#include <split-c/udpam.h>
#include <split-c/am.h>
#include <split-c/AM.h>

#define SPLITC_TAG		18


#ifdef 	SC_SEP_REQ_RPL
extern eb_t __sc_bundle_src;
extern eb_t __sc_bundle_dst;
extern ea_t __sc_endpoint_src;
extern ea_t __sc_endpoint_dst;
#else

extern eb_t __sc_bundle;
extern ea_t __sc_endpoint;

#define __sc_bundle_dst		__sc_bundle
#define __sc_bundle_src		__sc_bundle
#define __sc_endpoint_dst 	__sc_endpoint
#define __sc_endpoint_src 	__sc_endpoint
#endif



/* ........................
 *
 * Synchronization counters
 *
 * ........................
 */
typedef volatile int Counter;


EXTERN inline void splitc_poll()
INLINE({
	AM_Poll(__sc_bundle_src);
})


/* Native compiler on Paragon won't allow ifdef in macros. - ssl */
#if SC_PROFILE 
EXTERN inline void sync_ctr(Counter *ctr)
INLINE({
	while(*ctr) {
		sc_prof_data[PROF_SYNC]++;
		splitc_poll();
	}
})
#else
EXTERN inline void sync_ctr(Counter *ctr)
INLINE({
	while(*ctr) {
		splitc_poll();
	}
})
#endif

EXTERN inline int is_sync_ctr(Counter *ctr)
INLINE({
	splitc_poll(); return !*ctr;
})

/* ...............
 *
 * Global pointers
 *
 * ...............
 */

#include <split-c/pointers.h>


/* .............
 *
 * Handlers
 *
 * .............
 */

#include <split-c/handlers.h>

/* .............
 *
 * Global memory
 *
 * .............
 */

/* Constants specifying when to use segments
 * (these should go somewhere else... -TvE */
#define		STORE_STRUCT_LEN	60
#define		PUT_STRUCT_LEN		32

/* Defines splitc_lock and unlock for atomic handlers */
#include <split-c/lock.h>

/* Ack-ed reads and writes */
#include <split-c/memory.h>

/* Non-ack-ed stores */
#include <split-c/store.h>

#include <split-c/bulk.h>

/*************************************************************
 *
 * Global communication
 * 
 *************************************************************/
#include <split-c/com.h>

/* ..............
 *
 * Global control
 *
 * ..............
 */

EXTERN void all_set_PROCS(int new_PROCS);
#include <split-c/control.h> /* public stuff. Maybe should be here verbatim? */

/* ........................
 *
 * Barrier
 *
 * ........................
 */
#include <split-c/barrier.h>

/* .........................
 *
 * Dynamic memory allocation
 *
 * .........................
 */

/*EXTERN ___V_SPTR spread_malloc(unsigned num_elems, unsigned elem_sz);*/
/*EXTERN ___V_SPTR all_spread_malloc(unsigned num_elems, unsigned elem_s);*/
/*EXTERN void spread_free(___V_SPTR ptr);*/
/*EXTERN ___V_SPTR all_spread_malloc(unsigned size);*/
/*EXTERN void all_spread_free(___V_SPTR ptr);*/
#include <split-c/malloc.h>

/* .........................
 *
 * Timer library
 *
 * .........................
 */

EXTERN double get_seconds();

#include <atomic.h>

#endif
