/*									tab:8
 *
 * main.c - Main program for the HP PA machines with the Medusa
 *
 * 
 * "Copyright (c) 1993 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 		Richard P. Martin 
 * Version:		3
 * Creation Date:	Sat Nov 20 16:47:42 1993
 * Filename:		main.c
 * History:
 *	RPM	3	Tue Mar  5 16:28:29 1996
 *		added splitc_print_debug
 *	ssl	2	Mon Mar 14 11:14:45 1994
 *		added barrier to end of main
 */

#include <stdio.h>
#define SPLIT_C_INTERNAL_DEFINE	/* turns extern's to not externs */
#include <split-c/split-c.h>
#include <time.h>

#define DFLT_NUM_NODES 0
#define DFLT_PART NULL

const int ___PROCS;
const int ___PROCSM1;
const int ___PROCSLOG;
const int ___MYPROC;
const int ___AM_MAX_PKT;

int splitc_print_debug; 


/* return log2 if exact, else return -1 */
getLog(int i)
{
	int exp = 0;

	if(i <= 0) return -1;
	for(exp=0; (i&1)==0; i>>=1) exp++;
	return i==1 ? exp : -1;
}

init_consts()
{
	(int)___MYPROC = am_my_proc();
	(int)___PROCS =  am_procs();
	(int)___PROCSM1 = (int) ___PROCS - 1;
	(int)___PROCSLOG = getLog((int) ___PROCS);
	(int)___AM_MAX_PKT = am_max_size();

#if __PROCSLOG__ > 0
	/* number of processors is fixed power-of-two */
	if (___PROCSLOG != __PROCSLOG__)
	{
		on_one
			fprintf(stderr, "FATAL error: "
			    "attempt to set PROCS (%d) to other than\n"
			    "value compiled into program (%d)\n", ___PROCS,
			    __PROCS__);
		exit(1);
	}
#elif defined(__PROCSLOG__)
	/* number of processors is unknown power-of-two */
	if (___PROCSLOG == -1)
	{
		on_one
			fprintf(stderr, "FATAL error: "
			    "attempt to set PROCS (%d) to non power-of-two\n"
			    "value with program compiled for power of two\n",
			    ___PROCS);
		exit(1);
	}
#elif __PROCS__ > 0
	/* number of processors is fixed non-power-of-two */
	if (___PROCS != __PROCS__)
	{
		on_one
			fprintf(stderr, "FATAL error: "
			    "attempt to set PROCS (%d) to other than\n"
			    "value compiled into program (%d)\n", ___PROCS,
			    __PROCS__);
		exit(1);
	}
#endif

	/*spread_malloc_init();*/
	atomic_init();

	if (splitc_print_debug) {
	  fprintf(stderr," %d about to call split-c com_init\n",MYPROC);
	}

	com_init();
#ifdef SC_PROFILE	
	profile_init();
#endif	
}

void srandom(long seed)
{
  srand48(seed);
}

long random() 
{
  return (lrand48());
}


struct timeval splitc_base_time;

main(int argc, char **argv, char **envp)
{
  struct timezone tz;
  int i;
  int use_glunix; 
  /* create application */
  use_glunix =1;

  if (getenv("SPLITC_DEBUG") != NULL)
    splitc_print_debug=1;


  if (splitc_print_debug)
    printf("about to call am_enable()\n");


  /* fork child processes onto all nodes of application */
  if (getenv("SPLITC_NO_GLUNIX") != NULL) {
    use_glunix =0;
    am_enable(use_glunix); /* 1= use glunix, 0 = do not use glunix */
  }
  else {
    lam_glunix_spawn(&argc,argv,envp);
    am_enable(1);
  }

  if (splitc_print_debug) {
    fprintf(stderr,"completed am_enable\n");
    fflush(stderr);
  }

  init_consts();

  if (splitc_print_debug) {
    fprintf(stderr,"Node %d initialized constants\n",MYPROC);
    fflush(stderr);
  }

  gettimeofday(&splitc_base_time,&tz);

  if (splitc_print_debug) {
    fprintf(stderr,"Node %d entering split-c main\n",MYPROC);
    fflush(stderr);
  }
  splitc_main(argc, argv);

  am_disable();
}
