/*									tab:4
 *
 * X_handlers.c - Split-C functions for X
 *
 * "Copyright (c) 1993 by Steve Lumetta and The Regents of the University 
 * of California.  All rights reserved."
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * Author: 			Lok Tin Liu
 * Version:			5
 * Creation Date:	Sun Nov  7 16:59:46 1993
 * Filename:		X_handlers.c
 * History:
 *	ssl	4	Wed Feb 16 19:01:52 1994
 *		Cleaned up for use on the HPs.
 *	SL	3	Mon Jan 17 16:30:07 1994
 *		Fixed bug in declaration of CMPE_printGetChannel
 *	LTL	2	Thu Nov 25 15:41:14 1993
 *		Fixed bug in setting host handlers
 *	LTL	1	Sun Nov  7 17:00:02 1993
 *		Initial release. X stuff is not in libsplit-c anymore.
 */

/* This used to be X functions that are used to display images on the host
 * of the CM-5. It is modified to be used on one of the nodes (usually node 0).
 */
	 

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
 
static Display *X_Display = NULL;
static Window X_Window;
static GC X_GC;
static int X_Size;
static XImage *X_Image;
static char *X_Data;
static unsigned long X_Pixels[256];

void imageXDraw() 
{
	if (X_Image != NULL) {
		XPutImage(X_Display, X_Window, X_GC, X_Image, 0, 0, 0, 0,
				  X_Size, X_Size);
		listenX();
	}
}

listenX()
{
	XEvent event;

    while (XPending(X_Display) > 0) {
		XNextEvent(X_Display, &event);
		if (event.type == Expose) {
			while (XCheckTypedEvent(X_Display, Expose, &event));
			imageXDraw();
        }
    }
}

void X_init(int two, int three, int four)
{
	/* two - size of display window */
	
    int rgbOfs, rgbRange, i;
    Screen scr;
    int scrNum;
    XSizeHints size_hints;
	XColor SetColors[256];
	Colormap colorMap;
	Visual *vis;

	unsigned long masks[256]; /* These are ignored */

    X_Display = XOpenDisplay(NULL);
    if (X_Display == NULL) {
		fprintf(stderr, "Could not open display.\n");
		exit(1);
    }
    scrNum = DefaultScreen(X_Display);
    colorMap = DefaultColormap(X_Display, scrNum);
    X_GC = DefaultGC(X_Display, scrNum);
	
	X_Data = (char *)malloc(two*two);
    if (XAllocColorCells(X_Display, colorMap, True, masks, 0, X_Pixels, 4)) {
		for (i = 0; i < 4; i++) {
			SetColors[i].pixel = X_Pixels[i];
			SetColors[i].flags = DoRed|DoGreen|DoBlue;
		}
		SetColors[0].red = 0;         /* Blue, green, red, and white */
		SetColors[0].green = 0;
		SetColors[0].blue = 60000;
		SetColors[1].red = 0;
		SetColors[1].green = 60000;
		SetColors[1].blue = 0;
		SetColors[2].red = 60000;
		SetColors[2].green = 0;
		SetColors[2].blue = 0;
		SetColors[3].red = 60000;
		SetColors[3].green = 60000;
		SetColors[3].blue = 60000;
		XStoreColors(X_Display, colorMap, SetColors, 4);
		X_Image = XCreateImage(X_Display, DefaultVisual(X_Display, scrNum),
							   8, ZPixmap, 0, X_Data, two, two, 8, 0);
	}
	else {
		printf("You must use a color or grey-scale monitor.\n");
		exit(1);
/*		X_Pixels[0] = BlackPixel(X_Display, scrNum);
		X_Pixels[1] = X_Pixels[2] = X_Pixels[3] =
			WhitePixel(X_Display, scrNum);
		X_Image = XCreateImage(X_Display, DefaultVisual(X_Display, scrNum),
							   1, ZPixmap, 0, X_Data, two, two, 8, 0);*/

	}
	for (i = 0; i < two*two; i++)
		X_Data[i] = X_Pixels[0];
	X_Window = XCreateSimpleWindow(X_Display,
								   RootWindow(X_Display, scrNum),
								   100, 100, two, two, 5, 1, 0);
	if (X_Window == (Window)NULL) {
		fprintf(stderr, "Could not open window.\n");
		exit(1);
	}
	size_hints.flags = PPosition|PSize|PMinSize;
	size_hints.x = 100;
	size_hints.y = 100;
	size_hints.width = two;
	size_hints.height = two;
	size_hints.min_width = two;
    size_hints.min_height = two;
    XSetStandardProperties(X_Display, X_Window, "Split-C Display",
						   "Split-C Display", None, NULL, 0, &size_hints);
    XSelectInput(X_Display, X_Window, ExposureMask);
    XMapWindow(X_Display, X_Window);
    X_Size = two;
    listenX();
}

void X_point(int two, int three, int four)
{
	/* two - X position of point */
	/* three - Y position of point */
	/* four - color of point */

	X_Data[three*X_Size + two] = X_Pixels[four];
}

void X_show(int three, int four)
{
	imageXDraw();
}

