/*                                                                      tab:4

 * com.header - Header file to generate com.sc
 *
 * 
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author:                      Arvind Krishnamurthy
 * Version:                     6
 * Creation Date:       Apr  2 13:42:47 1993
 * Filename:            com.sc
 * History:              
 *      ssl     6       Thu Oct 28 23:41:04 1993
 *              com.h and cm/cmmd.h are now included with split-c.h
 *      LTL     4       Mon Aug 16 15:33:59 1993
 *              Fixed it up for version 112 to support CMMD
 *      AK      2       Fri Apr  2 13:42:47 1993
 *              Obtained from David's initial version of com library.
 */
static char _version_string_[] = "\nVersion:6:com.sc\0\n";

#include <split-c/split-c.h>

/*
   This library provides global communication support for Split-C, including
   broadcasts, reductions, and scans.  

   It provides operations in which all processors participate.
   There are two basic styles: a processor-oriented functional interface
   and a spread array interface.

   Currently only the processor interface is supported.

   Eventually it will also provide autonomous operations.

   An all_ routine end with a barrier, so that there is no conflict on
   use of internal state.  Routines that consume a spread array also
   begin with a barrier, to reduce the chance that it is still being
   produced when the routine is called.

 */

#define ADD(A, B)	(A + B)
#define MAX(A, B)	(A < B ? B : A)
#define MIN(A, B)	(A < B ? A : B)
#define MULT(A, B)	(A * B)
#define OR(A, B)	(A | B)
#define XOR(A, B)	(A ^ B)
#define AND(A, B)	(A & B)

static int ack_slot;
static Counter red_rctr, red_sctr;
static long long reduce_buf[2];
static Counter bcast_rctr, bcast_sctr;
static long long bcast_buf;

static int phase = 0;
static Counter scan_rctr[2], scan_sctr[2];
static long long scan_buf[20];
static long long temp_scan_val[20];
static long long scan_result_buf;


#define ADD_FN {\
		  int n=*nel;\
		  while (n-- > 0)\
		  *accum++ += *partial++;}

#define MULT_FN {\
		  int n=*nel;\
		  while (n-- > 0)\
		  *accum++ *= *partial++;}
#define MAX_FN {\
		  int n=*nel;\
		  while (n-- > 0){\
		  *accum = (*accum >= *partial)?(*accum):(*partial);\
				    accum++;partial++;	}}

#define MIN_FN {\
		  int n=*nel;\
		  while (n-- > 0){\
		  *accum = (*accum <= *partial)?(*accum):(*partial);\
				    accum++;partial++;	}}
#define OR_FN {\
		 int n=*nel;\
		 while (n-- > 0)\
		 *accum++ |= *partial++;}
		  
#define XOR_FN {\
		  int n=*nel;\
		  while (n-- > 0)\
		  *accum++ ^= *partial++;}

#define AND_FN {\
		  int n=*nel;\
		  while (n-- > 0)\
		  *accum++ &= *partial++;}
		    
/*Summation*/
static void i_sum (int *accum, int *partial, int *nel)
{
    ADD_FN;
}
		
static void ui_sum (unsigned int *accum, unsigned int *partial, int *nel)
{
  ADD_FN;
}

static void f_sum (float *accum, float *partial, int *nel)
{
  ADD_FN;
}
static void d_sum (double *accum, double *partial, int *nel)
{
  ADD_FN;
}

/*Multiplication*/
static void i_mult (int *accum, int *partial, int *nel)
{
  MULT_FN;
}
static void ui_mult (unsigned int *accum, unsigned int *partial, int *nel)
{
  MULT_FN;
}
static void f_mult (float *accum, float *partial, int *nel)
{
  MULT_FN;
}
static void d_mult (double *accum, double *partial, int *nel)
{
  MULT_FN;
}

/*Maximum*/
static void i_max (int *accum, int *partial, int *nel)
{
  MAX_FN;
}
static void ui_max (unsigned int *accum, unsigned int *partial, int *nel)
{
  MAX_FN;
}
static void f_max (float *accum, float *partial, int *nel)
{
  MAX_FN;
}
static void d_max (double *accum, double *partial, int *nel)
{
  MAX_FN;
}
/*Minimum*/
static void i_min (int *accum, int *partial, int *nel)
{
  MIN_FN;
}
static void ui_min (unsigned int *accum, unsigned int *partial, int *nel)
{
  MIN_FN;
}
static void f_min (float *accum, float *partial, int *nel)
{
  MIN_FN;
}
static void d_min (double *accum, double *partial, int *nel)
{
  MIN_FN;
}

/*Bitwise OR*/
static void i_or (int *accum, int *partial, int *nel)
{
  OR_FN;
}
static void ui_or (unsigned int *accum, unsigned int *partial, int *nel)
{
  OR_FN;
}

/*Bitwise exclusive OR*/
static void i_xor (int *accum, int *partial, int *nel)
{
  XOR_FN;
}
static void ui_xor (unsigned int *accum, unsigned int *partial, int *nel)
{
  XOR_FN;
}

/*Bitwise AND*/
static void i_and (int *accum, int *partial, int *nel)
{
  AND_FN;
}
static void ui_and (unsigned int *accum, unsigned int *partial, int *nel)
{
  AND_FN;
}

/*Reduction functions*/
int all_reduce_to_one_add (int val)
{
  gam_reduce(ew_base.segGroup,i_sum,(caddr_t)&val,sizeof(int),1,0);
  return val;
}

int all_reduce_to_all_add (int val)
{
  gam_reduce(ew_base.segGroup,i_sum,(caddr_t)&val,sizeof(int),1,0);
  return val;
}

unsigned int 
  all_reduce_to_one_uadd (unsigned int val)
{
  gam_reduce(ew_base.segGroup,ui_sum,(caddr_t)&val,sizeof(int),1,0);
  return val;
}

unsigned int 
  all_reduce_to_all_uadd (unsigned int val)
{
  gam_reduce(ew_base.segGroup,ui_sum,(caddr_t)&val,sizeof(int),1,0);
  return val;
}


float 
  all_reduce_to_one_fadd (float val) 
{
  gam_reduce(ew_base.segGroup,f_sum,(caddr_t)&val,sizeof(float),1,0);
  return val;
}

float 
  all_reduce_to_all_fadd (float val) 
{
  gam_reduce(ew_base.segGroup,f_sum,(caddr_t)&val,sizeof(float),1,0);
  return val;
}

double 
  all_reduce_to_one_dadd (double val)
{
  gam_reduce(ew_base.segGroup,d_sum,(caddr_t)&val,sizeof(double),1,0);
  return val;
}

double 
  all_reduce_to_all_dadd (double val)
{
  gam_reduce(ew_base.segGroup,d_sum,(caddr_t)&val,sizeof(double),1,0);
  return val;
}

int 
  all_bcast (int val)
{
  gam_bcast(ew_base.segGroup,0,(char*)&val,sizeof(int),0); /*global?*/
  return (val);
}

int 
  all_bcast_i (int val)
{
  gam_bcast(ew_base.segGroup,0,(char*)&val,sizeof(int),0); /*global?*/
  return (val);
}

double 
all_bcast_d (double val)
{
  gam_bcast(ew_base.segGroup,0,(char*)&val,sizeof(double),0); /*global?*/
  return (val);
}


unsigned int
all_bcast_u (unsigned int val)
{
  gam_bcast(ew_base.segGroup,0,(char*)&val,sizeof(unsigned int),0); /*global?*/
  return (val);
}

int all_reduce_to_one_mult (int val)
{
  gam_reduce(ew_base.segGroup,i_mult,(caddr_t)&val,sizeof(int),1,0);
  return val;
}

int all_reduce_to_all_mult (int val)
{
  gam_reduce(ew_base.segGroup,i_mult,(caddr_t)&val,sizeof(int),1,0);
  return val;
}
unsigned int 
  all_reduce_to_one_umult (unsigned int val)
{
  gam_reduce(ew_base.segGroup,ui_mult,(caddr_t)&val,sizeof(int),1,0);
  return val;
}
unsigned int 
  all_reduce_to_all_umult (unsigned int val)
{
  gam_reduce(ew_base.segGroup,ui_mult,(caddr_t)&val,sizeof(int),1,0);
  return val;
}
float 
  all_reduce_to_one_fmult (float val)
{
  gam_reduce(ew_base.segGroup,f_mult,(caddr_t)&val,sizeof(float),1,0);
  return val;
}
float 
  all_reduce_to_all_fmult (float val)
{
  gam_reduce(ew_base.segGroup,f_mult,(caddr_t)&val,sizeof(float),1,0);
  return val;
}
double 
  all_reduce_to_one_dmult (double val)
{
  gam_reduce(ew_base.segGroup,d_mult,(caddr_t)&val,sizeof(double),1,0);
  return val;
}
double 
  all_reduce_to_all_dmult (double val)
{
  gam_reduce(ew_base.segGroup,d_mult,(caddr_t)&val,sizeof(double),1,0);
  return val;
}

int 
  all_reduce_to_one_max (int val)
{
  gam_reduce(ew_base.segGroup,i_max,(caddr_t)&val,sizeof(int),1,0);
  return val;
}
int 
  all_reduce_to_all_max (int val)
{
  gam_reduce(ew_base.segGroup,i_max,(caddr_t)&val,sizeof(int),1,0);
  return val;
}

unsigned int 
  all_reduce_to_one_umax (unsigned int val)
{
  gam_reduce(ew_base.segGroup,ui_max,(caddr_t)&val,sizeof(int),1,0);
  return val;
}
unsigned int 
  all_reduce_to_all_umax (unsigned int val)
{
  gam_reduce(ew_base.segGroup,ui_max,(caddr_t)&val,sizeof(int),1,0);
  return val;
}
float 
  all_reduce_to_one_fmax (float val)
{
  gam_reduce(ew_base.segGroup,f_max,(caddr_t)&val,sizeof(float),1,0);
  return val;
}
float 
  all_reduce_to_all_fmax (float val)
{
  gam_reduce(ew_base.segGroup,f_max,(caddr_t)&val,sizeof(float),1,0);
  return val;
}
double 
  all_reduce_to_one_dmax (double val)
{
  gam_reduce(ew_base.segGroup,d_max,(caddr_t)&val,sizeof(double),1,0);
  return val;
}
double 
  all_reduce_to_all_dmax (double val)
{
  gam_reduce(ew_base.segGroup,d_max,(caddr_t)&val,sizeof(double),1,0);
  return val;
}
int 
  all_reduce_to_one_min (int val)
{
  gam_reduce(ew_base.segGroup,i_min,(caddr_t)&val,sizeof(int),1,0);
  return val;
}
int 
  all_reduce_to_all_min (int val)
{
  gam_reduce(ew_base.segGroup,i_min,(caddr_t)&val,sizeof(int),1,0);
  return val;
}
unsigned int 
  all_reduce_to_one_umin (unsigned int val)
{
  gam_reduce(ew_base.segGroup,ui_min,(caddr_t)&val,sizeof(int),1,0);
  return val;
}
unsigned int 
  all_reduce_to_all_umin (unsigned int val)
{
  gam_reduce(ew_base.segGroup,ui_min,(caddr_t)&val,sizeof(int),1,0);
  return val;
}
float 
  all_reduce_to_one_fmin (float val)
{
  gam_reduce(ew_base.segGroup,f_min,(caddr_t)&val,sizeof(float),1,0);
  return val;
}
float 
  all_reduce_to_all_fmin (float val)
{
  gam_reduce(ew_base.segGroup,f_min,(caddr_t)&val,sizeof(float),1,0);
  return val;
}
double
  all_reduce_to_one_dmin (double val)
{
    gam_reduce(ew_base.segGroup,d_min,(caddr_t)&val,sizeof(double),1,0);
    return val;
}
double
  all_reduce_to_all_dmin (double val)
{
    gam_reduce(ew_base.segGroup,d_min,(caddr_t)&val,sizeof(double),1,0);
    return val;
}

int 
  all_reduce_to_one_or (int val)
{
  gam_reduce(ew_base.segGroup,i_or,(caddr_t)&val,sizeof(int),1,0);
  return val;
  }
int 
  all_reduce_to_all_or (int val)
{
  gam_reduce(ew_base.segGroup,i_or,(caddr_t)&val,sizeof(int),1,0);
  return val;
  }
unsigned int 
  all_reduce_to_one_uor (unsigned int val)
{
  gam_reduce(ew_base.segGroup,ui_or,(caddr_t)&val,sizeof(int),1,0);
  return val;
}
unsigned int 
  all_reduce_to_all_uor (unsigned int val)
{
  gam_reduce(ew_base.segGroup,ui_or,(caddr_t)&val,sizeof(int),1,0);
  return val;
}
int 
  all_reduce_to_one_xor (int val)
{
  gam_reduce(ew_base.segGroup,i_xor,(caddr_t)&val,sizeof(int),1,0);
  return val;
}
int 
  all_reduce_to_all_xor (int val)
{
  gam_reduce(ew_base.segGroup,i_xor,(caddr_t)&val,sizeof(int),1,0);
  return val;
}
unsigned int 
  all_reduce_to_one_uxor (unsigned int val)
{
  gam_reduce(ew_base.segGroup,ui_xor,(caddr_t)&val,sizeof(int),1,0);
  return val;
}
unsigned int 
  all_reduce_to_all_uxor (unsigned int val)
{
  gam_reduce(ew_base.segGroup,ui_xor,(caddr_t)&val,sizeof(int),1,0);
  return val;
}
int 
all_reduce_to_one_and (int val)
{
  gam_reduce(ew_base.segGroup,i_and,(caddr_t)&val,sizeof(int),1,0);
  return val;
}
int 
all_reduce_to_all_and (int val)
{
  gam_reduce(ew_base.segGroup,i_and,(caddr_t)&val,sizeof(int),1,0);
  return val;
}
unsigned int 
all_reduce_to_one_uand (unsigned int val)
{
  gam_reduce(ew_base.segGroup,ui_and,(caddr_t)&val,sizeof(int),1,0);
  return val;
}
unsigned int 
all_reduce_to_all_uand (unsigned int val)
{
  gam_reduce(ew_base.segGroup,ui_and,(caddr_t)&val,sizeof(int),1,0);
  return val;
}

int 
all_scan_add (int val)
{
  int dest;
  int i, count;
  int bit_vec = 1;
  int scan_val;
  long long *local_buf = (long long *) scan_buf;
  scan_sctr[phase] = scan_rctr[phase] = 0;
  phase = (phase + 1) % 2;
  *((int *) &temp_scan_val[0]) = scan_val = val;
  for (count = 0; bit_vec != PROCS; count++, bit_vec = bit_vec << 1)
    {
      dest = MYPROC & (~bit_vec);
      if (dest != MYPROC)
	break;
    }
  store_sync_ctr (count * sizeof (int), &scan_rctr
		  [phase], &scan_sctr[phase]);
  for (i = 1; i <= count; i++)
    {
      scan_val = ADD (scan_val, *((int *) &local_buf[i]));
      *((int *) &temp_scan_val[i]) = scan_val;
    }
  __i_store_ctr (toglobal (dest, &scan_buf[count + 1]), scan_val, &scan_rctr[phase], &scan_sctr[phase]);
  count--;
  bit_vec = bit_vec >> 1;
  if (MYPROC != 0)
    {
      int result_val;
      store_sync_ctr (sizeof (int), &scan_rctr[phase], &scan_sctr[phase]);
      result_val = *((int
		      *) &scan_result_buf);
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store_ctr (toglobal (dest, &scan_result_buf), ADD (result_val, *((int *) &temp_scan_val[count])), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return ADD (result_val, val);
    }
  else
    {
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store_ctr (toglobal (dest, &scan_result_buf
			 ), *((int *) &temp_scan_val[count]), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return val;
    }
}
unsigned int 
all_scan_uadd (unsigned int val)
{
  int dest;
  int i, count;
  int bit_vec = 1;
  unsigned int scan_val;
  long long *local_buf = (long long *) scan_buf;
  scan_sctr[phase] = scan_rctr[phase] = 0;
  phase = (phase + 1) % 2;
  *((unsigned int *) &temp_scan_val[0]) = scan_val = val;
  for (count = 0; bit_vec != PROCS; count++, bit_vec = bit_vec << 1)
    {
      dest = MYPROC & (~bit_vec);
      if (dest != MYPROC)
	break;
    }
  store_sync_ctr
    (count * sizeof (unsigned int), &scan_rctr[phase], &scan_sctr[phase]);
  for (i = 1; i <= count; i++)
    {
      scan_val = ADD (scan_val, *((unsigned int *) &local_buf[i]));
      *((unsigned int *) &temp_scan_val[i]) = scan_val;
    }
  __i_store_ctr (toglobal (dest, &scan_buf[count + 1]), scan_val, &scan_rctr[phase], &scan_sctr[phase]);
  count--;
  bit_vec = bit_vec >> 1;
  if (MYPROC != 0)
    {
      unsigned int result_val;
      store_sync_ctr (sizeof
		      (unsigned int), &scan_rctr[phase], &scan_sctr[phase]);
      result_val = *((unsigned int *) &scan_result_buf);
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store_ctr (toglobal (dest, &scan_result_buf), ADD (result_val, *((unsigned int *) &temp_scan_val[count])), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return ADD (result_val, val);
    }
  else
    {
      for (; count >= 0; count--, bit_vec
	   = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store_ctr (toglobal (dest, &scan_result_buf), *((unsigned int *) &temp_scan_val[count]), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return val;
    }
}
float 
all_scan_fadd (float val)
{
  int dest;
  int i, count;
  int bit_vec = 1;
  float scan_val;
  long long *local_buf = (long long *) scan_buf;
  scan_sctr[phase] = scan_rctr[phase] = 0;
  phase = (phase + 1) % 2;
  *((float *) &temp_scan_val[0]) = scan_val = val;
  for (count = 0; bit_vec != PROCS; count++, bit_vec = bit_vec << 1)
    {
      dest = MYPROC & (~bit_vec);
      if (dest != MYPROC)
	break;
    }
  store_sync_ctr (count * sizeof (float
		  ), &scan_rctr[phase], &scan_sctr[phase]);
  for (i = 1; i <= count; i++)
    {
      scan_val = ADD (scan_val, *((float *) &local_buf[i]));
      *((float *) &temp_scan_val[i]) = scan_val;
    }
  __f_store_ctr (toglobal (dest, &scan_buf[count + 1]), scan_val, &scan_rctr[phase], &scan_sctr[phase]);
  count--;
  bit_vec = bit_vec >> 1;
  if (MYPROC != 0)
    {
      float result_val;
      store_sync_ctr (sizeof (float), &scan_rctr[phase], &scan_sctr[phase]
      );
      result_val = *((float *) &scan_result_buf);
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __f_store_ctr (toglobal (dest, &scan_result_buf), ADD (result_val, *((float *) &temp_scan_val[count])), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return ADD (result_val, val);
    }
  else
    {
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __f_store_ctr (toglobal
			 (dest, &scan_result_buf), *((float *) &temp_scan_val[count]), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return val;
    }
}
double 
all_scan_dadd (double val)
{
  int dest;
  int i, count;
  int bit_vec = 1;
  double scan_val;
  long long *local_buf = (long long *) scan_buf;
  scan_sctr[phase] = scan_rctr[phase] = 0;
  phase = (phase + 1) % 2;
  *((double *) &temp_scan_val[0]) = scan_val = val;
  for (count = 0; bit_vec != PROCS; count++, bit_vec = bit_vec << 1)
    {
      dest = MYPROC & (~bit_vec);
      if (dest != MYPROC)
	break;
    }
  store_sync_ctr (count * sizeof
		  (double), &scan_rctr[phase], &scan_sctr[phase]);
  for (i = 1; i <= count; i++)
    {
      scan_val = ADD (scan_val, *((double *) &local_buf[i]));
      *((double *) &temp_scan_val[i]) = scan_val;
    }
  __d_store_ctr (toglobal (dest, &scan_buf[count + 1]), scan_val, &scan_rctr[phase], &scan_sctr[phase]);
  count--;
  bit_vec = bit_vec >> 1;
  if (MYPROC != 0)
    {
      double result_val;
      store_sync_ctr (sizeof (double), &scan_rctr[phase], &scan_sctr
		      [phase]);
      result_val = *((double *) &scan_result_buf);
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __d_store_ctr (toglobal (dest, &scan_result_buf), ADD (result_val, *((double *) &temp_scan_val[count])), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return ADD (result_val, val);
    }
  else
    {
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __d_store_ctr
	    (toglobal (dest, &scan_result_buf), *((double *) &temp_scan_val[count]), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return val;
    }
}
int 
all_scan_mult (int val)
{
  int dest;
  int i, count;
  int bit_vec = 1;
  int scan_val;
  long long *local_buf = (long long *) scan_buf;
  scan_sctr[phase] = scan_rctr[phase] = 0;
  phase = (phase + 1) % 2;
  *((int *) &temp_scan_val[0]) = scan_val = val;
  for (count = 0; bit_vec != PROCS; count++, bit_vec = bit_vec << 1)
    {
      dest = MYPROC & (~bit_vec);
      if (dest != MYPROC)
	break;
    }
  store_sync_ctr (count * sizeof (int), &scan_rctr
		  [phase], &scan_sctr[phase]);
  for (i = 1; i <= count; i++)
    {
      scan_val = MULT (scan_val, *((int *) &local_buf[i]));
      *((int *) &temp_scan_val[i]) = scan_val;
    }
  __i_store_ctr (toglobal (dest, &scan_buf[count + 1]), scan_val, &scan_rctr[phase], &scan_sctr[phase]);
  count--;
  bit_vec = bit_vec >> 1;
  if (MYPROC != 0)
    {
      int result_val;
      store_sync_ctr (sizeof (int), &scan_rctr[phase], &scan_sctr[phase]);
      result_val = *((int
		      *) &scan_result_buf);
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store_ctr (toglobal (dest, &scan_result_buf), MULT (result_val, *((int *) &temp_scan_val[count])), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return MULT (result_val, val);
    }
  else
    {
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store_ctr (toglobal (dest, &scan_result_buf
			 ), *((int *) &temp_scan_val[count]), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return val;
    }
}
unsigned int 
all_scan_umult (unsigned int val)
{
  int dest;
  int i, count;
  int bit_vec = 1;
  unsigned int scan_val;
  long long *local_buf = (long long *) scan_buf;
  scan_sctr[phase] = scan_rctr[phase] = 0;
  phase = (phase + 1) % 2;
  *((unsigned int *) &temp_scan_val[0]) = scan_val = val;
  for (count = 0; bit_vec != PROCS; count++, bit_vec = bit_vec << 1)
    {
      dest = MYPROC & (~bit_vec);
      if (dest != MYPROC)
	break;
    }
  store_sync_ctr
    (count * sizeof (unsigned int), &scan_rctr[phase], &scan_sctr[phase]);
  for (i = 1; i <= count; i++)
    {
      scan_val = MULT (scan_val, *((unsigned int *) &local_buf[i]));
      *((unsigned int *) &temp_scan_val[i]) = scan_val;
    }
  __i_store_ctr (toglobal (dest, &scan_buf[count + 1]), scan_val, &scan_rctr[phase], &scan_sctr[phase]);
  count--;
  bit_vec = bit_vec >> 1;
  if (MYPROC != 0)
    {
      unsigned int result_val;
      store_sync_ctr (sizeof
		      (unsigned int), &scan_rctr[phase], &scan_sctr[phase]);
      result_val = *((unsigned int *) &scan_result_buf);
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store_ctr (toglobal (dest, &scan_result_buf), MULT (result_val, *((unsigned int *) &temp_scan_val[count])), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return MULT (result_val, val);
    }
  else
    {
      for (; count >= 0; count--, bit_vec
	   = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store_ctr (toglobal (dest, &scan_result_buf), *((unsigned int *) &temp_scan_val[count]), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return val;
    }
}
float 
all_scan_fmult (float val)
{
  int dest;
  int i, count;
  int bit_vec = 1;
  float scan_val;
  long long *local_buf = (long long *) scan_buf;
  scan_sctr[phase] = scan_rctr[phase] = 0;
  phase = (phase + 1) % 2;
  *((float *) &temp_scan_val[0]) = scan_val = val;
  for (count = 0; bit_vec != PROCS; count++, bit_vec = bit_vec << 1)
    {
      dest = MYPROC & (~bit_vec);
      if (dest != MYPROC)
	break;
    }
  store_sync_ctr (count * sizeof (float
		  ), &scan_rctr[phase], &scan_sctr[phase]);
  for (i = 1; i <= count; i++)
    {
      scan_val = MULT (scan_val, *((float *) &local_buf[i]));
      *((float *) &temp_scan_val[i]) = scan_val;
    }
  __f_store_ctr (toglobal (dest, &scan_buf[count + 1]), scan_val, &scan_rctr[phase], &scan_sctr[phase]);
  count--;
  bit_vec = bit_vec >> 1;
  if (MYPROC != 0)
    {
      float result_val;
      store_sync_ctr (sizeof (float), &scan_rctr[phase], &scan_sctr[phase
		      ]);
      result_val = *((float *) &scan_result_buf);
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __f_store_ctr (toglobal (dest, &scan_result_buf), MULT (result_val, *((float *) &temp_scan_val[count])), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return MULT (result_val, val);
    }
  else
    {
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __f_store_ctr (toglobal
			 (dest, &scan_result_buf), *((float *) &temp_scan_val[count]), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return val;
    }
}
double 
all_scan_dmult (double val)
{
  int dest;
  int i, count;
  int bit_vec = 1;
  double scan_val;
  long long *local_buf = (long long *) scan_buf;
  scan_sctr[phase] = scan_rctr[phase] = 0;
  phase = (phase + 1) % 2;
  *((double *) &temp_scan_val[0]) = scan_val = val;
  for (count = 0; bit_vec != PROCS; count++, bit_vec = bit_vec << 1)
    {
      dest = MYPROC & (~bit_vec);
      if (dest != MYPROC)
	break;
    }
  store_sync_ctr (count * sizeof
		  (double), &scan_rctr[phase], &scan_sctr[phase]);
  for (i = 1; i <= count; i++)
    {
      scan_val = MULT (scan_val, *((double *) &local_buf[i]));
      *((double *) &temp_scan_val[i]) = scan_val;
    }
  __d_store_ctr (toglobal (dest, &scan_buf[count + 1]), scan_val, &scan_rctr[phase], &scan_sctr[phase]);
  count--;
  bit_vec = bit_vec >> 1;
  if (MYPROC != 0)
    {
      double result_val;
      store_sync_ctr (sizeof (double), &scan_rctr[phase], &scan_sctr
		      [phase]);
      result_val = *((double *) &scan_result_buf);
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __d_store_ctr (toglobal (dest, &scan_result_buf), MULT (result_val, *((double *) &temp_scan_val[count])), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return MULT (result_val, val);
    }
  else
    {
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __d_store_ctr
	    (toglobal (dest, &scan_result_buf), *((double *) &temp_scan_val[count]), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return val;
    }
}
int 
all_scan_max (int val)
{
  int dest;
  int i, count;
  int bit_vec = 1;
  int scan_val;
  long long *local_buf = (long long *) scan_buf;
  scan_sctr[phase] = scan_rctr[phase] = 0;
  phase = (phase + 1) % 2;
  *((int *) &temp_scan_val[0]) = scan_val = val;
  for (count = 0; bit_vec != PROCS; count++, bit_vec = bit_vec << 1)
    {
      dest = MYPROC & (~bit_vec);
      if (dest != MYPROC)
	break;
    }
  store_sync_ctr (count * sizeof (int), &scan_rctr
		  [phase], &scan_sctr[phase]);
  for (i = 1; i <= count; i++)
    {
      scan_val = MAX (scan_val, *((int *) &local_buf[i]));
      *((int *) &temp_scan_val[i]) = scan_val;
    }
  __i_store_ctr (toglobal (dest, &scan_buf[count + 1]), scan_val, &scan_rctr[phase], &scan_sctr[phase]);
  count--;
  bit_vec = bit_vec >> 1;
  if (MYPROC != 0)
    {
      int result_val;
      store_sync_ctr (sizeof (int), &scan_rctr[phase], &scan_sctr[phase]);
      result_val = *((int
		      *) &scan_result_buf);
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store_ctr (toglobal (dest, &scan_result_buf), MAX (result_val, *((int *) &temp_scan_val[count])), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return MAX (result_val, val);
    }
  else
    {
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store_ctr (toglobal (dest, &scan_result_buf
			 ), *((int *) &temp_scan_val[count]), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return val;
    }
}
unsigned int 
all_scan_umax (unsigned int val)
{
  int dest;
  int i, count;
  int bit_vec = 1;
  unsigned int scan_val;
  long long *local_buf = (long long *) scan_buf;
  scan_sctr[phase] = scan_rctr[phase] = 0;
  phase = (phase + 1) % 2;
  *((unsigned int *) &temp_scan_val[0]) = scan_val = val;
  for (count = 0; bit_vec != PROCS; count++, bit_vec = bit_vec << 1)
    {
      dest = MYPROC & (~bit_vec);
      if (dest != MYPROC)
	break;
    }
  store_sync_ctr
    (count * sizeof (unsigned int), &scan_rctr[phase], &scan_sctr[phase]);
  for (i = 1; i <= count; i++)
    {
      scan_val = MAX (scan_val, *((unsigned int *) &local_buf[i]));
      *((unsigned int *) &temp_scan_val[i]) = scan_val;
    }
  __i_store_ctr (toglobal (dest, &scan_buf[count + 1]), scan_val, &scan_rctr[phase], &scan_sctr[phase]);
  count--;
  bit_vec = bit_vec >> 1;
  if (MYPROC != 0)
    {
      unsigned int result_val;
      store_sync_ctr (sizeof
		      (unsigned int), &scan_rctr[phase], &scan_sctr[phase]);
      result_val = *((unsigned int *) &scan_result_buf);
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store_ctr (toglobal (dest, &scan_result_buf), MAX (result_val, *((unsigned int *) &temp_scan_val[count])), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return MAX (result_val, val);
    }
  else
    {
      for (; count >= 0; count--, bit_vec
	   = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store_ctr (toglobal (dest, &scan_result_buf), *((unsigned int *) &temp_scan_val[count]), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return val;
    }
}
float 
all_scan_fmax (float val)
{
  int dest;
  int i, count;
  int bit_vec = 1;
  float scan_val;
  long long *local_buf = (long long *) scan_buf;
  scan_sctr[phase] = scan_rctr[phase] = 0;
  phase = (phase + 1) % 2;
  *((float *) &temp_scan_val[0]) = scan_val = val;
  for (count = 0; bit_vec != PROCS; count++, bit_vec = bit_vec << 1)
    {
      dest = MYPROC & (~bit_vec);
      if (dest != MYPROC)
	break;
    }
  store_sync_ctr (count * sizeof (float
		  ), &scan_rctr[phase], &scan_sctr[phase]);
  for (i = 1; i <= count; i++)
    {
      scan_val = MAX (scan_val, *((float *) &local_buf[i]));
      *((float *) &temp_scan_val[i]) = scan_val;
    }
  __f_store_ctr (toglobal (dest, &scan_buf[count + 1]), scan_val, &scan_rctr[phase], &scan_sctr[phase]);
  count--;
  bit_vec = bit_vec >> 1;
  if (MYPROC != 0)
    {
      float result_val;
      store_sync_ctr (sizeof (float), &scan_rctr[phase], &scan_sctr[phase]
      );
      result_val = *((float *) &scan_result_buf);
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __f_store_ctr (toglobal (dest, &scan_result_buf), MAX (result_val, *((float *) &temp_scan_val[count])), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return MAX (result_val, val);
    }
  else
    {
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __f_store_ctr (toglobal
			 (dest, &scan_result_buf), *((float *) &temp_scan_val[count]), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return val;
    }
}
double 
all_scan_dmax (double val)
{
  int dest;
  int i, count;
  int bit_vec = 1;
  double scan_val;
  long long *local_buf = (long long *) scan_buf;
  scan_sctr[phase] = scan_rctr[phase] = 0;
  phase = (phase + 1) % 2;
  *((double *) &temp_scan_val[0]) = scan_val = val;
  for (count = 0; bit_vec != PROCS; count++, bit_vec = bit_vec << 1)
    {
      dest = MYPROC & (~bit_vec);
      if (dest != MYPROC)
	break;
    }
  store_sync_ctr (count * sizeof
		  (double), &scan_rctr[phase], &scan_sctr[phase]);
  for (i = 1; i <= count; i++)
    {
      scan_val = MAX (scan_val, *((double *) &local_buf[i]));
      *((double *) &temp_scan_val[i]) = scan_val;
    }
  __d_store_ctr (toglobal (dest, &scan_buf[count + 1]), scan_val, &scan_rctr[phase], &scan_sctr[phase]);
  count--;
  bit_vec = bit_vec >> 1;
  if (MYPROC != 0)
    {
      double result_val;
      store_sync_ctr (sizeof (double), &scan_rctr[phase], &scan_sctr
		      [phase]);
      result_val = *((double *) &scan_result_buf);
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __d_store_ctr (toglobal (dest, &scan_result_buf), MAX (result_val, *((double *) &temp_scan_val[count])), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return MAX (result_val, val);
    }
  else
    {
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __d_store_ctr
	    (toglobal (dest, &scan_result_buf), *((double *) &temp_scan_val[count]), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return val;
    }
}
int 
all_scan_min (int val)
{
  int dest;
  int i, count;
  int bit_vec = 1;
  int scan_val;
  long long *local_buf = (long long *) scan_buf;
  scan_sctr[phase] = scan_rctr[phase] = 0;
  phase = (phase + 1) % 2;
  *((int *) &temp_scan_val[0]) = scan_val = val;
  for (count = 0; bit_vec != PROCS; count++, bit_vec = bit_vec << 1)
    {
      dest = MYPROC & (~bit_vec);
      if (dest != MYPROC)
	break;
    }
  store_sync_ctr (count * sizeof (int), &scan_rctr
		  [phase], &scan_sctr[phase]);
  for (i = 1; i <= count; i++)
    {
      scan_val = MIN (scan_val, *((int *) &local_buf[i]));
      *((int *) &temp_scan_val[i]) = scan_val;
    }
  __i_store_ctr (toglobal (dest, &scan_buf[count + 1]), scan_val, &scan_rctr[phase], &scan_sctr[phase]);
  count--;
  bit_vec = bit_vec >> 1;
  if (MYPROC != 0)
    {
      int result_val;
      store_sync_ctr (sizeof (int), &scan_rctr[phase], &scan_sctr[phase]);
      result_val = *((int
		      *) &scan_result_buf);
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store_ctr (toglobal (dest, &scan_result_buf), MIN (result_val, *((int *) &temp_scan_val[count])), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return MIN (result_val, val);
    }
  else
    {
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store_ctr (toglobal (dest, &scan_result_buf
			 ), *((int *) &temp_scan_val[count]), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return val;
    }
}
unsigned int 
all_scan_umin (unsigned int val)
{
  int dest;
  int i, count;
  int bit_vec = 1;
  unsigned int scan_val;
  long long *local_buf = (long long *) scan_buf;
  scan_sctr[phase] = scan_rctr[phase] = 0;
  phase = (phase + 1) % 2;
  *((unsigned int *) &temp_scan_val[0]) = scan_val = val;
  for (count = 0; bit_vec != PROCS; count++, bit_vec = bit_vec << 1)
    {
      dest = MYPROC & (~bit_vec);
      if (dest != MYPROC)
	break;
    }
  store_sync_ctr
    (count * sizeof (unsigned int), &scan_rctr[phase], &scan_sctr[phase]);
  for (i = 1; i <= count; i++)
    {
      scan_val = MIN (scan_val, *((unsigned int *) &local_buf[i]));
      *((unsigned int *) &temp_scan_val[i]) = scan_val;
    }
  __i_store_ctr (toglobal (dest, &scan_buf[count + 1]), scan_val, &scan_rctr[phase], &scan_sctr[phase]);
  count--;
  bit_vec = bit_vec >> 1;
  if (MYPROC != 0)
    {
      unsigned int result_val;
      store_sync_ctr (sizeof
		      (unsigned int), &scan_rctr[phase], &scan_sctr[phase]);
      result_val = *((unsigned int *) &scan_result_buf);
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store_ctr (toglobal (dest, &scan_result_buf), MIN (result_val, *((unsigned int *) &temp_scan_val[count])), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return MIN (result_val, val);
    }
  else
    {
      for (; count >= 0; count--, bit_vec
	   = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store_ctr (toglobal (dest, &scan_result_buf), *((unsigned int *) &temp_scan_val[count]), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return val;
    }
}
float 
all_scan_fmin (float val)
{
  int dest;
  int i, count;
  int bit_vec = 1;
  float scan_val;
  long long *local_buf = (long long *) scan_buf;
  scan_sctr[phase] = scan_rctr[phase] = 0;
  phase = (phase + 1) % 2;
  *((float *) &temp_scan_val[0]) = scan_val = val;
  for (count = 0; bit_vec != PROCS; count++, bit_vec = bit_vec << 1)
    {
      dest = MYPROC & (~bit_vec);
      if (dest != MYPROC)
	break;
    }
  store_sync_ctr (count * sizeof (float
		  ), &scan_rctr[phase], &scan_sctr[phase]);
  for (i = 1; i <= count; i++)
    {
      scan_val = MIN (scan_val, *((float *) &local_buf[i]));
      *((float *) &temp_scan_val[i]) = scan_val;
    }
  __f_store_ctr (toglobal (dest, &scan_buf[count + 1]), scan_val, &scan_rctr[phase], &scan_sctr[phase]);
  count--;
  bit_vec = bit_vec >> 1;
  if (MYPROC != 0)
    {
      float result_val;
      store_sync_ctr (sizeof (float), &scan_rctr[phase], &scan_sctr[phase]
      );
      result_val = *((float *) &scan_result_buf);
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __f_store_ctr (toglobal (dest, &scan_result_buf), MIN (result_val, *((float *) &temp_scan_val[count])), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return MIN (result_val, val);
    }
  else
    {
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __f_store_ctr (toglobal
			 (dest, &scan_result_buf), *((float *) &temp_scan_val[count]), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return val;
    }
}
double 
all_scan_dmin (double val)
{
  int dest;
  int i, count;
  int bit_vec = 1;
  double scan_val;
  long long *local_buf = (long long *) scan_buf;
  scan_sctr[phase] = scan_rctr[phase] = 0;
  phase = (phase + 1) % 2;
  *((double *) &temp_scan_val[0]) = scan_val = val;
  for (count = 0; bit_vec != PROCS; count++, bit_vec = bit_vec << 1)
    {
      dest = MYPROC & (~bit_vec);
      if (dest != MYPROC)
	break;
    }
  store_sync_ctr (count * sizeof
		  (double), &scan_rctr[phase], &scan_sctr[phase]);
  for (i = 1; i <= count; i++)
    {
      scan_val = MIN (scan_val, *((double *) &local_buf[i]));
      *((double *) &temp_scan_val[i]) = scan_val;
    }
  __d_store_ctr (toglobal (dest, &scan_buf[count + 1]), scan_val, &scan_rctr[phase], &scan_sctr[phase]);
  count--;
  bit_vec = bit_vec >> 1;
  if (MYPROC != 0)
    {
      double result_val;
      store_sync_ctr (sizeof (double), &scan_rctr[phase], &scan_sctr
		      [phase]);
      result_val = *((double *) &scan_result_buf);
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __d_store_ctr (toglobal (dest, &scan_result_buf), MIN (result_val, *((double *) &temp_scan_val[count])), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return MIN (result_val, val);
    }
  else
    {
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __d_store_ctr
	    (toglobal (dest, &scan_result_buf), *((double *) &temp_scan_val[count]), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return val;
    }
}
int 
all_scan_or (int val)
{
  int dest;
  int i, count;
  int bit_vec = 1;
  int scan_val;
  long long *local_buf = (long long *) scan_buf;
  scan_sctr[phase] = scan_rctr[phase] = 0;
  phase = (phase + 1) % 2;
  *((int *) &temp_scan_val[0]) = scan_val = val;
  for (count = 0; bit_vec != PROCS; count++, bit_vec = bit_vec << 1)
    {
      dest = MYPROC & (~bit_vec);
      if (dest != MYPROC)
	break;
    }
  store_sync_ctr (count * sizeof (int), &scan_rctr
		  [phase], &scan_sctr[phase]);
  for (i = 1; i <= count; i++)
    {
      scan_val = OR (scan_val, *((int *) &local_buf[i]));
      *((int *) &temp_scan_val[i]) = scan_val;
    }
  __i_store_ctr (toglobal (dest, &scan_buf[count + 1]), scan_val, &scan_rctr[phase], &scan_sctr[phase]);
  count--;
  bit_vec = bit_vec >> 1;
  if (MYPROC != 0)
    {
      int result_val;
      store_sync_ctr (sizeof (int), &scan_rctr[phase], &scan_sctr[phase]);
      result_val = *((int
		      *) &scan_result_buf);
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store_ctr (toglobal (dest, &scan_result_buf), OR (result_val, *((int *) &temp_scan_val[count])), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return OR (result_val, val);
    }
  else
    {
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store_ctr (toglobal (dest, &scan_result_buf)
			 ,*((int *) &temp_scan_val[count]), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return val;
    }
}
unsigned int 
all_scan_uor (unsigned int val)
{
  int dest;
  int i, count;
  int bit_vec = 1;
  unsigned int scan_val;
  long long *local_buf = (long long *) scan_buf;
  scan_sctr[phase] = scan_rctr[phase] = 0;
  phase = (phase + 1) % 2;
  *((unsigned int *) &temp_scan_val[0]) = scan_val = val;
  for (count = 0; bit_vec != PROCS; count++, bit_vec = bit_vec << 1)
    {
      dest = MYPROC & (~bit_vec);
      if (dest != MYPROC)
	break;
    }
  store_sync_ctr
    (count * sizeof (unsigned int), &scan_rctr[phase], &scan_sctr[phase]);
  for (i = 1; i <= count; i++)
    {
      scan_val = OR (scan_val, *((unsigned int *) &local_buf[i]));
      *((unsigned int *) &temp_scan_val[i]) = scan_val;
    }
  __i_store_ctr (toglobal (dest, &scan_buf[count + 1]), scan_val, &scan_rctr[phase], &scan_sctr[phase]);
  count--;
  bit_vec = bit_vec >> 1;
  if (MYPROC != 0)
    {
      unsigned int result_val;
      store_sync_ctr (sizeof
		      (unsigned int), &scan_rctr[phase], &scan_sctr[phase]);
      result_val = *((unsigned int *) &scan_result_buf);
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store_ctr (toglobal (dest, &scan_result_buf), OR (result_val, *((unsigned int *) &temp_scan_val[count])), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return OR (result_val, val);
    }
  else
    {
      for (; count >= 0; count--, bit_vec
	   = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store_ctr (toglobal (dest, &scan_result_buf), *((unsigned int *) &temp_scan_val[count]), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return val;
    }
}
int 
all_scan_xor (int val)
{
  int dest;
  int i, count;
  int bit_vec = 1;
  int scan_val;
  long long *local_buf = (long long *) scan_buf;
  scan_sctr[phase] = scan_rctr[phase] = 0;
  phase = (phase + 1) % 2;
  *((int *) &temp_scan_val[0]) = scan_val = val;
  for (count = 0; bit_vec != PROCS; count++, bit_vec = bit_vec << 1)
    {
      dest = MYPROC & (~bit_vec);
      if (dest != MYPROC)
	break;
    }
  store_sync_ctr (count * sizeof (int), &scan_rctr
		  [phase], &scan_sctr[phase]);
  for (i = 1; i <= count; i++)
    {
      scan_val = XOR (scan_val, *((int *) &local_buf[i]));
      *((int *) &temp_scan_val[i]) = scan_val;
    }
  __i_store_ctr (toglobal (dest, &scan_buf[count + 1]), scan_val, &scan_rctr[phase], &scan_sctr[phase]);
  count--;
  bit_vec = bit_vec >> 1;
  if (MYPROC != 0)
    {
      int result_val;
      store_sync_ctr (sizeof (int), &scan_rctr[phase], &scan_sctr[phase]);
      result_val = *((int
		      *) &scan_result_buf);
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store_ctr (toglobal (dest, &scan_result_buf), XOR (result_val, *((int *) &temp_scan_val[count])), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return XOR (result_val, val);
    }
  else
    {
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store_ctr (toglobal (dest, &scan_result_buf
			 ), *((int *) &temp_scan_val[count]), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return val;
    }
}
unsigned int 
all_scan_uxor (unsigned int val)
{
  int dest;
  int i, count;
  int bit_vec = 1;
  unsigned int scan_val;
  long long *local_buf = (long long *) scan_buf;
  scan_sctr[phase] = scan_rctr[phase] = 0;
  phase = (phase + 1) % 2;
  *((unsigned int *) &temp_scan_val[0]) = scan_val = val;
  for (count = 0; bit_vec != PROCS; count++, bit_vec = bit_vec << 1)
    {
      dest = MYPROC & (~bit_vec);
      if (dest != MYPROC)
	break;
    }
  store_sync_ctr
    (count * sizeof (unsigned int), &scan_rctr[phase], &scan_sctr[phase]);
  for (i = 1; i <= count; i++)
    {
      scan_val = XOR (scan_val, *((unsigned int *) &local_buf[i]));
      *((unsigned int *) &temp_scan_val[i]) = scan_val;
    }
  __i_store_ctr (toglobal (dest, &scan_buf[count + 1]), scan_val, &scan_rctr[phase], &scan_sctr[phase]);
  count--;
  bit_vec = bit_vec >> 1;
  if (MYPROC != 0)
    {
      unsigned int result_val;
      store_sync_ctr (sizeof
		      (unsigned int), &scan_rctr[phase], &scan_sctr[phase]);
      result_val = *((unsigned int *) &scan_result_buf);
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store_ctr (toglobal (dest, &scan_result_buf), XOR (result_val, *((unsigned int *) &temp_scan_val[count])), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return XOR (result_val, val);
    }
  else
    {
      for (; count >= 0; count--, bit_vec
	   = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store_ctr (toglobal (dest, &scan_result_buf), *((unsigned int *) &temp_scan_val[count]), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return val;
    }
}
int 
all_scan_and (int val)
{
  int dest;
  int i, count;
  int bit_vec = 1;
  int scan_val;
  long long *local_buf = (long long *) scan_buf;
  scan_sctr[phase] = scan_rctr[phase] = 0;
  phase = (phase + 1) % 2;
  *((int *) &temp_scan_val[0]) = scan_val = val;
  for (count = 0; bit_vec != PROCS; count++, bit_vec = bit_vec << 1)
    {
      dest = MYPROC & (~bit_vec);
      if (dest != MYPROC)
	break;
    }
  store_sync_ctr (count * sizeof (int), &scan_rctr
		  [phase], &scan_sctr[phase]);
  for (i = 1; i <= count; i++)
    {
      scan_val = AND (scan_val, *((int *) &local_buf[i]));
      *((int *) &temp_scan_val[i]) = scan_val;
    }
  __i_store_ctr (toglobal (dest, &scan_buf[count + 1]), scan_val, &scan_rctr[phase], &scan_sctr[phase]);
  count--;
  bit_vec = bit_vec >> 1;
  if (MYPROC != 0)
    {
      int result_val;
      store_sync_ctr (sizeof (int), &scan_rctr[phase], &scan_sctr[phase]);
      result_val = *((int
		      *) &scan_result_buf);
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store_ctr (toglobal (dest, &scan_result_buf), AND (result_val, *((int *) &temp_scan_val[count])), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return AND (result_val, val);
    }
  else
    {
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store_ctr (toglobal (dest, &scan_result_buf
			 ), *((int *) &temp_scan_val[count]), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return val;
    }
}
unsigned int 
all_scan_uand (unsigned int val)
{
  int dest;
  int i, count;
  int bit_vec = 1;
  unsigned int scan_val;
  long long *local_buf = (long long *) scan_buf;
  scan_sctr[phase] = scan_rctr[phase] = 0;
  phase = (phase + 1) % 2;
  *((unsigned int *) &temp_scan_val[0]) = scan_val = val;
  for (count = 0; bit_vec != PROCS; count++, bit_vec = bit_vec << 1)
    {
      dest = MYPROC & (~bit_vec);
      if (dest != MYPROC)
	break;
    }
  store_sync_ctr
    (count * sizeof (unsigned int), &scan_rctr[phase], &scan_sctr[phase]);
  for (i = 1; i <= count; i++)
    {
      scan_val = AND (scan_val, *((unsigned int *) &local_buf[i]));
      *((unsigned int *) &temp_scan_val[i]) = scan_val;
    }
  __i_store_ctr (toglobal (dest, &scan_buf[count + 1]), scan_val, &scan_rctr[phase], &scan_sctr[phase]);
  count--;
  bit_vec = bit_vec >> 1;
  if (MYPROC != 0)
    {
      unsigned int result_val;
      store_sync_ctr (sizeof
		      (unsigned int), &scan_rctr[phase], &scan_sctr[phase]);
      result_val = *((unsigned int *) &scan_result_buf);
      for (; count >= 0; count--, bit_vec = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store_ctr (toglobal (dest, &scan_result_buf), AND (result_val, *((unsigned int *) &temp_scan_val[count])), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return AND (result_val, val);
    }
  else
    {
      for (; count >= 0; count--, bit_vec
	   = bit_vec >> 1)
	{
	  dest = MYPROC | bit_vec;
	  __i_store_ctr (toglobal (dest, &scan_result_buf), *((unsigned int *) &temp_scan_val[count]), &scan_rctr[phase], &scan_sctr[phase]);
	}
      return val;
    }
}
void 
all_bcast_buffer (___V_SPTR val, int size)
{
  int left = 2 * MYPROC + 1;
  int right = 2 * MYPROC + 2;
  void *local = tolocal (val);
  ___V_GPTR leftBuf = toglobal (left, local);
  ___V_GPTR rightBuf = toglobal (right, local);
  reset_store_sync ();
  barrier ();
  if (MYPROC != 0)
    {
      store_sync (size);
    }
  if (left < PROCS)
    bulk_store (leftBuf, local, size);
  if (right < PROCS)
    bulk_store (rightBuf, local, size);
  barrier ();
  return;
}

void *
all_bcast_p (void *p)
{
  return (void *) all_bcast_i ((int) p);
}
void
init_bcast ()
{
  int left = 2 * MYPROC + 1;
  int right = 2 * MYPROC + 2;
  bcast_rctr = bcast_sctr = 0;
  if (left < PROCS)
    {
      bcast_rctr += sizeof (int);
      bcast_sctr += sizeof (int);
    }
  if (right < PROCS)
    {
      bcast_rctr += sizeof (int);
      bcast_sctr += sizeof (int);
    }
}
void
init_reduce ()
{
  if (MYPROC)
    red_rctr = red_sctr = sizeof (int);
  else
    red_rctr = red_sctr = 0;
}
void
init_scan ()
{
  scan_rctr[0] = scan_rctr[1] = scan_sctr[0] = scan_sctr[1] = 0;
}
int
com_init ()
{
  init_bcast ();
  init_reduce ();
  init_scan ();
  barrier ();
}
