/****************************************************************************/
/*                                                                          */
/* Copyright    **    **  ********  ********  **    **   ******     Limited */
/* Copyright    ***  ***  ********  ********  **   **   ********    Limited */
/* Copyright    ********  **           **     **  **    **    **    Limited */
/* Copyright    ** ** **  ******       **     ****      **    **    Limited */
/* Copyright    **    **  **           **     ** **     **    **    Limited */
/* Copyright    **    **  **           **     **  **    **    **    Limited */
/* Copyright    **    **  ********  ********  **   **   ********    Limited */
/* Copyright    **    **  ********  ********  **    **   ******     Limited */
/*                                                                          */
/*                      **       ****     ******    ******                  */
/*                     ***      ******   ********  ********                 */
/*                    ****     **    **  **    **  **   ***                 */
/*                      **      *******   *******      ***                  */
/*                      **           **        **    ***                    */
/*                      **           **        **   ***                     */
/*                      **     ********  ********  ********                 */
/*                      **      ******    ******   ********                 */
/*                                                                          */
/* Author:  Eric Barton - September 1992                                    */
/****************************************************************************/

/*
 * Copyright (c) 1992 by Meiko Ltd.
 */

#pragma ident "$Id: group.c,v 1.31 1994/08/30 15:16:40 jim Exp $"
/*             $Source: /home/cvs/master/cs2parts/ew/group.c,v $ */

/* 
 * Functions modified to call am_poll() when waiting on an elan event
 */

#define EW_SPINUSECS 16

#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <sys/types.h>
#include <elan/elan.h>

#include <ew/group.h>
#include "group_sys.h"

#include <elan/event.h>
#include <ew/ew.h>

#define DBG(x)


void	 gam_waitevent (void *c, ELAN_EVENT *e, void *t)
{
	while (!EVENT_READY(e))
	 	am_poll();
	ELAN_CLEAREVENT(e);

}
/*
 * This gsync is the fastest IF all the group members arrive nearly simultaneously.
 * In particular if the time between the root arriving and the others is small.
 * Otherwise because the polling backs off, there can be large delays in exiting
 * from the gsync.
 * 
 * *** CHANGE ME ***
 * The best solution would be to back off to a tree gsync when the backoff period
 * has gone above twice the time for a ew_gsync. 
 * We can't amke this change until the ew version numbe changes, though. Therefore
 * for now we use a faster polling rate. (see ew_t_gsync).
 */
void            gam_fgsync (EW_GROUP *ug)	/* network polling gsync */
{						/* (broadcast MUST be possible) */
   GROUP	  *g = (GROUP *)ug;
   register int    seq;
   register int	   usec = 1;
   

   if (g->g_waitType != ELAN_POLL_EVENT)	/* Mustn't poll forever */
   {
      gam_fbcast (&g->g_public, 0, NULL, 0, 1);	/* NULL broadcast is equivalent */
      return;
   }
   
   if (g->g_public.g_bcastVp == ELAN_INVALID_PROCESS) /* can't reduce in network */
   {
      gam_bcast (&g->g_public, 0, NULL, 0, 1);	/* NULL broadcast is equivalent */
      return;
   }

   seq = g->g_bcastSeq;				/* grab current seq # */
   g->g_fgsyncReadySeq = seq;			/* invalidate completion flags */
#if 0
   /* No longer used ! */
   g->g_fgsyncDoneSeq  = seq;
#endif
   g->g_bcastSeq       = ++seq;			/* get next seq # and flag ready */

   if (g->g_public.g_self == 0)			/* I'm the root process */
   {
      elan_setevent (ew_ctx, &g->g_t_gsyncGo);	/* => get thread to ? and ! */
   }


   while (g->g_fgsyncReadySeq != seq)	/* poll for completion */
   {					/* (got bcast seq)     */
       /* 
	* This is busy anyway, so it may as well spin on the real test.
	* The variable will be in the cache anyway, so it shouldn't hurt.
	* (But maybe not on the 403 ... )
	*/
       am_poll();
       elan_usecspin (usec);
       if ((usec <<= 1) > EW_SPINUSECS)
	   usec = EW_SPINUSECS;
   }
}

void            gam_gsync (EW_GROUP *g)		/* tree reducing gsync */
{
   gam_bcast (g, 0, NULL, 0, 1);			/* NULL broadcast is equivalent */
}

void            gam_fbcast (EW_GROUP *ug,	/* group */
			   int	     src,	/* group member sourcing data */
			   caddr_t   data,	/* source/dest of data */
			   int       size,	/* # bytes in data */
			   int       global)	/* data address same everywhere */
{
   GROUP	  *g = (GROUP *)ug;		/* network polling broadcast */
   int             nob;
   caddr_t         buf;
   int             seq;
   ELAN_EVENT     *bcastPeerDone;


   if (g->g_public.g_bcastVp == ELAN_INVALID_PROCESS) /* can't reduce in network */
   {
      gam_bcast (&g->g_public, src, data, size, global);	/* do it on a tree */
      return;
   }

   do
   {
      if (global)
      {
	 buf = data;
	 nob = size;
      }
      else
      {
	 buf = g->g_up->gp_buffer;

	 if ((nob = size) > g->g_bufSize)
	    nob = g->g_bufSize;
      }

      seq = g->g_bcastSeq + 1;			/* get next bcast seq # */
      bcastPeerDone = &g->g_bcastPeerDone[seq & 1];
      ELAN_CLEAREVENT (bcastPeerDone);		/* clear bcast done event */
      g->g_bcastSeq = seq;			/* THEN flag ready */

      if (g->g_public.g_self == src)		/* I'm the source */
      {
	 g->g_bcastDma.dma_size      = nob;
	 g->g_bcastDma.dma_source    = data;
	 g->g_bcastDma.dma_dest      = buf;
	 g->g_bcastDma.dma_destEvent = bcastPeerDone;

	 elan_setevent  (ew_ctx, &g->g_t_bcastGo);
	 gam_waitevent (ew_ctx, &g->g_bcastDone, g->g_waitType);
      }
      else
      {
	 gam_waitevent (ew_ctx, bcastPeerDone, g->g_waitType);

	 if (!global)
	    ew_bcopy (buf, data, nob);
      }
      data += nob;
      am_poll();
   } while ((size -= nob) > 0);
}

void            gam_bcast (EW_GROUP *ug,		/* group */
			  int 	    src,	/* group member sourcing data */
			  caddr_t   data,	/* source/dest of data */
			  int       size,	/* # bytes in data */
			  int       global)	/* data address same everywhere */
{
   GROUP	  *g = (GROUP *)ug;
   int             nob;
   caddr_t         sbuf;
   caddr_t         dbuf;
   u_int           srcProc;
   u_int	   seq;
   ELAN_EVENT     *bcastPeerDone;
   GPEER       *d;

   DBG (ew_dbg ("ew_bcast (%08x)\n", g));

   if (g->g_public.g_bcastVp != ELAN_INVALID_PROCESS)
   {
      /* Combine ready on tree.
       * root tells src everyone's ready.
       * src does broadcast */

      if (g->g_public.g_self == 0 && src != 0)	/* root finds src process # unless herself */
	 srcProc = ew_groupMember (ug, src);

      do
      {
	 if (global)
	 {
	    dbuf = data;
	    nob = size;
	 }
	 else
	 {
	    dbuf = g->g_up->gp_buffer;
	    if ((nob = size) > g->g_bufSize)
	       nob = g->g_bufSize;
	 }

	 seq = ++(g->g_bcastSeq);		/* get next bcast seq # */
	 bcastPeerDone = &g->g_bcastPeerDone[seq & 1];
	 ELAN_CLEAREVENT (bcastPeerDone);	/* clear bcast done event BEFORE parent ! ready */

	 g->g_up->gp_dma.dma_size = 0;		/* init prop ready? up tree */

	 if (g->g_public.g_self == 0 && src != 0) /* root init src ! ready */
	    g->g_up->gp_dma.dma_destProc = srcProc;

	 if (g->g_public.g_self == src)		/* broadcast source: init everyone ! data */
	 {
	    g->g_bcastDma.dma_size = nob;
	    g->g_bcastDma.dma_source = data;
	    g->g_bcastDma.dma_dest = dbuf;
	    g->g_bcastDma.dma_destEvent = bcastPeerDone;
	 }

	 DBG (ew_dbg ("ew_bcast (): initialised dmas\n"));

	 for (d = g->g_down;
	      d < g->g_endDown;
	      d = (GPEER *) ((caddr_t) d + g->g_peerSize))	/* branches ? ready */
	 {

	    gam_waitevent (ew_ctx, &d->gp_peerDone, g->g_waitType);
	 }

	 if (g->g_public.g_self != 0 ||		/* not root: parent ! ready */
	     src != 0)				/* root: src ! ready (and she's not me) */
	    elan_dma (ew_ctx, &g->g_up->gp_dma);

	 if (g->g_public.g_self == src)
	 {
	    if (g->g_public.g_self != 0)	/* wait for root to send ready */
	       gam_waitevent (ew_ctx, &g->g_up->gp_peerDone, g->g_waitType);
	    
	    elan_dma (ew_ctx, &g->g_bcastDma);
	    gam_waitevent (ew_ctx, &g->g_bcastDone, g->g_waitType);
	 }
	 else
	 {
	    gam_waitevent (ew_ctx, bcastPeerDone, g->g_waitType);

	    if (dbuf != data)
	       ew_bcopy (dbuf, data, nob);
	 }

	 data += nob;
         am_poll();
      } while ((size -= nob) > 0);
   }
   else
   {
      /* children ? ready 
       * parent ! ready
       * parent ? ready
       * children ! ready */

      do
      {
	 if ((nob = size) > g->g_bufSize)
	    nob = g->g_bufSize;

	 dbuf = global ? data : g->g_up->gp_buffer;
	 sbuf = g->g_public.g_self == src ? data : NULL;
	 
	 for (d = g->g_down;
	      d < g->g_endDown;
	      d = (GPEER *) ((caddr_t) d + g->g_peerSize))	/* branches ? ready */
	 {
	    gam_waitevent (ew_ctx, &d->gp_peerDone, g->g_waitType);

	    if (d->gp_proc <= src && src < d->gp_endProc)	/* this branch is source */
	       sbuf = d->gp_buffer;
	 }
	 
	 if (g->g_public.g_self != 0)
	 {
	    DBG (ew_dbg ("ew_bcast (): parent ! ready %08x (%08x)\n", g->g_up->gp_dma.dma_destEvent, sbuf));

	    g->g_up->gp_dma.dma_size = sbuf ? nob : 0;
	    g->g_up->gp_dma.dma_source = sbuf;

	    elan_dma (ew_ctx, &g->g_up->gp_dma);/* parent ! data/ready */
	 }

	 if (!sbuf)				/* getting data from parent */
	    sbuf = dbuf;
	 else if (sbuf != data)			/* getting data from child */
	 {
	    ew_bcopy (sbuf, data, nob);		/* must copy BEFORE child ! ready */
	    sbuf = data;
	 }

	 if (g->g_public.g_self != 0)
	 {
	    DBG (ew_dbg ("ew_bcast (): parent ? ready %08x\n", &g->g_up->gp_peerDone));

	    gam_waitevent (ew_ctx, &g->g_up->gp_peerDone, g->g_waitType);	/* parent ? ready/data */
	 }

	 for (d = g->g_down;
	      d < g->g_endDown;
	      d = (GPEER *) ((caddr_t) d + g->g_peerSize))	/* branches ! ready/data */
	 {

	    if (d->gp_proc <= src && src < d->gp_endProc)
	       d->gp_dma.dma_size = 0;		/* she has it already */
	    else
	    {
	       d->gp_dma.dma_size = nob;
	       d->gp_dma.dma_source = sbuf;
	       d->gp_dma.dma_dest = dbuf;
	    }

	    if ((d->gp_dma.dma_type  & 0xffff) == 0)
	       ew_exception (EW_EINTERNAL, "DMA has 0 type");
	    
	    elan_dma (ew_ctx, &d->gp_dma);
	 }

	 if (sbuf != data)			/* not got data in local buf */
	    ew_bcopy (sbuf, data, nob);		/* put it there (must have come from parent) */

	 while ((d = (GPEER *) ((caddr_t) d - g->g_peerSize)) >= g->g_down)	/* wait for completion */
	 {

	    gam_waitevent (ew_ctx, &d->gp_done, g->g_waitType);	/* in reverse order */
	 }

	 data += nob;
         am_poll();
      } while ((size -= nob) > 0);
   }
}

void            gam_reduce (EW_GROUP *ug,	/* group */
			   void    (*fn) (),	/* reduction function */
			   caddr_t   data,	/* data to be reduced */
			   int	     elsize,	/* bytes in 1 element of data */
			   int	     nelem,	/* # elements */
			   int       global)	/* data address same everywhere */
{
   GROUP	  *g = (GROUP *)ug;
   int		   size;
   int		   chunk;
   int             nob;
   u_int	   seq;
   caddr_t         accumulator;
   ELAN_EVENT     *bcastPeerDone;
   GPEER          *d;

   /* Combine ready on process tree. 
    * If broadcast is possible, root process signals ready to the rest with
    * a broadcast of the sequence # in everyone's parent GBRANCH.
    * Otherwise, ready is propagated back down the tree. */

   DBG (ew_dbg ("ew_reduce (%08x)\n", g));

   if ((size = nelem * elsize) > g->g_bufSize && /* > 1 packet */
       (nelem = g->g_bufSize / elsize) == 0)	/* 1 element won't fit */
      ew_exception (EW_ERANGE, "ew_reduce() elsize %d pktsize %d",elsize, g->g_bufSize);
      
   do
   {
      if ((nob = nelem * elsize) > size)	/* next chunk more than what's left */
      {
	 nob = size;				/* reduce chunk */
	 nelem = size / elsize;			/* reset # elements */
      }
      
      if (global)
	 accumulator = data;
      else
      {
	 accumulator = g->g_up->gp_buffer;
	 ew_bcopy (data, accumulator, nob);
      }

      /* init DMAs NOW so all do in parallel) 
       * Also note: setups could also be done in parallel between !parent
       * and ?parent, but this puts root's setups on the critical path.
       * Setups here put the leaves on the critical path (they initiate),
       * but they have less setup anyway (no children). */

      if (g->g_public.g_self != 0)		/* init parent ! partial */
      {
	 g->g_up->gp_dma.dma_size = nob;
	 g->g_up->gp_dma.dma_source = accumulator;
      }

      if (g->g_public.g_bcastVp == ELAN_INVALID_PROCESS) /* init children ! result */
	 for (d = g->g_down;
	      d < g->g_endDown;
	      d = (GPEER *) ((caddr_t) d + g->g_peerSize))
	 {
	    d->gp_dma.dma_size = nob;
	    d->gp_dma.dma_source = accumulator;
	    d->gp_dma.dma_dest = accumulator;
	 }
      else
      {
	 seq = ++g->g_bcastSeq;
	 bcastPeerDone = &g->g_bcastPeerDone[seq & 1]; /* new bcast event */
	 ELAN_CLEAREVENT (bcastPeerDone);	/* clear BEFORE parent ! ready */

	 if (g->g_public.g_self == 0)		/* init all ! result */
	 {
	    g->g_bcastDma.dma_size = nob;
	    g->g_bcastDma.dma_source = accumulator;
	    g->g_bcastDma.dma_dest = accumulator;
	    g->g_bcastDma.dma_destEvent = bcastPeerDone;
	 }
      }

      DBG (ew_dbg ("ew_reduce (): initialised dmas\n"));

      for (d = g->g_down;
	   d < g->g_endDown;
	   d = (GPEER *) ((caddr_t) d + g->g_peerSize))
      {
	 DBG (ew_dbg ("ew_reduce (): branch %d ? ready %08x\n",
		      ((caddr_t) d - (caddr_t) g->g_down) / g->g_peerSize, &d->gp_peerDone));

	 gam_waitevent (ew_ctx, &d->gp_peerDone, g->g_waitType);	/* subtree ? partial */

	 if (fn)
	    fn (accumulator, d->gp_buffer, &nelem); /* REDUCE !!! */
      }

      if (g->g_public.g_self == 0)		/* root */
      {
	 if (g->g_public.g_bcastVp == ELAN_INVALID_PROCESS) /* children ! result */
	 {
	    for (d = g->g_down;
		 d < g->g_endDown;
		 d = (GPEER *) ((caddr_t) d + g->g_peerSize))
	    {
	       DBG (ew_dbg ("ew_reduce (): branch %d ! ready %08x wait %08x (%08x)\n",
			    ((caddr_t) d - (caddr_t) g->g_down) / g->g_peerSize,
			    d->gp_dma.dma_destEvent, d->gp_dma.dma_sourceEvent, &d->gp_done));

	       elan_dma (ew_ctx, &d->gp_dma);
	    }

	    if (!global)			/* copy while waiting for comms  */
	       ew_bcopy (accumulator, data, nob);

	    /* Now wait for completion in reverse order to minimise chance of */
	    /* descheduling on waitevent */
	    while ((d = (GPEER *) ((caddr_t) d - g->g_peerSize)) >= g->g_down)
	       gam_waitevent (ew_ctx, &d->gp_done, g->g_waitType);
	 }
	 else					/* everyone ! result */
	 {
	    DBG (ew_dbg ("ew_reduce (): EVERYONE ! ready %08x wait %08x (%08x)\n",
		    g->g_bcastDma.dma_destEvent, g->g_bcastDma.dma_sourceEvent, &g->g_bcastDone));

	    elan_dma (ew_ctx, &g->g_bcastDma);

	    if (!global)			/* copy while waiting for comms */
	       ew_bcopy (accumulator, data, nob);

	    gam_waitevent (ew_ctx, &g->g_bcastDone, g->g_waitType);
	 }
      }
      else
      {
	 DBG (ew_dbg ("ew_reduce (): parent ! ready %08x\n", g->g_up->gp_dma.dma_destEvent));

	 elan_dma (ew_ctx, &g->g_up->gp_dma);	/* parent ! partial */

	 if (g->g_public.g_bcastVp == ELAN_INVALID_PROCESS)
	 {
	    DBG (ew_dbg ("ew_reduce (): parent ? ready %08x\n", &g->g_up->gp_peerDone));

	    gam_waitevent (ew_ctx, &g->g_up->gp_peerDone, g->g_waitType);	/* parent ? result */
	    
	    for (d = g->g_down;
		 d < g->g_endDown;
		 d = (GPEER *) ((caddr_t) d + g->g_peerSize))
	    {
	       DBG (ew_dbg ("ew_reduce (): branch %d ! ready %08x wait %08x (%08x)\n",
			    ((caddr_t) d - (caddr_t) g->g_down) / g->g_peerSize,
			    d->gp_dma.dma_destEvent, d->gp_dma.dma_sourceEvent, &d->gp_done));

	       elan_dma (ew_ctx, &d->gp_dma);	/* subtree ! result */
	    }

	    if (!global)			/* copy while waiting for comms */
	       ew_bcopy (accumulator, data, nob);

	    /* Now wait for completion in reverse order to minimise chance of */
	    /* descheduling on waitevent */
	    while ((d = (GPEER *) ((caddr_t) d - g->g_peerSize)) >= g->g_down)
	       gam_waitevent (ew_ctx, &d->gp_done, g->g_waitType);
	 }
	 else
	 {
	    DBG (ew_dbg ("ew_reduce (): bcast ? ready %08x\n", bcastPeerDone));

	    gam_waitevent (ew_ctx, bcastPeerDone, g->g_waitType);
	    
	    if (!global)
	       ew_bcopy (accumulator, data, nob);
	 }
      }

      data += nob;
   } while ((size -= nob) > 0);
}

