/*                                                                      tab:4
 *
 * string.h - Global-string library. Public definitions.
 *
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author:        Andrea Dusseau
 * Version        5
 * Creation Date: Thu Oct  1 14:45:08 PDT 1992
 * Filename:      gstr.h
 * History:
 *	ssl	5	Tue Nov 16 12:05:45 1993
 *		Replaced global tags with macros.
 *	LTL	4	Mon Aug 16 15:58:36 1993
 *		Changed the way extern and inline are handled for gcc-2.4.5.
 *      This is for version 112 of libsplitc.
 *	SL	3	Thu Feb 11 13:21:27 1993
 *		Added include for split-c.h
 * 	TvE	2	Mon Jan  4 17:28:05 PST 1993
 *		Renamed to string.h.
 */

#ifndef __STRING_H__
#define __STRING_H__

#include <split-c/split-c.h>

/*
 * STRING_INTERNAL_DEFINE is defined when compiling the actual code for the
 * library so that a normal version for all the inlined functions is produced.
 * It is off when any user code is being compiled.
 */
#undef EXTERN
#ifdef STRING_INTERNAL_DEFINE
#      define EXTERN 
#      define GSTR_INLINE(code) {code}
#else
#      define EXTERN extern
#      define GSTR_INLINE(code) ;
#endif

typedef ___C_GPTR g_str;

#define g_index(s, c) g_strchr(s, c)
#define g_rindex(s, c) g_strrchr(s, c)

EXTERN inline void  g_strinit();
EXTERN inline char *gstrn_get_ctr(char *dest, g_str src, int maxlen, Counter *flag);
EXTERN inline char *gstrn_read(char *dest, g_str src, int maxlen);
EXTERN inline char *gstrn_get(char *dest, g_str src, int maxlen);
EXTERN inline g_str gstrn_put_ctr(g_str dest, char *src, int maxlen, Counter *flag);
EXTERN inline g_str gstrn_put(g_str dest, char *src, int maxlen);
EXTERN inline g_str gstrn_write(g_str dest, char *src, int maxlen);
EXTERN inline g_str g_strncpy(g_str dest, g_str src, int maxlen);
EXTERN inline char *gstr_get_ctr(char *dest, g_str src, Counter *flag);
EXTERN inline char *gstr_read(char *dest, g_str src);
EXTERN inline char *gstr_get(char *dest, g_str src);
EXTERN inline g_str gstr_put_ctr(g_str dest, char *src, Counter *flag);
EXTERN inline g_str gstr_write(g_str dest, char *src);
EXTERN inline g_str gstr_put(g_str dest, char *src);
EXTERN inline g_str g_strcpy(g_str dest, g_str src);
EXTERN inline char *gstrcat_get_ctr(char *dst, g_str src, Counter *flag);
EXTERN inline char *gstrcat_read(char *dst, g_str src);
EXTERN inline char *gstrcat_get(char *dst, g_str src);
EXTERN inline g_str gstrcat_put_ctr(g_str dst, char *src, Counter *flag);
EXTERN inline g_str gstrcat_write(g_str dst, char *src);
EXTERN inline g_str gstrcat_put(g_str dst, char *src);
EXTERN inline g_str g_strcat(g_str dst, g_str src);
EXTERN inline char *gstrncat_get_ctr(char *dst, g_str src, int maxlen, Counter *flag);
EXTERN inline char *gstrncat_read(char *dst, g_str src, int maxlen);
EXTERN inline char *gstrncat_get(char *dst, g_str src, int maxlen);
EXTERN inline g_str gstrncat_put_ctr(g_str dst, char *src, int maxlen, Counter *flag);
EXTERN inline g_str gstrncat_write(g_str dst, char *src, int maxlen);
EXTERN inline g_str gstrncat_put(g_str dst, char *src, int maxlen);
EXTERN inline g_str g_strncat(g_str dst, g_str src, int maxlen);
EXTERN inline int   g_strlen(g_str s);
EXTERN inline int   g_strncmp(g_str gs1, g_str gs2, int maxlen);
EXTERN inline int   g_strcmp(g_str gs1, g_str gs2);
EXTERN inline char *g_strdup(g_str gs);
EXTERN inline int   g_strcasecmp(g_str gs1, g_str gs2);
EXTERN inline g_str g_strchr(g_str s, int c);
EXTERN inline g_str g_strrchr(g_str s, int c);
EXTERN inline g_str g_strpbrk(g_str gs1, g_str gs2);
EXTERN inline int   g_strspn(g_str gs1, g_str gs2);
EXTERN inline int   g_strcspn(g_str gs1, g_str gs2);
EXTERN inline g_str g_strstr(g_str gs1, g_str gs2);

#include <split-c/string_inline.h>

#endif






