/*									tab:4
 *
 * string_inline.h - Global-string library.  Inline functions.
 *
 * "Copyright (c) 1993 and The Regents of the University 
 * of California.  All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 			Andrea C. Dusseau
 * Version:			17
 * Creation Date:	Thu Oct  1 14:45:08 PDT 1992
 * Filename:		string_inline.h
 * History:
 *	ssl	15	Sun Nov 21 21:18:22 1993
 *		Replaced global pointer arithmetic with macro calls.
 *      Allows compilation by native compiler.
 *	ssl	14	Thu Nov 18 01:04:00 1993
 *		Modified to allow compilation with native compiler.
 *	ssl	13	Tue Nov 16 12:07:49 1993
 *		Replaced global tags with macros.
 *	ssl	12	Thu Oct 28 22:01:35 1993
 *		Include am.h instead of cm5.h
 *	ssl	11	Fri Oct 22 13:17:22 1993
 *		Moved _version_string_ to string.sc
 *	LTL	10	Mon Aug 16 16:01:01 1993
 *		Fixed it up for version 112 to support CMMD
 *	ACD	8	Wed Jun 16 10:58:51 1993
 *		Rewrite so that use bulk_get_ctr and bulk_put_ctr instead of
 *              specialized CMAM handlers in most cases.  Removed gst_init()
 *              requirement.
 *	ACD	5	Wed Jan 27 13:40:43 1993
 *		Same problem as previous one.  Now have </usr/include/string.h>
 *              hard-coded so that we can have same version in our libraries
 *              as the released version.
 *	ACD	4	Mon Jan 25 11:47:06 1993
 *		Changed #include<usr_include/string.h> to #include<string.h>
 *      ACD     3        Tue Jan  5 12:49:53 PST 1993
 *              Changed references to numWrOuts/numRdOuts to
 *                numPutOuts/numGetOuts respectively
 * 	TvE	2	Mon Jan  4 17:28:05 PST 1993
 */


#ifndef __STRING_INLINE_H__
#define __STRING_INLINE_H__

#include "/usr/include/string.h"
#include "am.h"

#define min(a, b)  (int)((a) <= (b) ? (a) : (b))

typedef struct g_str_helper_type {
  volatile int flag;
  ___C_GPTR result;
} g_strhelper;

extern int numGetOuts, numPutOuts;

void g_strlen_h(int proc, char *s, int *flag, int *lengthaddr);
void g_strchr_h(int proc, char *s, int c, g_strhelper *info);
void g_strrchr_h(int proc, char *s, int c, g_strhelper *info);

/* ...............................................................
 *
 * User Routines 
 *
 * ...............................................................
 */

EXTERN inline void  g_strinit()
GSTR_INLINE({
  fprintf(stderr, "SPLIT-C NOTE: It is no longer necessary to call g_strinit().\n");
})

/* ------------------ gstrn_{get{_ctr}, read} ---------------------- */
EXTERN inline char *gstrn_get_ctr(char *dest, g_str src, int maxlen, Counter *flag)
GSTR_INLINE({
  int len;
  int srclen;

  if (toproc(src) == MYPROC) return strncpy(dest, (char *)tolocal(src), maxlen);
  
  srclen = g_strlen(src) + 1;
  len = min(maxlen, srclen);
  __get_ctr(dest, src, len, flag);
  return dest;

})


EXTERN inline char *gstrn_read(char *dest, g_str src, int maxlen)
GSTR_INLINE({
  Counter flag = 0;
  gstrn_get_ctr(dest, src, maxlen, &flag);
  sync_ctr(&flag);
  return dest;
})


EXTERN inline char *gstrn_get(char *dest, g_str src, int maxlen)
GSTR_INLINE({
  return gstrn_get_ctr(dest, src, maxlen, &numGetOuts);
})


/* -------------------- gstrn{_put{_ctr}, _write} ----------------------- */
EXTERN inline g_str gstrn_put_ctr (g_str dest, char *src, int maxlen, Counter *flag)
GSTR_INLINE({
  int size;

  if (toproc(dest) == MYPROC)
	  return ___global_cast(strncpy((char *)tolocal(dest),src, maxlen));

  size = strlen(src) + 1;
  size = min(size, maxlen);
  bulk_put_ctr(dest, src, size, flag);
  return dest;
})

EXTERN inline g_str gstrn_put(g_str dest, char *src, int maxlen)
GSTR_INLINE({
  return gstrn_put_ctr(dest, src, maxlen, &numPutOuts);
})

EXTERN inline g_str gstrn_write(g_str dest, char *src, int maxlen)
GSTR_INLINE({
  Counter flag = 0;

  gstrn_put_ctr(dest, src, maxlen, &flag);
  sync_ctr(&flag);
  return dest;
})

/* ----------------------- g_strncpy ------------------------- */

EXTERN inline g_str g_strncpy(g_str dest, g_str src, int maxlen)
GSTR_INLINE({
  
  if (toproc(dest) == MYPROC)
    if (toproc(src) == MYPROC) 
      return ___global_cast(strncpy((char *)tolocal(dest), (char *)tolocal(src),
									maxlen));
    else
      return ___global_cast(gstrn_read((char *)tolocal(dest), src, maxlen));
  else
    if (toproc(src) == MYPROC)
      return gstrn_write(dest, (char *)tolocal(src), maxlen);
    else {
		fprintf(stderr, "FATAL error: three way string communication"
				"no longer allowed in libsplit-c\n");
		exit(1);

      /* 
       * Unfortunately, three-way communication is not possible w/ CMAM.
       *  Have to do a local copy and then a write. . . 
       */
		/*char buffer[maxlen];*/
		/*gstrn_read(buffer, src, maxlen);*/
		/*gstrn_write(dest, buffer, maxlen);*/
		/*return dest;*/
    }
})


/* ----------------------- gstr_{get{_ctr}, read} ------------------------- */
EXTERN inline char *gstr_get_ctr(char *dest, g_str src, Counter *flag)
GSTR_INLINE({  
  int len;

  if (toproc(src) == MYPROC) return strcpy(dest, (char *)tolocal(src));

  len = g_strlen(src) + 1;
  bulk_get_ctr(dest, src, len, flag);
  return dest;
})


EXTERN inline char *gstr_read(char *dest, g_str src)
GSTR_INLINE({  
  Counter flag = 0;
  gstr_get_ctr(dest, src, &flag);
  sync_ctr(&flag);
  return dest;
})


EXTERN inline char *gstr_get(char *dest, g_str src)
GSTR_INLINE({  
  return gstr_get_ctr(dest, src, &numGetOuts);
})


/* ------------------------ gstr_{put{_ctr}, write} ----------------------- */
EXTERN inline g_str gstr_put_ctr(g_str dest, char *src, Counter *flag)
GSTR_INLINE({
  int size;

  size = strlen(src) + 1;
  bulk_put_ctr(dest, src, size, flag);
  return dest;
})

EXTERN inline g_str gstr_write(g_str dest, char *src)
GSTR_INLINE({
  Counter flag = 0;
  gstr_put_ctr(dest, src, &flag);
  sync_ctr(&flag);
  return dest;
})



EXTERN inline g_str gstr_put(g_str dest, char *src)
GSTR_INLINE({
  return gstr_put_ctr(dest, src, &numPutOuts);
})


/* ----------------------- g_strcpy ------------------------- */
EXTERN inline g_str g_strcpy(g_str dest, g_str src)
GSTR_INLINE({  
  if (toproc(dest) == MYPROC)
    if (toproc(src) == MYPROC) 
      return ___global_cast(strcpy((char *)tolocal(dest), (char *)tolocal(src)));
    else
      return ___global_cast(gstr_read((char *)tolocal(dest), src));
  else
    if (toproc(src) == MYPROC)
      return gstr_write(dest, (char *)tolocal(src));
    else {
      int size = g_strlen(src) + 1; 
      {
		  fprintf(stderr, "FATAL error: three way string communication"
				  "no longer allowed in libsplit-c\n");
		  exit(1);
		  /* 
		   * Unfortunately, three-way communication is not possible w/ CMAM.
		   *  Have to do a local copy and then a write. . . 
		   */
		  /*char buffer[size];*/
		  /*gstr_read(buffer, src);*/
		  /*gstr_write(dest, buffer);*/
		  /*return dest;*/
      }
    }
})



/* ------------------------ gstrcat_{get{_ctr}, read} ---------------------- */
EXTERN inline char *gstrcat_get_ctr(char *dst, g_str src, Counter *flag)
GSTR_INLINE({
  int len;
  char *dest_end;

  if (toproc(src) == MYPROC)  return strcat(dst, (char *)tolocal(src));

  len = g_strlen(src) + 1;
  printf("%d: Length of xfer: %d\n", MYPROC, len);
  dest_end = dst + strlen(dst);
  printf("%d: dst: %x, dst_end: %x, Offset: %d\n", MYPROC, dst, dest_end, strlen(dst));
  bulk_get_ctr(dest_end, src, len, flag);
  return dst;
})


EXTERN inline char *gstrcat_read(char *dst, g_str src)
GSTR_INLINE({
  Counter flag = 0;
  gstrcat_get_ctr(dst, src, &flag);
  sync_ctr(&flag);
  return dst;
})


EXTERN inline char *gstrcat_get(char *dst, g_str src)
GSTR_INLINE({
  return gstrcat_get_ctr(dst, src, &numGetOuts);
})



/* ------------------------ gstrcat{_put{_ctr}, _write} ---------------------- */
EXTERN inline g_str gstrcat_put_ctr(g_str dst, char *src, Counter *flag)
GSTR_INLINE({  
  int size;
  ___C_GPTR dest_end;

  if (toproc(dst) == MYPROC) return ___global_cast(strcat(tolocal(dst), src));

  size = strlen(src) + 1;
  dest_end = toglobal(toproc(dst), tolocal(dst)+g_strlen(dst));
  bulk_put_ctr(dest_end, src, size, flag);
  return dst;
})


EXTERN inline g_str gstrcat_write(g_str dst, char *src)
GSTR_INLINE({  
  Counter flag = 0;
  gstrcat_put_ctr(dst, src, &flag);
  sync_ctr(&flag);
  return dst;
})


EXTERN inline g_str gstrcat_put(g_str dst, char *src)
GSTR_INLINE({  
  return gstrcat_put_ctr(dst, src, &numPutOuts);
})

/* ------------------------ g_strcat ---------------------- */
EXTERN inline g_str g_strcat(g_str dst, g_str src)
GSTR_INLINE({

  if (toproc(src) == MYPROC)
    if (toproc(dst) == MYPROC) 
      return ___global_cast(strcat((char *)tolocal(dst), (char *)tolocal(src)));
    else
      return gstrcat_write(dst, (char *)tolocal(src));
  else 
    if (toproc(dst) == MYPROC)
      return ___global_cast(gstrcat_read((char *)tolocal(dst), src));
    else {
      int size = g_strlen(src) + 1;
      {
		fprintf(stderr, "FATAL error: three way string communication"
				"no longer allowed in libsplit-c\n");
		exit(1);
		/*char buffer[size];*/
		/*gstr_read(buffer, src);*/
		/*gstrcat_write(dst, buffer);*/
		/*return dst;*/
      }
    }
})

/* ------------------------ gstrncat{_get{_ctr}, _read} ---------------------- */
EXTERN inline char *gstrncat_get_ctr(char *dst, g_str src, int maxlen, Counter *flag)
GSTR_INLINE({
  char *dest_end;
  if (toproc(src) == MYPROC) 
    return strncat(dst, (char *)tolocal(src), maxlen);

  dest_end = dst + strlen(dst);
  gstrn_get_ctr(dest_end, src, maxlen, flag);
  return dst;
})


EXTERN inline char *gstrncat_read(char *dst, g_str src, int maxlen)
GSTR_INLINE({
  char *dest_end;

  if (toproc(src) == MYPROC) 
    return strncat(dst, (char *)tolocal(src), maxlen);

  dest_end = dst + strlen(dst);
  gstrn_read(dest_end, src, maxlen);
  return dst;
})


EXTERN inline char *gstrncat_get(char *dst, g_str src, int maxlen)
GSTR_INLINE({
  char *dest_end;

  if (toproc(src) == MYPROC) 
    return strncat(dst, (char *)tolocal(src), maxlen);

  dest_end = dst + strlen(dst);
  gstrn_get(dest_end, src, maxlen);
  return dst;
})

/* ------------------------ gstrncat{_put{_ctr}, _write} ---------------------- */
EXTERN inline g_str gstrncat_put_ctr(g_str dst, char *src, int maxlen, Counter *flag)
GSTR_INLINE({
  ___C_GPTR dest_end;
  int size;

  if (toproc(dst) == MYPROC) 
    return ___global_cast(strncat((char *)tolocal(dst), src, maxlen));
  
  size = strlen(src) + 1;
  size = min(size, maxlen);
  dest_end = toglobal(toproc(dst), tolocal(dst)+g_strlen(dst));
  bulk_put_ctr(dest_end, src, size, flag);
  return dst;
})


EXTERN inline g_str gstrncat_write(g_str dst, char *src, int maxlen)
GSTR_INLINE({
  int flag = 0;
  gstrncat_put_ctr(dst, src, maxlen, &flag);
  sync_ctr(&flag);
  return dst;
})


EXTERN inline g_str gstrncat_put(g_str dst, char *src, int maxlen)
GSTR_INLINE({
  return gstrncat_put_ctr(dst, src, maxlen, &numPutOuts);
})

/* ------------------------ g_strncat ---------------------- */
EXTERN inline g_str g_strncat(g_str dst, g_str src, int maxlen)
GSTR_INLINE({
  if (toproc(src) == MYPROC)
    if (toproc(dst) == MYPROC) 
      return ___global_cast(strncat((char *)tolocal(dst), (char *)tolocal(src),
									maxlen));
    else
      return gstrncat_write(dst, (char *)tolocal(src), maxlen);
  else 
    if (toproc(dst) == MYPROC)
      return ___global_cast(gstrncat_read((char *)tolocal(dst), src, maxlen));
    else {
		fprintf(stderr, "FATAL error: three way string communication"
				"no longer allowed in libsplit-c\n");
		exit(1);
		/*char buffer[maxlen];*/
		/*gstrn_read(buffer, src, maxlen);*/
		/*gstrncat_write(dst, buffer, maxlen);*/
		/*return dst;*/
    }
})


/* ------------------------ g_strlen ------------------------ */
EXTERN inline int g_strlen(g_str s)
GSTR_INLINE({ 
  Counter flag = 0;
  int length;

  if (toproc(s) == MYPROC) return strlen((char *)tolocal(s));

  am_request_3(toproc(s), g_strlen_h, (char *)tolocal(s), &flag, &length);
  am_wait(&flag, 1);
  return length;
})

/* ------------------------ g_strncmp ---------------------- */
EXTERN inline int g_strncmp(g_str gs1, g_str gs2, int maxlen)
GSTR_INLINE({
  char b1[maxlen];
  char b2[maxlen];

  if (toproc(gs1) == MYPROC) {
    if (toproc(gs2) == MYPROC)
      return strncmp((char *)tolocal(gs1), (char *)tolocal(gs2), maxlen);
    else
      return strncmp((char *)tolocal(gs1), gstrn_read(b2, gs2, maxlen), maxlen);
  }
  
  if (toproc(gs2) == MYPROC) 
    return strncmp(gstrn_read(b1, gs1, maxlen), (char *)tolocal(gs2), maxlen);

  return strncmp(gstrn_read(b1, gs1, maxlen), gstrn_read(b2, gs2, maxlen), maxlen);
})


/* ------------------------ g_strcmp ---------------------- */
EXTERN inline int g_strcmp(g_str gs1, g_str gs2)
GSTR_INLINE({
  int size;

  if (toproc(gs1) == MYPROC)
    if (toproc(gs2) == MYPROC) 
      return strcmp((char *)tolocal(gs1), (char *)tolocal(gs2));
    else {
	char *addr = (char *)tolocal(gs1);
	size = min(strlen(addr) + 1, g_strlen(gs2) + 1);
	{
	  char buffer[size];
	  return strncmp(addr, gstrn_read(buffer, gs2, size), size);   
	}
      }
  else {
    if (toproc(gs2) == MYPROC) {
      char *addr = (char *)tolocal(gs2);
	size = min(strlen(addr) + 1, g_strlen(gs1) + 1);
	{
	  char buffer[size];
	  return strncmp(gstrn_read(buffer, gs1, size), addr, size);   
	}
      }
    else { 
      size = min(g_strlen(gs1) + 1, g_strlen(gs2) + 1);
      {
	char b1[size];
	char b2[size];
	return strncmp(gstrn_read(b1, gs1, size), gstrn_read(b2, gs2, size), size);
      }
    }
  }
})

/* --------------------- g_strdup ------------------------- */
EXTERN inline char *g_strdup(g_str gs)
GSTR_INLINE({
  char *s;
  char *tmp;
  int len;
  Counter flag = 0;

  if (toproc(gs) == MYPROC) return strdup((char *)tolocal(gs));

  len = g_strlen(gs) + 1;
  if ((s = (char *)malloc(len)) == NULL) return NULL;

  bulk_get_ctr(s, gs, len, &flag);
  sync_ctr(&flag);
  
  for (tmp = s; *tmp++ = toupper(*tmp); );
  return s;
})

/* ------------------------ g_strcasecmp ---------------------- */
EXTERN inline int g_strcasecmp(g_str gs1, g_str gs2)
GSTR_INLINE({
  int size;

  if (toproc(gs1) == MYPROC)
    if (toproc(gs2) == MYPROC) 
      return strcasecmp((char *)tolocal(gs1), (char *)tolocal(gs2));
    else {
	char *addr = (char *)tolocal(gs1);
	size = min(strlen(addr) + 1, g_strlen(gs2) + 1);
	{
	  char buffer[size];
	  return strncasecmp(addr, gstrn_read(buffer, gs2, size), size);   
	}
      }
  else {
    if (toproc(gs2) == MYPROC) {
      char *addr = (char *)tolocal(gs2);
	size = min(strlen(addr) + 1, g_strlen(gs1) + 1);
	{
	  char buffer[size];
	  return strncasecmp(gstrn_read(buffer, gs1, size), addr, size);   
	}
      }
    else { 
      size = min(g_strlen(gs1) + 1, g_strlen(gs2) + 1);
      {
	char b1[size];
	char b2[size];
	return strncasecmp(gstrn_read(b1, gs1, size), gstrn_read(b2, gs2, size), size);
      }
    }
  }
})

/* ---------------------- g_strchr ---------------------------- */
EXTERN inline g_str g_strchr(g_str s, int c)
GSTR_INLINE({
  g_strhelper info;
  int proc;

  proc = toproc(s);
  if (proc == MYPROC) return ___global_cast(strchr((char *)tolocal(s), c));

  info.flag = 0;

  am_request_3(proc, g_strchr_h, (char *)tolocal(s), c, &info);
  am_wait(&(info.flag), 1);
  return info.result;
})

/* ---------------------- g_strrchr ---------------------------- */
EXTERN inline g_str g_strrchr(g_str s, int c)
GSTR_INLINE({
  g_strhelper info;
  int proc;

  proc = toproc(s);
  if (proc == MYPROC) return ___global_cast(strrchr((char *)tolocal(s), c));

  info.flag = 0;

  am_request_3(proc, g_strrchr_h, (char *)tolocal(s), c, &info);
  am_wait(&(info.flag), 1);
  return info.result;
})

/* --------------------- g_strpbrk -------------------------- */
EXTERN inline g_str g_strpbrk(g_str gs1, g_str gs2)
GSTR_INLINE({
  int size1;
  int size2;

  if (toproc(gs1) == MYPROC) 
    if (toproc(gs2) == MYPROC)
      return ___global_cast(strpbrk((char *)tolocal(gs1), (char *)tolocal(gs2)));
    else {
      size2 = g_strlen(gs2) + 1;
      {
	char buffer[size2];
	return ___global_cast(strpbrk((char *)tolocal(gs1),
								  gstrn_read(buffer, gs2, size2)));
      }
    }
  else {
    if (toproc(gs2) == MYPROC) {
      size1 = g_strlen(gs1) + 1;
      {
	char buffer[size1];
	return toglobal(toproc(gs1), strpbrk(gstrn_read(buffer, gs1, size1),
										 (char *)tolocal(gs2)));
      }
    }
    else {
      size1 = g_strlen(gs1) + 1;
      size2 = g_strlen(gs2) + 1;
      {
	char b1[size1];
	char b2[size2];
	return toglobal (toproc(gs1), strpbrk(gstrn_read(b1, gs1, size1), 
						     gstrn_read(b2, gs2, size2)));
      }
    }
  }
})

/* --------------------- g_strspn -------------------------- */
EXTERN inline int g_strspn(g_str gs1, g_str gs2)
GSTR_INLINE({
  int size1;
  int size2;

  if (toproc(gs1) == MYPROC) 
    if (toproc(gs2) == MYPROC)
      return strspn((char *)tolocal(gs1), (char *)tolocal(gs2));
    else {
      size2 = g_strlen(gs2) + 1;
      {
	char b2[size2];
	return strspn((char *)tolocal(gs1), gstrn_read(b2, gs2, size2));
      }
    }
  else if (toproc(gs2) == MYPROC) {
    size1 = g_strlen(gs1) + 1;
    {
      char b1[size1];
      return strspn(gstrn_read(b1, gs1, size1), (char *)tolocal(gs2));
    }
  }
  else {
    size1 = g_strlen(gs1) + 1;
    size2 = g_strlen(gs2) + 1;
    {
      char b1[size1];
      char b2[size2];
      return strspn(gstrn_read(b1, gs1, size1), gstrn_read(b2, gs2, size2));
    }
  }
})

/* --------------------- g_strcspn -------------------------- */
EXTERN inline int g_strcspn(g_str gs1, g_str gs2)
GSTR_INLINE({
  int size1;
  int size2;

  if (toproc(gs1) == MYPROC)
    if (toproc(gs2) == MYPROC)
      return strcspn((char *)tolocal(gs1), (char *)tolocal(gs2));
    else {
      size2 = g_strlen(gs2) + 1;
      {
	char b2[size2];
	return strcspn((char *)tolocal(gs1), gstrn_read(b2, gs2, size2));
      }
    }
  else if (toproc(gs2) == MYPROC) {
    size1 = g_strlen(gs1) + 1;
    {
      char b1[size1];
      return strcspn(gstrn_read(b1, gs1, size1), (char *)tolocal(gs2));
    }
  }

  size1 = g_strlen(gs1) + 1;
  size2 = g_strlen(gs2) + 1;
  {
    char b1[size1];
    char b2[size2];
    return strcspn(gstrn_read(b1, gs1, size1), gstrn_read(b2, gs2, size2));
  }
})


/* --------------------- g_strstr -------------------------- */
EXTERN inline g_str g_strstr(g_str gs1, g_str gs2)
GSTR_INLINE({
  int size1;
  int size2;
   if (toproc(gs1) == MYPROC)
    if (toproc(gs2) == MYPROC)
      return ___global_cast(strstr((char *)tolocal(gs1), (char *)tolocal(gs2)));
    else {
      size2 = g_strlen(gs2) + 1;
      {
	char buffer[size2];
	return ___global_cast(strstr((char *)tolocal(gs1),
								 gstrn_read(buffer, gs2, size2)));
      }
    }
     
  else if (toproc(gs2) == MYPROC) {
    size1 = g_strlen(gs1) + 1;
    {
      char buffer[size1];
      return toglobal(toproc(gs1), 
			     strstr(gstrn_read(buffer, gs1, size1), (char *)tolocal(gs2)));
    }
  }
  else {
    
    size1 = g_strlen(gs1) + 1;
    size2 = g_strlen(gs2) + 1;
    {
      char b1[size1];
      char b2[size2];
      return toglobal (toproc(gs1), strstr(gstrn_read(b1, gs1, size1), 
						  gstrn_read(b2, gs2, size2)));
    }
  }
})

#endif /* __STRING_INLINE_H__ */






