/* 									tab:8
 *
 * memory.sc - Global memory accesses with acks (get, put, read and write)
 * 
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author:	Thorsten von Eicken
 * Version:	7
 * Filename:	memory.sc
 * History:
 *	ssl	5	Sun Dec  5 22:59:24 1993
 *		am and handlers now use unsigned instead of long long as arguments.
 *      doubles will be passed in fp registers and long longs will be dword
 *      aligned.
 *	ssl	4	Tue Nov 16 12:21:40 1993
 *	  Replaced global tags with macros.
 * 	TvE	3	Wed Dec 23 14:41:44 PST 1992
 *		Create from split-c.sc.
 */
static char _version_string_[] = "\nVersion:5:memory.sc\0\n";

#include <split-c/split-c.h>

/* .......................................
 *
 * Keep track of outstanding GETs and PUTs
 *
 * .......................................
 */

Counter numGetOuts = 0;
Counter numPutOuts = 0;

void __decr_put_ctr(void) { numPutOuts--; }

/* ..........................
 *
 * GETs with explicit counter
 *
 * ..........................
 */

void __c_get_handler(int retNode, char *addr, char *retAddr, Counter *ctr)
{
	am_reply_3(retNode, __c_get_reply_handler, retAddr, *addr, ctr);
}

void __sh_get_handler(int retNode, short *addr, short *retAddr, Counter *ctr)
{
	am_reply_3(retNode, __sh_get_reply_handler, retAddr, *addr, ctr);
}

void __i_get_handler(int retNode, int *addr, int *retAddr, Counter *ctr)
{
	am_reply_3(retNode, __i_get_reply_handler, retAddr, *addr, ctr);
}

void __f_get_handler(int retNode, int *addr, int *retAddr, Counter *ctr)
{
	am_reply_3(retNode, __f_get_reply_handler, retAddr, *addr, ctr);
}

void __d_get_handler(int retNode, double *addr, double *retAddr, Counter *ctr)
{
	/* some machines pass doubles in fp registers. force passing in int
	   registers. - ssl */
	unsigned *word0 = (unsigned*) addr;
	am_reply_4(retNode, __d_get_reply_handler, retAddr, *word0, *(word0+1),
			  ctr);
}

#ifdef __GNUC__
void __ll_get_handler(int retNode, long long *addr, long long *retAddr,
		      Counter *ctr)
{
	/* pass as two unsigned so that HPs won't dword align long long - ssl */
	unsigned *word0 = (unsigned*)addr;
	am_reply_4(retNode, __ll_get_reply_handler, retAddr, *word0, *(word0+1),
			  ctr);
}
#endif

void __c_get_reply_handler(int src_node, char *addr, int data, Counter *ctr)
{
	addr[0] = data;
	__decr_ctr(ctr);
}

void __sh_get_reply_handler(int src_node, short *addr, int data, Counter *ctr)
{
	addr[0] = data;
	__decr_ctr(ctr);
}

void __i_get_reply_handler(int src_node, int *addr, int data, Counter *ctr)
{
	addr[0] = data;
	__decr_ctr(ctr);
}

void __f_get_reply_handler(int src_node, int *addr, int data, Counter *ctr)
{
	addr[0] = data;
	__decr_ctr(ctr);
}

/* must talk unsigned instead of double as args because some machines
   pass doubles in fp registers. - ssl */
void __d_get_reply_handler(int src_node, double *addr, unsigned word0,
													 unsigned word1, Counter *ctr)
{
	*((unsigned*)addr) = word0;
	*((unsigned*)addr+1) = word1;
	__decr_ctr(ctr);
}

#ifdef __GNUC__
/* pass as two unsigned so that HPs won't dword align long long - ssl */
void __ll_get_reply_handler(int src_node, long long *addr, unsigned word0,
														unsigned word1, Counter *ctr)
{
	*((unsigned*)addr) = word0;
	*((unsigned*)addr+1) = word1;
	__decr_ctr(ctr);
}
#endif


/* ..........................
 *
 * PUTs with explicit counter
 *
 * ..........................
 */

void __c_put_handler(int retNode, char *addr, int data, Counter *ctr)
{
	*addr = data;
	am_reply_1(retNode, __decr_ctr_h, ctr);
}

void __sh_put_handler(int retNode, short *addr, int data, Counter *ctr)
{
	*addr = data;
	am_reply_1(retNode, __decr_ctr_h, ctr);
}

void __i_put_handler(int retNode, int *addr, int data, Counter *ctr)
{
	*addr = data;
	am_reply_1(retNode, __decr_ctr_h, ctr);
}

void __f_put_handler(int retNode, float *addr, int data, Counter *ctr)
{
	*(int *)addr = data;
	am_reply_1(retNode, __decr_ctr_h, ctr);
}

/* ..........................
 *
 * PUTs with implicit counter
 *
 * ..........................
 */

/* must pass as two unsigneds because some machines pass doubles in fp
   registers. - ssl */
void __d_put_implicit_handler(int retNode, double *addr, unsigned word0,
							  unsigned word1)
{
	*((unsigned*)addr) = word0;
	*((unsigned*)addr+1) = word1;
	am_reply_0(retNode, __decr_put_ctr);
}

#ifdef __GNUC__
void __ll_put_implicit_handler(int retNode, long long *addr, unsigned word0,
							   unsigned word1)
{
	*((unsigned*)addr) = word0;
	*((unsigned*)addr+1) = word1;
	am_reply_0(retNode, __decr_put_ctr);
}
#endif

/* ......
 *
 * WRITEs
 *
 * ......
 */

/* Special counter for double/long-long writes */
Counter __write_ctr=0;
void __decr_write_ctr() { __write_ctr = 0; }

/* pass as unsigned so it will not pass in fp registers */
void __d_write_handler(int retNode, double *addr, unsigned word0,
					   unsigned word1)
{
	*((unsigned*)addr) = word0;
	*((unsigned*)addr+1) = word1;
	am_reply_0(retNode, __decr_write_ctr);
}

#ifdef __GNUC__
void __ll_write_handler(int retNode, long long *addr, unsigned word0,
						unsigned word1)
{
	*((unsigned*) addr) = word0;
	*((unsigned*) addr+1) = word1;
	am_reply_0(retNode, __decr_write_ctr);
}
#endif





