/*									tab:2
 *
 * am.h - Paragon Active Messages include file
 *
 * "Copyright (c) 1994 by Lok Tin Liu and The Regents of the University 
 * of California.  All rights reserved."
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 			Lok Tin Liu
 * Version:			165
 * Creation Date:	Wed Aug 17 13:58:55 1994
 * Filename:		am.h
 * Description:
 *    This file contains the PAM header file developed by LTL at UCB.
 *    It contains no Intel proprietary information.
 * History:
 *	LTL	160	Sun Jul 23 13:04:29 1995
 *		Made am_request_df() and am_reply_df() consistent with GAM spec 1.1.
 *	LTL	157	Sun Jul 23 12:14:25 1995
 *		Removed inline declaration for am_enq_7i(), am_poll(), and am_wait()
 *    so that we can use cc to compile instead of gcc
 *	LTL	151	Thu Jul  6 21:04:29 1995
 *		PAM Release 1.6
 */



#ifndef PAM_H
#define PAM_H

#include <values.h>

#define AM_MAX_PKT MAXINT


/* an entry in the receive buffer -- fill in one 8-word cache line */
typedef struct pam_recbuf_entry {
  unsigned long flag; /* set when entry is  valid */
	unsigned long arg0, arg1, arg2, arg3, arg4, arg5, arg6;
} pam_recbuf_t;

typedef unsigned int vnn_t;
typedef void (*handler_t)();
		

extern int am_enable(int max_outstanding, volatile pam_recbuf_t **recbuf_base,
										 volatile pam_recbuf_t **recbuf_bound);


extern int am_disable();
extern int am_procs();
extern vnn_t am_my_proc();
extern int am_max_size();

extern int am_enq_7i(vnn_t dest_node, int placeholder, int arg0,
										 int arg1, int arg2,
										 int arg3, int arg4,
										 int arg5,	int arg6);

extern void am_poll();


extern void am_wait(volatile int *flag, int value);

extern void am_barrier();

#define barrier() am_barrier()

#ifdef PAM_INTERNAL
extern int am_request_5(vnn_t dest, handler_t handler, int arg0,
															 int arg1, int arg2, int arg3, int arg4);
extern int am_send_5(vnn_t dest, handler_t handler, int arg0,
															 int arg1, int arg2, int arg3, int arg4);
#else
extern int am_request_5(vnn_t dest, handler_t handler, ...);
extern int am_send_5(vnn_t dest, handler_t handler, ...);
#endif

#define am_request_4(dest, handler, int_arg0, int_arg1, int_arg2, int_arg3) \
		am_request_5(dest, handler, int_arg0, int_arg1, int_arg2, int_arg3, 0)

#define am_request_3(dest, handler, int_arg0, int_arg1, int_arg2) \
		am_request_5(dest, handler, int_arg0, int_arg1, int_arg2, 0, 0)		
								 
#define am_request_2(dest, handler, int_arg0, int_arg1) \
		am_request_5(dest, handler, int_arg0, int_arg1, 0, 0, 0)
		
#define am_request_1(dest, handler, int_arg0) \
		am_request_5(dest, handler, int_arg0, 0, 0, 0, 0)

#define am_request_0(dest, handler) \
		am_request_5(dest, handler, 0, 0, 0, 0, 0)		

extern int am_request_df(vnn_t dest, handler_t handler,
												 int arg1, int arg2, double arg3);

#ifdef PAM_INTERNAL
extern int am_reply_5(vnn_t dest, handler_t handler, int arg0, int arg1,
											int arg2, int arg3, int arg4);
#else
extern int am_reply_5(vnn_t dest, handler_t handler, ...);
#endif
														 
#define am_reply_4(dest, handler, int_arg0, int_arg1, int_arg2, int_arg3) \
		am_reply_5(dest, handler, int_arg0, int_arg1, int_arg2, int_arg3, 0)

#define am_reply_3(dest, handler, int_arg0, int_arg1, int_arg2) \
		am_reply_5(dest, handler, int_arg0, int_arg1, int_arg2, 0, 0)		
								 
#define am_reply_2(dest, handler, int_arg0, int_arg1) \
		am_reply_5(dest, handler, int_arg0, int_arg1, 0, 0, 0)
		
#define am_reply_1(dest, handler, int_arg0) \
		am_reply_5(dest, handler, int_arg0, 0, 0, 0, 0)

#define am_reply_0(dest, handler) \
		am_reply_5(dest, handler, 0, 0, 0, 0, 0)		

extern int am_reply_df(vnn_t dest, handler_t handler, 
															int arg1, int arg2, double arg3);
		

extern int am_store_async(vnn_t dest_node, void *src_buf, void *dest_buf,
							int len, handler_t request_handler, void *request_handler_arg1,
							handler_t endfunc, void *endfunc_arg1);

extern int am_store(vnn_t dest_node, void *src_buf, void *dest_buf, int len, 
						 handler_t request_handler, void *request_handler_arg1);

extern int am_get(vnn_t dest_node, void *src_buf, void *dest_buf, int len,
					 handler_t reply_handler, void *handler_arg1);
												
#endif







