/*									tab:8
 *
 * bulk.c - Handlers for bulk functions.
 *
 * 
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 		Steve Luna
 * Version:		31
 * Creation Date:	Sat Nov 27 23:07:14 1993
 * Filename:		bulk.c
 * History:
 *	LTL	29	Mon Jul 24 22:03:26 1995
 *		Use am_store_async() instead of am_store() for store
 *	LTL	8	Sun Jul 10 15:06:21 1994
 *		Modified for TCPAM
 *	RPM	2	Sun Nov 28 22:33:34 1993
 *		new version for PMAM interface, bonged segments
 */

#include <split-c/split-c.h>



void __GetComplete(int src_node, void *buf, int len, void *info)
{
  if (info != NULL) 
    __decr_ctr(info);
}


void __bulk_put_handler(int dest, void *buf, int len, void *cnt)
{

  if (cnt != NULL) 
    am_reply_1(dest,__PutCtrCompleteAtSource,cnt);
}

void __PutCtrCompleteAtSource(int src_node, Counter *ctr)
{
  (*ctr)--;

}

/* ............................
 *
 * STOREs with explicit counter
 *
 * ............................
 */


static void store_handler(int src_node, void *buf, int len, void *arg)
{
	recvBytes += (int)len;
}


/* STORE bulk */
void __store(___V_GPTR gPtr, void *lPtr, int len)
{
	Processor pid = toproc(gPtr);
	void *addr    = tolocal(gPtr);
	int i;

	if(pid == MYPROC) {
		sentBytes += len;
		memcpy(addr,lPtr, len);
		recvBytes += len;
	} else {
	 if (len <= ___AM_MAX_PKT) {
 		 /* endfunc is (void *)0; the message processor won't enqueue the
 		  * endfunc when the DMA on the sender completes */
		 am_store_async(pid, lPtr, addr, len, store_handler, 0,
										 (void *)0, (void *)0);										
   }
	 else {
     for (i = ___AM_MAX_PKT; i < len; i += ___AM_MAX_PKT) {
       /* endfunc is (void *)0; the message processor won't enqueue the
		    * endfunc when the DMA on the sender completes */				 
   		 am_store_async(pid, lPtr, addr, ___AM_MAX_PKT, store_handler, 0,
										 (void *)0, (void *)0);								
			 (char *)addr += ___AM_MAX_PKT;
			 (char *)lPtr += ___AM_MAX_PKT;			 
		 }
		 i -= ___AM_MAX_PKT;
		 /* endfunc is (void *)0; the message processor won't enqueue the
			* endfunc when the DMA on the sender completes */
		 am_store_async(pid, lPtr, addr, len - i, store_handler, 0,
										 (void *)0, (void *)0);		 
	 }
	 sentBytes += len;
	}
}

void
bulk_store(___V_GPTR gPtr, void *lPtr, int len)
{
  __store(gPtr, lPtr, len);
}







