/*									tab:8
 *
 * malloc_int.h - data structures for split-c malloc
 *
 * 
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 		Steve Luna
 * Version:		1
 * Creation Date:	Tue Dec  7 05:42:49 1993
 * Filename:		malloc_int.h
 * History:
 */

/*
 * Strategy: this is a binned memory allocator. There are separate bin tables
 * for global and local heaps. A bin table keeps a separate
 * free list for each bin size. Bin sizes are multiples of the bin granularity.
 * When a bin is empty, a fixed-size block (roughly a VM page) is allocated
 * and fragmented into the bin size. The per-bin free list is organized in
 * two levels: it is a list of blocks allocated for that bin size, and each
 * block holds a list of free fragments. Fragments larger than the largest
 * bin size are stored on a linked list. Elements on this list are allocated
 * using a best fit strategy. 
 */

#define	SC_MALLOC_SHFT	(5)
#define	SC_MALLOC_GRAN	(1<<SC_MALLOC_SHFT)
#define	SC_MALLOC_BINS	((1100/SC_MALLOC_GRAN)+1)
#define SC_MAX_BIN_SIZE ((SC_MALLOC_BINS-1)*SC_MALLOC_GRAN)

#define	SC_MALLOC_MINFRAGS	(10)	/* min frags in a block */
#define SC_MALLOC_MINBLK	(4*1024)	/* min block size */
#define WHICH_BIN(global, elmt) ((global) ? (&__gsc_malloc_bin[elmt]) : \
								 (&__lsc_malloc_bin[elmt]))
#define WHICH_QUEUE(global) ((global) ? (&__gbig_queue) : (&__lbig_queue))
#define MAGIC_KEY 0x12345678

typedef struct sc_malloc_frag sc_malloc_frag;
typedef struct sc_malloc_block sc_malloc_block;

struct sc_malloc_frag
{
    union
	{
		sc_malloc_frag *next;	/* frag free: linear list in block */
		sc_malloc_block *block;	/* frag used: point back to block */
	}
    chain;
    int size;
	unsigned key;				/* identifies a valid frag when freeing */
	int padding;				/* make dword aligned */
};

struct sc_malloc_block
{
    sc_malloc_block *next;	/* circular list */
    sc_malloc_block *head;	/* point back to bin table */
    sc_malloc_frag *free;	/* linear list */
    int padding;		    /* dword align */
};

#ifndef SC_MALLOC_INTERNAL_DEFINE
extern sc_malloc_block __gsc_malloc_bin[SC_MALLOC_BINS];
extern sc_malloc_block __lsc_malloc_bin[SC_MALLOC_BINS];
extern sc_malloc_frag __gbig_queue;
extern sc_malloc_frag __lbig_queue;
#endif
