/*									tab:8
 *
 * memory.h - Global memory accesses with acks (get, put, read and write)
 *            Implementation dependent definitions and declarations.
 *
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 		Arvind Krishnamurthy
 * Version:		68
 * Creation Date:	Tue Sep 22 23:54:02 1992
 * Filename:		memory.h
 * History:
 *	LTL	67	Mon Jul 24 21:13:24 1995
 *		Removed #ifdef PMAM
 *	ssl	49	Mon Mar 21 09:47:12 1994
 *		sync for HPAM can use poll but PMAM cannot
 *	ssl	48	Sun Mar 20 14:22:13 1994
 *		sync on hps can use poll now
 *	ssl	46	Fri Mar 11 16:12:35 1994
 *		sync must use am_wait_until_zero for hpam, because am_poll does
 *      not retransmit
 *	ssl	45	Sun Dec  5 23:02:43 1993
 *		am and handlers now use unsigned instead of long long as arguments.
 *      doubles will be passed in fp registers and long longs will be dword
 *      aligned.
 *	ssl	44	Sat Nov 27 22:57:58 1993
 *		Moved __get* __put* __write to bulk.h. Included split-c.h
 *	ssl	43	Tue Nov 23 03:26:44 1993
 *		__d_get_ctr, __d_put, ... store int's instead of doubles.
 *      i860 passes doubles in fp registers.
 *	ssl	42	Tue Nov 16 12:00:34 1993
 *		Replaced global tags with macros.
 *	ssl	41	Thu Oct 28 21:59:55 1993
 *		Include am.h instead of cm5.h
 *      Took out __put_rvalue stuff.
 *	AK	40	Thu Jun 17 15:13:53 1993
 *		Commented out the check for < 8 in __get_ctr.
 *		Compiler does the check currently.
 *	SL	38	Thu May 13 17:01:47 1993
 *		Added SC_PROFILE information
 *	AK	37	Tue Mar 16 14:33:33 1993
 *		_*_struct_reads return pointers, not structures.
 *	AK	34	Sat Jan 23 01:47:53 1993
 *		Modified __read to report errors.
 * 	TvE	32	Sat Dec 19 00:15:28 PST 1992
 *		Started major clean-up and overhaul.
 *		Extracted from split-c_int.h
 *	AK	1	Sat Nov 28 21:17:44 1992
 *		changed include file name from cmam2-4
 * 	TvE	31	Tue Sep 29 12:47:46 PDT 1992
 *		changed include file names
 */

#ifndef __MEMORY_H
#define __MEMORY_H

/* .....................
 *
 * Profiling information
 *
 * .....................
 */

#include <stdio.h>
#include <split-c/split-c.h>

#ifdef SC_PROFILE
#include <split-c/profile.h>
#endif

extern inline void __get_ctr(void *lPtr, ___V_GPTR gPtr, int len,
							 Counter *ctr);
extern inline void __put_ctr(___V_GPTR gPtr, void *lPtr, int len,
							 Counter *ctr);
extern inline void __put(___V_GPTR gPtr, void *lPtr, int len);
extern inline void __write(___V_GPTR gPtr, void *lPtr, int len);

/* .......................................
 *
 * Keep track of outstanding GETs and PUTs
 *
 * .......................................
 */

/* A local variable which is incremented and then we wait for it to go to zero.
   Used as the implicit flag in all non-blocking operations */
extern Counter numGetOuts;
extern Counter numPutOuts;

EXTERN inline void __decr_ctr_h(int src_node, Counter *ctr)
INLINE({
	*ctr -= 1;
})

EXTERN inline void __decr_ctr(Counter *ctr)
INLINE({
	*ctr -= 1;
})

EXTERN inline void __incr_ctr(Counter *ctr)
INLINE({
	*ctr += 1;
})

/* .........................
 *
 * Synchronize GETs and PUTs
 *
 * .........................
 */



/* Keep ifdef's out of macros. Native compiler won't take them. - ssl */
#if SC_PROFILE /* Waiting for synchronization */
EXTERN inline void sync(void)
INLINE({
	while(numGetOuts || numPutOuts) {
		sc_prof_data[PROF_SYNC]++;
		am_poll();
	}
})
#else

EXTERN inline void sync(void)
INLINE({
	while (numGetOuts || numPutOuts) {
		am_poll();
	}
}) 
#endif		

/* Check whether all of this node's split-phase ops have completed */
EXTERN inline int is_sync(void)
INLINE({
	am_poll();
	return !(numGetOuts || numPutOuts);
})

/* ..........................
 *
 * GETs with explicit counter
 *
 * ..........................
 */

/* Handler on the remote end to send value back */
extern void __c_get_handler (int retNode, char *addr,
			     char *retAddr, Counter *ctr);
extern void __sh_get_handler(int retNode, short *addr,
			     short *retAddr, Counter *ctr);
extern void __i_get_handler (int retNode, int *addr,
			     int *retAddr, Counter *ctr);
extern void __f_get_handler (int retNode, int *addr,
			     int *retAddr, Counter *ctr);
extern void __d_get_handler (int retNode, double *addr,
			     double *retAddr, Counter *ctr);
#ifdef __GNUC__
extern void __ll_get_handler(int retNode, long long *addr,
			     long long *retAddr, Counter *ctr);
#endif

/* Handlers on the local end to receive value */
extern void __c_get_reply_handler (int src_node, char *addr, int data,
																	 Counter *ctr);
extern void __sh_get_reply_handler(int src_node, short *addr, int data,
																	 Counter *ctr);
extern void __i_get_reply_handler (int src_node, int *addr, int data,
																	 Counter *ctr);
extern void __f_get_reply_handler (int src_node, int *addr, int data,
																	 Counter *ctr);
extern void __d_get_reply_handler (int src_node, double *addr,  unsigned word0,
								   unsigned word1, Counter *ctr);
#ifdef __GNUC__
extern void __ll_get_reply_handler(int src_node, long long *adr,
																	 unsigned word0, unsigned word1,
																	 Counter *ctr);
#endif

/* GET character with explicit counter */
void __c_get_ctr(char *lPtr, ___V_GPTR gPtr, Counter *ctr);
EXTERN inline void __c_get_ctr(char *lPtr, ___V_GPTR gPtr, Counter *ctr)
INLINE({
	int  pid   = toproc(gPtr);
	char *addr = tolocal(gPtr);

	if (pid == MYPROC) {
		*lPtr = *addr;
	} else {
		__incr_ctr(ctr);
		am_request_3(pid, __c_get_handler, addr, lPtr, ctr);
	}
})

/* GET short with explicit counter */
EXTERN inline void __sh_get_ctr(short *lPtr, ___V_GPTR gPtr, Counter *ctr)
INLINE({
	int  pid    = toproc(gPtr);
	short *addr = tolocal(gPtr);

	if (pid == MYPROC) {
		*lPtr = *addr;
	} else {
		__incr_ctr(ctr);
		am_request_3(pid, __sh_get_handler, addr, lPtr, ctr);
	}
})

/* GET int with explicit counter */
EXTERN inline void __i_get_ctr(int *lPtr, ___V_GPTR gPtr, Counter *ctr)
INLINE({
	int  pid  = toproc(gPtr);
	int *addr = tolocal(gPtr);

	if (pid == MYPROC) {
		*lPtr = *addr;
	} else {
		__incr_ctr(ctr);
		am_request_3(pid, __i_get_handler, addr, lPtr, ctr);
	}
})

/* GET float with explicit counter */
EXTERN inline void __f_get_ctr(float *lPtr, ___V_GPTR gPtr, Counter *ctr)
INLINE({
	int  pid    = toproc(gPtr);
	float *addr = tolocal(gPtr);

	if (pid == MYPROC) {
		*lPtr = *addr;
	} else {
		__incr_ctr(ctr);
		am_request_3(pid, __f_get_handler, addr, lPtr, ctr);
	}
})

/* GET float with explicit counter */
EXTERN inline void __d_get_ctr(double *lPtr, ___V_GPTR gPtr, Counter *ctr)
INLINE({
	int  pid     = toproc(gPtr);
	double *addr = tolocal(gPtr);

	if (pid == MYPROC) {
		*lPtr = *addr;
	} else {
		__incr_ctr(ctr);
		am_request_3(pid, __d_get_handler, addr, lPtr, ctr);
	}
})

#ifdef __GNUC__
/* GET float with explicit counter */
EXTERN inline void __ll_get_ctr(long long *lPtr, ___V_GPTR gPtr,Counter *ctr)
INLINE({
	int  pid        = toproc(gPtr);
	long long *addr = tolocal(gPtr);

	if (pid == MYPROC) {
		*lPtr = *addr;
	} else {
		__incr_ctr(ctr);
		am_request_3(pid, __ll_get_handler, addr, lPtr, ctr);
	}
})
#endif


/* ..........................
 *
 * GETs with implicit counter
 *
 * ..........................
 */

EXTERN inline void __c_get(char *lPtr, ___V_GPTR gPtr)
INLINE({__c_get_ctr(lPtr, gPtr, &numGetOuts); })

EXTERN inline void __sh_get(short *lPtr, ___V_GPTR gPtr)
INLINE({ __sh_get_ctr(lPtr, gPtr, &numGetOuts); })

EXTERN inline void __i_get(int *lPtr, ___V_GPTR gPtr)
INLINE({ __i_get_ctr(lPtr, gPtr, &numGetOuts); })

EXTERN inline void __f_get(float *lPtr, ___V_GPTR gPtr)
INLINE({ __f_get_ctr(lPtr, gPtr, &numGetOuts); })

EXTERN inline void __d_get(double *lPtr, ___V_GPTR gPtr)
INLINE({ __d_get_ctr(lPtr, gPtr, &numGetOuts); })

#ifdef __GNUC__	
EXTERN inline void __ll_get(long long *lPtr, ___V_GPTR gPtr)
INLINE({ __ll_get_ctr(lPtr, gPtr, &numGetOuts); })
#endif	

EXTERN inline void __struct_i_get(void *lPtr, ___V_GPTR gPtr)
INLINE({ __i_get_ctr(lPtr, gPtr, &numGetOuts); })

#ifdef __GNUC__
EXTERN inline void __struct_d_get(void *lPtr, ___V_GPTR gPtr)
INLINE({ __ll_get_ctr(lPtr, gPtr, &numGetOuts); })
#endif

/* ..........................
 *
 * PUTs with explicit counter
 *
 * ..........................
 */

/* Handler on the remote end to write value */
extern void __c_put_handler (int retNode, char *addr,  int data, Counter *ctr);
extern void __sh_put_handler(int retNode, short *addr, int data, Counter *ctr);
extern void __i_put_handler (int retNode, int *addr,   int data, Counter *ctr);
extern void __f_put_handler (int retNode, float *addr, int data, Counter *ctr);

/* PUT character with explicit counter */
EXTERN inline void __c_put_ctr(___V_GPTR gPtr, char val, Counter *ctr)
INLINE({
	int pid    = toproc(gPtr);
	char *addr = tolocal(gPtr);

	if (pid == MYPROC) {
		*addr = val;
	} else {
		__incr_ctr(ctr);
		am_request_3(pid, __c_put_handler, addr, val, ctr);
	}
})

/* PUT short with explicit counter */
EXTERN inline void __sh_put_ctr(___V_GPTR gPtr, short val, volatile int *ctr)
INLINE({
	int pid     = toproc(gPtr);
	short *addr = tolocal(gPtr);

	if (pid == MYPROC) {
		*addr = val;
	} else {
		__incr_ctr(ctr);
		am_request_3(pid, __sh_put_handler, addr, val, ctr);
	}
})

/* PUT int with explicit counter */
EXTERN inline void __i_put_ctr(___V_GPTR gPtr, int val, volatile int *ctr)
INLINE({
	int pid   = toproc(gPtr);
	int *addr = tolocal(gPtr);

	if (pid == MYPROC) {
		*addr = val;
	} else {
		__incr_ctr(ctr);
		am_request_3(pid, __i_put_handler, addr, val, ctr);
	}
})

/* PUT float with explicit counter */
EXTERN inline void __f_put_ctr(___V_GPTR gPtr, float val, volatile int *ctr)
INLINE({
	int pid     = toproc(gPtr);
	float *addr = tolocal(gPtr);

	if (pid == MYPROC) {
		*addr = val;
	} else {
		__incr_ctr(ctr);
		am_request_3(pid, __f_put_handler, addr, *(int *)&val, ctr);
	}
})

/* PUT double/long-long with explicit counter don't fit into a packet... */

EXTERN void __put_ctr(___V_GPTR gPtr, void *lPtr, int len, Counter *ctr);

/* ..........................
 *
 * PUTs with implicit counter
 *
 * ..........................
 */

extern void __d_put_implicit_handler (int retNode, double *addr,
									  unsigned word0, unsigned word1);

#ifdef __GNUC__
extern void __ll_put_implicit_handler(int retNode, long long *addr,
									  unsigned word0, unsigned word1);
#endif

EXTERN inline void __c_put(___V_GPTR gPtr, char val)
INLINE({ __c_put_ctr(gPtr, val, &numPutOuts); })

EXTERN inline void __sh_put(___V_GPTR gPtr, short val)
INLINE({ __sh_put_ctr(gPtr, val, &numPutOuts); })

EXTERN inline void __i_put(___V_GPTR gPtr, int val)
INLINE({ __i_put_ctr(gPtr, val, &numPutOuts); })

EXTERN inline void __f_put(___V_GPTR gPtr, float val)
INLINE({ __f_put_ctr(gPtr, val, &numPutOuts); })

/* PUT double with implicit counter.
 * Fits in one packet whereas explicit counter version doesn't */
EXTERN inline void __d_put(___V_GPTR gPtr, double val)
INLINE({
	int pid      = toproc(gPtr);
	double *addr = tolocal(gPtr);
	unsigned *word0 = (unsigned*)&val;

	if (pid == MYPROC) {
		*addr = val;
	} else {
		__incr_ctr(&numPutOuts);
		/* pass as two unsigneds because some machines pass doubles in
		   fp registers. - ssl */
		am_request_3(pid, __d_put_implicit_handler, addr, *word0,
					*(word0+1));
	}
})

#ifdef __GNUC__
/* PUT double with implicit counter.
 * Fits in one packet whereas explicit counter version doesn't */
EXTERN inline void __ll_put(___V_GPTR gPtr, long long val)
INLINE({
	int pid         = toproc(gPtr);
	long long *addr = tolocal(gPtr);
	unsigned *word0 = (unsigned*) &val;

	if (pid == MYPROC) {
		*addr = val;
	} else {
		__incr_ctr(&numPutOuts);
		/* pass as two unsigned so that HPs don't dword align long long */
		am_request_3(pid, __ll_put_implicit_handler, addr, *word0,
					*(word0+1));
	}
})
#endif

/* PUT 4 byte struct with implicit counter */
typedef struct { int data; } ___Struct_i;
EXTERN inline void __struct_i_put(___V_GPTR gPtr, ___Struct_i val)
INLINE({
	int pid           = toproc(gPtr);
	___Struct_i *addr = tolocal(gPtr);

	if (pid == MYPROC) {
		*addr = val;
	} else {
		__incr_ctr(&numPutOuts);
		am_request_3(pid, __i_put_handler, addr,
		       val.data, &numPutOuts);
	}
})

#ifdef __GNUC__
/* PUT 8 byte struct with implicit counter */
typedef struct { long long data; } ___Struct_d;
EXTERN inline void __struct_d_put(___V_GPTR gPtr, ___Struct_d val)
INLINE({

	int pid           = toproc(gPtr);
	___Struct_d *addr = tolocal(gPtr);

	if (pid == MYPROC) {
		*addr = val;
	} else {
		__incr_ctr(&numPutOuts);
		am_request_2(pid, __ll_put_implicit_handler, addr, val.data);
	}
})
#endif

EXTERN void __put(___V_GPTR gPtr, void *lPtr, int len);

/* .....
 *
 * READs
 *
 * .....
 */

/* READ character */
EXTERN inline char __c_read(___V_GPTR gPtr)
INLINE({
	char val;
	Counter ctr=0;
	
	__c_get_ctr(&val, gPtr, &ctr);
	sync_ctr(&ctr);
	return val;
})

/* READ short */
EXTERN inline short __sh_read(___V_GPTR gPtr)
INLINE({
	short val;
	Counter ctr=0;
	
	__sh_get_ctr(&val, gPtr, &ctr);
	sync_ctr(&ctr);
	return val;
})

/* READ int */
EXTERN inline int __i_read(___V_GPTR gPtr)
INLINE({
	int val;
	Counter ctr=0;

	__i_get_ctr(&val, gPtr, &ctr);
	sync_ctr(&ctr);
	return val;
})

/* READ float */
EXTERN inline float __f_read(___V_GPTR gPtr)
INLINE({
	float val;
	Counter ctr=0;

	__f_get_ctr(&val, gPtr, &ctr);
	sync_ctr(&ctr);
	return val;
})

/* READ double */
EXTERN inline double __d_read(___V_GPTR gPtr)
INLINE({
	double val;
	Counter ctr=0;

	__d_get_ctr(&val, gPtr, &ctr);
	sync_ctr(&ctr);
	return val;
})

#ifdef __GNUC__
/* READ long long */
EXTERN inline long long __ll_read(___V_GPTR gPtr)
INLINE({
	long long val;
	Counter ctr=0;

	__ll_get_ctr(&val, gPtr, &ctr);
	sync_ctr(&ctr);
	return val;
})
#endif

EXTERN	___Struct_i val_i;     /* this is not re-entrant */  

/* READ 4 byte struct */
EXTERN inline ___Struct_i * __struct_i_read(___V_GPTR gPtr)
INLINE({

	Counter ctr=0;

	__i_get_ctr(&val_i.data, gPtr, &ctr);
	sync_ctr(&ctr);
	return &val_i;
})

#ifdef __GNUC__
EXTERN 	___Struct_d val_d;   /* this is not re-entrant */

/* READ 8 bytes struct */
EXTERN inline ___Struct_d * __struct_d_read(___V_GPTR gPtr)
INLINE({

	Counter ctr=0;

	__ll_get_ctr(&val_d.data, gPtr, &ctr);
	sync_ctr(&ctr);
	return &val_d;
})
#endif

/* ......
 *
 * WRITEs
 *
 * ......
 */

extern Counter __write_ctr;
extern void __d_write_handler (int retNode, double *addr, unsigned word0,
							   unsigned word1);
#ifdef __GNUC__
extern void __ll_write_handler(int retNode, long long *addr, unsigned word0,
							   unsigned word1);
#endif

/* Why do these functions return a value??? -TvE */

/* WRITE character */
EXTERN inline char __c_write(___V_GPTR gPtr, char val)
INLINE({
	Counter ctr=0;

	__c_put_ctr(gPtr, val, &ctr);
	sync_ctr(&ctr);
	return val;
})

/* WRITE short */
EXTERN inline short __sh_write(___V_GPTR gPtr, short val)
INLINE({
	Counter ctr=0;

	__sh_put_ctr(gPtr, val, &ctr);
	sync_ctr(&ctr);
	return val;
})

/* WRITE int */
EXTERN inline int __i_write(___V_GPTR gPtr, int val)
INLINE({
	Counter ctr=0;

	__i_put_ctr(gPtr, val, &ctr);
	sync_ctr(&ctr);
	return val;
})

/* WRITE float */
EXTERN inline float __f_write(___V_GPTR gPtr, float val)
INLINE({
	Counter ctr=0;

	__f_put_ctr(gPtr, val, &ctr);
	sync_ctr(&ctr);
	return val;

})

/* WRITE double
 * Fits in one packet whereas explicit counter PUT doesn't */
EXTERN inline double __d_write(___V_GPTR gPtr, double val)
INLINE({
	int pid      = toproc(gPtr);
	double *addr = tolocal(gPtr);
	unsigned *word0 = (unsigned*)&val;

	if (pid == MYPROC) {
		*addr = val;
	} else {
		__incr_ctr(&__write_ctr);
		/* pass as unsigned so it will not pass in fp registers */
		am_request_3(pid, __d_write_handler, addr, *word0, *(word0+1));
		sync_ctr(&__write_ctr);
	}
	return val;
})

#ifdef __GNUC__
/* WRITE long long
 * Fits in one packet whereas explicit counter PUT doesn't */
EXTERN inline long long __ll_write(___V_GPTR gPtr, long long val)
INLINE({
	int pid         = toproc(gPtr);
	long long *addr = tolocal(gPtr);
	unsigned *word0 = (unsigned*) &val;

	if (pid == MYPROC) {
		*addr = val;
	} else {
		__incr_ctr(&__write_ctr);
		/* pass as unsigned so it won't dword align */
		am_request_3(pid, __ll_write_handler, addr, *word0, *(word0+1));
		sync_ctr(&__write_ctr);
	}
	return val;
})
#endif

/* WRITE 4 byte struct */
EXTERN inline ___Struct_i __struct_i_write(___V_GPTR gPtr, ___Struct_i val)
INLINE({
	__i_write(gPtr, val.data);
	return val;
})

#ifdef __GNUC__
/* WRITE 8 byte struct */
EXTERN inline ___Struct_d __struct_d_write(___V_GPTR gPtr, ___Struct_d val)
INLINE({
	__ll_write(gPtr, val.data);
	return val;
})
#endif

EXTERN void __write(___V_GPTR gPtr, void *lPtr, int len);

/* ................................
 *
 * Ununderstandable garbage... -TvE
 *
 * ................................
 */

/* A wrapper required to return the structure */
EXTERN inline void __read_local(void *lPtr, ___V_GPTR gPtr, int len)
INLINE({
	volatile int flag=0;

	__get_ctr(lPtr, gPtr, len, &flag);
	while (flag != 0) am_poll();
})

/* This function should never be executed. Used
   internally by the compiler but should eventually
   be replaced by a C expression or by __read_local. */
EXTERN inline void * __read (___V_GPTR p, int size)
INLINE({
        printf ("Split-C internal error: please report the bug\n");
	return NULL;
})

/* what is this doing in the library? - ssl */
/*EXTERN inline foo(___V_GPTR gPtr, char val)
INLINE({
    __c_write(gPtr, val);
})*/
#endif __MEMORY_H
