/*									tab:8
 *
 * pointers.h - Make global pointers happen. Compiler specific stuff.
 *
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 		Thorsten von Eicken
 * Version:		12
 * Creation Date:	Sat Dec 19 00:47:44 PST 1992
 * Filename:		pointers.h
 * History:
 *	ssl	11	Sun Nov 21 21:19:30 1993
 *		Alternate representation for globPtrType when compiling
 *      with native compiler.
 *	ssl	10	Tue Nov 16 12:03:14 1993
 *		Replaced global tags with macros.
 *	AK	8	Tue Mar 16 14:32:44 1993
 *		Modified num_procs* to handle when procs is explicitly spec.
 *	AK	6	Tue Mar 16 14:32:00 1993
 *		Fixed the bug in ___post_assign_spread_pointer
 * 	TvE	3	Tue Jan 19 00:10:10 PST 1993
 *		PROCS_POWER_OF_TWO changed to ___PROCS_POWER_OF_TWO
 * 	TvE	2	Sat Dec 19 00:48:06 PST 1992
 *		Stuff taken from split-c_int.h
 */

#ifndef __POINTERS_H
#define __POINTERS_H

/* Mhh, do we really want to do this? -TvE */
#ifndef NULL
#  define NULL (0)
#endif

/* .............................
 *
 * Global pointer representation
 *
 *..............................
 */

/* This declaration is picked up by the split-c compiler and used as the
 * representation for global pointers. */

/*************************************************************
 * A long long representation for global pointers will not work
 * on native compilers. - ssl
 *************************************************************/

#if !defined(__GNUC__) || defined(UNIX_SM)
typedef struct {
	Processor pid;
	void *addr;
} ___globPtrType;
#else
typedef long long ___globPtrType;
#endif

typedef	struct {
	Processor	pid;
	void *		addr;
} ___splitPtrAsStruct;

typedef union {
	___splitPtrAsStruct	asStruct;
	___V_GPTR		asLong;
} ___splitPtrAsUnion;

#ifdef NO_SPLITCC
#define SET_PROC(x, y)  { (x).asStruct.pid = (Processor)(y); }
#define SET_ADDR(x, y)  { (x).asStruct.addr = (void *)(y); }
#define GET_PROC(x)     ( ((___splitPtrAsUnion*)(&x))->asStruct.pid )
#define GET_ADDR(x)     ( ((___splitPtrAsUnion*)(&x))->asStruct.addr )
#define ___V_GPTR_CAST
#else
#define SET_PROC(x, y)  { (x).asStruct.pid = (Processor)(y); }
#define SET_ADDR(x, y)  { (x).asStruct.addr = (void *)(y); }
#define GET_PROC(x)     ( ((___splitPtrAsUnion)(x)).asStruct.pid )
#define GET_ADDR(x)     ( ((___splitPtrAsUnion)(x)).asStruct.addr )
#define ___V_GPTR_CAST  (___V_GPTR)
#endif

/* .............................
 *
 * Constructing global addresses
 *
 * .............................
 */

/* Creates a global pointer which is proc:local */
EXTERN inline ___V_GPTR toglobal(Processor proc, void *local)
INLINE({
	___splitPtrAsUnion tempPtr;

	if(local == NULL) proc = 0;
	SET_PROC(tempPtr, proc);
	SET_ADDR(tempPtr, local);
	return tempPtr.asLong;
})

/* Creates a spread pointer which is proc:local */
EXTERN inline ___V_SPTR tospread(Processor proc, void *local)
INLINE({
	___splitPtrAsUnion tempPtr;

	if(local == NULL) proc = 0;
	SET_PROC(tempPtr, proc);
	SET_ADDR(tempPtr, local);
	return tempPtr.asLong;
})

/* Accessor function to get the processor part */
EXTERN inline Processor toproc(___V_GPTR globPtr)
INLINE({
	return GET_PROC(globPtr);
})

/* Accessor function to get the address part of a global pointer */
EXTERN inline void *tolocal(___V_GPTR globPtr)
INLINE({
	return GET_ADDR(globPtr);
})

/* .........................
 *
 * Casts to global addresses
 *
 * .........................
 */

/* Cast a pointer into a global pointer - fill in pid with MYPROC */
EXTERN inline ___V_GPTR ___global_cast(void *local)
INLINE({
	___splitPtrAsUnion tempPtr;
	SET_PROC(tempPtr, (local == NULL) ? 0 : MYPROC);
	SET_ADDR(tempPtr, local);
	return tempPtr.asLong;
})

/* Like toglobal but without the null pointer check */
EXTERN inline ___V_GPTR ___create_global_ptr(Processor proc, void *local)
INLINE({
	___splitPtrAsUnion tempPtr;
	SET_PROC(tempPtr, proc);
	SET_ADDR(tempPtr, local);
	return tempPtr.asLong;
})

/* Cast a pointer into a spread pointer - fill in pid with 0 */
/* was: ___cons_spread_ptr */
EXTERN inline ___V_SPTR ___spread_cast(void *local)
INLINE({
	___splitPtrAsUnion tempPtr;
	SET_PROC(tempPtr, 0);
	SET_ADDR(tempPtr, local);
	return tempPtr.asLong;
})

/* Like tospread but without the null pointer check */
EXTERN inline ___V_SPTR ___create_spread_ptr(Processor proc, void *local)
/* was: ___make_spread */
INLINE({
	___splitPtrAsUnion tempPtr;
	SET_PROC(tempPtr, proc);
	SET_ADDR(tempPtr, local);
	return tempPtr.asLong;
})

/* ..........................
 *
 * Global pointer comparisons
 *
 * ..........................
 */

EXTERN inline int ___null_ptr_check(___V_GPTR gPtr)
INLINE ({
	return GET_ADDR(gPtr) == 0;
})

EXTERN inline int ___not_null_ptr_check(___V_GPTR gPtr)
INLINE ({
	return GET_ADDR(gPtr) != 0;
})

EXTERN inline int ___eq_ptr_check(___V_GPTR gPtr1, ___V_GPTR gPtr2)
INLINE ({
	return GET_ADDR(gPtr1) == GET_ADDR(gPtr2)
	    && GET_PROC(gPtr1) == GET_PROC(gPtr2);
})

EXTERN inline int ___neq_ptr_check(___V_GPTR gPtr1, ___V_GPTR gPtr2)
INLINE ({
	return GET_ADDR(gPtr1) != GET_ADDR(gPtr2)
	    || GET_PROC(gPtr1) != GET_PROC(gPtr2);

})

/* .............................
 *
 * Arithmetic on global pointers
 *
 * .............................
 */

/* Index into global array */
#if 0
/* This turns out to be better than the next version.. -TvE */
EXTERN inline ___V_GPTR ___inc_addr_part(___V_GPTR gPtr, int index)
INLINE({
	___splitPtrAsUnion tempPtr = (___splitPtrAsUnion)gPtr;
	tempPtr.asStruct.addr += index;
	return tempPtr.asLong;
})
#else
EXTERN inline ___V_GPTR ___inc_addr_part(___V_GPTR gPtr, int index)
INLINE({
	___splitPtrAsUnion tempPtr;

	SET_PROC (tempPtr, toproc (gPtr));
	SET_ADDR (tempPtr, (void *)((int)GET_ADDR(gPtr) + index));
	return tempPtr.asLong;
})
#endif

/* Index into a spread array */
/* Had to break this into two procedures. Native compiler */
/* won't take ifdef's in macros. - ssl */
#ifdef ___PROCS_POWER_OF_TWO
EXTERN inline ___V_SPTR ___index_spread_ptr
	(___V_SPTR	sPtr,		/* base pointer */
	 int		index,		/* index into array */
	 int		size)		/* local element size */
INLINE({
	const int	proc = GET_PROC(___V_GPTR_CAST sPtr);
	void *const	local = GET_ADDR(___V_GPTR_CAST sPtr);
	int		newProc;
	void *		newLocal;
	___splitPtrAsUnion tempPtr;

	newProc  = (proc + index) & PROCSM1;
	newLocal = (void *)((int)local + ((proc + index)>>PROCSLOG)*size);

	SET_PROC(tempPtr, newProc);
	SET_ADDR(tempPtr, newLocal);
	return tempPtr.asLong;
})
#else
EXTERN inline ___V_SPTR ___index_spread_ptr
	(___V_SPTR	sPtr,		/* base pointer */
	 int		index,		/* index into array */
	 int		size)		/* local element size */
INLINE({
	const int	proc = GET_PROC(___V_GPTR_CAST sPtr);
	void *const	local = GET_ADDR(___V_GPTR_CAST sPtr);
	int		newProc;
	void *		newLocal;
	___splitPtrAsUnion tempPtr;

	newProc  = (proc + index) % PROCS;
	newLocal = local + ((proc + index)/PROCS)*size;

	SET_PROC(tempPtr, newProc);
	SET_ADDR(tempPtr, newLocal);
	return tempPtr.asLong;
})
#endif

/* Index into spread array, get processor part only */
/* Had to break this into two procedures. Native compiler */
/* won't take ifdef's in macros. - ssl */
#ifdef ___PROCS_POWER_OF_TWO
EXTERN inline int ___procs_of_index(___V_SPTR sPtr, int index)
INLINE({
	int proc = GET_PROC(___V_GPTR_CAST sPtr);

	return (proc + index) & PROCSM1;
})
#else
EXTERN inline int ___procs_of_index(___V_SPTR sPtr, int index)
INLINE({
	int proc = GET_PROC(___V_GPTR_CAST sPtr);

	return (proc + index) % PROCS;
})
#endif

/* Index into spread array formed by a cast from local ptr to a spread ptr,
 * get processor part only */
#ifdef ___PROCS_POWER_OF_TWO
EXTERN inline int ___procs_of_index0 (___V_SPTR sPtr, int index)
INLINE({
	return index & PROCSM1;
})
#else
EXTERN inline int ___procs_of_index0 (___V_SPTR sPtr, int index)
INLINE({
	return index % PROCS;
})
#endif

/* This needs some fixing...  -TvE */
EXTERN inline int ___num_chunks(___V_SPTR sPtr, int index)
INLINE({
	int proc = GET_PROC(___V_GPTR_CAST sPtr);
	return (proc + index);
})

/* This needs some fixing...  -TvE */
EXTERN inline int ___num_chunks0(___V_SPTR sPtr, int index)
INLINE({
	return index;
})

/* ..................................
 *
 * Post-arithmetic on global pointers
 *
 * ..................................
 */

EXTERN inline ___V_GPTR ___post_inc_glob_ptr
	(___V_GPTR *glob_ptr_ptr, int size)
INLINE({
	___V_GPTR const temp = *glob_ptr_ptr;
	SET_ADDR(*(___splitPtrAsUnion*)glob_ptr_ptr,
		(void *)((int)GET_ADDR(*glob_ptr_ptr) + size));
	return temp;
})

EXTERN inline ___V_GPTR ___post_dec_glob_ptr
	(___V_GPTR *glob_ptr_ptr, int size)
INLINE({
	___V_GPTR const temp = *glob_ptr_ptr;
	SET_ADDR(*(___splitPtrAsUnion*)glob_ptr_ptr,
		(void *)((int)GET_ADDR(*glob_ptr_ptr) - size));
	return temp;
})

EXTERN inline ___V_SPTR ___post_assign_spread_ptr
	(___V_SPTR *spread_ptr_ptr, ___V_SPTR spreadPtr)
INLINE({
        ___V_SPTR const temp = *spread_ptr_ptr;

	*spread_ptr_ptr = spreadPtr;
	return temp;
})

#endif __POINTERS_H
