/*                                                                      tab:8
 *
 * atomic.h - Atomic operations.
 *
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author:		Andrea Dusseau
 * Version:		9
 * Creation Date:	Thu Oct 22 15:49:07 PDT 1992
 * Filename:		atomic.sc
 * History:
 *	ssl	9	Wed Jan 19 11:41:46 1994
 *		rpc removed
 *	ssl	8	Tue Nov 16 11:49:43 1993
 *		Replaced global tags with macros.
 *	ssl	7	Thu Oct 28 21:57:35 1993
 *		Include am.h instead of cm5.h. 
 *	LTL	6	Mon Aug 16 15:16:22 1993
 *		Included "cm5.h" for version 112 to support CMMD
 *	SL	4	Tue Jan 19 13:59:21 1993
 *		Added prototypes for atomic sync. and rpc routines
 *  ACD 3   Tue Jan  5 13:01:24 PST 1993
 *      Made definitions of all atomic routines allow varargs
 * 	TvE	2	Mon Jan  4 22:09:23 PST 1993
 *		Sanitized a bit.
 *	AK	1	Sat Nov 28 21:16:27 1992
 *		Changed include file name from cmam-2-4
 */

#ifndef __ATOMIC_H__
#define	__ATOMIC_H__

/*
 * ATOMIC_INTERNAL_DEFINE is defined when compiling the actual code for the
 * library so that a normal version for all the inlined functions is produced.
 * It is off when any user code is being compiled.
 */
#undef EXTERN
#ifdef ATOMIC_INTERNAL_DEFINE
#      define EXTERN 
#else
#      define EXTERN extern
#endif

extern mutex_t Atomic_mutex;
extern mutex_t AtomicFunc_mutex;

/* ....................................
 *
 * Canned atomic operations on integers
 *
 * ....................................
 */
EXTERN inline int cmp_and_swap(___I_GPTR loc, int test, int new);
EXTERN inline int test_and_set(___I_GPTR loc);
EXTERN inline int exchange(___I_GPTR loc, int new_val);
EXTERN inline int fetch_and_add(___I_GPTR loc, int val);

/* .....................................................
 *
 * Building blocks for general atomic and rpc operations
 *
 * .....................................................
 */
#ifndef ATOMIC_INTERNAL_DEFINE
EXTERN void atomic_no_sync(void (*fun)(), ___V_GPTR gptr, ...);
EXTERN void atomic(void (*fun)(), ___V_GPTR gptr, ...);
EXTERN void atomic_long(void (*fun)(), ___V_GPTR gptr, ...);
EXTERN int atomic_i(void (*fun)(), ___V_GPTR gptr, ...);
EXTERN double atomic_d(void (*fun)(), ___V_GPTR gptr, ...);
#endif

/* ...................................................
 *
 * Synchronization primitives for atomic and rpc calls
 *
 * ...................................................
 */
EXTERN void all_atomic_sync();

#include "am.h"
/* Inline definitions to speed up things */
#include "atomic_inline.h" 

#endif /* __ATOMIC_H__ */







