/*									tab:4
 *
 * profile.sc - Split-C profiling routines and data
 *
 * "Copyright (c) 1993 by Steve Lumetta and The Regents of the University 
 * of California.  All rights reserved."
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 			Steve Lumetta
 * Version:			2
 * Creation Date:	Thu May 13 17:13:53 1993
 * Filename:		profile.sc
 * History:
 *	ssl	2	Sat Nov 20 20:42:21 1993
 *		Replaced barrier with am_barrier.
 *	SL	1	Thu May 13 17:14:04 1993
 *		First written
 */

#ifdef SC_PROFILE /* File is empty otherwise */

#include <stdio.h>
#include <split-c/split-c.h>
#include <split-c/control.h>
#include <split-c/com.h>
#include <split-c/profile.h>
#include <math.h>

int sc_prof_data[NUM_PROF_TYPES];
char *prof_filename = NULL;
static FILE *prof_file;
static double last_call;

void profile_init()
{
    int i;

    on_one {
        if (prof_filename != NULL) {
            prof_file = fopen(prof_filename, "w");
            if (prof_file == NULL) {
                fprintf(stderr,
                        "fopen of \"%s\" for use as profile file failed.",
                        prof_filename);
                prof_file = stdout;
            }
        }
        else
            prof_file = stdout;
    }
    prof_file = (FILE *)all_bcast_i((int)prof_file);
    for (i = 0; i < NUM_PROF_TYPES; i++)
        sc_prof_data[i] = 0;
	last_call = get_seconds();
}

void all_do_profile(char *comment)
{
    int i;
    char out_str[NUM_PROF_TYPES*12+10];
    double passed_time;
    
    barrier();
	passed_time = get_seconds() - last_call;
	sprintf(out_str, "%5d", MYPROC);
	for (i = 0; i < NUM_PROF_TYPES; i++)
		sprintf(out_str + i*12 + 5, "%12d", sc_prof_data[i]);
    on_one {
        fprintf(prof_file, "%s\n(%.6f seconds since last call)\n\n", comment,
				passed_time);
    }
	barrier();
	fprintf(prof_file, "%s\n", out_str);
	barrier();
	on_one {
        fprintf(prof_file, "\n\n", comment);
		fflush(prof_file);
	}
	barrier();  /* Give equal starting times */
    for (i = 0; i < NUM_PROF_TYPES; i++)
        sc_prof_data[i] = 0;
	last_call = get_seconds();
}

#endif


