/* Subroutines used for code generation on the DEC Alpha.
   Copyright (C) 1992, 1993, 1994 Free Software Foundation, Inc.
   Contributed by Richard Kenner (kenner@vlsi1.ultra.nyu.edu)

   Vector, Aux, and Dist additions and T3D support Copyright (C) 1993,1994 
      United States Government as represented by the Director, 
      National Security Agency.
      All rights reserved

This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */


#include <stdio.h>
#include "config.h"
#include "rtl.h"
#include "regs.h"
#include "hard-reg-set.h"
#include "real.h"
#include "insn-config.h"
#include "conditions.h"
#include "insn-flags.h"
#include "output.h"
#include "insn-attr.h"
#include "flags.h"
#include "recog.h"
#include "reload.h"
#include "expr.h"
#include "obstack.h"
#include "tree.h"

/* Save information from a "cmpxx" operation until the branch or scc is
   emitted.  */

rtx alpha_compare_op0, alpha_compare_op1;
int alpha_compare_fp_p;

/* Save the name of the current function as used by the assembler.  This
   is used by the epilogue.  */

char *alpha_function_name;

/* Non-zero if inside of a function, because the Alpha asm can't
   handle .files inside of functions.  */

static int inside_function = FALSE;

/* Nonzero if the current function needs gp.  */

int alpha_function_needs_gp;

extern char *version_string;

/* Declarations of static functions.  */
static void alpha_set_memflags_1  PROTO((rtx, int, int, int));
static void add_long_const	PROTO((FILE *, HOST_WIDE_INT, int, int, int));

/* Returns 1 if VALUE is a mask that contains full bytes of zero or ones.  */

int
zap_mask (value)
     HOST_WIDE_INT value;
{
  int i;

  for (i = 0; i < HOST_BITS_PER_WIDE_INT / HOST_BITS_PER_CHAR;
       i++, value >>= 8)
    if ((value & 0xff) != 0 && (value & 0xff) != 0xff)
      return 0;

  return 1;
}

/* Returns 1 if OP is either the constant zero or a register.  If a
   register, it must be in the proper mode unless MODE is VOIDmode.  */

int
reg_or_0_operand (op, mode)
      register rtx op;
      enum machine_mode mode;
{
  return op == const0_rtx || register_operand (op, mode);
}

/* Return 1 if OP is a constant in the range of 0-63 (for a shift) or
   any register.  */

int
reg_or_6bit_operand (op, mode)
     register rtx op;
     enum machine_mode mode;
{
  return ((GET_CODE (op) == CONST_INT
	   && (unsigned HOST_WIDE_INT) INTVAL (op) < 64)
	  || register_operand (op, mode));
}


/* Return 1 if OP is an 8-bit constant or any register.  */

int
reg_or_8bit_operand (op, mode)
     register rtx op;
     enum machine_mode mode;
{
  return ((GET_CODE (op) == CONST_INT
	   && (unsigned HOST_WIDE_INT) INTVAL (op) < 0x100)
	  || register_operand (op, mode));
}

/* Return 1 if OP is an 8-bit constant.  */

int
cint8_operand (op, mode)
     register rtx op;
     enum machine_mode mode;
{
  return (GET_CODE (op) == CONST_INT
	  && (unsigned HOST_WIDE_INT) INTVAL (op) < 0x100);
}

/* Return 1 if the operand is a valid second operand to an add insn.  */

int
add_operand (op, mode)
     register rtx op;
     enum machine_mode mode;
{
  if (GET_CODE (op) == CONST_INT)
    return ((unsigned HOST_WIDE_INT) (INTVAL (op) + 0x8000) < 0x10000
	    || ((INTVAL (op) & 0xffff) == 0
		&& (((INTVAL (op) >> 31) & 0xffffffff) == 0xffffffff
		    || INTVAL (op) >> 31 == 0)));

  return register_operand (op, mode);
}

/* Return 1 if the operand is a valid second operand to a sign-extending
   add insn.  */

int
sext_add_operand (op, mode)
     register rtx op;
     enum machine_mode mode;
{
  if (GET_CODE (op) == CONST_INT)
    return ((unsigned HOST_WIDE_INT) INTVAL (op) < 255
	    || (unsigned HOST_WIDE_INT) (- INTVAL (op)) < 255);

  return register_operand (op, mode);
}

/* Return 1 if OP is the constant 4 or 8.  */

int
const48_operand (op, mode)
     register rtx op;
     enum machine_mode mode;
{
  return (GET_CODE (op) == CONST_INT
	  && (INTVAL (op) == 4 || INTVAL (op) == 8));
}

/* Return 1 if OP is a valid first operand to an AND insn.  */

int
and_operand (op, mode)
     register rtx op;
     enum machine_mode mode;
{
  if (GET_CODE (op) == CONST_DOUBLE && GET_MODE (op) == VOIDmode)
    return (zap_mask (CONST_DOUBLE_LOW (op))
	    && zap_mask (CONST_DOUBLE_HIGH (op)));

  if (GET_CODE (op) == CONST_INT)
    return ((unsigned HOST_WIDE_INT) INTVAL (op) < 0x100
	    || (unsigned HOST_WIDE_INT) ~ INTVAL (op) < 0x100
	    || zap_mask (INTVAL (op)));

  return register_operand (op, mode);
}

/* Return 1 if OP is a valid first operand to an IOR or XOR insn.  */

int
or_operand (op, mode)
     register rtx op;
     enum machine_mode mode;
{
  if (GET_CODE (op) == CONST_INT)
    return ((unsigned HOST_WIDE_INT) INTVAL (op) < 0x100
	    || (unsigned HOST_WIDE_INT) ~ INTVAL (op) < 0x100);

  return register_operand (op, mode);
}

/* Return 1 if OP is a constant that is the width, in bits, of an integral
   mode smaller than DImode.  */

int
mode_width_operand (op, mode)
     register rtx op;
     enum machine_mode mode;
{
  return (GET_CODE (op) == CONST_INT
	  && (INTVAL (op) == 8 || INTVAL (op) == 16 || INTVAL (op) == 32));
}

/* Return 1 if OP is a constant that is the width of an integral machine mode
   smaller than an integer.  */

int
mode_mask_operand (op, mode)
     register rtx op;
     enum machine_mode mode;
{
#if HOST_BITS_PER_WIDE_INT == 32
  if (GET_CODE (op) == CONST_DOUBLE)
    return CONST_DOUBLE_HIGH (op) == 0 && CONST_DOUBLE_LOW (op) == -1;
#endif

  return (GET_CODE (op) == CONST_INT
	  && (INTVAL (op) == 0xff
	      || INTVAL (op) == 0xffff
#if HOST_BITS_PER_WIDE_INT == 64
	      || INTVAL (op) == 0xffffffff
#endif
	      ));
}

/* Return 1 if OP is a multiple of 8 less than 64.  */

int
mul8_operand (op, mode)
     register rtx op;
     enum machine_mode mode;
{
  return (GET_CODE (op) == CONST_INT
	  && (unsigned HOST_WIDE_INT) INTVAL (op) < 64
	  && (INTVAL (op) & 7) == 0);
}

/* Return 1 if OP is the constant zero in floating-point.  */

int
fp0_operand (op, mode)
     register rtx op;
     enum machine_mode mode;
{
  return (GET_MODE (op) == mode
	  && GET_MODE_CLASS (mode) == MODE_FLOAT && op == CONST0_RTX (mode));
}

/* Return 1 if OP is the floating-point constant zero or a register.  */

int
reg_or_fp0_operand (op, mode)
     register rtx op;
     enum machine_mode mode;
{
  return fp0_operand (op, mode) || register_operand (op, mode);
}

/* Return 1 if OP is a register or a constant integer.  */


int
reg_or_cint_operand (op, mode)
    register rtx op;
    enum machine_mode mode;
{
     return GET_CODE (op) == CONST_INT || register_operand (op, mode);
}

/* Return 1 if OP is something that can be reloaded into a register;
   if it is a MEM, it need not be valid.  */

int
some_operand (op, mode)
     register rtx op;
     enum machine_mode mode;
{
  if (mode != VOIDmode && GET_MODE (op) != VOIDmode && mode != GET_MODE (op))
    return 0;

  switch (GET_CODE (op))
    {
    case REG:  case MEM:  case CONST_DOUBLE:
    case CONST_INT:  case LABEL_REF:  case SYMBOL_REF:  case CONST:
      return 1;

    case SUBREG:
      return some_operand (SUBREG_REG (op), VOIDmode);
    }

  return 0;
}

/* Return 1 if OP is a valid operand for the source of a move insn.  */

int
input_operand (op, mode)
     register rtx op;
     enum machine_mode mode;
{
  if (mode != VOIDmode && GET_MODE (op) != VOIDmode && mode != GET_MODE (op))
    return 0;

  if (GET_MODE_CLASS (mode) == MODE_FLOAT && GET_MODE (op) != mode)
    return 0;

  switch (GET_CODE (op))
    {
#ifndef __T3D__
    case LABEL_REF:
    case SYMBOL_REF:
    case CONST:
      return mode == DImode;
#endif

    case REG:
      return 1;

    case SUBREG:
      if (register_operand (op, mode))
	return 1;
      /* ... fall through ... */
    case MEM:
      return mode != HImode && mode != QImode && general_operand (op, mode);

    case CONST_DOUBLE:
      return GET_MODE_CLASS (mode) == MODE_FLOAT && op == CONST0_RTX (mode);

    case CONST_INT:
      return mode == QImode || mode == HImode || add_operand (op, mode);
    }

  return 0;
}

/* Return 1 if OP is a SYMBOL_REF for a function known to be in this
   file.  */

int
current_file_function_operand (op, mode)
     rtx op;
     enum machine_mode mode;
{
  return (GET_CODE (op) == SYMBOL_REF
	  && (SYMBOL_REF_FLAG (op)
	      || op == XEXP (DECL_RTL (current_function_decl), 0)));
}

/* Return 1 if OP is a valid operand for the MEM of a CALL insn.  */

int
call_operand (op, mode)
     rtx op;
     enum machine_mode mode;
{
  if (mode != Pmode)
    return 0;

  return (GET_CODE (op) == SYMBOL_REF
	  || (GET_CODE (op) == REG && REGNO (op) == 27));
}

/* Return 1 if OP is a valid Alpha comparison operator.  Here we know which
   comparisons are valid in which insn.  */

int
alpha_comparison_operator (op, mode)
     register rtx op;
     enum machine_mode mode;
{
  enum rtx_code code = GET_CODE (op);

  if (mode != GET_MODE (op) || GET_RTX_CLASS (code) != '<')
    return 0;

  return (code == EQ || code == LE || code == LT
	  || (mode == DImode && (code == LEU || code == LTU)));
}

/* Return 1 if OP is a signed comparison operation.  */

int
signed_comparison_operator (op, mode)
     register rtx op;
     enum machine_mode mode;
{
  switch (GET_CODE (op))
    {
    case EQ:  case NE:  case LE:  case LT:  case GE:   case GT:
      return 1;
    }

  return 0;
}

/* Return 1 if this is a divide or modulus operator.  */

int
divmod_operator (op, mode)
     register rtx op;
     enum machine_mode mode;
{
  switch (GET_CODE (op))
    {
    case DIV:  case MOD:  case UDIV:  case UMOD:
      return 1;
    }

  return 0;
}

/* Return 1 if this memory address is a known aligned register plus
   a constant.  It must be a valid address.  This means that we can do
   this as an aligned reference plus some offset.

   Take into account what reload will do.

   We could say that out-of-range stack slots are alignable, but that would
   complicate get_aligned_mem and it isn't worth the trouble since few
   functions have large stack space.  */

int
aligned_memory_operand (op, mode)
     register rtx op;
     enum machine_mode mode;
{
  if (GET_CODE (op) == SUBREG)
    {
      if (GET_MODE (op) != mode)
	return 0;
      op = SUBREG_REG (op);
      mode = GET_MODE (op);
    }

  if (reload_in_progress && GET_CODE (op) == REG
      && REGNO (op) >= FIRST_PSEUDO_REGISTER)
    op = reg_equiv_mem[REGNO (op)];

  if (GET_CODE (op) != MEM || GET_MODE (op) != mode
      || ! memory_address_p (mode, XEXP (op, 0)))
    return 0;

  op = XEXP (op, 0);

  if (GET_CODE (op) == PLUS)
    op = XEXP (op, 0);

  return (GET_CODE (op) == REG
	  && (REGNO (op) == STACK_POINTER_REGNUM
	      || op == hard_frame_pointer_rtx
	      || (REGNO (op) >= FIRST_VIRTUAL_REGISTER
		  && REGNO (op) <= LAST_VIRTUAL_REGISTER)));
}

/* Similar, but return 1 if OP is a MEM which is not alignable.  */

int
unaligned_memory_operand (op, mode)
     register rtx op;
     enum machine_mode mode;
{
  if (GET_CODE (op) == SUBREG)
    {
      if (GET_MODE (op) != mode)
	return 0;
      op = SUBREG_REG (op);
      mode = GET_MODE (op);
    }

  if (reload_in_progress && GET_CODE (op) == REG
      && REGNO (op) >= FIRST_PSEUDO_REGISTER)
    op = reg_equiv_mem[REGNO (op)];

  if (GET_CODE (op) != MEM || GET_MODE (op) != mode)
    return 0;

  op = XEXP (op, 0);

  if (! memory_address_p (mode, op))
    return 1;

  if (GET_CODE (op) == PLUS)
    op = XEXP (op, 0);

  return (GET_CODE (op) != REG
	  || (REGNO (op) != STACK_POINTER_REGNUM
	      && op != hard_frame_pointer_rtx
	      && (REGNO (op) < FIRST_VIRTUAL_REGISTER
		  || REGNO (op) > LAST_VIRTUAL_REGISTER)));
}

/* Return 1 if OP is any memory location.  During reload a pseudo matches.  */

int
any_memory_operand (op, mode)
     register rtx op;
     enum machine_mode mode;
{
  return (GET_CODE (op) == MEM
	  || (GET_CODE (op) == SUBREG && GET_CODE (SUBREG_REG (op)) == REG)
	  || (reload_in_progress && GET_CODE (op) == REG
	      && REGNO (op) >= FIRST_PSEUDO_REGISTER)
	  || (reload_in_progress && GET_CODE (op) == SUBREG
	      && GET_CODE (SUBREG_REG (op)) == REG
	      && REGNO (SUBREG_REG (op)) >= FIRST_PSEUDO_REGISTER));
}

int
symbol_ref_operand (op, mode)
     register rtx op;
     enum machine_mode mode;
{
  return (GET_CODE (op) == SYMBOL_REF || GET_CODE (op) == LABEL_REF
	  || GET_CODE (op) == CONST);
}


/* REF is an alignable memory location.  Place an aligned SImode
   reference into *PALIGNED_MEM and the number of bits to shift into
   *PBITNUM.  */

void
get_aligned_mem (ref, paligned_mem, pbitnum)
     rtx ref;
     rtx *paligned_mem, *pbitnum;
{
  rtx base;
  HOST_WIDE_INT offset = 0;

  if (GET_CODE (ref) == SUBREG)
    {
      offset = SUBREG_WORD (ref) * UNITS_PER_WORD;
      if (BYTES_BIG_ENDIAN)
	offset -= (MIN (UNITS_PER_WORD, GET_MODE_SIZE (GET_MODE (ref)))
		   - MIN (UNITS_PER_WORD,
			  GET_MODE_SIZE (GET_MODE (SUBREG_REG (ref)))));
      ref = SUBREG_REG (ref);
    }

  if (GET_CODE (ref) == REG)
    ref = reg_equiv_mem[REGNO (ref)];

  if (reload_in_progress)
    base = find_replacement (&XEXP (ref, 0));
  else
    base = XEXP (ref, 0);

  if (GET_CODE (base) == PLUS)
    offset += INTVAL (XEXP (base, 1)), base = XEXP (base, 0);

  *paligned_mem = gen_rtx (MEM, SImode,
			   plus_constant (base, offset & ~3));
  MEM_IN_STRUCT_P (*paligned_mem) = MEM_IN_STRUCT_P (ref);
  MEM_VOLATILE_P (*paligned_mem) = MEM_VOLATILE_P (ref);
  RTX_UNCHANGING_P (*paligned_mem) = RTX_UNCHANGING_P (ref);

#ifdef __T3D__
  *pbitnum = GEN_INT (((~offset) & 3) * 8);
#else
  *pbitnum = GEN_INT ((offset & 3) * 8);
#endif
}

/* Similar, but just get the address.  Handle the two reload cases.  */

rtx
get_unaligned_address (ref)
     rtx ref;
{
  rtx base;
  HOST_WIDE_INT offset = 0;

  if (GET_CODE (ref) == SUBREG)
    {
      offset = SUBREG_WORD (ref) * UNITS_PER_WORD;
      if (BYTES_BIG_ENDIAN)
	offset -= (MIN (UNITS_PER_WORD, GET_MODE_SIZE (GET_MODE (ref)))
		   - MIN (UNITS_PER_WORD,
			  GET_MODE_SIZE (GET_MODE (SUBREG_REG (ref)))));
      ref = SUBREG_REG (ref);
    }

  if (GET_CODE (ref) == REG)
    ref = reg_equiv_mem[REGNO (ref)];

  if (reload_in_progress)
    base = find_replacement (&XEXP (ref, 0));
  else
    base = XEXP (ref, 0);

  if (GET_CODE (base) == PLUS)
    offset += INTVAL (XEXP (base, 1)), base = XEXP (base, 0);

  return plus_constant (base, offset);
}

/* Subfunction of the following function.  Update the flags of any MEM
   found in part of X.  */

static void
alpha_set_memflags_1 (x, in_struct_p, volatile_p, unchanging_p)
     rtx x;
     int in_struct_p, volatile_p, unchanging_p;
{
  int i;

  switch (GET_CODE (x))
    {
    case SEQUENCE:
    case PARALLEL:
      for (i = XVECLEN (x, 0) - 1; i >= 0; i--)
	alpha_set_memflags_1 (XVECEXP (x, 0, i), in_struct_p, volatile_p,
			      unchanging_p);
      break;

    case INSN:
      alpha_set_memflags_1 (PATTERN (x), in_struct_p, volatile_p,
			    unchanging_p);
      break;

    case SET:
      alpha_set_memflags_1 (SET_DEST (x), in_struct_p, volatile_p,
			    unchanging_p);
      alpha_set_memflags_1 (SET_SRC (x), in_struct_p, volatile_p,
			    unchanging_p);
      break;

    case MEM:
      MEM_IN_STRUCT_P (x) = in_struct_p;
      MEM_VOLATILE_P (x) = volatile_p;
      RTX_UNCHANGING_P (x) = unchanging_p;
      break;
    }
}

/* Given INSN, which is either an INSN or a SEQUENCE generated to
   perform a memory operation, look for any MEMs in either a SET_DEST or
   a SET_SRC and copy the in-struct, unchanging, and volatile flags from
   REF into each of the MEMs found.  If REF is not a MEM, don't do
   anything.  */

void
alpha_set_memflags (insn, ref)
     rtx insn;
     rtx ref;
{
  /* Note that it is always safe to get these flags, though they won't
     be what we think if REF is not a MEM.  */
  int in_struct_p = MEM_IN_STRUCT_P (ref);
  int volatile_p = MEM_VOLATILE_P (ref);
  int unchanging_p = RTX_UNCHANGING_P (ref);

  if (GET_CODE (ref) != MEM
      || (! in_struct_p && ! volatile_p && ! unchanging_p))
    return;

  alpha_set_memflags_1 (insn, in_struct_p, volatile_p, unchanging_p);
}

int
alpha_emit_set_symbol (target, src)
     rtx target,src;
{
#ifdef __T3D__
  rtx t, tt, ttt, t1, base, offset; 

  if (reload_in_progress)
    {
      t = tt = ttt = target;
    }
  else
    {
      t = gen_reg_rtx (DImode);
      tt = gen_reg_rtx (DImode);
      ttt = gen_reg_rtx (DImode);
    }

  if (GET_CODE(src) == CONST && GET_CODE (XEXP (src, 0)) == PLUS)
    {
      offset = XEXP( XEXP (src, 0), 1);
      base = XEXP( XEXP (src, 0), 0);
    }
  else
    {
      base = src;
      offset = 0;
    }

  emit_insn (gen_alpha_laum (t, const0_rtx, base));
  emit_insn (gen_ashldi3 (t, t, GEN_INT (32)));
  emit_insn (gen_alpha_lalm (t, t, base));

  if (offset)
    {
      t1 = emit_insn (gen_alpha_lal (tt, t, base));
      REG_NOTES (t1) = gen_rtx (EXPR_LIST, REG_EQUAL, base, REG_NOTES (t1));
      t1 = emit_insn (gen_adddi3 (ttt, tt, offset));
      REG_NOTES (t1) = gen_rtx (EXPR_LIST, REG_EQUAL, src, REG_NOTES (t1));
    }
  else
    {
      t1 = emit_insn (gen_alpha_lal (ttt, t, base));
      REG_NOTES (t1) = gen_rtx (EXPR_LIST, REG_EQUAL, src, REG_NOTES (t1));
    }

  if (!reload_in_progress)
    emit_move_insn (target, ttt);

  return 1;
#else
  return 0;
#endif
}


/* Try to output insns to set TARGET equal to the constant C if it can be
   done in less than N insns.  Returns 1 if it can be done and the
   insns have been emitted.  If it would take more than N insns, zero is
   returned and no insns and emitted.  */

int
alpha_emit_set_const (target, c, n)
     rtx target;
     HOST_WIDE_INT c;
     int n;
{
  HOST_WIDE_INT new = c;
  int i, bits;

#if HOST_BITS_PER_WIDE_INT == 64
  /* We are only called for SImode and DImode.  If this is SImode, ensure that
     we are sign extended to a full word.  This does not make any sense when
     cross-compiling on a narrow machine.  */

  if (GET_MODE (target) == SImode)
    c = (c & 0xffffffff) - 2 * (c & 0x80000000);
#endif

  /* If this is a sign-extended 32-bit constant, we can do this in at most
     three insns, so do it if we have enough insns left.  We always have
     a sign-extended 32-bit constant when compiling on a narrow machine.  */

  if (HOST_BITS_PER_WIDE_INT != 64
      || (c >> 31 & 0xffffffff) == 0xffffffff || c >> 31 == 0)
    {
      HOST_WIDE_INT low = (c & 0xffff) - 2 * (c & 0x8000);
      HOST_WIDE_INT tmp1 = c - low;
      HOST_WIDE_INT high
	= ((tmp1 >> 16) & 0xffff) - 2 * ((tmp1 >> 16) & 0x8000);
      HOST_WIDE_INT extra = 0;

      /* If HIGH will be interpreted as negative but the constant is
	 positive, we must adjust it to do two ldha insns.  */

      if ((high & 0x8000) != 0 && c>>31 == 0)
	{
	  extra = 0x4000;
	  tmp1 -= 0x40000000;
	  high = ((tmp1 >> 16) & 0xffff) - 2 * ((tmp1 >> 16) & 0x8000);
	}

      if (c == low || (low == 0 && extra == 0))
	{
	  emit_move_insn (target, GEN_INT (c));
	  return 1;
	}
      else if (n >= 2 + (extra != 0))
	{
	  emit_move_insn (target, GEN_INT (low));
	  if (extra != 0)
	    emit_insn (gen_add2_insn (target, GEN_INT (extra << 16)));

	  emit_insn (gen_add2_insn (target, GEN_INT (high << 16)));
	  return 1;
	}
    }

  /* If we couldn't do it that way, try some other methods (that depend on
     being able to compute in the target's word size).  But if we have no
     instructions left, don't bother.  Also, don't even try if this is 
     SImode (in which case we should have already done something, but
     do a sanity check here).  */

  if (n == 1 || HOST_BITS_PER_WIDE_INT < 64 || GET_MODE (target) != DImode)
    return 0;

  /* First, see if can load a value into the target that is the same as the
     constant except that all bytes that are 0 are changed to be 0xff.  If we
     can, then we can do a ZAPNOT to obtain the desired constant.  */

  for (i = 0; i < 64; i += 8)
    if ((new & ((HOST_WIDE_INT) 0xff << i)) == 0)
      new |= (HOST_WIDE_INT) 0xff << i;

  if (alpha_emit_set_const (target, new, n - 1))
    {
      emit_insn (gen_anddi3 (target, target, GEN_INT (c | ~ new)));
      return 1;
    }

  /* Find, see if we can load a related constant and then shift and possibly
     negate it to get the constant we want.  Try this once each increasing
     numbers of insns.  */

  for (i = 1; i < n; i++)
    {
      /* First try complementing.  */
      if (alpha_emit_set_const (target, ~ c, i))
	{
	  emit_insn (gen_one_cmpldi2 (target, target));
	  return 1;
	}

      /* First try to form a constant and do a left shift.  We can do this
	 if some low-order bits are zero; the exact_log2 call below tells
	 us that information.  The bits we are shifting out could be any
	 value, but here we'll just try the 0- and sign-extended forms of
	 the constant.  To try to increase the chance of having the same
	 constant in more than one insn, start at the highest number of
	 bits to shift, but try all possibilities in case a ZAPNOT will
	 be useful.  */

      if ((bits = exact_log2 (c & - c)) > 0)
	for (; bits > 0; bits--)
	  if (alpha_emit_set_const (target, c >> bits, i)
	      || alpha_emit_set_const (target,
				       ((unsigned HOST_WIDE_INT) c) >> bits,
				       i))
	    {
	      emit_insn (gen_ashldi3 (target, target, GEN_INT (bits)));
	      return 1;
	    }

      /* Now try high-order zero bits.  Here we try the shifted-in bits as
	 all zero and all ones.  */

      if ((bits = HOST_BITS_PER_WIDE_INT - floor_log2 (c) - 1) > 0)
	for (; bits > 0; bits--)
	  if (alpha_emit_set_const (target, c << bits, i)
	      || alpha_emit_set_const (target,
				       ((c << bits)
					| (((HOST_WIDE_INT) 1 << bits) - 1)),
				       i))
	    {
	      emit_insn (gen_lshrdi3 (target, target, GEN_INT (bits)));
	      return 1;
	    }

      /* Now try high-order 1 bits.  We get that with a sign-extension.
	 But one bit isn't enough here.  */
      
      if ((bits = HOST_BITS_PER_WIDE_INT - floor_log2 (~ c) - 2) > 0)
	for (; bits > 0; bits--)
	  if (alpha_emit_set_const (target, c << bits, i)
	      || alpha_emit_set_const (target,
				       ((c << bits)
					| (((HOST_WIDE_INT) 1 << bits) - 1)),
				       i))
	    {
	      emit_insn (gen_ashrdi3 (target, target, GEN_INT (bits)));
	      return 1;
	    }
    }

  /*
   * finally try to just load as if it were a symbol
   */
  if (n==3)
    {
      HOST_WIDE_INT c1,c2,c3,c4;
      rtx t, tt, t1;
      
      c1 = (c >> 48) & 0xffff;
      c2 = (c >> 32) & 0xffff;
      c3 = (c >> 16) & 0xffff;
      c4 = c & 0xffff;
      
      
      /* fix up carries */
      if (c4 & 0x8000)
	{
	  c4 |= ~c4 & ~0xffff;
	  c3++;
	  if (c3 > 0xffff)
	    {
	      c3 = 0;
	      c2++;
	      if (c2 > 0xffff)
		{
		  c2 = 0;
		  c1++;
		  if (c1 > 0xffff)
		    c1 = 0;
		}
	    }
	}
      if (c3 & 0x8000)
	{
	  c3 |= ~c3 & ~0xffff;
	  c2++;
	  if (c2 > 0xffff)
	    {
	      c2 = 0;
	      c1++;
	      if (c1 > 0xffff)
		c1 = 0;
	    }
	}
      if (c2 & 0x8000)
	{
	  c2 |= ~c2 & ~0xffff;
	  c1++;
	  if (c1 > 0xffff)
	    c1 = 0;
	}
      if (c1 & 0x8000)
	{
	  c1 |= ~c1 & ~0xffff;
	}
      
      c1 = c1 << 16;
      c3 = c3 << 16;
      
      if (reload_in_progress)
	{
	  t = tt = target;
	}
      else
	{
	  t = gen_reg_rtx (DImode);
	  tt = gen_reg_rtx (DImode);
	}

      emit_insn (gen_adddi3(t, const0_rtx, GEN_INT (c1)));
      emit_insn (gen_adddi3(t, t, GEN_INT (c2)));
      emit_insn (gen_ashldi3 (t, t, GEN_INT (32)));
      emit_insn (gen_adddi3(t, t, GEN_INT (c3)));
      t1 = emit_insn (gen_adddi3(tt, t, GEN_INT (c4)));
      REG_NOTES (t1) = gen_rtx (EXPR_LIST, REG_EQUAL, GEN_INT (c), REG_NOTES (t1));

      if (!reload_in_progress)
	emit_move_insn (target, tt);

      return 1;
    }
  return 0;
}

/* Adjust the cost of a scheduling dependency.  Return the new cost of
   a dependency LINK or INSN on DEP_INSN.  COST is the current cost.  */

int
alpha_adjust_cost (insn, link, dep_insn, cost)
     rtx insn;
     rtx link;
     rtx dep_insn;
     int cost;
{
  rtx set;

  /* If the dependence is an anti-dependence, there is no cost.  For an
     output dependence, there is sometimes a cost, but it doesn't seem
     worth handling those few cases.  */

  if (REG_NOTE_KIND (link) != 0)
    return 0;

  /* If INSN is a store insn and DEP_INSN is setting the data being stored,
     we can sometimes lower the cost.  */

  if (recog_memoized (insn) >= 0
      && (get_attr_type (insn) == TYPE_STI || get_attr_type (insn) == TYPE_STR)
      && (set = single_set (dep_insn)) != 0
      && GET_CODE (PATTERN (insn)) == SET
      && rtx_equal_p (SET_DEST (set), SET_SRC (PATTERN (insn))))
    switch (get_attr_type (dep_insn))
      {
      case TYPE_LDI:
      case TYPE_LDR:
	/* No savings here.  */
	return cost;

      case TYPE_IMULL:
      case TYPE_IMULQ:
	/* In these cases, we save one cycle.  */
	return cost - 2;

      default:
	/* In all other cases, we save two cycles.  */
	return MAX (0, cost - 4);
      }

  /* Another case that needs adjustment is an arithmetic or logical
     operation.  It's cost is usually one cycle, but we default it to
     two in the MD file.  The only case that it is actually two is
     for the address in loads and stores.  */

  if (recog_memoized (dep_insn) >= 0
      && get_attr_type (dep_insn) == TYPE_IADDLOG)
    switch (get_attr_type (insn))
      {
      case TYPE_LDI:
      case TYPE_LDR:
      case TYPE_STI:
      case TYPE_STR:
	return cost;

      default:
	return 2;
      }

  /* The final case is when a compare feeds into an integer branch.  The cost
     is only one cycle in that case.  */

  if (recog_memoized (dep_insn) >= 0
      && get_attr_type (dep_insn) == TYPE_ICMP
      && recog_memoized (insn) >= 0
      && get_attr_type (insn) == TYPE_IBR)
    return 2;

  /* Otherwise, return the default cost. */

  return cost;
}

/* Print an operand.  Recognize special options, documented below.  */

void
print_operand (file, x, code)
    FILE *file;
    rtx x;
    char code;
{
  int i;

  switch (code)
    {
    case 'r':
      /* If this operand is the constant zero, write it as "$31".  */
      if (GET_CODE (x) == REG)
	fprintf (file, "%s", reg_names[REGNO (x)]);
      else if (x == CONST0_RTX (GET_MODE (x)))
#ifdef __T3D__
	fprintf (file, "r31");
#else
	fprintf (file, "$31");
#endif
      else
	output_operand_lossage ("invalid %%r value");

      break;

    case 'R':
      /* Similar, but for floating-point.  */
      if (GET_CODE (x) == REG)
	fprintf (file, "%s", reg_names[REGNO (x)]);
      else if (x == CONST0_RTX (GET_MODE (x)))
#ifdef __T3D__
	fprintf (file, "f31");
#else
	fprintf (file, "$f31");
#endif
      else
	output_operand_lossage ("invalid %%R value");

      break;

    case 'N':
      /* Write the 1's complement of a constant.  */
      if (GET_CODE (x) != CONST_INT)
	output_operand_lossage ("invalid %%N value");

      fprintf (file, "%ld", ~ INTVAL (x));
      break;

    case 'P':
      /* Write 1 << C, for a constant C.  */
      if (GET_CODE (x) != CONST_INT)
	output_operand_lossage ("invalid %%P value");

      fprintf (file, "%ld", (HOST_WIDE_INT) 1 << INTVAL (x));
      break;

    case 'h':
      /* Write the high-order 16 bits of a constant, sign-extended.  */
      if (GET_CODE (x) != CONST_INT)
	output_operand_lossage ("invalid %%h value");

      if (INTVAL(x) < 0)
	{
	  int t = -INTVAL(x);
	  t = t >> 16;
	  fprintf (file, "%ld", -t);
	}
      else
	fprintf (file, "%ld", INTVAL (x) >> 16);
      break;

    case 'L':
      /* Write the low-order 16 bits of a constant, sign-extended.  */
      if (GET_CODE (x) != CONST_INT)
	output_operand_lossage ("invalid %%L value");

      fprintf (file, "%ld", (INTVAL (x) & 0xffff) - 2 * (INTVAL (x) & 0x8000));
      break;

    case 'm':
      /* Write mask for ZAP insn.  */
      if (GET_CODE (x) == CONST_DOUBLE)
	{
	  HOST_WIDE_INT mask = 0;
	  HOST_WIDE_INT value;

	  value = CONST_DOUBLE_LOW (x);
	  for (i = 0; i < HOST_BITS_PER_WIDE_INT / HOST_BITS_PER_CHAR;
	       i++, value >>= 8)
	    if (value & 0xff)
	      mask |= (1 << i);

	  value = CONST_DOUBLE_HIGH (x);
	  for (i = 0; i < HOST_BITS_PER_WIDE_INT / HOST_BITS_PER_CHAR;
	       i++, value >>= 8)
	    if (value & 0xff)
	      mask |= (1 << (i + sizeof (int)));

	  fprintf (file, "%ld", mask & 0xff);
	}

      else if (GET_CODE (x) == CONST_INT)
	{
	  HOST_WIDE_INT mask = 0, value = INTVAL (x);

	  for (i = 0; i < 8; i++, value >>= 8)
	    if (value & 0xff)
	      mask |= (1 << i);

	  fprintf (file, "%ld", mask);
	}
      else
	output_operand_lossage ("invalid %%m value");
      break;

    case 'M':
      /* 'b', 'w', or 'l' as the value of the constant.  */
      if (GET_CODE (x) != CONST_INT
	  || (INTVAL (x) != 8 && INTVAL (x) != 16 && INTVAL (x) != 32))
	output_operand_lossage ("invalid %%M value");

      fprintf (file, "%s",
	       INTVAL (x) == 8 ? "b" : INTVAL (x) == 16 ? "w" : "l");
      break;

    case 'U':
      /* Similar, except do it from the mask.  */
      if (GET_CODE (x) == CONST_INT && INTVAL (x) == 0xff)
	fprintf (file, "b");
      else if (GET_CODE (x) == CONST_INT && INTVAL (x) == 0xffff)
	fprintf (file, "w");
#if HOST_BITS_PER_WIDE_INT == 32
      else if (GET_CODE (x) == CONST_DOUBLE
	       && CONST_DOUBLE_HIGH (x) == 0
	       && CONST_DOUBLE_LOW (x) == -1)
	fprintf (file, "l");
#else
      else if (GET_CODE (x) == CONST_INT && INTVAL (x) == 0xffffffff)
	fprintf (file, "l");
#endif
      else
	output_operand_lossage ("invalid %%U value");
      break;

    case 's':
      /* Write the constant value divided by 8.  */
      if (GET_CODE (x) != CONST_INT
	  && (unsigned HOST_WIDE_INT) INTVAL (x) >= 64
	  && (INTVAL (x) & 7) != 8)
	output_operand_lossage ("invalid %%s value");

      fprintf (file, "%ld", INTVAL (x) / 8);
      break;

    case 'S':
      /* Same, except compute (64 - c) / 8 */

      if (GET_CODE (x) != CONST_INT
	  && (unsigned HOST_WIDE_INT) INTVAL (x) >= 64
	  && (INTVAL (x) & 7) != 8)
	output_operand_lossage ("invalid %%s value");

      fprintf (file, "%ld", (64 - INTVAL (x)) / 8);
      break;

    case 'C':
      /* Write out comparison name.  */
      if (GET_RTX_CLASS (GET_CODE (x)) != '<')
	output_operand_lossage ("invalid %%C value");

      if (GET_CODE (x) == LEU)
	fprintf (file, "ule");
      else if (GET_CODE (x) == LTU)
	fprintf (file, "ult");
      else
	fprintf (file, "%s", GET_RTX_NAME (GET_CODE (x)));
      break;

    case 'D':
      /* Similar, but write reversed code.  We can't get an unsigned code
	 here.  */
      if (GET_RTX_CLASS (GET_CODE (x)) != '<')
	output_operand_lossage ("invalid %%D value");

      fprintf (file, "%s", GET_RTX_NAME (reverse_condition (GET_CODE (x))));
      break;

    case 'E':
      /* Write the divide or modulus operator.  */
      switch (GET_CODE (x))
	{
#ifndef __T3D__
	case DIV:
	  fprintf (file, "div%s", GET_MODE (x) == SImode ? "l" : "q");
	  break;
	case UDIV:
	  fprintf (file, "div%su", GET_MODE (x) == SImode ? "l" : "q");
	  break;
	case MOD:
	  fprintf (file, "rem%s", GET_MODE (x) == SImode ? "l" : "q");
	  break;
	case UMOD:
	  fprintf (file, "rem%su", GET_MODE (x) == SImode ? "l" : "q");
	  break;
#endif
	default:
	  output_operand_lossage ("invalid %%E value");
	  break;
	}
      break;

    case 'A':
      /* Write "_u" for unaligned access.  */
      if (GET_CODE (x) == MEM && GET_CODE (XEXP (x, 0)) == AND)
	fprintf (file, "_u");
      break;

    case 0:
      if (GET_CODE (x) == REG)
	fprintf (file, "%s", reg_names[REGNO (x)]);
      else if (GET_CODE (x) == MEM)
	output_address (XEXP (x, 0));
      else
	output_addr_const (file, x);
      break;

    default:
      output_operand_lossage ("invalid %%xn code");
    }
}

/* Do what is necessary for `va_start'.  The argument is ignored;
   We look at the current function to determine if stdarg or varargs
   is used and fill in an initial va_list.  A pointer to this constructor
   is returned.  */

struct rtx_def *
alpha_builtin_saveregs (arglist)
     tree arglist;
{
  rtx block, addr, argsize;
  tree fntype = TREE_TYPE (current_function_decl);
  int stdarg = (TYPE_ARG_TYPES (fntype) != 0
		&& (TREE_VALUE (tree_last (TYPE_ARG_TYPES (fntype)))
		    != void_type_node));

  /* Compute the current position into the args, taking into account
     both registers and memory.  Both of these are already included in
     current_function_args_info.  */

#ifdef __T3D__
  argsize = GEN_INT (current_function_args_info.nargs * UNITS_PER_WORD);
#else
  argsize = GEN_INT (current_function_args_info * UNITS_PER_WORD);
#endif

  /* SETUP_INCOMING_VARARGS moves the starting address base up by 48,
     storing fp arg registers in the first 48 bytes, and the integer arg
     registers in the next 48 bytes.  This is only done, however, if any
     integer registers need to be stored.

     If no integer registers need be stored, then we must subtract 48 in
     order to account for the integer arg registers which are counted in
     argsize above, but which are not actually stored on the stack.  */

#ifdef __T3D__
  addr = virtual_incoming_args_rtx;
#else
  addr = (current_function_args_info <= 6
	  ? plus_constant (virtual_incoming_args_rtx, 6 * UNITS_PER_WORD)
	  : plus_constant (virtual_incoming_args_rtx, - (6 * UNITS_PER_WORD)));
#endif

  /* Allocate the va_list constructor */
  block = assign_stack_local (BLKmode, 2 * UNITS_PER_WORD, BITS_PER_WORD);
  RTX_UNCHANGING_P (block) = 1;
  RTX_UNCHANGING_P (XEXP (block, 0)) = 1;

  /* Store the address of the first integer register in the
     __va_base member.  */
  emit_move_insn (change_address (block, Pmode, XEXP (block, 0)),
		  force_operand (addr, NULL_RTX));

  /* Store the argsize as the __va_offset member.  */
  emit_move_insn (change_address (block, Pmode,
				  plus_constant (XEXP (block, 0),
						 UNITS_PER_WORD)),
		  force_operand (argsize, NULL_RTX));

  /* Return the address of the va_list constructor, but don't put it in a
     register.  Doing so would fail when not optimizing and produce worse
     code when optimizing.  */
  return XEXP (block, 0);
}

/* This page contains routines that are used to determine what the function
   prologue and epilogue code will do and write them out.  */

/* Compute the size of the save area in the stack.  */

int
alpha_sa_size ()
{
  int size = 0;
  int i;

#ifdef __T3D__
  size = 5+6+8;          /* ra, ssib, ci, shtop, fp + saved r + sav f*/
#else
  for (i = 0; i < FIRST_PSEUDO_REGISTER; i++)
    if (! fixed_regs[i] && ! call_used_regs[i] && regs_ever_live[i])
      size++;

  /* If some registers were saved but not reg 26, reg 26 must also
     be saved, so leave space for it.  */
  if (size != 0 && ! regs_ever_live[26])
    size++;
#endif

  /* Our size must be even (multiple of 16 bytes).  */
  if (size & 1)
    size ++;

  return size * 8;
}

/* Return 1 if this function can directly return via $26.  */

int
direct_return ()
{
  return (reload_completed && alpha_sa_size () == 0
	  && get_frame_size () == 0
	  && current_function_pretend_args_size == 0);
}

/* Write a version stamp.  Don't write anything if we are running as a
   cross-compiler.  Otherwise, use the versions in /usr/include/stamp.h.  */

#ifndef __T3D__
#ifndef CROSS_COMPILE
#include <stamp.h>
#endif
#endif

void
alpha_write_verstamp (file)
     FILE *file;
{
#ifdef MS_STAMP
  char *p;

  fprintf (file, "\t.verstamp %d %d ", MS_STAMP, LS_STAMP);
  for (p = version_string; *p != ' ' && *p != 0; p++)
    fprintf (file, "%c", *p == '.' ? ' ' : *p);
  fprintf (file, "\n");
#endif
}

/* Write code to add constant C to register number IN_REG (possibly 31)
   and put the result into OUT_REG.  Use TEMP_REG as a scratch register;
   usually this will be OUT_REG, but should not be if OUT_REG is 
   STACK_POINTER_REGNUM, since it must be updated in a single instruction.
   Write the code to FILE.  */

static void
add_long_const (file, c, in_reg, out_reg, temp_reg)
     FILE *file;
     HOST_WIDE_INT c;
     int in_reg, out_reg, temp_reg;
{
  HOST_WIDE_INT low = (c & 0xffff) - 2 * (c & 0x8000);
  HOST_WIDE_INT tmp1 = c - low;
  HOST_WIDE_INT high = ((tmp1 >> 16) & 0xffff) - 2 * ((tmp1 >> 16) & 0x8000);
  HOST_WIDE_INT extra = 0;

  /* We don't have code to write out constants larger than 32 bits.  */
#if HOST_BITS_PER_LONG_INT == 64
  if ((unsigned HOST_WIDE_INT) c >> 32 != 0)
    abort ();
#endif

  /* If HIGH will be interpreted as negative, we must adjust it to do two
     ldha insns.  Note that we will never be building a negative constant
     here.  */

  if (high & 0x8000)
    {
      extra = 0x4000;
      tmp1 -= 0x40000000;
      high = ((tmp1 >> 16) & 0xffff) - 2 * ((tmp1 >> 16) & 0x8000);
    }

  if (low != 0)
    {
      int result_reg = (extra == 0 && high == 0) ? out_reg : temp_reg;

      if (low >= 0 && low < 255)
#ifdef __T3D__
	fprintf (file, "\taddq r%d,%d,r%d\n", in_reg, low, result_reg);
#else
	fprintf (file, "\taddq $%d,%d,$%d\n", in_reg, low, result_reg);
#endif
      else
#ifdef __T3D__
	fprintf (file, "\tlda r%d,%d(r%d)\n", result_reg, low, in_reg);
#else
	fprintf (file, "\tlda $%d,%d($%d)\n", result_reg, low, in_reg);
#endif

      in_reg = result_reg;
    }

  if (extra)
    {
#ifdef __T3D__
      fprintf (file, "\tldah r%d,%d(r%d)\n", out_reg, extra, in_reg);
#else
      fprintf (file, "\tldah $%d,%d($%d)\n", out_reg, extra, in_reg);
#endif
      in_reg = out_reg;
    }

  if (high)
#ifdef __T3D__
    fprintf (file, "\tldah r%d,%d(r%d)\n", out_reg, high, in_reg);
#else
    fprintf (file, "\tldah $%d,%d($%d)\n", out_reg, high, in_reg);
#endif
}

/* Write function prologue.  */

#ifndef __T3D__
void
output_prolog (file, size)
     FILE *file;
     int size;
{
  HOST_WIDE_INT out_args_size
    = ALPHA_ROUND (current_function_outgoing_args_size);
  HOST_WIDE_INT sa_size = alpha_sa_size ();
  HOST_WIDE_INT frame_size
    = (out_args_size + sa_size
       + ALPHA_ROUND (size + current_function_pretend_args_size));
  HOST_WIDE_INT reg_offset = out_args_size;
  HOST_WIDE_INT start_reg_offset = reg_offset;
  HOST_WIDE_INT actual_start_reg_offset = start_reg_offset;
  int int_reg_save_area_size = 0;
  rtx insn;
  unsigned reg_mask = 0;
  int i;

  /* Ecoff can handle multiple .file directives, so put out file and lineno.
     We have to do that before the .ent directive as we cannot switch
     files within procedures with native ecoff because line numbers are
     linked to procedure descriptors.
     Outputting the lineno helps debugging of one line functions as they
     would otherwise get no line number at all. Please note that we would
     like to put out last_linenum from final.c, but it is not accesible.  */

  if (write_symbols == SDB_DEBUG)
    {
      ASM_OUTPUT_SOURCE_FILENAME (file,
				  DECL_SOURCE_FILE (current_function_decl));
      if (debug_info_level != DINFO_LEVEL_TERSE)
        ASM_OUTPUT_SOURCE_LINE (file,
				DECL_SOURCE_LINE (current_function_decl));
    }

  /* The assembly language programmer's guide states that the second argument
     to the .ent directive, the lex_level, is ignored by the assembler,
     so we might as well omit it.  */
     
  fprintf (file, "\t.ent ");
  assemble_name (file, alpha_function_name);
  fprintf (file, "\n");
  ASM_OUTPUT_LABEL (file, alpha_function_name);
  inside_function = TRUE;

  /* Set up offsets to alpha virtual arg/local debugging pointer.  */

  alpha_auto_offset = -frame_size + current_function_pretend_args_size;
  alpha_arg_offset = -frame_size + 48;

  /* If we need a GP (we have a LDSYM insn or a CALL_INSN), load it first. 
     Even if we are a static function, we still need to do this in case
     our address is taken and passed to something like qsort.  */

  alpha_function_needs_gp = 0;
  for (insn = get_insns (); insn; insn = NEXT_INSN (insn))
    if ((GET_CODE (insn) == CALL_INSN)
	|| (GET_RTX_CLASS (GET_CODE (insn)) == 'i'
	    && GET_CODE (PATTERN (insn)) != USE
	    && GET_CODE (PATTERN (insn)) != CLOBBER
	    && (get_attr_type (insn) == TYPE_LDSYM
		|| get_attr_type (insn) == TYPE_ISUBR)))
      {
	alpha_function_needs_gp = 1;
	break;
      }

  if (alpha_function_needs_gp)
    fprintf (file, "\tldgp $29,0($27)\n");

  /* Put a label after the GP load so we can enter the function at it.  */
  assemble_name (file, alpha_function_name);
  fprintf (file, "..ng:\n");

  /* Adjust the stack by the frame size.  If the frame size is > 4096
     bytes, we need to be sure we probe somewhere in the first and last
     4096 bytes (we can probably get away without the latter test) and
     every 8192 bytes in between.  If the frame size is > 32768, we
     do this in a loop.  Otherwise, we generate the explicit probe
     instructions. 

     Note that we are only allowed to adjust sp once in the prologue.  */

  if (frame_size < 32768)
    {
      if (frame_size > 4096)
	{
	  int probed = 4096;
	  int regnum = 2;	/* $1 is static chain, so start with $2.  */

	  fprintf (file, "\tldq $%d,-%d($30)\n", regnum++, probed);

	  while (probed + 8192 < frame_size)
	    fprintf (file, "\tldq $%d,-%d($30)\n", regnum++, probed += 8192);

	  /* We only have to do this probe if we aren't saving registers.  */
	  if (sa_size == 0 && probed + 4096 < frame_size)
	    fprintf (file, "\tldq $%d,-%d($30)\n", regnum++, probed += 4096);

	  if (regnum > 9)
	    abort ();
	}

      if (frame_size != 0)
	fprintf (file, "\tlda $30,-%d($30)\n", frame_size);
    }
  else
    {
      /* Here we generate code to set R4 to SP + 4096 and set R5 to the
	 number of 8192 byte blocks to probe.  We then probe each block
	 in the loop and then set SP to the proper location.  If the
	 amount remaining is > 4096, we have to do one more probe if we
	 are not saving any registers.  */

      HOST_WIDE_INT blocks = (frame_size + 4096) / 8192;
      HOST_WIDE_INT leftover = frame_size + 4096 - blocks * 8192;

      add_long_const (file, blocks, 31, 5, 5);

      fprintf (file, "\tlda $4,4096($30)\n");

      assemble_name (file, alpha_function_name);
      fprintf (file, "..sc:\n");

      fprintf (file, "\tldq $6,-8192($4)\n");
      fprintf (file, "\tsubq $5,1,$5\n");
      fprintf (file, "\tlda $4,-8192($4)\n");

      fprintf (file, "\tbne $5,");
      assemble_name (file, alpha_function_name);
      fprintf (file, "..sc\n");

      fprintf (file, "\tlda $30,-%d($4)\n", leftover);

      if (leftover > 4096 && sa_size == 0)
	fprintf (file, "\tldq $2,%d($30)\n", leftover - 4096);
    }

  /* Describe our frame.  */
  fprintf (file, "\t.frame $%d,%d,$26,%d\n", 
	   (frame_pointer_needed
	    ? HARD_FRAME_POINTER_REGNUM : STACK_POINTER_REGNUM),
	   frame_size, current_function_pretend_args_size);
    
  /* Save register 26 if any other register needs to be saved.  */
  if (sa_size != 0)
    {
      reg_mask |= 1 << 26;
      fprintf (file, "\tstq $26,%d($30)\n", reg_offset);
      reg_offset += 8;
      int_reg_save_area_size += 8;
    }

  /* Now save any other used integer registers required to be saved.  */
  for (i = 0; i < 32; i++)
    if (! fixed_regs[i] && ! call_used_regs[i] && regs_ever_live[i] && i != 26)
      {
	reg_mask |= 1 << i;
	fprintf (file, "\tstq $%d,%d($30)\n", i, reg_offset);
	reg_offset += 8;
	int_reg_save_area_size += 8;
      }

  /* Print the register mask and do floating-point saves.  */
  if (reg_mask)
    fprintf (file, "\t.mask 0x%x,%d\n", reg_mask,
	     actual_start_reg_offset - frame_size);

  start_reg_offset = reg_offset;
  reg_mask = 0;

  for (i = 0; i < 32; i++)
    if (! fixed_regs[i + 32] && ! call_used_regs[i + 32]
	&& regs_ever_live[i + 32])
      {
	reg_mask |= 1 << i;
	fprintf (file, "\tstt $f%d,%d($30)\n", i, reg_offset);
	reg_offset += 8;
      }

  /* Print the floating-point mask, if we've saved any fp register.  */
  if (reg_mask)
    fprintf (file, "\t.fmask 0x%x,%d\n", reg_mask,
	     actual_start_reg_offset - frame_size + int_reg_save_area_size);

  /* If we need a frame pointer, set it from the stack pointer.  Note that
     this must always be the last instruction in the prologue.  */
  if (frame_pointer_needed)
    fprintf (file, "\tbis $30,$30,$15\n");

  /* End the prologue and say if we used gp.  */
  fprintf (file, "\t.prologue %d\n", alpha_function_needs_gp);
}

/* Write function epilogue.  */

void
output_epilog (file, size)
     FILE *file;
     int size;
{
  rtx insn = get_last_insn ();
  HOST_WIDE_INT out_args_size
    = ALPHA_ROUND (current_function_outgoing_args_size);
  HOST_WIDE_INT sa_size = alpha_sa_size ();
  HOST_WIDE_INT frame_size
    = (out_args_size + sa_size
       + ALPHA_ROUND (size + current_function_pretend_args_size));
  HOST_WIDE_INT reg_offset = out_args_size;
  HOST_WIDE_INT frame_size_from_reg_save = frame_size - reg_offset;
  int restore_fp
    = frame_pointer_needed && regs_ever_live[HARD_FRAME_POINTER_REGNUM];
  int i;

  /* If the last insn was a BARRIER, we don't have to write anything except
     the .end pseudo-op.  */
  if (GET_CODE (insn) == NOTE)
    insn = prev_nonnote_insn (insn);
  if (insn == 0 || GET_CODE (insn) != BARRIER)
    {
      int fp_offset;

      /* If we have a frame pointer, restore SP from it.  */
      if (frame_pointer_needed)
	fprintf (file, "\tbis $15,$15,$30\n");

      /* Restore all the registers, starting with the return address
	 register.  */
      if (sa_size != 0)
	{
	  fprintf (file, "\tldq $26,%d($30)\n", reg_offset);
	  reg_offset += 8;
	}

      /* Now restore any other used integer registers that that we saved,
	 except for FP if it is being used as FP, since it must be
	 restored last.  */

      for (i = 0; i < 32; i++)
	if (! fixed_regs[i] && ! call_used_regs[i] && regs_ever_live[i]
	    && i != 26)
	  {
	    if (i == HARD_FRAME_POINTER_REGNUM && frame_pointer_needed)
	      fp_offset = reg_offset;
	    else
	      fprintf (file, "\tldq $%d,%d($30)\n", i, reg_offset);
	    reg_offset += 8;
	  }

      for (i = 0; i < 32; i++)
	if (! fixed_regs[i + 32] && ! call_used_regs[i + 32]
	    && regs_ever_live[i + 32])
	  {
	    fprintf (file, "\tldt $f%d,%d($30)\n", i, reg_offset);
	    reg_offset += 8;
	  }

      /* If the stack size is large and we have a frame pointer, compute the
	 size of the stack into a register because the old FP restore, stack
	 pointer adjust, and return are required to be consecutive
	 instructions.   */
      if (frame_size > 32767 && restore_fp)
	add_long_const (file, frame_size, 31, 1, 1);

      /* If we needed a frame pointer and we have to restore it, do it
	 now.  This must be done in one instruction immediately
	 before the SP update.  */
      if (restore_fp)
	fprintf (file, "\tldq $15,%d($30)\n", fp_offset);

      /* Now update the stack pointer, if needed.  Only one instruction must
	 modify the stack pointer.  It must be the last instruction in the
	 sequence and must be an ADDQ or LDA instruction.  If the frame
	 pointer was loaded above, we may only put one instruction here.  */

      if (frame_size > 32768 && restore_fp)
	fprintf  (file, "\taddq $1,$30,$30\n");
      else
	add_long_const (file, frame_size, 30, 30, 1);

      /* Finally return to the caller.  */
      fprintf (file, "\tret $31,($26),1\n");
    }

  /* End the function.  */
  fprintf (file, "\t.end ");
  assemble_name (file, alpha_function_name);
  fprintf (file, "\n");
  inside_function = FALSE;
}
#endif


/* Debugging support.  */

#include "gstab.h"

/* Count the number of sdb related labels are generated (to find block
   start and end boundaries).  */

int sdb_label_count = 0;

/* Next label # for each statement.  */

static int sym_lineno = 0;

/* Count the number of .file directives, so that .loc is up to date.  */

static int num_source_filenames = 0;

/* Name of the file containing the current function.  */

static char *current_function_file = "";

/* Offsets to alpha virtual arg/local debugging pointers.  */

long alpha_arg_offset;
long alpha_auto_offset;

/* Emit a new filename to a stream.  */

void
alpha_output_filename (stream, name)
     FILE *stream;
     char *name;
{
  static int first_time = TRUE;
  char ltext_label_name[100];

  if (first_time)
    {
      first_time = FALSE;
      ++num_source_filenames;
      current_function_file = name;
      fprintf (stream, "\t.file\t%d ", num_source_filenames);
      output_quoted_string (stream, name);
      fprintf (stream, "\n");
      if (!TARGET_GAS && write_symbols == DBX_DEBUG)
	fprintf (stream, "\t#@stabs\n");
    }

  else if (!TARGET_GAS && write_symbols == DBX_DEBUG)
    {
      ASM_GENERATE_INTERNAL_LABEL (ltext_label_name, "Ltext", 0);
      fprintf (stream, "%s ", ASM_STABS_OP);
      output_quoted_string (stream, name);
      fprintf (stream, ",%d,0,0,%s\n", N_SOL, &ltext_label_name[1]);
    }

  else if (name != current_function_file
      && strcmp (name, current_function_file) != 0)
    {
      if (inside_function && ! TARGET_GAS)
	fprintf (stream, "\t#.file\t%d ", num_source_filenames);
      else
	{
	  ++num_source_filenames;
	  current_function_file = name;
	  fprintf (stream, "\t.file\t%d ", num_source_filenames);
	}

      output_quoted_string (stream, name);
      fprintf (stream, "\n");
    }
}

/* Emit a linenumber to a stream.  */

#ifdef __T3D__
void
alpha_output_lineno (stream, line)
     FILE *stream;
     int line;
{
  ++sym_lineno;
  fprintf (stream, "\n\t;loc\t%d,%d,%d\nLN$%d_%s::\n",
	   num_source_filenames? 1 : 0,
	   line, sym_lineno, sym_lineno, alpha_function_name);
}
#else
void
alpha_output_lineno (stream, line)
     FILE *stream;
     int line;
{
  if (! TARGET_GAS && write_symbols == DBX_DEBUG)
    {
      /* mips-tfile doesn't understand .stabd directives.  */
      ++sym_lineno;
      fprintf (stream, "$LM%d:\n\t%s %d,0,%d,$LM%d\n",
	       sym_lineno, ASM_STABN_OP, N_SLINE, line, sym_lineno);
    }
  else
    fprintf (stream, "\n\t.loc\t%d %d\n", num_source_filenames, line);
}
#endif

t3d_alignment_log (bits)
   unsigned int bits;
{
  int cnt = -1;

  while (bits)
    {
      cnt++;
      bits = bits >> 1;
    }
  cnt = cnt - 3;		/* adjust to bytes */
  if (cnt < 3)
    return 3;
  if (cnt > 9)
    return 9;
  return cnt;
}


/* Generate text section delimiters */

static char *
fixname(name)
     char *name;
{
  static char buf[64];
  char *p1, *p2;

  p1 = name;
  p2 = buf;

  while (*p1)
    {
      if (isalnum(*p1) || *p1 == '_' || *p1 == '@' || *p1 == '$' || *p1 == '.')
	{
	  *p2++ = *p1++;
	}
      else
	{
	  *p2++ = '$';
	  p1++;
	}
    }

  *p2 = '\0';
  return buf;
}

char *
t3d_text_section_delimit(name)
  char *name;
{
  static char buf[64];

  if (!name) 
    name = "mainfile";

  name = fixname(name);
  no_section_section();
  sprintf(buf, "\t.psect %s@code,code\n", name);
  return buf;
}

char *
t3d_data_section_delimit(name)
     char *name;
{
  static char buf[64];
  char *atdata;

  if (!name) 
    {
      name = "mainfile";
      atdata = "@maindata";
    }
  else if (!index(name, '@'))
    atdata = "@data";		/* prevents name conflicts in seg names */
  else
    atdata = "";
  name = fixname(name);
  no_section_section ();
  sprintf(buf, "\t.psect %s%s,data\n", name, atdata);
  return buf;
}

char *
t3d_comm_section_delimit(name)
     char *name;
{
  static char buf[64];

  no_section_section();
  name = fixname(name);
  sprintf(buf, "\t.psect %s@common,common\n", name);
  return buf;
}
/* Output at beginning of assembler file 
 * Expansion of ASM_FILE_START
 */

void 
t3d_start_file(file, main_input_filename)
     FILE *file;
     char *main_input_filename;
{
  char *p;
  fprintf(file, "\t.ident ");
  
  if (isdigit (*main_input_filename)) putc('$', file);
  for (p = main_input_filename; *p != NULL; p++)
    if (*p == '.' || *p == '/' || *p == '-') putc('$', file);
    else putc(*p, file);
  putc('\n', file);
  
  fprintf(file, "\t.macro loc DEPTH,LINENO,SYM\n");
  fprintf(file, "$LN'SYM:\n");
  fprintf(file, "\t.endm loc\n");
  fprintf(file, ";SOURCE_FILE:%s\n", main_input_filename);
}

/* Output at end of assembler file 
 * Expansion of ASM_FILE_END
 */

void 
t3d_end_file(file)
     FILE *file;
{
  no_section_section ();
  t3d_dump_externals (file);
  fprintf(file, "\t.end\n");
}

char *t3d_labelstr = ":\n";
/*
 * save info for a call info word
 */
int t3d_nargs;
int t3d_nregargs;
int t3d_regarg_mask;

/*
 * calculate the caller info word
 */
t3d_ci_info (args)
     tree args;
{
  int mask = 0;
  int cnt = 0;
  int reg_cnt = 0;

  for (; args; args = TREE_CHAIN (args))
    {
      rtx reg = DECL_INCOMING_RTL(args);
      cnt++;
      if (reg && GET_CODE(reg) == REG && FUNCTION_ARG_REGNO_P(REGNO(reg)))
	{
	  reg_cnt++;
	  if (REGNO(reg) > 32) 
	    mask |= 1 << (55 - REGNO(reg));
	}
    }
  if ((reg_cnt == 6) && (mask & 4))
    {
      reg_cnt = 7;
      mask &= 0xf8;
    }
  return ((cnt<<8) | mask | reg_cnt);
}
/*
 * output a CRAY standard "SSIB".  Info comes from CAM Refernece manual,
 * section A.2.5.8:
 *
 *      word 0:
 *           63:56 unused
 *	     55:48 length of function name
 *           47:40 static version of expected ci info
 *           39:32 version (?)
 *           31:24 kind (1 = leaf, 0 = non-leaf)
 *           23:16 lang (scc=8)
 *           15:00 length of ssib
 *      word 1:
 *           63:32 Saved R bit map (callee saved)
 *           31:00 Saved F bit map
 *      word 2: entry point address
 *      word 3: language specific info
 *      word 4: machine specific info
 *      word 5+: subprogram name
 */
t3d_output_ssib(file, fname, decl, size)
     FILE *file;
     char *fname;
     tree decl;
     int size;
{
  char buf[100];
  int i;
  int ireg_mask, freg_mask;
  ireg_mask = freg_mask = 0;

  for (i = 0; i < 32; i++)
    if (! fixed_regs[i] && ! call_used_regs[i] && regs_ever_live[i] && i != 26)
      ireg_mask |= 1 << i;

  for (i = 32; i < 64; i++)
    if (! fixed_regs[i] && ! call_used_regs[i] && regs_ever_live[i])
      freg_mask |= 1 << (i-32);

  sprintf(buf,"%s@ssib", fname);

  fprintf(file, "%s",
	  t3d_data_section_delimit(buf));
  fprintf(file, "\t.quad\t^X%x%08x\n",
	  strlen(fname)<<16 | (t3d_ci_info(DECL_ARGUMENTS(decl))&0xff)<<8 | 1,
	  leaf_function_p()<<24 | 12<<16 | 40);
  if (ireg_mask)
    fprintf(file, "\t.quad\t^X%x%08x\n", ireg_mask, freg_mask);
  else if (freg_mask)
    fprintf(file, "\t.quad\t^X%x\n", freg_mask);
  else
    fprintf(file, "\t.quad\t0\n");
  fprintf(file, "\t.quad\t%s\n", fname);
  fprintf(file, "\t.quad\t%d\n", 0);
  fprintf(file, "\t.quad\t%d\n", 0);
  ASM_OUTPUT_ASCII(file, fname, strlen(fname));
}

t3d_skip(fp, size)
     FILE *fp;
     int size;
{
  int x;
  while (size)
    {
      x = (size > (1024*1024)) ? (1024*1024) : size;
      fprintf (fp, "\t.blkb %d\n", x);
      size -= x;
    }
}

#define MAX_EXTERN_NAMES 10000
static int extern_names_index = 0;
static char *extern_names[MAX_EXTERN_NAMES];
/*
 * save up external names for end of section output
 */
t3d_output_external (name)
     char *name;
{
  int i;
  if (extern_names_index >= MAX_EXTERN_NAMES)
    abort();
  /* a bit of overhead to make sure we don't duplicate */
  for (i=0; i<extern_names_index; i++)
    if (!strcmp (name, extern_names[i]))
      return;
  extern_names[extern_names_index++] = name;
}

t3d_dump_externals (file)
     FILE *file;
{
  tree t, lookup_name();
  no_section_section ();

  while (extern_names_index)
    {
      extern_names_index--;
      t = get_identifier (extern_names[extern_names_index]);
      t = lookup_name(t);
      if (!t || TREE_CODE_CLASS (TREE_CODE (t)) != 'd' || DECL_EXTERNAL (t))
	fprintf(file, "\t.extern\t%s\n", extern_names[extern_names_index]);
    }
}

rtx t3d_prefetch_queue;
static rtx offset_mask;
static rtx annex_bits;
rtx last_dtb_user;

gen_dtb_set (newaddr, addr)
     rtx newaddr, addr;
{
  rtx offset, offset_align, pe, temp, label;

  if (!last_dtb_user)
    last_dtb_user = gen_reg_rtx(DImode);  
#if HOST_BITS_PER_LONG_INT == 64
  if (!t3d_prefetch_queue)
    {
      t3d_prefetch_queue = GEN_INT (0x8200180000);
      offset_mask = GEN_INT (0xffffffffffff);
      annex_bits = GEN_INT (0x1800000000);
    }
#else
  if (!t3d_prefetch_queue)
    {
      t3d_prefetch_queue = GEN_INT (0x820018);
      offset_mask = GEN_INT (0xffff);
      annex_bits = GEN_INT (0x18);
    }
#endif
  offset_align = gen_reg_rtx (DImode);
  pe = gen_reg_rtx (DImode);
#ifdef NOT_DEBUGGED_STUFF
  label = gen_label_rtx();
#endif

  emit_insn (gen_lshrdi3 (pe, addr, GEN_INT (48)));

#ifdef NOT_DEBUGGED_STUFF
  emit_move_insn (newaddr, addr);
  emit_insn (gen_cmpdi (pe, GEN_INT(3)));
  emit_insn (gen_beq (label));
#endif

#if HOST_BITS_PER_LONG_INT == 64
  emit_insn (gen_anddi3 (newaddr, addr, offset_mask));
  temp = force_reg (DImode, annex_bits);
#else
  temp = force_reg (DImode, offset_mask);
  emit_insn (gen_ashldi3 (temp, temp, GEN_INT (48)));
  emit_insn (gen_one_cmpldi2 (temp, temp));
  emit_insn (gen_anddi3 (newaddr, addr, temp));
  temp = force_reg (DImode, annex_bits);
  emit_insn (gen_ashldi3 (temp, temp, GEN_INT (32)));
#endif

  emit_insn (gen_iordi3 (newaddr, newaddr, temp));
  emit_insn (gen_anddi3 (offset_align, newaddr, GEN_INT (-32)));
  emit_insn (gen_alpha_stq_c (newaddr, offset_align, pe, newaddr, last_dtb_user));
#ifdef NOT_DEBUGGED_STUFF
  emit_label (label);
#endif
}

static int loop_label_num = 0;

/* Emit code to perform a block get.

   Assumes that structures are aligned on word boundries

   OPERANDS[0] is the destination, as a REG, clobbered.
   OPERANDS[1] is a register for the fifo queue
   OPERANDS[2] is the source pointer as a REG, clobbered.
   OPERANDS[3] is a reg or const_int size 
   OPERANDS[4] is a scratch register
   OPERANDS[5] is another scratch register */

char *
output_block_get (operands, size_is_const)
     rtx *operands;
     int size_is_const;
{
  int n_bytes;
  int offset;
  rtx temp;

  if (size_is_const)
    {
      n_bytes = INTVAL (operands[3]);
      if (n_bytes == 0)
	return "";
      
      /* unroll blocks less than prefetch queue  */
      if (n_bytes <= 8*16)
	{
	  
	  for (offset = 0; offset < n_bytes; offset += 8)
	    {
	      output_asm_insn("fetch 0(%2)", operands);
	      output_asm_insn("addq %2,8,%2", operands);
	    }
	  
	  output_asm_insn ("mb", operands);
	  
	  for (offset = 0; offset < n_bytes; offset += 8)
	    {
	      char buf[30];
	      output_asm_insn ("ldq %2,0(%1)", operands);
	      
	      if (offset+8 < n_bytes)
		{
		  sprintf(buf,"stq %%2,%d(%%0)", offset);
		  output_asm_insn (buf, operands);
		}
	    }
	  
	  if (n_bytes % 8 == 0)
	    {
	      char buf[30];
	      /* Store the last word.  */
	      sprintf(buf,"stq %%2,%d(%%0)", offset-8);
	      output_asm_insn (buf, operands);
	    }
	  else
	    {
	      char buf[30];
	      switch (n_bytes % 8)
		{
		case 1:
		  operands[3] = gen_rtx (CONST_INT, VOIDmode, 0x80);
		  break;
		case 2:
		  operands[3] = gen_rtx (CONST_INT, VOIDmode, 0xc0);
		  break;
		case 3:
		  operands[3] = gen_rtx (CONST_INT, VOIDmode, 0xe0);
		  break;
		case 4:
		  operands[3] = gen_rtx (CONST_INT, VOIDmode, 0xf0);
		  break;
		case 5:
		  operands[3] = gen_rtx (CONST_INT, VOIDmode, 0xf8);
		  break;
		case 6:
		  operands[3] = gen_rtx (CONST_INT, VOIDmode, 0xfc);
		  break;
		case 7:
		  operands[3] = gen_rtx (CONST_INT, VOIDmode, 0xfe);
		  break;
		}
	      output_asm_insn ("zapnot %2,%3,%2", operands);
	      sprintf(buf,"ldq %%4,%d(%%0)", offset-8);
	      output_asm_insn (buf, operands);
	      output_asm_insn ("zap %4,%5,%4", operands);
	      output_asm_insn ("bis %2,%4,%2", operands);
	      sprintf(buf,"stq %%2,%d(%%0)", offset-8);
	      output_asm_insn (buf, operands);
	    }
	  return "";
	}
    }
  /* copy with a loop */
    {
      char str1[20], str2[20], str3[20], buf[50], *p1, *p2, *p3;
      int i;

      ASM_GENERATE_INTERNAL_LABEL(str1, "LBC", loop_label_num);
      p1 = &str1[1];		/* skip * */
      ASM_GENERATE_INTERNAL_LABEL(str2, "LBC1", loop_label_num);
      p2 = &str2[1];		/* skip * */
      ASM_GENERATE_INTERNAL_LABEL(str3, "LBE", loop_label_num++);
      p3 = &str3[1];

      if (size_is_const)
	{
	  sprintf (buf, "lda %%5,%d(r31)", (n_bytes/8));
	  output_asm_insn (buf, operands);
	  output_asm_insn ("addq %2,120,%2", operands);
	}
      else
	{
	  output_asm_insn ("srl %3,3,%5", operands);  /* word count */
	  output_asm_insn ("lda %4,15(r31)", operands);
	  output_asm_insn ("cmple %5,15,%3", operands);
	  output_asm_insn ("cmovne %3,%5,%4", operands);
	  sprintf (buf, "laum %%3,%s(r31)", p1);
	  output_asm_insn (buf, operands);
	  output_asm_insn ("s8addq %4,%2,%2", operands);
	  output_asm_insn ("sll %3,32,%3", operands);
	  output_asm_insn ("s8addq %4,0,%4", operands);
	  sprintf (buf, "lalm %%3,%s(%%3)", p1);
	  output_asm_insn (buf, operands);
	  sprintf (buf, "lal %%3,%s(%%3)", p1);
	  output_asm_insn (buf, operands);
	  output_asm_insn ("subq %3,%4,%3", operands);
	  output_asm_insn ("jmp r31,(%3),0", operands);
	}
      
      for (i=0; i<15; i++)
	{
	  sprintf (buf, "subq %%2,%d,%%4", (15-i)*8);
	  output_asm_insn (buf, operands);
	  output_asm_insn ("fetch 0(%4)", operands);
	}

      sprintf (buf, "\n%s:\n\tmb", p1); 
      output_asm_insn (buf, operands);
      output_asm_insn ("cmple %5,15,%4", operands);
      sprintf (buf, "bne %%4,%s", p3);
      output_asm_insn (buf, operands);
      output_asm_insn ("subq %5, 15, %5", operands);

      sprintf (buf, "\n%s:\n\tldq %%4,0(%%1)", p2); 
      output_asm_insn (buf, operands);
      output_asm_insn ("subq %5,1,%5", operands);
      output_asm_insn ("fetch 0(%2)", operands);
      output_asm_insn ("addq %2,8,%2", operands);
      output_asm_insn ("stq %4,0(%0)", operands);
      output_asm_insn ("addq %0,8,%0", operands);
      sprintf (buf, "bgt %%5,%s", p2);
      output_asm_insn (buf, operands);

      output_asm_insn ("lda %5,15(r31)", operands);

      sprintf (buf, "\n%s:\n\tldq %%4,0(%%1)", p3); 
      output_asm_insn (buf, operands);
      output_asm_insn ("subq %5,1,%5", operands);
      output_asm_insn ("stq %4,0(%0)", operands);
      output_asm_insn ("addq %0,8,%0", operands);
      sprintf (buf, "bgt %%5,%s", p3);
      output_asm_insn (buf, operands);
      
    }

  return "";
}
/* Emit code to perform a block move.

   Restriction: If the length argument is non-constant, alignment
   must be 8.

   OPERANDS[0] is the destination pointer as a REG, clobbered.
   OPERANDS[1] is the source pointer as a REG, clobbered.
   if SIZE_IS_CONSTANT
     OPERANDS[2] is a register for temporary storage.
     OPERANDS[5] is the size as a CONST_INT
   else
     OPERANDS[2] is a REG which will contain the size, clobbered.
   OPERANDS[3] is a register for temporary storage.
   OPERANDS[4] is a register for temporary storage.
   OPERANDS[6] is the alignment safe to use, as a CONST_INT.  */

char *
output_block_move (operands, size_is_constant)
     rtx *operands;
     int size_is_constant;
{
  int align = INTVAL (operands[6]);
  unsigned long n_bytes;

  /* We can't move more than four bytes at a time because the PA
     has no longer integer move insns.  (Could use fp mem ops?)  */
  if (align > 8)
    align = 8;

  if (size_is_constant)
    {
      unsigned long offset;
      rtx temp;

      n_bytes = INTVAL (operands[5]);
      if (n_bytes == 0)
	return "";

      if (align >= 8)
	{
	  /* Don't unroll too large blocks.  */
	  if (n_bytes > 256)
	    goto copy_with_loop;

	  output_asm_insn ("ldq %2,0(%1)", operands);

	  for (offset = 8; offset < n_bytes;
	       offset += (offset+8 < n_bytes) ? 16 : 8)
	    {
	      char buf[30];

	      sprintf(buf,"ldq %%3,%d(%%1)", offset);
	      output_asm_insn (buf, operands);
	      if (offset+8 < n_bytes)
		{
		  sprintf(buf,"ldq %%4,%d(%%1)", offset+8);
		  output_asm_insn (buf, operands);
		}
	      sprintf(buf,"stq %%2,%d(%%0)", offset-8);
	      output_asm_insn (buf, operands);
	      if (offset+8 < n_bytes)
		{
		  sprintf(buf,"stq %%3,%d(%%0)", offset);
		  output_asm_insn (buf, operands);

		  temp = operands[4];
		  operands[4] = operands[3];
		  operands[3] = operands[2];
		  operands[2] = temp;
		} 
	      else
		{
		  temp = operands[4];
		  operands[4] = operands[2];
		  operands[2] = operands[3];
		  operands[3] = temp;
		}
	    }

	  if (n_bytes % 8 == 0)
	    {
	      char buf[30];
	      /* Store the last word.  */
	      sprintf(buf,"stq %%2,%d(%%0)", offset-8);
	      output_asm_insn (buf, operands);
	    }
	  else
	    {
	      char buf[30];
	      switch (n_bytes % 8)
		{
		case 1:
		  operands[5] = gen_rtx (CONST_INT, VOIDmode, 0x80);
		  break;
		case 2:
		  operands[5] = gen_rtx (CONST_INT, VOIDmode, 0xc0);
		  break;
		case 3:
		  operands[5] = gen_rtx (CONST_INT, VOIDmode, 0xe0);
		  break;
		case 4:
		  operands[5] = gen_rtx (CONST_INT, VOIDmode, 0xf0);
		  break;
		case 5:
		  operands[5] = gen_rtx (CONST_INT, VOIDmode, 0xf8);
		  break;
		case 6:
		  operands[5] = gen_rtx (CONST_INT, VOIDmode, 0xfc);
		  break;
		case 7:
		  operands[5] = gen_rtx (CONST_INT, VOIDmode, 0xfe);
		  break;
		}
	      output_asm_insn ("zapnot %2,%5,%2", operands);
	      sprintf(buf,"ldq %%3,%d(%%0)", offset-8);
	      output_asm_insn (buf, operands);
	      output_asm_insn ("zap %3,%5,%3", operands);
	      output_asm_insn ("bis %2,%3,%2", operands);
	      sprintf(buf,"stq %%2,%d(%%0)", offset-8);
	      output_asm_insn (buf, operands);
	    }
	  return "";
	}

      if (align >= 4 && n_bytes >= 4)
	{
	  output_asm_insn ("ldl %2,0(%1)", operands);

	  for (offset = 4; offset < n_bytes; offset += 4)
	    {
	      char buf[30];

	      sprintf(buf,"ldl %%3,%d(%%1)", offset);
	      output_asm_insn (buf, operands);
	      sprintf(buf,"stl %%2,%d(%%0)", offset-4);
	      output_asm_insn (buf, operands);

	      temp = operands[2];
	      operands[2] = operands[3];
	      operands[3] = temp;
	    }
	  if (n_bytes % 4 == 0)
	    {
	      char buf[30];
	      /* Store the last word.  */
	      sprintf(buf,"stl %%2,%d(%%0)", offset-4);
	      output_asm_insn (buf, operands);
	    }
	  else
	    {
	      char buf[30];
	      /* Store the last, partial word.  */
	      switch (n_bytes % 4)
		{
		case 1:
		  operands[5] = gen_rtx (CONST_INT, VOIDmode, 0x08);
		  break;
		case 2:
		  operands[5] = gen_rtx (CONST_INT, VOIDmode, 0x0c);
		  break;
		case 3:
		  operands[5] = gen_rtx (CONST_INT, VOIDmode, 0x0e);
		  break;
		}
	      sprintf(buf,"ldl %%3,%d(%%0)", offset-4);
	      output_asm_insn (buf, operands);
	      output_asm_insn ("zap %3,%5,%3", operands);
	      output_asm_insn ("zapnot %2,%5,%2", operands);
	      output_asm_insn ("bis %2,%3,%2", operands);
	      sprintf(buf,"stl %%2,%d(%%0)", offset-4);
	      output_asm_insn (buf, operands);
	    }
	  return "";
	}

      abort();
    }

  if (align != 8)
    abort();
     
 copy_with_loop:

  if (size_is_constant)
    {
      /* Size is compile-time determined, and also not
	 small (such small cases are handled above).  */
      operands[5] = gen_rtx (CONST_INT, VOIDmode, n_bytes);
      output_asm_insn ("lda %2,%5(r31)", operands);
    }

  {
    char str1[20], str2[20], buf[50], *p1, *p2;

    ASM_GENERATE_INTERNAL_LABEL(str1, "LBC", loop_label_num++);
    p1 = &str1[1];		/* skip * */
    ASM_GENERATE_INTERNAL_LABEL(str2, "LBCE", loop_label_num++);
    p2 = &str2[1];

    output_asm_insn ("ldq %3,0(%1)", operands);

    sprintf (buf, "\n%s:\n\tldq %%4,8(%%1)", p1);
    output_asm_insn (buf, operands);
    output_asm_insn ("subq %2,8,%2", operands);
    output_asm_insn ("stq %3,0(%0)", operands);
    output_asm_insn ("ldq %3,16(%1)", operands);
    output_asm_insn ("addq %1,16,%1", operands);
    sprintf (buf, "ble %%2,%s", p2);
    output_asm_insn (buf, operands);
    output_asm_insn ("subq %2,8,%2", operands);
    output_asm_insn ("stq %4,8(%0)", operands);
    output_asm_insn ("addq %0,16,%0", operands);
    sprintf (buf, "bgt %%2,%s\n%s:\n", p1,p2);
    output_asm_insn (buf, operands);
  }
  /* The counter is negative, >= -4.  The remaining number of bytes are
     determined by the two least significant bits.  */

  if (size_is_constant)
    {
      if (n_bytes % 8 != 0)
	{
	  char buf[30];
	  switch (n_bytes % 8)
	    {
	    case 1:
	      operands[4] = gen_rtx (CONST_INT, VOIDmode, 0x80);
	      break;
	    case 2:
	      operands[4] = gen_rtx (CONST_INT, VOIDmode, 0xc0);
	      break;
	    case 3:
	      operands[4] = gen_rtx (CONST_INT, VOIDmode, 0xe0);
	      break;
	    case 4:
	      operands[4] = gen_rtx (CONST_INT, VOIDmode, 0xf0);
	      break;
	    case 5:
	      operands[4] = gen_rtx (CONST_INT, VOIDmode, 0xf8);
	      break;
	    case 6:
	      operands[4] = gen_rtx (CONST_INT, VOIDmode, 0xfc);
	      break;
	    case 7:
	      operands[4] = gen_rtx (CONST_INT, VOIDmode, 0xfe);
	      break;
	    }
	  output_asm_insn ("ldq %1,0(%0)", operands);
	  output_asm_insn ("zap %1,%4,%1", operands);
	  output_asm_insn ("zapnot %2,%4,%2", operands);
	  output_asm_insn ("bis %1,%2,%1", operands);
	  output_asm_insn ("stq %1,0(%0)", operands);
	}
    }
  else
    {
      /* need to do last word, potentially */

    }
  return "";
}

#ifdef __T3D__
/* Write function prologue.  */

void
output_prolog (file, size)
     FILE *file;
     int size;
{
  HOST_WIDE_INT frame_size = ((size + current_function_outgoing_args_size
			       + current_function_pretend_args_size
			       + alpha_sa_size () + 15) & ~15);
  HOST_WIDE_INT reg_offset = frame_size - 8;
  HOST_WIDE_INT start_reg_offset = reg_offset;
  HOST_WIDE_INT actual_start_reg_offset = start_reg_offset;
  rtx insn;
  int reg_offset_base_reg = 30;
  int i;

  ASM_OUTPUT_LABEL (file, alpha_function_name);
  inside_function = TRUE;

  /* Set up offsets to alpha virtual arg/local debugging pointer.  */

  alpha_auto_offset = 0;
  alpha_arg_offset = 0;


  /* Adjust the stack by the frame size.  If the frame size is > 4096
     bytes, we need to be sure we probe somewhere in the first and last
     4096 bytes (we can probably get away without the latter test) and
     every 8192 bytes in between.  If the frame size is > 32768, we
     do this in a loop.  Otherwise, we generate the explicit probe
     instructions. 

     Note that we are only allowed to adjust sp once in the prologue.  */

  if (frame_size < 32768)
    {
      if (frame_size > 4096)
	{
	  int probed = 4096;
	  int regnum = 2;

	  fprintf (file, "\tldq r%d,-%d(r30)\n", regnum++, probed);

	  while (probed + 8192 < frame_size)
	    fprintf (file, "\tldq r%d,-%d(r30)\n", regnum++, probed += 8192);

	  if (probed + 4096 < frame_size)
	    fprintf (file, "\tldq r%d,-%d(r30)\n", regnum++, probed += 4096);

	  if (regnum > 9)
	    abort ();
	}

      if (frame_size != 0)
	fprintf (file, "\tlda r30,-%d(r30)\n", frame_size);
    }
  else
    {
      /* Here we generate code to set R4 to SP + 4096 and set R5 to the
	 number of 8192 byte blocks to probe.  We then probe each block
	 in the loop and then set SP to the proper location.  If the
	 amount remaining is > 4096, we have to do one more probe.  */

      HOST_WIDE_INT blocks = (frame_size + 4096) / 8192;
      HOST_WIDE_INT leftover = frame_size + 4096 - blocks * 8192;

      add_long_const (file, blocks, 31, 5, 5);

      fprintf (file, "\tlda r4,4096(r30)\n");
      fprintf (file, "%s..sc:\n", alpha_function_name);
      fprintf (file, "\tldq r6,-8192(r4)\n");
      fprintf (file, "\tsubq r5,1,r5\n");
      fprintf (file, "\tlda r4,-8192(r4)\n");
      fprintf (file, "\tbne r5,%s..sc\n", alpha_function_name);
      fprintf (file, "\tlda r30,-%d(r4)\n", leftover);

      if (leftover > 4096)
	fprintf (file, "\tldq r2,%d(r30)\n", leftover - 4096);
    }

  /* If reg_offset is "close enough" to 2**15 that one of the offsets would
     overflow a store instruction, compute the base of the register save
     area into r28 temporarily.  this will become the frame pointer.  */
  if (reg_offset >= 32768 - alpha_sa_size () && alpha_sa_size () != 0)
    {
      add_long_const (file, frame_size, 30, 28, 28);
      reg_offset_base_reg = 28;
      reg_offset = start_reg_offset = -8;
    }

  /* Save register 26 if it is used or if any other register needs to
     be saved.  */
  if (regs_ever_live[26] || alpha_sa_size () != 0)
    fprintf (file, "\tstq r26,%d(r%d)\n", reg_offset, reg_offset_base_reg);
  reg_offset -= 8;

  /* Save the frame pointer in correct stack slot */
  if (regs_ever_live[15] || frame_pointer_needed)
    fprintf (file, "\tstq r15,%d(r%d)\n", reg_offset, reg_offset_base_reg);
  reg_offset -= 8;

  /* shared_stack top not implemented */
  reg_offset -= 8;

  /* set up pointer to static info block */
  fprintf (file, "\tlaum r4,%s@ssib(r31)\n", alpha_function_name);
  fprintf (file, "\tsll r4,32,r4\n");
  fprintf (file, "\tlalm r4,%s@ssib(r4)\n", alpha_function_name);
  fprintf (file, "\tlal r4,%s@ssib(r4)\n", alpha_function_name);
  fprintf (file, "\tstq r4,%d(r%d)\n", reg_offset, reg_offset_base_reg);
  reg_offset -= 8;

  /* ci register is next */
  if (regs_ever_live[25])
    fprintf (file, "\tstq r25,%d(r%d)\n", reg_offset, reg_offset_base_reg);
  reg_offset -= 8;

  /* 
  /* Now save any other used integer registers required to be saved.  */
  for (i = 31; i >= 0; i--)
    if (! fixed_regs[i] && ! call_used_regs[i] && regs_ever_live[i])
      {
	if (i == 26 || i == 15 || i == 25)
	  continue;
	fprintf (file, "\tstq r%d,%d(r%d)\n",
		 i, reg_offset, reg_offset_base_reg);
	reg_offset -= 8;
      }

  start_reg_offset = reg_offset;
  for (i = 31; i >= 0; i--)
    if (! fixed_regs[i + 32] && ! call_used_regs[i + 32]
	&& regs_ever_live[i + 32])
      {
	fprintf (file, "\tstt f%d,%d(r%d)\n",
		 i, reg_offset, reg_offset_base_reg);
	reg_offset -= 8;
      }

  /* If we need a frame pointer, set it from the stack pointer.  Note that
     this must always be the last instruction in the prologue.  */
  if (frame_pointer_needed)
    {
      if (frame_size < 32768)
	fprintf (file, "\tlda r15,%d(r30)\n", frame_size);
      else
	fprintf (file, "\tbis r28,r28,r15\n"); /* already in 28 */
    }
}

/* Write function epilogue.  */

void
output_epilog (file, size)
     FILE *file;
     int size;
{
  rtx insn = get_last_insn ();
  HOST_WIDE_INT frame_size = ((size + current_function_outgoing_args_size
			       + current_function_pretend_args_size
			       + alpha_sa_size () + 15) & ~15);
  HOST_WIDE_INT reg_offset = frame_size - 8;
  HOST_WIDE_INT frame_size_from_reg_save = frame_size - reg_offset;
  int reg_offset_base_reg = 30;
  int i;
  extern rtx last_gsync_dest;

  /* If the last insn was a BARRIER, we don't have to write anything except
     the .end pseudo-op.  */
  if (GET_CODE (insn) == NOTE)
    insn = prev_nonnote_insn (insn);
  if (insn == 0 || GET_CODE (insn) != BARRIER)
    {
      int fp_offset;

      /* If we have a frame pointer, restore from it.  */
      if (frame_pointer_needed)
	{
	  fprintf (file, "\tbis r15,r15,r28\n");
	  reg_offset_base_reg = 28;
	  reg_offset = -8;
	}

      /* If the register save area is out of range, put its address into
	 r28.  */
      if (reg_offset >= 32768 - alpha_sa_size () && alpha_sa_size () != 0)
	{
	  add_long_const (file, reg_offset, 30, 28, 28);
	  reg_offset_base_reg = 28;
	  reg_offset = -8;
	}

      /* Restore all the registers, starting with the return address
	 register.  */
      if (regs_ever_live[26] || alpha_sa_size () != 0)
	fprintf (file, "\tldq r26,%d(r%d)\n", reg_offset, reg_offset_base_reg);
      reg_offset -= 8;

      /* set up to restore the frame pointer from correct stack slot */
      fp_offset = reg_offset;
      reg_offset -= 8;

      /* shared_stack top not implemented */
      reg_offset -= 8;

      /* pointer to static info block not restored */
      reg_offset -= 8;

      /* ci register is not restored */
      reg_offset -= 8;

      /* Now restore any other used integer registers that that we saved,
	 except for FP if it is being used as FP, since it must be
	 restored last.  */

      for (i = 32; i >= 0; i--)
	if (! fixed_regs[i] && ! call_used_regs[i] && regs_ever_live[i])
	  {
	    if (i == 15 || i == 26 || i == 25)
	      continue;
	    fprintf (file, "\tldq r%d,%d(r%d)\n",
		     i, reg_offset, reg_offset_base_reg);
	    reg_offset -= 8;
	  }
      
      for (i = 32; i >= 0; i--)
	if (! fixed_regs[i + 32] && ! call_used_regs[i + 32]
	    && regs_ever_live[i + 32])
	  {
	    fprintf (file, "\tldt f%d,%d(r%d)\n",
		     i, reg_offset, reg_offset_base_reg);
	    reg_offset -= 8;
	  }
      
      /* If we needed a frame pointer and we have to restore it, do it
	 now.  This must be done in one instruction immediately
	 before the SP update.  */
      if (frame_pointer_needed || regs_ever_live[FRAME_POINTER_REGNUM])
	fprintf (file, "\tldq r15,%d(r%d)\n", fp_offset, reg_offset_base_reg);

      /* adjust the stack pointer */
      if (!frame_pointer_needed && frame_size < 32768)
	fprintf (file, "\tlda r30,%d(r30)\n", frame_size);
      else
	fprintf (file, "\tbis r28,r28,r30\n"); /* already in r28 */

      /* Finally return to the caller.  */
      fprintf (file, "\tret r31,(r26)\n");
    }

  /* Show that we know this function if it is called again.  */
  SYMBOL_REF_FLAG (XEXP (DECL_RTL (current_function_decl), 0)) = 1;
  /* reset static rtx's used in each function */
  t3d_prefetch_queue = offset_mask = annex_bits = last_dtb_user = last_gsync_dest = 0;
}
#endif

#ifdef __T3D__
/* If this function is called, there are arguments in registers that need
   to be moved to the stack so that va_start, etc., can process them
   correctly. We don't have to move explicitly named arguments; hence, the
   maximum number of args to save on the stack at any call site for this
   function will be 6-cum.cum.  Write code that at runtime tests the Call
   Information Word (reg 25) to determine the type of the argument to know
   whether to move a floating point value from the appropriate integer register
   before saving the register to the stack. The type bits in reg 25 are
   in bit positions 3-7. */
void
setup_incoming_varargs (cum)
     struct t3d_cum_args cum;
{
  int i;
  rtx ci, cond_reg, pat, ireg, freg, skiplabel, endlabel, insn;

  insn = get_last_insn();
  ci = gen_rtx (REG, DImode, 25);
  cond_reg = gen_rtx (REG, DImode, 0); /* use return value register */
  for (i = cum.cum; i < 6; i++)
    {
      skiplabel = gen_label_rtx();
      endlabel = gen_label_rtx();
      ireg = gen_rtx (REG, DImode, 16 + i);
      freg = gen_rtx (REG, DFmode, 32 + 16 + i);
      /* set condition by testing the right bit in reg 25 */
      if (i != 5)
	insn = emit_insn (gen_anddi3 (cond_reg,ci,
				      gen_rtx (CONST_INT, DImode,
					       1<<(7-i))));
      else
	insn = emit_insn (gen_anddi3 (cond_reg,ci,
				      gen_rtx (CONST_INT, DImode, 1)));
      insn = emit_jump_insn_after (gen_rtx (SET, VOIDmode, pc_rtx,
		  gen_rtx (IF_THEN_ELSE, VOIDmode,
			   gen_rtx (EQ, VOIDmode, cond_reg, const0_rtx),
			   gen_rtx (LABEL_REF, VOIDmode, skiplabel), pc_rtx)),
				   insn);
      insn = emit_move_insn (gen_rtx (MEM, DFmode,
				      plus_constant (virtual_incoming_args_rtx,
						     i * UNITS_PER_WORD)),
			     freg);
      insn = emit_jump_insn_after (gen_rtx (SET, VOIDmode, pc_rtx,
					    gen_rtx (LABEL_REF, VOIDmode,
						     endlabel)),
				   insn);
      insn = emit_label_after (skiplabel, insn);
      insn = emit_move_insn (gen_rtx (MEM, DImode,
				      plus_constant (virtual_incoming_args_rtx,
						     i * UNITS_PER_WORD)),
			     ireg);
      insn = emit_label_after (endlabel, insn);
    }
}
#endif
