#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>

#include "myrelo.h"
#include "mysymbol.h"
#include "../gsyms.h"

#define NWORDS(x)  (((x)+7)>>3)
#define LINESZ		256	/* max line length expected (only include sdb lines) */
#define GRABSZ          100	/* number of table elements to grab at a time */

/* global info extracted from the .s file */
int *lineno_map, lineno_cnt;
char *source_file_name;
char *module_name;

struct psd *psd; int n_psd, max_psd;
#define PSD_INDEX(x) ((x) * NWORDS( sizeof (struct psd))+1)
#define INDEX_PSD(x) (((x)-1) / NWORDS( sizeof (struct psd)))
struct rbs *rbs; int n_rbs, max_rbs;
#define RBS_INDEX(x) ((x) * NWORDS( sizeof (struct rbs))+1)
#define INDEX_RBS(x) (((x)-1) / NWORDS( sizeof (struct rbs)))
#define TDS_INDEX(x) ((x) * NWORDS( sizeof (struct tds))+1)
#define INDEX_TDS(x) (((x)-1) / NWORDS( sizeof (struct tds)))
struct vds *vds; int n_vds, max_vds;
#define VDS_INDEX(x) ((x) * NWORDS( sizeof (struct vds))+1)
#define INDEX_VDS(x) (((x)-1) / NWORDS( sizeof (struct vds)))
struct nld *nld; int n_nld, max_nld;
#define NLD_INDEX(x) ((x) * NWORDS( sizeof (struct nld))+1)
#define INDEX_NLD(x) (((x)-1) / NWORDS( sizeof (struct nld)))
struct uld *uld; int n_uld, max_uld;
#define ULD_INDEX(x) ((x) * NWORDS( sizeof (struct uld))+1)
#define INDEX_ULD(x) (((x)-1) / NWORDS( sizeof (struct uld)))

long *name_buf; int n_name, max_name;

struct tds default_tds[] =
{
  {"T_NULL", 6, 0, 1, SYM_DF_OCTAL, 3, 0, SYM_DF_DECIMAL},
  {"T_VOID", 6, 0, 1, SYM_DF_OCTAL, 3, 64, SYM_DF_DECIMAL},
  {"T_CHAR", 6, 0, 1, SYM_DF_ASCII, 3, 8, SYM_DF_ASCII},
  {"T_SHORT", 7, 0, 1, SYM_DF_DECIMAL, 3, 32, SYM_DF_DECIMAL},
  {"T_INT", 5, 0, 1, SYM_DF_DECIMAL, 3, 64, SYM_DF_DECIMAL},
  {"T_LONG", 6, 0, 1, SYM_DF_DECIMAL, 3, 64, SYM_DF_DECIMAL},
  {"T_FLOAT", 7, 0, 1, SYM_DF_FLOAT, 3, 32, SYM_DF_FLOAT},
  {"T_DOUBLE", 8, 0, 1, SYM_DF_FLOAT, 3, 64, SYM_DF_FLOAT},
  {"T_STRUCT", 8, 0, 1, SYM_DF_DECIMAL, 3, 64, SYM_TC_STRUCTURE},
  {"T_UNION", 7, 0, 1, SYM_DF_DECIMAL, 3, 64, SYM_TC_STRUCTURE},
  {"T_ENUM", 6, 0, 1, SYM_DF_DECIMAL, 3, 64, SYM_TC_ENUMERATION},
  {"T_MOE", 5, 0, 1, SYM_DF_DECIMAL, 3, 64, SYM_TC_ENUMERATION},
  {"T_UCHAR", 7, 0, 1, SYM_DF_ASCII, 3, 8, SYM_DF_ASCII},
  {"T_USHORT", 8, 0, 1, SYM_DF_DECIMAL, 3, 32, SYM_DF_UDECIMAL},
  {"T_UINT", 6, 0, 1, SYM_DF_DECIMAL, 3, 64, SYM_DF_UDECIMAL},
  {"T_ULONG", 7, 0, 1, SYM_DF_DECIMAL, 3, 64, SYM_DF_UDECIMAL}
};

main(argc, argv)
     int argc;
     char **argv;
{
  FILE *objfile, *asmfile, *outfile;
  char *buf;
  int cur_len;

  if (argc != 4)
    {
      fprintf(stderr, "usage: fixsyms objfile asmfile outfile\n");
      exit(1);
    }

  if ((objfile = fopen(argv[1], "r")) == NULL)
    {
      fprintf(stderr, "can't open %s as objfile\n", argv[1]);
      exit(1);
    }

  if (!strcmp(argv[1],argv[3]))
      unlink(argv[1]);

  if ((asmfile = fopen(argv[2], "r")) == NULL)
    {
      fprintf(stderr, "can't open %s as asmfile\n", argv[2]);
      exit(1);
    }

  if ((outfile = fopen(argv[3], "w")) == NULL)
    {
      fprintf(stderr, "can't open %s as objfile\n", argv[3]);
      exit(1);
    }

  buf = malloc (50*sizeof (long));
  cur_len = 50;

  psd = (struct psd *) malloc (GRABSZ * sizeof(struct psd));
  max_psd = GRABSZ;
  n_psd = 0;

  rbs = (struct rbs *) malloc (GRABSZ * sizeof(struct rbs));
  max_rbs = GRABSZ;
  n_rbs = 0;

  vds = (struct vds *) malloc (GRABSZ * sizeof(struct vds));
  max_vds = GRABSZ;
  n_vds = 0;

  nld = (struct nld *) malloc (GRABSZ * sizeof(struct nld));
  max_nld = GRABSZ;
  n_nld = 0;

  uld = (struct uld *) malloc (GRABSZ * sizeof(struct uld));
  max_uld = GRABSZ;
  n_uld = 0;

  name_buf = (long *) malloc (GRABSZ * sizeof(long));
  max_name = GRABSZ;
  n_name = 0;

  parse_asm_file(asmfile);

  while (1)
    {
      struct tbl_hdr *th;
      if (fread (buf, sizeof(struct tbl_hdr), 1, objfile) == 0)
	break;
      th = (struct tbl_hdr *)buf;
      if (cur_len < th->hdr_len)
	{
	  buf = realloc(buf, th->hdr_len * sizeof(long));
	  th = (struct tbl_hdr *)buf;
	  cur_len = th->hdr_len;
	}
      fread (&buf[sizeof (struct tbl_hdr)], sizeof(long),
	     th->hdr_len - (sizeof (struct tbl_hdr)/sizeof(long)), objfile);

      switch (th->hdr_type)
	{
	case 011:
#ifdef DEBUG_SCAN
	  fprintf(stderr, "current mth\n");
	  print_mth(buf);
	  print_psd_from_mth(buf, buf);
	  print_rbs_from_mth(buf, buf);
	  print_tds_from_mth(buf, buf);
	  print_vds_from_mth(buf, buf);
	  print_nld_from_mth(buf, buf);
	  print_uld_from_mth(buf, buf);
#endif
	  /* this is the "mth", so don't write as we are replacing it */
	  break;

	case 047:
	  /* we parse this table twice, once to get the correspondence between
	     line numbers and addresses, then again to get the "real" symbols */
	  read_dpd_lnums(buf, buf);
	  read_dpd_table(buf, buf);
	  write_dpd_table(buf, buf, outfile);
	  break;

	case 040:
	  /* this is "module compete", so first output our mth */
	  write_mth(outfile, asmfile);
	  fwrite(buf, sizeof(long), th->hdr_len, outfile);
	  break;


	default:
	  fwrite(buf, sizeof(long), th->hdr_len, outfile);
	  break;
	}
    }
}

read_dpd_lnums(dpd, buf)
     struct dpd_tbl *dpd;
     long *buf;
{
  struct pdtent *pdtent, *pdtentend;
  struct pdtblck *pdtblck, *pdtblckend;
  int offset = 0, t = 0;
  int last_lnum = 1;

  pdtblck = (struct pdtblck *)&buf[dpd->hdl];
  pdtblckend = (struct pdtblck *)&buf[dpd->hdl+dpd->bdl];

  while (pdtblck < pdtblckend)
    {
      char *name = ((char *)pdtblck)+sizeof(struct pdtblck);
      int nlen = pdtblck->pdtbknl;

      if (n_rbs >= max_rbs)
	{
	  max_rbs += GRABSZ;
	  rbs = realloc(rbs, sizeof (struct rbs) * max_rbs);
	}

      rbs[n_rbs].rbs_nlen= nlen;	/* name length */
      rbs[n_rbs].rbs_ints= put_name(&rbs[n_rbs].rbs_name, name, nlen);
      
      rbs[n_rbs].rbs_cb= pdtblck->pdtbkcb;	/* set: common block */
      rbs[n_rbs].rbs_bl= pdtblck->pdtbkl;	/* block location */
      rbs[n_rbs].rbs_bc= pdtblck->pdtbkc;	/* block contents */
      rbs[n_rbs].rbs_bt= pdtblck->pdtbkt;	/* block type */
      rbs[n_rbs].rbs_iblk= t;			/* block index */
      rbs[n_rbs].dummy = offset;
      n_rbs++;

      t+=2;

      pdtblck++;			/* skip structure */
      pdtblck = (struct pdtblck *)(((long *) pdtblck) + NWORDS (nlen)); /* skip name  */
      offset += NWORDS(sizeof (struct pdtblck)) + NWORDS (nlen);
    }


  pdtent = (struct pdtent *)&buf[dpd->hdl+dpd->bdl];
  pdtentend = (struct pdtent *)&buf[dpd->hdl+dpd->bdl+dpd->sdl];

  while (pdtent < pdtentend)
    {
      char *name = ((char *)pdtent)+sizeof(struct pdtent);
      int nlen = pdtent->pdtenl;
      
      if (!strncmp(name, "LN$", 3))
	{
	  int num = atoi(name+3);
	  if (!num)
	    {
	      fprintf(stderr, "invalid LN$ symbol %s\n", name);
	    }
	  else
	    {
	      int i;

	      last_lnum = num;
	      uld[num].uld_addr.si_sc = 1;
	      uld[num].uld_addr.si_irbs = find_rlb(pdtent->pdtebi);
	      uld[num].uld_addr.si_offset = pdtent->pdtepv;
	      uld[num].uld_type = SYM_LT_START_OTHER;
	      pdtent->pdtepv = 1000; /* WWCHACK */
	      for (i=0; i<n_psd; i++)
		{
		  if (psd[i].psd_eaddr.si_sc == 63 &&
		      psd[i].psd_eaddr.si_offset == uld[num].uld_lnum)
		    {
		      psd[i].psd_eaddr = uld[num].uld_addr;
		      if (psd[i].psd_parent == 1) /* only for top level functions */
			psd[i].psd_eaddr.si_offset = 0; /* WWChack */
		    }
		  if (psd[i].psd_climit.si_sc == 63 &&
		      psd[i].psd_climit.si_offset == uld[num].uld_lnum)
		    {
		      psd[i].psd_climit = uld[num].uld_addr;
		    }
		}
#ifdef DEBUG_SCAN
	      fprintf(stderr, "found entry %d line %d offset %o\n", num,
		      uld[num].uld_lnum, uld[num].uld_addr.si_offset);
#endif
	    }
	}
      pdtent++;			/* skip structure */
      pdtent = (struct pdtent *)(((long *) pdtent) + NWORDS(nlen)); /* skip name */
    }
}

read_dpd_table(dpd, buf)
     struct dpd_tbl *dpd;
     long *buf;
{
  struct pdtblck *pdtblck, *pdtblckend;
  struct pdtent *pdtent, *pdtentend;
  int offset = 0, t = 0;
  int last_lnum = 1;

#if 0
  pdtblck = (struct pdtblck *)&buf[dpd->hdl];
  pdtblckend = (struct pdtblck *)&buf[dpd->hdl+dpd->bdl];

  while (pdtblck < pdtblckend)
    {
      char *name = ((char *)pdtblck)+sizeof(struct pdtblck);
      int nlen = pdtblck->pdtbknl;
      union sinfo addr;
      int ipsd;
      
      addr.si_sc = 1;
      addr.si_irbs = RBS_INDEX(t);
      addr.si_offset = 0;
      t++;

      ipsd = find_psd (addr);

      if (ipsd)
	{
	  if (n_nld >= max_nld)
	    {
	      max_nld += GRABSZ;
	      nld = realloc(nld, sizeof (struct nld) * max_nld);
	    }
	  nld[n_nld].nld_ints = put_name(&nld[n_nld].nld_name, name, nlen);
	  nld[n_nld].nld_nlen = nlen;
	  nld[n_nld].nld_u = 0;	/* user defined */
	  nld[n_nld].nld_sinfo = addr;
	  nld[n_nld].nld_ipsd = ipsd;
	  nld[n_nld].nld_lnum = psd[INDEX_PSD(nld[n_nld].nld_ipsd)].psd_entry;
	  nld[n_nld].nld_type = SYM_LT_EXPORTED_ENTRY;
	  nld[n_nld].nld_enlen = 0;
	  nld[n_nld].nld_ename = 0;
	  n_nld++;
	}

      pdtblck++;			/* skip structure */
      pdtblck = (struct pdtblck *)(((long *) pdtblck) + NWORDS (nlen)); /* skip name  */
      offset += NWORDS(sizeof (struct pdtblck)) + NWORDS (nlen);
    }

#endif

  pdtent = (struct pdtent *)&buf[dpd->hdl+dpd->bdl];
  pdtentend = (struct pdtent *)&buf[dpd->hdl+dpd->bdl+dpd->sdl];

  while (pdtent < pdtentend)
    {
      char *name = ((char *)pdtent)+sizeof(struct pdtent);
      int nlen = pdtent->pdtenl;
      
      if (strncmp(name, "LN$", 3) && strncmp(name, "$L", 2)) /* skip compiler gened */
	{
	  union sinfo addr;
	  int ipsd;

	  addr.si_sc = 1;
	  addr.si_irbs = find_rlb(pdtent->pdtebi);
	  addr.si_offset = pdtent->pdtepv;
	  ipsd = find_psd (addr);

	  if (ipsd)		/* make sure we have a scope for this, otherwise skip */
	    {
	      if (n_nld >= max_nld)
		{
		  max_nld += GRABSZ;
		  nld = realloc(nld, sizeof (struct nld) * max_nld);
		}
	      nld[n_nld].nld_ints = put_name(&nld[n_nld].nld_name, name, nlen);
	      nld[n_nld].nld_nlen = nlen;
	      nld[n_nld].nld_u = 0;	/* user defined */
	      nld[n_nld].nld_sinfo = addr;
	      nld[n_nld].nld_ipsd = ipsd;
	      nld[n_nld].nld_lnum = psd[INDEX_PSD(nld[n_nld].nld_ipsd)].psd_entry;
	      nld[n_nld].nld_type = SYM_LT_EXPORTED_ENTRY;
	      nld[n_nld].nld_enlen = 0;
	      nld[n_nld].nld_ename = 0;
	      n_nld++;
#if 0
	      nld[n_nld].nld_ints = put_name(&nld[n_nld].nld_name, name, nlen);
	      nld[n_nld].nld_nlen = nlen;
	      nld[n_nld].nld_u = 0;	/* user defined */
	      nld[n_nld].nld_sinfo = addr;
	      nld[n_nld].nld_sinfo.si_offset += 0600;	/* WWChack */
	      nld[n_nld].nld_ipsd = ipsd;
	      nld[n_nld].nld_lnum = psd[INDEX_PSD(nld[n_nld].nld_ipsd)].psd_entry;
	      nld[n_nld].nld_type = SYM_LT_END_PROLOGUE;
	      nld[n_nld].nld_enlen = 0;
	      nld[n_nld].nld_ename = 0;
	      n_nld++;
#endif
	    }
	}
      pdtent++;			/* skip structure */
      pdtent = (struct pdtent *)(((long *) pdtent) + NWORDS(nlen)); /* skip name */
    }
}

write_dpd_table(dpd, buf, file)
     struct dpd_tbl *dpd;
     long *buf;
     FILE *file;
{
  int oldpos, newpos, oldsdl;
  struct pdtent *pdtent, *pdtentend;

  oldpos = newpos = dpd->hdl+dpd->bdl;
  oldsdl = dpd->sdl;

  pdtent = (struct pdtent *)&buf[dpd->hdl+dpd->bdl];
  pdtentend = (struct pdtent *)&buf[dpd->hdl+dpd->bdl+dpd->sdl];

  while (pdtent < pdtentend)
    {
      char *name = ((char *)pdtent)+sizeof(struct pdtent);
      int nlen = pdtent->pdtenl;
      
      if (strncmp(name, "LN$", 3)) /* skip compiler gened */
	{
	  if (oldpos != newpos)
	    memcpy ((char *)&buf[newpos], (char *)&buf[oldpos],
		    (NWORDS(sizeof(struct pdtent)) + NWORDS(nlen)) * sizeof(long));

	  newpos = newpos + NWORDS(sizeof(struct pdtent)) + NWORDS(nlen);
	}
      pdtent++;			/* skip structure */
      pdtent = (struct pdtent *)(((long *) pdtent) + NWORDS(nlen)); /* skip name */
      oldpos = oldpos + NWORDS(sizeof(struct pdtent)) + NWORDS(nlen);
    }

  dpd->sdl = newpos-(dpd->hdl+dpd->bdl);
  dpd->d.tbl_len = dpd->hdl+dpd->bdl+dpd->sdl+dpd->srl;

  fwrite(&buf[0], sizeof(long), dpd->hdl, file);
  fwrite(&buf[dpd->hdl], sizeof(long), dpd->bdl, file);
  fwrite(&buf[dpd->hdl+dpd->bdl], sizeof(long), dpd->sdl, file);
  fwrite(&buf[dpd->hdl+dpd->bdl+oldsdl], sizeof(long), dpd->srl, file);
}

write_mth(file, asmfile)
     FILE *file;
     FILE *asmfile;
{
  struct mth mth;
  int offset;


  mth.mth_tt = 011;
  mth.mth_lang = 4;
  mth.mth_nlen = strlen (module_name);
  mth.mth_slen = strlen (source_file_name);
  mth.mth_hlen = NWORDS(sizeof (struct mth)) - 1 + NWORDS(mth.mth_nlen)
		  + NWORDS(mth.mth_slen);
  mth.mth_tv = 2;		/* table version */
  mth.mth_csq = 3;		/* call sequence */
  mth.mth_ov = 0;		/* set: variable descriptors are ordered */
  mth.mth_ol = 1;		/* set: label descriptors are ordered */
  mth.mth_so = 1;		/* set: scalar optimization */
  mth.mth_av = 0;		/* set: automatic vectorization */
  mth.mth_dce = 1;		/* set: dead code elimination */
  mth.mth_dbs = 0;		/* set: data breakpoint support */
  mth.mth_fts = 0;		/* set: flow trace support */
  mth.mth_sl = 1;		/* set: label generated for every statement */

  mth.mth_vdslen = VDS_INDEX(1)-1; /* variable descriptor length */
  mth.mth_nldlen = NLD_INDEX(1)-1; /* name label desc. length */
  mth.mth_uldlen = ULD_INDEX(1)-1; /* unnamed label desc. length */
  mth.mth_psdlen = PSD_INDEX(1)-1; /* scope descriptor length */
  mth.mth_fdslen = VDS_INDEX(1)-1; /* field  descriptor length */
  mth.mth_rbl = RBS_INDEX(1)-1;	/*  referenced block desc. len */
  mth.mth_asl = NWORDS(sizeof (struct alts)); /* alt. storage desc block len */
  mth.mth_ddl = 0;

  offset = mth.mth_hlen+1;

  mth.mth_lcbt = 0;		/* common block */
  if (mth.mth_lcbt)
    mth.mth_icbt = offset;
  else
    mth.mth_icbt = 0;

  offset += mth.mth_lcbt;


  mth.mth_lrbs = RBS_INDEX(n_rbs)-1; /* referenced block */
  if (mth.mth_lrbs)
    mth.mth_irbs = offset;
  else
    mth.mth_irbs = 0;

  offset += mth.mth_lrbs;

  mth.mth_lnts = n_name;	/* name table block */
  if (mth.mth_lnts)
    mth.mth_ints = offset;
  else
    mth.mth_ints = 0;

  offset += mth.mth_lnts;


  mth.mth_lpsd = PSD_INDEX(n_psd)-1; /* scope directory */
  if (mth.mth_lpsd)
    mth.mth_ipsd = offset;
  else
    mth.mth_ipsd = 0;

  offset += mth.mth_lpsd;

  mth.mth_ltds = NWORDS(sizeof(default_tds));		/* type section */
  if (mth.mth_ltds)
    mth.mth_itds = offset;
  else
    mth.mth_itds = 0;

  offset += mth.mth_ltds;

  mth.mth_lfds = 0;		/* field_section */
  if (mth.mth_lfds)
    mth.mth_ifds = offset;
  else
    mth.mth_ifds = 0;

  offset += mth.mth_lfds;


  mth.mth_lvds = VDS_INDEX(n_vds)-1; /* variable section */
  if (mth.mth_lvds)
    mth.mth_ivds = offset;
  else
    mth.mth_ivds = 0;

  offset += mth.mth_lvds;

  mth.mth_lnld = NLD_INDEX(n_nld)-1; /* named label section */
  if (mth.mth_lnld)
    mth.mth_inld = offset;
  else
    mth.mth_inld = 0;

  offset += mth.mth_lnld;


  mth.mth_luld = ULD_INDEX(n_uld)-1; /* unnamed label section */
  if (mth.mth_luld)
    mth.mth_iuld = offset;
  else
    mth.mth_iuld = 0;

  offset += mth.mth_luld;

  mth.mth_llcs = 0;		/* long constant */
  if (mth.mth_llcs)
    mth.mth_ilcs = offset;
  else
    mth.mth_ilcs = 0;

  offset += mth.mth_llcs;

  mth.mth_lalts = 0;		/* alt storage */
  if (mth.mth_lalts)
    mth.mth_ialts = offset;
  else
    mth.mth_ialts = 0;

  offset += mth.mth_lalts;

  mth.mth_llds = 0;		/* soruce or listing descriptor section */
  if (mth.mth_llds)
    mth.mth_ilds = offset;
  else
    mth.mth_ilds = 0;

  offset += mth.mth_llds;

  mth.mth_tlen = offset-1;

  strncpy(mth.mth_pname, "AC", 8);	 /* compiler name */
  strncpy(mth.mth_plevel, "2.6beta", 8);    /* WWChack compiler version */
  strncpy(mth.mth_pdate, "12/19/94", 8); /* WWChack compiler date */
  {
    struct stat buf;
    struct tm *tp;

    if (fstat (fileno(asmfile), &buf)<0)
      {
	fprintf(stderr, "can't stat asmfile\n");
	exit(1);
      }
    if (!(tp = localtime(&buf.st_ctime)))
      {
	fprintf(stderr, "can't get localtime\n");
	exit(1);
      }

    sprintf(mth.mth_cdate, "%02d/%02d/%02d", tp->tm_mon+1, tp->tm_mday, tp->tm_year);
    sprintf(mth.mth_ctime, "%02d:%02d:%02d", tp->tm_hour, tp->tm_min, tp->tm_sec);
  }

  mth.mth_sref = 1;		/* source reference flag */
  mth.mth_sline = 1;		/* line number in source of first mod line */
  mth.mth_slave = 0;		/*  */
  mth.mth_dex = 0;		/*  */
  mth.mth_w15 = 0;		/* unused word 15 */


#ifdef DEBUG_DUMP
  fprintf(stderr,"new mth\n");
  print_mth (&mth);
  print_psd(psd, n_psd);
  print_rbs(rbs, n_rbs);
  print_tds(default_tds, mth.mth_ltds);
  print_vds(vds, n_vds);
  print_nld(nld, n_nld);
  print_uld(&uld[1], n_uld-1);
#endif

  /* header */
  fwrite (&mth, sizeof (long), NWORDS(sizeof (struct mth))-1  , file); 
  fwrite (module_name, sizeof (long), NWORDS (mth.mth_nlen), file);
  fwrite (source_file_name, sizeof (long), NWORDS (mth.mth_slen), file);

  /* referenced blocks */
  fwrite (rbs, sizeof (long), mth.mth_lrbs, file);

  /* name block */
  fwrite (name_buf, sizeof (long), n_name, file);

  /* proceedure/scope directory blocks */
  fwrite (psd, sizeof (long), mth.mth_lpsd, file);

  /* type section */
  fwrite (default_tds, sizeof (long), mth.mth_ltds, file);

  /* variable description section */
  fwrite (vds, sizeof (long), mth.mth_lvds, file);

  /* named label section */
  fwrite (nld, sizeof (long), mth.mth_lnld, file);

  /* unnamed labels, skip first */
  fwrite (&uld[1], sizeof (long), mth.mth_luld, file);
}


int put_name(buf, src, count)
     char *buf, *src;
     int count;
{
  int ret;
  strncpy (buf, src, 8);	/* put away first word in buffer */

  if (count <= 8)
    return 0;			/* all done */

  src += 8;
  count -= 8;
  ret = n_name+1;

  if (n_name + (count/8) >= max_name)
    {
      max_name += GRABSZ;
      name_buf = (long *) realloc (name_buf, max_name * sizeof (long));
    }

  while (count > 0)
    {
      strncpy ((char *)(&name_buf[n_name++]), src, 8);
      src += 8;
      count -= 8;
    }
  return ret;
}

int find_rlb(offset)
     int offset;
{
  int i;
  
  for (i=0; i < n_rbs; i++)
    if (rbs[i].dummy == offset)
      return RBS_INDEX(i);
  return 0;
}

int find_psd(addr)
     union sinfo addr;
{
  int i;

  for (i=n_psd-1; i >= 0; i--)	/* look backwards so we get the most restrictive */
    if (psd[i].psd_eaddr.si_irbs == addr.si_irbs
	&& psd[i].psd_eaddr.si_offset <= addr.si_offset
	&& psd[i].psd_climit.si_offset >= addr.si_offset)
      return PSD_INDEX(i);
  return 0;
}

parse_asm_file (fp)
     FILE *fp;
{
  char *p;
  int current_line_num = 0;
  int current_psd = 0;

  while (1)
    {
      char buf[LINESZ];
      if (!fgets(buf, LINESZ, fp))
	{
	  psd[current_psd].psd_exit = current_line_num;
	  psd[current_psd].psd_climit.si_sc = 63; /* address to be set later */
	  psd[current_psd].psd_climit.si_offset = current_line_num; 
	  break;
	}

      /* punt a trailing \n if there */
      if (p = strrchr(buf,'\n'))
	*p = '\000';

      /* check for a "loc" line */
      {
	int depth, line, sym;
	if (sscanf(buf, " ;loc %d, %d, %d", &depth, &line, &sym) == 3)
	  {
#ifdef DEBUG_PARSE
	    fprintf(stderr, "found loc LN%d: %d\n", sym, line);
#endif
	    if (sym >= max_uld)
	      {
		max_uld = sym + GRABSZ;
		uld = (struct uld *) realloc (uld, max_uld * sizeof(struct uld));

	      }
	    uld[sym].uld_lnum = line;
	    current_line_num = line;
	    if (sym > n_uld)
	      n_uld = sym;
	    continue;
	  }
      }

      /* check for a "begin" line */
      {
	int val;
	
	if (sscanf(buf, " ;.begin %d\n", &val) == 1)
	  {
#ifdef DEBUG_PARSE
	    fprintf (stderr, "found begin %d\n", val);
#endif
	    if (n_psd >= max_psd)
	      {
		max_psd += GRABSZ;
		psd = realloc(psd, sizeof (struct psd) * max_psd);
	      }
	    psd[n_psd].psd_nlen = strlen (buf);
	    psd[n_psd].psd_ints = put_name(psd[n_psd].psd_name, buf,
					   psd[n_psd].psd_nlen);
	    insert_psd (n_psd, current_psd);
	    psd[n_psd].psd_entry = current_line_num;
	    psd[n_psd].psd_decl = current_line_num;
	    psd[n_psd].psd_eaddr.si_sc = 63; /* address to be set later */
	    psd[n_psd].psd_eaddr.si_offset = current_line_num; /* 0 */
	    psd[n_psd].psd_icbt = 0;
	    psd[n_psd].psd_itds = 0;
	    psd[n_psd].psd_ivds = VDS_INDEX(n_vds);
	    psd[n_psd].psd_inld = NLD_INDEX(n_nld);
	    psd[n_psd].psd_iuld = ULD_INDEX(n_uld);
	    current_psd = n_psd;
	    n_psd++;
	    continue;
	  }
      }
      
      /* check for a "end" line */
      {
	int val;
	
	if (sscanf(buf, " ;.bend %d\n", &val) == 1)
	  {
#ifdef DEBUG_PARSE
	    fprintf (stderr, "found bend %d\n", val);
#endif
	    if (n_psd >= max_psd)
	      {
		max_psd += GRABSZ;
		psd = realloc(psd, sizeof (struct psd) * max_psd);
	      }
	    psd[current_psd].psd_climit.si_sc = 63; /* address to be set later */
	    psd[current_psd].psd_climit.si_offset = current_line_num;
	    psd[current_psd].psd_exit = current_line_num;
	    psd[current_psd].psd_cvds = n_vds - INDEX_VDS(psd[current_psd].psd_ivds);
	    psd[current_psd].psd_cnld = n_nld - INDEX_NLD(psd[current_psd].psd_inld);
	    psd[current_psd].psd_culd = n_uld - INDEX_ULD(psd[current_psd].psd_iuld);
	    current_psd = INDEX_PSD(psd[current_psd].psd_parent);
	    continue;
	  }
      }

      /* check for a "variable definitions" line */
      {
	char name[LINESZ];
	char val[LINESZ];
	int int_val;
	int sc, type;
	
	if (sscanf(buf, " ;.def %s .val %s .scl %d; .type 0x%x; .endef\n",
		   name, val, &sc, &type) == 4)
	  {
#ifdef DEBUG_PARSE
	    fprintf (stderr, "found variable defn:%s val:%s sc:%d type:0x%x\n",
		     name, val, sc, type);
#endif
	    switch ((type & N_TMASK) >> 4)
	      {
	      case DT_NON:
		if (sc != C_EXT && sc != C_AUTO && sc != C_ARG)
		  continue;

		if (n_vds >= max_vds)
		  {
		    max_vds += GRABSZ;
		    vds = realloc(vds, sizeof (struct vds) * max_vds);
		  }
		vds[n_vds].vds_nlen = strlen (name);
		if (name[vds[n_vds].vds_nlen-1] == ';')
		  {
		    name[vds[n_vds].vds_nlen-1] = 0;
		    vds[n_vds].vds_nlen--;	
		  }
		vds[n_vds].vds_ints = put_name(vds[n_vds].vds_name, name,
					       vds[n_vds].vds_nlen);
		vds[n_vds].vds_ipsd = PSD_INDEX(current_psd);
		vds[n_vds].vds_itds = TDS_INDEX(type & N_BTMASK);
		vds[n_vds].vds_r = 0;
		vds[n_vds].vds_e = 0;
		vds[n_vds].vds_a = 0; /* assign flag */
		vds[n_vds].vds_lhs = 0; /* LHS flag */
		vds[n_vds].vds_d = 1; /* data loaded flag */
		vds[n_vds].vds_p = 0; /* passed flag */
		vds[n_vds].vds_rhs = 2; /* RHS flag */
		vds[n_vds].vds_pe = 0; /* pe resident flag */
		vds[n_vds].vds_i = 0; /* arg intent */
		vds[n_vds].vds_pu = 0; /* public module */
		vds[n_vds].vds_pr = 0; /* private module */
		vds[n_vds].vds_dv = 0; /* fortran90 dope vector */
		vds[n_vds].vds_refvar = 0; /* c++ reference variable */

		switch (sc) {
		case C_EXT:
		  vds[n_vds].vds_ic = SYM_IC_EXPORTED;
		  vds[n_vds].vds_sinfo.si_sc = SYM_SC_MEMORY;
		  vds[n_vds].vds_sinfo.si_irbs = 0;
		  vds[n_vds].vds_sinfo.si_offset = 0;
		  break;

		case C_ARG:
		  int_val = atoi(val);
		  vds[n_vds].vds_ic = SYM_IC_BY_VALUE;
		  vds[n_vds].vds_sinfo.si_sc = SYM_SC_FXR_OFFSET;
		  vds[n_vds].vds_sinfo.si_rnum = 249; /* WWChack? (this is what cc does? */
		  vds[n_vds].vds_sinfo.si_offset = int_val*8;
		  break;

		case C_AUTO:
		  int_val = atoi(val);
		  vds[n_vds].vds_ic = SYM_IC_LOCAL;
		  vds[n_vds].vds_sinfo.si_sc = SYM_SC_STACK;
		  vds[n_vds].vds_sinfo.si_irbs = 0;
		  /* in the following line, we multipy by 8 because
		     si_offset is in bits */
		  vds[n_vds].vds_sinfo.si_offset = int_val*8;
		  break;
		}
		n_vds++;
		break;
	      case DT_PTR:
	      case DT_FCN:
	      case DT_ARY:
		break;
	      }
	    continue;
	  }
      }


	  /* check for a "real" file name */
     if (!strncmp(";SOURCE_FILE:", buf, 13))
	{
	  source_file_name = malloc (strlen(buf)-12);
	  strcpy (source_file_name, &buf[13]);
#ifdef DEBUG_PARSE
	  fprintf(stderr, "found sourcefile %s\n", source_file_name);
#endif
	  continue;
	}
      if (!strncmp("	.ident ", buf, 8))
	{
#ifdef DEBUG_PARSE
	  fprintf(stderr, "found module %s\n", module_name);
#endif
	  module_name = malloc (strlen(buf)-7);
	  strcpy (module_name, &buf[8]);
	  if (n_psd >= max_psd)
	    {
	      max_psd += GRABSZ;
	      psd = realloc(psd, sizeof (struct psd) * max_psd);
	    }
	  psd[n_psd].psd_nlen = strlen (module_name);
	  psd[n_psd].psd_ints = put_name(psd[n_psd].psd_name, module_name,
					 psd[n_psd].psd_nlen);
	  psd[n_psd].psd_parent = 0;
	  psd[n_psd].psd_child = 0;
	  psd[n_psd].psd_sibling = 0;
	  psd[n_psd].psd_entry = 0;
	  psd[n_psd].psd_decl = 1;
	  psd[n_psd].psd_eaddr.si_sc = 0;
	  psd[n_psd].psd_eaddr.si_irbs = 0;
	  psd[n_psd].psd_eaddr.si_offset = 0;
	  psd[n_psd].psd_climit.si_sc = 0;
	  psd[n_psd].psd_climit.si_irbs = 0;
	  psd[n_psd].psd_climit.si_offset = 0;
	  psd[n_psd].psd_icbt = 0;
	  psd[n_psd].psd_itds = 1;
	  psd[n_psd].psd_ctds = (sizeof (default_tds)/sizeof (struct tds));
	  psd[n_psd].psd_ivds = 0;
	  psd[n_psd].psd_inld = 0;
	  psd[n_psd].psd_iuld = 0;
	  current_psd = n_psd;
	  n_psd++;
	  continue;
	}
    }
}

insert_psd (child, parent)
     int child, parent;
{
  psd[child].psd_parent = PSD_INDEX(parent);
  if (!psd[parent].psd_child)
    {
      psd[parent].psd_child = PSD_INDEX(child);
    }
  else		/* insert in sibling list of child */
    {
      int t = INDEX_PSD(psd[parent].psd_child);
      while (psd[t].psd_sibling)
	t =  INDEX_PSD(psd[t].psd_sibling);
      psd[t].psd_sibling = PSD_INDEX(child);
    }
  psd[child].psd_sibling = 0;
  psd[child].psd_child = 0;
}
