/* USMID @(#)mpp/relo.h	81.1	09/27/93 14:43:36 */

/* COPYRIGHT CRAY RESEARCH, INC.
 * UNPUBLISHED -- ALL RIGHTS RESERVED UNDER
 * THE COPYRIGHT LAWS OF THE UNITED STATES */

#ifndef _RELO_H
#define _RELO_H

/*
 *  This is the include file which contains the C language
 *  table definitions for relocatable output from assemblers
 *  and compilers.  All tables are variable sized.  The first
 *  word of each table contains a table type code, a table
 *  length, and an optional block index field.
 *
 *  These tables are described in further detail in relo(5)
 *  and in the document entitled "CRAY Object File Format".
 */

/*
 *  A common field declaration, used within structures
 *  throughout this include file, is defined as follows.
 *  This field declaration is "undefined" at the end of this
 *  include file.
 */

#define	BYTES_IN_ASCII	8

#define FIELD unsigned long

typedef union {
  long	l;
  char	c [BYTES_IN_ASCII];
} ASCII;

#define NON_CRAY_WORD_SIZE 32

#ifdef CRAY	/* Define macros to manipulate large bit fields */

#  define LG(f_type, f_name, f_size) \
f_type	f_name	: f_size

#else

#  define LG(f_type, f_name, f_size) \
f_type		: f_size - NON_CRAY_WORD_SIZE; \
f_type	f_name	: NON_CRAY_WORD_SIZE

#endif

/*
 *  The following definitions represent table type codes (found
 *  in the first word of all tables).
 *
 *  The type codes for the LHT, BLD, SYM, CMB, DMT, and GNT
 *  tables are defined within this include file.  However,
 *  these tables are described elsewhere.
 */
typedef enum {
  LHT_TYPE = 001,    /* Library (build) header table            */
  BLD_TYPE = 002,    /* Build directory table                   */
  MTT_TYPE = 010,    /* Module termination table                */
  XRL_TYPE = 014,    /* Extended relocation table               */
  REL_TYPE,          /* Relocation table                        */
  TXT_TYPE,          /* Text table                              */
  PDT_TYPE,          /* Program Descriptor table                */
  GNT_TYPE = 027,    /* Global symbol table                     */
  DMT_TYPE = 040,    /* Distributed module termination table    */
  DFI_TYPE = 042,    /* Distributed field initialization table  */
  DEX_TYPE,          /* Distributed Expression table            */
  DPD_TYPE = 047     /* Distributed Program Descriptor table    */
  } table_type;
  
/*
 *  The following table header structure is used for the
 *  PDT_TYPE, TXT_TYPE, REL_TYPE, XRL_TYPE, MTT_TYPE, LHT_TYPE,
 *  and BLD_TYPE tables.
 */
struct tbl_hdr {
  FIELD		hdr_type	:  7;	/* Table type			*/
  FIELD		hdr_undef	:  9;  /* (Unused, reserved by CRI)    */
  FIELD		hdr_bi		: 16;	/* Block index (optional)	*/
  FIELD		hdr_len		: 32;	/* Table length			*/
};

/*
 *  The following structure defines fields within a program
 *  description table (PDT) header.
 */
struct pdttabl {
	LG(FIELD, pdthdr   ,64);	/* Use tbl_hdr structure here	*/
	FIELD pdthdsz      : 32;	/* Word size of header area	*/
	FIELD pdtblsz      : 32;	/* Word size of block area	*/
	FIELD pdtensz      : 32;	/* Word size of entry area	*/
	FIELD pdtexsz      : 32;	/* Word size of external area	*/
	FIELD pdtmdy       : 64;	/* MM/DD/YY - this compilation	*/
	FIELD pdthms       : 64;	/* HH:MM:SS - this compilation	*/
	FIELD pdtcmpid     : 64;	/* Generating product name	*/
	FIELD pdtcmpvr     : 64;	/* Generating product version	*/
	FIELD pdtosvr      : 64;	/* Host OS version		*/
	FIELD pdtudt       : 64;	/* UNICOS time stamp (date)	*/
	FIELD              :  1;	/* (Unused, reserved by CRI)	*/
	FIELD pdtfe        :  1;	/* Fatal error flag (1==true)	*/
	FIELD pdtbd        :  1;	/* Block data module (1==true)	*/
	FIELD pdtmpa       :  1;	/* Module passed address flag	*/
	FIELD pdtdc        :  1;	/* Dual case names flag(1==true)*/
	FIELD pdtusr       :  8;	/* (Unused, reserved for user)	*/
	FIELD              :  1;	/* (Unused, reserved by CRI)	*/
	FIELD pdtmf        :  2;	/* Module flag for Fortran 90:  */
	                        	/*   0 = independent            */
	                        	/*   1 = first of a module      */
	                        	/*   2 = in a module set        */
	                        	/*   3 = last of a module       */
	FIELD pdtfnl       :  8;	/* Char count in file name	*/
	FIELD pdtmnl       :  8;	/* Char count in module name	*/
	FIELD pdtss        : 32;	/* Stack size requirement	*/
	LG(FIELD, pdtuqnm  ,64);	/* Unique ID for module name    */
	/*
	 *  All remaining fields found in the PDT table header,
	 *  if any, are dynamic in that they are defined
	 *  depending on the pdthdsz, pdtfnl, and pdtmnl fields
	 *  which are defined above and the total size (10
	 *  decimal words) of the above fields.
	 *
	 *  The module name field, if it exists (pdtmnl != 0),
	 *  is the very last field found in the PDT table
	 *  header.  The module name is stored in the minimum
	 *  number of words required to store pdtmnl characters.
	 *  This name is left justified and zero filled within
	 *  this field.  The module name field is exactly
	 *  ((pdtmnl + 7) / 8) words long (where "/" yields the
	 *  quotient as the whole, not fractional, portion of
	 *  the division).
	 *
	 *  The file name field, if it exists (pdtfnl != 0),
	 *  directly precedes the module name field.  If the
	 *  module name field does not exist, the file name
	 *  field is the very last field found in the PDT
	 *  table header.  The file name is stored in the
	 *  minimum number of words required to store pdtfnl
	 *  characters.  This name is left justified and zero
	 *  filled within this field.  The file name field is
	 *  exactly ((pdtfnl + 7) / 8) words long (where "/"
	 *  is defined as above).
	 *
	 *  If (10 + ((pdtmnl + 7) / 8) + ((pdtfnl + 7) / 8))
	 *  is not equal to pdthdsz then there are one or more
	 *  fields which immediately follow the pdtuqnm field
	 *  and precede the file name field, if it exists, and
	 *  the module name field, if it exists.  The first
	 *  word contains the word length of each field within
	 *  this area.  Following this word are the fields
	 *  themselves.  This area is described in further
	 *  detail in relo(4F).
	 */
};

typedef struct pdttabl pdt_t;

typedef struct {
  FIELD          : 32;
  FIELD mul      : 16;
  FIELD cmtl     :  8;
  FIELD mtl      :  8;
} pdtex_t;

/*
 *  The next structure is the portion of the PDT which describes
 *  a single block.
 */
struct pdtblck {                /* word 0                       */
	FIELD pdtbkcb    :  1;	/* Common block flag (1==true)	*/
	FIELD pdtbkl     :  3;	/* Block location		*/
	FIELD pdtbkc     :  3;	/* Block contents		*/
	FIELD pdtbkt     :  3;	/* Block type			*/
	FIELD pdtbal     :  3;	/* Block align flag		*/
	FIELD pdtbef     :  1;	/* Block entry flag		*/
        FIELD pddux      : 14;
	FIELD            : 26;	/* (Unused, reserved by CRI)	*/
        FIELD pddum      :  10;  /* Number of wds in d block     */
                                /* word 1                       */
	FIELD pdtbusr    : 40;	/* (Unused, reserved for user)	*/
	FIELD pdtbknl    :  8;	/* Char count in block name	*/
        FIELD pdgroup    : 16;  /* Block group number           */
                                /* word 2                       */
	FIELD pdtbkln    : 64;	/* Block length (words)		*/
	/*
	 *  The block name follows in the minimum number of
	 *  words required to store pdtbknl characters.  This
	 *  name is left justified and zero filled within this
	 *  field.
	 */
};

/*
 *  Constants for the pdtbkl (block location) field
 */
#define BKL_CM     0		/* Common memory, also		*/
				/*    CRAY-2 FGP instr memory	*/
#define BKL_LM     1		/* CRAY-2 local memory, also	*/
				/*    CRAY-2 FGP local memory	*/
#define BKL_LMA    2		/* CRAY-2 local memory absolute	*/
				/*    (unused by UNICOS)	*/
#define BKL_EM     3		/* CRAY X-MP extended memory	*/

#define BLK_AX     4            /* CRAY X-MP auxiliary memory   */
/*
 *  Constants for the pdtbkc (block contents) field
 */
#define BKC_UNK    0		/* Unknown			*/
#define BKC_IX     1		/* Instructions	only		*/
#define BKC_DT     2		/* Data only			*/
#define BKC_BS     3		/* Data only with no text (bss) */
#define BKC_CN     4		/* Constants only		*/
#define BKC_ZD     5		/* Data only with no text	*/
				/*    (zero fill assumed)	*/

/*
 *  Constants for the pdtbkt (block type) field
 */
#define BKT_CM     0		/* Regular common block		*/
#define BKT_TCM    2		/* Task common block		*/
#define BKT_DBF    4		/* Dynamic block		*/

/*
 *  Constants for the pdtbal (block align) flag
 */
#define BAL_NA     0		/* No alignment			*/
#define BAL_AL     1		/* Align to instruction buffer	*/
				/*    boundary			*/

/*
 *  The next structure is the portion of the PDT which describes
 *  a single entry point.  Entry points with pdtebi==0 are
 *  constants.
 */
struct pdtent {
	FIELD pdtepf     :  1;	/* Primary entry flag (1==true)	*/
	FIELD            : 15;	/* (Unused, reserved by CRI)	*/
	FIELD pdtenl     :  8;	/* Char count in entry name	*/
	FIELD pdterm     :  3;	/* Suggested relocation mode	*/
	FIELD            : 13;	/* (Unused, reserved by CRI)	*/
	FIELD pdteusr    :  8;	/* (Unused, reserved for user)	*/
	FIELD pdtebi     : 16;	/* Block index			*/
        FIELD pdtepv     : 64;	/* Entry point (signed) value 	*/
	/*
	 *  The entry name follows in the minimum number of
	 *  words required to store pdtenl characters.  This
	 *  name is left justified and zero filled within this
	 *  field.
	 */
};

/*
 *  The next structure is the portion of the PDT which describes
 *  a single external.
 */
struct pdtext  {
	FIELD pdtxmn     :  1;	/* Module specification		*/
	FIELD            :  1;	/* (Unused, reserved by CRI)	*/
	FIELD pdtxnl     :  6;	/* Char count in external name	*/
	FIELD pdtxsf     :  3;	/* Soft external (1==true)	*/
	FIELD pdtxct     :  2;	/* Call tree information	*/
	FIELD pdtxpa     :  1;	/* External passed as argument	*/
	FIELD            : 42;	/* (Unused, reserved by CRI)	*/
	FIELD pdtxusr    :  8;	/* (Unused, reserved for user)	*/
	/*
	 *  The external name follows in the minimum number of
	 *  words required to store pdtxnl characters.  This
	 *  name is left justified and zero filled within this
	 *  field.
	 */
};

/*
 *  Constants for pdtxct (call tree information) field
 */
#define XCT_EXT    0		/* Regular external		*/
#define XCT_THDO   1		/* External is task head	*/
#define XCT_THDS   2		/* External is both task head	*/
				/*    and subprogram		*/

/*
 *  A text table (TXT) contains hdr_len words.  One text table
 *  contains zero or more subtables.  Each subtable has a two
 *  word header.  Fields within this two word header are defined
 *  as follows.
 */
struct txtitem {
	long  txtinc     : 17;	/* Incr between dups (signed)	*/
	LG(FIELD, txtsba ,38);	/* Starting bit address		*/
	FIELD txtnbl     :  6;	/* Number of bits in last word	*/
	FIELD txtusr1    :  3;	/* (Unused, reserved for user)	*/
	FIELD txtusr2    :  5;	/* (Unused, reserved for user)	*/
	FIELD            :  8;	/* (Unused, reserved by CRI)	*/
	FIELD txtndp     : 19;	/* Number of duplications	*/
	FIELD txtntw     : 32;	/* Number of text words		*/
	/*
	 *  Text words immediately follow the two word header,
	 *  the number of which is specified by txtntw.
	 */
};

/*
 *  A relocation table (REL) contains hdr_len words.  One
 *  relocation table contains zero or more single word
 *  relocation entries.  Fields within these single word entries
 *  are defined as follows.
 */
struct relitem {
	FIELD relrt      :  1;	/* Relocation type		*/
	FIELD relri      : 16;	/* Relocation index		*/
	LG(FIELD, relrba ,38);	/* Rightmost bit address	*/
	FIELD relfl      :  6;	/* Field length in bits to	*/
				/*    relocate			*/
	FIELD relrm      :  3;	/* Relocation mode		*/
};

/*
 *  An extended relocation table (XRL) contains hdr_len words.
 *  One extended relocation table contains zero or more double
 *  word relocation entries.  Fields within these double word
 *  entries are defined as follows.
 */
struct xrlitem {
	FIELD xrlrt      :  1;	/* Relocation type		*/
	FIELD xrlri      : 16;	/* Relocation index		*/
	FIELD xrlusr     :  8;	/* (Unused, reserved for user)	*/
	FIELD            : 23;	/* (Unused, reserved by CRI)	*/
	FIELD xrlsp      :  3;	/* Special relocation		*/
	FIELD xrln       :  1;	/* Sign before relocation	*/
	FIELD xrlsr      :  3;	/* Sign specification of result	*/
	FIELD xrlfl      :  6;	/* Field length in bits to	*/
				/*    relocate			*/
        FIELD xrlrm      :  3;	/* Relocation mode		*/
	FIELD            : 26;	/* (Unused, reserved by CRI)	*/
	LG(FIELD, xrlrba ,38);	/* Rightmost bit address	*/
};

/*
 *  Constants for the relrt and xrlrt (relocation type) fields
 */
#define RT_BLK     0		/* Block entry			*/
#define RT_EXT     1		/* External entry		*/

/*
 *  Constants for the xrlsp (special relocation) field
 */
#define SP_NONE    0		/* No special relocation	*/
#define SP_RVHF    1		/* Reversed halves relocation	*/

/*
 *  Constants for the xrlsr (sign specification of result) field
 */
#define SR_NONE    0		/* Sign does not matter		*/
#define SR_POS     1		/* Field must be positive	*/
#define SR_NEG     2		/* Field must be negative	*/
#define SR_EXT     3		/* Field is sign extended	*/

/*
 *  Constants for the pdterm, relrm, and xrlrm (relocation mode)
 *  fields
 */
#define RM_WORD    0		/* Word address			*/
#define RM_HALF    1		/* Half word address		*/
#define RM_PARC    2		/* Parcel address		*/
#define RM_BYTE    3		/* Byte address			*/
#define RM_BIT     6		/* Bit address			*/
#define RM_ENTR    7		/* Relocation mode from		*/
				/*    associated entry (pdterm);*/
				/*    legal on external		*/
				/*    references only.		*/

/*
 *  A module termination table (MTT) contains hdr_len words.
 *  Within UNICOS, this table is composed of two words, a header
 *  and a checksum word.  Checksums are not generated by UNICOS
 *  products.
 *
 *  This table is represented by the following structure.
 */
struct mtttabl {
	LG(FIELD, mtthdr     ,64);	/* Use tbl_hdr structure here	*/
	LG(FIELD, mttcksm    ,64);	/* Checksum			*/
};



/**********************************************************************/


typedef struct dtbl_hdr {
#						define TBL_TYPE	7
  table_type	tbl_type	: TBL_TYPE;	/* Table type                         */
#						define FHD_LEN	8
  FIELD		fhd_len		: FHD_LEN;	/* Fixed header length                */
#
  FIELD				: 1;		/* Unused; reserved by CRI            */
#						define DTBL_LEN	48
  LG(FIELD,	tbl_len	,	 DTBL_LEN);	/* Number of words in the table       */
						/* including the header words         */
} dtbl_hdr_t;

/*--------------------------------------------------------------------*/
/*                                                                    */
/*                 Distributed Program Descriptor                     */
/*                                                                    */
/*--------------------------------------------------------------------*/

typedef enum {
	lang_Unknown = 0,
	lang_Assembly = 1,
	lang_Fortran_77 = 4,
	lang_Standard_C = 5
} dpdlang;

typedef struct dpd_tbl {
  dtbl_hdr_t	d		;		/* Table header structure (dpd = 047) */
  /* -------------------------------------------------------------------------------- */
#						define DPDFE		1
  FIELD		fe		: DPDFE;	/* Fatal error flag                   */
#						define DPDUAB		1
  FIELD		uab		: DPDUAB;	/* Unconditionally Allocate Blocks    */
#						define DPDMPA	1
  FIELD		mpa		: DPDMPA;	/* Module Passed Address flag         */
#						define DPDDC	1
  FIELD		dc		: DPDDC;	/* Dual case names permitted          */
#						define DPDVER	8
  FIELD		ver		: DPDVER;	/* Version number of tables           */
#						define DPDUSR	8
  FIELD		usr		: DPDUSR;	/* Reserved for user                  */
#
  FIELD				: 20;		/* Not used                           */
#						define DPDLANG	4
  dpdlang	lang		: DPDLANG;	/* Source language                    */
#						define DPDNPE	20
  FIELD		npe		: DPDNPE;	/* If nonzero, N$PES already bound    */
  /* -------------------------------------------------------------------------------- */
#						define DPDHDL	32
  FIELD		hdl		: DPDHDL;	/* Word count for the header          */
#						define DPDBDL	32
  FIELD		bdl		: DPDBDL;	/* Word count of block descriptor     */
  /* -------------------------------------------------------------------------------- */
#						define DPDSDL	32
  FIELD		sdl		: DPDSDL;	/* Word count of symbolic definition  */
#						define DPDSRL	32
  FIELD		srl		: DPDSRL;	/* Word count of symbolic reference   */
  /* -------------------------------------------------------------------------------- */
  FIELD 	mdy	:64	;		/* ASCII date of compilation          */
  FIELD		hms     :64	;		/* ASCII time of compilation          */
  FIELD		cmpid	:64	;		/* ASCII generating product name      */
  FIELD		cmpvr	:64	;		/* ASCII generating product version   */
  FIELD		osvr    :64	;		/* ASCII operating system version     */
#						define DPDUDT	64
  LG(FIELD,	udt		, DPDUDT);	/* Compilation time from date(2)      */
  /* -------------------------------------------------------------------------------- */
#						define DPDMNI	32
  FIELD		mni		: DPDMNI;	/* module name index                  */
#						define DPDMNL	32
  FIELD		mnl		: DPDMNL;	/* module name length                 */
  /* -------------------------------------------------------------------------------- */
#						define DPDFNI	32
  FIELD		fni		: DPDFNI;	/* file name index                    */
#						define DPDFNL	32
  FIELD		fnl		: DPDFNL;	/* file name length                   */
  /* -------------------------------------------------------------------------------- */
#						define DPDTADI	32
  FIELD		tadi		: DPDTADI;	/* target architecture descriptor     */
#						define DPDTADL	32
  FIELD		tadl		: DPDTADL;	/* length                             */
  /* -------------------------------------------------------------------------------- */
#						define DPDCFI	32
  FIELD		cfi		: DPDCFI;	/* commentary field index             */
#						define DPDCFL	32
  FIELD		cfl		: DPDCFL;	/* commentary field length            */
  /* -------------------------------------------------------------------------------- */
#						define DPDMUFI	32
  FIELD		mufi		: DPDMUFI;	/* module use field index             */
#						define DPDMUFL	32
  FIELD		mufl		: DPDMUFL;	/* module use field length            */
  /* -------------------------------------------------------------------------------- */
  LG(FIELD,filler02		, 64);		/* reserved for CRI.                  */
  LG(FIELD,filler03		, 64);		/* reserved for CRI.                  */
  LG(FIELD,filler04		, 64);		/* reserved for CRI.                  */
} dpd_t;

/*
 * Define a union for machine characteristics
 */

#define	MC_LENGTH	3	/* The length of the characteristics section in words */

typedef union {
  struct {
    LG(FIELD,	length		, 64);	/* The number of machine targeting words */
    ASCII	primary_name	;	/* ASCII primary machine name            */
    LG(FIELD,filler05		, 49);	/* Unused; reserved for CRI              */
    FIELD	bmm		: 1;	/* Bit matrix multiply required          */
    FIELD	a32		: 1;	/* 32-bit A registers required           */
    FIELD	cor		: 1;	/* Control operand range interrupts      */
    FIELD	cls		: 1;	/* Cluster registers required            */
    FIELD	str		: 1;	/* Status register required              */
    FIELD	nbd		: 1;	/* Cannot use bidirectional memory       */
    FIELD	hpm		: 1;	/* Hardware Performance monitor required */
    FIELD	avl		: 1;	/* Additional vector logical             */
    FIELD	nvr		: 1;	/* Cannot use vector recursion           */
    FIELD	vr		: 1;	/* Vector recursion required             */
    FIELD	rvl		: 1;	/* Must read vector length               */
    FIELD	pc		: 1;	/* Programmable clock required           */
    FIELD	cig		: 1;	/* Compressed index, gather/scatter      */
    FIELD	ema		: 1;	/* EMA hardware is required              */
    FIELD	vp		: 1;	/* Vector pop count required             */
  } cray1;
  struct {
    LG(FIELD,	length		, 64);	/* The number of machine targeting words */
    ASCII	primary_name	;	/* ASCII primary machine name            */
    LG(FIELD,filler06		, 61);	/* Unused; reserved for CRI              */
    FIELD	cmr		: 1;	/* Bidirectional memory is required      */
    FIELD	tgt		: 1;	/* Tailgating is required                */
    FIELD	far		: 1;	/* Fetch/increment hardware required     */
  } cray2;
  struct {
    FIELD	dpdtanpe	: 20;	/* Number of processors bound to module  */
    LG(FIELD,filler07		, 10);	/* Unused; reserved for CRI              */
    FIELD       dpdtaosp	: 1;	/* Requires T3D OS partition             */
    FIELD       dpdtahwp	: 1;	/* Requires T3D hardware partition       */
    FIELD       dpdtaeuk	: 8;	/* Number of eureka barriers requested   */
    FIELD       dpdtabar	: 8;	/* Number of barriers requested          */
    FIELD       dpdtafl		: 16;	/* The number of machine targeting words */
    ASCII       dpdtapmn	;	/* ASCII primary machine name            */
    LG(FIELD,filler08		, 64);	/* Not currently used                    */
  } t3d;
} mc_t;

/*--------------------------------------------------------------------*/
/*                                                                    */
/*                Block Descriptor Entry                              */
/*                                                                    */
/*--------------------------------------------------------------------*/

typedef enum { off, on } flag;

/* Values of the dpdb.ksc field:                                      */

typedef enum {ksc_local = 0, ksc_common = 1} dpdbksc;

/* Values of the dpdb.kl field:                                       */

typedef enum {kl_common = 0,
              kl_stack  = 4,
              kl_heap   = 6} dpdbkl;

/* Values of the dpdb.kc field:                                       */

typedef enum {kc_unknown = 0,
                kc_code = 1,
                kc_data = 2,
                kc_bss  = 3,
                kc_constants = 4,
                kc_bssz = 5} dpdbkc;

/* Values of the dpdb.kt field:                                       */

typedef enum { kt_regular = 0, kt_distributed = 1 } dpdbkt;

/* Values of the dpdb.al field:                                       */

typedef enum { al_unaligned = 0, 
                 al_aligned = 1,
                 al_dex_aligned = 2} dpdbal;

typedef struct {
#						define DPDBKSC	1
  dpdbksc	ksc		: DPDBKSC;	/* Common block flag                  */
#						define DPDBKL	3
  dpdbkl	kl		: DPDBKL;	/* Block location                     */
#						define DPDBKC	3
  dpdbkc	kc		: DPDBKC;	/* Block contents                     */
#						define DPDBKT	3
  dpdbkt	kt		: DPDBKT;	/* Block type                         */
#						define DPDBAL	3
  dpdbal	al		: DPDBAL;	/* Block alignment                    */
#						define DPDBKLNI	1
  flag		klni		: DPDBKLNI;	/* kln field is an expression         */
#
  LG(FIELD,filler09		, 40);		/* Unused; reserved for CRI	      */
#						define DPDBKDL	10
  FIELD		kdl		: DPDBKDL;	/* Number of words in distribution    */
                                		/* descriptor section                 */
  /* -------------------------------------------------------------------------------- */
#						define DPDBALV	32
  FIELD		alv		: DPDBALV;	/* block alignment value              */
#						define DPDBUSR	8
  FIELD		usr		: DPDBUSR;	/* reserved for user                  */
#						define DPDBKNL	8
  FIELD		knl		: DPDBKNL;	/* block name length                  */ 
#						define DPDBKGP	16
  FIELD		kgp		: DPDBKGP;	/* block group number                 */
  /* -------------------------------------------------------------------------------- */
#						define DPDBKLN	64
  LG(FIELD,	kln		, DPDBKLN);	/* Size of block in bits              */
} dpdb_t;

/*--------------------------------------------------------------------*/
/*                                                                    */
/*           Block Distribution Descriptors                           */
/*                                                                    */
/* dpdb.kdl words follow as several three-word entries                */
/*                                                                    */
/*--------------------------------------------------------------------*/

typedef struct {
#						define DPDDPEI	1
  flag		pei		: DPDDPEI;	/* The pe field is an expression      */
#						define DPDDBKI	1
  flag		bki		: DPDDBKI;	/* The bk field is an expression      */
#						define DPDDCYI	1
  flag		cyi		: DPDDCYI;	/* The cy field is an expression      */
#
  FIELD				: 29;
#						define DPDDPE	32
  FIELD		pe		: DPDDPE;	/* Number of processors over which    */
						/* this dimension should be dis-      */
						/* tributed.                          */
  /* -------------------------------------------------------------------------------- */
#						define DPDDBK	64
  LG(FIELD,	bk		, DPDDBK);	/* Initial block size allocated to    */
						/* each processor.                    */
#						define DPDDCY	64
  LG(FIELD,	cy		, DPDDCY);	/* Number of cycles                   */
} dpdd_t;

/*--------------------------------------------------------------------*/
/*                                                                    */
/*              SYMBOLIC DEFINITIONS                                  */
/*                                                                    */
/*--------------------------------------------------------------------*/

/* Values of the dpdsd.t field:					      */

typedef enum {
	t_constant	= 0,
	t_mem_location	= 4,
	t_init_entry	= 5
} dpdsdt;

typedef struct dpdsd {
#						define DPDSDT	4
  dpdsdt	t		: DPDSDT;	/* Type of definition		      */
#						define DPDSDVI	1
  flag		vi		: DPDSDVI;	/* The v field is an expression       */
#
  FIELD				: 11;		/* Unused; reserved for CRI	      */
#						define DPDSDNL	8
  FIELD		nl		: DPDSDNL;	/* Name length                        */
#						define DPDSDUSR	8
  FIELD		usr		: DPDSDUSR;	/* Unused: reserved for user          */
#						define DPDSDBI	32
  FIELD		bi		: DPDSDBI;	/* Block index                        */
  /* -------------------------------------------------------------------------------- */
#						define DPDSDV	64
  LG(FIELD,	v		, DPDSDV);	/* value                              */
  /* This is followed by the entry point name (as many words as are   */
  /* required.                                                        */
} dpdsd_t;

/*--------------------------------------------------------------------*/
/*                                                                    */
/*                     SYMBOLIC REFERENCES                            */
/*                                                                    */
/*--------------------------------------------------------------------*/

/* Values of the dpdsr.n field:					      */

typedef enum {
	n_uncondit	= 0,
	n_condit	= 1,
	n_uncondit_sp	= 8,
	n_condit_sp	= 9
} dpdsrn;

typedef struct dpdsr {
#						define DPDSRNL	8
  FIELD		nl		: DPDSRNL;	/* Number of characters in external   */
#						define DPDSRN	4
  dpdsrn	n		: DPDSRN;	/* Reference "necessity"              */
#						define DPDSRPA	1
  flag		pa		: DPDSRPA;	/* Passed as an argument from here    */
#
  LG(FIELD,filler10		, 43);		/* ---------------------------------- */
#						define DPDSRUSR	8
  FIELD		usr		: DPDSRUSR;	/* Reserved for user                  */
  /* -------------------------------------------------------------------------------- */
  /* This is followed by the entry point name (as many words as are   */
  /* required.                                                        */
} dpdsr_t;

/*--------------------------------------------------------------------*/
/*                                                                    */
/*            Distributed Expression Table                            */
/*                                                                    */
/*--------------------------------------------------------------------*/

typedef struct dex_tbl {
  dtbl_hdr_t	d			;	/* Table header structure (dex = 043) */
  /* -------------------------------------------------------------------------------- */
#						define DEXTRES	1
  flag		tres		: DEXTRES;	/* All expressions have been resolved */
#						define DEXIEL	8
  FIELD		iel		: DEXIEL;	/* Length of an initial expression    */
						/* entry.  Must be > 0.               */
  FIELD				: 15;		/* ---------------------------------- */
#						define DEXISL	40
  LG(FIELD,	isl		, DEXISL);	/* Total length of initial entry      */
						/* section.                           */
} dex_t;

/* DEX Expression entry header */

typedef enum {
	long_life_expression	= 0,
	short_term_expression	= 1,
	long_term_macro		= 8,
	short_term_macro	= 9
} dexetyp;

typedef enum { 
	rt_integer	= 0,
	rt_IEEE_fp	= 1,
	rt_IEEE_complex	= 2
} dexert;

typedef enum {
	rs_signed	= 0,
	rs_unsigned	= 1
} dexers;

typedef struct dexe_tbl {
#						define DEXETYP	4
  dexetyp	typ		: DEXETYP;	/* Expression type                   */
#						define DEXERES	1
  flag		res		: DEXERES;	/* Already resolved                  */
#						define DEXERT	4
  dexert	rt		: DEXERT;	/* Result type of expression         */
#						define DEXERP	8
  FIELD		rp		: DEXERP;	/* Result Precision                  */
#						define DEXERS	1
  dexers	rs		: DEXERS;	/* Result signed/unsigned            */
#
  FIELD				: 6;	/*                                   */
#						define DEXHDL	4
  FIELD		hdl		: DEXHDL;	/* word length of header structure   */
#						define DEXEL		16
  FIELD		l		: DEXEL;	/* length                            */
#						define DEXOPRL	20
  FIELD		oprl		: DEXOPRL;	/* size of operator section          */
  /* -------------------------------------------------------------------------------- */
  FIELD				: 16;		/*                                   */
#						define DEXREO	48
  LG(FIELD,	reo		, DEXREO);	/* Word offset of residual part      */
} dexe_t;

/* Evaluation Mode Descriptor */

typedef enum {
	dexetint	= 0,
	dexetfp		= 1,
	dexetcxr	= 2,
	dexetcxi	= 3
} EMD_type;

typedef enum {
	dexep8		= 0,
	dexep16		= 1,
	dexep32		= 2,
	dexep64		= 3,
	dexep128	= 4
} EMD_precision;

typedef enum {
	dexesn		= 0,
	dexeusn		= 1
} EMD_sign;

typedef enum {
	dexernear	= 0,
	dexertozr	= 1,
	dexernegi	= 2,
	dexerposi	= 3
} EMD_rounding;

typedef struct emd {
  EMD_type	OT : 2;
  EMD_precision	OP : 3;
  EMD_sign	OS : 1;
  EMD_rounding	OR : 2;
} EMD_t;

typedef enum {
	EMD_int_8_sn		=	0,
	EMD_int_8_usn		=	4,
	EMD_int_16_sn		=	8,
	EMD_int_16_usn		=	12,
	EMD_int_32_sn		=	16,
	EMD_int_32_usn		=	20,
	EMD_int_64_sn		=	24,
	EMD_int_64_usn		=	28,
	EMD_int_128_sn		=	32,
	EMD_int_128_usn		=	36,
	EMD_fp_16_sn_near	=	72,
	EMD_fp_16_sn_tozr	=	73,
	EMD_fp_16_sn_negi	=	74,
	EMD_fp_16_sn_posi	=	75,
	EMD_fp_32_sn_near	=	80,
	EMD_fp_32_sn_tozr	=	81,
	EMD_fp_32_sn_negi	=	82,
	EMD_fp_32_sn_posi	=	83,
	EMD_fp_64_sn_near	=	88,
	EMD_fp_64_sn_tozr	=	89,
	EMD_fp_64_sn_negi	=	90,
	EMD_fp_64_sn_posi	=	91,
	EMD_fp_128_sn_near	=	96,
	EMD_fp_128_sn_tozr	=	97,
	EMD_fp_128_sn_negi	=	98,
	EMD_fp_128_sn_posi	=	99,
	EMD_cxr_16_sn_near	=	136,
	EMD_cxr_16_sn_tozr	=	137,
	EMD_cxr_16_sn_negi	=	138,
	EMD_cxr_16_sn_posi	=	139,
	EMD_cxr_32_sn_near	=	144,
	EMD_cxr_32_sn_tozr	=	145,
	EMD_cxr_32_sn_negi	=	146,
	EMD_cxr_32_sn_posi	=	147,
	EMD_cxr_64_sn_near	=	152,
	EMD_cxr_64_sn_tozr	=	153,
	EMD_cxr_64_sn_negi	=	154,
	EMD_cxr_64_sn_posi	=	155,
	EMD_cxr_128_sn_near	=	160,
	EMD_cxr_128_sn_tozr	=	161,
	EMD_cxr_128_sn_negi	=	162,
	EMD_cxr_128_sn_posi	=	163,
	EMD_cxi_16_sn_near	=	200,
	EMD_cxi_16_sn_tozr	=	201,
	EMD_cxi_16_sn_negi	=	202,
	EMD_cxi_16_sn_posi	=	203,
	EMD_cxi_32_sn_near	=	208,
	EMD_cxi_32_sn_tozr	=	209,
	EMD_cxi_32_sn_negi	=	210,
	EMD_cxi_32_sn_posi	=	211,
	EMD_cxi_64_sn_near	=	216,
	EMD_cxi_64_sn_tozr	=	217,
	EMD_cxi_64_sn_negi	=	218,
	EMD_cxi_64_sn_posi	=	219,
	EMD_cxi_128_sn_near	=	224,
	EMD_cxi_128_sn_tozr	=	225,
	EMD_cxi_128_sn_negi	=	226,
	EMD_cxi_128_sn_posi	=	227
} EMD_types;

/*--------------------------------------------------------------------*/
/*                                                                    */
/*                       Expression Operations                        */
/*                                                                    */
/*--------------------------------------------------------------------*/

#define	DEX_OPERATOR_SIZE	8     /* Size of DEX operator in bits */
#define DEX_LABEL_MASK		0xf

#define GET_DEX_LABEL_NUM(OPR)	((OPR) & DEX_LABEL_MASK)

#define DEX_OPR_TYPE_BITS	8
typedef enum {
  Control,
  Entry,
  Mathematical,
  Logical,
  Relational,
  Comparison
} t_dex_opr_types;

typedef enum ops {
  dexcerr = 0,     /* Control ERRor                                   */
  dexcrtop,        /* Control ReTurn OP                               */
  dexcrtn,         /* Control ReTurn N                                */
  dexctdm = 4,     /* Control Toggle Debugging Mode                   */
  dexcss2 = 6,     /* Control Save State 2                            */
  dexcss1,         /* Control Save State 1                            */
  dexccop,         /* Control Convert OPerand                         */
  dexcsem = 12,    /* Control Select Evaluation Mode                  */
  dexccem,         /* Control Convert to Evaluation Mode              */
  dexcso0,         /* Control Save Operand to memory 0                */
  dexcso1,         /* Control Save Operand to memory 1                */
  dexccpy,         /* Control CoPY                                    */
  dexcswp,         /* Control SWaP                                    */
  dexcadp = 20,    /* Control Advance Data Pointer                    */
  dexcsdp8,        /* Control Set Data Pointer to 8-bit value         */
  dexcsdp16,       /* Control Set Data Pointer to 16-bit value        */
  dexcsdp24,       /* Control Set Data Pointer to 24-bit value        */
  dexcmac,         /* Control MAcro Call                              */
  dexcmar,         /* Control MAcro Return                            */
  dexcpos = 27,    /* Control Previous Operand Stream                 */
  dexcsaz0 = 32,  dexcsaz1,   dexcsaz2,   dexcsaz3,   dexcsaz4,
  dexcsaz5,       dexcsaz6,   dexcsaz7,   dexcsaz8,   dexcsaz9,
  dexcsaz10,      dexcsaz11,  dexcsaz12,  dexcsaz13,  dexcsaz14,
  dexcsaz15,       /* Control Skip ahead if zero                      */
  dexcsanz0 = 48, dexcsanz1,  dexcsanz2,  dexcsanz3,  dexcsanz4,
  dexcsanz5,      dexcsanz6,  dexcsanz7,  dexcsanz8,  dexcsanz9,
  dexcsanz10,     dexcsanz11, dexcsanz12, dexcsanz13, dexcsanz14,
  dexcsanz15,       /* Control Skip ahead if non-zero                 */
  dexcsa0 = 64,   dexcsa1,    dexcsa2,    dexcsa3,    dexcsa4,
  dexcsa5,        dexcsa6,    dexcsa7,    dexcsa8,    dexcsa9,
  dexcsa10,       dexcsa11,   dexcsa12,   dexcsa13,   dexcsa14,
  dexcsa15,       /* Control Skip                                     */
  dexcst0 = 80,   dexcst1,    dexcst2,    dexcst3,    dexcst4,
  dexcst5,        dexcst6,    dexcst7,    dexcst8,    dexcst9,
  dexcst10,       dexcst11,   dexcst12,   dexcst13,   dexcst14,
  dexcst15,       /* Control Target                                   */
  dexn8b,         /* Enter 8-bit                                      */
  dexn8be,        /* Enter 8-bit sign-extended                        */
  dexn16b,        /* Enter 16-bit                                     */
  dexn16be,       /* Enter 16-bit sign-extended                       */
  dexn24b,        /* Enter 24-bit                                     */
  dexn24be,       /* Enter 24-bit sign-extended                       */
  dexn32b,        /* Enter 32-bit                                     */
  dexn32be,       /* Enter 32-bit sign-extended                       */
  dexn64b,        /* Enter 64-bit                                     */
  dexn128b,       /* Enter 128-bit                                    */
  dexnso0 = 106,  /* Enter Saved Operand in memory 0                  */
  dexnso1,        /* Enter Saved Operand in memory 1                  */
  dexnzer = 108,  /* Enter zero                                       */
  dexnone,        /* Enter one                                        */
  dexnnpe = 112,  /* Enter N$PES                                      */
  dexnba  = 114,  /* Enter block bit address                          */
  dexnbl,         /* Enter block bit length                           */
  dexnsrv = 118,  /* Enter symbolic reference value                   */
  dexnsrt,        /* Enter symbolic reference's definition type       */
  dexndex = 122,  /* Enter DEX                                        */
  dexmadd = 128,  /*                                                  */
  dexmsub,
  dexmmul,
  dexmdiv,
  dexmrem,
  dexmpow,
  dexmlg2,
  dexlshl = 160,
  dexlshr,
  dexlashr,
  dexlmskl = 166,
  dexlmskr,
  dexland = 170,
  dexlxor,
  dexlor,
  dexlnot,
  dexrlt = 192,
  dexrle,
  dexrgt,
  dexrge,
  dexreq = 198,
  dexrne,
  dexsmin = 208,
  dexsmax,
  dexLAST
} ops_t;

/*--------------------------------------------------------------------*/
/*                                                                    */
/*            Distributed Field Initialization                        */
/*                                                                    */
/*--------------------------------------------------------------------*/

typedef enum {
	ivs_supplied	= 0,
	ivs_derived	= 8
} dfiivs;

typedef struct {
  dtbl_hdr_t	d			;	/* Table header structure (dfi = 042) */
  /* -------------------------------------------------------------------------------- */
  FIELD				: 32;		/* Unused; Reserved for CRI           */
#						define DFIBI		32
  FIELD		bi		: DFIBI;	/* Block index                        */
} dfi_hdr_t;

/*            Initialization Entry                                    */

typedef struct dfi_tbl {
#						define DFIIVS	4
  FIELD		ivs		: DFIIVS;	/* Initialization value source        */
#						define DFIINCI	1
  flag		inci		: DFIINCI;	/* "Increment is index"               */
#						define DFIREPI	1
  flag		repi		: DFIREPI;	/* "Replication is index"             */
#						define DFISBOI	1
  flag		sboi		: DFISBOI;	/* "Starting bit offset is index"     */
#						define DFIUSR	8
  FIELD		usr		: DFIUSR;	/*                                    */
#
  FIELD				: 17;		/*                                    */
#						define DFIFL		32
  FIELD		fl		: DFIFL;	/* Field length                       */
  /* -------------------------------------------------------------------------------- */
#						define DFIINC	32
  FIELD		inc		: DFIINC;	/* Increment for replications         */
#						define DFIREP	32
  FIELD		rep		: DFIREP;	/* Replication count                  */
  /* -------------------------------------------------------------------------------- */
#						define DFISBO	64
  LG(FIELD,	sbo		, DFISBO);	/* Starting bit offset                */
} dfi_t;

/*--------------------------------------------------------------------*/
/*                                                                    */
/*                 Distributed Module Termination                     */
/*                                                                    */
/*--------------------------------------------------------------------*/

typedef struct dmt_tbl {
  dtbl_hdr_t	d			;	/* Table header structure (dmt = 040) */
  /* -------------------------------------------------------------------------------- */
#						define DMTCKSM	64
  LG(FIELD,	cksm		, DMTCKSM);	/* Optional checksum                  */
} dmt_t;

/*
 *  Undefine the common field declarations
 */
#undef FIELD

#endif /* !_RELO_H */


