/* USMID @(#)include/symbol.h	81.0	09/26/93 15:32:56 */

/* COPYRIGHT CRAY RESEARCH, INC. 
 * UNPUBLISHED -- ALL RIGHTS RESERVED UNDER 
 * THE COPYRIGHT LAWS OF THE UNITED STATES */

#ifndef _SYMBOL_H
#define _SYMBOL_H

/*
 * Symbolic information as included in a relocatable or executable
 * file. The first word of each symbol table is almost a table-header
 * structure as defined in relo.h where the constants for table-type
 * are also defined.
 */

#define	Align	int :0


/*
 * global name table header
 */
struct	gnt {
	unsigned	gnt_type:7;		/* table type GNT_TYPE */
	unsigned	gnt_nl:8;		/* name length */
	unsigned	:13;
	unsigned	gnt_vers:4;		/* table version */
	unsigned	gnt_wc:32;		/* table length */
	char		gnt_name[1];		/* overlay name */
};

/*
 * global symbol entry
 */
struct gse {
	unsigned	gse_bflag:1;		/* block flag */
	unsigned	gse_cb:1;		/* common block flag */
	unsigned	gse_bl:3;		/* block location */
	unsigned	gse_bc:3;		/* block contents */
	unsigned	gse_bt:3;		/* block type */
	unsigned	gse_nl:8;		/* name length */
	unsigned	gse_bnl:8;		/* block name length */
	unsigned	:5;
	unsigned	gse_blen:32;		/* length of block (words) */
	unsigned long	gse_value;		/* value */
	char		gse_name[1];		/* name */
};


/*
 * module table header
 */
struct mth {
	unsigned mth_tt:7;	/* table type SYM_TYPE */
	unsigned mth_lang:9;	/* language */
	unsigned mth_hlen:8;	/* header length */
	unsigned mth_nlen:8;	/* module name length */
	unsigned mth_tlen:32;	/* table length */

	/*
	 * descriptor lengths are for a single element
	 * of that type
	 */
	unsigned mth_tv:8;	/* table version */
	unsigned mth_csq:8;	/* call sequence */
	unsigned mth_vdslen:8;	/* variable descriptor length */
	unsigned mth_nldlen:8;	/* name label descriptor length */
	unsigned mth_uldlen:8;	/* unnamed label descriptor length */
	unsigned mth_psdlen:8;	/* scope descriptor length */
	unsigned mth_fdslen:8;	/* field  descriptor length */

	unsigned mth_ov:1;	/* set: variable descriptors are ordered */
	unsigned mth_ol:1;	/* set: label descriptors are ordered */
	unsigned mth_ddl:6;	/* distributed descriptor length */
	Align;

	/*
	 * indexes are 1's based and are offset from the beginning
	 * of the table entry.  A zero means none exist.
	 * ifoo -> index to foo
	 * lfoo -> length of foo
	 */
	unsigned mth_icbt:32;	/* common block */
	unsigned mth_lcbt:32;

	unsigned mth_irbs:32;	/* referenced block */
	unsigned mth_lrbs:32;

	unsigned mth_ints:32;	/* name table */
	unsigned mth_lnts:32;

	unsigned mth_ipsd:32;	/* scope directory */
	unsigned mth_lpsd:32;

	unsigned mth_itds:32;	/* type section */
	unsigned mth_ltds:32;

	unsigned mth_ifds:32;	/* field section */
	unsigned mth_lfds:32;

	unsigned mth_ivds:32;	/* variable section */
	unsigned mth_lvds:32;

	unsigned mth_inld:32;	/* named label section */
	unsigned mth_lnld:32;

	unsigned mth_iuld:32;	/* unnamed label section */
	unsigned mth_luld:32;

	unsigned mth_ialts:32;	/* alternate storage information */
	unsigned mth_lalts:32;

	unsigned mth_ilcs:32;	/* long constant */
	unsigned mth_llcs:32;

	unsigned mth_ilds:32;	/* source or listing descriptor section */
	unsigned mth_llds:32;

	long mth_w15;		/* reserved */

	char mth_pname[8];	/* compiler name */
	char mth_plevel[8];	/* compiler version */
	char mth_pdate[8];	/* compiler date */
	char mth_cdate[8];	/* date module was compiled */
	char mth_ctime[8];	/* time module was compiled */

	unsigned mth_so:1;	/* set: scalar optimization */
	unsigned mth_av:1;	/* set: automatic vectorization */
	unsigned mth_dce:1;	/* set: dead code elimination */
	unsigned mth_dbs:1;	/* set: data breakpoint support */
	unsigned mth_fts:1;	/* set: flow trace support */
	unsigned mth_sl:1;	/* set: label generated for every statement */
	unsigned mth_unused0:22;
	unsigned mth_slave:1;
	unsigned mth_sref:1;
	unsigned mth_slen:10;
	unsigned mth_sline:24;
	Align;

	unsigned mth_rbl:8;
	unsigned mth_asl:8;
	Align;

	unsigned long mth_dex;
	unsigned long mth_unused1;
	unsigned long mth_unused2;
        unsigned long mth_unused3;

	char mth_mname[1];
	/* rest of module name */
};

/*
 * pascal module table header fields
 */
struct mth_pascal {
	unsigned mth_inloc:32;	/* nonlocal section index */
	unsigned mth_lnloc:32;	/* nonlocal section length */
};
/*
 * CAL module table header fields
 */
struct mth_cal {
	unsigned mth_iqtab:32;	/* descriptor section index */
	unsigned mth_lqtab:32;	/* descriptor section length */

	unsigned mth_iqvar:32;	/* variable section index */
	unsigned mth_lqvar:32;	/* variable section length */

	unsigned mth_iqlab:32;	/* label section index */
	unsigned mth_lqlab:32;	/* label section length */
};

/*
 * common block table header
 */
struct cbt {
	unsigned cbt_tt:7;	/* table type CMB_TYPE */
	unsigned cbt_lang:9;	/* language */
	unsigned cbt_hlen:8;	/* length of header */
	unsigned cbt_nlen:8;	/* name length */
	unsigned cbt_tlen:32;	/* table length */

	long cbt_chksum;	/* checksum */

	unsigned cbt_type:16;	/* common block type */
				/* subfields of cbt_type duplicate information
				 * from the gse structure in the gnt-table
				 */
	Align;

	unsigned cbt_tv:8;	/* table version */
	unsigned cbt_vdslen:8;	/* variable descriptor length */
	unsigned cbt_fdslen:8;	/* filed descriptor length */
	unsigned cbt_ov:1;	/* set: variable descriptors ordered */
	Align;

	/*
	 * indexes are 1's based and are offset from the beginning
	 * of the table entry.  A zero means none exist.
	 * ifoo -> index to foo
	 * lfoo -> length of foo
	 */
	unsigned cbt_irbs:32;	/* referenced block section */
	unsigned cbt_lrbs:32;
	unsigned cbt_ints:32;	/* name table section */
	unsigned cbt_lnts:32;
	unsigned cbt_itds:32;	/* type section */
	unsigned cbt_ltds:32;
	unsigned cbt_ifds:32;	/* field descriptor section */
	unsigned cbt_lfds:32;
	unsigned cbt_ivds:32;	/* variable descriptor section */
	unsigned cbt_lvds:32;
	unsigned cbt_ialts:32;	/* alternate storage section */
	unsigned cbt_lalts:32;

	long cbt_w11;		/* reserved */
	long cbt_w12;		/* reserved */

	char cbt_name[1];	/* common block name */
	/* rest of name */
};

/*
 * CAL common block table fields
 */
struct cbt_cal {
	unsigned cbt_iqtab:32;	/* descriptor section */
	unsigned cbt_lqtab:32;
	unsigned cbt_iqvar:32;	/* variable section */
	unsigned cbt_lqvar:32;
};

/* 
 * referenced block section
 */
struct rbs {
	char rbs_name[8];	/* first 8 chars of name */
	unsigned rbs_nlen:8;	/* name length */
	unsigned rbs_ints:24;	/* index to rest of name */
	int :6;			/* reserved */
	unsigned rbs_cb:1;	/* set: common block */
	unsigned rbs_bl:3;	/* block location */
	unsigned rbs_bc:3;	/* block contents */
	unsigned rbs_bt:3;	/* block type */
	unsigned rbs_iblk:16;	/* block index */
	unsigned dummy:64;
};

/*
 * type descriptor section
 */
struct tds {
	char tds_name[8];	/* first 8 chars of type name */
	unsigned tds_nlen:8;	/* length of type name */
	unsigned tds_ints:24;	/* index to rest of name */
	unsigned tds_ipsd:20;	/* procedure directory index */
	unsigned tds_df:5;	/* display format code */
	Align;
	unsigned tds_lent:18;	/* number of words in this type descriptor */
	unsigned tds_bitlen:38;	/* type variable length in bits */
	unsigned tds_tc:8;	/* type class */
};

/*
 * array descriptor
 */
struct tds_ad {
	unsigned ad_nd:8;	/* number of dimensions */
	unsigned ad_o:1;	/* set: elements in column order */
	unsigned ad_p:1;	/* set: array is packed */
	unsigned ad_a:1;	/* set: packed array left adjusted */
	unsigned ad_ielem:20;	/* type section element index */
	unsigned ad_isubs:20;	/* type section subscript index */
	Align;

	unsigned ad_lt:3;	/* lower bound type */
	unsigned ad_ut:3;	/* upper bound type */
	/*
	 * next 8 go here
	 */
	Align;
	/*
	 * continue with 9 lower/upper bound types per word
	 */
	/*
	 * When needed, the LD and UD fields are found here.
	 * they are either a constant or index of a variable
	 * descriptor.
	 */
};

/*
 * pointer descriptor
 */
struct tds_pd {
	unsigned pd_idom:20; /* index of domain type in type section */
	Align;
};

/*
 * structured variable descriptor
 */
struct tds_svd {
	unsigned svd_ifield:20;	/* field index */
	unsigned svd_cfield:20;	/* number of fields */
	unsigned svd_type:8;	/* structured variable type */
	Align;
	/* language specific information goes here */
};

/*
 * subrange descriptor
 */
struct tds_sd {
	unsigned sd_irange:20; /* index of range type in type section */
	Align;
	long sd_min;	/* minimum value in subrange */
	long sd_max;	/* maximum value in subrange */
};

/*
 * enumeration descriptor
 */
struct tds_ed {
	unsigned ed_count:20; /* number of enumerated constants */
	Align;
	struct {	
		long ed_cname;	/* first 8 chars of name */
		unsigned ed_cnlen:8; /* length of name */
		unsigned ed_icname:24; /* index of name */
		Align;
		unsigned long ed_value; /* value of enumeration */
	} ed_e[1];
	/*
	 * there are ed_count of these structures,
	 * just keep indexing!
	 */
};

/*
 * set descriptor
 */
struct tds_setd {
	unsigned setd_ibase:20; /* base type index in type section */
	Align;
};

/*
 * pascal types go here.  (incomplete)
 */

/*
 * fortran types go here.  (incomplete)
 */

/*
 * storage information format
 */
union sinfo {
	struct {
		unsigned usi_sc:6;
		unsigned usi_iparent:20;
		unsigned :10;
		unsigned usi_count:8;
		unsigned usi_index:20;
	} si_u0;
	struct {
		unsigned :32;
		unsigned usi_value:32;
	} si_u1;
	struct {
		unsigned :6;
		unsigned usi_rtype:4;
		unsigned usi_rnum:16;
		unsigned :6;
		unsigned usi_treg:16;
		unsigned usi_nreg:16;
	} si_u2;
	struct {
		unsigned :6;
		unsigned usi_irbs:16;
		unsigned :4;
		long usi_offset:38;
	} si_u3;

#define	si_sc		si_u0.usi_sc	/* commented below */
#define	si_count	si_u0.usi_count
#define	si_ialts	si_u0.usi_index
#define	si_ivar		si_u0.usi_index
#define	si_irbs		si_u3.usi_irbs
#define	si_offset	si_u3.usi_offset
#define	si_iparent	si_u0.usi_iparent
#define	si_value	si_u1.usi_value
#define	si_ilcon 	si_u1.usi_value
#define	si_treg		si_u2.usi_treg
#define	si_nreg		si_u2.usi_nreg
#define	si_rtype	si_u2.usi_rtype
#define	si_rnum		si_u2.usi_rnum

	/*
	 * class 0: more than one storage location
		si_count		count of storage descriptors in the
					alternate storage information section
		si_ialts		index of first storage descriptor for
					this object in the alternate storage
					information section
	 * class 1: memory block
	 * class 2: indirect off memory block
		si_irbs			index of load block descriptor in the
					referenced block section
		si_offset		bit offset from base of block
	 * class 3: stack
	 * class 4: indirect off stack
		si_offset		bit offset from base of stack frame
	 * class 5: small constant
		si_value		value of constant
	 * class 6: long constant
		si_ilcon		index of the value of the constant in
					the long constant section
	 * class 7: assigned to fixed register
	 * class 8: indirect off fixed register
		si_treg			type of register
		si_nreg			register number
	 * class 11: fixed in structured type
		si_iparent		index of parent type in type section
		si_offset		bit offset of field
	 * class 12: based variable
		si_ivar			index of the variable descriptor for
					the corresponding pointer variable in
					the variable section
	 * class 13: based descriptor
		si_iparent		index of the descriptor for the parent
					variable in the variable section
		si_offset		bit offset of field
	 * class 14: formal parameter passed by address
	 * class 15: formal parameter passed by value
		si_offset		bit offset of argument word in the
					argument list
	 * class 16: task common relocatable index
		si_irbs			index of load block descriptor in the
					referenced block section
		si_offset		bit offset from index into task common
	 * class 17: external with offset
		si_ilcon		index into the long constant section
					of an offset to the badded to the
					external address
	 * class 18: offset from fixed register
		si_rtype		register type to which base address
					is assigned
		si_rnum			register number
		si_offset		bit offset of variable from base
					address specified in register
	 */
};

/*
 * variable descriptors
 */
struct vds {
	char vds_name[8];		/* first 8 chars of name */
	unsigned vds_nlen:8;		/* length of name */
	unsigned vds_ints:24;		/* index to rest of name */
	unsigned vds_ipsd:20;		/* index to scope */
	Align;
	union sinfo vds_sinfo;		/* information about location */
	unsigned vds_itds:20;		/* index of descriptor in type section */
	unsigned vds_r:1;		/* set: variable was referenced */
	unsigned vds_e:1;		/* set: variable is equivalenced */
	unsigned vds_ic:8;		/* identifier class */
	unsigned vds_a:1;	/* assign flag */
	unsigned vds_lhs:2;	/* LHS flag */
	unsigned vds_d:2;	/* data loaded flag */
	unsigned vds_p:2;	/* passed flag */
	unsigned vds_rhs:2;	/* RHS flag */
	unsigned vds_pe:1;	/* pe resident flag */
	unsigned vds_i:2;	/* arg intent */
	unsigned vds_pu:1;	/* public module */
	unsigned vds_pr:1;	/* private module */
	unsigned vds_dv:2;	/* fortran90 dope vector */
	unsigned vds_refvar:2;	/* c++ reference variable */
	
      };

/*
 * alternate storage descriptor
 */
struct alts {
	union sinfo alts_sino;		/* information about storage location */
	unsigned alts_ivds:20;		/* index of variable descriptor */
	unsigned alts_ialts:20;		/* index of next entry */
	unsigned alts_stav: 4;		/* storage availability */
	Align;
};

/*
 * common block list
 */
struct cbl {
	char cbl_name[8];		/* first 8 chars of name */

	unsigned cbl_nlen:8;		/* length of name */
	unsigned cbl_ints:24;		/* index to rest of name */
	Align;
	long cbl_chksum;		/* checksum */
};

/*
 * procedure/scope directory
 */
struct psd {
	char psd_name[8];		/* first 8 chars of name */
	unsigned psd_nlen:8;		/* length of name */
	unsigned psd_ints:24;		/* index to rest of name */
	unsigned psd_enlen:8;		/* pascal: external name length */
	unsigned psd_ename:24;		/* pascal: external name index */
	union sinfo psd_eaddr;			/* entry address */
	union sinfo psd_climit;		/* "exit" address */
	unsigned psd_parent:32;		/* index into scope directory for parent */
	unsigned psd_sibling:32;	/*  " " " " for next sibling */
	unsigned psd_child:32;		/*  " " " " of first child */
	unsigned psd_decl:24;		/* source line of declaration */
	Align;
	unsigned psd_entry:24;		/* source line of start of code */
	unsigned psd_exit:24;		/* source line of end of code */
	Align;
	unsigned psd_icbt:32;		/* common block list */
	unsigned psd_ccbt:32;
	unsigned psd_itds:32;		/* type descriptors */
	unsigned psd_ctds:32;
	unsigned psd_ivds:32;		/* variables */
	unsigned psd_cvds:32;
	unsigned psd_inld:32;		/* named labels */
	unsigned psd_cnld:32;
	unsigned psd_iuld:32;		/* unnamed labels */
	unsigned psd_culd:32;
};

/*
 * pascal procedure/scope directory
 */
struct psd_pascal {
	unsigned psd_inloc:32;		/* non-local index */
	unsigned psd_cnloc:32;		/* non-local count */
	unsigned psd_pt:4;		/* procedure type */
	unsigned psd_et:4;		/* external type */
	Align;
};

/*
 * named label descriptor section
 */
struct nld {
	char nld_name[8];		/* first 8 chars of name */
	unsigned nld_nlen:8;		/* length of name */
	unsigned nld_ints:24;		/* index to rest of name */
	unsigned nld_ipsd:20;		/* index to scope */
	unsigned nld_u:1;		/* set: user defined */
	Align;
	union sinfo nld_sinfo;		/* storage information */
	unsigned nld_lnum:24;		/* source line number */
	unsigned nld_type:8;		/* label type */
	unsigned nld_enlen:8;		/* pascal: external name length */
	unsigned nld_ename:24;		/* pascal: external name index */
};

/*
 * unnamed label descriptors
 */
struct uld {
	union sinfo uld_addr;		/* relocatable address of label */
	unsigned uld_lnum:24;		/* source line number of label */
	unsigned uld_type:8;		/* label type */
	Align;
};

#undef	Align

/*
 * language field
 */
#define	SYM_LANG_PASCAL		1
#define	SYM_LANG_FORTRAN	2
#define	SYM_LANG_CAL		3
#define	SYM_LANG_C		4

/*
 * block location
 */
#define	SYM_BL_COMMON		0
#define	SYM_BL_LOCAL		1
#define	SYM_BL_CONSTANT		2
#define	SYM_BL_LOCAL_ABSOLUTE	2
#define	SYM_BL_EXTENDED		3

/*
 * block contents
 */
#define	SYM_BC_UNKNOWN		0
#define	SYM_BC_INSTRUCTION	1
#define	SYM_BC_DATA		2
#define	SYM_BC_BSS		3
#define	SYM_BC_CONSTANT		4

/*
 * block types
 */
#define	SYM_BT_COMMON		0
#define	SYM_BT_TASK_COMMON	2
#define	SYM_BT_DYNAMIC_COMMON	4

/*
 * display format/type class
 */
#define	SYM_DF_OCTAL		0
#define	SYM_DF_DECIMAL		1
#define	SYM_DF_UDECIMAL		2
#define	SYM_DF_FLOAT		3
#define	SYM_DF_COMPLEX		4
#define	SYM_DF_LOGICAL1		5
#define	SYM_DF_LOGICAL2		6
#define	SYM_DF_LOGICAL3		7
#define	SYM_DF_WADDR		8
#define	SYM_DF_PADDR		9
#define	SYM_DF_ASCII		10

/*
 * display format
 */
#define	SYM_DF_DOUBLE		20
#define	SYM_DF_BINARY		21
#define	SYM_DF_HEX		22
#define	SYM_DF_INSTRUCTION	23
#define	SYM_DF_SPECIAL_OCTAL	24
#define	SYM_DF_SPECIAL_HEX	25
#define	SYM_DF_SPECIAL_BINARY	26

/*
 * type class
 */
#define	SYM_TC_CFT_STRING	20
#define	SYM_TC_STRING		21
#define	SYM_TC_PASCAL_TEXT	22
#define	SYM_TC_ARRAY		30
#define	SYM_TC_POINTER		31
#define	SYM_TC_STRUCTURE	32
#define	SYM_TC_SUBRANGE		33
#define	SYM_TC_ENUMERATION	34
#define	SYM_TC_SET		35
#define	SYM_TC_PASCAL_FILE	36
#define	SYM_TC_PASCAL_VARIANT_PART	37
#define	SYM_TC_PASCAL_VARIANT	38
#define	SYM_TC_LRLTRAN_VECTOR	39
#define	SYM_TC_LRLTRAN_VECTOR_DESCRIPTOR	40

/*
 * upper and lower bound types of array descriptors
 */
#define	SYM_BT_CONSTANT		0	/* both (upper and lower) */
#define	SYM_BT_VARIABLE		1	/* both */
#define	SYM_BT_EXPRESION	2	/* both */
#define	SYM_BT_ASSUMED		3	/* upper only */
#define	SYM_BT_ZERO		4	/* lower only */
#define	SYM_BT_ONE		5	/* lower only */
#define	SYM_BT_RUNTIME		6	/* upper only */
#define	SYM_BT_EXTENT		7	/* upper only */

/*
 * types
 */
#define	SYM_TYPE_RECORD		1
#define	SYM_TYPE_STRUCTURE	2
#define	SYM_TYPE_UNION		3
#define	SYM_TYPE_PSTRUCTURE	4

/*
 * identifier Classes
 */
#define	SYM_IC_STRUCTURE	0
#define	SYM_IC_LOCAL		1
#define	SYM_IC_EXPORTED		2
#define	SYM_IC_IMPORTED		3
#define	SYM_IC_COMMON		4
#define	SYM_IC_CONSTANT		5
#define	SYM_IC_BY_ADDRESS	6
#define	SYM_IC_BY_VALUE		7
#define	SYM_IC_RESULT		8
#define	SYM_IC_ARRAY_BOUND	9
#define	SYM_IC_TMP		10

/*
 * storage Class
 */
#define	SYM_SC_MULTI		0
#define	SYM_SC_MEMORY		1
#define	SYM_SC_INDR_MEMORY	2
#define	SYM_SC_STACK		3
#define	SYM_SC_INDR_STACK	4
#define	SYM_SC_SMALL_CONSTANT	5
#define	SYM_SC_CONSTANT		6
#define	SYM_SC_REGISTER		7
#define	SYM_SC_INDR_REGISTER	8
#define	SYM_SC_EXTERNAL		9
#define	SYM_SC_UNFIXED		10
#define	SYM_SC_ELEMENT		11
#define	SYM_SC_BASED_VARIABLE	12
#define	SYM_SC_BASED_DESCRIPTOR	13
#define	SYM_SC_BY_ADDRESS	14
#define	SYM_SC_BY_VALUE		15
#define	SYM_SC_TASK_COMMON	16
#define	SYM_SC_EXT_OFFSET	17
#define	SYM_SC_FXR_OFFSET	18

/*
 * register types
 */
#define	SYM_RT_UNKNOWN		0
#define	SYM_RT_A		1
#define	SYM_RT_B		2
#define	SYM_RT_S		3
#define	SYM_RT_T		4
#define	SYM_RT_V		5
#define	SYM_RT_SB		6
#define	SYM_RT_ST		7
#define	SYM_RT_SM		8
#define SYM_RT_R		9
#define SYM_RT_F		10

/*
 * procedure types
 */
#define	SYM_PT_PROGRAM		0
#define	SYM_PT_MODULE		1
#define	SYM_PT_PROCEDURE	2
#define	SYM_PT_FUNCTION		3

/*
 * external types
 */
#define	SYM_ET_LOCAL		0
#define	SYM_ET_IMPORTED		1
#define	SYM_ET_EXPORTED		2
#define	SYM_ET_FORMAL		3

/*
 * label Types
 */
#define	SYM_LT_UNKNOWN		0
#define	SYM_LT_LABEL		1
#define	SYM_LT_EXPORTED_ENTRY	2
#define	SYM_LT_IMPORTED_ENTRY	3
#define	SYM_LT_LOCAL_ENTRY	4
#define	SYM_LT_EXIT		5
#define	SYM_LT_END_PROLOGUE	6
#define	SYM_LT_START_PROLOGUE	7
#define	SYM_LT_START_FOR	11
#define	SYM_LT_TOP_FOR		12
#define	SYM_LT_BOTTOM_FOR	13
#define	SYM_LT_END_FOR		14
#define	SYM_LT_START_WHILE	15
#define	SYM_LT_TOP_WHILE	16
#define	SYM_LT_END_WHILE	17
#define	SYM_LT_START_REPEAT	18
#define	SYM_LT_TOP_REPEAT	19
#define	SYM_LT_END_REPEAT	20
#define	SYM_LT_START_IF		21
#define	SYM_LT_START_THEN	22
#define	SYM_LT_START_ELSE	23
#define	SYM_LT_END_IF		24
#define	SYM_LT_START_OTHER	25

#endif /* !_SYMBOL_H */
