/* Vector extensions to GCC
   
   Copyright (C) 1993, 1994 United States Government as represented by
   the Director, National Security Agency.
   All rights reserved.

This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include "config.h"
#include "rtl.h"
#include "tree.h"

int hard_regno_vector_length = 0;
int current_vector_length = 0;
int default_vector_length = 0;
int current_vector_stride = 0;
int default_vector_stride = 0;

int global_vector_length = 0;

/*
   First some utility functions to extract vector length and other
   things from rtx's and trees, etc. */

#ifdef REGISTER_VECTOR_NAMES
char *reg_vector_names[] = REGISTER_VECTOR_NAMES;
#else
char **reg_vector_names;
#endif

int
rtx_vlength (x)
     rtx x;
{
  switch (GET_CODE (x))
    {
    case REG:
    case MEM:
      return XINT(x, 1);
    case SUBREG:
      return XINT(x, 2);
    case ASHIFT:
    case ROTATE:
    case LSHIFTRT:
    case ASHIFTRT:
    case ROTATERT:
    case SMAX:
    case SMIN:
    case UMAX:
    case UMIN:
    case PLUS:
    case MINUS:
    case MULT:
    case DIV:
    case MOD:
    case UDIV:
    case UMOD:
    case AND:
    case IOR:
    case XOR:
    case NOT:
    case EQ:
    case NE:
    case GE:
    case GT: 
    case LE:
    case LT:
    case GEU:
    case GTU: 
    case LEU:
    case LTU:
    case NEG:
    case CALL:
      return rtx_vlength (XEXP (x,0));
    case SIGN_EXTEND:
    case ZERO_EXTEND:
    case TRUNCATE:
    case FLOAT_EXTEND:
    case FLOAT_TRUNCATE:
    case FLOAT:
    case FIX:
    case UNSIGNED_FLOAT: 
      return rtx_vlength (XEXP (x,0));
    case IF_THEN_ELSE:
      return rtx_vlength (XEXP (x,1));
    case SEQUENCE:
    case PARALLEL:
      return rtx_vlength (XVECEXP (x, 0, 0));
    case USE:
    case CLOBBER:
      return rtx_vlength (XEXP (x,0));
    case QUEUED:
      return rtx_vlength (XEXP (x,3));
    case SET:
      return rtx_vlength (SET_DEST (x));
    case INSN:
      return rtx_vlength (PATTERN(x));
    case EXPR_LIST:
      return rtx_vlength (XEXP (x,0));
    case HIGH:
    case LO_SUM:
      return rtx_vlength (XEXP (x,0));
    case GSYNC:
      return rtx_vlength (XEXP (x,0));
    case ASM_OPERANDS:
      if (ASM_OPERANDS_INPUT_LENGTH (x))
	return rtx_vlength (ASM_OPERANDS_INPUT (x, 0));
      return 0;
    case SYMBOL_REF:
    case LABEL_REF:
    case JUMP_INSN:
    case CALL_INSN:
    case BARRIER:
    case NOTE:
    case CODE_LABEL:
    case CONST_INT:
    case CONST_DOUBLE:
    case CONST:
    case ASM_INPUT:
    case UNSPEC:
    case UNSPEC_VOLATILE:
    case ZERO_EXTRACT:
      return 0;
    default:
      abort();
    }
}

int
rtx_vsize (x, mode)
     rtx x;
     enum machine_mode mode;
{
  int t = rtx_vlength (x);

  if (!mode)
    mode = GET_MODE (x);

  return (t ? t : 1) * GET_MODE_SIZE (mode);
}

int
rtx_vnregs (x, mode)
     rtx x;
     enum machine_mode mode;
{
  int t = rtx_vlength (x);
  
  if (!mode)
    mode = GET_MODE (x);

  return (t ? t : 1) * ((GET_MODE_SIZE (mode) + UNITS_PER_WORD - 1)
			/ UNITS_PER_WORD);
}
     
int
rtx_vstride (x)
     rtx x;
{
  switch (GET_CODE (x))
    {
    case REG:
    case MEM:
      return XINT (x, 2);
    case SUBREG:
      return XINT (x, 3);
    case ASHIFT:
    case ROTATE:
    case LSHIFTRT:
    case ASHIFTRT:
    case ROTATERT:
    case PLUS:
    case MINUS:
    case MULT:
    case DIV:
    case MOD:
    case UDIV:
    case UMOD:
    case AND:
    case IOR:
    case XOR:
    case NOT:
    case EQ:
    case NE:
    case GE:
    case GT: 
    case LE:
    case LT:
    case GEU:
    case GTU: 
    case LEU:
    case LTU:
    case CALL:
      return rtx_vstride (XEXP (x,0));
    case SIGN_EXTEND:
    case ZERO_EXTEND:
    case TRUNCATE:
    case FLOAT_EXTEND:
    case FLOAT_TRUNCATE:
    case FLOAT:
    case FIX:
    case UNSIGNED_FLOAT: 
     return rtx_vstride (XEXP (x,0));
    case SEQUENCE:
    case PARALLEL:
      return rtx_vstride (XVECEXP (x, 0, 0));
    case USE:
    case CLOBBER:
      return rtx_vstride (XEXP (x,0));
    case QUEUED:
      return rtx_vstride (XEXP (x,3));
    case SET:
      return rtx_vstride (SET_DEST (x));
    case INSN:
      return rtx_vstride (PATTERN(x));
    case EXPR_LIST:
      return rtx_vstride (XEXP (x,0));
    case HIGH:
    case LO_SUM:
      return rtx_vstride (XEXP (x,0));
    case GSYNC:
      return rtx_vstride (XEXP (x,0));
    case ASM_OPERANDS:
      if (ASM_OPERANDS_INPUT_LENGTH (x))
	return rtx_vstride (ASM_OPERANDS_INPUT (x, 0));
      return 0;
    case SYMBOL_REF:
    case LABEL_REF:
    case JUMP_INSN:
    case CALL_INSN:
    case BARRIER:
    case CODE_LABEL:
    case NOTE:
    case CONST:
    case ASM_INPUT:
      return 0;
    case CONST_INT:
    case CONST_DOUBLE:
      return GET_MODE_SIZE (GET_MODE (x));
    default:
      abort();
    }
}

int
rtx_vector_p (x)
     rtx x;
{
  if (rtx_vlength (x))
    return 1;
  return 0;
}

tree
tree_vector_type (type)
     tree type;
{
  tree vals;
  while (1)
    {
      switch(TREE_CODE (type))
	{
	case PARM_DECL:
	case VAR_DECL:
	  type = TREE_TYPE(type);
	  break;
	  
	case ARRAY_TYPE:
	  vals = TYPE_DOMAIN(type);
	  if (!vals)
	    return type;	/* no domain specified, a valid vector type */
	  if (global_vector_length ||
	      ((TREE_INT_CST_LOW(TYPE_MAX_VALUE(vals))
		- TREE_INT_CST_LOW(TYPE_MIN_VALUE(vals)))
	       < MAX_VECTOR_LENGTH))
	    return type;	/* valid vector length */
	  return 0;		/* invalid vector length */

	case POINTER_TYPE:
	  return 0;
	  break;

	case ERROR_MARK:
	  return 0;

	default:
	  if (!TREE_TYPE(type))
	    return 0;
	  type = TREE_TYPE(type);
	}
    }
}

int
tree_vlength (t)
     tree t;
{
  tree vals;
  t = tree_vector_type (t);

  if (!t || !tree_vector_p (t))
    return 0;

  vals = TYPE_DOMAIN(t);
  if (!vals)
    return 0;
  return (1 + TREE_INT_CST_LOW(TYPE_MAX_VALUE(vals))
	  - TREE_INT_CST_LOW(TYPE_MIN_VALUE(vals)));
}

int
tree_vstride (t)
     tree t;
{
  tree tt = tree_vector_type(t);

  if (!tt || !tree_vector_p (tt))
    return 0;

  return TYPE_STRIDE (tt);
}

int
tree_vector_p (t)
     tree t;
{
  t = tree_vector_type (t);
  if (!t) 
    return 0;

  switch (TREE_CODE (TREE_TYPE (t)))
    {
    case RECORD_TYPE:
    case UNION_TYPE:
      if (TYPE_MODE (TREE_TYPE (t)) != ABLKmode)
	return 0;
    case AUX_ENUMERAL_TYPE:
    case AUX_INTEGER_TYPE:
    case AUX_REAL_TYPE:
    case AUX_POINTER_TYPE:
      return 1;
    }
  return 0;
}

tree
tree_vbase_type (type)
     tree type;
{
  tree t1;
  if (!(t1 = tree_vector_type(type)))
    return type;

  return TREE_TYPE(t1);
}

void
set_vector_rtx (x)
     rtx x;
{
  current_vector_length = rtx_vlength (x);
  if (current_vector_length)
    current_vector_stride = rtx_vstride (x);
  else
    current_vector_stride = 0;
}

void
set_vector_tree (t)
     tree t;
{
  current_vector_length = tree_vlength (t);
  if (current_vector_length)
    current_vector_stride = tree_vstride (t);
  else
    current_vector_stride = 0;
}

void
set_vector_int (vl, vs)
     int vl, vs;
{
  current_vector_length = vl;
  if (current_vector_length)
    current_vector_stride = vs;
  else
    current_vector_stride = 0;
}

void
set_hardreg_rtx (x)
     rtx x;
{
  hard_regno_vector_length = rtx_vlength (x);
}

void
set_hardreg_mode (mode, size)
int size;
enum machine_mode mode;
{
  if (size)
    {
      hard_regno_vector_length
	= size / ((GET_MODE_SIZE(mode) + UNITS_PER_WORD - 1) / UNITS_PER_WORD);
      if (hard_regno_vector_length == 1)
	hard_regno_vector_length = 0;
    }
  else
    hard_regno_vector_length = 0;
}

void
set_hardreg_tree (t)
     tree t;
{
  if (tree_vector_p (t))
    hard_regno_vector_length = tree_vlength (t);
  else
    hard_regno_vector_length = 0;
}

void
modify_vector_rtx (r)
     rtx r;
{
  switch (GET_CODE (r))
    {
    case SUBREG:
      XINT (r, 2) = current_vector_length;
      XINT (r, 3) = current_vector_stride;
      break;

    case MEM:
    case REG:
      XINT (r, 1) = current_vector_length;
      XINT (r, 2) = current_vector_stride;
      break;

    default:
      abort ();
    }
  current_vector_length = default_vector_length;
  current_vector_stride = default_vector_stride;
}

