/* Vector flow analysis for GNU compiler.
   Copyright (C) 1987, 1988, 1992 Free Software Foundation, Inc.

   Vector, Aux, and Dist additions Copyright (C) 1993,1994 
      United States Government as represented by the Director, 
      National Security Agency.
      All rights reserved

This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

/* JMD 7/14/93:  This file incorporates some of the code from flow.c
   in modified form to analyze and modify vector lengths.

   The first step is dividing the function into basic blocks.
   find_basic_blocks_for_vector_flow does this.

   ** find_basic_blocks_for_vector_flow **

   find_basic_blocks_for_vector_flow divides the current function's rtl
   into basic blocks.  It records the beginnings and ends of the
   basic blocks in the vectors basic_block_head and basic_block_end,
   and the number of blocks in n_basic_blocks.

   find_basic_blocks_for_vector_flow also finds any unreachable loops
   and deletes them. */


#include <stdio.h>
#include "config.h"
#include "rtl.h"
#include "basic-block.h"
#include "insn-config.h"
#include "regs.h"
#include "hard-reg-set.h"
#include "flags.h"
#include "output.h"

#include "obstack.h"
#define obstack_chunk_alloc xmalloc
#define obstack_chunk_free free

/* added by JMD 12/1/92 for vector flow in ac */
#define BOTTOM -1
#define VLOOPTHRESHHOLD 0

extern void free ();

/* List of labels that must never be deleted.  */
extern rtx forced_labels;

/* Get the basic block number of an insn.
   This info should not be expected to remain available
   after the end of life_analysis.  */

/* This is the limit of the allocated space in the following two arrays.  */

static int max_uid_for_flow;

#define BLOCK_NUM(INSN)  uid_block_number[INSN_UID (INSN)]

/* This is where the BLOCK_NUM values are really stored.
   This is set up by find_basic_blocks_for_vector_flow and used there and in life_analysis,
   and then freed.  */

static short *uid_block_number;

/* Number of basic blocks in the current function.  */

int n_basic_blocks;

/* Element N is first insn in basic block N.
   This info lasts until we finish compiling the function.  */

rtx *basic_block_head;

/* Element N is last insn in basic block N.
   This info lasts until we finish compiling the function.  */

rtx *basic_block_end;

/* Element N is a regset describing the registers live
   at the start of basic block N.
   This info lasts until we finish compiling the function.  */

regset *basic_block_live_at_start;

/* Element N is nonzero if control can drop into basic block N
   from the preceding basic block.  Freed after life_analysis.  */

static char *basic_block_drops_in;

/* Element N is depth within loops of the last insn in basic block number N.
   Freed after life_analysis.  */

static short *basic_block_loop_depth;

/* Element N nonzero if basic block N can actually be reached.
   Vector exists only during find_basic_blocks_for_vector_flow.  */

static char *block_live_static;

/* Depth within loops of basic block being scanned for lifetime analysis,
   plus one.  This is the weight attached to references to registers.  */

static int loop_depth;

/* Set of registers that may be eliminable.  These are handled specially
   in updating regs_ever_live.  */

static HARD_REG_SET elim_reg_set;

/* desired length for vector insns.  Set in toplev.c in response to
   -fstrip-mine{-n} user options */
extern int global_vector_length;

int *basic_block_vector_in_length;
int *basic_block_vector_out_length;
int *basic_block_vector_gen_length;
int **basic_block_predecessors;
int **basic_block_successors;
rtx *new_vector_regs1;
rtx *new_vector_regs2;
  
/* Forward declarations */
static void find_basic_blocks_for_vector_flow ();
/* The following added by JMD 11/24/92 for ac */
static void modify_vector_lengths ();
static rtx make_vector_loop ();
static void find_block_flow_info ();
static void find_vector_flow_info ();
static void init_block_vector ();
static void print_vector_flow ();
static void update_vector_regs ();
static void update_mem_refs ();
/* end of JMD additions */
static void mark_label_ref ();
void allocate_for_life_analysis (); /* Used also in stupid_life_analysis */
void dump_flow_info ();

/* Find basic blocks of the current function and perform only the
   data flow analysis necessary to change vector lengths.  Most of
   this function is copied from flow_analysis in flow.c.
   F is the first insn of the function and NREGS the number of register numbers
   in use.  */

void
analyze_vector_lengths (f, nregs, file)
     rtx f;
     int nregs;
     FILE *file;
{
  register rtx insn;
  register int i;
  rtx reg;
  int *tem, regcount, save_vl;
  rtx nonlocal_label_list = nonlocal_label_rtx_list ();

#ifdef ELIMINABLE_REGS
  static struct {int from, to; } eliminables[] = ELIMINABLE_REGS;
#endif

  /* Record which registers will be eliminated.  We use this in
     mark_used_regs. */

  CLEAR_HARD_REG_SET (elim_reg_set);

#ifdef ELIMINABLE_REGS
  for (i = 0; i < sizeof eliminables / sizeof eliminables[0]; i++)
    SET_HARD_REG_BIT (elim_reg_set, eliminables[i].from);
#else
  SET_HARD_REG_BIT (elim_reg_set, FRAME_POINTER_REGNUM);
#endif

  /* Count the basic blocks.  Also find maximum insn uid value used.  */

  {
    register RTX_CODE prev_code = JUMP_INSN;
    register RTX_CODE code;

    max_uid_for_flow = 0;

    for (insn = f, i = 0; insn; insn = NEXT_INSN (insn))
      {
	code = GET_CODE (insn);
	if (INSN_UID (insn) > max_uid_for_flow)
	  max_uid_for_flow = INSN_UID (insn);
	if (code == CODE_LABEL
	    || (GET_RTX_CLASS (code) == 'i'
		&& (prev_code == JUMP_INSN
		    || (prev_code == CALL_INSN
			/* && nonlocal_label_list != 0 */)
		    || prev_code == BARRIER)))
	  i++;
	if (code != NOTE)
	  prev_code = code;
      }
  }

#ifdef AUTO_INC_DEC
  /* Leave space for insns we make in some cases for auto-inc.  These cases
     are rare, so we don't need too much space.  */
  max_uid_for_flow += max_uid_for_flow / 10;
#endif

  /* Allocate some tables that last till end of compiling this function
     and some needed only in find_basic_blocks_for_vector_flow and life_analysis.  */

  n_basic_blocks = i;
  basic_block_head = (rtx *) oballoc (n_basic_blocks * sizeof (rtx));
  basic_block_end = (rtx *) oballoc (n_basic_blocks * sizeof (rtx));
  basic_block_drops_in = (char *) alloca (n_basic_blocks);
  basic_block_loop_depth = (short *) alloca (n_basic_blocks * sizeof (short));
  /*****************************************************************
    Code added by JMD for flow analysis for vector lengths in ac.
    ****************************************************************/
  basic_block_vector_in_length = (int*) alloca (n_basic_blocks * sizeof (int));
  basic_block_vector_out_length = (int*) alloca (n_basic_blocks * sizeof (int));
  basic_block_vector_gen_length = (int*) alloca (n_basic_blocks * sizeof (int));
  bzero (basic_block_vector_in_length, n_basic_blocks * sizeof (int));
  bzero (basic_block_vector_out_length, n_basic_blocks * sizeof (int));
  bzero (basic_block_vector_gen_length, n_basic_blocks * sizeof (int));
  basic_block_predecessors = (int **) alloca (n_basic_blocks * sizeof (int *));
  basic_block_successors = (int **) alloca (n_basic_blocks * sizeof (int *));
  uid_block_number
    = (short *) alloca ((max_uid_for_flow + 1) * sizeof (short));

  find_basic_blocks_for_vector_flow (f, nonlocal_label_list);

  /*****************************************************************
    Here is the place to modify vector lengths using the previous
    basic block analysis.  For now try not to do live variable analysis.
    ****************************************************************/

  tem = (int *) alloca (n_basic_blocks * n_basic_blocks * sizeof (int));
  bzero (tem, n_basic_blocks * n_basic_blocks * sizeof (int));
  init_block_vector (basic_block_predecessors, tem, n_basic_blocks, n_basic_blocks * sizeof (int));
  tem = (int *) alloca (n_basic_blocks * n_basic_blocks * sizeof (int));
  bzero (tem, n_basic_blocks * n_basic_blocks * sizeof (int));
  init_block_vector (basic_block_successors, tem, n_basic_blocks, n_basic_blocks * sizeof (int));

  find_block_flow_info (f, basic_block_predecessors, basic_block_successors,nonlocal_label_list);
  find_vector_flow_info (f, basic_block_predecessors, basic_block_successors);
  /* set up vector for mapping old vector regs to new vector regs */
  new_vector_regs1 = (rtx *) alloca (max_reg_num() * sizeof (rtx));
  bzero (new_vector_regs1, max_reg_num() * sizeof (rtx));
  new_vector_regs2 = (rtx *) alloca (max_reg_num() * sizeof (rtx));
  bzero (new_vector_regs2, max_reg_num() * sizeof (rtx));
  regcount = max_reg_num();
  for (i = 0; i < regcount; i++)
    if (regno_reg_rtx[i]
	&& rtx_vlength(regno_reg_rtx[i]) > global_vector_length)
      {
	int taillength;
	set_vector_int (global_vector_length, rtx_vstride (regno_reg_rtx[i]));
	new_vector_regs1[i] = gen_reg_rtx (GET_MODE (regno_reg_rtx[i]));
	if (taillength = rtx_vlength(regno_reg_rtx[i]) % global_vector_length)
	  {
	    set_vector_int (taillength, rtx_vstride (regno_reg_rtx[i]));
	    new_vector_regs2[i] = gen_reg_rtx (GET_MODE (regno_reg_rtx[i]));
	  }
      }
  modify_vector_lengths (f, nregs);
  if (file)
    print_vector_flow (file);

  basic_block_drops_in = 0;
  uid_block_number = 0;
  basic_block_loop_depth = 0;
}

/* Find all basic blocks of the function whose first insn is F.
   Store the correct data in the tables that describe the basic blocks,
   set up the chains of references for each CODE_LABEL, and
   delete any entire basic blocks that cannot be reached.

   NONLOCAL_LABEL_LIST is the same local variable from flow_analysis.  */

static void
find_basic_blocks_for_vector_flow (f, nonlocal_label_list)
     rtx f, nonlocal_label_list;
{
  register rtx insn;
  register int i;
  register char *block_live = (char *) alloca (n_basic_blocks);
  register char *block_marked = (char *) alloca (n_basic_blocks);
  /* List of label_refs to all labels whose addresses are taken
     and used as data.  */
  rtx label_value_list = 0;

  block_live_static = block_live;
  bzero (block_live, n_basic_blocks);
  bzero (block_marked, n_basic_blocks);

  /* Initialize with just block 0 reachable and no blocks marked.  */
  if (n_basic_blocks > 0)
    block_live[0] = 1;

  /* Initialize the ref chain of each label to 0.  */
  /* Record where all the blocks start and end and their depth in loops.  */
  /* For each insn, record the block it is in.  */
  /* Also mark as reachable any blocks headed by labels that
     must not be deleted.  */

  {
    register RTX_CODE prev_code = JUMP_INSN;
    register RTX_CODE code;
    int depth = 1;

    for (insn = f, i = -1; insn; insn = NEXT_INSN (insn))
      {
	code = GET_CODE (insn);
	if (code == NOTE)
	  {
	    if (NOTE_LINE_NUMBER (insn) == NOTE_INSN_LOOP_BEG)
	      depth++;
	    else if (NOTE_LINE_NUMBER (insn) == NOTE_INSN_LOOP_END)
	      depth--;
	  }
	/* A basic block starts at label, or after something that can jump.  */
	else if (code == CODE_LABEL
		 || (GET_RTX_CLASS (code) == 'i'
		     && (prev_code == JUMP_INSN
			 || (prev_code == CALL_INSN
			     && GET_CODE (PATTERN (insn)) != CLOBBER
			     /* && nonlocal_label_list != 0 */)
			 || prev_code == BARRIER)))
	  {
	    basic_block_head[++i] = insn;
	    basic_block_end[i] = insn;
	    basic_block_loop_depth[i] = depth;
	    if (code == CODE_LABEL)
	      {
		LABEL_REFS (insn) = insn;
		/* Any label that cannot be deleted
		   is considered to start a reachable block.  */
		if (LABEL_PRESERVE_P (insn))
		  block_live[i] = 1;
	      }
	  }
	else if (GET_RTX_CLASS (code) == 'i')
	  {
	    basic_block_end[i] = insn;
	    basic_block_loop_depth[i] = depth;
	  }

	/* Make a list of all labels referred to other than by jumps.  */
	if (code == INSN || code == CALL_INSN)
	  {
	    rtx note = find_reg_note (insn, REG_LABEL, 0);
	    if (note != 0)
	      label_value_list = gen_rtx (EXPR_LIST, VOIDmode, XEXP (note, 0),
					  label_value_list);
	  }

	BLOCK_NUM (insn) = i;

	/* Don't separate a CALL_INSN from following CLOBBER insns.  This is
	   a kludge that will go away when each CALL_INSN records its
	   USE and CLOBBERs.  */

	if (code != NOTE
	    && ! (prev_code == CALL_INSN && code == INSN
		  && GET_CODE (PATTERN (insn)) == CLOBBER))
	  prev_code = code;
      }
    if (i + 1 != n_basic_blocks)
      abort ();
  }

  /* Don't delete the labels that are referenced by non-jump instructions.  */
  {
    register rtx x;
    for (x = label_value_list; x; x = XEXP (x, 1))
      block_live[BLOCK_NUM (XEXP (x, 0))] = 1;
  }

  /* Record which basic blocks control can drop in to.  */

  {
    register int i;
    for (i = 0; i < n_basic_blocks; i++)
      {
	register rtx insn = PREV_INSN (basic_block_head[i]);
	/* TEMP1 is used to avoid a bug in Sequent's compiler.  */
	register int temp1;
	while (insn && GET_CODE (insn) == NOTE)
	  insn = PREV_INSN (insn);
	temp1 = insn && GET_CODE (insn) != BARRIER;
	basic_block_drops_in[i] = temp1;
      }
  }

  /* Now find which basic blocks can actually be reached
     and put all jump insns' LABEL_REFS onto the ref-chains
     of their target labels.  */

  if (n_basic_blocks > 0)
    {
      int something_marked = 1;

      /* Find all indirect jump insns and mark them as possibly jumping
	 to all the labels whose addresses are explicitly used.
	 This is because, when there are computed gotos,
	 we can't tell which labels they jump to, of all the possibilities.  */

      for (insn = f; insn; insn = NEXT_INSN (insn))
	if (GET_CODE (insn) == JUMP_INSN
	    && GET_CODE (PATTERN (insn)) == SET
	    && SET_DEST (PATTERN (insn)) == pc_rtx
	    && (GET_CODE (SET_SRC (PATTERN (insn))) == REG
		|| GET_CODE (SET_SRC (PATTERN (insn))) == MEM))
	  {
	    rtx x;
	    for (x = label_value_list; x; x = XEXP (x, 1))
	      mark_label_ref (gen_rtx (LABEL_REF, VOIDmode, XEXP (x, 0)),
			      insn, 0);
	    for (x = forced_labels; x; x = XEXP (x, 1))
	      mark_label_ref (gen_rtx (LABEL_REF, VOIDmode, XEXP (x, 0)),
			      insn, 0);
	  }

      /* Find all call insns and mark them as possibly jumping
	 to all the nonlocal goto handler labels.  */

      for (insn = f; insn; insn = NEXT_INSN (insn))
	if (GET_CODE (insn) == CALL_INSN)
	  {
	    rtx x;
	    for (x = nonlocal_label_list; x; x = XEXP (x, 1))
	      mark_label_ref (gen_rtx (LABEL_REF, VOIDmode, XEXP (x, 0)),
			      insn, 0);
	    /* ??? This could be made smarter:
	       in some cases it's possible to tell that certain
	       calls will not do a nonlocal goto.

	       For example, if the nested functions that do the
	       nonlocal gotos do not have their addresses taken, then
	       only calls to those functions or to other nested
	       functions that use them could possibly do nonlocal
	       gotos.  */
	  }

      /* Pass over all blocks, marking each block that is reachable
	 and has not yet been marked.
	 Keep doing this until, in one pass, no blocks have been marked.
	 Then blocks_live and blocks_marked are identical and correct.
	 In addition, all jumps actually reachable have been marked.  */

      while (something_marked)
	{
	  something_marked = 0;
	  for (i = 0; i < n_basic_blocks; i++)
	    if (block_live[i] && !block_marked[i])
	      {
		block_marked[i] = 1;
		something_marked = 1;
		if (i + 1 < n_basic_blocks && basic_block_drops_in[i + 1])
		  block_live[i + 1] = 1;
		insn = basic_block_end[i];
		if (GET_CODE (insn) == JUMP_INSN)
		  mark_label_ref (PATTERN (insn), insn, 0);
	      }
	}

      /* Now delete the code for any basic blocks that can't be reached.
	 They can occur because jump_optimize does not recognize
	 unreachable loops as unreachable.  */

      for (i = 0; i < n_basic_blocks; i++)
	if (!block_live[i])
	  {
	    insn = basic_block_head[i];
	    while (1)
	      {
		if (GET_CODE (insn) == BARRIER)
		  abort ();
		if (GET_CODE (insn) != NOTE)
		  {
		    PUT_CODE (insn, NOTE);
		    NOTE_LINE_NUMBER (insn) = NOTE_INSN_DELETED;
		    NOTE_SOURCE_FILE (insn) = 0;
		  }
		if (insn == basic_block_end[i])
		  {
		    /* BARRIERs are between basic blocks, not part of one.
		       Delete a BARRIER if the preceding jump is deleted.
		       We cannot alter a BARRIER into a NOTE
		       because it is too short; but we can really delete
		       it because it is not part of a basic block.  */
		    if (NEXT_INSN (insn) != 0
			&& GET_CODE (NEXT_INSN (insn)) == BARRIER)
		      delete_insn (NEXT_INSN (insn));
		    break;
		  }
		insn = NEXT_INSN (insn);
	      }
	    /* Each time we delete some basic blocks,
	       see if there is a jump around them that is
	       being turned into a no-op.  If so, delete it.  */

	    if (block_live[i - 1])
	      {
		register int j;
		for (j = i; j < n_basic_blocks; j++)
		  if (block_live[j])
		    {
		      rtx label;
		      insn = basic_block_end[i - 1];
		      if (GET_CODE (insn) == JUMP_INSN
			  /* An unconditional jump is the only possibility
			     we must check for, since a conditional one
			     would make these blocks live.  */
			  && simplejump_p (insn)
			  && (label = XEXP (SET_SRC (PATTERN (insn)), 0), 1)
			  && INSN_UID (label) != 0
			  && BLOCK_NUM (label) == j)
			{
			  PUT_CODE (insn, NOTE);
			  NOTE_LINE_NUMBER (insn) = NOTE_INSN_DELETED;
			  NOTE_SOURCE_FILE (insn) = 0;
			  if (GET_CODE (NEXT_INSN (insn)) != BARRIER)
			    abort ();
			  delete_insn (NEXT_INSN (insn));
			}
		      break;
		    }
	      }
	  }
    }
}

/* Check expression X for label references;
   if one is found, add INSN to the label's chain of references.

   CHECKDUP means check for and avoid creating duplicate references
   from the same insn.  Such duplicates do no serious harm but
   can slow life analysis.  CHECKDUP is set only when duplicates
   are likely.  */

static void
mark_label_ref (x, insn, checkdup)
     rtx x, insn;
     int checkdup;
{
  register RTX_CODE code;
  register int i;
  register char *fmt;

  /* We can be called with NULL when scanning label_value_list.  */
  if (x == 0)
    return;

  code = GET_CODE (x);
  if (code == LABEL_REF)
    {
      register rtx label = XEXP (x, 0);
      register rtx y;
      if (GET_CODE (label) != CODE_LABEL)
	abort ();
      /* If the label was never emitted, this insn is junk,
	 but avoid a crash trying to refer to BLOCK_NUM (label).
	 This can happen as a result of a syntax error
	 and a diagnostic has already been printed.  */
      if (INSN_UID (label) == 0)
	return;
      CONTAINING_INSN (x) = insn;
      /* if CHECKDUP is set, check for duplicate ref from same insn
	 and don't insert.  */
      if (checkdup)
	for (y = LABEL_REFS (label); y != label; y = LABEL_NEXTREF (y))
	  if (CONTAINING_INSN (y) == insn)
	    return;
      LABEL_NEXTREF (x) = LABEL_REFS (label);
      LABEL_REFS (label) = x;
      block_live_static[BLOCK_NUM (label)] = 1;
      return;
    }

  fmt = GET_RTX_FORMAT (code);
  for (i = GET_RTX_LENGTH (code) - 1; i >= 0; i--)
    {
      if (fmt[i] == 'e')
	mark_label_ref (XEXP (x, i), insn, 0);
      if (fmt[i] == 'E')
	{
	  register int j;
	  for (j = 0; j < XVECLEN (x, i); j++)
	    mark_label_ref (XVECEXP (x, i, j), insn, 1);
	}
    }
}


static void
modify_vector_lengths (f, nregs)
     rtx f;
     int nregs;
{
  int i;
  rtx block, insn, endmarker;

  for (i=0; i < n_basic_blocks; i++)
    {
      if (basic_block_vector_gen_length[i] == 0)
	continue;
      block = basic_block_head [i];
      endmarker = next_nonnote_insn (basic_block_end [i]);
      for (insn = block; insn != endmarker;
	   insn = NEXT_INSN (insn))
	{
	  if (rtx_vlength (insn) > MAX_VECTOR_LENGTH)
	    insn = make_vector_loop (insn);
	}
    }
}

struct insnchain {
  rtx insn;
  struct insnchain *next;
}*ic;

static rtx
make_vector_loop (insn)
     rtx insn;
{
  rtx prev = PREV_INSN (insn);
  rtx next = NEXT_INSN (insn);
  rtx firstvinsn = insn;
  rtx newinsn, pattern, lastvinsn, src, dest, mr, loopstart;
  rtx fcinsn;
  int i,j, veclen, itercount, rem, meminc, vectorinc, offset;
  struct insnchain **uses, **antidep, **followers;
  rtx *lastset, *flowdep, *outputdep;
  enum machine_mode mode;
  enum rtx_code code;
  struct vectormemref {
    rtx insn;
    rtx mref;
    rtx *memaddr;
    rtx incpattern;
    struct vectormemref *next;
  }*vmr, *vmrtemp;

  veclen = rtx_vlength (insn);
  itercount = veclen / global_vector_length;
  rem = veclen % global_vector_length;

  /* Find scalar insns and move them ahead of this vector insn
     whenever dependences permit.  This will increase register
     pressure on aux scalar registers but should reduce vector
     memory traffic.  Don't move scalar insns between the last
     vector insn and block end. */
  for (lastvinsn = basic_block_end [BLOCK_NUM (insn)];
       rtx_vlength (lastvinsn) == 0;
       lastvinsn = PREV_INSN (lastvinsn))
    ;   /* at this point lastvinsn should be the last vector insn in block */
  lastset = (rtx *) alloca (max_reg_num() * sizeof (rtx ));
  bzero (lastset, max_reg_num() * sizeof (rtx ));
  uses = (struct insnchain **) alloca (max_reg_num() *
				      sizeof (struct insnchain *));
  bzero (uses, max_reg_num() *  sizeof (struct insnchain *));
  flowdep = (rtx *) alloca (max_uid_for_flow * sizeof (rtx ));
  bzero (flowdep, max_uid_for_flow * sizeof (rtx ));
  antidep = (struct insnchain **) alloca (max_uid_for_flow *
					  sizeof (struct insnchain *));
  bzero (antidep, max_uid_for_flow * sizeof (struct insnchain *));
  outputdep = (rtx *) alloca (max_uid_for_flow * sizeof (rtx ));
  bzero (outputdep, max_uid_for_flow * sizeof (rtx ));
  followers = (struct insnchain **) alloca (max_uid_for_flow *
				      sizeof (struct insnchain *));
  bzero (followers, max_uid_for_flow *  sizeof (struct insnchain *));

  /* The following code computes def-use chains for dependence
     analysis between firstvinsn and lastvinsn.  It is not in a
     separate function because of alloca. */
  for (fcinsn = lastvinsn; fcinsn != PREV_INSN (firstvinsn); fcinsn = PREV_INSN (fcinsn))
    {
      if (GET_CODE (fcinsn) == NOTE)
	continue;
      if (GET_CODE (PATTERN (fcinsn)) == SET)
	{
	  rtx src, dest;
	  src = SET_SRC (PATTERN (fcinsn));
	  dest = SET_DEST (PATTERN (fcinsn));

	  while (GET_CODE (dest) == SUBREG
		 || GET_CODE (dest) == ZERO_EXTRACT
		 || GET_CODE (dest) == SIGN_EXTRACT
		 || GET_CODE (dest) == STRICT_LOW_PART)
	    dest = XEXP (dest, 0);
	  if (GET_CODE (dest) == REG)
	    {
	      for (ic = uses [REGNO (dest)]; ic; ic = ic->next)
		flowdep [INSN_UID (ic->insn)] = fcinsn;
	      followers [INSN_UID (fcinsn)] = uses [REGNO (dest)];
	      uses [REGNO (dest)] = NULL;
	      if (lastset [REGNO (dest)])
		{
		  outputdep [INSN_UID(lastset[REGNO(dest)])] = fcinsn;
		  ic = (struct insnchain *) alloca (sizeof (struct insnchain));
		  ic->insn = lastset[REGNO(dest)];
		  ic->next = followers [INSN_UID (fcinsn)];
		  followers [INSN_UID (fcinsn)] = ic;
		}
	      lastset [REGNO (dest)] = fcinsn;
	    }
	  else if (GET_CODE (dest) == MEM)
	    {
	      src = XEXP (dest, 0);
	      while (GET_CODE (src) == SUBREG
		     || GET_CODE (src) == ZERO_EXTRACT
		     || GET_CODE (src) == SIGN_EXTRACT
		     || GET_CODE (src) == STRICT_LOW_PART
		     || GET_CODE (src) == PRE_DEC
		     || GET_CODE (src) == PRE_INC
		     || GET_CODE (src) == POST_DEC
		     || GET_CODE (src) == POST_INC)
		src = XEXP (src, 0);
	      if (GET_CODE (src) == REG)
		{
		  ic = (struct insnchain *) alloca (sizeof (struct insnchain));
		  ic->insn = fcinsn;
		  ic->next = uses [REGNO (src)];
		  uses [REGNO (src)] = ic;
		  if (lastset [REGNO (src)])
		    {
		      ic = (struct insnchain *) alloca (sizeof (struct insnchain));
		      ic->insn = fcinsn;
		      ic->next = antidep [INSN_UID(lastset[REGNO(src)])];
		      antidep [INSN_UID(lastset[REGNO(src)])] = ic;
		      ic = (struct insnchain *) alloca (sizeof (struct insnchain));
		      ic->insn = lastset[REGNO(src)];
		      ic->next = followers [INSN_UID(fcinsn)];
		      followers [INSN_UID(fcinsn)] = ic;
		    }
		}
	      else if (GET_CODE (src) == PLUS
		       || GET_CODE (src) == MINUS
		       || GET_CODE (src) == LO_SUM)
		{
		  rtx opnd;
		  opnd = XEXP (src, 0);
		  while (GET_CODE (opnd) == SUBREG
			 || GET_CODE (opnd) == ZERO_EXTRACT
			 || GET_CODE (opnd) == SIGN_EXTRACT
			 || GET_CODE (opnd) == STRICT_LOW_PART
			 || GET_CODE (opnd) == PRE_DEC
			 || GET_CODE (opnd) == PRE_INC
			 || GET_CODE (opnd) == POST_DEC
			 || GET_CODE (opnd) == POST_INC)
		    opnd = XEXP (opnd, 0);
		  if (GET_CODE (opnd) == REG)
		    {
		      ic = (struct insnchain *) alloca (sizeof (struct insnchain));
		      ic->insn = fcinsn;
		      ic->next = uses [REGNO (opnd)];
		      uses [REGNO (opnd)] = ic;
		      if (lastset [REGNO (opnd)])
			{
			  ic = (struct insnchain *) alloca (sizeof (struct insnchain));
			  ic->insn = fcinsn;
			  ic->next = antidep [INSN_UID(lastset[REGNO(opnd)])];
			  antidep [INSN_UID(lastset[REGNO(opnd)])] = ic;
			  ic = (struct insnchain *) alloca (sizeof (struct insnchain));
			  ic->insn = lastset[REGNO(opnd)];
			  ic->next = followers [INSN_UID(fcinsn)];
			  followers [INSN_UID(fcinsn)] = ic;
			}
		    }
		  opnd = XEXP (src, 1);
		  while (GET_CODE (opnd) == SUBREG
			 || GET_CODE (opnd) == ZERO_EXTRACT
			 || GET_CODE (opnd) == SIGN_EXTRACT
			 || GET_CODE (opnd) == STRICT_LOW_PART
			 || GET_CODE (opnd) == PRE_DEC
			 || GET_CODE (opnd) == PRE_INC
			 || GET_CODE (opnd) == POST_DEC
			 || GET_CODE (opnd) == POST_INC)
		    opnd = XEXP (opnd, 0);
		  if (GET_CODE (opnd) == REG)
		    {
		      ic = (struct insnchain *) alloca (sizeof (struct insnchain));
		      ic->insn = fcinsn;
		      ic->next = uses [REGNO (opnd)];
		      uses [REGNO (opnd)] = ic;
		      if (lastset [REGNO (opnd)])
			{
			  ic = (struct insnchain *) alloca (sizeof (struct insnchain));
			  ic->insn = fcinsn;
			  ic->next = antidep [INSN_UID(lastset[REGNO(opnd)])];
			  antidep [INSN_UID(lastset[REGNO(opnd)])] = ic;
			  ic = (struct insnchain *) alloca (sizeof (struct insnchain));
			  ic->insn = lastset[REGNO(opnd)];
			  ic->next = followers [INSN_UID(fcinsn)];
			  followers [INSN_UID(fcinsn)] = ic;
			}
		    }
		}
	    }

	  while (GET_CODE (src) == SUBREG
		 || GET_CODE (src) == ZERO_EXTRACT
		 || GET_CODE (src) == SIGN_EXTRACT
		 || GET_CODE (src) == STRICT_LOW_PART)
	    src = XEXP (src, 0);
	  if (GET_CODE (src) == REG)
	    {
	      ic = (struct insnchain *) alloca (sizeof (struct insnchain));
	      ic->insn = fcinsn;
	      ic->next = uses [REGNO (src)];
	      uses [REGNO (src)] = ic;
	      if (lastset [REGNO (src)])
		{
		  ic = (struct insnchain *) alloca (sizeof (struct insnchain));
		  ic->insn = fcinsn;
		  ic->next = antidep [INSN_UID(lastset[REGNO(src)])];
		  antidep [INSN_UID(lastset[REGNO(src)])] = ic;
		  ic = (struct insnchain *) alloca (sizeof (struct insnchain));
		  ic->insn = lastset[REGNO(src)];
		  ic->next = followers [INSN_UID(fcinsn)];
		  followers [INSN_UID(fcinsn)] = ic;
		}
	    }
	  else if (GET_CODE (src) == MEM)
	    {
	      src = XEXP (src, 0);
	      while (GET_CODE (src) == SUBREG
		     || GET_CODE (src) == ZERO_EXTRACT
		     || GET_CODE (src) == SIGN_EXTRACT
		     || GET_CODE (src) == STRICT_LOW_PART
		     || GET_CODE (src) == PRE_DEC
		     || GET_CODE (src) == PRE_INC
		     || GET_CODE (src) == POST_DEC
		     || GET_CODE (src) == POST_INC)
		src = XEXP (src, 0);
	      if (GET_CODE (src) == REG)
		{
		  ic = (struct insnchain *) alloca (sizeof (struct insnchain));
		  ic->insn = fcinsn;
		  ic->next = uses [REGNO (src)];
		  uses [REGNO (src)] = ic;
		  if (lastset [REGNO (src)])
		    {
		      ic = (struct insnchain *) alloca (sizeof (struct insnchain));
		      ic->insn = fcinsn;
		      ic->next = antidep [INSN_UID(lastset[REGNO(src)])];
		      antidep [INSN_UID(lastset[REGNO(src)])] = ic;
		      ic = (struct insnchain *) alloca (sizeof (struct insnchain));
		      ic->insn = lastset[REGNO(src)];
		      ic->next = followers [INSN_UID(fcinsn)];
		      followers [INSN_UID(fcinsn)] = ic;
		    }
		}
	      else if (GET_CODE (src) == PLUS
		       || GET_CODE (src) == MINUS
		       || GET_CODE (src) == LO_SUM)
		{
		  rtx opnd;
		  opnd = XEXP (src, 0);
		  while (GET_CODE (opnd) == SUBREG
			 || GET_CODE (opnd) == ZERO_EXTRACT
			 || GET_CODE (opnd) == SIGN_EXTRACT
			 || GET_CODE (opnd) == STRICT_LOW_PART)
		    opnd = XEXP (opnd, 0);
		  if (GET_CODE (opnd) == REG)
		    {
		      ic = (struct insnchain *) alloca (sizeof (struct insnchain));
		      ic->insn = fcinsn;
		      ic->next = uses [REGNO (opnd)];
		      uses [REGNO (opnd)] = ic;
		      if (lastset [REGNO (opnd)])
			{
			  ic = (struct insnchain *) alloca (sizeof (struct insnchain));
			  ic->insn = fcinsn;
			  ic->next = antidep [INSN_UID(lastset[REGNO(opnd)])];
			  antidep [INSN_UID(lastset[REGNO(opnd)])] = ic;
			  ic = (struct insnchain *) alloca (sizeof (struct insnchain));
			  ic->insn = lastset[REGNO(opnd)];
			  ic->next = followers [INSN_UID(fcinsn)];
			  followers [INSN_UID(fcinsn)] = ic;
			}
		    }
		  opnd = XEXP (src, 1);
		  while (GET_CODE (opnd) == SUBREG
			 || GET_CODE (opnd) == ZERO_EXTRACT
			 || GET_CODE (opnd) == SIGN_EXTRACT
			 || GET_CODE (opnd) == STRICT_LOW_PART)
		    opnd = XEXP (opnd, 0);
		  if (GET_CODE (opnd) == REG)
		    {
		      ic = (struct insnchain *) alloca (sizeof (struct insnchain));
		      ic->insn = fcinsn;
		      ic->next = uses [REGNO (opnd)];
		      uses [REGNO (opnd)] = ic;
		      if (lastset [REGNO (opnd)])
			{
			  ic = (struct insnchain *) alloca (sizeof (struct insnchain));
			  ic->insn = fcinsn;
			  ic->next = antidep [INSN_UID(lastset[REGNO(opnd)])];
			  antidep [INSN_UID(lastset[REGNO(opnd)])] = ic;
			  ic = (struct insnchain *) alloca (sizeof (struct insnchain));
			  ic->insn = lastset[REGNO(opnd)];
			  ic->next = followers [INSN_UID(fcinsn)];
			  followers [INSN_UID(fcinsn)] = ic;
			}
		    }
		}
	    }
	  else if (! CONSTANT_P (src))
	    {
	      rtx opnd;
	      opnd = XEXP (src, 0); /* assume at least one operand */
	      while (GET_CODE (opnd) == SUBREG
		     || GET_CODE (opnd) == ZERO_EXTRACT
		     || GET_CODE (opnd) == SIGN_EXTRACT
		     || GET_CODE (opnd) == STRICT_LOW_PART)
		opnd = XEXP (opnd, 0);
	      if (GET_CODE (opnd) == REG)
		{
		  ic = (struct insnchain *) alloca (sizeof (struct insnchain));
		  ic->insn = fcinsn;
		  ic->next = uses [REGNO (opnd)];
		  uses [REGNO (opnd)] = ic;
		  if (lastset [REGNO (opnd)])
		    {
		      ic = (struct insnchain *) alloca (sizeof (struct insnchain));
		      ic->insn = fcinsn;
		      ic->next = antidep [INSN_UID(lastset[REGNO(opnd)])];
		      antidep [INSN_UID(lastset[REGNO(opnd)])] = ic;
		      ic = (struct insnchain *) alloca (sizeof (struct insnchain));
		      ic->insn = lastset[REGNO(opnd)];
		      ic->next = followers [INSN_UID(fcinsn)];
		      followers [INSN_UID(fcinsn)] = ic;
		    }
		}
	      if (GET_RTX_CLASS (GET_CODE (src)) != '1')
		opnd = XEXP (src, 1);
	      else
		opnd = NULL;
	      while (opnd
		     && (GET_CODE (opnd) == SUBREG
			 || GET_CODE (opnd) == ZERO_EXTRACT
			 || GET_CODE (opnd) == SIGN_EXTRACT
			 || GET_CODE (opnd) == STRICT_LOW_PART)) 
		opnd = XEXP (opnd, 0);
	      if (opnd && GET_CODE (opnd) == REG)
		{
		  ic = (struct insnchain *) alloca (sizeof (struct insnchain));
		  ic->insn = fcinsn;
		  ic->next = uses [REGNO (opnd)];
		  uses [REGNO (opnd)] = ic;
		  if (lastset [REGNO (opnd)])
		    {
		      ic = (struct insnchain *) alloca (sizeof (struct insnchain));
		      ic->insn = fcinsn;
		      ic->next = antidep [INSN_UID(lastset[REGNO(opnd)])];
		      antidep [INSN_UID(lastset[REGNO(opnd)])] = ic;
		      ic = (struct insnchain *) alloca (sizeof (struct insnchain));
		      ic->insn = lastset[REGNO(opnd)];
		      ic->next = followers [INSN_UID(fcinsn)];
		      followers [INSN_UID(fcinsn)] = ic;
		    }
		}
	    }
	}
    }

  /* This code moves forward through the code.  That way we
     may be able to move a number of scalar insns whose only dependences
     are on each other. */
  newinsn = firstvinsn;
  while (newinsn != lastvinsn)
    {
      int uid = INSN_UID(newinsn);
      if (GET_CODE (PATTERN (newinsn)) == SET
	  && ! rtx_vlength (newinsn)
	  && (! (flowdep[uid] || antidep[uid] || outputdep [uid])))
	{
	  /* then we have a movable scalar insn; move it
	     and update the insns that depend on it */
	  
	  struct insnchain *temp1, *temp2;
	  for (ic = followers [uid]; ic; ic = ic->next)
	    {
	      rtx finsn = ic->insn;
	      int fuid = INSN_UID (finsn);
	      if (flowdep[fuid] == newinsn)
		flowdep[fuid] = 0;
	      if (outputdep[fuid] == newinsn)
		outputdep[fuid] = 0;
	      if (antidep[fuid])
		if (antidep[fuid]->insn == newinsn)
		  antidep[fuid] = antidep[fuid]->next;
		else
		  {
		    temp1 = antidep[fuid]->next;
		    temp2 = antidep[fuid];
		    while (temp1)
		      if (temp1->insn == newinsn)
			/* then remove newinsn from the chain */
			{
			  temp2->next = temp1->next;
			  temp1 = temp2->next;
			}
		      else
			{
			  temp2 = temp1;
			  temp1 = temp1->next;
			}
		  }
	    }
	  emit_insn_before (PATTERN (newinsn), insn);
	  newinsn = delete_insn (newinsn);
	  if (GET_CODE (newinsn) == NOTE)
	    newinsn = next_nonnote_insn (newinsn);
	}
      else
	newinsn = next_nonnote_insn (newinsn);
    }

  newinsn = PREV_INSN (firstvinsn);
  loopstart = firstvinsn;
  vmr = vmrtemp = NULL;
      
  /* find last insn in vector sequence with same vector length */
  for (lastvinsn = firstvinsn;
       lastvinsn != next_nonnote_insn (basic_block_end[BLOCK_NUM (insn)]);
       lastvinsn = next_nonnote_insn (lastvinsn))
    if (rtx_vlength (lastvinsn) != veclen)
      {
	lastvinsn = prev_nonnote_insn (lastvinsn);
	break;
      }
  if ((! lastvinsn)
      || lastvinsn == next_nonnote_insn (basic_block_end[BLOCK_NUM (insn)]))
    /* then lastvinsn should be the last insn in block */
    lastvinsn = basic_block_end[BLOCK_NUM (insn)];
  /* find memory references in the vector instructions */
  for (insn = firstvinsn;
       insn && insn != next_nonnote_insn (lastvinsn);
       insn = next_nonnote_insn (insn))
    {
      pattern = PATTERN (insn);
      if ( (code = GET_CODE (pattern)) == SET)
	{
	  rtx mref, mrpat, mop0, mop1;
	  rtx *memaddr;
	  if (GET_CODE (SET_SRC (pattern)) == MEM)
	    {
	      mref = SET_SRC (pattern);
	      memaddr = &(SET_SRC (pattern));
	    }
	  else if (GET_CODE (SET_DEST (pattern)) == MEM)
	    {
	      mref = SET_DEST (pattern);
	      memaddr = &(SET_DEST (pattern));
	    }
	  else if (GET_CODE (SET_SRC (pattern)) == ASM_OPERANDS)
	    {
	      int length = XVECLEN (SET_SRC (pattern), 4);
	      int i, found = 0;

	      for (i=0; i < length; i++)
		if (GET_CODE (XVECEXP (SET_SRC (pattern), 4, i)) == MEM)
		  {
		    mref = XVECEXP (SET_SRC (pattern), 4, i);
		    memaddr = &XVECEXP (SET_SRC (pattern), 4, i);
		    found = 1;
		  }
	      if (!found)
		continue;
	    }
	  else
	    continue;
	  mrpat = XEXP (mref, 0);
	  if (GET_CODE (mrpat) == REG
	      && rtx_vector_p (mrpat))
	    /* then this is scatter or gather; do not add to list */
	    continue;
	  else if (GET_CODE (mrpat) == PLUS)
	    {
	      mop0 = XEXP (mrpat, 0);
	      mop1 = XEXP (mrpat, 1);
	      if ((GET_CODE (mop0) == REG
		   && rtx_vector_p (mop0))
		  || (GET_CODE (mop1) == REG
		      && rtx_vector_p (mop1)))
		continue;
	    }  
	  vmrtemp = vmr;
	  vmr = (struct vectormemref *) alloca (sizeof (struct vectormemref));
	  vmr->insn = insn;
	  vmr->mref = mref;
	  vmr->memaddr = memaddr;
	  vmr->next = vmrtemp;
	}
    }
  if (vmr)   /* prepare for memory adjustments */
    {
      struct vectormemref *v1, *v2;
      rtx oldref, newref, oldaddr, offsetreg, increg, inc_const;
      for (vmrtemp = vmr; vmrtemp; vmrtemp = vmrtemp->next) 
	{
	  mr = copy_rtx_if_shared (vmrtemp->mref);
	  mode = GET_MODE (mr);
	  meminc = XINT (mr, 2);
	  oldref = XEXP (mr,0);
	  newref = gen_reg_rtx (SImode);
	  offsetreg = gen_reg_rtx (SImode);
	  increg = gen_reg_rtx (SImode);
	  
	  newinsn = emit_insn_after (gen_rtx (SET, SImode, offsetreg, const0_rtx),
				     newinsn);
	  inc_const = (gen_rtx (CONST_INT, SImode,
				global_vector_length * meminc));
	  newinsn = emit_insn_after (gen_rtx (SET, SImode, increg, inc_const),
				     newinsn);
	  vmrtemp->incpattern
	    = gen_rtx (SET, SImode, offsetreg,
		       gen_rtx (PLUS, SImode, offsetreg, increg));
	  newinsn = emit_insn_after (gen_rtx (SET, SImode, newref, oldref),
				     newinsn);
	  /* Now look for identical memory references in other vmr's */
	  v1 = v2 = vmrtemp;
	  while (v1->next)
	    {
	      v1 = v1->next;
	      if (rtx_equal_p (v1->mref, mr))
		{
		  *(v1->memaddr) = mr;
		  v2->next = v1->next;
		  v1 = v2;
		}
	      else
		v2 = v1;
	    }
	  XEXP (mr, 0) = gen_rtx (PLUS, SImode, newref, offsetreg);
	  XINT (mr, 1) = global_vector_length;
	  *(vmrtemp->memaddr) = mr;
	}
    }

  
  if (itercount <= VLOOPTHRESHHOLD) /* emit itercount sequences of vector insns
				       for each existing sequence followed by
				       a shorter sequence if rem > 0 */
    {
      newinsn = lastvinsn;        /* to initialize for loop */
      for (i=0; i < itercount; i++)
	{
	  for (insn = firstvinsn;
	       insn && insn != next_nonnote_insn (lastvinsn);
	       insn = next_nonnote_insn (insn))
	    {
	      pattern = copy_rtx_if_shared (PATTERN (insn));
	      update_vector_regs (pattern, new_vector_regs1);
	      newinsn = emit_insn_after (pattern, newinsn);
	    }
	  if (vmr)   /* emit insns for memory adjustments */
	    {
	      for (vmrtemp = vmr; vmrtemp; vmrtemp = vmrtemp->next) 
		newinsn = emit_insn_after (copy_rtx_if_shared (vmrtemp->incpattern), newinsn);
	    }
	}
      if (rem)
	{
	  for (insn = firstvinsn;
	       insn && insn != next_nonnote_insn (lastvinsn);
	       insn = next_nonnote_insn (insn))
	    {
	      pattern = copy_rtx_if_shared (PATTERN (insn));
	      update_vector_regs (pattern, new_vector_regs2);
	      newinsn = emit_insn_after (pattern, newinsn);
	    }
	}
      /* Now delete old vector insns */
      insn = firstvinsn;
      while (insn != next_nonnote_insn (lastvinsn))
	insn = delete_insn (insn);
      /* at this point newinsn should have the right value to return */
    }
  else  /* convert the vector insns into a loop */
    {
      rtx ivar, startlabel, contlabel, endlabel, ccrtx;
      struct insnchain *iclast;
      
      ivar = gen_reg_rtx (SImode);
      startlabel = gen_label_rtx ();
      contlabel = gen_label_rtx ();
      endlabel = gen_label_rtx ();
      newinsn = emit_insn_after (gen_rtx (SET, VOIDmode, ivar,
					   CONST0_RTX(SImode)), newinsn);
      newinsn = emit_note_after (NOTE_INSN_LOOP_BEG, newinsn);
      newinsn = emit_label_after (startlabel, newinsn);
      insn = firstvinsn;
      ic = iclast = NULL;
      while (insn != next_nonnote_insn (lastvinsn))
	{
	  pattern = copy_rtx_if_shared (PATTERN (insn));
	  update_vector_regs (pattern, new_vector_regs1);
	  emit_insn_before (pattern, insn);
	  if (iclast)
	    {
	      iclast->next = (struct insnchain *) alloca (sizeof (struct insnchain));
	      iclast = iclast->next;
	      iclast->insn = insn;
	      iclast->next = NULL;
	    }
	  else
	    {
	      ic = iclast = (struct insnchain *) alloca (sizeof (struct insnchain));
	      iclast->insn = insn;
	      iclast->next = NULL;
	    }
	  insn = next_nonnote_insn (insn);
	}
      if (! insn) /* then lastvinsn is the last real insn */
	if (! (insn = NEXT_INSN (lastvinsn)))
	  insn = emit_note_after (NOTE_INSN_DELETED, lastvinsn);
      /* At this point insn should have a valid value for what follows. */
      ccrtx = gen_rtx (REG, CCmode, 0);
      emit_insn_before (gen_rtx (SET, VOIDmode, ivar,
		           gen_rtx (PLUS, SImode, ivar,
			      CONST1_RTX(SImode))), insn);
      newinsn = emit_note_before (NOTE_INSN_LOOP_CONT, insn);
      emit_label_after (contlabel, newinsn);
      emit_insn_before (gen_rtx (SET, VOIDmode, ccrtx,
			   gen_rtx (COMPARE, CCmode, ivar,
			      gen_rtx (CONST_INT, SImode, itercount))), insn);
      newinsn = emit_jump_insn_before (gen_rtx (SET, VOIDmode, pc_rtx,
		        gen_rtx (IF_THEN_ELSE, VOIDmode,
			  gen_rtx (GE, VOIDmode, ccrtx, const0_rtx),
			    gen_rtx (LABEL_REF, VOIDmode, endlabel), pc_rtx)),
			     insn);
      if (vmr)   /* emit memory adjustments */
	for (vmrtemp = vmr; vmrtemp; vmrtemp = vmrtemp->next) 
	  newinsn = emit_insn_after (copy_rtx_if_shared (vmrtemp->incpattern),
				     newinsn);
      newinsn = emit_jump_insn_after (gen_rtx (SET, VOIDmode, pc_rtx,
					       gen_rtx (LABEL_REF, VOIDmode,
							startlabel)), newinsn);
      newinsn = emit_barrier_after (newinsn);
      newinsn = emit_note_after (NOTE_INSN_LOOP_END, newinsn);
      newinsn = emit_label_after (endlabel, newinsn);
      if (rem)
	{
	  rtx src0, src1;
	  
	  if (vmr)   /* emit insns for memory adjustments */
	    {
	      for (vmrtemp = vmr; vmrtemp; vmrtemp = vmrtemp->next) 
		newinsn = emit_insn_after (copy_rtx_if_shared (vmrtemp->incpattern), newinsn);
	    }
	  for (; ic; ic = ic->next)
	    {
	      pattern = copy_rtx_if_shared (PATTERN (ic->insn));
	      update_vector_regs (pattern, new_vector_regs2);
	      update_mem_refs (pattern, rem);
	      newinsn = emit_insn_after (pattern, newinsn);
	      delete_insn (ic->insn);
	    }
	}
      else
	for (; ic; ic = ic->next)
	  delete_insn (ic->insn);
    }
  return newinsn;
}

static void
init_block_vector (vector, space, nelts, bytes_per_elt)
     int  **vector;
     int  *space;
     int nelts;
     int bytes_per_elt;
{
  register int i;
  register int *p = space;

  for (i = 0; i < nelts; i++)
    {
      vector[i] = p;
      p += bytes_per_elt / sizeof (*p);
    }
}

static void
find_block_flow_info (f, basic_block_predecessors, basic_block_successors, nonlocal_label_list)
     rtx f, nonlocal_label_list;
     int **basic_block_predecessors, **basic_block_successors;
{
  rtx insn;
  RTX_CODE code;
  int i;

  /* Most of the work has already been done by find_basic_blocks_for_vector_flow.
     basic_block_drops_in handles nonbranching flow */
  for (i=0; i < n_basic_blocks; i++)
    {
      if (basic_block_drops_in[i])
	{
	  basic_block_predecessors[i][i-1] = 1;
	  basic_block_successors[i-1][i] = 1;
	}
      insn = basic_block_end [i];
      code = GET_CODE (insn);
      while (code == NOTE)
	{
	  insn = PREV_INSN (insn);
	  code = GET_CODE (insn);
	}
      if (code == JUMP_INSN)
	{
	  int target;
	  if (JUMP_LABEL (insn))
	    {
	      target = BLOCK_NUM (JUMP_LABEL (insn));
	      basic_block_successors[i][target] = 1;
	      basic_block_predecessors[target][i] = 1;
	    }
	  else if (GET_CODE (PATTERN (insn)) == ADDR_VEC
		   || GET_CODE (PATTERN (insn)) == ADDR_DIFF_VEC)
	    {
	      int elnum;
	      rtx pat = PATTERN (insn);
	      int diff_vec_p = GET_CODE (pat) == ADDR_DIFF_VEC;
	      int len = XVECLEN (pat, diff_vec_p);

	      for (elnum = 0; elnum < len; elnum++)
		if (target = XINT (XEXP (XVECEXP (pat, diff_vec_p, elnum), 0), 0))
		  {
		    target = uid_block_number[target];
		    basic_block_successors[i][target] = 1;
		    basic_block_predecessors[target][i] = 1;
		  }
	    }
	}
      else
	if (code == CALL_INSN)
	  {
	    rtx x;
	    for (x = nonlocal_label_list; x; x = XEXP (x, 1))
	      {
		int target;
		target = BLOCK_NUM (x);
		basic_block_successors[i][target] = 1;
		basic_block_predecessors[target][i] = 1;
	      }
	  }
    }
}

static void
find_vector_flow_info (f, basic_block_predecessors, basic_block_successors)
     rtx f;
     int **basic_block_predecessors, **basic_block_successors;
{
  rtx insn;
  RTX_CODE code;
  int i,j, changes = 1;

  /* set initial GEN and OUT values for each block */
  for (i=0; i < n_basic_blocks; i++)
    {
      if ( (insn = basic_block_head[i]) == basic_block_end[i])
	/* we have a single insn block */
	{
	  if (GET_CODE (insn) == CALL_INSN)
	    basic_block_vector_out_length[i] = BOTTOM;
	  else
	    {
	      basic_block_vector_out_length[i] = rtx_vlength (insn);
	      basic_block_vector_gen_length[i] = rtx_vlength (insn);
	    }
	  continue;
	}
      /* Now start at the beginning of the multi-insn block to find
	 the first generated vector length */
      for (insn = basic_block_head[i]; insn != next_nonnote_insn (basic_block_end[i]); insn = next_nonnote_insn (insn))
	if (rtx_vlength (insn))
	  {
	    basic_block_vector_gen_length[i] = rtx_vlength (insn);
	    break;  /* since we want only the first value */
	  }
      if (insn == next_nonnote_insn (basic_block_end[i]))
	{
	  insn = basic_block_end[i];
	  while (insn && GET_CODE (insn) == NOTE)
	    insn = PREV_INSN (insn);
	  if (GET_CODE (insn) == CALL_INSN)
	    basic_block_vector_out_length[i] = BOTTOM;
	  continue;
	}
	  
      /* Now start at the end of the multi-insn block to find
	 the last generated vector length */
      for (insn = basic_block_end[i]; insn != prev_nonnote_insn (basic_block_head[i]); insn = prev_nonnote_insn (insn))
	{
	  if (rtx_vlength (insn))
	    {
	      basic_block_vector_out_length[i] = rtx_vlength (insn);
	      break;
	    }
	  if (GET_CODE (insn) == CALL_INSN)
	    {
	      basic_block_vector_out_length[i] = BOTTOM;
	      break;
	    }
	}
    }
  /* Now iterate to a fixed point */
  while (changes)
    {
      changes = 0;
      for (i=0; i < n_basic_blocks; i++)
	{
	  for (j=0; j < n_basic_blocks; j++)
	    {
	      if (basic_block_predecessors[i][j])
		if (basic_block_vector_out_length[j])
		  if (! basic_block_vector_in_length[i])
		    {
		      basic_block_vector_in_length[i] = basic_block_vector_out_length[j];
		      changes = 1;
		    }
		  else
		    if (basic_block_vector_in_length[i] != basic_block_vector_out_length[j])
		      if (basic_block_vector_in_length[i] != BOTTOM)
			{
			  basic_block_vector_in_length[i] = BOTTOM;
			  changes = 1;
			}
	    }
	  if (! basic_block_vector_gen_length[i])
	    if (basic_block_vector_in_length[i] != basic_block_vector_out_length[i])
	      {
		basic_block_vector_out_length[i] = basic_block_vector_in_length[i];
		changes = 1;
	      }
	}
    }
}

static void
print_vector_flow (file)
     FILE *file;
{
  int i,j;

  for (i=0; i < n_basic_blocks; i++)
    {
      fprintf (file, "\n;;Basic Block %d:  insn %d to insn %d\n",
	       i,INSN_UID(basic_block_head[i]),INSN_UID(basic_block_end[i]));
      fprintf (file, "Predecessors:\t");
      for (j=0;j < n_basic_blocks;j++)
	if (basic_block_predecessors[i][j])
	  fprintf (file, " %d", j);
      fprintf (file, "\n");
      fprintf (file, "Successors:\t");
      for (j=0;j < n_basic_blocks;j++)
	if (basic_block_successors[i][j])
	  fprintf (file, " %d", j);
      fprintf (file, "\n");
      fprintf (file, "In Vector Length:\t%d\n", basic_block_vector_in_length[i]);
      fprintf (file, "Gen Vector Length:\t%d\n", basic_block_vector_gen_length[i]);
      fprintf (file, "Out Vector Length:\t%d\n", basic_block_vector_out_length[i]);
    }
}

static void
update_vector_regs (pattern, newregs)
     rtx pattern, newregs[];
{
  rtx src0, src1, src2, src, *loc0, *loc1, *loc2, dest, dest0;

  src0 = src1 = src2 = dest0 = NULL_RTX;
  loc0 = loc1 = loc2 = NULL;
  if (GET_CODE (pattern) == CLOBBER)
    {
      if (GET_CODE (XEXP (pattern, 0)) == REG
	  && rtx_vector_p (XEXP (pattern, 0)))
	XEXP (pattern, 0) = newregs [REGNO (XEXP (pattern, 0))];
      return;
    }
  if (GET_CODE (pattern) == MEM)
    {
      if (GET_CODE (XEXP (pattern, 0)) == REG
	  && rtx_vector_p (XEXP (pattern, 0)))
	XEXP (pattern, 0) = newregs [REGNO (XEXP (pattern, 0))];
      else if (GET_CODE (XEXP (pattern, 0)) == PLUS)
	{
	  loc0 = &XEXP (XEXP (pattern, 0), 0);
	  loc1 = &XEXP (XEXP (pattern, 0), 1);
	  if (GET_CODE (*loc0) == REG
	      && rtx_vector_p (*loc0))
	    *loc0 = newregs [REGNO (*loc0)];
	  if (GET_CODE (*loc1) == REG
	      && rtx_vector_p (*loc1))
	    *loc1 = newregs [REGNO (*loc1)];
	}
      return;
    }
  src = SET_SRC (pattern);
  switch (GET_RTX_CLASS (GET_CODE (src)))
    {
    case 'o': if (GET_CODE (src) == REG
		  && newregs [REGNO (src)])
                SET_SRC (pattern) = newregs [REGNO (src)];
              else if (GET_CODE (src) == MEM)
		if (GET_CODE (XEXP (src, 0)) == REG
		    && rtx_vector_p (XEXP (src, 0)))
		  XEXP (src, 0) = newregs [REGNO (XEXP (src, 0))];
		else if (GET_CODE (XEXP (src, 0)) == PLUS)
		  {
		    loc0 = &XEXP (XEXP (src, 0), 0);
		    loc1 = &XEXP (XEXP (src, 0), 1);
		    if (GET_CODE (*loc0) == REG
			&& rtx_vector_p (*loc0))
		      *loc0 = newregs [REGNO (*loc0)];
		    if (GET_CODE (*loc1) == REG
			&& rtx_vector_p (*loc1))
		      *loc1 = newregs [REGNO (*loc1)];
		  }
              break;
    case '1': src0 = XEXP (src, 0);
              break;
    case '<':
    case '2':
    case 'c': src0 = XEXP (src, 0);
              loc0 = &XEXP (src, 0);
              while (GET_CODE (src0) == SUBREG
		     || GET_CODE (src0) == ZERO_EXTRACT
		     || GET_CODE (src0) == SIGN_EXTRACT)
		{
		  loc0 = &XEXP (src0, 0);;
		  src0 = XEXP (src0, 0);
		}
              src1 = XEXP (src, 1);
              loc1 = &XEXP (src, 1);
              while (GET_CODE (src1) == SUBREG
		     || GET_CODE (src1) == ZERO_EXTRACT
		     || GET_CODE (src1) == SIGN_EXTRACT)
		{
		  loc1 = &XEXP (src1, 0);
		  src1 = XEXP (src1, 0);
		}
              break;
    case '3': src0 = XEXP (src, 0);
              loc0 = &XEXP (src, 0);
              while (GET_CODE (src0) == SUBREG
		     || GET_CODE (src0) == ZERO_EXTRACT
		     || GET_CODE (src0) == SIGN_EXTRACT)
		{
		  loc0 = &XEXP (src0, 0);;
		  src0 = XEXP (src0, 0);
		}
              src1 = XEXP (src, 1);
              loc1 = &XEXP (src, 1);
              while (GET_CODE (src1) == SUBREG
		     || GET_CODE (src1) == ZERO_EXTRACT
		     || GET_CODE (src1) == SIGN_EXTRACT)
		{
		  loc1 = &XEXP (src1, 0);
		  src1 = XEXP (src1, 0);
		}
              src2 = XEXP (src, 2);
              loc2 = &XEXP (src, 2);
              while (GET_CODE (src2) == SUBREG
		     || GET_CODE (src2) == ZERO_EXTRACT
		     || GET_CODE (src2) == SIGN_EXTRACT)
		{
		  loc2 = &XEXP (src2, 0);
		  src2 = XEXP (src2, 0);
		}
              break;
    case 'x': if (GET_CODE (src) == ASM_OPERANDS)
                /* then do everything for the source here */
                {
		  int length = ASM_OPERANDS_INPUT_LENGTH(src);
		  int i;
		  		  
		  for (i=0; i<length; i++)
		    if (GET_CODE (ASM_OPERANDS_INPUT(src, i)) == REG
			&& rtx_vector_p (ASM_OPERANDS_INPUT(src, i))
			&& newregs [REGNO (ASM_OPERANDS_INPUT(src, i))])
		      ASM_OPERANDS_INPUT(src, i)
			= newregs [REGNO (ASM_OPERANDS_INPUT(src, i))];
		}
              /* no break in case one of the x's may fit the following */
    case 'b': if (GET_CODE (src) == SUBREG
		  || GET_CODE (src) == ZERO_EXTRACT
		  || GET_CODE (src) == SIGN_EXTRACT)
                {
		  src0 = XEXP (src, 0);
		  while (GET_CODE (src0) == SUBREG
			 || GET_CODE (src0) == ZERO_EXTRACT
			 || GET_CODE (src0) == SIGN_EXTRACT)
		    {
		      src = src0;
		      src0 = XEXP (src, 0);
		    }
		}
              break;
    default:  abort ();
    }
  if (src0
      && GET_CODE (src0) == REG
      && newregs [REGNO (src0)])
    if (loc0)
      *loc0 = newregs [REGNO (src0)];
    else
      XEXP (src, 0) = newregs [REGNO (src0)];
  if (src1
      && GET_CODE (src1) == REG
      && newregs [REGNO (src1)])
    if (loc1)
      *loc1 = newregs [REGNO (src1)];
    else
      XEXP (src, 1) = newregs [REGNO (src1)];
  if (src2
      && GET_CODE (src2) == REG
      && newregs [REGNO (src2)])
    if (loc2)
      *loc2 = newregs [REGNO (src2)];
    else
      XEXP (src, 2) = newregs [REGNO (src2)];
  dest = SET_DEST (pattern);
  if (GET_CODE (dest) == REG
      && newregs [REGNO (dest)])
    SET_DEST (pattern) = newregs [REGNO (dest)];
  else if (GET_CODE (dest) == MEM)
    if (GET_CODE (XEXP (dest, 0)) == REG
	&& rtx_vector_p (XEXP (dest, 0)))
      XEXP (dest, 0) = newregs [REGNO (XEXP (dest, 0))];
    else if (GET_CODE (XEXP (dest, 0)) == PLUS)
      {
	loc0 = &XEXP (XEXP (dest, 0), 0);
	loc1 = &XEXP (XEXP (dest, 0), 1);
	if (GET_CODE (*loc0) == REG
	    && rtx_vector_p (*loc0))
	  *loc0 = newregs [REGNO (*loc0)];
	if (GET_CODE (*loc1) == REG
	    && rtx_vector_p (*loc1))
	  *loc1 = newregs [REGNO (*loc1)];
      }
    else if (GET_CODE (dest) == SUBREG
	     || GET_CODE (dest) == ZERO_EXTRACT
	     || GET_CODE (dest) == SIGN_EXTRACT
	     || GET_CODE (dest) == STRICT_LOW_PART)
    {
      dest0 = XEXP (dest, 0);
      while (GET_CODE (dest0) == SUBREG
	     || GET_CODE (dest0) == ZERO_EXTRACT
	     || GET_CODE (dest0) == SIGN_EXTRACT
	     || GET_CODE (dest0) == STRICT_LOW_PART)
	{
	  dest = dest0;
	  dest0 = XEXP (dest, 0);
	}
      if (GET_CODE (dest0) == REG
	  && newregs [REGNO (dest0)])
	XEXP (dest, 0) = newregs [REGNO (dest0)];
    }
}

static void
update_mem_refs (x, length)
     rtx x;
     int length;
{
  rtx memref;

  if (GET_CODE (x) == SET)
    if (GET_CODE (SET_DEST (x)) == MEM)
      {
	set_vector_int (length, VECTOR_STRIDE (SET_DEST (x)));
	memref = gen_rtx (MEM, GET_MODE (SET_DEST (x)),
				 XEXP (SET_DEST (x), 0));
	MEM_IN_STRUCT_P (memref) = MEM_IN_STRUCT_P (SET_DEST (x));
	SET_DEST (x) = memref;
      }
    else if (GET_CODE (SET_SRC (x)) == MEM)
      {
	set_vector_int (length, VECTOR_STRIDE (SET_SRC (x)));
	memref = gen_rtx (MEM, GET_MODE (SET_SRC (x)),
				 XEXP (SET_SRC (x), 0));
	MEM_IN_STRUCT_P (memref) = MEM_IN_STRUCT_P (SET_SRC (x));
	SET_SRC (x) = memref;
      }
}
